package clarion.acs.gs;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class GoalList implements GoalStructure {

  Global global;           // global settings.
  TaskClarion taskClarion; // task-specific clarion settings.
  int    netIdx;           // net index.
  int    curActiveGoalIdx; // current active goal index.
  short  goalReps[][];    // all of the possible goals.
  short  gsParamDVs[];     // each parameter dimension of a goal.
  int    size_gc;          // size of goal chunk.
  int    size_gs;          // size of goal structure.
  int    gsActionMadeMode; // GS action by top level only or by both.
  double threshold_GS;     // goal threshold.
  int    gsBlaOption;      // BLA option: constant or recency-based.
  double gsBla;            // BLA of GS.
  double gsBlaParams[];    // parameters of recency-based BLA.
  long   currentTime;      // current time for calculating BLA.
  double P_GS;             // coordination of external vs goal actions.
  double P_EXT;            // coordination of external vs goal actions.
  GoalChunk[] goalItems;      // recording current existing goals.
  StochasticDecider decider; // used for selecting active goal.

  public GoalList( Global g )
  {
      global = g;
      taskClarion = global.getTaskClarion();
      netIdx = global.getGsNetIdx();
      size_gc = global.getGoalChunkSize();
      size_gs = global.getGoalStructSize();
      gsActionMadeMode = global.getGoalDecisionMode();
      threshold_GS = global.getGoalActivationThreshold();
      gsBlaOption = global.getGoalBlaOption();
      if( gsBlaOption == global.CONSTANT_OPTION )
          gsBla = global.getGoalBla();
      else
          gsBlaParams = global.getGoalBlaParas();
      P_GS = global.getGoalCorConstant();
      P_EXT = global.getExternCorConstant();
      goalReps = taskClarion.getGoals();
      gsParamDVs = taskClarion.getGoalDims();
      goalItems = new GoalChunk[ size_gs ];
      /*
      for( int i = 0; i < size_gs; i++ )
      {
           short[][] goal = global.getGoals( i );
           Feature[] features = new Feature[ size_gc ];
           for( int j = 0; j < size_gc; j++ )
                features[i] = new Feature( goalFeatLabels[ j ], (short)global.FEAT_GS,
                                           (short)j, goal[ i ], global );
           goals[ i ] = new GoalChunk( global.CHK_STATE, goalChunkLabels[ i ],
                                       features,global);
           goals[ i ].setBlaParams( gsBlaParams );

      }
      */
      for( int i = 0; i < size_gs; i++ ){
           Feature[] features = new Feature[ size_gc + 1 ];
           for( int j = 0; j < size_gc + 1; j++ )
                if( j < size_gc ){    // parameter dimension.
                    short[] con = new short[ gsParamDVs[j] ];
                    for( int k = 0; k < con.length; k++ )
                         con[k] = 0;
                    con[goalReps[i][j]] = 1;   // goal parameter value in this dimension.
                    features[j] = new Feature( "goal param dim " + j, (short)Global.FEAT_INTERNAL,
                                               (short)(Global.FEAT_GOAL + j), con, global );
                }
                else{        // goal dimension.
                    short[] con = new short[ size_gs ];
                    for( int k = 0; k < size_gs; k++ )
                         con[k] = 0;
                    con[i] = 1;    // the ith goal.
                    features[j] = new Feature( "goal dim ", (short)global.FEAT_INTERNAL,
                                               (short)Global.FEAT_GOAL, con, global );
                }
           goalItems[ i ] = new GoalChunk( global.CHK_GOAL, "goal item " + i,
                                           features, global );
      }
      decider = new StochasticDecider( 0.1 );
      global.setGS( this );
  }

  public boolean isEmpty()
  {
     boolean emptyFlag = true;
     for( int i = 0; i < goalItems.length; i++ )
          if( !goalItems[i].isRemoved() ){
              emptyFlag = false;
              break;
          }
     return emptyFlag;
  }

  public int getSize()
  {
     return size_gs;
  }

  public int getDoNothingAction()
  {
     return 2 * size_gs;
  }

  public int getActiveGoal()
  {
     int currentGoal = getActiveItem().getFeature(size_gc).getValues()[0];
     return currentGoal;
  }

  public GoalChunk getActiveItem()
  {
      return goalItems[ curActiveGoalIdx ];
  }

  public double getThreshlod()
  {
     return threshold_GS;
  }

  public void setCurrentTime( long time )
  {
     currentTime = time;
  }

  public boolean actionSet( int idx )
  {
     GoalChunk gc = goalItems[ idx ];
     if( gc.isRemoved() )
         return false;
     if( !gc.isActivated() )
         gc.setActivated( true );
     gc.increaseUsage( currentTime );
     return true;
  }

  public boolean actionReset( int idx )
  {
     GoalChunk gc = goalItems[ idx ];
     gc.setRemoved( true );
     return true;
  }

  public boolean actionNoGoal()
  {
     return false;
  }

  public boolean goalCompete()
  {
     int activatedGoalNum = 0;
     int indices[] = new int[ size_gs ];
     double inputs[] = new double[ size_gs ];
     for( int i = 0; i < goalItems.length; i++ ){
          GoalChunk gc = goalItems[ i ];
          if( gc.isRemoved() || !gc.isActivated() )
              continue;
          gc.calcBLA( currentTime );
          double bla = gc.getBLA();
          if( bla > threshold_GS ){
              inputs[ activatedGoalNum ] = bla;
              indices[ activatedGoalNum++ ] = i;
          }
     }
     if( activatedGoalNum == 0 ){
         if( curActiveGoalIdx == global.EMPTY )
             return false;
         else
             return true;
     }
     int idx = decider.getRecommendation( inputs, activatedGoalNum );
     if( curActiveGoalIdx == indices[ idx ] )
         return false;
     else{
         curActiveGoalIdx = indices[ idx ];
         return true;
     }
  }

  public void goalFinished( boolean popFlag )
  {
     GoalChunk gc = goalItems[ curActiveGoalIdx ];
     gc.setRemoved( true );
  }

  public void updateState( double[] state )
  {
     goalItems[ curActiveGoalIdx ].getContent( state );
  }
}