package clarion.acs.gs;
import  java.util.Vector;
import  java.util.Stack;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class GoalStack implements GoalStructure {

  Global global;           // global settings.
  TaskClarion taskClarion; // task-specific clarion settings.
  int    netIdx;           // net index.
  int    curActiveGoalIdx; // current active goal index.
  short  goalReps[][];    // all of the possible goals.
  short  gsParamDVs[];     // each parameter dimension of a goal.
  int    size_gc;          // size of goal chunk, actually the # of parameter dims.
  int    size_gs;          // size of goal structure.
  int    gsActionMadeMode; // GS action by top level only or by both.
  double P_GS;             // coordination of external vs goal actions.
  double P_EXT;            // coordination of external vs goal actions.
  GoalChunk goalItems[];     // the goal items.
  Stack  stack;              // stack for goals.
  StochasticDecider decider; // used for selecting active goal.

  public GoalStack( Global g )
  {
      global = g;
      taskClarion = global.getTaskClarion();
      netIdx = global.getGsNetIdx();
      size_gc = global.getGoalChunkSize();  // actually the # of parameter dims.
      size_gs = global.getGoalStructSize();
      goalReps = taskClarion.getGoals();
      gsParamDVs = taskClarion.getGoalDims();
      gsActionMadeMode = global.getGoalDecisionMode();
      P_GS = global.getGoalCorConstant();
      P_EXT = global.getExternCorConstant();
      goalItems = new GoalChunk[ size_gs ];
      for( int i = 0; i < size_gs; i++ ){
           Feature[] features = new Feature[ size_gc + 1 ];
           for( int j = 0; j < size_gc + 1; j++ )
                if( j < size_gc ){    // parameter dimension.
                    short[] con = new short[ gsParamDVs[j] ];
                    for( int k = 0; k < con.length; k++ )
                         con[k] = 0;
                    con[goalReps[i][j]] = 1;   // goal parameter value in this dimension.
                    features[j] = new Feature( "goal param dim " + j, (short)Global.FEAT_INTERNAL,
                                               (short)(Global.FEAT_GOAL + j), con, global );
                }
                else{        // goal dimension.
                    short[] con = new short[ size_gs ];
                    for( int k = 0; k < size_gs; k++ )
                         con[k] = 0;
                    con[i] = 1;    // the ith goal.
                    features[j] = new Feature( "goal dim ", (short)global.FEAT_INTERNAL,
                                               (short)Global.FEAT_GOAL, con, global );
                }
           goalItems[ i ] = new GoalChunk( global.CHK_GOAL, "goal item " + i,
                                           features, global );
      }
      stack = new Stack();
      //stack.setSize( size_gs );
      decider = new StochasticDecider( 0.1 );
      global.setGS( this );
  }

  public boolean isEmpty()
  {
     return stack.empty();
  }

  public int getSize()
  {
     return size_gs;
  }

  public int getDoNothingAction()
  {
     return 2 * size_gs;
  }

  public int getActiveGoal()
  {
     if( stack.empty() )
         return -1;
     int currentGoal = getActiveItem().getFeature(size_gc).getValues()[0];
     return currentGoal;
  }

  public GoalChunk getActiveItem()
  {
     if( stack.empty() )
         return null;
     int idx = ((Integer)stack.peek()).intValue();
     return goalItems[idx];
  }

  // push action
  public boolean actionSet( int idx )
  {
     if( !stack.empty() && curActiveGoalIdx == idx )
         return false;
     curActiveGoalIdx = idx;
     stack.push( new Integer( curActiveGoalIdx ) );
     return true;
  }

  // pop action.
  public boolean actionReset( int idx )
  {
     if( stack.empty() ){
         curActiveGoalIdx = Global.EMPTY;
         return false;
     }
     curActiveGoalIdx = ((Integer)( stack.pop() )).intValue();
     return true;
  }

  // no-goal-action.
  public boolean actionNoGoal()
  {
     return false;
  }

  // when a goal is completed, no automatic goal pop.
  public void goalFinished( boolean popFlag )
  {
     if( popFlag )
         actionReset( 0 );
  }

  public void updateState( double[] state )
  {
     goalItems[ curActiveGoalIdx ].getContent( state );
  }

}