package clarion.acs.net;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;
import clarion.coordinate.*;

public class AcsNet {

  // global settings.
  protected Global global;
  // the user-defined CLARION settings.
  protected TaskClarion taskClarion;
  // the net indice of ACS.
  protected int netIdx;
  //  ACS # of knowledge types.
  protected int COMP_NUM;
  // current chosen component
  protected int chosenComp;
  // current chosen action, actually is an index of an action in the action set.
  // the actual action is a dim-value pairs.
  protected short[] chosenAction;
  protected int outputDimNum, outputNum;
  protected int inputNum, exInputNum, gsInputNum, wmInputNum;
  // current states. state = EX + GS + WM.
  protected double[] state, exState, gsState, wmState;
   // current sensory input.
  protected Feature sensoryInput[];
  // variables for components.
  protected boolean    netCompFlags[];
  protected AcsNetComp netComps[];
  protected double[][] outputActivations;    // action strengths of all of the components.
  protected double     reinforcement;
  // variables for combining components.
  protected int  combMode1;      // fix or variable.
  protected int  combMode2;      // stochastic or percentage.
  protected double[] succRates;  // success rate of each component, for variable mode.
  protected double[] c3s, c4s;   // used for variable mode.
  protected StochasticComb stochComb;
  protected WeightSumComb wsComb;
  protected CorrectionComb ctComb;

  public AcsNet( int netIdx, Global global )
  {
     this.netIdx = netIdx;
     this.global = global;
     taskClarion = global.getTaskClarion();
     COMP_NUM = global.getAcsKnowledgeNum();
     netCompFlags = new boolean[ COMP_NUM ];
     netComps = new AcsNetComp[ COMP_NUM ];
     // initialize states.
     if( global.isExNet(netIdx) )
         exInputNum = global.getIdnInputNum( netIdx );
     else
         exInputNum = global.getSensoryInputLength();
     exState = new double[ exInputNum ];
     if( global.getGsOption() != Global.GS_NON ){
         int gsNetIdx = global.getGsNetIdx();
         //gsInputNum = global.getIdnInputNum( gsNetIdx );
         gsInputNum = global.getGsInputLength();
         gsState = new double[ gsInputNum ];
     }
     if( global.isWmOn() ){
         int wmNetIdx = global.getWmNetIdx();
         //wmInputNum = global.getIdnInputNum( wmNetIdx );
         wmInputNum = global.getWmInputLength();
         wmState = new double[ wmInputNum ];
     }
     inputNum = exInputNum + gsInputNum + wmInputNum;
     state = new double[ inputNum ];
     // initialize IDN.
     netCompFlags[ global.BL ] = global.isIdnOn( netIdx );
     if( netCompFlags[global.BL] )
         netComps[ global.BL ] = new IDN( inputNum, netIdx, global );
     // initialize ARS.
     netCompFlags[global.RER] = global.isRerOn( netIdx );
     if( netCompFlags[global.RER] )
         netComps[ global.RER ] = new RerRuleSet( netIdx, global );
     netCompFlags[global.IRL] = global.isIrlOn( netIdx );
     if( netCompFlags[global.IRL] )
         netComps[ global.IRL ] = new IrlRuleSet( netIdx, global );
     netCompFlags[global.FR] = global.isFrOn( netIdx );
     if( netCompFlags[global.FR] )
         netComps[ global.FR ] = new FixRuleSet( netIdx, global );
     outputDimNum = global.getIdnOutputDimNum( netIdx );
     // initialize action activations  of each component.
     outputNum = global.getIdnOutputNum( netIdx );
     outputActivations = new double[ COMP_NUM ][ outputNum ];
     // initialize variables for combining components.
     combMode1 = global.getAcsCrossCombineMode(netIdx, 0 );
     combMode2 = global.getAcsCrossCombineMode(netIdx, 1 );
     succRates = new double[ COMP_NUM ];
     chosenAction = new short[ outputDimNum ];
     double[] paras = global.getAcsCrossCombineParas( netIdx );
     if( combMode1 == global.STOCH_COMB_MODE ){
         stochComb = new StochasticComb( combMode2, COMP_NUM, paras, global );
         if( combMode2 == global.VAR_COMB_MODE ){
             c3s = global.getAcsCrossCombineVarParaAbove(netIdx);
             c4s = global.getAcsCrossCombineVarParaUnder(netIdx);
         }
     }
     else{ // bottom-up or top-guide.
         if( combMode2 < global.CT_COMB_MODE )
             wsComb = new WeightSumComb( combMode2, outputNum, paras,
                          global.getOverallActionSelectionTemp(netIdx) );
         else
             ctComb = new CorrectionComb();
     }
  }

  public void getExState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          arr[i] = exState[i];
  }

  public void getGsState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          arr[i] = gsState[i];
  }

  public void getWmState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          arr[i] = wmState[i];
  }

  public String[] getRerRules()
  {
     if( netComps[global.RER] != null )
         return ((RerRuleSet)netComps[ global.RER ]).getRules();
     else
         return (String[])null;
  }

  public String[] getIrlRules()
  {
     if( netComps[global.IRL] != null )
         return ((IrlRuleSet)netComps[ global.IRL ]).getRules();
     else
        return (String[])null;
  }

  public String[] getFixRules()
  {
     if( netComps[global.FR] != null )
         return ((FixRuleSet)netComps[ global.FR ]).getRules();
     else
         return (String[])null;
  }

  public double getPM( int idx )
  {
     if( netCompFlags[ idx ] )
         return netComps[ idx ].getPM();
     else
         return 0.0;
  }

  public double getNM( int idx )
  {
     if( netCompFlags[ idx ] )
         return netComps[ idx ].getNM();
     else
         return 0.0;
  }

  public short[] getChosenAction()
  {
     return chosenAction;
  }

  public Chunk[][] getAllConclusions()
  {
     Chunk[][] chunks = new Chunk[COMP_NUM][];
     for( int i = 0; i < COMP_NUM; i++ ){
          if( netCompFlags[i] )
              chunks[i] = netComps[i].getAllConclusions();
     }
     return chunks;
  }

  public double[][] getActionActivations()
  {
     return outputActivations;
  }

  // no NACS RT involved.
  public int getRT()
  {
     int RT = 0;
     int RT_BL = 0;
     int RT_TL = 0;
     if( combMode1 == Global.STOCH_COMB_MODE ){
         RT = netComps[ chosenComp ].getResponseTime( 0 );
         if( chosenComp == Global.BL )
             RT_BL = RT;
         else
             RT_TL = RT;
     }
     else{
        for( int i = 1; i < COMP_NUM; i++ )
             if( netCompFlags[i] )
                 RT_TL = Math.max(RT_TL, netComps[i].getResponseTime( 0 ));
        RT_BL = netComps[Global.BL].getResponseTime( 0 );
     }
     return taskClarion.getOverallRT( RT_BL, RT_TL );
  }

  // NACS RT involved as TLAT.
  public int getRT( int nacsRT )
  {
     int RT = 0;
     int RT_BL = 0;
     int RT_TL = 0;
     if( combMode1 == Global.STOCH_COMB_MODE ){
         RT = netComps[ chosenComp ].getResponseTime( nacsRT );
         if( chosenComp == Global.BL )
             RT_BL = RT;
         else
             RT_TL = RT;
     }
     else{
        for( int i = 1; i < COMP_NUM; i++ )
             if( netCompFlags[i] )
                 RT_TL = Math.max(RT_TL, netComps[i].getResponseTime( nacsRT ));
        RT_BL = netComps[Global.BL].getResponseTime( nacsRT );
     }
     return taskClarion.getOverallRT( RT_BL, RT_TL );
  }

  // maybe revised further: divide the chunk into EX, GS and WM states if exists.
  public void setStates( Chunk chunk )
  {
     int dimNum = chunk.getDimNum();
     int offset = 0;
     for( int i = 0; i < dimNum; i++ ) {
          short[] arr = chunk.getFeature(i).getContent();
          for( int j = 0; j < arr.length; j++ )
               state[ offset + j ] = arr[ j ];
          offset += arr.length;
     }
  }

  public void setSensoryInput( Feature[] sensoryInput )
  {
     this.sensoryInput = sensoryInput;
  }

  public void setGsState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          gsState[i] = arr[i];
  }

  public void setWmState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          wmState[i] = arr[i];
  }

  public void setCurDesiredOutput( Feature[] desiredOutput )
  {
     ((IDN)netComps[ Global.BL ]).setCurDesiredOutput( desiredOutput );
  }

  public void setCurrentTime( long curTime )
  {
     for( int i = 0; i < netCompFlags.length; i++ )
          if( netCompFlags[ i ] )
              netComps[ i ].setCurrentTime( (int)curTime );
  }

  public void setReinforcement( double reward )
  {
     reinforcement = reward;
  }

  // this means stochastic combination used( used to be the percentage combination )
  public void decideAction( int compIdx )
  {
     // decide an action.
     chosenComp = compIdx;
     if( netCompFlags[ chosenComp ] )
         chosenAction = netComps[ chosenComp ].decideAction();
     else{
         for( int i = 0; i < chosenAction.length; i++ )
              chosenAction[i] = Global.EMPTY;
     }
     if( !global.isOverrideByIdn() )  // added on 06/23/03
         return;
     // the chosen action is empty and IDN is on, then let IDN determine the action.
     if( chosenAction[0] == Global.EMPTY && netCompFlags[global.BL] ){
         chosenComp = global.BL;
         chosenAction = netComps[ chosenComp ].decideAction();
     }
     // the chosen action is still empty, randomly select an action.
     if( chosenAction[0] == Global.EMPTY ){
         chosenAction = new short[ outputDimNum ];
         for( int i = 0; i < outputDimNum; i++ )
              chosenAction[i] = (short)( global.getIdnOutputDimVals(netIdx)[i] * Math.random() );
     }
  }

  // reasoning for determining an action to perform given current input.
  public void reasoning()
  {
     // get active net comps.
     netCompFlags[Global.BL] = global.isIdnOn( netIdx );
     netCompFlags[Global.RER] = global.isRerOn( netIdx );
     netCompFlags[Global.IRL] = global.isIrlOn( netIdx );
     netCompFlags[Global.FR] = global.isFrOn( netIdx );
     for( int i = 0; i < COMP_NUM; i++ ){
          if( netCompFlags[ i ] ){
              netComps[ i ].setStates( state, exState, gsState, wmState );
              netComps[ i ].reasoning();
          }
          if( combMode1 > Global.STOCH_COMB_MODE && combMode2 < Global.CT_COMB_MODE ){
              // if it is weight-sum combination.
              if( netCompFlags[ i ] )
                  netComps[ i ].getOutputActivations( outputActivations[ i ] );
              else
                  for( int j = 0; j < outputNum; j++ )
                       outputActivations[ i ][ j ] = 0.0;
          }
     }
     // combine the results.
     if( combMode1 == Global.STOCH_COMB_MODE )
         stochCombine();
     else{ // bottom-up verification or top-guide
         if( combMode2 < Global.CT_COMB_MODE )
             weightSumCombine();
         else
             correctionCombine();
     }
  }

  // need be overrided by GS, WM net.
  public void perform()
  {
  }

  // update each component.
  public void update( double reward )
  {
     reinforcement = reward;
     // update IDN component.
     if( global.getIdnLearningOption( netIdx ) == global.QL )
         netComps[ global.BL ].setStates( state, exState, gsState, wmState );
     netComps[ global.BL ].update( reinforcement );
     // update other components.
     for( int i = 1; i < COMP_NUM; i++ ){
         if( netCompFlags[ i ] ){
             double threshold = 0.0;
             switch(i) {
               case Global.RER:
                    threshold = global.getRerPosThreshold( netIdx );
                    break;
               case Global.IRL:
                    threshold = global.getIrlPosThreshold( netIdx );
                    break;
             }
             boolean posFlag = ((IDN)netComps[ global.BL ]).getPositivity( i, threshold );
             netComps[ i ].update( posFlag );
         }
     }
  }

  // form current state including EX, GS, WM if exists.
  public void formState()
  {
     // form externl sensory input.
     if( global.isExNet( netIdx ) ){
         short exIdnInputDVs[];
         short exIdnInputFeatureInfo[][];
         exIdnInputDVs = global.getIdnInputDimVals( netIdx );
         exIdnInputFeatureInfo = global.getIdnInputFeatureInfo( netIdx );
         for( int i = 0; i < sensoryInput.length; i++ ){
              int idx = searchFeature( sensoryInput[ i ].getType(), sensoryInput[ i ].getOrder(),
                                       exIdnInputFeatureInfo );
              if( idx != -1 ){ // the feature exists in the input to this net.
                  int offset = 0;
                  for( int j = 0; j < idx; j++ )
                       offset += exIdnInputDVs[ j ];
                  short[] content = sensoryInput[ i ].getContent();
                  for( int j = 0; j < exIdnInputDVs[ idx ]; j++ )
                       exState[ offset + j ] = content[ j ];
              }
              else
                  System.out.println( "not an eligible dim in AcsNet.java!" );
         }
     }
     else{
         int offset = 0;
         for( int i = 0; i < sensoryInput.length; i++ ){
              short[] content = sensoryInput[ i ].getContent();
              for( int j = 0; j < content.length; j++ )
                   exState[ offset + j ] = content[ j ];
              offset += content.length;
         }
     }
     // form state including EX, GS, WM.
     for( int i = 0; i < exInputNum; i++ )
          state[i] = exState[i];
     int offset = exInputNum;
     if( global.getGsOption() != Global.GS_NON ){
         for( int i = 0; i < gsInputNum; i++ )
              state[offset+i] = gsState[i];
         offset += gsInputNum;
     }
     if( global.isWmOn() ){
         for( int i = 0; i < wmInputNum; i++ )
              state[offset+i] = wmState[i];
     }
  }

  public void offlineTrainByEm()
  {
     //((IDN)netComps[global.BL]).offlineTrainByEm();
     int emIterNum = global.getEmTrainIterNum();
     NACS nacs = global.getNACS();
     EMAcsSample[] emSamples = nacs.getEmAcsSamples( netIdx, global.ACS );
     int emTrainSampleNum = emSamples.length;
     Vector sands = new Vector( emTrainSampleNum * emIterNum );
     for( int j = 0; j < emIterNum; j++ )
       for( int k = 0; k < emTrainSampleNum; k++ )
            sands.addElement( new Integer(k) );
     while( !sands.isEmpty() ){
       // decompose the step chunk.
       int idx = ((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
       Chunk curState = emSamples[idx].getCurState();
       Chunk curAction = emSamples[idx].getCurAction();
       Chunk nextState = emSamples[idx].getNextState();
       double reward = emSamples[idx].getReward();
       // convert the state chunk into input to each component of the net.
       setStates( curState );
       // reasoning.
       reasoning();
       // deliver chosen action.
       for( int i = 0; i < outputDimNum; i++ )
            chosenAction[i] = curAction.getFeature(i).getValues()[0];
       deliverAction();
       // convert the nextState chunk into input to each component of the net.
       if( nextState != null )
           setStates( nextState );
       // update each component
       update( reward );
    }
  }

  public void offlineTrainByAem()
  {
     //((IDN)netComps[global.BL]).offlineTrainByAem();
     int aemIterNum = global.getAemTrainAcsIterNum();
     NACS nacs = global.getNACS();
     AbsEMSample[] aemSamples = nacs.getAemSamples( netIdx );
     int aemTrainSampleNum = aemSamples.length;
     Vector sands = new Vector( aemTrainSampleNum * aemIterNum );
     for( int j = 0; j < aemIterNum; j++ )
       for( int k = 0; k < aemTrainSampleNum; k++ )
            sands.addElement( new Integer(k) );
     while( !sands.isEmpty() ){
       int idx = ((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
       Chunk curState = aemSamples[idx].getCurState();
       Chunk curAction = aemSamples[idx].getCurAction();
       Chunk nextState = aemSamples[idx].getNextState();
       double reward = aemSamples[idx].getReward();
       double[] stateDistribution = aemSamples[idx].getStateDistributions();
       Chunk[] nextStates = aemSamples[idx].getPotentialNextStates();
       // convert the state chunk into input to each component of the net.
       setStates( curState );
       // reasoning.
       reasoning();
       // deliver chosen action.
       for( int i = 0; i < outputDimNum; i++ )
            chosenAction[i] = curAction.getFeature(i).getValues()[0];
       deliverAction();
       // convert the nextState chunk into input to each component of the net.
       if( nextState != null )
           setStates( nextState );
       // update each component
       update( reward, nextStates, stateDistribution );
     }
  }

  // deliver currently chosen action to each component.
  protected void deliverAction()
  {
     for( int i = 0; i < COMP_NUM; i++ ){
          if( netCompFlags[ i ] )
              netComps[ i ].setChosenAction( chosenAction );
     }
  }

/************************** private methods ***********************************/
  private int searchFeature( int type,  int order, short[][] info )
  {
      int idx = -1;
      for( int i = 0; i < info.length; i++ ){
           if( info[i][0] == type && info[i][1] == order ){
               idx = i;
               break;
           }
      }
      return idx;
  }

  private void stochCombine()
  {
      for( int i = 0; i < chosenAction.length; i++ )
           chosenAction[i] = Global.EMPTY;
      // percentage combination, only one set of measuses needed.
      if( combMode2 == global.VAR_COMB_MODE ) {
          double[] PMs = new double[ COMP_NUM ];
          double[] NMs = new double[ COMP_NUM ];
          for( int i = 0; i < COMP_NUM; i++ ) {
               PMs[ i ] = getPM( i );
               NMs[ i ] = getNM( i );
               succRates[i] = (c3s[i] + PMs[i]) / (c4s[i] + PMs[i] + NMs[i]);
          }
          stochComb.setSuccRate( succRates );
          stochComb.calcProbabilities();
      }
      // select which part: BL, RER, IRL or FR to decide action.
      chosenComp = stochComb.selectComponent();
      // decide an action.
      decideAction( chosenComp );
      // set the chosen action.
      deliverAction();
   }

   private void weightSumCombine()
   {
      for( int i = 0; i < chosenAction.length; i++ )
             chosenAction[i] = Global.EMPTY;
      if( combMode2 == global.VAR_COMB_MODE ){
          // calculate PM and NM.
          double[] PMs = new double[ COMP_NUM ];
          double[] NMs = new double[ COMP_NUM ];
          for( int i = 0; i < COMP_NUM; i++ ) {
               PMs[ i ] = getPM( i );
               NMs[ i ] = getNM( i );
               succRates[i] = (c3s[i] + PMs[i])/(c4s[i] + PMs[i] + NMs[i]);
          }
          wsComb.setSuccRate( succRates );
          wsComb.calcWeights();
      }
      // decide action.
      wsComb.setActivations( outputActivations );
      wsComb.combination( global.getIdnOutputDimVals(netIdx), chosenAction );
      // set the chosen action.
      deliverAction();
  }

  // the correction combination method should be overrided by user's code.
  private void correctionCombine()
  {
     //ctComb.correctionComb();
     if( combMode1 == Global.BU_COMB_MODE )
         taskClarion.runBUCorrection();
     if( combMode1 == Global.TG_COMB_MODE )
         taskClarion.runTGCorrection();
  }

  // used for offline training using AEM.
  private void update( double reward, Chunk[] nextStates, double[] stateDistribution )
  {
     reinforcement = reward;
     // update IDN component.
     if( global.getIdnLearningOption( netIdx ) == global.QL )
         netComps[ global.BL ].setStates( state, exState, gsState, wmState );
     ((IDN)netComps[ global.BL ]).update( reinforcement, nextStates, stateDistribution );
     // update other components.
     for( int i = 1; i < COMP_NUM; i++ ){
         if( netCompFlags[ i ] ){
             double threshold = 0.0;
             switch(i) {
               case 1:
                    threshold = global.getRerPosThreshold( netIdx );
                    break;
               case 2:
                    threshold = global.getIrlPosThreshold( netIdx );
                    break;
               case 3:
                    threshold = reward;
                    //threshold = 0.1;
                    break;
             }
             boolean posFlag = ((IDN)netComps[ global.BL ]).getPositivity( i, threshold );
             netComps[ i ].update( posFlag );
         }
     }
  }
}