package clarion.acs.net;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;
import clarion.tools.*;

public class IDN implements AcsNetComp{

  // global settings.
  Global global;
  // task-specific CLARION settings.
  TaskClarion taskClarion;
  // the net indice.
  int netIdx;
  // learning option: q-learning, simplified q, BP, fix or by MCS.
  int learnOption;
  // current desired output.
  Feature desiredOutput[];
  // variables for the net
  BPNet net;
  int INPUT_NUM, OUTPUT_DIM_NUM;
  double[] state;  // current state including EX, GS, WM.
  short[] actionDVs;
  // used for MCS's minotoring conclusions.
  int c24;
  Chunk[] blResults;
  // current chosen action is a dim-value pairs.
  short[] chosenAction;
  short[] suggestedAction;
  // paramenters for rule positivity measurement: one Q(x,a) and one maxQ(y,b)
  double qValue1, qValue2;
  // Q(x,a)s and maxQ(y,b)s (with multiple dims)
  double qValue1s[], qValue2s[];
  // used for updating PM and NM.
  boolean positivity;
  // variables for action decision making.
  double temperature;
  StochasticDecider actionDecider;
  int currentTime;

  public IDN( int inputNum, int nIdx, Global g )
  {
     INPUT_NUM = inputNum;
     netIdx = nIdx;
     global = g;
     taskClarion = global.getTaskClarion();
     learnOption = global.getIdnLearningOption( netIdx );
     if( learnOption == Global.BP )
         net = new BPNet( inputNum, netIdx, global, Global.ACS );
     else
         net = new QBPNet( inputNum, netIdx, global );
     OUTPUT_DIM_NUM = global.getIdnOutputDimNum( netIdx );
     state = new double[ INPUT_NUM ];
     actionDVs = global.getIdnOutputDimVals( netIdx );
     chosenAction = new short[ OUTPUT_DIM_NUM ];
     suggestedAction = new short[ OUTPUT_DIM_NUM ];
     // initialize the conclusions for monitoring.
     if( global.isMcsOn() ){
         int c24 = global.getMonitorAcsBottomConcNum();
         blResults = new Chunk[ c24 ];
     }
     qValue1s = new double[ OUTPUT_DIM_NUM ];
     qValue2s = new double[ OUTPUT_DIM_NUM ];
     temperature = global.getIdnActionSelectionTemp( netIdx );
     actionDecider = new StochasticDecider( net.getOutput().length, temperature );
  }

  // need be replaced.
  public double[] getOutputActivations()
  {
     return net.getOutput();
  }

  // override if neccessary, use averaging Q-values to calculate activation of
  // each action since each action has multiple dims.
  public void getOutputActivations( double[] output )
  {
     getOutput( output );
  }

  public void getOutput( double[] output )
  {
     net.getOutput( output );
  }

  public double getSuccRate()
  {
     return net.getSuccRate();
  }

  public double getPM()
  {
     return net.getPM();
  }

  public double getNM()
  {
     return net.getNM();
  }

  public Chunk[] getAllConclusions()
  {
     Chunk[] chunks = new Chunk[1];
     int type = Global.CHK_ACTION;
     String label = Global.CHUNK_TYPE_LABELS[type];
     Feature[] featList = new Feature[ OUTPUT_DIM_NUM ];
     for( int i = 0; i < OUTPUT_DIM_NUM; i++ ){
          int typeIdx = global.getIdnOutputFeatureInfo( netIdx )[i][0];
          int featIdx = global.getIdnOutputFeatureInfo( netIdx )[i][1];
          String featLabel = global.getFeatureNames( typeIdx, featIdx );
          int featLen = global.getFeatureLength( typeIdx, featIdx );
          short[] values = new short[1];
          values[0] = suggestedAction[i];
          featList[i] = new Feature( featLabel, (short)typeIdx, (short)featIdx, (short)featLen,
                                     values, global );
     }
     chunks[0] = new Chunk( type, label, featList, global );
     return chunks;
  }

  public int getResponseTime( int BLAT )
  {
     calcRT( BLAT );
     return net.getResponseTime();
  }

  public boolean getPositivity( int ruleType, double threshold )
  {
     boolean posFlag = false;
     double reward = (net instanceof QBPNet)?((QBPNet)net).getReinforcement():0.0;
     double qDiscount = (net instanceof QBPNet)?((QBPNet)net).getQDiscount():0.0;
     if( ruleType == Global.RER )
         posFlag = taskClarion.getRerPositivity(learnOption,reward,qDiscount,qValue2,qValue1,threshold);
     if( ruleType == Global.IRL )
         posFlag = taskClarion.getIrlPositivity(learnOption,reward,qDiscount,qValue2,qValue1,threshold);
     if( ruleType == Global.FR )
         posFlag = taskClarion.getFrPositivity(learnOption,reward,qDiscount,qValue2,qValue1,threshold);
     return posFlag;
     /*
     if( learnOption == Global.QL ) {
         QBPNet qNet = (QBPNet)net;
         if( qNet.getReinforcement() + qNet.getQDiscount() * qValue2 - qValue1 > value )
             positivity = true;
         else
             positivity = false;
     }
     else if( learnOption == Global.SQL )
          {
              QBPNet qNet = (QBPNet)net;
              if( qNet.getReinforcement() > value )
                  positivity = true;
              else
                  positivity = false;
          }
          else {
              if( value > 0 )
                  positivity = true;
              else
                  positivity = false;
          }
     return positivity;
     */
  }

   public void setStates( double[] state, double[] exState, double[] gsState, double[] wmState )
   {
      for( int i = 0; i < INPUT_NUM; i++ )
           this.state[ i ] = state[ i ];
      if( learnOption != Global.QL )
          net.setInput( this.state );
      else
          ((QBPNet)net).setState( this.state );
   }

  public void setState( Chunk chunk )
  {
     int dimNum = chunk.getDimNum();
     int offset = 0;
     for( int i = 0; i < dimNum; i++ ) {
          short[] arr = chunk.getFeature(i).getContent();
          for( int j = 0; j < arr.length; j++ )
               state[ offset + j ] = arr[ j ];
          offset += arr.length;
     }
     if( learnOption != Global.QL )
         net.setInput( state );
     else
         ((QBPNet)net).setState( state );
  }

  public void setChosenAction( short[] action )
  {
      for( int i = 0; i < OUTPUT_DIM_NUM; i++ )
           chosenAction[i] = action[i];
  }

  // used for updating IDN with BP learning
  public void setCurDesiredOutput( Feature[] desiredOutput )
  {
     net.setDesiredOutput( desiredOutput );
  }

  public void setCurrentTime( int curTime )
  {
     currentTime = curTime;
  }

  public void reasoning()
  {
     net.forwardPass();
  }

  // this means IDN decide the action at current step.
  public short[] decideAction()
  {
     short[] action = actionDecider.getRecommendation( net.getOutput(), actionDVs );
     for( int i = 0; i < OUTPUT_DIM_NUM; i++ )
          suggestedAction[i] = action[i];
     return action;
  }

  public void update( double reward )
  {
     if( chosenAction[0] == Global.EMPTY )  // chosen action is empty.
         return;
     // qvalues of Q(x,a).
     if( learnOption != Global.BP )
         ((QBPNet)net).getQVals( chosenAction, qValue1s );
     // qvalues of maxQ(y,b)
     if( learnOption == Global.QL ){
         net.forwardPass();
         ((QBPNet)net).getMaxQVals( qValue2s );
     }
     // calcluate qvalue1 and qvalue2 used for positivity measurement.
     calcPositiveMeatureParas();
     // update BL network.
     switch( learnOption  ){
        case Global.QL:
             ((QBPNet)net).update( chosenAction, reward, qValue1s, qValue2s );
             break;
        case Global.SQL:
             ((QBPNet)net).update( chosenAction, reward, qValue1s );
             break;
        case Global.BP:
             net.backwardPass();
             break;
     }
  }

  // update for offline training using AEM.
  public void update( double reward, Chunk[] nextStates, double[] stateDistribution )
  {
     if( chosenAction[0] == Global.EMPTY )  // chosen action is empty.
         return;
     // qvalue of Q(x,a).
     if( learnOption != Global.BP )
         ((QBPNet)net).getQVals( chosenAction, qValue1s );
     // qvalue of maxQ(y,b)
     if( learnOption == Global.QL ){
         net.forwardPass();
         ((QBPNet)net).getMaxQVals( qValue2s );
     }
     // calcluate qvalue1 and qvalue2 used for positivity measurement.
     calcPositiveMeatureParas();
     // update QL using frequency estimates represented by the AEM.
     if( learnOption == Global.QL ){
           // qvalue of next potential states.
           double[][] qValues = new double[nextStates.length][OUTPUT_DIM_NUM];
           for( int i = 0; i < nextStates.length; i++ ) {
                // set current state.
                setState( nextStates[i] );
                net.forwardPass();
               ((QBPNet)net).getMaxQVals( qValues[i] );
           }
           for( int i = 0; i <  OUTPUT_DIM_NUM; i++ ){
                double value = 0.0;
                for( int j = 0; j < nextStates.length; j++ )
                     value += stateDistribution[j] * qValues[j][i];
                qValue2s[i] = value;
           }
          ((QBPNet)net).update( chosenAction, reward, qValue1s, qValue2s );
     }
     // update SQL.
     if( learnOption == Global.SQL )
         ((QBPNet)net).update( chosenAction, reward, qValue1s );
     // update BP.
     if( learnOption == Global.BP ){
         net.setDesiredOutput( chosenAction );
         net.backwardPass();
     }
  }

  public void update( boolean posFlag )
  {
  }

  public void calcRT( int BLAT )
  {
     net.calcRT( BLAT );
  }

/******************************* private methods ******************************/
  // overrided if neccessary, using averaging Q values to get qValue2( maxQ(y,b) )
  // and qValue1( Q(x,a) ). ( r + Qdiscount * maxQ(y,b) - Q(x,a) )
  private void calcPositiveMeatureParas()
  {
     qValue1 = qValue2 = 0.0;
     for( int i = 0; i < OUTPUT_DIM_NUM; i++ ){
          qValue1 += qValue1s[i];
          qValue2 += qValue2s[i];
     }
     qValue1 /= OUTPUT_DIM_NUM;
     qValue2 /= OUTPUT_DIM_NUM;
  }
}