package clarion.acs.net;
import clarion.acs.net.auxiliary.*;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.net.auxiliary.*;

public class IrlRule extends TLRule {

   final static double LOG_2 = Math.log( 2 );
   protected int groupId;     // decide which IRL subset is used.
   protected int ruleIdx;
   protected double c5, c6;
   // for deletion, generalization, specialization.
   protected double threshold4, threshold6, threshold7;
   protected IrlRuleForm ruleForm;

   public IrlRule( int groupIdx, int netIdx, int setIdx, int ruleIdx, Global global )
   {
       super( netIdx, Global.IRL, setIdx, global );
       this.groupId = groupIdx;
       this.ruleIdx = ruleIdx;
       c5 = global.getIrlIGParaAbove( netIdx );
       c6 = global.getIrlIGParaUnder( netIdx );
       threshold4 = global.getIrlDeleteThreshold( netIdx );
       threshold6 = global.getIrlGenThreshold( netIdx );
       threshold7 = global.getIrlSpecThreshold( netIdx );
       ruleForm = global.getTaskClarion().getIrlRule( groupIdx, netIdx, setIdx, ruleIdx );
   }

   public boolean isUseless()
   {
       double pm = attributes.getPM();
       double nm = attributes.getNM();
       /*
       double ig = Math.log((pm+c5)/(pm+nm+c6))/LOG_2;
       if( ig < threshold4 )
           return true;
       else
           return false;
       */
       //return taskClarion.getIrlIG( netIdx, pm, nm, threshold4 ) < 0.0;
       boolean flag = taskClarion.getIrlIG( netIdx, pm, nm, threshold4 ) < 0.0;
       //if( flag ){
       //    System.out.println( "match rate = " + (pm+1)/(pm+nm+2) );
       //}

       return flag;
   }

   public boolean decideAction( short[] state, short[] exState, short[] gsState, short[] wmState )
   {
       ruleSupport = ruleForm.determineAction( state, exState, gsState, wmState, suggestedAction );
       if( Math.abs( ruleSupport - 1.0 ) < 0.01 )
           return true;
       else
           return false;
   }

   public boolean canGeneralize()
   {
       double pm = attributes.getPM();
       double nm = attributes.getNM();
       return taskClarion.getIrlIG( netIdx, pm, nm, threshold6 ) >= 0.0;
   }

   public boolean canSpecialize()
   {
       double pm = attributes.getPM();
       double nm = attributes.getNM();
       return taskClarion.getIrlIG( netIdx, pm, nm, threshold7 ) < 0.0;
   }

   public void generalize()
   {
       ruleForm.generalize();
   }

   public void specialize()
   {
       ruleForm.specialize();
   }

/*************************** OVERRIDED METHODS *********************************/
   public String toString()
   {
       String str = "IRL RULE( " + netIdx + ", " + setIdx + ", " + ruleIdx + " ) = \n";
       double pm = attributes.getPM();
       str += "pm = " + String.valueOf( pm );
       double nm = attributes.getNM();
       str += ", nm = " + String.valueOf( nm );
       double rate = (pm+c5)/(pm+nm+c6);
       str += ", match rate = " + String.valueOf( rate );
       str += "\n";
       return str;
  }
}