package clarion.acs.net;
import java.util.Vector;
import java.util.TreeMap;
import clarion.acs.net.auxiliary.*;
import clarion.system.*;
import clarion.tools.*;

public class IrlRuleSet extends TLRuleSet{

   protected int groupId;          // decide which IRL subset is used.
   protected int curRuleSetIdx;
   protected int SET_NUM;          // # of rule sets.
   protected boolean useupFlag;    // check if all rules are used up.
   protected Vector activeRuleList;

   public IrlRuleSet( int netIdx, Global global )
   {
       super( netIdx, Global.IRL, global );
       groupId = global.getAgentGroupId();
       MAX_AGE = (int)( 1 / global.getIrlDensity( netIdx ) );
       SET_NUM = global.getIrlSetNum( netIdx );
       initNewSet();
   }

   public void update( boolean successFlag )
   {
      super.update( successFlag );
      // check usefulness of each rule.
      int idx = 0;
      while( idx < activeRuleList.size() ){
           int ruleIdx = ((Integer)activeRuleList.elementAt( idx )).intValue();
           IrlRule rule = (IrlRule)ruleList[ruleIdx];
           if( rule.isUseless() ){

               //System.out.println( "rule set = " + curRuleSetIdx + ", the deleted rule's pm " + rule.getPM()
               //                    + ", the deleted rule's nm " + rule.getNM() );

               ruleList[ruleIdx] = null;
               activeRuleList.remove( idx );
           }
           else{
               idx++;
               if( rule.canGeneralize() )
                   rule.generalize();
               if( rule.canSpecialize() )
                   rule.specialize();

           }
      }
      if( activeRuleList.isEmpty() ){
          if( (++curRuleSetIdx) < SET_NUM )
              initNewSet();
          else if( !useupFlag ){
                   useupFlag = true;
                   System.out.println( "NO IRL RULES!" );
                 }
      }
      // discount statistics.
      trialIdx++;
      for( int i = 0; i < activeRuleList.size(); i++ ){
           idx = ((Integer)activeRuleList.elementAt( i )).intValue();
           ruleList[idx].aging();
           if( trialIdx >= episodeLength )
               ruleList[idx].discountStatistics();
      }
      if( trialIdx >= episodeLength )
          trialIdx = 0;
   }

/*************************** PRIVATE METHODS **********************************/
   private void initNewSet()
   {
       //System.out.println( "net idx = " + netIdx + ", cur set idx = " + curRuleSetIdx );
       //System.out.println( "set num = " +  global.getIrlRuleNums( netIdx ).length );

       RULE_NUM = global.getIrlRuleNums( netIdx )[ curRuleSetIdx ];
       UTL_FLAG = global.isRuleUtilityOn( netIdx, global.IRL, curRuleSetIdx );
       ruleList = new IrlRule[ RULE_NUM ];
       for( int i = 0; i < RULE_NUM; i++ )
            ruleList[ i ] = new IrlRule( groupId, netIdx, curRuleSetIdx, i, global );
       activeRuleList = new Vector( RULE_NUM );
       for( int i = 0; i < RULE_NUM; i++ )
            activeRuleList.addElement( new Integer(i) );
       ruleUtls = new double[ RULE_NUM ];
       recommendRules = new short[ RULE_NUM ];
       if( global.isActionChunkBlaOn( netIdx, global.IRL, curRuleSetIdx ) )
           actionBLAoption = global.getAcsChunkBlaOption( netIdx, global.IRL, curRuleSetIdx );
       else
           actionBLAoption = global.EMPTY;
   }
}