 /************************** class RerRule begin ********************************/
 package clarion.acs.net;
 import java.util.Arrays;
 import clarion.acs.net.auxiliary.*;
 import clarion.system.*;
 import clarion.common.*;
 import clarion.tools.*;

 public class RerRule extends TLRule {

     // basic match: match all dims. general match: match all dims but one.
     protected final int BASIC_MATCH = 0, GENERAL_MATCH = 1;
     protected final int WITH = 0, WITHOUT = 1;    // used for stats.
     protected final int PM = 0, NM = 1, RATE = 2; // used for stats.
     protected int genSpecMode;        // general-to-specific or specific-to-general.
     protected short condValNums[];    // condition values in each dim.
    /*location of the max matching rate in the array statistics.*/
     protected int[] potentialLocs[];
     // if extracted, generalized or specialized.
     protected boolean MATCH_ALL_FLAG;     // if a match-all rule.
     protected boolean extractFlag, geneFlag, specFlag;
     protected boolean basicCondMatchFlag, basicMatchFlag;
     protected boolean geneCondMatchFlag, geneMatchFlag;
     protected boolean actionMatchFlag;
     protected double  c1, c2;          // used for calculating IG.
     protected float[][][][] statistics;
     /* the 3rd dimension records statistics of condition conditions[i][j]
        0: statistics WITH the condition "conditions[i][j]=1"
        1: statistics WITHOUT the condition "conditions[i][j]=1"
        the 4th dimension
        0: Positive Matches, 1: Negative Matches, 2: Rate PM+1/PN+NM+2
      */

     public RerRule( int netIdx, Global global )
     {
        super( netIdx, Global.RER, 0, global );
     }

     public RerRule( int netIdx, Global global, short[] state, short[] action, boolean mFlag )
     {
         super( netIdx, Global.RER, 0, global );
         genSpecMode = global.getRerGenSpecMode( netIdx );
         MATCH_ALL_FLAG = mFlag;
         extractFlag = true;  // new extracted rule.
         condValNums = new short[ INPUT_DIM_NUM ];
         potentialLocs = new int[2][2];
         // initialize stats.
         statistics = new float[ INPUT_DIM_NUM ][][][];
         for( int i = 0; i < inputDVs.length; i++ )
             statistics[i] = new float[ inputDVs[i] ][2][3];
         // initialzie states and suggest action.
         conditions = new short[ INPUT_DIM_NUM ][];
         for( int i = 0; i < inputDVs.length; i++ )
              conditions[ i ] = new short[ inputDVs[i] ];
         convertToRuleConditions( state );

         //suggestedAction = new short[ OUTPUT_DIM_NUM ];

         for( int i = 0; i < action.length; i++ )
              suggestedAction[i] = action[i];
     }

/*************************** BOOLEAN METHODS **********************************/
     public boolean isBasicConditionMatched()
     {
         return basicCondMatchFlag;
     }

     public boolean isGeneralConditionMatched()
     {
         return geneCondMatchFlag;
     }

     public boolean isBasicMatched()
     {
         return basicMatchFlag;
     }

     public boolean isGeneralMatched()
     {
         return geneMatchFlag;
     }

     public boolean isExtracted()
     {
         return extractFlag;
     }

     public boolean isGenerailized()
     {
         return geneFlag;
     }

     public boolean isSpecialized()
     {
         return specFlag;
     }

/***************************** GET METHODS ************************************/
     /* get the location of the max matching rate in the statistics array. */
     /* used for rule generalization and specialization. */
     public int[] getMaxIGLoc( int mode )
     {
        return potentialLocs[ mode ];
     }

     public double[] getMatchRate()
     {
        double[] matches = new double[2];
        matches[PM] = attributes.getPM();
        matches[NM] = attributes.getNM();
        return matches;
     }

     // get a match rate of a minor variation of the rule condition.
     // mode = 0: generailization, = 1: specification.
     public double[] getMaxMatches( int idx1, int idx2, int mode )
     {
         double[] matches = new double[2];
         matches[PM] = statistics[ idx1 ][ idx2 ][ 1 - mode ][ PM ];
         matches[NM] = statistics[ idx1 ][ idx2 ][ 1 - mode ][ NM ];
         return matches;
     }

/***************************** SET METHODS ************************************/
     public void setCondValNums( short[] arr )
     {
        for( int i = 0; i < arr.length; i++ )
              condValNums[ i ] = arr[ i ];
     }

     public void setPotentialLocs( int[][] arr )
     {
        for( int i = 0; i < arr.length; i++ )
         for( int j = 0; j < arr[ i ].length; j++ )
              potentialLocs[ i ][ j ] = arr[ i ][ j ];
     }

     public void setStatistics( float[][][][] stats )
     {
        for( int i = 0; i < stats.length; i++ )
         for( int j = 0; j  < stats[ i ].length; j++ )
          for( int k = 0; k < stats[ i ][ j ].length; k++ )
           for( int m = 0; m < stats[ i ][ j ][ k ].length; m++ )
                statistics[ i ][ j ][ k ][ m ] = stats[ i ][ j][ k ][ m ];
     }

     public void setGenerailizedFlag( boolean flag )
     {
         geneFlag = flag;
     }

     public void setSpecializedFlag( boolean flag )
     {
         specFlag = flag;
     }

     public void resetStatistics()
     {
        super.resetStatistics();
        for( int i = 0; i < potentialLocs.length; i++ )
          for( int j = 0; j < potentialLocs[i].length; j++ )
               potentialLocs[ i ][ j ] = 0;
        for( int i = 0; i < INPUT_DIM_NUM; i++)
          for( int j = 0; j < inputDVs[i]; j++)
            for( int k = 0; k < 2; k++ )
              for( int m = 0; m < 3; m++ )
                   statistics[i][j][k][m] = 0;
     }

     /* search the location of the max matching rate in the array statistics.
        mode 0: generalization, mode 1: specialization. */
     public int[] searchMaxIGLoc( int mode )
     {
        double val, maxVal = 0.0;
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
            for( int j = 0; j < inputDVs[i]; j++ )
                if( conditions[i][j] == mode )
                {
                    val = statistics[i][j][1-mode][RATE];
                    if( maxVal < val )
                    {
                        maxVal = val;
                        potentialLocs[mode][0] = i;
                        potentialLocs[mode][1] = j;
                    }
                }
         return potentialLocs[mode];
     }

   // add a value of one dimension to the rule.
   // used for generaliztion.
   public void addCondition( int idx1, int idx2 )
   {
        conditions[ idx1 ][ idx2 ] = 1;
        condValNums[ idx1 ]++;
   }

   // del a value of one dimension to the rule.
   // used for specialization.
   public void delCondition( int idx1, int idx2 )
   {
        if( condValNums[ idx1 ] <= 1 )
            return;
        conditions[ idx1 ][ idx2 ] = 0;
        condValNums[ idx1 ]--;
   }

   // test if the current rule can be generalized.
   public boolean canGeneralized()
   {
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
             if( i == potentialLocs[Global.GENERALIZATION ][0] && condValNums[ i ] < inputDVs[i] )
                 return true;
        return false;
   }

   // test if the current rule can be specialized.
   public boolean canSpecialized()
   {
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
              if( i == potentialLocs[Global.SPECIFICATION][0] && condValNums[ i ] > 1  )
                  return true;   // modified 07/08/03
        return false;
   }

/************************** RULE MATCH METHODS *********************************/
   // if the chosen action matches the conclusion of the rule.
   public void matchAction( short[] action )
   {
        // check if the action is same as suggestedAction.
        boolean bValue = isActionMatched( action );
        if( bValue ){
            actionMatchFlag = true;
            basicMatchFlag = basicCondMatchFlag;
            geneMatchFlag = geneCondMatchFlag;
        }
        else{
            actionMatchFlag = false;
            basicMatchFlag = false;
            geneMatchFlag = false;
        }
   }

   // if current state matches the rule condition and calculate the rule support.
   public boolean decideAction( short[] state, short[] exState, short[] gsState, short[] wmState )
   {
        extractFlag = geneFlag = specFlag = false;
        int misMatchNum = isConditionMatched( state );
        if( misMatchNum == 0 ){  // match for all dims.
            basicCondMatchFlag = true;
            geneCondMatchFlag = false;
        }
        if( misMatchNum == 1 ){  // match for all dims except one.
            basicCondMatchFlag = false;
            geneCondMatchFlag = true;
        }
        if( misMatchNum > 1 ){
            basicCondMatchFlag = false;
            geneCondMatchFlag = false;
        }
        return basicCondMatchFlag;
   }

/************************** RULE UPDATE METHODS *******************************/
   public void update( boolean successFlag, int curTime )
   {
        super.update( basicCondMatchFlag, successFlag, curTime );
        if( basicMatchFlag )
            updateStatistics( successFlag, BASIC_MATCH );
        if( geneMatchFlag )
            updateStatistics( successFlag, GENERAL_MATCH );
   }

   // the method is only used when the rule is a match-all rule.
   public void updateMatchAll( boolean successFlag )
   {
        attributes.setAge( 0 ); // it means the rule is applied this step.
        updateStatistics( successFlag, BASIC_MATCH );
   }

   /* update the rule statistics whenever a match exists.  */
   public void updateStatistics( boolean successFlag, int matchMode )
   {
        int matchType;
        if( successFlag )
            matchType = PM;
        else
            matchType = NM;
        if( matchMode == BASIC_MATCH ) // match all of dims.
            attributes.updateMatches( matchType );
        else
        {  // match all of dims but one.
           Arrays.sort( inputConditions[ DIFF_DIM_IDX ] );
           // update the stats of the input condition values in the mismatched dim
           for( int i = 0; i < inputConditions[DIFF_DIM_IDX].length; i++ )
           {
                int condIdx = inputConditions[ DIFF_DIM_IDX ][ i ];
                statistics[DIFF_DIM_IDX][condIdx][WITH][matchType] += 1;
           }
           // update the stats of the rule condition values except the above ones
           // in the mismatched dim.
           for( int i = 0; i < inputDVs[ DIFF_DIM_IDX ]; i++ )
           {
                if( conditions[DIFF_DIM_IDX][i] == 1 )
                {
                    int val = Arrays.binarySearch(inputConditions[DIFF_DIM_IDX],(short)i);
                    if( val < 0 ) // not included in the input condition.
                        statistics[DIFF_DIM_IDX][i][WITHOUT][matchType] += 1;
                }
           }
           return;
        }
        if( MATCH_ALL_FLAG )  // a match-all rule.
            return;
        // update statistics of the input conditions.
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
         for( int j = 0; j < inputConditions[i].length; j++ )
              statistics[i][inputConditions[i][j]][WITH][matchType] += 1;
        // update statistics of the rule conditions except the above ones.
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
        {
             Arrays.sort( inputConditions[i] );
             for( int j = 0; j < inputDVs[i]; j++ )
             {    // if not included in the input condition.
                  if(conditions[i][j]==1 && Arrays.binarySearch(inputConditions[i],(short)j)<0)
                     statistics[i][j][WITHOUT][matchType] += 1;
            }
        }
        // calculate all of the match rates.
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
          for( int j = 0; j < inputDVs[i]; j++ )
               for( int k = 0; k < 2; k++ )
               {
                    float val1 = statistics[i][j][k][PM];
                    float val2 = statistics[i][j][k][NM];
                    statistics[i][j][k][RATE] = (val1 + 1) / (val1 + val2 + 2);
               }
   }

   // discount the rule statistics after each trial.
   public void discountStatistics()
   {
       attributes.discountMatches( DISCOUNT );
       for( int i = 0; i < INPUT_DIM_NUM; i++ )
         for( int j = 0; j < inputDVs[i]; j++ )
            for( int k = 0; k < 2; k++ )
            {
                 statistics[i][j][k][PM] *= DISCOUNT;
                 statistics[i][j][k][NM] *= DISCOUNT;
                 float val1 = statistics[i][j][k][PM];
                 float val2 = statistics[i][j][k][NM];
                 statistics[i][j][k][RATE] = (val1 + 1) / (val1 + val2 + 2);
            }
   }

/************************** OVERRIDE METHODS **********************************/
   // convert the rule into a text string.
    public String toString()
    {
         int MAX_INPUT_DIM_NUM = INPUT_DIM_NUM;
         String ruleString = "{ ";
         for( int i = 0; i < MAX_INPUT_DIM_NUM; i++ ){
              ruleString += "(";
              for( int j = 0; j < inputDVs[i]; j++ ){
                   if( conditions[ i ][ j ] == 1 )
                      ruleString += ( j + 1 ) + " ";
              }
              ruleString += ")";
              if( i < MAX_INPUT_DIM_NUM - 1 )
                  ruleString += ",";
         }
         ruleString += " }";
         //ruleString += " --> action " + String.valueOf( suggestedAction+1 );
         ruleString += " --> action ( ";
         for( int i = 0; i < suggestedAction.length; i++ )
              ruleString += ( suggestedAction[i] + 1 ) + " ";
         ruleString += ") \n";
         return ruleString;
   }

   public int hashCode()
   {
       int num = Math.min( 32, outputDVs[0] );
       int code = 0;
       for( int i = 0; i < num; i++ )
            code = (code <<= 1) + conditions[0][i];
       return code;
   }

   public boolean equals( Object obj )
   {
       if( !(obj instanceof RerRule ) || obj == null )
            return false;
       RerRule otherRule = (RerRule)obj;
       // if this covers other rule
       int retVal = this.cover( otherRule );
       if( retVal >= 0 || retVal == -1 )
           return false;
       // if other rule covers this rule.
       retVal = otherRule.cover( this );
       if( retVal >= 0 || retVal == -1 )
           return false;
       return true;
   }

   protected Object clone( )
   {
         short[] state = new short[ global.getIdnInputNum( netIdx ) ];
         RerRule rule = new RerRule( netIdx, global, state, suggestedAction,  MATCH_ALL_FLAG );
         rule.setCondition( conditions );
         rule.setCondValNums( condValNums );
         rule.setPotentialLocs( potentialLocs );
         rule.setStatistics( statistics );
         rule.setAttributes( attributes );
         return rule;
   }

/************************** PRIVATE METHODS ************************************/
   private void convertToRuleConditions( short[] state )
   {    // decide the source of the rule condition for each dim.
        int NON_MASK_NUM = 1;
        int masks[] = new int[ INPUT_DIM_NUM ];
        if( ( genSpecMode == global.DBL_DIRECT && Math.random() < 0.5 ) ||
              genSpecMode == global.GEN_TO_SPEC )
        {
            for( int i = 0; i < INPUT_DIM_NUM; i++ )
                 masks[ i ] = 1;
            int count = 0;
            while( count < NON_MASK_NUM )
            {
                int dimIdx = (int)(Math.random() * INPUT_DIM_NUM);
                if( masks[ dimIdx ] == 1 )
                {
                    masks[ dimIdx ] = 0;
                    count++;
                }
            }
        }
        else
            for( int i = 0; i < INPUT_DIM_NUM; i++ )
                 masks[ i ] = 0;
        // generate the rule conditions.
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
          for( int j = 0; j < inputDVs[i]; j++)
                 conditions[i][j] = 0;
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
        {
             for( int j = 0; j < inputDVs[i]; j++ )
             {
                  if( masks[ i ] == 0 )
                      conditions[ i ][ j ] = state[ inputOffsets[i] + j ];
                  else
                      conditions[ i ][ j ] = 1;
                  if( conditions[ i ][ j ] == 1 )
                      condValNums[ i ]++;
             }
        }
   }
 }