package clarion.acs.net;
import clarion.system.*;

public class RerRuleEx extends RerRule {

     protected float[][][] specStats;
     /*  record statistics of all the values in conditions in order to determine
         which dimension will be specialized. the 3rd dimension:
         0: Positive Matches, 1: Negative Matches, 2: Rate PM + 1/ PN + NM + 2  */
     protected float[][] genStats;
     /* record statistics of all the dimensions of conditions in order to determine
        which dimension will be generalized. the 2nd dimension:
        0: Positive Matches, 1: Negative Matches, 2: Rate PM + 1/ PN + NM + 2 */
     protected int[][] dimGenSpecInfo;
     /* used for which dimension of conditions can be generalized or specialized
        only record the valid dimensions which has non-zero values in its elements
        1st dimension contains the indices of all the valid dimensions.
        2nd dimension indicates the corresponding dimension can be generalized or
        specialized. = 0: can be generalized; = 1: can be specialized */
     protected int VALID_DIM_NUM, genIdx, specIdx[];
     /* VALID_DIM_NUM means the # of dimensions which has non-zero elements
        genIdx: indicating the dimension which can generalized at the most priority.
        specIdx: indicating the dimension and one of its elements which can be
        specialized at the most priority.*/

 public RerRuleEx( int netIdx, Global g, short[] state, short[] action, boolean mFlag, int vDimNum )
 {
        super( netIdx, g, state, action, mFlag );
        VALID_DIM_NUM = vDimNum;
        genStats = new float[INPUT_DIM_NUM][3];
        specStats = new float[INPUT_DIM_NUM][][];
        for( int i = 0; i < inputDVs.length; i++ )
             specStats[i] = new float[inputDVs[i]][3];
        dimGenSpecInfo = new int[VALID_DIM_NUM][2];
        if( !MATCH_ALL_FLAG )
            setValidDims( state );
 }

 public void setValidDims( short[] state )
 {
     int count = 0;
     for( int i = 0; i < INPUT_DIM_NUM; i++ )
      for( int j = 0; j < inputDVs[i]; j++ )
      {
           if( state[ inputOffsets[i] + j ] == 1 )
           {
               dimGenSpecInfo[count++][0] = i;
               break;
           }
      }
 }

 public int getDimGenSpecInfoIdx( int idx )
 {
     int retVal = 0;
     for( int i = 0; i < dimGenSpecInfo.length; i++ )
          if( dimGenSpecInfo[ i ][ 0 ] == idx )
          {
              retVal = i;
              break;
          }
     return retVal;
 }

  /* the following three methods functioned as super.getOptCondition() */
 public double[] getMaxMatches( int mode )
 {
     return(mode == 0) ? getMaxGenMatches(): getMaxSpecMatches();
 }

 public double[] getMaxGenMatches()
 {
     double val1, val2, val, maxVal = 0.0;
     double[] maxVals = new double[2];
     for( int i = 0; i < VALID_DIM_NUM; i++ ){
          if( dimGenSpecInfo[i][1] == 0 ){
              val1 = genStats[dimGenSpecInfo[i][0]][0];
              val2 = genStats[dimGenSpecInfo[i][0]][1];
              val = ( val1 + 1 ) / ( val1 + val2 + 2 );
              if( val > maxVal ){
                  maxVal = val;     // modified by xi zhang 11/28/01
                  maxVals[PM] = val1;  // modified by xi zhang 5/5/03
                  maxVals[NM] = val2;  // modified by xi zhang 5/5/03
                  genIdx = dimGenSpecInfo[i][0];
              }
           }
     }
     return maxVals;
 }

 public double[] getMaxSpecMatches()
 {
     double val1, val2, val, maxVal = 0.0;
     double[] maxVals = new double[2];
     specIdx = new int[2];
     for( int i = 0; i < VALID_DIM_NUM; i++ ){
          if( dimGenSpecInfo[i][1] == 1 ){
              for( int j = 0; j < inputDVs[i]; j++ ){
                   val1 = specStats[dimGenSpecInfo[i][0]][j][0];
                   val2 = specStats[dimGenSpecInfo[i][0]][j][1];
                   val = ( val1 + 1 ) / ( val1 + val2 + 2 );
                   if( val > maxVal ){
                       maxVal = val;        // modified by xi zhang 11/28/01
                       maxVals[PM] = val1;  // modified by xi zhang 5/5/03
                       maxVals[NM] = val2;  // modified by xi zhang 5/5/03
                       specIdx[0] = dimGenSpecInfo[i][0];
                       specIdx[1] = j;
                   }
              }
          }
     }
     return maxVals;
 }

 public void setDimGenSpecInfo( int[][] info )
 {
     for( int i = 0; i < info.length; i++ )
      for( int j = 0; j < info[ i ].length; j++ )
           dimGenSpecInfo[ i ][ j ] = info[ i ][ j ];
 }

 public void setGenStats( float[][] info )
 {
     for( int i = 0; i < info.length; i++ )
       for( int j = 0; j < info[ i ].length; j++ )
            genStats[ i ][ j] = info[ i ][ j ];
 }

 public void setSpecStats( float[][][] info )
 {
     for( int i = 0; i < info.length; i++ )
       for( int j = 0; j < info[ i ].length; j++ )
         for( int k = 0; k < info[ i ][ j ].length; k++ )
              specStats[ i ][ j ][ k ] = info[ i ][ j ][ k ];
 }

 public void resetStatistics()
 {
     attributes.resetMatches();
     for( int i = 0; i < INPUT_DIM_NUM; i++)
     {
          for( int j = 0; j < 3; j++ )
               genStats[i][j] = 0;
          for( int j = 0; j < inputDVs[i]; j++)
            for( int k = 0; k < 3; k++ )
                 specStats[i][j][k] = 0;
     }
 }

 // test if the current rule can be generalized.
 public boolean canGeneralized()
 {
     for( int i = 0; i < VALID_DIM_NUM; i++ )
          if( dimGenSpecInfo[i][1] == 0 )
              return true;
     return false;
 }

 // test if the current rule can ba specialized.
 public boolean canSpecialized()
 {
     for( int i = 0; i < VALID_DIM_NUM; i++ )
          if( dimGenSpecInfo[i][1] == 1 )
              return true;
     return false;
 }

 public void generalizeADim()
 {
     for( int i = 0; i < inputDVs[i]; i++ )
          conditions[ genIdx ][ i ] = 1;
     int idx = getDimGenSpecInfoIdx( genIdx );
     // it means this dimension can only be specialized and not generalized in the future.
     dimGenSpecInfo[ idx ][ 1 ] = 1;
 }

 public void specializeADim()
 {
     for( int i = 0; i < inputDVs[i]; i++ )
          conditions[ specIdx[0] ][ i ] = 0;
     conditions[ specIdx[0] ][ specIdx[1] ] = 1;
     int idx = getDimGenSpecInfoIdx( specIdx[0] );
     //it means this dimension can only be generalized and not specialized in the future.
     dimGenSpecInfo[ idx ][ 1 ] = 0;
 }

 public void updateStatistics( boolean successFlag, int matchMode )
 {
     // match current condition and action.
     int matchType;
     if( successFlag )
         matchType = PM;
     else
         matchType = NM;
     if( matchMode == BASIC_MATCH ) // match all of dims.
         attributes.updateMatches( matchType );
     else
     {   // match all of dims but one.
         // update the stats of the input condition values in the mismatched dim
         for( int i = 0; i < inputConditions[DIFF_DIM_IDX].length; i++ )
         {
              int condIdx = inputConditions[ DIFF_DIM_IDX ][ i ];
              specStats[DIFF_DIM_IDX][condIdx][matchType] += 1;
         }
         genStats[DIFF_DIM_IDX][matchType] += 1;
         return;
     }
     if( MATCH_ALL_FLAG )
         return;
     // update statistics of the input conditions.
     for( int i = 0; i < INPUT_DIM_NUM; i++ )
       for( int j = 0; j < inputConditions[i].length; j++ )
            specStats[i][inputConditions[i][j]][matchType] += 1;
 }

 // discount the rule statistics after each trial.
 public void discountStatistics()
 {
     attributes.discountMatches( DISCOUNT );
     for( int i = 0; i < INPUT_DIM_NUM; i++ )
     {
          genStats[i][0] *= DISCOUNT;
          genStats[i][1] *= DISCOUNT;
          for( int j = 0; j < inputDVs[i]; j++ )
          {
               specStats[i][j][0] *= DISCOUNT;
               specStats[i][j][1] *= DISCOUNT;
          }
     }
 }

/******************************* PROTECTED METHODS *****************************/
 protected Object clone( )
 {
     short[] state = new short[ global.getIdnInputNum( netIdx ) ];
     RerRuleEx rule = new RerRuleEx( netIdx, global, state, suggestedAction,
                                     MATCH_ALL_FLAG, VALID_DIM_NUM );
     rule.setDimGenSpecInfo( dimGenSpecInfo );
     rule.setGenStats( genStats );
     rule.setSpecStats( specStats );
     return rule;
 }
}