package clarion.acs.net.auxiliary;
import java.util.HashSet;
import java.util.Iterator;
import clarion.system.*;
import clarion.acs.net.*;

public class ActiveRule {

  protected RerRule rule;
  protected HashSet children;
  protected int netIdx;
  protected TaskClarion taskClarion;

  public ActiveRule( RerRule rule, int netIdx, TaskClarion taskClarion )
  {
      this.rule = rule;
      children = new HashSet();
      this.netIdx = netIdx;
      this.taskClarion = taskClarion;
  }

  public int getCount()
  {
      return children.size();
  }

  public RerRule getRule()
  {
      return rule;
  }

  public void setRule( RerRule rule )
  {
     this.rule = rule;
  }

  public Object[] getChildren()
  {
      return children.toArray();
  }

  public double getPM()
  {
     return rule.getPM();
  }

  public double getNM()
  {
     return rule.getNM();
  }

  public double getOverallPM()
  {
     double pm = getPM();
     Iterator iter = children.iterator();
     while( iter.hasNext() ){
         RerRule chd = (RerRule)iter.next();
         pm += chd.getPM();
     }
     return pm;
  }

  public double getOverallNM()
  {
     double pm = getNM();
     Iterator iter = children.iterator();
     while( iter.hasNext() ){
         RerRule chd = (RerRule)iter.next();
         pm += chd.getNM();
     }
     return pm;
  }

  public void addChildren( ActiveRule activeRule )
  {
      /*
      System.out.println( "before, added activeRule = \n" + activeRule );
      System.out.println( "before, key rule = \n" + rule );
      System.out.println( "before, children = " );
      Object[] chds = children.toArray();
      for( int i = 0; i < children.size(); i++ )
           System.out.print( chds[i] );
       */

      addChild( activeRule.getRule() );
      if( activeRule.getCount() == 0 )
          return;
      Object[] rules = activeRule.getChildren();
      for( int i = 0; i < rules.length; i++ )
           addChild( (RerRule)rules[i] );

      /*
      System.out.println( "after, children = " );
      chds = children.toArray();
      for( int i = 0; i < children.size(); i++ )
           System.out.print( chds[i] );
      */
  }

  public RerRule[] removeChildren()
  {
     RerRule[] rules = new RerRule[ children.size() ];
     int count = 0;
     Iterator iter = children.iterator();
     while( iter.hasNext() ){
          RerRule chd = (RerRule)iter.next();
          if( rule.cover( chd ) == -2 ){ // one-direction cover.
              if( chd.cover( rule ) == -2 ){  // two-direction cover.
                  iter.remove();
                  double ig = taskClarion.getRerIG( netIdx, rule.getMatchRate()[0], rule.getMatchRate()[1],
                                                    chd.getMatchRate()[0], chd.getMatchRate()[1] );
                  if( ig < 0.0 )
                      this.setRule( chd );
              }
          }
          else{
             rules[count++] = chd;
             iter.remove();
          }
     }
     // fill in the restored children.
     RerRule[] restoreChildren = new RerRule[ count ];
     for( int i = 0; i < count; i++ )
          restoreChildren[i] = rules[i];
     return restoreChildren;
  }

  public int cover( ActiveRule otherRule )
  {
      return rule.cover( otherRule.getRule() );
  }

  public String toString()
  {
     String str = "ACTIVE RULE: " + rule;
     Iterator iter = children.iterator();
     while( iter.hasNext() ){
         RerRule chd = (RerRule)(iter.next());
         str += "-----------INACTIVE RULE: " + chd;
     }
     return str;
  }

  public int hashCode()
  {
     return 0;
  }

  public boolean equals( Object obj )
  {
     if( !(obj instanceof ActiveRule) || obj == null )
         return false;
     return rule.equals( ((ActiveRule)obj).getRule() );
  }

  private boolean addChild( RerRule child )
  {
      if( child == null || rule.equals( child ) ){
          //System.out.println( "the child cannot be added !" );
          return false;
      }
      boolean containFlag = children.contains( child );
      /*
      boolean equalFlag = false;
      for( int i = 0; i < children.size(); i++ )
           if( child.equals( children.toArray()[i] ) ){
               equalFlag = true;
               break;
           }
      if( equalFlag != containFlag ){
          System.out.println( "SOME ERRORS IN HASHSET CONTAIN" );
          System.out.println( "child = " + child );
          System.exit( 1 );
      }
      */
      //if( !equalFlag  )
      if( !containFlag )
          children.add( child );
      else{
          //System.out.println( "child is one of the children, child = \n" + child );

          Iterator iter = children.iterator();
          while( iter.hasNext() ){
              RerRule chd = (RerRule)iter.next();
              if( chd.equals( child ) ){
                  double ig = taskClarion.getRerIG( netIdx, chd.getMatchRate()[0], chd.getMatchRate()[1],
                                                    child.getMatchRate()[0], child.getMatchRate()[1] );
                  if( ig < 0.0 ){
                      iter.remove();
                      children.add( child );
                  }
                  break;
              }
          }
      }
      return true;
  }

}