 package clarion.acs.net.auxiliary;
 import java.util.Arrays;
 import clarion.system.*;
 import clarion.common.*;
 import clarion.tools.*;

 /**
  * this rule is for implementing action rule representation.
  */
 public class TLRule{
   protected final static double DISCOUNT = 0.9;  // used for udpating stats.
   // global settings.
   protected Global global;
   // the user-defined CLARION settings.
   protected TaskClarion taskClarion;
   // the net indice.
   protected int netIdx, ruleType, setIdx;
   protected int INPUT_NUM, OUTPUT_NUM;
   protected int INPUT_DIM_NUM, OUTPUT_DIM_NUM;
   protected short inputDVs[], outputDVs[];
   protected short inputOffsets[], outputOffsets[];
   protected short conditions[][], suggestedAction[];
   protected short inputConditions[][];   // input state in form of dim-vals.
   protected int   DIFF_DIM_IDX;          // used for general match.
   protected boolean partialMatchFlag;
   protected double  ruleSupport;
   protected int blaOption;
   protected double ruleCost;
   protected RuleAttributes attributes;

   public TLRule( int netIdx, int ruleType, int setIdx, Global global )
   {
       this.netIdx = netIdx;
       this.ruleType = ruleType;
       this.setIdx = setIdx;
       this.global = global;
       taskClarion = global.getTaskClarion();
       INPUT_NUM = global.getARInputNum( netIdx, ruleType );
       OUTPUT_NUM = global.getAROutputNum( netIdx, ruleType );
       INPUT_DIM_NUM = global.getARInputDimNum( netIdx, ruleType );
       OUTPUT_DIM_NUM = global.getAROutputDimNum( netIdx, ruleType );
       inputDVs = global.getARInputDimVals( netIdx, ruleType );
       outputDVs = global.getAROutputDimVals( netIdx, ruleType );
       inputOffsets = new short[ INPUT_DIM_NUM ];
       outputOffsets = new short[ OUTPUT_DIM_NUM ];
       CommonMethods.calcOffsets( inputOffsets, inputDVs );
       CommonMethods.calcOffsets( outputOffsets, outputDVs );
       inputConditions = new short[ INPUT_DIM_NUM ][];
       suggestedAction = new short[ OUTPUT_DIM_NUM ];
       if( global.isRuleBlaOn( netIdx, ruleType, setIdx ) )
           blaOption = global.getActionRuleBlaOption( netIdx, ruleType, setIdx );
       else
           blaOption = global.EMPTY;
       // initialize rule attributes.
       attributes = new RuleAttributes( netIdx, ruleType, setIdx, global );
       partialMatchFlag = global.isRulePartMatchOn( netIdx, ruleType, setIdx );
   }

/*###############################################################################################
   This method and several others need be revised, DIM INFOS of state(IDN input) may be different
   from those of rules since DIMS of rules are a subset of DIMS of IDN.
################################################################################################*/
   /**
    * if current state matches the rule condition and calculate the rule support.
    * also get the different dimension index if only one dim is mismatched.
    * returns the # of mismatched dimensions.
    */
   public int isConditionMatched( short[] input )
   {
      convertToInputConditions( input );
      DIFF_DIM_IDX = -1;
      ruleSupport = 0.0;
      double[] activations = new double[INPUT_DIM_NUM];
      for( int i = 0; i < INPUT_DIM_NUM; i++ )
           activations[i] = 0.0;
      int misMatchNum = 0;
      for( int i = 0; i < INPUT_DIM_NUM; i++ ){
           boolean dimMatchFlag = true;
           for( int j = 0; j < inputConditions[i].length; j++ ){
                int condIdx = inputConditions[i][j];
                if( conditions[i][condIdx] == 0 ){
                    dimMatchFlag = false;
                    break;
                }
           }
           if( dimMatchFlag )
               activations[i] = 1.0;
           else{
               misMatchNum++;
               DIFF_DIM_IDX = i;
           }
      }
      // calculate rule support.
      if( partialMatchFlag ){
          double[] weights = taskClarion.getRerWeights(netIdx);
          switch( ruleType ){
                  case Global.RER:
                       weights = taskClarion.getRerWeights(netIdx);
                       break;
                  case Global.IRL:
                       weights = taskClarion.getIrlWeights(netIdx);
                       break;
                  case Global.FR:
                       weights = taskClarion.getFrWeights(netIdx);
                       break;
          }
          for( int i = 0; i < INPUT_DIM_NUM; i++ )
               ruleSupport += activations[i] * weights[i];
      }
      else{
          if( misMatchNum == 0 )
              ruleSupport = 1.0;
          else
              ruleSupport = 0.0;
      }
      return misMatchNum;
   }

   public boolean isActionMatched( short[] action )
   {
       return CommonMethods.isEqual( suggestedAction, action );
   }

   // this may be overrided.
   public boolean decideAction( short[] state, short[] exState, short[] gsState, short[] wmState )
   {
       boolean matchFlag = (isConditionMatched(state) == 0);
       return matchFlag;
   }

   public short[][] getCondition()
   {
       return conditions;
   }

   public short[] getCondition( int dimIdx )
   {
       return conditions[ dimIdx ];
   }

   public void getAction( short[] action )
   {
       for( int i = 0; i < action.length; i++ )
            action[i] = suggestedAction[i];
   }

   public Chunk getActionChunk()
   {
      int type = Global.CHK_ACTION;
      String label = Global.CHUNK_TYPE_LABELS[type];
      Feature[] featList = new Feature[ OUTPUT_DIM_NUM ];
      for( int i = 0; i < OUTPUT_DIM_NUM; i++ ){
           int typeIdx = global.getAROutputFeatureInfo( netIdx, ruleType )[i][0];
           int featIdx = global.getAROutputFeatureInfo( netIdx, ruleType )[i][1];
           String featLabel = global.getFeatureNames( typeIdx, featIdx );
           int featLen = global.getFeatureLength( typeIdx, featIdx );
           short[] values = new short[1];
           values[0] = suggestedAction[i];
           featList[i] = new Feature( featLabel, (short)typeIdx, (short)featIdx, (short)featLen,
                                      values, global );
      }
      Chunk chunk = new Chunk( type, label, featList, global );
      return chunk;
   }

   public int getAge()
   {
       return attributes.getAge();
   }

   public double getPM()
   {
       return attributes.getPM();
   }

   public double getNM()
   {
       return attributes.getNM();
   }

   public int getResponseTime()
   {
       return attributes.getResponseTime();
   }

   public double getRuleSupport()
   {
       return ruleSupport;
   }

   public double getUtility()
   {
       attributes.calcUtility( ruleCost );
       return attributes.getUtility();
   }

   public void setCondition( short[][] cond )
   {
      for( int i = 0; i < cond.length; i++ )
       for( int j = 0; j < cond[ i ].length; j++ )
            conditions[ i ][ j ] = cond[ i ][ j ];
   }

   /*
   public void setSuggestedAction( short[] action )
   {
      //for( int i = 0; i < OUTPUT_DIM_NUM; i++ )
      //    suggestedAction[i] = action[i];
      suggestedAction = action;
   }
   */

   public void setAge( int age )
   {
       attributes.setAge( age );
   }

   public void setPM( float pm )
   {
       attributes.setPM( pm );
   }

   public void setNM( float nm )
   {
       attributes.setNM( nm );
   }

   /*
   public void setRuleCost()
   {
       attributes.setCost( ruleCost );
   }
   */

   public void setAttributes( RuleAttributes feat )
   {
       attributes = (RuleAttributes)feat.clone();
   }

   public void aging()
   {
       attributes.updateAge();
   }

   public void discountStatistics()
   {
       attributes.discountMatches( DISCOUNT );
   }

   public void resetStatistics()
   {
       attributes.resetMatches();
   }

   public void calcBLA( int curTime )
   {
       attributes.calcBLA( curTime );
   }

   public void calcTLDT( double chunkBLA )
   {
       attributes.calcTLDT( chunkBLA );
   }

   public void calcRT( int TLAT )
   {
       attributes.calcTLRT( TLAT );
   }

   public void update( boolean actionMatchFlag, boolean successFlag, int curTime )
   {
       attributes.setAge( 0 );
       if( blaOption == global.VARIED_OPTION )
           attributes.updateUsages( curTime );
       if( actionMatchFlag )
           attributes.updateMatches( successFlag );
   }

   // inputConditions used for converting the original input state into input conditions.
   public void convertToInputConditions( short[] state )
   {
        for( int i = 0; i < INPUT_DIM_NUM; i++ ){
             short[] values = new short[ inputDVs[i] ];
             int count = 0;
             for( int j = 0; j < inputDVs[i]; j++ )
                  if( state[ inputOffsets[i] + j ] == 1 )
                      values[ count++ ] = (short)j;
             inputConditions[ i ] = new short[ count ];
             for( int j = 0; j < count; j++ )
                  inputConditions[ i ][ j ] = values[ j ];
        }
   }

   /**
    * if this rule covers other rule by both condition and action.
    * returns:
    *   -2: cover completely;
    *   -1: more than one dim is different
    * >= 0: the different dim index.
    */
   public int cover( TLRule rule )
   {
      if( rule == null )
          return -1;
      // check if the actions are the same.
      short[] otherAction = new short[ OUTPUT_DIM_NUM ];
      rule.getAction( otherAction );
      for( int i = 0; i < suggestedAction.length; i++ )
           if( otherAction[i] != suggestedAction[i] )
               return -1;
      //check if the conditions are the same.
      int retVal = -2;
      short[][] otherCondition = rule.getCondition();
      for( int i = 0; i < INPUT_DIM_NUM; i++ )
      {
           int count = 0;
           for( int j = 0; j < inputDVs[i]; j++ )
                if( conditions[i][j] == 0 && otherCondition[i][j] == 1 )
                    count++;
           if( count > 0 )
           {
               if( retVal == -2 )   // first different dim
                   retVal = i;
               else                // more than one dim different.
               {
                   retVal = -1;
                   break;
               }
           }
       }
       return retVal;
  }
 }