package clarion.acs.net.auxiliary;
import java.util.*;
import java.util.TreeMap;
import clarion.acs.net.*;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class TLRuleSet implements AcsNetComp {
   protected int MATCH_CONDITION = 0, MATCH_BOTH = 1, ALL = 2;
   // global settings.
   protected Global global;
   // task-specific clarion settings.
   protected TaskClarion taskClarion;
   // the net indice.
   protected int netIdx;
   protected int ruleType;
   protected TLRule[] ruleList;
   protected int RULE_NUM;     // rule # of current rule set, the length of ruleList.
   protected int MAX_AGE;      // relative to freq of applying a rule
   protected int genMode;      // generalization mode.
   protected int INPUT_DIM_NUM;   // # of input dims
   protected int OUTPUT_DIM_NUM;  // # of output dims
   protected int INPUT_NUM;      // length of the input.
   protected int OUTPUT_NUM;     // length of the output.
   protected short[] inputDVs;      // # of values int each input dim.
   protected short[] outputDVs;     // # of values int each output dim.
   protected short[] inputOffsets;  // the start position if each dim in the input array.
   protected short[] outputOffsets; // the start position if each dim in the output array.
   protected boolean partialMatchFlag;  // partial match flag.
   protected boolean UTL_FLAG;        // utility flag.
   protected double  ruleUtls[];      // rule utilities used for action decision.
   protected double  outputActivations[];  // output activations used for weight-sum combination
   protected TreeMap activationList = new TreeMap(); // the activations of the the activated actions.
   protected short[] state, exState, gsState, wmState;  // current states.
   protected int c20;  // used for MCS's minotoring conclusions.
   protected Chunk[] tlResults;
   protected int ruleMatchNum, recommendRule;
   protected short[] chosenAction, recommendRules;
   protected String ruleReps[];
  // used for cross combination.
   protected double  SUCC_C3, SUCC_C4;
   protected double  succRate, sumPM, sumNM;
   // used for updating stats.
   protected int episodeLength, trialIdx;
   // used for RT.
   protected int actionBLAoption;
   protected int currentTime, responseTime;
   protected TreeMap actionUsageInfo = new TreeMap();  // action usages info.
   protected StochasticDecider actionDecider;
   // used for the key to TreeMap.
   protected int[] multipliers;

   public TLRuleSet ( int netIdx, int ruleType, Global global )
   {
       this.netIdx = netIdx;
       this.ruleType = ruleType;
       this.global = global;
       taskClarion = global.getTaskClarion();
       INPUT_DIM_NUM = global.getARInputDimNum( netIdx, ruleType );
       OUTPUT_DIM_NUM = global.getAROutputDimNum( netIdx, ruleType );
       INPUT_NUM = global.getARInputNum( netIdx, ruleType );
       OUTPUT_NUM = global.getAROutputNum( netIdx, ruleType );
       inputDVs = global.getARInputDimVals( netIdx, ruleType );
       outputDVs = global.getAROutputDimVals( netIdx, ruleType );
       inputOffsets = new short[INPUT_DIM_NUM];
       outputOffsets = new short[OUTPUT_DIM_NUM];
       CommonMethods.calcOffsets( inputOffsets, inputDVs );
       CommonMethods.calcOffsets( outputOffsets, outputDVs );
       UTL_FLAG = global.isRuleUtilityOn( netIdx, ruleType, 0 );
       outputActivations = new double[ OUTPUT_NUM ];
       chosenAction = new short[ OUTPUT_DIM_NUM ];
       // initialize the conclusions for monitoring.
       if( global.isMcsOn() )
           c20 = global.getMonitorAcsTopConcNum();
       ruleReps = new String[ 3 ];
       episodeLength = global.getEpisodeLength();
       // initialize the variables used for cross combination.
       if( global.getAcsCrossCombineMode( netIdx, 1 ) == Global.VAR_COMB_MODE ){
           SUCC_C3 = global.getAcsCrossCombineVarParaAbove( netIdx )[ruleType];
           SUCC_C4 = global.getAcsCrossCombineVarParaUnder( netIdx )[ruleType];
       }
       // initialize the action usages for calculating BLA if the bla option is set.
       if( global.isActionChunkBlaOn( netIdx, ruleType, 0 ) )
           actionBLAoption = global.getAcsChunkBlaOption( netIdx, ruleType, 0 );
       else
           actionBLAoption = Global.EMPTY;
       if( actionBLAoption == Global.VARIED_OPTION )
           actionUsageInfo = new TreeMap();  // action usages.
       // action decider.
       double temper = global.getARActionSelectionTemp( netIdx, ruleType );
       actionDecider = new StochasticDecider( temper );
       // used for the key to TreeMap.
       multipliers = new int[ OUTPUT_DIM_NUM ];
       multipliers[ OUTPUT_DIM_NUM-1 ] = 1;
       for( int i = OUTPUT_DIM_NUM-2; i >= 0; i-- )
            multipliers[i] = multipliers[i+1] * outputDVs[i+1];
   }

   public String[] getRules()
   {
      return ruleReps;
   }

   public String getAllRules()
   {
      return ruleReps[ ALL ];
   }

   // matching rules ( only matching condition ) only from the active rules.
   public String getMatchConditionRules()
   {
      return ruleReps[ MATCH_CONDITION ];
   }

   public String getMatchRules()
   {
      return ruleReps[ MATCH_BOTH ];
   }

   public double getPM()
   {
      sumPM = 0.0;
      for( int i = 0; i < RULE_NUM; i++ )
           if( ruleList[ i ] != null )
               sumPM += ruleList[ i ].getPM();
      return sumPM;
   }

   public double getNM()
   {
      sumNM = 0.0;
      for( int i = 0; i < RULE_NUM; i++ )
           if( ruleList[ i ] != null )
               sumNM += ruleList[ i ].getNM();
      return sumNM;
   }

   public double getSuccRate()
   {
      getPM();
      getNM();
      succRate = ( SUCC_C3 + sumPM ) / ( SUCC_C4 + sumPM + sumNM );
      return succRate;
   }

   // get all the activated actions' activations.
   public double[] getOutputActivations()
   {
      return outputActivations;
   }

   // get all the activated actions' activations.
   public void getOutputActivations( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           arr[ i ] = outputActivations[ i ];
   }

   public Chunk[] getAllConclusions()
   {
      tlResults = fillInFinalResults();
      return tlResults;
   }

   public int getResponseTime( int TLAT )
   {
      calcRT( TLAT );
      return responseTime;
   }

   public void setChosenAction( short[] action )
   {
      for( int i = 0; i < OUTPUT_DIM_NUM; i++ )
           chosenAction[i] = action[i];
   }

   public void setStates( double[] state, double[] exState, double[] gsState, double[] wmState )
   {
      if( this.state == null )
          this.state = new short[ state.length ];
      for( int i = 0; i < state.length; i++ )
           this.state[ i ] = (short)state[ i ];
      if( this.exState == null )
          this.exState = new short[ exState.length ];
      for( int i = 0; i < exState.length; i++ )
           this.exState[ i ] = (short)exState[ i ];
      if( this.gsState == null && gsState != null ){
          this.gsState = new short[ gsState.length ];
          for( int i = 0; i < gsState.length; i++ )
               this.gsState[ i ] = (short)gsState[ i ];
      }
      if( this.wmState == null && wmState != null ){
          this.wmState = new short[ wmState.length ];
          for( int i = 0; i < wmState.length; i++ )
               this.wmState[ i ] = (short)wmState[ i ];
      }
   }

   public void setCurrentTime( int curTime )
   {
      currentTime = curTime;
   }

   // get recommended action from current rule set.
   public short[] decideAction()
   {
      if( ruleMatchNum == 0 ){
          short[] action = new short[ OUTPUT_DIM_NUM ];
          for( int i = 0; i < action.length; i++ )
               action[i] = Global.EMPTY;
          return action;
      }
      int recommendIdx = 0;
      recommendIdx = actionDecider.getRecommendation( ruleUtls, ruleMatchNum );
      recommendRule = recommendRules[ recommendIdx ];
      short[] recommendAction = new short[ OUTPUT_DIM_NUM ];
      ruleList[recommendRule].getAction( recommendAction );
      return recommendAction;
   }

   public void reasoning()
   {
      ruleMatchNum = 0;
      activationList.clear();
      for( int i = 0; i < OUTPUT_NUM; i++ )
           outputActivations[ i ] = 0.0;
      for( int i = 0; i < RULE_NUM; i++ ){
           ruleUtls[ i ] = 0.0;
           recommendRules[ i ] = -1;
      }
      ruleReps[ ALL ] = "All rules: \n";
      ruleReps[ MATCH_CONDITION ] = "Match condition rules: \n";
      for( int i = 0; i < RULE_NUM; i++ ){
           if( ruleList[i] == null )
               continue;
           boolean matchFlag = ruleList[i].decideAction( state, exState, gsState, wmState );
           ruleReps[ ALL ] += ruleList[i];
           if( matchFlag ){
               ruleReps[ MATCH_CONDITION ] += ruleList[i];
               recommendRules[ ruleMatchNum ] = (short)i;
               if( UTL_FLAG )
                   ruleUtls[ ruleMatchNum ] = ruleList[i].getUtility();
               else
                   ruleUtls[ ruleMatchNum ] = 1;
               ruleMatchNum++;
               updateActionActivations( ruleList[i] );
          }
          else{
              if( partialMatchFlag ) // if partial matched, needs further revision.
                  updateActionActivations( ruleList[i] );
          }
      }
      fillInOutputActivation();
   }

   public void update( boolean successFlag )
   {
      ruleReps[ MATCH_BOTH ] = "Match both condition and action rules: \n";
      for( int i = 0; i < ruleMatchNum; i++ ){
           int idx = recommendRules[i];
           boolean actionMatchFlag = ruleList[ idx ].isActionMatched( chosenAction );
           if( actionMatchFlag ){
               ruleReps[ MATCH_BOTH ] += ruleList[ idx ];
               updateActionUsage();
           }
           ruleList[idx].update( actionMatchFlag, successFlag, currentTime );
      }
   }

   public void update( double reward )
   {
   }

   public void calcRT( int TLAT )
   {
      responseTime = 1500;
      if( actionBLAoption == Global.EMPTY || ruleMatchNum == 0 )
          return;
      double actionBla = calcChosenActionBLA();
      for( int i = 0; i < ruleMatchNum; i++ ){
           int idx = recommendRules[i];
           if( ruleList[ idx ].isActionMatched( chosenAction ) ){
               ruleList[idx].calcBLA( currentTime );
               ruleList[idx].calcTLDT( actionBla );
               ruleList[idx].calcRT( TLAT );
               if( responseTime < ruleList[idx].getResponseTime() )
                   responseTime = ruleList[idx].getResponseTime();
           }
      }
  }

   public void updateActionUsage()
   {
       if( actionBLAoption != Global.VARIED_OPTION )
           return;
      // update action BLAs.
      KeyToValue key = new KeyToValue( chosenAction, multipliers );
      ActionUsageInfo value = (ActionUsageInfo)actionUsageInfo.get(key);
      value.increaseUsage( currentTime );
   }

/****************************** PRIVATE METHODS ********************************/
  // calculate the BLA of currently chosen action, used for RT.
  private double calcChosenActionBLA()
  {
      if( actionBLAoption == global.CONSTANT_OPTION )
          return global.getAcsChunkBla( netIdx, ruleType, 0 );
      KeyToValue key = new KeyToValue( chosenAction, multipliers );
      if( !actionUsageInfo.containsKey( key ) )
          actionUsageInfo.put(key,new ActionUsageInfo(currentTime,global.getAcsChunkBlaParas(netIdx,ruleType,0)));
      ActionUsageInfo value = (ActionUsageInfo)actionUsageInfo.get(key);
      value.calcBla( currentTime );
      return value.getBla();
  }

  private void updateActionActivations( TLRule rule )
  {
     double supportThreshold = 0.5;
     double support = rule.getRuleSupport();
     if( support < supportThreshold )
         return;
     short[] action = new short[ OUTPUT_DIM_NUM ];
     rule.getAction( action );

     if( action[0] == Global.EMPTY )
         return;

     KeyToValue key = new KeyToValue( action, multipliers );
     if( activationList.containsKey( key ) ){
         double value = ((Double)activationList.get( key )).doubleValue();
         value = Math.max( support, value );
         activationList.put(key, new Double(value));
     }
     else
         activationList.put(key, new Double(support));
  }

  private void fillInOutputActivation()
  {
     Object[] keys = activationList.keySet().toArray();
     Object[] values = activationList.values().toArray();
     int actionNum = keys.length;
     for( int i = 0; i < actionNum; i++ ){
          short[] action = ((KeyToValue)keys[i]).getKeys();
          for( int j = 0; j < OUTPUT_DIM_NUM; j++ ){
               int idx = outputOffsets[j] + action[j];

               if( idx >= outputActivations.length || idx < 0 ){
                   System.out.println( "action = " + action[j] + ", idx = " + idx );
               }

               outputActivations[idx] += Math.max(outputActivations[idx], ((Double)values[i]).doubleValue());
         }
     }
  }

  // fill in the most activated conclusions from top level.
  private Chunk[] fillInFinalResults()
  {
     int resultNum = ruleMatchNum;
     double[] strengths = new double[ resultNum ];
     for( int i = 0; i < resultNum; i++ )
          strengths[i] = ruleList[recommendRules[i]].getRuleSupport();
     SortableElement[] sortableElements = new SortableElement[ resultNum ];
     for( int i = 0; i < resultNum; i++ )
          sortableElements[i] = new SortableElement( strengths[i], recommendRules[i] );
     Arrays.sort( sortableElements );
     int c = Math.min( c20, resultNum );
     int len = (global.isMcsOn())? c : resultNum;
     Chunk[] results = new Chunk[ len ];
     try{
        for( int i = 0; i < len; i++ ){
             int idx = sortableElements[resultNum-i-1].getChunkIdx();
             results[i] = (Chunk)(ruleList[idx].getActionChunk().clone());
        }
     }
     catch( Exception e ){
            System.out.println( e + ", clone chunk error in TLRuleSet.java." );
     }
     return results;
  }
}