package clarion.acs.wm;
import  clarion.system.*;
import  clarion.nacs.*;

public class WM{

  Global global;          // global settings.
  int    netIdx;          // net index.
  int    size_WM;         // size of WM.
  double threshold_WM;    // activation threshold.
  int    blaOption;       // BLA option: constant or recency-based.
  double blaParams[];     // three parameters for BLA: c, d, iBLA.
  long   currentTime;     // used for calculating BLA.
  short  wmItemLengths[]; // the length of each item.
  int    offsets[];       // start points of each item in input.
  GKSChunk items[];       // items.

  public WM( Global g )
  {
      global = g;
      netIdx = global.getWmNetIdx();
      size_WM = global.getWmSize();
      threshold_WM = global.getWmActivationThreshold();
      blaOption = global.getWmBlaOption();
      blaParams = global.getWmBlaParas();
      wmItemLengths = global.getWmItemLengths();
      offsets = new int[ size_WM ];
      offsets[ 0 ] = 0;
      for( int i = 1; i < size_WM; i++ )
           offsets[ i ] = offsets[ i-1 ] + wmItemLengths[ i-1 ];
      items = new GKSChunk[ size_WM ];
      global.setWM( this );
  }

  public int getSize()
  {
      return size_WM;
  }

  public int getDoNothingAction()
  {
     return 2 * size_WM + 1;
  }

  public int getWmLength()
  {
      int wmLength = 0;
      for( int i = 0; i < items.length; i++ )
           wmLength += items[i].getLength();
      return wmLength;
  }

  public GKSChunk[] getItems()
  {
      GKSChunk[] wmItems = new GKSChunk[size_WM];
      try{
         for( int i = 0; i < size_WM; i++ )
              wmItems[i] = (items[i]==null) ? null : (GKSChunk)(items[i].clone());
      }
      catch(  CloneNotSupportedException e ) {
         System.out.println( "clone error from GKSChunk in WM." );
      }
      return wmItems;
  }

  public void setCurrentTime( long time )
  {
      currentTime = time;
  }

/************************ Working Memory actions ******************************/
  public boolean actionSet( int idx, GKSChunk chunk )
  {
      if( items[ idx ] != null ){
         if( items[ idx ].equals( chunk ) ){
             if( items[idx].getBlaOption() == Global.VARIED_OPTION )
                 items[ idx ].increaseUsage( currentTime );
             return false;
         }
      }
      try{
             items[ idx ] = (GKSChunk)chunk.clone();
             if( items[idx].getBlaOption() == Global.VARIED_OPTION ){
                 items[ idx ].setBlaParams( blaParams );
                 items[ idx ].resetUsage();
                 items[ idx ].increaseUsage( currentTime );
             }
      }
      catch( CloneNotSupportedException e ){
         System.out.println( "clone error!" );
      }
      return true;
  }

  public boolean actionReset( int idx )
  {
      if( items[ idx ] == null )
          return false;
      items[ idx ] = null;
      return true;
  }

  public boolean actionResetAll()
  {
      boolean retVal = false;
      for( int i = 0; i < size_WM; i++ )
           if( items[ i ] != null ){
               actionReset( i );
               retVal = true;
           }
      return retVal;
  }

  public boolean actionNo()
  {
     return false;
  }

  // get current WM state as input to WMNet.
  public void updateState( double[] state )
  {
     for( int i = 0; i < size_WM; i++ ){
          if( items[i] == null ){
              for( int j = 0; j < wmItemLengths[i]; j++ )
                   state[ offsets[i] + j ] = 0.0;
          }
          else{
              items[i].calcBLA( currentTime );
              if( items[i].getBLA() < threshold_WM )   // input with strengh 0
                  for( int j = 0; j < wmItemLengths[i]; j++ )
                       state[ offsets[i] + j ] = 0.0;
              else{  // input with strengh 1
                  short[] content = new short[ wmItemLengths[i] ];
                  items[i].getContent( content );
                  for( int j = 0; j < wmItemLengths[i]; j++ )
                       state[ offsets[i] + j ] = content[ j ];
              }
          }
     }
  }
}