package clarion.acs.wm;
import clarion.acs.net.*;
import clarion.system.*;
import clarion.nacs.*;

public class WMNet extends AcsNet {

  WM wm;    // working memory.
  GKSChunk retrievalResults[];   // retrieval results from NACS.

  public WMNet( int nIdx, Global g )
  {
     super( nIdx, g );
     wm = new WM( global );
  }

  public void setRetrievalResult( GKSChunk[] result, int slot )
  {
      retrievalResults = result;
      GKSChunk chunk = selectRetrievalResult();
      boolean updateFlag = false;
      if( chunk != null )
          updateFlag = wm.actionSet( slot, chunk );
      // update current state.
      if( updateFlag )
          wm.updateState( wmState );
  }

  // reasoning for determining an action to perform given current input.
  public void reasoning()
  {
     // get active net comps.
     if( global.getWmActionDecisionMode() == Global.ACTION_BY_TL ){
         for( int i = 0; i < COMP_NUM; i++ ){
              if( netCompFlags[ i ] ){
                  netComps[ i ].setStates( state, exState, gsState, wmState );
                  netComps[ i ].reasoning();
              }
         }
         // select which part: BL, RER, IRL or FR to decide action.
         chosenComp = Global.FR;
         // decide an action.
         decideAction( chosenComp );
         // set the chosen action.
         deliverAction();
     }
     else
         super.reasoning();
  }

  public void perform()
  {

     //System.out.println( "WMNet.java, performed action = " + chosenAction[0] );

     int slot = 0;
     boolean updateFlag = false;
     if( chosenAction[0] < wm.getSize() )    // a set action.
     {
         slot = chosenAction[0];
         GKSChunk chunk = selectRetrievalResult();
         if( chunk != null )
             updateFlag = wm.actionSet( slot, chunk );
     }
     else if( chosenAction[0] < 2 * wm.getSize() )   // a reset action.
          {
              slot = chosenAction[0] - wm.getSize();
              updateFlag = wm.actionReset( slot );
          }
          else if( chosenAction[0] == 2 * wm.getSize() )  // reset-all action
                   updateFlag = wm.actionResetAll();
               else
                   updateFlag = wm.actionNo();        // do-nothing action.
     // update current state.
     if( updateFlag )
         wm.updateState( wmState );
  }

  private GKSChunk selectRetrievalResult()
  {
     if( retrievalResults == null )
         return null;
     else{
         int idx = (int)(Math.random() * retrievalResults.length);
         return retrievalResults[ idx ];
     }
  }
}