// largely based on Chris Terry's CLARION code
package clarion.common;
import clarion.system.*;
import clarion.tools.*;

public class BPNet{
    // global settings.
    Global  global;
    protected int nInput, nHidden, nOutput;
    protected double[] Input,Output,Hidden;
    protected double[] OutputMomentum,HiddenMomentum;
    protected double[] OutputThresholds,HiddenThresholds;
    protected double[] OutputDeriv,HiddenDeriv;
    protected double[] Errors, HiddenErrors;
    protected double[] DesiredOutput;
    protected double[][] InputToHiddenWeights, HiddenToOutputWeights;
    protected double[][] ItoHWeights, HtoOWeights;   // used for restoration.
    protected double[][] InputToHiddenMomentum, HiddenToOutputMomentum;
    protected double  HINITWEIGHT = 0.05;    // brackets for the initial weights
    protected double  LINITWEIGHT = -0.05;
    protected double  HINITTHRESHOLD = 1;    // brackets for the initial weights
    protected double  LINITTHRESHOLD = -1;
    protected double Momentum;     // used for update of weights.
    protected double Eta;          // learning rate.
    protected double RZero;        // tolerance of error...sorta
    protected boolean debug;
    protected double sumSqErrors;
    // variables for RT.
    protected int    BLRT, BLPT, BLDT, BLAT;
    // variables for success rate.
    protected double PM, NM, succRate;
    protected double SUCC_C3 = 1.0, SUCC_C4 = 2.0;
    // variables for dimension values.
    protected short[] inputDVs, outputDVs;
    // the start index of each action's dim in net output.
    protected short[] outputOffsets;

    // old constructor.
    public BPNet(int numInput, int numHidden, int numOutput )
    {
        // # of units in input, output, hidden layers
        nInput = numInput;
        nOutput = numOutput;
        nHidden = numHidden;
        // value of input, hidden and output vectors
        Input = new double[nInput];
        Hidden = new double[nHidden];
        Output = new double[nOutput];
        // weight vectors
        InputToHiddenWeights = new double[nInput][nHidden];
        HiddenToOutputWeights = new double[nHidden][nOutput];
        ItoHWeights = new double[nInput][nHidden];
        HtoOWeights = new double[nHidden][nOutput];
        // momentum vectors
        InputToHiddenMomentum = new double[nInput][nHidden];
        HiddenToOutputMomentum = new double[nHidden][nOutput];
        HiddenMomentum = new double[nHidden];
        OutputMomentum = new double[nOutput];
        // error terms in hidden and output layers
        HiddenErrors = new double[nHidden];
       	Errors = new double[nOutput];
        // bias in hidden and output layers
        HiddenThresholds = new double[nHidden];
        OutputThresholds = new double[nOutput];
        // derivative values of activation functions in
        // hidden and output layers
        HiddenDeriv = new double[nHidden];
        OutputDeriv = new double[nOutput];
        // target output
        DesiredOutput = new double[nOutput];
        Eta = World.TASK_PARA_ARRAY[ World.NET_LRATE ];
        Momentum = World.TASK_PARA_ARRAY[ World.NET_MOMENTUM ];
        // initialize weights.
        initWeights( 0 );
    }

    // new constructor.
    public BPNet( int inputNum, int netIdx, Global global, int netType )
    {
        nInput = inputNum;
        this.global = global;
        if( netType == Global.ACS ){
            //nInput = global.getIdnInputNum( netIdx );
            nHidden = global.getIdnHiddenNum( netIdx );
            nOutput = global.getIdnOutputNum( netIdx );
            Eta = global.getIdnBPLearningRate( netIdx );
            Momentum = global.getIdnBPMomentum( netIdx );
            inputDVs = global.getIdnInputDimVals( netIdx );
            outputDVs = global.getIdnOutputDimVals( netIdx );
            if( global.getAcsCrossCombineMode(netIdx, 1 ) == Global.VAR_COMB_MODE ){
                SUCC_C3 = global.getAcsCrossCombineVarParaAbove( netIdx )[Global.BL];
                SUCC_C4 = global.getAcsCrossCombineVarParaAbove( netIdx )[Global.BL];
            }
            if( global.isRTOn() ){
                BLPT = global.getPT_BL();
                BLAT = global.getAT_BL();
                BLDT = global.getDT_BL();
            }
        }
        else{  // AMN of NACS
            //nInput = global.getAmnInputNum( netIdx );
            nHidden = global.getAmnHiddenNum( netIdx );
            nOutput = global.getAmnOutputNum( netIdx );
            Eta = global.getAmnBPLearningRate( netIdx );
            Momentum = global.getAmnBPMomentum( netIdx );
            inputDVs = global.getAmnInputDimVals( netIdx );
            outputDVs = global.getAmnOutputDimVals( netIdx );
            if( global.isRTOn() )   // need be updated.
            {
                BLPT = global.getPT_BL();
                BLAT = global.getAT_BL();
                BLDT = global.getDT_BL();
            }
        }
        // start points of each dim in 1-dim state.
        outputOffsets = new short[ outputDVs.length ];
        //outputOffsets[ 0 ] = 0;
        //for( int i = 1; i < outputDVs.length; i++ )
             //outputOffsets[i] = (short)(outputOffsets[i-1] + outputDVs[i-1]);
        CommonMethods.calcOffsets( outputOffsets, outputDVs );
        // weight vectors
        InputToHiddenWeights = new double[nInput][nHidden];
        HiddenToOutputWeights = new double[nHidden][nOutput];
        ItoHWeights = new double[nInput][nHidden];
        HtoOWeights = new double[nHidden][nOutput];
        // momentum vectors
        InputToHiddenMomentum = new double[nInput][nHidden];
        HiddenToOutputMomentum = new double[nHidden][nOutput];
        HiddenMomentum = new double[nHidden];
        OutputMomentum = new double[nOutput];
        // value of input, hidden and output vectors
        Input = new double[nInput];
        Hidden = new double[nHidden];
        Output = new double[nOutput];
        // bias in hidden and output layers
        HiddenThresholds = new double[nHidden];
        OutputThresholds = new double[nOutput];
        // derivative values of activation functions in
        // hidden and output layers
        HiddenDeriv = new double[nHidden];
        OutputDeriv = new double[nOutput];
        // error terms in hidden and output layers
        HiddenErrors = new double[nHidden];
       	Errors = new double[nOutput];
        // desired output
	       DesiredOutput = new double[nOutput];
        // initialize weights.
       	initWeights( 0 );
    }

    public void reinitWeights()
    {
       initWeights( 0 );
    }

    public void restoreInitWeights()
    {
       initWeights( 1 );
    }

/***************************** GET AND SET METHODS ****************************/
    public int getnInput()
    {
       return nInput;
    }

    public int getnHidden()
    {
       return nHidden;
    }

    public int getnOutput()
    {
      	return nOutput;
    }

    public void setLearningRate( double value )
    {
        Eta = value;
    }

    public void setMomentum( double value )
    {
        Momentum = value;
    }

    // utility methods
    public void getInput( double[] fillMe )
    {
        for( int i = 0; i < nInput; i++ )
             fillMe[i] = Input[i];
    }

    public double[] getOutput()
    {
        return Output;
    }

    public void getOutput( double[] out )
    {
       	for( int i = 0; i < nOutput; i++ )
             out[ i ] = Output[i];
    }

    public double getOutput( int idx )
    {
       return Output[ idx ];
    }

    public double getOutput( int dimIdx, int offset )
    {
       if( outputOffsets[ dimIdx ] + offset >= Output.length ||
           outputOffsets[ dimIdx ] + offset < 0 )
           System.out.println( "dim offset = " + outputOffsets[ dimIdx ] + ", offset = " + offset +
                               ", output len = " + Output.length );

       return Output[ outputOffsets[ dimIdx ] + offset ];
    }

    public void setInput( int[] newInputs )
    {
       	for( int i = 0; i < nInput; i++ )
             Input[i] = newInputs[i];
    }

    public void setInput( double[] newInputs )
    {
       	for( int i = 0; i < nInput; i++ )
             Input[i] = newInputs[i];
    }

    public void setInput( int index, double value )
    {
       	Input[index] = value;
    }

    public void setDesiredOutput( double[] dOutputs )
    {
       	for( int i = 0; i < nOutput; i++ )
             DesiredOutput[i] = dOutputs[i];
    }

    public void setDesiredOutput( int index, double value )
    {
       	DesiredOutput[index] = value;
    }

    public void setDesiredOutput( short[] action )
    {
        for( int i = 0; i < outputDVs.length; i++ )
          for( int j = 0; j < outputDVs[i]; j++ )
               DesiredOutput[ outputOffsets[i] + j ] = (j==action[i]) ? 1 : 0;
    }

    public void setDesiredOutput( Feature[] output )
    {
        for( int i = 0; i < outputDVs.length; i++ )
          for( int j = 0; j < outputDVs[i]; j++ )
               DesiredOutput[ outputOffsets[i] + j ] = output[i].getActivations()[j];
    }

    public double getSumSqErrors()
    {
       	return sumSqErrors;
    }

    public double getPM()
    {
        return PM;
    }

    public double getNM()
    {
        return NM;
    }

    public double getSuccRate()
    {
       succRate = ( SUCC_C3 + PM ) / ( SUCC_C4 + PM + NM );
       return succRate;
    }

    public int getResponseTime()
    {
       return BLRT;
    }

    public void resetMatches()
    {
        PM = 0.0;
        NM = 0.0;
    }

    public void updateMatches( boolean flag )
    {
        if( flag )
            PM += 1.0;
        else
            NM += 1.0;
    }

    public void forwardPass()
    {
       	computeHiddenActivation();
        computeOutputActivation();
    }

    public void backwardPass()
    {
       	computeErrors();
        computeHiddenErrors();
        modifyHiddenToOutput();
        modifyInputToHidden();
    }

    public void calcRT( int BLAT )
    {
        this.BLAT = BLAT;
        BLRT = global.getTaskClarion().getBLOverallRT( BLPT, BLDT, BLAT );
    }

/************************** PROTECTED METHODS *********************************/
    /********************* forward pass methods **************************/
    protected void computeHiddenActivation()
    {
       	double sum,v;
       	for (int j = 0; j < nHidden; j++)
        {
            		sum = HiddenThresholds[j];
            		for ( int i = 0; i < nInput; i++ )
	             sum += Input [i] * InputToHiddenWeights[i][j];
            		v = activation( sum );
              Hidden[j] = v;
            		HiddenDeriv[j] = derivative(v);
        }
    }

    protected void computeOutputActivation ()
    {
       	double sum,v;
       	for (int j = 0; j < nOutput; j++)
        {
            		sum = OutputThresholds[j];
            		for( int i = 0; i < nHidden; i++ )
     	             sum += Hidden[i] * HiddenToOutputWeights[i][j];
            		v = activation(sum);
            		Output[j] = v;
            		OutputDeriv[j] = derivative(v);
       	}
    }

    /****************** backward pass methods *****************************/
    // computer the errors of output layer
    protected void computeErrors()
    {
        sumSqErrors = 0;
       	for( int i = 0; i < nOutput; i++ ){
             double err = DesiredOutput[i] - Output[i];
             Errors[i] = err * OutputDeriv[i];
             sumSqErrors += err * err;
        }
       	sumSqErrors = sumSqErrors / 2;
        if( debug )
            System.out.println( sumSqErrors );
    }

    protected void computeHiddenErrors ()
    {
       	double h_err;
       	for ( int i = 0; i < nHidden; i++ )
        {
              h_err = 0.0;
              for ( int j = 0; j < nOutput; j++ )
     	              h_err += HiddenToOutputWeights[i][j] * Errors[j];
            		HiddenErrors[i] = h_err * HiddenDeriv[i];
        }
    }

    // modify the weights of hidden to output layer
    protected void modifyHiddenToOutput ()
    {
        double x;
        for( int j = 0; j < nOutput; j++ ) {
      	    if( Math.abs( Errors[j] ) > RZero ) {
               		x = Errors[j];
                // udpate HiddenToOutputWeights
                for ( int i = 0; i < nHidden; i++ ){
        	             if( Momentum > RZero ){
                        			/* momentum */
                       			HiddenToOutputMomentum[i][j] = Eta * x * Hidden[i] +
                                          Momentum * HiddenToOutputMomentum[i][j];
                       			HiddenToOutputWeights[i][j] += HiddenToOutputMomentum[i][j];
               		     } else
                		       	HiddenToOutputWeights[i][j] += Eta * x * Hidden[i];
                }

                // update OutputThresholds
                if ( Momentum > RZero ){
                     OutputMomentum[j] =  Eta * x + Momentum * OutputMomentum[j];
                     OutputThresholds[j] += OutputMomentum[j];
                }
                else
                		    OutputThresholds[j] += Eta * x;
      	    }
       	}
    }

    // modify the weights of input to hidden layer
    protected void modifyInputToHidden ()
    {
       	double x;
       	for( int j = 0; j < nHidden; j++ ){
         	   if( Math.abs( HiddenErrors[j] ) > RZero ){
                		x = HiddenErrors[j];
                  // update the InputToHiddenWeights
	               	 for (int i = 0; i < nInput; i++){
                 		    if( Momentum > RZero ){
                         		InputToHiddenMomentum[i][j] = Eta * x* Input[i]
                                       + Momentum * InputToHiddenMomentum[i][j];
                        			InputToHiddenWeights[i][j] += InputToHiddenMomentum[i][j];
		                     }
                       else
                         		InputToHiddenWeights[i][j] += Eta * x * Input[i];
                		}
                  // upate the HiddenThresholds
                		if(Momentum > RZero){
               		    HiddenMomentum[j] = Eta * x + Momentum * HiddenMomentum[j];
               		    HiddenThresholds[j] += HiddenMomentum[j];
                		}
                  else
            	   	    HiddenThresholds[j] += Eta * x;
             }
     	 }
    }

/****************************** PRIVATE METHODS *******************************/
    private void initWeights( int mode )
    {
       for(int i = 0; i < nInput; i++)
           for (int j = 0; j < nHidden; j++)
      	InputToHiddenMomentum[i][j] = Momentum;

      	for (int i = 0; i < nHidden; i++){
	      //  HiddenThresholds[i] = randomThresholds();
            HiddenThresholds[i] = 1;
	          	HiddenMomentum[i] = 0;
      	}

       for(int i = 0; i < nHidden; i++)
        for(int j = 0; j < nOutput; j++)
            HiddenToOutputMomentum[i][j] = Momentum;

      	for(int i = 0; i < nOutput; i++){
      	   //OutputThresholds[i] = randomThresholds();
           OutputThresholds[i] = 1;
       	   OutputMomentum[i] = 0;
       	}

        if( mode == 0 ){
       	    for( int i = 0; i < nInput; i++ )
             	for( int j = 0; j < nHidden; j++ ){
                   InputToHiddenWeights[i][j] = randomWeights();
                   ItoHWeights[i][j] = InputToHiddenWeights[i][j];
               }
            for (int i = 0; i < nHidden; i++)
                for (int j = 0; j < nOutput; j++){
              		     HiddenToOutputWeights[i][j] = randomWeights();
                     HtoOWeights[i][j] = HiddenToOutputWeights[i][j];
                }

        }
        else
        {
       	    for( int i = 0; i < nInput; i++ )
            	for( int j = 0; j < nHidden; j++ )
      	   	       InputToHiddenWeights[i][j] = ItoHWeights[i][j];
            for(int i = 0; i < nHidden; i++)
              for(int j = 0; j < nOutput; j++)
                  HiddenToOutputWeights[i][j] = HtoOWeights[i][j];
        }
    }

    private double randomWeights()
    {
       	double sign,value;
        if ( Math.random() < 0.5 )
             sign = LINITWEIGHT;
        else
             sign = HINITWEIGHT;

        value = Math.random();
        return( value * sign );
    }

    private double randomThresholds()
    {
       	double sign,value;
        if ( Math.random() < 0.5 )
             sign = LINITTHRESHOLD;
        else
             sign = HINITTHRESHOLD;

        value = Math.random();
        return( value * sign );
    }

    // activation function
    private double activation(double value)
    {
       	// can switch other activation functions here
       	// simple sigmoid function
       	return (1.0 / (1.0 + Math.exp( -value)));
    }

    // derivative of the activation function
    private double derivative( double value )
    {
       	return( value * ( 1 - value ) );
    }
}
