package clarion.common;
import java.util.Arrays;
import clarion.system.*;

public class Chunk implements Cloneable {

   protected Global global;       // gloabl settings.
   // variables for chunk.
   protected Feature[] featList;
   protected int    chunkType;      // different types of chunks.
   protected String chunkLabel;     // chunk's label.
   // variables for BLA.
   protected int    chunkBlaOption;  // constent or varied.
   protected int    MAX_USAGE_NUM = 1000;  // used for chunk usage.
   protected int    usageNum;       // the # of using the chunk.
   protected long   usageRecords[]; // record the times when the chunk was used.
   protected double c, d, iBLA;     // three parameters for BLA: c, d, iBLA.
   protected double BLA;            // base level activation.
   protected double strength;       // chunk strength used for conclusion chunks.

   public Chunk( int type, String label, Feature[] list, Global g )
   {
       chunkType = type;
       chunkLabel = label;
       global = g;
       chunkBlaOption = global.getChunkBlaOption();
       if( chunkBlaOption == global.CONSTANT_OPTION )
           BLA = global.getChunkBla();
       else{
           c = global.getChunkBlaParas()[ global.BLA_C ];
           d = global.getChunkBlaParas()[ global.BLA_D ];
           iBLA = global.getChunkBlaParas()[ global.BLA_I ];
           usageRecords = new long[ MAX_USAGE_NUM ];
       }
       featList = new Feature[ list.length ];
       for( int i = 0; i < list.length; i++ )
            featList[ i ] = (Feature)(list[ i ].clone());
   }

   public Chunk( int type, String label, Chunk[] chunks, Global g )
   {
       chunkType = type;
       chunkLabel = label;
       global = g;
       chunkBlaOption = global.getChunkBlaOption();
       if( chunkBlaOption == global.CONSTANT_OPTION )
           BLA = global.getChunkBla();
       else{
           c = global.getChunkBlaParas()[ global.BLA_C ];
           d = global.getChunkBlaParas()[ global.BLA_D ];
           iBLA = global.getChunkBlaParas()[ global.BLA_I ];
           usageRecords = new long[ MAX_USAGE_NUM ];
       }
       int featLen = 0;
       for( int i = 0; i < chunks.length; i++ )
            featLen += chunks[i].getDimNum();
       featList = new Feature[ featLen ];
       int count = 0;
       for( int i = 0; i < chunks.length; i++ )
         for( int j = 0; j < chunks[i].getDimNum(); j++ )
              featList[ count++ ] = (Feature)(chunks[i].getFeature(j).clone());
   }

   public void append( int type, String label, Chunk[] chunks )
   {
       chunkType = type;
       chunkLabel = label;
       chunkBlaOption = global.getChunkBlaOption();
       if( chunkBlaOption == global.CONSTANT_OPTION )
           BLA = global.getChunkBla();
       else{
           c = global.getChunkBlaParas()[ global.BLA_C ];
           d = global.getChunkBlaParas()[ global.BLA_D ];
           iBLA = global.getChunkBlaParas()[ global.BLA_I ];
           usageRecords = new long[ MAX_USAGE_NUM ];
       }
       Feature[] list = new Feature[ featList.length ];
       for( int i = 0; i < featList.length; i++ )
            list[ i ] = (Feature)(featList[ i ].clone());
       int featLen = featList.length;
       for( int i = 0; i < chunks.length; i++ )
            featLen += chunks[i].getDimNum();
       featList = new Feature[ featLen ];
       int count = 0;
       for( int i = 0; i < list.length; i++ )
            featList[ count++ ] = (Feature)(list[ i ].clone());
       for( int i = 0; i < chunks.length; i++ )
         for( int j = 0; j < chunks[i].getDimNum(); j++ )
              featList[ count++ ] = (Feature)(chunks[i].getFeature(j).clone());
   }

   public void reinit()
   {
       resetUsage();
   }

   public int getChunkType()
   {
       return chunkType;
   }

   public int getDimNum()
   {
       return featList.length;
   }

   public String getLabel( )
   {
       return chunkLabel;
   }

   public int getLength()
   {
      int length = 0;
      for( int i = 0; i < featList.length; i++ )
           length += featList[i].getLength();
      return length;
   }

   public int getBlaOption()
   {
       return chunkBlaOption;
   }

   public int getUsageNum()
   {
       return usageNum;
   }

   public long[] getUsageRecords()
   {
       return usageRecords;
   }

   public void getUsageRecords( long[] arr )
   {
       for( int i = 0; i < arr.length; i++ )
            arr[ i ] = usageRecords[ i ];
   }

   public double getBLA()
   {
       return BLA;
   }

   public double getStrength()
   {
       return strength;
   }

   public int getFeatureNum()
   {
       return featList.length;
   }

   public Feature[] getFeatures()
   {
       return featList;
   }

   public void getFeatures( Feature[] other )
   {
       other = new Feature[ featList.length ];
       for( int i = 0; i < featList.length; i++ )
            other[ i ] = (Feature)(featList[ i ].clone());
   }

   public Feature getFeature( int idx )
   {
       return (Feature)(featList[ idx ].clone());
   }

   public void getContent( short[] arr )
   {
      int offset = 0;
      for( int i = 0; i < featList.length; i++ )
      {
           short[] content = new short[ featList[i].getLength() ];
           featList[i].getContent( content );
           for( int j = 0; j < content.length; j++ )
                arr[ offset + j ] = content[ j ];
           offset += content.length;
      }
   }

   public void getContent( double[] arr )
   {
      int offset = 0;
      for( int i = 0; i < featList.length; i++ )
      {
           short[] content = new short[ featList[i].getLength() ];
           featList[i].getContent( content );
           for( int j = 0; j < content.length; j++ )
                arr[ offset + j ] = content[ j ];
           offset += content.length;
      }
   }

   public void setLabel( String str )
   {
       chunkLabel = str;
   }

   public void setBlaOption( int option )
   {
       chunkBlaOption = option;
   }

   // used for constent BLA option.
   public void setBla( double bla )
   {
      BLA = bla;
   }

   public void setBLA_C( double val )
   {
       c = val;
   }

   public void setBLA_D( double val )
   {
       d = val;
   }

   public void setBLA_INIT( double val )
   {
       iBLA = val;
   }

   public void setBlaParams( double[] blaParams )
   {
       c = blaParams[ global.BLA_C ];
       d = blaParams[ global.BLA_D ];
       iBLA = blaParams[ global.BLA_I ];
   }

   public void setUsageNum( int num )
   {
       usageNum = (short)num;
   }

   public void setUsageRecords( long[] records )
   {
       for( int i = 0; i < usageRecords.length; i++ )
            usageRecords[ i ] = records[ i ];
   }

   public void setStrength( double val )
   {
       strength = val;
   }

   public void increaseUsage( long currentTime )
   {
       if( usageNum > 0 && currentTime == usageRecords[ usageNum-1 ] ){
           //System.out.println("usage has already update in Chunk.java");
           return;
       }
       if( usageNum < MAX_USAGE_NUM-1 )
           usageRecords[ usageNum++ ] = currentTime;
       else
           System.out.println("error usage update in Chunk.java");
   }

   public void resetUsage()
   {
       usageNum = 0;
       for( int i = 0; i < usageRecords.length; i++ )
            usageRecords[ i ] = 0;
   }

   public void calcBLA( long currentTime )
   {
       if( chunkBlaOption == global.CONSTANT_OPTION )
           return;
       double val = iBLA;
       for( int j = 0; j < usageNum; j++ ){
            long timeLaps = currentTime - usageRecords[j];
            if( timeLaps <= 0 ){
                System.out.println( "error usage recording in Chunk.java" );
                break;
            }
            val += c * Math.pow( timeLaps, -d );
       }
       BLA = val;
       if( BLA < 0.01 )
           BLA = 0.01;
   }

   public int searchFeature( String label )
   {
       if( label == "" )
           return -1;

       int idx = -1;
       for( int i = 0; i < featList.length; i++ )
            if( label.compareTo( featList[ i ].getLabel() ) == 0 )
            {
                idx = i;
                break;
            }

      return idx;
   }

   public boolean isEqual( Feature[] features )
   {
       if( featList.length != features.length )
           return false;
       for( int i = 0; i < featList.length; i++ ) {
            if( !featList[ i ].equals( features[ i ] ) )
                return false;
       }
       return true;
   }

/************************* OVERRIDE METHODS ************************************/
    public String toString()
    {
       //String str = "Label: " + chunkLabel + ", Features: ";
       String str = "feature list = ";
       for( int i = 0; i < featList.length; i++ )
            //str += featList[ i ] + "\n";
            str += featList[ i ] + ", ";
       str += "\n";
       str += "strength = " + String.valueOf( strength ) + "\n";
       return str;
    }

    // needs be cnanged.
    public int hashCode()
    {
       int sum = 0;
       for( int i = 0; i < featList.length; i++ )
            sum += featList[ i ].hashCode();

       return sum;
    }

    public boolean equals( Object other )
    {
       if( other == null || !(other instanceof Chunk ) )
           return false;
       Chunk otherChunk = (Chunk)other;
       if( chunkType != otherChunk.getChunkType() )
           return false;
       Feature[] otherFeatures = otherChunk.getFeatures();
       return isEqual( otherFeatures );
    }

    public Object clone() throws CloneNotSupportedException
    {
       Chunk chunk = new Chunk( chunkType, chunkLabel, featList, global );
       chunk.setStrength( strength );
       chunk.setBlaOption( chunkBlaOption );
       if( chunkBlaOption == global.CONSTANT_OPTION )
           chunk.setBla( BLA );
       else{
           chunk.setBLA_C( c );
           chunk.setBLA_D( d );
           chunk.setBLA_INIT( iBLA );
           chunk.setUsageNum( usageNum );
           chunk.setUsageRecords( usageRecords );
       }
       return chunk;
    }
}