package clarion.common;
import java.lang.*;
import java.util.Vector;
import clarion.*;
import clarion.tools.*;

public class CommonChunk implements Cloneable {

    final short MAX_USAGE_NUM = 1000;  // used for chunk usage.

    // variables for identifying chunk.
    protected int chunkType;        // represent different type of chunks: visuopatial,
                                    // auditory/verbal and other information.
    protected short  CHUNK_ID;      // the chunk identification NO.
    protected String chunkLabel;    // chunk's label.
    protected Vector content;       // the content of the chunk.

    protected short DIM_NUM;        // the # of dimensions in the chunk.
    protected short dimTypes[];     // the dimension types, 0: discrete; 1:continuous.
    protected short dimValNums[];   // the # of values in each dimension.
    protected int   chunkLength;    // the length of the chunk by summarizing the length
                                    // of each dimension.
    // variables for BLA.
    protected short usageNum;       // the # of using the chunk.
    protected long  currentTime;    // current time used for calculating the BLA.
    protected long  usageRecords[]; // record the time when the chunk was used.
    protected double BLA;            // base level activation.
    protected double INIT_BLA;      // initial base level activation.
    protected double BLA_C = 2.0;   // used for calculating the BLA.
    protected double BLA_D = -0.5;  // used for calculating the BLA.

    public CommonChunk( int cId, int type, int dNum, int[] dValNums,
                        int[] dTypeInfo, double initBla )
    {
        CHUNK_ID = (short)cId;
        chunkType = type;
        DIM_NUM = (short)dNum;
        INIT_BLA = (float)initBla;

        dimValNums = new short[ DIM_NUM ];
        for( int i = 0; i < DIM_NUM; i++ )
             dimValNums[ i ] = (short)dValNums[ i ];

        dimTypes = new short[ DIM_NUM ];
        for( int i = 0; i < DIM_NUM; i++ )
             dimTypes[ i ] = (short)dTypeInfo[ i ];

        usageRecords = new long[ MAX_USAGE_NUM ];
        content = new Vector( DIM_NUM );
    }

    public CommonChunk( int cId, int type, int dNum, short[] dValNums,
                        short[] dTypeInfo, double initBla )
    {
        CHUNK_ID = (short)cId;
        chunkType = type;
        DIM_NUM = (short)dNum;
        INIT_BLA = (float)initBla;

        dimValNums = new short[ DIM_NUM ];
        for( int i = 0; i < DIM_NUM; i++ )
             dimValNums[ i ] = dValNums[ i ];

        dimTypes = new short[ DIM_NUM ];
        for( int i = 0; i < DIM_NUM; i++ )
             dimTypes[ i ] = dTypeInfo[ i ];

        usageRecords = new long[ MAX_USAGE_NUM ];
        content = new Vector( DIM_NUM );
    }

    public CommonChunk( int cId, int type, int dNum, short[] dValNums,
                        short[] dTypeInfo, double[] blaParas )
    {
        CHUNK_ID = (short)cId;
        chunkType = type;
        DIM_NUM = (short)dNum;

        // set BLA variables.
        BLA_C = (float)blaParas[ 0 ];
        BLA_D = (float)blaParas[ 1 ];
        INIT_BLA = (float)blaParas[ 2 ];


        dimValNums = new short[ DIM_NUM ];
        for( int i = 0; i < DIM_NUM; i++ )
             dimValNums[ i ] = dValNums[ i ];

        dimTypes = new short[ DIM_NUM ];
        for( int i = 0; i < DIM_NUM; i++ )
             dimTypes[ i ] = dTypeInfo[ i ];

        usageRecords = new long[ MAX_USAGE_NUM ];
        content = new Vector( DIM_NUM );
    }

    public void reinit( int type )
    {
        chunkType = type;
        usageNum = 0;
        currentTime = 0;

        for( int i = 0; i < MAX_USAGE_NUM; i++ )
             usageRecords[ i ] = 0;
    }

    public void reinit( int type, Vector con )
    {
        chunkType = type;
        usageNum = 0;
        currentTime = 0;
        chunkLabel = "";

        for( int i = 0; i < MAX_USAGE_NUM; i++ )
             usageRecords[ i ] = 0;

        content.removeAllElements();
        if( con == null )
            return;

        int len = con.size();
        content = new Vector( len );
        DIM_NUM = (short)len;
        dimValNums = new short[ len ];
        dimTypes = new short[ len ];
        for( int i = 0; i < len; i++ )
        {
             content.addElement( con.elementAt(i) );
             dimValNums[ i ] = (short)((short[])con.elementAt( i )).length;
        }
    }

    public int getChunkID( )
    {
        return CHUNK_ID;
    }

    public int getDimNum()
    {
       return DIM_NUM;
    }

    public int getChunkType()
    {
       return chunkType;
    }

    public int getUsageNum()
    {
       return usageNum;
    }

    public long getCurrentTime()
    {
       return currentTime;
    }

    public int getLength()
    {
       chunkLength = 0;
       for( int i = 0; i < DIM_NUM; i++ )
            chunkLength += dimValNums[ i ];
       return chunkLength;
    }

    public short[] getDimTypes()
    {
       return dimTypes;
    }

    public void getDimTypes( int[] arr )
    {
       for( int i = 0; i < arr.length; i++ )
            arr[ i ] = dimTypes[ i ];
    }

    public short[] getDimValNums()
    {
       return dimValNums;
    }

    public void getDimValNums( int[] arr )
    {
       for( int i = 0; i < arr.length; i++ )
            arr[ i ] = dimValNums[ i ];
    }

    public long[] getUsageRecords()
    {
       return usageRecords;
    }

    public void getUsageRecords( long[] arr )
    {
       for( int i = 0; i < arr.length; i++ )
            arr[ i ] = usageRecords[ i ];
    }

    public double getInitBLA()
    {
       return INIT_BLA;
    }

    public double getBLA( )
    {
       return BLA;
    }

    public String getLabel( )
    {
       return chunkLabel;
    }

    public Vector getContent()
    {
       return content;
    }

    public void getContent( short[][] arr )
    {
       for( int i = 0; i < arr.length; i++ )
        for( int j = 0; j < arr[ 0 ].length; j++ )
             arr[ i ][ j ] = ((short[])content.elementAt(i))[j];
    }

    public void setChunkType( int type )
    {
        chunkType = type;
    }

    public void setUsageNum( int num )
    {
        usageNum = (short)num;
    }

    public void setCurrentTime( long curTime )
    {
        currentTime = curTime;
    }

    public void setDimTypes( short[] dTypeInfo )
    {
        for( int i = 0; i < dimTypes.length; i++ )
             dimTypes[ i ] = dTypeInfo[ i ];
    }

    public void setUsageRecords( long[] records )
    {
        for( int i = 0; i < usageRecords.length; i++ )
             usageRecords[ i ] = records[ i ];
    }

    public void setLabel( String str )
    {
       chunkLabel = str;
    }

    public void setBlaParas( double[] paras )
    {
       BLA_C = paras[ 0 ];
       BLA_D = paras[ 1 ];
       INIT_BLA = (float)paras[ 2 ];
    }

    public void setContent( short[][] con )
    {
        content.removeAllElements();
        for( int i = 0; i < con.length; i++ )
             content.insertElementAt( con[ i ], i );
    }

    public void setContent( Vector con )
    {
        content.removeAllElements();
        for( int i = 0; i < con.size(); i++ )
        {
             Object obj = con.elementAt( i );
             content.insertElementAt( obj, i );
        }
    }

    public boolean isEqual( short[][] arr )
    {
       for( int i = 0; i < DIM_NUM; i++ )
            if( !CommonMethods.isEqual( arr[ i ], (short[])(content.elementAt(i)) ) )
                return false;
       return true;
    }

    public void calcBLA( )
    {
        double val = INIT_BLA;
        for( int j = 0; j < usageNum; j++ )
        {
             long timeDiff = currentTime - usageRecords[j];
             val += BLA_C * Math.pow( timeDiff, BLA_D );
        }

        BLA = (float)val;
    }

    public void increaseUsage()
    {
        usageRecords[ usageNum++ ] = currentTime;
    }

    public void increaseUsage( long time )
    {
        setCurrentTime( time );
        usageRecords[ usageNum++ ] = currentTime;
    }

    public void resetUsage()
    {
        usageNum = 0;
        for( int i = 0; i < usageRecords.length; i++ )
             usageRecords[ i ] = 0;
    }

/************************* OVERRIDE METHODS ************************************/
    public String toString()
    {
       //String str = "chunk \"" + chunkLabel + "\" = " ;
       String str = "CHUNK ID = " + String.valueOf( CHUNK_ID ) + ", ";
       str += "CHUNK TYPE = " + String.valueOf( chunkType ) + ", ";
       str += "CHUNK LABEL = " + chunkLabel + "\n";
       str += "CHUNK CONTENT = \n";

       str += "[ ";
       for( int i = 0; i < DIM_NUM; i++ )
       {
            str += "( ";
            short[] arr = (short[])content.elementAt(i);
            int count = 0;
            for( int j = 0; j < arr.length; j++ )
            {
                 if( arr[ j ] == 1 )
                 {
                     if( count == 0 )
                         str += String.valueOf( j + 1 );
                     else
                         str += ", " + String.valueOf( j + 1 );
                     count++;
                 }
             }
            str += " )";
       }
       str += " ] \n";

       return str;
    }

    // needs be cnanged.
    public int hashCode()
    {
       return 0;
    }

    public boolean equals( Object other )
    {
       if( other == null || !(other instanceof CommonChunk) )
           return false;

       CommonChunk otherChunk = (CommonChunk)other;
       if( !otherChunk.getLabel().equals( chunkLabel ) )
           return false;

       if( DIM_NUM != otherChunk.getDimNum() )
           return false;

       for( int i = 0; i < DIM_NUM; i++ )
            if( !CommonMethods.isEqual( (short[])otherChunk.getContent().elementAt( i ),
                                        (short[])(content.elementAt(i)) ) )
                return false;
       return true;
    }

    public Object clone() throws CloneNotSupportedException
    {
       CommonChunk chunk = new CommonChunk( CHUNK_ID, chunkType, DIM_NUM, dimValNums,
                                            dimTypes, INIT_BLA );
       // from this class
       chunk.reinit( chunkType, content );
       chunk.setCurrentTime( currentTime );
       chunk.setUsageNum( usageNum );
       chunk.setUsageRecords( usageRecords );
       chunk.setLabel( chunkLabel );

       return chunk;
    }
}