package clarion.common;
import java.util.Arrays;
import clarion.system.*;

public class Feature implements Cloneable {

   Global  global;      // global settings.

   String   label;       // the label of the feature.
   short    type;        // the type of the feature.
   short    order;       // order in the whole feature list.
   short    length;      // the length of this feature.
   short[]  values;      // the values this feature has.
   double[] activations; // activation of each feature value.

   public Feature( String label, short type, short order, short[] con, Global g )
   {
      this.label = label;
      this.type = type;
      this.order = order;
      length = (short)con.length;
      int valNum = 0;
      short[] content = new short[ length ];
      activations = new double[ length ];
      for( int i = 0; i < length; i++ )
           if( con[ i ] == 1 ){
               content[ valNum++ ] = (short)i;
               activations[ i ] = 1.0;
           }
      values = new short[ valNum ];
      for( int i = 0; i < valNum; i++ )
           values[ i ] = content[ i ];
      Arrays.sort( values );
      global = g;
   }

   public Feature( String label, short type, short order, double[] actVals, Global g )
   {
      this.label = label;
      this.type = type;
      this.order = order;
      length = (short)actVals.length;
      int valNum = 0;
      short[] content = new short[ length ];
      activations = new double[ length ];
      for( int i = 0; i < length; i++ ){
           activations[ i ] = actVals[ i ];
           if( activations[ i ] > 0.0 )
               content[ valNum++ ] = (short)i;
      }
      values = new short[ valNum ];
      for( int i = 0; i < valNum; i++ )
           values[ i ] = content[ i ];
      Arrays.sort( values );
      global = g;
   }

   public Feature( String label, short type, short order, short length, short[] vals, Global g )
   {
      this.label = label;
      this.type = type;
      this.order = order;
      this.length = length;
      values = new short[ vals.length ];
      activations = new double[ length ];
      for( int i = 0; i < vals.length; i++ ){
           values[ i ] = vals[ i ];
           activations[ values[i] ] = 1.0;
      }
      Arrays.sort( values );
      global = g;
   }

   public boolean isTheSameFeature( Feature other )
   {
      return getType()==other.getType() && getOrder()==other.getOrder();
   }

   public String getLabel()
   {
      return label;
   }

   public int getType()
   {
      return type;
   }

   public int getOrder()
   {
      return order;
   }

   public int getLength()
   {
      return length;
   }

   public int getValueNum()
   {
      return values.length;
   }

   public short[] getValues()
   {
      return values;
   }

   public void getValues( short[] vals )
   {
      for( int i = 0; i < vals.length; i++ )
           vals[ i ] = values[ i ];
   }

   public short[] getContent()
   {
      short[] content = new short[ length ];
      for( int i = 0; i < length; i++ )
           content[ i ] = 0;
      for( int i = 0; i < values.length; i++ )
           content[ values[ i ] ] = 1;
      return content;
   }

   public void getContent( short[] content )
   {
      for( int i = 0; i < length; i++ )
           content[ i ] = 0;
      for( int i = 0; i < values.length; i++ )
           content[ values[ i ] ] = 1;
   }

   public double[] getActivations()
   {
      return activations;
   }

   public void getActivations( double[] vals )
   {
      for( int i = 0; i < vals.length; i++ )
           activations[ i ] = vals[ i ];
   }

   public void setValues( short[] arr )
   {
      values = new short[ arr.length ];
      for( int i = 0; i < arr.length; i++ )
           values[ i ] = arr[ i ];
      Arrays.sort( values );
   }

   // if this feature is contained in other feature.
   public boolean isContained( Feature other )
   {
       if( values.length > other.getValueNum() )
           return false;
       short[] con1 = this.getContent();
       short[] con2 = other.getContent();
       for( int i = 0; i < length; i++ )
            if( con1[i] == 1 && con2[i] == 0 )
                return false;
       return true;
   }

/************************* OVERRIDE METHODS ************************************/
   public String toString()
   {
      String[] names = global.getFeatureValueNames( type, order );
      //String str = "( " + label + ":";
      String str = "( ";
      for( int i = 0; i < values.length; i++ )
           str += names[ values[i] ] + " ";
      str += ")";
      return str;
   }

   public int hashCode()
   {
      /*
      int sum = 0;
      for( int i = 0; i < values.length; i++ )
           sum += values[ i ];
      return sum;
      */
      Arrays.sort( values );
      int num = Math.min( 32, length );
      int code = 0;
      for( int i = 0; i < num; i++ )
           code = (code <<= 1) + ((Arrays.binarySearch(values, (short)i) >= 0) ? 1 : 0);
      return code;
   }

   public boolean equals( Object other )
   {
      if( other == null || !(other instanceof Feature ) )
          return false;
      Feature otherFeat = (Feature)other;
      if( label.compareTo( otherFeat.getLabel()) != 0 )
          return false;
      if( type != otherFeat.getType() )
          return false;
      if( order != otherFeat.getOrder() )
          return false;
      if( values.length != otherFeat.getValueNum() )
          return false;
      Arrays.sort( otherFeat.getValues() );
      return Arrays.equals( values, otherFeat.getValues());
   }

   public Object clone()
   {
      Feature feature = new Feature( label, type, order, length, values, global );
      return feature;
   }
}