package clarion.common;
import clarion.system.*;

public class RuleAttributes {

  // global settings.
   protected Global global;
   // task-specific clarion settings.
   protected TaskClarion taskClarion;
   // ACS rule or NACS rule
   protected boolean acsRuleFlag;
   // the net indice.
   protected int netIdx;
   // the rule type.
   int ruleType;
   // the rule set the rule belongs to
   int ruleSet;
   // age from last application of this rule.
   int age;
   // RT variables.
   int TLRT, TLPT, TLDT, actionTime;
   int assocAppTime;
   double t0, t1, chunkBLA;
   // for base-level activation.
   int blaOption;
   int MAX_USAGE_NUM = 2000, usedNum, usages[];
   double BLA, c, d, iBLA;
   // used for utility
   int utlOption;
   //double utility, benefit, cost, v, c7, c8, PM, NM;
   double utility, PM, NM;

   // counstructor of the class RuleReature.
   // used for NACS rules.
   public RuleAttributes( Global global )
   {
       this.global = global;
       taskClarion = global.getTaskClarion();
       acsRuleFlag = false;
       actionTime = global.getDT_TL_actionTime();
       t0 = global.getDT_TL_t0();
       t1 = global.getDT_TL_t1();
       if( global.isAssocBlaOn() ){
           blaOption = global.getAssocBlaOption();
           if( blaOption == global.CONSTANT_OPTION )
               BLA = global.getAssocBla();
           else{
               c = global.getAssocBlaParas()[global.BLA_C];
               d = global.getAssocBlaParas()[global.BLA_D];
               iBLA = global.getAssocBlaParas()[global.BLA_I];
           }
       }
       else
           blaOption = global.EMPTY;
       //MAX_USAGE_NUM = global.getMaxUsageNum();
       usages = new int[ MAX_USAGE_NUM ];
       utlOption = global.getAssocUtilityOption();
       //benefit = global.getAssocUtilityParas()[global.UTL_BENE];
       //cost = global.getAssocUtilityParas()[global.UTL_COST];
       //v = global.getAssocUtilityParas()[global.UTL_V];
       //c7 = global.getAssocUtilityParas()[global.UTL_C7];
       //c8 = global.getAssocUtilityParas()[global.UTL_C8];
   }

   // used for ACS rules.
   public RuleAttributes( int netIdx, int ruleType, int ruleSet, Global global )
   {
       this.netIdx = netIdx;
       this.ruleType = ruleType;
       this.ruleSet = ruleSet;
       this.global = global;
       taskClarion = global.getTaskClarion();
       acsRuleFlag = true;
       TLPT = global.getPT_TL();
       actionTime = global.getDT_TL_actionTime();
       t0 = global.getDT_TL_t0();
       t1 = global.getDT_TL_t1();
       if( global.isRuleBlaOn( netIdx, ruleType, ruleSet ) ){
           blaOption = global.getActionRuleBlaOption( netIdx, ruleType, ruleSet );
           if( blaOption == global.CONSTANT_OPTION )
               BLA = global.getActionRuleBla( netIdx, ruleType, ruleSet );
           else{
               c = global.getActionRuleBlaParas( netIdx, ruleType, ruleSet )[global.BLA_C];
               d = global.getActionRuleBlaParas( netIdx, ruleType, ruleSet )[global.BLA_D];
               iBLA = global.getActionRuleBlaParas( netIdx, ruleType, ruleSet )[global.BLA_I];
           }
       }
       else
           blaOption = global.EMPTY;
       //MAX_USAGE_NUM = global.getMaxUsageNum();
       usages = new int[ MAX_USAGE_NUM ];
       if( global.isRuleUtilityOn( netIdx, ruleType, ruleSet ) ){
           utlOption = global.getRuleUtilityOption( netIdx, ruleType, ruleSet );
           //benefit = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_BENE];
           //cost = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_COST];
           //v = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_V];
           //c7 = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_C7];
           //c8 = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_C8];
       }
   }

   public void reinit()
   {
       PM = 0.0;
       NM = 0.0;
       usedNum = 0;
       for( int i = 0; i < usages.length; i++ )
            usages[i] = 0;
   }

   public int getAge()
   {
       return age;
   }

   public double getPM()
   {
       return PM;
   }

   public double getNM()
   {
       return NM;
   }

   public double getUtility()
   {
       return utility;
   }

   public double getBLA()
   {
       return BLA;
   }

   public int getResponseTime()
   {
       return TLRT;
   }

   public int getAssocRuleAppTime()
   {
      return assocAppTime;
   }

   public void setAge( int age )
   {
       this.age = age;
   }

   public void setPM( double PM )
   {
       this.PM = PM;
   }

   public void setNM( double NM )
   {
       this.NM = NM;
   }

   public void setUsedNum( int usedNum )
   {
      this.usedNum = usedNum;
   }

   public void setUsages( int[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           usages[ i ] = arr[ i ];
   }

   /*
   public void setCost( double c )
   {
       cost = c;
   }
   */

   public void setChunkBLA( double val )
   {
       chunkBLA = val;
   }

   public void resetMatches()
   {
       PM = 0.0;
       NM = 0.0;
   }

   public void discountMatches( double discount )
   {
       PM *= discount;
       NM *= discount;
   }

   public void calcTLRT( int TL_AT )
   {
      TLPT = taskClarion.getPT_TL( global.getPT_BL() );
      int TLAT = TL_AT;
      TLRT = taskClarion.getTLOverallRT( TLPT, TLDT, TLAT );
   }

   public void calcTLDT( double chunkBLA )
   {
      this.chunkBLA = chunkBLA;
      //TLDT = (int)( actionTime + t0 / BLA + t1 / chunkBLA );
      TLDT = taskClarion.getDT_TL( global.getDT_BL(), actionTime, BLA, chunkBLA );
   }

   // flag indicates involving NACS or not.
   public void calcTLAT( boolean flag, int chunkRetrTime )
   {
      calcAssocRuleAppTime();
      int TLAT = taskClarion.getAT_TL( flag, assocAppTime, chunkRetrTime );
   }

   // associative rule application time for one iteration.
   public void calcAssocRuleAppTime()
   {
      /*
      double t4 = global.getAssocAppTime_t4();
      double t5 = global.getAssocAppTime_t5();
      assocRuleAppTime = (int)(t4 + t5 / BLA);
      */
      assocAppTime = taskClarion.getAssocAppTime( BLA );
   }

   public void calcUtility( double cost )
   {
      /*
       if( utlOption == global.VARIED_OPTION ){
           benefit = ( c7 + PM ) / ( c8 + PM + NM );
           cost = c;
       }
       utility = benefit - v * cost;
       */
       //this.cost = cost;
       if( acsRuleFlag ){
           switch( ruleType ){
              case Global.RER:
                   utility = taskClarion.getRerUtility( netIdx, ruleSet, PM, NM );
                   break;
              case Global.IRL:
                   utility = taskClarion.getIrlUtility( netIdx, ruleSet, PM, NM );
                   break;
              case Global.FR:
                   utility = taskClarion.getFrUtility( netIdx, ruleSet, PM, NM );
                   break;
          }
       }
       else
          utility = taskClarion.getAssocUtility( PM, NM );
   }

   public void calcBLA( int curTime )
   {
       if( blaOption != Global.VARIED_OPTION )
           return;
       else{
           double sum = iBLA;
           for( int i = 0; i < usedNum; i++ ) {
                int timeLaps = curTime - usages[ i ];
                if( timeLaps <= 0 ){
                    break;
                }
                sum += c * Math.pow( (double)timeLaps, -d );
           }
           BLA = sum;
       }
       if( BLA < 0.01 )
           BLA = 0.01;
   }

   public void updateAge()
   {
       age++;
   }

   public void updateMatches( boolean flag )
   {
      if( flag )
          PM += 1.0;
      else
          NM += 1.0;
   }

   public void updateMatches( int matchType )
   {
      if( matchType == global.POS_MATCH_TYPE )
          PM += 1.0;
      if( matchType == global.NEG_MATCH_TYPE )
          NM += 1.0;
   }

   public void updateUsages( int curTime )
   {
       age = 0;
       if( usedNum > 0 && curTime == usages[ usedNum-1 ] ){
           return;
       }
       if( usedNum < MAX_USAGE_NUM-1 )
           usages[ usedNum++ ] = curTime;
       else
           System.out.println( "error usage update in RuleAttributes.java" );
   }

   public Object clone()
   {
      RuleAttributes feature = new RuleAttributes( netIdx, ruleType, ruleSet, global );
      feature.setAge( age );
      feature.setUsedNum( usedNum );
      feature.setUsages( usages );
      feature.setPM( PM );
      feature.setNM( NM );
      feature.setChunkBLA( chunkBLA );
      return feature;
   }
}