// this class is a generic one for coordinating actions of EXTERNAL and
// GOAL STACK.
package clarion.coordinate;
import clarion.system.*;

public class ExGScoordinator {
   final int DO_NOTHING = -2, NO_AVAILABLE = -1;
   Global global;   // global settings.
   TaskAgent agent;
   short performActionType, chosenActions[][];
   double exProb, gsProb;

   // constructor for stochastic combination method.
   // mode indicates fixed or varied. = 0: fixed; = 1: varied.
   public ExGScoordinator(){}

   public ExGScoordinator( TaskAgent agent, Global global )
   {
      this.agent = agent;
      this.global = global;
      exProb = global.getExternCorConstant();
      gsProb = global.getGoalCorConstant();
   }

   public void setChosenActions( short[][] actions )
   {
      chosenActions = actions;
   }

   // coordination of EX and GS as described in the description.
   public void coordinate()
   {
      int exAction = checkExActions( chosenActions );
      int gsAction = NO_AVAILABLE;
      int wmAction = NO_AVAILABLE;
      if( global.getGsOption() != Global.GS_NON )
          gsAction = checkGsAction( chosenActions );
      if( global.isWmOn() )
          wmAction = checkWmAction( chosenActions );
      if( exAction == NO_AVAILABLE || exAction == DO_NOTHING ){
          if( gsAction >= 0 )
              performActionType = Global.GS_TYPE;
      }
      if( gsAction == NO_AVAILABLE || gsAction == DO_NOTHING ){
          if( exAction >= 0 )
              performActionType = Global.EX_TYPE;
      }
      if( ( exAction == NO_AVAILABLE || exAction == DO_NOTHING ) &&
          ( gsAction == NO_AVAILABLE || gsAction == DO_NOTHING ) )
          performActionType = Global.EMPTY;
      randomCoordinate();
      if( performActionType == Global.EX_TYPE ){
          for( int i = 0; i < chosenActions.length; i++ )
               if( global.isExNet(i) && chosenActions[i][0] >= 0 )
                   agent.performAction( i, chosenActions[i] );
          if( wmAction >= 0 )
              global.getACS().performWmAction();
          if( gsAction >= 0 )
              global.getACS().performGsAction();
      }
      if( performActionType == Global.GS_TYPE ){
          if( wmAction >= 0 )
              global.getACS().performWmAction();
          if( gsAction >= 0 )
              global.getACS().performGsAction();
      }
   }

   // using "random selection" method to select EX or GS as currently
   // chosen action type provided that both chosen actions from EX and
   // GS are available and not do nothing.
   private void randomCoordinate()
   {
      double randVal = Math.random();
      if( randVal < exProb )
          performActionType = Global.EX_TYPE;
      else
          performActionType = Global.GS_TYPE;
   }

   // check if EX action is DO_NOTHING or NOT_AVAILABLE
   private int checkExActions( short[][] actions )
   {
      int value = NO_AVAILABLE;
      for( int i = 0; i < actions.length; i++ )
           if( global.isExNet(i) ){
               if( actions[i][0] > 0 )   // not do-nothing.
                   value = 1;
               if( actions[i][0] == 0 && value == NO_AVAILABLE )
                   value = DO_NOTHING;
           }
      return value;
   }

   // check if GS action is DO_NOTHING or NOT_AVAILABLE
   private int checkGsAction( short[][] actions )
   {
      int value = NO_AVAILABLE;
      for( int i = 0; i < actions.length; i++ )
           if( global.isGsNet(i) ){
               if( actions[i][0] != Global.EMPTY &&
                   actions[i][0] != global.getGS().getDoNothingAction() )   // not do-nothing.
                   value = 1;
               if( actions[i][0] == global.getGS().getDoNothingAction() && value == NO_AVAILABLE )
                   value = DO_NOTHING;
           }
      return value;
   }

   // check if WM action is DO_NOTHING or NOT_AVAILABLE
   private int checkWmAction( short[][] actions )
   {
      int value = NO_AVAILABLE;
      for( int i = 0; i < actions.length; i++ )
           if( global.isWmNet(i) ){
               if( actions[i][0] != Global.EMPTY &&
                   actions[i][0] != global.getWM().getDoNothingAction() )   // not do-nothing.
                   value = 1;
               if( actions[i][0] == global.getWM().getDoNothingAction() && value == NO_AVAILABLE )
                   value = DO_NOTHING;
           }
      return value;
   }
}