package clarion.coordinate;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class StochasticComb {

   Global global;
   int componentNum, actionNum;
   int chosenComponent, chosenAction;
   double probArr[], betaArr[], succArr[];
   StochasticDecider actionDecider;

   // mode = 0: fixed method which means the probs are fixed,
   // mode = 1: varied method which means the probs are got by calculating.
   public StochasticComb( int mode, int compNum, double[] arr )
   {
      componentNum = compNum;

      if( mode == 0 )      // fixed mode.
      {
          probArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               probArr[ i ] = arr[ i ];
      }
      else                // varied mode.
      {
          probArr = new double[ componentNum ];
          betaArr = new double[ componentNum ];
          succArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               betaArr[ i ] = arr[ i ];
      }

      actionDecider = new StochasticDecider( actionNum,
                          World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );
   }

   public StochasticComb( int mode, int compNum, double[] arr, Global g )
   {
      componentNum = compNum;
      if( mode == 0 )      // fixed mode.
      {
          probArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               probArr[ i ] = arr[ i ];
      }
      else                // varied mode.
      {
          probArr = new double[ componentNum ];
          betaArr = new double[ componentNum ];
          succArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               betaArr[ i ] = arr[ i ];
      }

      global = g;
      actionDecider = new StochasticDecider();
   }

   // used in varied percentage combination method.
   public void setSuccRate( double[] arr )
   {
      for( int i = 0; i < componentNum; i++ )
           succArr[ i ] = arr[ i ];
   }

   public void calcProbabilities()
   {
      double sum = 0;
      for( int i = 0; i < componentNum; i++ )
           sum += betaArr[ i ] * succArr[ i ];

      for( int i = 0; i < componentNum; i++ )
           probArr[ i ] = betaArr[ i ] * succArr[ i ] / sum;
   }

   // using "radomly selected" approach to select one component on which
   // the chosen actions are based, such as action chosing is based on
   // BL, FR, RER, IRL etc.
   public int selectComponent()
   {
      double randomVal = Math.random();
      chosenComponent = 0;
      double threshold = 0.0;
      for( int i = 0; i < componentNum; i++ )
      {
           threshold += probArr[ i ];
           if( randomVal < threshold )
           {
               chosenComponent = i;
               break;
           }
      }
      return chosenComponent;
   }

   // based on a chosen component, within an aciton type such as EX, GS, WM
   // action, select one action from that type using "randomly selected"
   // approach.
   public int selectAction( int actNum )
   {
      actionNum = actNum;
      double threshold = 0.0;
      double randomVal = Math.random();
      chosenAction = 0;
      double interval = 1.0 / actionNum;
      for( int i = 0; i < actionNum; i++ )
      {
           threshold += interval;
           if( randomVal < threshold )
           {
               chosenAction = i;
               break;
           }
      }
      return chosenAction;
   }

   // based on a chosen component, within an aciton type such as EX, GS, WM
   // action, select one action from that type using "boltzmann distribution
   // based on utility" approach.
   public int selectAction( double[] utilArr )
   {
      chosenAction = 0;
      chosenAction = actionDecider.getRecommendation( utilArr );
      return chosenAction;
   }

   public int selectAction( int netIdx, int knowledgeType, double[] utilArr )
   {
      chosenAction = 0;
      actionDecider.setTemperature( global.getARActionSelectionTemp( netIdx, knowledgeType ) );
      chosenAction = actionDecider.getRecommendation( utilArr );
      return chosenAction;
   }
}
