package clarion.coordinate;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class WeightSumComb{

   int componentNum, actionNumArr;
   double weightArr[], activationArr[][];
   double betaArr[], succArr[];
   StochasticDecider actionDecider;

   // mode = 0: fixed method which means the weights are fixed,
   // mode = 1: varied method which means the weights are got by calculating.
   public WeightSumComb( int mode, int actNumArr, double[] arr, double temperature )
   {
      actionNumArr = actNumArr;
      componentNum = arr.length;
      if( mode == 0 )     // fixed mode
      {
          weightArr = new double[ componentNum ];
            for( int i = 0; i < componentNum; i++ )
                 weightArr[ i ] = arr[ i ];
      }
      else               // varied mode
      {
          weightArr = new double[ componentNum ];
          betaArr = new double[ componentNum ];
          succArr = new double[ componentNum ];
          for( int i = 0; i < componentNum; i++ )
               betaArr[ i ] = arr[ i ];
      }
      activationArr = new double[ componentNum ][];
      for( int j = 0; j < componentNum; j++ )
           activationArr[j] = new double[ actionNumArr ];
      //actionDecider = new StochasticDecider( World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );
      actionDecider = new StochasticDecider( temperature );
   }

   public void getWeights( double[] arr )
   {
      for( int j = 0; j < componentNum; j++ )
           arr[ j ] =  weightArr[ j ];
   }

   // used in varied mode.
   public void setSuccRate( double[] arr )
   {
       for( int j = 0; j < componentNum; j++ )
            succArr[ j ] = arr[ j ];
   }

   // set the activatitons of each action from one type indexed by "typeIdx".
   // the activatitons of each action from each type is from different
   // components such as BL, FR, RER, IRL etc.
   public void setActivations( double[][] arr )
   {
      for( int i = 0; i < componentNum; i++ )
        for( int j = 0; j < actionNumArr; j++ )
             activationArr[ i ][ j ] = arr[ i ][ j ];
   }

   public void calcWeights()
   {
       double sum = 0;
       for( int j = 0; j < componentNum; j++ )
             sum += betaArr[ j ] * succArr[ j ];
       for( int j = 0; j < componentNum; j++ )
            weightArr[ j ] = betaArr[ j ] * succArr[ j ] / sum;
   }

   // old version
   public int combination()
   {
       int chosenAction = 0;
       int actionNum = actionNumArr;
       double[] inputArr = new double[ actionNum ];
       for( int i = 0; i < actionNum; i++ ){
            inputArr[i] = 0;
            for( int j = 0; j < componentNum; j++ ){
                 double val = weightArr[j] * activationArr[j][i];
                 inputArr[i] += val;
            }
       }
       chosenAction = actionDecider.getRecommendation( inputArr );
       return chosenAction;
   }

   //new version get the chosen action of one action type indexed by "netIdx".
   public void combination( short[] dimVals, short[] action )
   {
      int actionNum = actionNumArr;
      double[] inputArr = new double[ actionNum ];
      for( int i = 0; i < actionNum; i++ ){
           inputArr[ i ] = 0;
           for( int j = 0; j < componentNum; j++ ){
                double val = weightArr[j] * activationArr[j][i];
                inputArr[ i ] += val;
           }
      }
      short[] chosenAction = actionDecider.getRecommendation( inputArr, dimVals );
      for( int i = 0; i < action.length; i++ )
           action[i] = chosenAction[i];
   }
}
