package clarion.mcs;
import clarion.tools.*;
import clarion.system.*;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.common.*;

public class MCS {

   // global CLARION settings.
   final int ACS_COMP_NUM = Global.ACS_COMP_NUM;
   Global global;
   int  acsNetNum;
   int  goalAction;
   double  reinforcement;
   short[] acsInputDims[][], acsOutputDims[][];
   short[] nacsInputDims[], nacsOutputDims[];
   short[] acsReasonMethods[], nacsReasonMethods;
   short[] acsLearnMethods[], nacsLearnMethods;
   short[] offsets = new short[ 5 ];
   double[] input, output, state, goal, wm, drives, buffer;
   // the BP net for evaluation.
   BPNet evalNet;
   // the BP net for goal action.
   BPNet gsNet;
   // the BP nets for ACS filters.
   BPNet acsInputFilterNets[][], acsOutputFilterNets[][];
   // the BP nets for NACS filters.
   BPNet nacsInputFilterNets[], nacsOutputFilterNets[];
   // the BP nets for ACS reasoning.
   BPNet acsReasonNets[][];
   // the BP nets for NACS reasoning.
   BPNet nacsReasonNets[];
   // the BP nets for ACS learning.
   BPNet acsLearnNets[][];
   // the BP nets for NACS learning.
   BPNet nacsLearnNets[];
   MonitorBuf monitorBuf;
   StochasticDecider decider;

   public MCS( ACS acs, NACS nacs, Drives drives, Global global )
   {
      this.global = global;
      global.setMCS( this );
      acsNetNum = global.getAcsNetNum();
      // the BP nets for filters.
      acsInputFilterNets = new BPNet[acsNetNum][ACS_COMP_NUM];
      acsOutputFilterNets = new BPNet[acsNetNum][ACS_COMP_NUM];
      nacsInputFilterNets = new BPNet[2];
      nacsOutputFilterNets = new BPNet[2];
      // the BP nets for reasoning.
      acsReasonNets = new BPNet[acsNetNum][ACS_COMP_NUM];
      nacsReasonNets = new BPNet[2];
      // the BP nets for learning.
      acsLearnNets = new BPNet[acsNetNum][ACS_COMP_NUM];
      nacsLearnNets = new BPNet[2];
      monitorBuf = new MonitorBuf( global );
      offsets[0] = 0;
      offsets[1] = (short)(offsets[0] + acs.getSensoryLength());
      offsets[2] = (short)(offsets[1] + acs.getGsLength());
      offsets[3] = (short)(offsets[2] + acs.getWmLength());
      offsets[4] = (short)(offsets[3] + drives.getLength());
      int len = offsets[4] + monitorBuf.getLength();
      input = new double[ len ];
   }

   public void setState( short[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           state[i] = arr[i];
   }

   public void setGoal( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           goal[i] = arr[i];
   }

   public void setWM( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           wm[i] = arr[i];
   }

   public void setDrvies( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           drives[i] = arr[i];
   }

   public void setBuffer( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           buffer[i] = arr[i];
   }

   public MonitorBuf getMonitorBuffer()
   {
      return monitorBuf;
   }

   public double getReinforcement()
   {
      getOutputs( evalNet );
      reinforcement = decider.getRecommendation( output, output.length );
      return reinforcement;
   }

   public int getGoalAction()
   {
      getOutputs( gsNet );
      goalAction = decider.getRecommendation( output, output.length );
      return goalAction;
   }

   public short[] getAcsInputDims( int netIdx, int compIdx )
   {
      getOutputs( acsInputFilterNets[netIdx][compIdx] );
      return acsInputDims[netIdx][compIdx];
   }

   public short[] getAcsOutputDims( int netIdx, int compIdx )
   {
      getOutputs( acsOutputFilterNets[netIdx][compIdx] );
      return acsOutputDims[netIdx][compIdx];
   }

   public short[] getNacsInputDims( int compIdx )
   {
      getOutputs( nacsInputFilterNets[compIdx] );
      return nacsInputDims[compIdx];
   }

   public short[] getNacsOutputDims( int compIdx )
   {
      getOutputs( nacsOutputFilterNets[compIdx] );
      return nacsOutputDims[compIdx];
   }

   public int getAcsReasonMethod( int netIdx, int compIdx )
   {
      getOutputs( acsReasonNets[netIdx][compIdx] );
      return acsReasonMethods[netIdx][compIdx];
   }

   public int getNacsReasonMethod( int compIdx )
   {
      getOutputs( nacsReasonNets[compIdx] );
      return nacsReasonMethods[compIdx];
   }

   public int getAcsLearnMethod( int netIdx, int compIdx )
   {
      getOutputs( acsLearnNets[netIdx][compIdx] );
      return acsLearnMethods[netIdx][compIdx];
   }

   public int getNacsLearnMethod( int compIdx )
   {
      getOutputs( nacsLearnNets[compIdx] );
      return nacsLearnMethods[compIdx];
   }

   private void getOutputs( BPNet net )
   {
      fillInput();
      net.setInput( input );
      net.forwardPass();
      output = net.getOutput();
   }

   private void fillInput()
   {
      fillInput( 0, state );
      fillInput( 1, goal );
      fillInput( 2, wm );
      fillInput( 3, drives );
      fillInput( 4, buffer );
   }

   private void fillInput( int idx, double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           input[ offsets[idx] + i ] = arr[i];
   }
}