package clarion.mcs;
import java.util.Vector;
import clarion.system.*;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.common.*;
import clarion.tools.*;

public class MonitorBuf {

   final int BL = Global.BL;
   final int TL = Global.TL;
   final int COMB = 2;
   /**
    *  global CLARION settings.
    */
   Global global;
   /**
    *  the parameters in the performance and learning section.
    */
   int c20 = 2, c21 = 2, c22 = 1, c23 = 1;
   int c24 = 2, c25 = 2, c26 = 1, c27 = 1;
   int c28 = 2, c29 = 2, c30 = 2, c31 = 2;
   int acsNetNum, acsNetCompNum, bufferLength;
   double[] buffer;
   double[] acsStrengths[][], nacsStrengths[];
   double[] acsProgress[][], nacsProgress[];
   double[] acsImprove[][], nacsImprove[];
   Vector   preAcsChunks[][], preNacsChunks[];
   Vector   curAcsChunks[][], curNacsChunks[];

   public MonitorBuf( Global global )
   {
      this.global = global;
      c20 = global.getMonitorAcsTopConcNum();
      c24 = global.getMonitorAcsBottomConcNum();
      c28 = global.getMonitorAcsTopEpisodeNum();
      c30 = global.getMonitorAcsBottomEpisodeNum();
      c21 = global.getMonitorNacsTopConcNum();
      c25 = global.getMonitorNacsBottomConcNum();
      c29 = global.getMonitorNacsTopEpisodeNum();
      c31 = global.getMonitorNacsBottomEpisodeNum();
      acsNetNum = global.getAcsNetNum();
      acsNetCompNum = global.getAcsKnowledgeNum();
      // relative strengths.
      acsStrengths = new double[acsNetNum][][];
      for( int i = 0; i < acsNetNum; i++ ){
           int c = ( i == BL ) ? c24 : c20;
           acsStrengths[i] = new double[acsNetCompNum][c];
      }
      //nacsStrengths = new double[2][];
      nacsStrengths = new double[3][];
      nacsStrengths[TL] = new double[c21];
      nacsStrengths[BL] = new double[c25];
      nacsStrengths[COMB] = new double[ Math.max(c21, c25) ];
      // progress.
      acsProgress = new double[acsNetNum][][];
      for( int i = 0; i < acsNetNum; i++ ){
           int c = ( i == BL ) ? c26 : c22;
           acsProgress[i] = new double[acsNetCompNum][c];
      }
      nacsProgress = new double[2][];
      nacsProgress[TL] = new double[c23];
      nacsProgress[BL] = new double[c27];
      // improvements.
      acsImprove = new double[acsNetNum][2][];
      for( int i = 0; i < acsNetNum; i++ ){
           acsImprove[i][TL] = new double[c28];
           acsImprove[i][BL] = new double[c30];
      }
      nacsImprove = new double[2][];
      nacsImprove[TL] = new double[c29];
      nacsImprove[BL] = new double[c31];
      //init the bufer
      bufferLength = 0;
      bufferLength += CommonMethods.calcLength( acsStrengths );
      bufferLength += CommonMethods.calcLength( nacsStrengths );
      bufferLength += CommonMethods.calcLength( acsProgress );
      bufferLength += CommonMethods.calcLength( nacsProgress );
      bufferLength += CommonMethods.calcLength( acsImprove );
      bufferLength += CommonMethods.calcLength( nacsImprove );
      buffer = new double[ bufferLength ];
      // init variables for storing the chunks.
      preAcsChunks = new Vector[acsNetNum][acsNetCompNum];
      for( int i = 0; i < acsNetNum; i++ )
        for( int j = 0; j < acsNetCompNum; j++ ) {
             int c = ( i == BL ) ? c24 : c20;
             preAcsChunks[i][j] = new Vector( c );
      }
      curAcsChunks = new Vector[acsNetNum][acsNetCompNum];
      for( int i = 0; i < acsNetNum; i++ )
        for( int j = 0; j < acsNetCompNum; j++ ) {
             int c = ( i == BL ) ? c24 : c20;
             curAcsChunks[i][j] = new Vector( c );
      }
      //preNacsChunks = new Vector[2];
      preNacsChunks = new Vector[3];
      preNacsChunks[TL] = new Vector( c21 );
      preNacsChunks[BL] = new Vector( c25 );
      preNacsChunks[COMB] = new Vector( Math.max(c21, c25) );
      //curNacsChunks = new Vector[2];
      curNacsChunks = new Vector[3];
      curNacsChunks[TL] = new Vector( c21 );
      curNacsChunks[BL] = new Vector( c25 );
      curNacsChunks[COMB] = new Vector( Math.max(c21, c25) );
   }

   public int getLength()
   {
      return buffer.length;
   }

   public Vector getAcsChunks( int netIdx, int compIdx )
   {
      return curAcsChunks[netIdx][compIdx];
   }

   public Vector getNacsChunks( int level )
   {
      return curNacsChunks[level];
   }

   public double[] getBuffer()
   {
      int offset = 0;
      for( int i = 0; i < acsStrengths.length; i++ )
       for( int j = 0; j < acsStrengths[i].length; j++ )
        for( int k = 0; k < acsStrengths[i][j].length; k++ )
             buffer[ offset++ ] = acsStrengths[i][j][k];
      for( int i = 0; i < nacsStrengths.length; i++ )
       for( int j = 0; j < nacsStrengths[i].length; j++ )
             buffer[ offset++ ] = nacsStrengths[i][j];
      for( int i = 0; i < acsProgress.length; i++ )
       for( int j = 0; j < acsProgress[i].length; j++ )
        for( int k = 0; k < acsProgress[i][j].length; k++ )
             buffer[ offset++ ] = acsProgress[i][j][k];
      for( int i = 0; i < nacsProgress.length; i++ )
       for( int j = 0; j < nacsProgress[i].length; j++ )
             buffer[ offset++ ] = nacsProgress[i][j];
      for( int i = 0; i < acsImprove.length; i++ )
       for( int j = 0; j < acsImprove[i].length; j++ )
        for( int k = 0; k < acsImprove[i][j].length; k++ )
             buffer[ offset++ ] = acsImprove[i][j][k];
      for( int i = 0; i < nacsImprove.length; i++ )
       for( int j = 0; j < nacsImprove[i].length; j++ )
             buffer[ offset++ ] = nacsImprove[i][j];
      return buffer;
   }

   /**
    * netIdx = network NO.
    * compIdx = IDN, RER, IRL, FR.
    */
   public double[] getAcsRelStrengths( int netIdx, int compIdx )
   {
      calcRelStrengths( curAcsChunks[netIdx][compIdx], acsStrengths[netIdx][compIdx] );
      return acsStrengths[netIdx][compIdx];
   }

   /**
    * level = TL or BL.
    */
   public double[] getNacsRelStrengths( int level )
   {
      calcRelStrengths( curNacsChunks[level], nacsStrengths[level] );
      return nacsStrengths[ level ];
   }

   /**
    * netIdx = network NO.
    * compIdx = IDN, RER, IRL, FR.
    */
   public double[] getAcsProgress( int netIdx, int compIdx )
   {
      return acsProgress[netIdx][compIdx];
   }

   /**
    * level = TL or BL.
    */
   public double[] getNacsProgress( int level )
   {
      return nacsProgress[ level ];
   }

   /**
    * netIdx = network NO.
    * level = TL or BL.
    */
   public double[] getAcsImprovement( int netIdx, int level )
   {
      return acsImprove[ netIdx ][ level ];
   }

   /**
    * level = TL or BL.
    */
   public double[] getNacsImprovement( int level )
   {
      return nacsImprove[ level ];
   }

   public void setCurAcsConclusions( int netIdx, int compIdx, Chunk[] chunks )
   {
      if( chunks == null )
          return;
      try{
          preAcsChunks[netIdx][compIdx].removeAllElements();
          // move conclusions from current to previous.
          int num = curAcsChunks[netIdx][compIdx].size();
          for( int i = 0; i < num; i++ )
               preAcsChunks[netIdx][compIdx].addElement( curAcsChunks[netIdx][compIdx].remove(0) );
          // fill in current conclusions.
          num = (compIdx == BL)? Math.min( c24, chunks.length ):Math.min( c20, chunks.length );
          for( int i = 0; i < num; i++ )
               curAcsChunks[netIdx][compIdx].addElement((Chunk)(chunks[i].clone()));
      }
      catch( CloneNotSupportedException e ){
      }
   }

   public void setCurNacsConclusions( int level, Chunk[] chunks )
   {
      if( chunks == null )
          return;
      try{
          preNacsChunks[level].removeAllElements();
          // move conclusions from current to previous.
          int num = curNacsChunks[level].size();
          for( int i = 0; i < num; i++ )
               preNacsChunks[level].addElement( curNacsChunks[level].remove(0) );
          // fill in current conclusions.
          num = Math.min( curNacsChunks[level].capacity(), chunks.length );
          for( int i = 0; i < num; i++ )
               curNacsChunks[level].addElement((Chunk)(chunks[i].clone()));

          /*
          System.out.println( "chunks in montor buffer in MonitorBuf.java = " );
          for( int i = 0; i < curNacsChunks[level].size(); i++ )
               System.out.println( curNacsChunks[level].elementAt(i) );
          */

      }
      catch( CloneNotSupportedException e ){
      }
   }

   public void setAcsEpiSuccRate( int netIdx, int level, double succRate )
   {
      int len = acsImprove[netIdx][level].length;
      for( int i = 1; i < len; i++ )
           acsImprove[netIdx][level][i-1] = acsImprove[netIdx][level][i];
      acsImprove[netIdx][level][len-1] = succRate;
   }

   public void setNacsEpiSuccRate( int level, double succRate )
   {
      int len = nacsImprove[level].length;
      for( int i = 1; i < len; i++ )
           nacsImprove[level][i-1] = nacsImprove[level][i];
      nacsImprove[level][len-1] = succRate;
   }

   public void calcAcsRelStrengths()
   {
      for( int i = 0; i < acsNetNum; i++ )
        for( int j = 0; j < acsNetCompNum; j++ )
             calcRelStrengths( curAcsChunks[i][j], acsStrengths[i][j] );
   }

   public void calcNacsRelStrengths()
   {
      calcRelStrengths( curNacsChunks[BL], nacsStrengths[BL] );
      calcRelStrengths( curNacsChunks[TL], nacsStrengths[TL] );
      calcRelStrengths( curNacsChunks[COMB], nacsStrengths[COMB] );
   }

   public void calcAcsProgress()
   {
      for( int i = 0; i < acsNetNum; i++ )
        for( int j = 0; j < acsNetCompNum; j++ )
             calcProgress( preAcsChunks[i][j], curAcsChunks[i][j], acsProgress[i][j] );
   }

   public void calcNacsProgress()
   {
      calcProgress( preNacsChunks[BL], curNacsChunks[BL], nacsProgress[BL] );
      calcProgress( preNacsChunks[TL], curNacsChunks[TL], nacsProgress[TL] );
   }

   private void calcRelStrengths( Vector chunks, double[] strengths )
   {
      double sum = 0.0;

      //System.out.println( "chunk size = " + chunks.size() );
      //System.out.println( "strength len = " + strengths.length );

      for( int i = 0; i < chunks.size(); i++ )
           sum += ((Chunk)chunks.elementAt(i)).getStrength();
      for( int i = 0; i < chunks.size(); i++ )
           strengths[i] = ((Chunk)chunks.elementAt(i)).getStrength()/sum;
   }

   private void calcProgress( Vector preChunks, Vector curChunks, double[] progress )
   {
      if( preChunks.isEmpty() )
          return;
      int chgNum = preChunks.size();
      for( int i = 0; i < preChunks.size(); i++ ){
           Chunk chunk1 = (Chunk)(preChunks.elementAt(i));
           for( int j = 0; j < curChunks.size(); j++ ){
                Chunk chunk2 = (Chunk)(curChunks.elementAt(j));
                if( chunk1.equals( chunk2 ) ){
                    chgNum--;
                    break;
                }
           }
      }
      for( int i = 1; i < progress.length; i++ )
           progress[i-1] = progress[i];
      progress[progress.length-1] = (double)chgNum / preChunks.size();
   }
}