package clarion.ms;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.acs.gs.*;

public class MS {

   final int DR_NET_NUM = 3;
   // global settings.
   Global global;
   /**
    * goal structure, the one in ACS.
    */
   GoalStructure gs;
   /**
    *  goal actions for ACS and MCS.
    */
   int acsGoalAction, mcsGoalAction;
   /**
    * the goals for ACS and MCS by performing the corresponding action.
    */
   GoalChunk acsGoal, mcsGoal;
   /**
    * the sensory input to the drive networks.
    */
   double[] sensoryInput;
   /**
    * the drive outputs from the drive networks.
    */
   double[] driveOutput[] = new double[DR_NET_NUM][];
   /**
    * the BP drive networks
    */
   BPNet[] driveNets = new BPNet[DR_NET_NUM];

   public MS( Global global, GoalStructure gs )
   {
      this.global = global;
      this.gs = gs;
      global.setMS( this );
   }

   public void setAcsGoalAction( int action )
   {
      acsGoalAction = action;
   }

   public void setMcsGoalAction( int action )
   {
      mcsGoalAction = action;
   }

   public void setSensoryInput( double[] input )
   {
      sensoryInput = input;
   }

   public GoalChunk getAcsGoal()
   {
      return acsGoal;
   }

   public GoalChunk getMcsGoal()
   {
      return mcsGoal;
   }

   /**
    *  get the drive strengths with the net number - netIdx.
    */
   public double[] getDriveStrength( int netIdx )
   {
      driveNets[netIdx].setInput( sensoryInput );
      driveNets[netIdx].forwardPass();
      return driveNets[netIdx].getOutput();
   }

   /**
    *  offline training a drive network with the net number - netIdx.
    */
   public void offTrainDriveNet( int netIdx )
   {
      driveNets[netIdx].forwardPass();
      driveNets[netIdx].backwardPass();
   }
}