package clarion.nacs;
import java.awt.*;
import java.util.Arrays;
import java.util.Vector;
import clarion.common.*;
import clarion.system.*;
import clarion.tools.*;

public class AMNet extends BPNet{

  final int AAM_TYPE = 0, HAM_TYPE = 1;  // different types of AMN.
  // global settings.
  Global global;
  // amn net index
  int netIdx;
  // GKS
  GKS gks;
  // for offline training
  boolean assocAssimOn;
  boolean emOn;
  double  emTrainNacsProb;
  // retrieval results.
  Feature[] results;
  // the variables for AMN.
  int    netType;            // the type of AMN: AAM or HAM.
  int    inputDimNum;        // the # of dims of input.
  int    outputDimNum;       // the # of dims of output.
  short  inputDVs[];         // the # of values in each dim of input.
  short  outputDVs[];        // th3 # of values in each dim of input.
  short  inputFeatInfo[][];  // input's each feature's type and order in that type.
  short  outputFeatInfo[][]; // output's each feature's type and order in that type.
  // need be added to description.
  double noiseThreshold;
  int    RT;
  int    t_AMN;   // one pass time.

  // old constructor.
  public AMNet( int inputNum, int hiddenNum, int outputNum, int dNum, int[] dValNums )
  {
      super( inputNum, hiddenNum, outputNum );
      inputDimNum = dNum;
      inputDVs = new short[ inputDimNum ];
      for( int i = 0; i < inputDimNum; i++ )
           inputDVs[ i ] = (short)dValNums[ i ];
  }

  // old constructor.
  public AMNet( int inputNum, int hiddenNum, int outputNum, int dNum, short[] dValNums )
  {
      super( inputNum, hiddenNum, outputNum );
      inputDimNum = dNum;
      inputDVs = new short[ inputDimNum ];
      for( int i = 0; i < inputDimNum; i++ )
           inputDVs[ i ] = (short)dValNums[ i ];
  }

  // new constructor, # of input nodes should be equal to # of output nodes.
  public AMNet( int nIdx, Global g, GKS gks )
  {
      super( g.getAmnInputNum(nIdx), nIdx, g, Global.NACS );
      netIdx = nIdx;
      global = g;
      this.gks = gks;
      assocAssimOn = global.isAssocAssimiOn();
      emOn = global.isEMon();
      if( emOn )
          emTrainNacsProb = global.getEmTrainProb();
      netType = global.getAmnType( netIdx );
      inputDVs = global.getAmnInputDimVals( netIdx );
      outputDVs = global.getAmnOutputDimVals( netIdx );
      inputDimNum = inputDVs.length;
      outputDimNum = outputDVs.length;
      inputFeatInfo = global.getAmnInputFeatureInfo( netIdx );
      outputFeatInfo = global.getAmnOutputFeatureInfo( netIdx );
      if( global.isRTOn() )
          t_AMN = global.getTaskClarion().getAmnPassTime();
  }

  public void reinit()
  {
      super.reinitWeights();
  }

  public Feature[] getResults()
  {
      return results;
  }

  // get an output filtering noise.
  public void getOutput( double[] arr )
  {
      for( int i = 0; i < arr.length; i++ )
           if( Output[ i ] < noiseThreshold )
               arr[ i ] = 0.0;
           else
               arr[ i ] = Output[ i ];
  }

  // get an output filtering noise.
  public void getOutput( double[][] arr )
  {
      int offset = 0;
      for( int i = 0; i < outputDimNum; i++ )
      {
           for( int j = 0; j <  outputDVs[ i ]; j++ )
           {
                if( Output[ offset + j ] < noiseThreshold )
                    arr[ i ][ j ] = 0.0;
                else
                    arr[ i ][ j ] = Output[ offset + j ];
           }
           offset += outputDVs[i];
      }
  }

  public Feature[] getOutputFeatures( boolean multiFlag )
  {
      Feature[] outputFeatures = new Feature[ outputDimNum ];
      double[][] values = new double[ outputDimNum ][];
      int offset = 0;
      for( int i = 0; i < outputDimNum; i++ ){
           values[ i ] = new double[ outputDVs[ i ] ];
           for( int j = 0; j < outputDVs[ i ]; j++ ){
                if( Output[ offset + j ] < noiseThreshold )
                    values[ i ][ j ] = 0.0;
                else
                    values[ i ][ j ] = Output[ offset + j ];
           }
           short fType = outputFeatInfo[ i ][ 0 ];
           short fOrder = outputFeatInfo[ i ][ 1 ];
           String fLabel = global.getFeatureNames( fType, fOrder );
           if( !multiFlag ){
               int idx = CommonMethods.getMaxValueIdx( values[i] );
               for( int j = 0; j < values[i].length; j++ )
                    if( j != idx )
                        values[i][j] = 0.0;
           }
           outputFeatures[ i ] = new Feature( fLabel, fType, fOrder, values[ i ], global );
           offset += outputDVs[i];
      }
      return outputFeatures;
  }

  public void resetRT()
  {
      RT = 0;
  }

  public int getRT()
  {
      return RT;
  }

  public void setNoiseThreshold( double th )
  {
      noiseThreshold = th;
  }

  // each element of argument 'arr' represents a feature.
  public void setInput( int[] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           Input[ offset + arr[i] ] = 1;
           offset += inputDVs[i];
      }
  }

  // set a chunk as input.
  public void setInput( short[][] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           for( int j = 0; j < inputDVs[i]; j++ )
                Input[ offset + j ] = arr[ i ][ j ];
           offset += inputDVs[i];
      }
  }

  // set the strength of a chunk as input.
  public void setInput( double[][] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           for( int j = 0; j < inputDVs[i]; j++ )
                Input[ offset + j ] = arr[ i ][ j ];
           offset += inputDVs[i];
      }
  }

  // set several chunks as input
  public void setInput( short[][][] arr )
  {
      int offset = 0, dimOffset = 0;
      for( int i = 0; i < arr.length; i++ ){
           for( int j = 0; j < arr[i].length; j++ ){
                for( int k = 0; k < inputDVs[ dimOffset ]; k++ )
                     Input[ offset + k ] = arr[ i ][ j ][ k ];
                offset += inputDVs[ dimOffset ];
                dimOffset++;
           }
      }
  }

  public void setInput( Feature[] features )
  {
      int offset = 0;
      for( int i = 0; i < inputDimNum; i++ ){
           if( i < features.length ){
               if( features[i] != null ){
                   short[] content = features[ i ].getContent();
                   for( int j = 0; j < inputDVs[ i ]; j++ )
                        Input[ offset + j ] = content[ j ];
               }
               else
                   for( int j = 0; j < inputDVs[ i ]; j++ )
                        Input[ offset + j ] = 0;
           }
           else
               for( int j = 0; j < inputDVs[ i ]; j++ )
                    Input[ offset + j ] = 0;
           offset += inputDVs[ i ];
      }
  }

  // each element of argument 'arr' represents a feature.
  public void setDesiredOutput( short[] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           DesiredOutput[ offset + arr[i] ] = 1;
           offset += outputDVs[i];
      }
  }

  // set a chunk as desired output.
  public void setDesiredOutput( short[][] arr )
  {
      int offset = 0;
      for( int i = 0; i < arr.length; i++ ){
           for( int j = 0; j < outputDVs[i]; j++ )
                DesiredOutput[ offset + j ] = arr[ i ][ j ];
           offset += outputDVs[i];
      }
  }

  public void setDesiredOutput( Feature[] features )
  {
      int offset = 0;
      for( int i = 0; i < outputDimNum; i++ ){
           if( i < features.length ){
               if( features[i] != null ){
                   short[] content = features[ i ].getContent();
                   for( int j = 0; j < outputDVs[ i ]; j++ )
                        DesiredOutput[ offset + j ] = content[ j ];
               }
               else
                   for( int j = 0; j < outputDVs[ i ]; j++ )
                        DesiredOutput[ offset + j ] = 0;
           }
           else
               for( int j = 0; j < outputDVs[ i ]; j++ )
                    DesiredOutput[ offset + j ] = 0;
           offset += outputDVs[ i ];
      }
  }

  public void training()
  {
      forwardPass();
      backwardPass();
  }

  public boolean retrieval( boolean multiFlag )
  {
      RT = 0;
      if( netType == Global.AAM ){
          int iterNum = global.getAmnIterNum( netIdx );
          for( int i = 0; i < iterNum; i++ ){
               forwardPass();
               setInput( Output );
               if( global.isRTOn() )
                   RT += t_AMN;
          }
       }
      else{
           forwardPass();
           RT += t_AMN;
      }
      results = getOutputFeatures( multiFlag );
      return true;
  }

  public void update()
  {
      //System.out.println( "AMN update, assocAssimOn = " + assocAssimOn );

      if( !assocAssimOn ) // associative rule assimilaiton is off.
          return;
      // offline training.
      if( emOn && Math.random() < emTrainNacsProb ){
          EMNacsSample[] nacsSamples = gks.getEmNacsSamples( netIdx );
          offlineTrain( nacsSamples );
      }
  }

/**************************** PRIVATE METHODS *********************************/
  // offline training AMN using NACS EM.
  private void offlineTrain( EMNacsSample[] samples )
  {
      int emIterNum = global.getEmTrainIterNum();
      int emSampleNum = samples.length;
      /*
      System.out.println( "the iter num = " + emIterNum );
      System.out.println( "the offline NACS samples = " );
      for( int i = 0; i < samples.length; i++ )
            System.out.println( samples[i] );
      */
      Vector sands = new Vector( emSampleNum * emIterNum );
      for( int j = 0; j < emIterNum; j++ )
        for( int k = 0; k < emSampleNum; k++ )
             sands.addElement( new Integer(k) );

      /*
      System.out.println( "AMN update, vector num = " + sands.size() );
      setInput( global.getNACS().getGks().getChunk(0).getFeatures() );
      forwardPass();
      System.out.println( "before training, given the input, the outputs = \n" );
      CommonMethods.showValues( getOutput() );
      */


      while( !sands.isEmpty() ) {
        int idx = ((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
        setInput( samples[idx].getInput().getFeatures() );
        forwardPass();
        setDesiredOutput( samples[idx].getOutput().getFeatures() );
        backwardPass();
      }

      /*
      System.out.println( "after training, the given input = \n" + global.getNACS().getGks().getChunk(0) );
      setInput( global.getNACS().getGks().getChunk(0).getFeatures() );
      System.out.println( "the output = \n" );
      forwardPass();
      CommonMethods.showValues( getOutput() );
      */
      /*
      for( int i = 0; i < samples.length; i++ ){
           System.out.println( "the sample = \n" + samples[i] );
           setInput( samples[i].getInput().getFeatures() );
           System.out.println( "the output = \n" );
           forwardPass();
           CommonMethods.showValues( getOutput() );
      }
      */
  }

  private int searchFeature( int type,  int order, short[][] info )
  {
      int idx = -1;
      for( int i = 0; i < info.length; i++ ){
           if( info[ i ][ 0 ] == type && info[ i ][ 1 ] == order ){
               idx = i;
               break;
           }
      }
      return idx;
  }
}