package clarion.nacs;
import java.util.Arrays;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class AbsEM {

  final int SAMPLE_TYPE_NUM = 2;
  final int AFN_TYPE = 0, SFN_TYPE = 1;
  Global global;
  int netIdx;
  int ACTION_NUM, REWARD_NUM;
  int[] INPUT_NUMS, OUTPUT_NUMS;
  //short acsActionReps[][];
  int   maxSampleNum;
  Vector stateSamples;
  Vector actionSamples;
  Vector rewardSamples;
  int    aemTrainSampleNum;
  int    nextPotentialStateNum;
  double stateDistribution[];
  Chunk  potentialNextStates[];
  AbsEMSample selectedSamples[];
  ProbBPNet[] nets;
  double temperature;
  StochasticDecider decider;

  public AbsEM( int nIdx, Global g )
  {
     netIdx = nIdx;
     global = g;
     maxSampleNum = global.getAemStateNumBound();
     //ACTION_NUM = global.getAcsActionNum( netIdx );
     ACTION_NUM = maxSampleNum;
     //REWARD_NUM = global.getAcsRewardNum( netIdx );
     REWARD_NUM = maxSampleNum;
     INPUT_NUMS = new int[ SAMPLE_TYPE_NUM ];
     OUTPUT_NUMS = new int[ SAMPLE_TYPE_NUM ];
     INPUT_NUMS[ AFN_TYPE ] = maxSampleNum;      // state.
     OUTPUT_NUMS[ AFN_TYPE ] = ACTION_NUM;       // action.
     INPUT_NUMS[ SFN_TYPE ] = INPUT_NUMS[AFN_TYPE] + ACTION_NUM;  // state + action.
     OUTPUT_NUMS[ SFN_TYPE ] = INPUT_NUMS[AFN_TYPE] + REWARD_NUM; // state + reward.
     nets = new ProbBPNet[ SAMPLE_TYPE_NUM ];
     nets[AFN_TYPE] = new ProbBPNet(INPUT_NUMS[AFN_TYPE],OUTPUT_NUMS[AFN_TYPE],
                          global.getAemTrainedByEmBPLearningRate());
     nets[SFN_TYPE] = new ProbBPNet(INPUT_NUMS[SFN_TYPE],OUTPUT_NUMS[SFN_TYPE],
                          global.getAemTrainedByEmBPLearningRate());
     stateSamples = new Vector( maxSampleNum );
     actionSamples = new Vector( maxSampleNum );
     rewardSamples = new Vector( maxSampleNum );
     aemTrainSampleNum = global.getAemTrainAcsSampleNum();
     temperature = global.getAemTrainAcsOutputTemp();
     decider = new StochasticDecider( temperature );
     nextPotentialStateNum = global.getAemPotentialNextStateNum();
  }

  public AbsEMSample[] getSamples()
  {
     int sampleTrainNum = Math.min( stateSamples.size(), aemTrainSampleNum );
     Vector sands = new Vector( stateSamples.size() );
     for( int i = 0; i < stateSamples.size(); i++ )
          sands.addElement( new Integer(i) );
     selectedSamples = new AbsEMSample[ sampleTrainNum ];
     for( int i = 0; i < sampleTrainNum; i++ ) {
          int stateIdx=((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
          double[] inputs = new double[ INPUT_NUMS[AFN_TYPE] ];
          for( int j = 0; j < inputs.length; j++ )
               inputs[ j ] = 0.0;
          inputs[ stateIdx ] = 1.0;
          double[] outputs = new double[ OUTPUT_NUMS[AFN_TYPE] ];
          nets[ AFN_TYPE ].setInputs( inputs );
          nets[ AFN_TYPE ].forward();
          nets[ AFN_TYPE ].getOutputs( outputs );
          int actionIdx = decider.getRecommendation( outputs, 0, actionSamples.size() );
          inputs = new double[ INPUT_NUMS[SFN_TYPE] ];
          for( int j = 0; j < inputs.length; j++ )
               inputs[ j ] = 0.0;
          inputs[ stateIdx ] = 1.0;
          inputs[ maxSampleNum + actionIdx ] = 1.0;
          outputs = new double[ OUTPUT_NUMS[SFN_TYPE] ];
          nets[ SFN_TYPE ].setInputs( inputs );
          nets[ SFN_TYPE ].forward();
          nets[ SFN_TYPE ].getOutputs( outputs );
          int nextStateIdx = decider.getRecommendation(outputs, 0, stateSamples.size());
          int rewardIdx = decider.getRecommendation(outputs, maxSampleNum, maxSampleNum + rewardSamples.size());
          fetchStateDistribution( outputs );
          try{
             Chunk curState = (Chunk)(((Chunk)stateSamples.elementAt(stateIdx)).clone());
             Chunk nextState = (Chunk)(((Chunk)stateSamples.elementAt(nextStateIdx)).clone());
             Chunk action = (Chunk)(((Chunk)actionSamples.elementAt(actionIdx)).clone());
             double reward = ((Double)rewardSamples.elementAt(rewardIdx)).doubleValue();
             selectedSamples[i] = new AbsEMSample( curState, action, nextState, reward,
                                                   stateDistribution, potentialNextStates );
          }
          catch( Exception e){
                 System.out.println( "clone error from AbsEM!" );
          }
     }
     return selectedSamples;
  }

  public void offlineTrain( EMAcsSample item )
  {
     if( stateSamples.size() > maxSampleNum )
         return;
     // decompose different components of the item.
     Chunk curState = item.getCurState();
     Chunk nextState = item.getNextState();
     Chunk action = item.getCurAction();
     double reward = item.getReward();
     int stateIdx = global.EMPTY;
     int nextStateIdx = global.EMPTY;
     int actionIdx = global.EMPTY;
     int rewardIdx = global.EMPTY;
     try{
         // map current state.
         for( int i = 0; i < stateSamples.size(); i++ ) {
              if( stateSamples.elementAt(i).equals(curState)) {
                  stateIdx = i;
                  break;
              }
         }
         if( stateIdx == global.EMPTY ) {
             stateSamples.addElement( curState.clone() );
             stateIdx = stateSamples.size() - 1;
         }
         // map next state.
         for( int i = 0; i < stateSamples.size(); i++ ){
           if( stateSamples.elementAt(i).equals(nextState)){
               nextStateIdx = i;
               break;
            }
         }
         if( nextStateIdx == global.EMPTY ) {
             stateSamples.addElement( nextState.clone() );
             nextStateIdx = stateSamples.size() - 1;
         }
         // map action.
         for( int i = 0; i < actionSamples.size(); i++ ){
           if( actionSamples.elementAt(i).equals(action)){
               actionIdx = i;
               break;
           }
         }
         if( actionIdx == global.EMPTY ) {
             actionSamples.addElement( action.clone() );
             actionIdx = actionSamples.size() - 1;
         }
         // map reward.
         Double rewardSample = new Double(reward);
         for( int i = 0; i < rewardSamples.size(); i++ ){
           if( rewardSamples.elementAt(i).equals(rewardSample)){
               rewardIdx = i;
               break;
           }
         }
         if( rewardIdx == global.EMPTY ) {
             rewardSamples.addElement( rewardSample );
             rewardIdx = rewardSamples.size() - 1;
         }
     }
     catch( Exception e ) {
         System.out.println( "clone error FROM AbsEM!!" );
     }
     // train nets.
     for( int i = 0; i < SAMPLE_TYPE_NUM; i++ ) {
          // fill input
          double[] inputs = new double[ INPUT_NUMS[i] ];
          for( int j = 0; j < inputs.length; j++ )
               inputs[ j ] = 0.0;
          inputs[ stateIdx ] = 1.0;
          if( i == SFN_TYPE )
              inputs[ maxSampleNum + actionIdx ] = 1.0;
          nets[ i ].setInputs( inputs );
          // fill output
          double[] outputs = new double[ OUTPUT_NUMS[i] ];
          for( int j = 0; j < outputs.length; j++ )
               outputs[ j ] = 0.0;
          if( i == AFN_TYPE )
              outputs[ actionIdx ] = 1.0;
          else{
              outputs[ nextStateIdx ] = 1.0;
              outputs[ maxSampleNum + rewardIdx ] = 1.0;
          }
          nets[ i ].setDesiredOutputs( outputs );
          // training.
          nets[ i ].forward();
          nets[ i ].backward();
     }
  }

  public String testing()
  {
     int stateSampleNum = stateSamples.size();
     int actionSampleNum = actionSamples.size();
     int rewardSampleNum = rewardSamples.size();
     String disStr = "AEM distributions \n";
     for( int i = 0; i < stateSampleNum; i++ ) {
          double[][] stateProbs = new double[ actionSampleNum ][ stateSampleNum ];
          double[] actionProbs = new double[ actionSampleNum ];
          double[][] rewardProbs = new double[ actionSampleNum ][ rewardSampleNum ];
          int stateIdx = i;
          disStr += "Current state ---- " + (Chunk)(stateSamples.elementAt(stateIdx));
          // get action distributions.
          disStr += "Action distribution:  \n";
          double[] inputs = new double[ INPUT_NUMS[AFN_TYPE] ];
          double[] outputs = new double[ OUTPUT_NUMS[AFN_TYPE] ];
          for( int j = 0; j < inputs.length; j++ )
               inputs[ j ] = 0.0;
          inputs[ stateIdx ] = 1.0;
          nets[ AFN_TYPE ].setInputs( inputs );
          nets[ AFN_TYPE ].forward();
          nets[ AFN_TYPE ].getOutputs( outputs );
          double sum = 0.0;
          for( int j = 0; j < actionSampleNum; j++ )
               sum += outputs[ j ];
          for( int j = 0; j < actionSampleNum; j++ )
               actionProbs[j] = outputs[ j ] / sum;
          for( int j = 0; j < actionSampleNum; j++ ){
               disStr += "action -- " + (Chunk)(actionSamples.elementAt(j));
               disStr += String.valueOf(actionProbs[j]) + "\n";
          }
          // get next state and reward distributions.
          disStr += "Next state distribution: \n";
          for( int j = 0; j < actionSampleNum; j++ ){
               int actionIdx = j;
               disStr += "action -- " + (Chunk)(actionSamples.elementAt(actionIdx));
               inputs = new double[ INPUT_NUMS[SFN_TYPE] ];
               outputs = new double[ OUTPUT_NUMS[SFN_TYPE] ];
               for( int k = 0; k < inputs.length; k++ )
                    inputs[ k ] = 0.0;
               inputs[ stateIdx ] = 1.0;
               inputs[ maxSampleNum + actionIdx ] = 1.0;
               nets[ SFN_TYPE ].setInputs( inputs );
               nets[ SFN_TYPE ].forward();
               nets[ SFN_TYPE ].getOutputs( outputs );
               sum = 0.0;
               for( int k = 0; k < stateSampleNum; k++ )
                    sum += outputs[ k ];
               for( int k = 0; k < stateSampleNum; k++ )
                    stateProbs[actionIdx][k] = outputs[k] / sum;
               sum = 0.0;
               for( int k = 0; k < rewardSampleNum; k++ )
                    sum += outputs[ maxSampleNum + k ];
               for( int k = 0; k < rewardSampleNum; k++ )
                    rewardProbs[actionIdx][k] = outputs[maxSampleNum+k] / sum;
               for( int k = 0; k < stateSampleNum; k++ ){
                    disStr += "next state - " + (Chunk)(stateSamples.elementAt(k));
                    disStr += String.valueOf( stateProbs[actionIdx][k] ) + "\n";
               }
               for( int k = 0; k < rewardSampleNum; k++ ){
                    disStr += "reward -- " + ((Double)rewardSamples.elementAt(k)).doubleValue();
                    disStr += String.valueOf( rewardProbs[actionIdx][k] ) + "\n";
               }
          }
    }
    //System.out.println( disStr );
    return disStr;
  }

/**************************** private methods *********************************/
  private void fetchStateDistribution( double[] arr )
  {
    int pStateNum = Math.min( stateSamples.size(), nextPotentialStateNum );
    stateDistribution = new double[ pStateNum ];
    potentialNextStates = new Chunk[ pStateNum ];
    double[] outputs = new double[ stateSamples.size() ];
    for( int i = 0; i < stateSamples.size(); i++ )
         outputs[ i ] = arr[ i ];
    for( int i = 0; i < pStateNum; i++ ) {
         int maxLoc = -1;
         double maxValue = -1.0;
         for( int j = 0; j < stateSamples.size(); j++ )
              if( outputs[j] > maxValue ){
                  maxValue = outputs[j];
                  maxLoc = j;
              }
         outputs[maxLoc] = -1.0;
         stateDistribution[i] = maxValue;
         try{
           potentialNextStates[i] = (Chunk)(((Chunk)(stateSamples.elementAt(maxLoc))).clone());
         }
         catch( Exception e){
                System.out.println( "clone error! from AEM !!!" );
         }
    }
    double sum = 0.0;
    for( int i = 0; i < pStateNum; i++ )
         sum += stateDistribution[i];
    for( int i = 0; i < pStateNum; i++ )
         stateDistribution[i] /= sum;
  }
}