package clarion.nacs;
import java.util.Vector;
import java.util.Arrays;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;
import clarion.tools.linkedlist.*;

public class AssocRule implements CloneableListElement {

  Global global;
  TaskClarion taskClarion;
  int   age;
  short condition[];
  short conclusion;
  double c14, c15;  // used for calculating the support by intermixed reasoning type.
  double support;
  RuleAttributes attrib;
 /*********************** rule feature need be added.**************************/

  public AssocRule( short[] con, short conc, Global global )
  {
     condition = new short[ con.length ];
     for( int i = 0; i < con.length; i++ )
          condition[ i ] = con[ i ];
     conclusion = conc;
     this.global = global;
     this.taskClarion = global.getTaskClarion();
     c14 = global.getRBR_BalConstant();
     c15 = global.getSBR_BalConstant();
     attrib = new RuleAttributes( global );
  }

  // is chunkIdx in the condition
  public boolean isInCondition( int cond )
  {
    for( int i = 0; i < condition.length; i++ )
         if( condition[i] == (int)cond )
             return true;
     return false;
  }

  // is con the condition.
  public boolean isCondition( short[] con )
  {  // con is sorted.
     if( condition.length != con.length )
         return false;
     for( int i = 0; i < condition.length; i++ )
          if( condition[i] != con[i] )
              return false;
     return true;
  }

  public int getAge()
  {
     return age;
  }

  public int getConclusion()
  {
     return conclusion;
  }

  public int getConditionNum()
  {
     return condition.length;
  }

  public short[] getCondition()
  {
     return condition;
  }

  public void getCondition( short[] con )
  {
     for( int i = 0; i < con.length; i++ )
          con[ i ] = condition[ i ];
  }

  public double getBLA()
  {
     return attrib.getBLA();
  }

  public void resetAge()
  {
     age = 0;
  }

  public void increaseAge()
  {
     age++;
  }

  public void setConclusion( short conc )
  {
     conclusion = conc;
  }

  public void setCondition( short[] con )
  {
     condition = new short[ con.length ];
     for( int i = 0; i < con.length; i++ )
          condition[ i ] = con[ i ];
  }

  public double getSupport()
  {
     return support;
  }

  public void calculateSupport( Vector con, double[] weights )
  {
     support = 0.0;
     double ruleSupport = 0.0;
     double simiSupport = 0.0;
     // calculate rule support.
     double[] conditionStrengths = new double[ condition.length ];
     for( int i = 0; i < condition.length; i++ )
          conditionStrengths[ i ] = 0.0;
     for( int i = 0; i < con.size(); i++ ){
          int idx = ( (Integer)con.elementAt(i) ).intValue();
          for( int j = 0; j < condition.length; j++ ){
               if( idx == condition[ j ] ){
                   GKSChunk chunk = global.getNACS().getGks().getChunk(idx);
                   conditionStrengths[ j ] = chunk.getStrength();
                   break;
               }
          }
     }
     for( int i = 0; i < weights.length; i++ )
          ruleSupport += weights[ i ] * conditionStrengths[ i ];
     // calculate similarity support.
     double[] similarities = new double[ condition.length ];
     for( int i = 0; i < condition.length; i++ ){
          similarities[i] = 0.0;
          GKSChunk destChunk = global.getNACS().getGks().getChunk( condition[i] );
          for( int j = 0; j < con.size(); j++ ){
               int idx = ( (Integer)con.elementAt(j) ).intValue();
               GKSChunk souChunk = global.getNACS().getGks().getChunk( idx );
               double simi = souChunk.getStrength() * taskClarion.getSimilarity( souChunk, destChunk );
               if( simi > similarities[i] )
                   similarities[i] = simi;
          }
     }
     simiSupport = 1.0;
     for( int i = 0; i < similarities.length; i++ )
          if( similarities[i] < simiSupport )
              simiSupport = similarities[i];
     // combine the supports from both rule-based and similarity-based.

     //System.out.println("rule support = " + ruleSupport + ", simi support = " + simiSupport );

     support = Math.max( c14 * ruleSupport, c15 * simiSupport );
  }

  public void increaseUsage( long currentTime )
  {
     attrib.updateUsages( (int)currentTime );
  }

  public double calcBLA( long currentTime )
  {
     attrib.calcBLA( (int)currentTime );
     return attrib.getBLA();
  }

  public int calcAssocRuleAppTime()
  {
     attrib.calcAssocRuleAppTime();
     return attrib.getAssocRuleAppTime();
  }

/************************** OVERRIDED METHODS ********************************/
  public String toString()
  {
     String str = "( ";
     for( int i = 0; i < condition.length; i++ )
          str += "chunk_" + String.valueOf( condition[ i ] ) + " ";
     str += ") -----> chunk_" + String.valueOf( conclusion ) + "\n";

     return str;
  }

  public int hashCode()
  {
     int sum = 0;
     for( int i = 0; i < condition.length; i++ )
          sum += condition[ i ];
     sum += conclusion;

     return sum;
  }

  public boolean equals( Object other )
  {
     if( other == null || !( other instanceof AssocRule ) )
         return false;

     AssocRule otherRule = (AssocRule)other;
     if( conclusion != otherRule.getConclusion() )
         return false;
     return isCondition( otherRule.getCondition() );
  }

  public Object clone() throws CloneNotSupportedException
  {
     AssocRule rule = new AssocRule( condition, conclusion, global );
     return rule;
  }
}
