/*
   This class AssocRuleGrp is a associative rule group formed by same
   conclusion.
*/
package clarion.nacs;
import java.util.Vector;
import clarion.system.*;
import clarion.tools.linkedlist.*;

public class AssocRuleGrp{

  Global global;
  int groupID;
  int MAX_RULE_AGE;
  boolean emOn;
  double threshold_ae;
  short activeRuleConditions[][];
  int assocRuleAppTime;
  long currentTime;
  LinkedList rules;

  public AssocRuleGrp( int id, Global g )
  {
     groupID = id;
     global = g;
     emOn = global.isEMon();
     threshold_ae = global.getAssocExtThreshold();
     rules = new LinkedList();

  }

  public int getGroupID()
  {
     return groupID;
  }

  public int getRuleNum()
  {
     return rules.getCount();
  }

  public List getRules()
  {
     return rules;
  }

  public short[][] getActiveRuleConditions()
  {
     return activeRuleConditions;
  }

  public int getAssocRuleAppTime()
  {
     return assocRuleAppTime;
  }

  public void setMaxRuleAge( int age )
  {
     MAX_RULE_AGE = age;
  }

  public void setCurrentTime( long curTime )
  {
     currentTime = curTime;
  }

  public void setRules( LinkedList list )
  {
     try{
         rules = (LinkedList)( list.clone() );
     }
     catch(  CloneNotSupportedException e ){
         System.out.println( "clone error " );
     }
  }

  public boolean insertRule( short[] condition, short conclusion )
  {
     AssocRule newRule = new AssocRule( condition, conclusion, global );
     for( int i = 0; i < rules.getCount(); i++ ){
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          if( rule.equals( newRule ) ){   // the rule is evocated.
              rule.resetAge();
              if( global.isAssocBlaOn() && global.getAssocBlaOption() == Global.VARIED_OPTION )
                  rule.increaseUsage( currentTime );
              return false;
          }
     }
     rules.insertTail( newRule );
     return true;
  }

  public boolean removeRule( short[] condition, short conclusion )
  {
     boolean retVal = false;
     int i = 0;
     while( i < rules.getCount() ) {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          // note: 'condition' will be sorted in this method.
          if( rule.isCondition( condition ) ) {
              rules.removeFrom( i );
              retVal = true;
          }
          else
              i++;
     }
     return retVal;
  }

  // remove all of the rules has chunkIdx as one of its condition.
  public boolean removeCondition( int chunkIdx )
  {
     boolean retVal = false;
     int i = 0;
     while( i < rules.getCount() )
     {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          if( rule.isInCondition( chunkIdx ) )
          {
              rules.removeFrom( i );
              retVal = true;
          }
          else
              i++;
     }
     return retVal;
  }

  public double drawConclusion( Vector con )
  {
     if( global.isRTOn() )
         assocRuleAppTime = 0;
     double support = 0.0;
     for( int i = 0; i < rules.getCount(); i++ ) {
          AssocRule rule = ( AssocRule )rules.getFrom( i );
          double[] weights = global.getTaskClarion().getAssocWeights( rule.getConditionNum() );
          rule.calculateSupport( con, weights );
          double val = rule.getSupport();

          //System.out.println( "the support = " + val );

          if( val > support )
              support = val;
          if( val > 0.0 ){  // current rule is evocated.
              rule.resetAge();
              if( global.isAssocBlaOn() && global.getAssocBlaOption() == Global.VARIED_OPTION ){
                  rule.calcBLA( currentTime );
                  if( global.isRTOn() )
                      assocRuleAppTime += rule.calcAssocRuleAppTime();
                  rule.increaseUsage( currentTime );
              }
          }
     }
     // get all of the activated rules
     if( emOn ){
         short[][] conditions = new short[rules.getCount()][];
         int activeRuleCount = 0;
         for( int i = 0; i < rules.getCount(); i++ ) {
              AssocRule rule = ( AssocRule )rules.getFrom( i );
              if( rule.getSupport() > threshold_ae ){
                  conditions[activeRuleCount] = new short[rule.getConditionNum()];
                  rule.getCondition( conditions[activeRuleCount++] );
              }
         }
         activeRuleConditions = new short[activeRuleCount][];
         for( int i = 0; i < activeRuleCount; i++ ){
              activeRuleConditions[i] = new short[conditions[i].length];
              for( int j = 0; j < conditions[i].length; j++ ){
                   activeRuleConditions[i][j] = conditions[i][j];

                   //System.out.println( "the rule condition = " + conditions[i][j] );
              }
         }
     }
     return support;
  }

  // search a rule with the condition.
  public boolean searchRule( short[] condition )
  {
     boolean retVal = false;
     if( global.isRTOn() )
         assocRuleAppTime = 0;
     for( int i = 0; i < rules.getCount(); i++ ) {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          if( rule.isCondition( condition ) ){
              rule.resetAge();
              if( global.isAssocBlaOn() && global.getAssocBlaOption() == Global.VARIED_OPTION ){
                  rule.calcBLA( currentTime );
                  if( global.isRTOn() )
                      assocRuleAppTime += rule.calcAssocRuleAppTime();
                  rule.increaseUsage( currentTime );
              }
              retVal = true;
              break;
          }
     }
     return retVal;
  }

/*************************** CHUNK UPDATE METHODS *****************************/
  public void update()
  {
     /*
     for( int i = 0; i < rules.getCount(); i++ )
     {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          rule.increaseAge();
          if( rule.getAge() >= MAX_RULE_AGE )
          {
              rules.removeFrom( i );
              i--;
          }
     }
     */
     int i = 0;
     while( i < rules.getCount() )
     {
          AssocRule rule = (AssocRule)( rules.getFrom( i ) );
          rule.increaseAge();
          if( rule.getAge() >= MAX_RULE_AGE ){
              System.out.println( "the removed rule = " + rule );

              rules.removeFrom( i );
          }
          else
              i++;
     }
  }

/************************** OVERRIDED METHODS ********************************/
  public String toString()
  {
     String str = "";
     for( int i = 0; i < rules.getCount(); i++ ){
          str += "Rule_" + String.valueOf( i + 1 ) + " is: ";
          str += (AssocRule)( rules.getFrom( i ) );
     }
     return str;
  }

  public Object clone() throws CloneNotSupportedException
  {
      AssocRuleGrp ruleGrp = new AssocRuleGrp( groupID, global );
      ruleGrp.setRules( rules );
      return ruleGrp;
  }

}