package clarion.nacs;
import java.util.Vector;
import java.util.Arrays;
import clarion.system.*;
import clarion.tools.*;
import clarion.common.*;

public class EM {

  final int ACS = 0, AEM = 1;
  final double EMPTY_BLA = -1.0;
  Global global;
  int netIdx;
  int emType, amnType, chunkTypeNum;
  int MAX_ITEM_NUM, curItemNum;
  long   currentTime;
  int    blaOption;
  double BLA, blaParas[];
  double threshold_EM, itemBlas[];
  int    uselessItemNum, uselessItems[];
  Vector existChunkList, emptyChunkList;
  int combChunkDimOffsets[];
  int chunkNacsTypes[];
  String chunkNacsLabels[];
  int stepChunkDimOffsets[];
  int chunkAcsTypes[];
  String chunkAcsLabels[];
  //short acsActionReps[][];
  EMAcsSample[] acsSamples;
  EMNacsSample[] nacsSamples;
  EMChunk[] items;
  StochasticDecider itemDecider;

  public EM( int nIdx, Global g, int type )
  {
      netIdx = nIdx;
      global = g;
      emType = type;
      if( emType == Global.NACS )
          amnType = global.getAmnType( netIdx );
      MAX_ITEM_NUM = global.getEmCapacity();
      threshold_EM = global.getThreshold_EM();
      itemBlas = new double[ MAX_ITEM_NUM ];
      blaOption = global.getEmBlaOption( emType );
      if( blaOption == global.CONSTANT_OPTION )
          BLA = global.getEmBla( emType );
      else
          blaParas = global.getEmBlaParas( emType );
      existChunkList = new Vector( MAX_ITEM_NUM );
      emptyChunkList = new Vector( MAX_ITEM_NUM );
      initChunkLists();
      items = new EMChunk[ MAX_ITEM_NUM ];
      itemDecider = new StochasticDecider( 0.1 );

      combChunkDimOffsets =  new int[ 3 ];
      if( emType == Global.NACS ){
          int cueDimNum = global.getAmnInputDimNum( netIdx );
          int retrievalDimNum = global.getAmnOutputDimNum( netIdx );
          combChunkDimOffsets[ 0 ] = 0;
          combChunkDimOffsets[ 1 ] = cueDimNum;
          combChunkDimOffsets[ 2 ] = combChunkDimOffsets[ 1 ] + retrievalDimNum;
          chunkNacsTypes = new int[ 2 ];
          chunkNacsTypes[ 0 ] = Global.CHK_CUE;
          chunkNacsTypes[ 1 ] = Global.CHK_RETRIEVAL;
          chunkNacsLabels = new String[ 2 ];
          chunkNacsLabels[ 0 ] = Global.CHUNK_TYPE_LABELS[ global.CHK_CUE ];
          chunkNacsLabels[ 1 ] = Global.CHUNK_TYPE_LABELS[ global.CHK_RETRIEVAL ];
      }
      stepChunkDimOffsets = new int[ 3 ];
      if( emType == Global.NACS ){
          int stateDimNum = global.getIdnInputDimNum( netIdx );
          int actionDimNum = global.getIdnOutputDimNum( netIdx );
          //int rewardDimNum = global.getAcsRewardDimNum( netIdx );
          int rewardDimNum = 1;
          stepChunkDimOffsets = new int[ global.STEP_COMP_NUM + 1 ];
          stepChunkDimOffsets[ 0 ] = 0;
          stepChunkDimOffsets[ 1 ] = stateDimNum;
          stepChunkDimOffsets[ 2 ] = stepChunkDimOffsets[ 1 ] + actionDimNum;
          if( global.getIdnLearningOption(netIdx) == global.QL ) // next state exist
              stepChunkDimOffsets[ 3 ] = stepChunkDimOffsets[ 2 ] + stateDimNum;
          else // next state doesn't exist
              stepChunkDimOffsets[ 3 ] = stepChunkDimOffsets[ 2 ];
          stepChunkDimOffsets[ 4 ] = stepChunkDimOffsets[ 3 ] + rewardDimNum;
          chunkAcsTypes = new int[ global.STEP_COMP_NUM ];
          chunkAcsTypes[ 0 ] = Global.CHK_STATE;
          chunkAcsTypes[ 1 ] = Global.CHK_ACTION;
          chunkAcsTypes[ 2 ] = Global.CHK_STATE;
          chunkAcsTypes[ 3 ] = Global.CHK_REWARD;
          chunkAcsLabels = new String[ global.STEP_COMP_NUM ];
          chunkAcsLabels[ 0 ] = Global.CHUNK_TYPE_LABELS[ global.CHK_STATE ];
          chunkAcsLabels[ 1 ] = Global.CHUNK_TYPE_LABELS[ global.CHK_ACTION ];
          chunkAcsLabels[ 2 ] = Global.CHUNK_TYPE_LABELS[ global.CHK_STATE ];
          chunkAcsLabels[ 3 ] = Global.CHUNK_TYPE_LABELS[ global.CHK_REWARD ];
      }
      //acsActionReps = global.getAcsActionRepresentations( netIdx );
  }

  // initialize the list of existing or empty chunks.
  public void initChunkLists()
  {
      existChunkList.removeAllElements();
      emptyChunkList.removeAllElements();
      for( int i = 0; i < MAX_ITEM_NUM; i++ )
      {
           Integer elem = new Integer( i );
           emptyChunkList.addElement( elem );
      }
  }

  public void reinit()
  {
     initChunkLists();
  }

  // trainType: ACS or AEM.
  public EMAcsSample[] getAcsSamples( int trainType )
  {
     int trainSampleNum;
     if( trainType == ACS )
         trainSampleNum = global.getEmTrainItemNum();
     else // AEM
         trainSampleNum = global.getAemTrainedByEmItemNum();
     // search all of the STEP chunks.
     int[] stepChunkIndices = searchStepChunks();
     int sampleNum = Math.min( stepChunkIndices.length, trainSampleNum );
     //System.out.println("EM sample num = " + stepChunkIndices.length );
     Vector sands = new Vector( stepChunkIndices.length );
     for( int i = 0; i < stepChunkIndices.length; i++ )
          sands.addElement( new Integer(stepChunkIndices[i]) );
     acsSamples = new EMAcsSample[ sampleNum ];
     for( int i = 0; i < sampleNum; i++ ) {
          int idx=((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
          Chunk curState = fetchAcsChunk( global.CUR_STATE, items[idx] );
          Chunk curAction = fetchAcsChunk( global.CUR_ACTION, items[idx] );
          Chunk nextState = fetchAcsChunk( global.NEXT_STATE, items[idx] );
          double reward = fetchAcsChunk( global.REWARD, items[idx] ).getFeature(0).getActivations()[0];
          //int actionIdx = fetchAcsIndex( global.CUR_ACTION, curAction );
          //int rewardIdx = fetchAcsIndex( global.REWARD,reward );
          acsSamples[i] = new EMAcsSample( curState,curAction,nextState,reward );
     }
     return acsSamples;
  }

  public EMNacsSample[] getNacsSamples()
  {
     int trainSampleNum = global.getEmTrainItemNum();
     int sampleNum = Math.min( existChunkList.size(), trainSampleNum );
     Vector sands = new Vector( existChunkList.size() );
     for( int i = 0; i < existChunkList.size(); i++ )
          sands.addElement( new Integer(i) );
     nacsSamples = new EMNacsSample[ sampleNum ];
     for( int i = 0; i < sampleNum; i++ ) {
          int idx=((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
          if( amnType == global.HAM ){
              Chunk condition = fetchNacsChunk( 0, items[idx] );
              Chunk conclusion = fetchNacsChunk( 1, items[idx] );
              nacsSamples[i] = new EMNacsSample( condition, conclusion );
           }
           else{
              int chunkType = global.getAmnInputChunkType(netIdx);
              String chunkLabel = global.CHUNK_TYPE_LABELS[ chunkType ];
              Feature[] features = items[idx].getFeatures();
              Chunk input = new Chunk( chunkType, chunkLabel, features, global );
              nacsSamples[i] = new EMNacsSample( input, input );
           }
     }
     return nacsSamples;
  }

  public void setCurrentTime( long time )
  {
     currentTime = time;
  }

  // check if the BLA of a chunk falls below a certain threshold, should be called first.
  public void checkValid()
  {
     curItemNum = existChunkList.size();
     if( curItemNum == 0 )
         return;
     calcItemBlas();
     uselessItemNum = 0;
     uselessItems = new int[ curItemNum ];
     for( int i = 0; i < curItemNum; i++ ) {
          int idx = (( Integer )existChunkList.elementAt( i )).intValue();
          if( itemBlas[ idx ] < threshold_EM )
              uselessItems[ uselessItemNum++ ] = idx;
     }
     for( int i = 0; i < uselessItemNum; i++ )
          removeItem( uselessItems[i] );
  }

  public void insertItem( Chunk chunk )
  {
    Chunk item = null;
     try {
            item = (Chunk)(chunk.clone());
     }
     catch( Exception e ) {
            System.out.println( "clone error from EM! " );
            //System.exit( 1 );
     }
     curItemNum = existChunkList.size();
     int emptyIdx = -1;
     for( int i = 0; i < curItemNum; i++ ) {
          emptyIdx = (( Integer )existChunkList.elementAt( i )).intValue();
          if( items[ emptyIdx ].equals( item ) ) { // the item exists in EM.
              //System.out.println( "EM item: " + items[emptyIdx] + " == inserted item: " + item );
              if( blaOption == Global.VARIED_OPTION )
                  items[ emptyIdx ].increaseUsage( currentTime );
              itemBlas[ emptyIdx ] = EMPTY_BLA; // here, means bla need updated.
              return;
          }
     }
     // insert this new item.
     if( emptyChunkList.isEmpty() ) { // EM is full
         emptyIdx = searchRemovableItem();
         //System.out.println( "EM is full1" );
     }
     else {
         emptyIdx = (( Integer )( emptyChunkList.remove( 0 ) )).intValue();
         existChunkList.addElement( new Integer( emptyIdx ) );
     }
     items[ emptyIdx ] = new EMChunk( item, currentTime, global );
     items[ emptyIdx ].setBlaOption( blaOption );
     if( blaOption != Global.CONSTANT_OPTION ){
         items[ emptyIdx ].setBlaParams( blaParas );
         items[ emptyIdx ].increaseUsage( currentTime );
     }
     else
         items[ emptyIdx ].setBla( BLA );
     itemBlas[ emptyIdx ] = EMPTY_BLA;  // here, means bla need updated.
  }

/********************************* private methods ****************************/
  private void calcItemBlas()
  {
     curItemNum = existChunkList.size();
     for( int i = 0; i < MAX_ITEM_NUM; i++ )
          itemBlas[ i ] = EMPTY_BLA;
     if( blaOption == global.CONSTANT_OPTION ) {
         for( int i = 0; i < curItemNum; i++ ) {
              int idx = (( Integer )existChunkList.elementAt( i )).intValue();
              itemBlas[ idx ] = BLA;
         }
         return;
     }
     for( int i = 0; i < curItemNum; i++ ) {
          int idx = (( Integer )existChunkList.elementAt( i )).intValue();
          items[ idx ].calcBLA( currentTime );
          itemBlas[ idx ] = items[ idx ].getBLA();
     }
  }

  private void removeItem( int chunkIdx )
  {
     Integer elem = new Integer( chunkIdx );
     existChunkList.removeElement( elem );
     emptyChunkList.addElement( elem );
     itemBlas[ chunkIdx ] = EMPTY_BLA;
  }

  private int searchRemovableItem()
  {
     curItemNum = existChunkList.size();
     int inputNum = 0;
     int[] itemIndices = new int[ curItemNum ];
     double[] inputs = new double[ curItemNum ];
     for( int i = 0; i < MAX_ITEM_NUM; i++ ) {
          if( itemBlas[ i ] > 0.0 ) {
              inputs[ inputNum ] = 1 / itemBlas[ i ];
              itemIndices[ inputNum++ ] = i;
          }
     }
     int idx = itemDecider.getRecommendation( inputs, inputNum );
     return itemIndices[ idx ];
  }

  private int[] searchStepChunks()
  {
     curItemNum = existChunkList.size();
     int[] indices = new int[curItemNum];
     int stepChunkNum = 0;
     for( int i = 0; i < curItemNum; i++ ){
          int idx = ((Integer)(existChunkList.elementAt(i))).intValue();
          if( items[idx].getChunkType() == global.CHK_STEP )
              indices[stepChunkNum++] = idx;
     }
     int[] stepIndices = new int[ stepChunkNum ];
     for( int i = 0; i < stepChunkNum; i++ )
          stepIndices[i] = indices[i];
     return stepIndices;
  }

  private Chunk fetchNacsChunk( int idx, EMChunk item )
  {
     //System.out.println( "the item feature num = " + item.getFeatureNum() );

     int start = combChunkDimOffsets[ idx ];
     int end = combChunkDimOffsets[ idx+1 ];
     Feature[] features = new Feature[ end-start];
     for( int i = 0; i < end - start; i++ )
          features[ i ] = item.getFeature( start + i );
     return new Chunk( chunkNacsTypes[ idx ], chunkNacsLabels[ idx ], features, global );
  }

  private Chunk fetchAcsChunk( int idx, EMChunk item )
  {
     int start = stepChunkDimOffsets[ idx ];
     int end = stepChunkDimOffsets[ idx+1 ];
     if( start == end )
         return (Chunk)null;
     Feature[] features = new Feature[ end-start];
     for( int i = 0; i < end - start; i++ )
          features[ i ] = item.getFeature( start + i );
     return new Chunk( chunkAcsTypes[ idx ], chunkAcsLabels[ idx ], features, global );
  }

  /*
  private int fetchAcsIndex( int type, Chunk chunk )
  {
     int idx = -1;
     int start = stepChunkDimOffsets[ type ];
     int end = stepChunkDimOffsets[ type + 1 ];
     short[] chunkDimValues = new short[ end - start ];
     for( int i = 0; i < end - start; i++ )
          chunkDimValues[i] = ((Feature)( chunk.getFeature(i) )).getValues()[0];
     if( type == global.CUR_ACTION ) {
         for( int i = 0; i < acsActionReps.length; i++ )
              if( Arrays.equals( chunkDimValues, acsActionReps[i] ) ) {
                  idx = i;
                  break;
              }
     }
     else // assume reward has only one feature.
         idx = chunkDimValues[ 0 ];
     return idx;
  }
  */
}