package clarion.nacs;
import java.util.Vector;
import clarion.common.*;
import clarion.tools.*;
import clarion.system.*;

public class GKChunk extends CommonChunk{

   short CONC_IDX;         // the conclusion dimension.
   short cueDims[];        // the active dimensions in cue.
   short dimMatches[];     // record the matched dimensions with the cue.
   float featureWeights[]; // weights used for calculating the activation.
   double activation;      // activation used for selecting the chunk.

   boolean availableFlag;
   boolean activatedFlag;

   public GKChunk( int cId, int type, int dNum, int[] dValNums, int[] dTypes,
                   double initBla, float[] wArr )
   {
      super( cId, type, dNum, dValNums, dTypes, initBla );

      dimMatches = new short[ getDimNum() ];
      featureWeights = new float[ getDimNum() ];
      for( int i = 0; i < getDimNum(); i++ )
           featureWeights[ i ] = wArr[ i ];
   }

   public GKChunk( int cId, int type, int dNum, short[] dValNums, short[] dTypes,
                   double initBla, float[] wArr )
   {
      super( cId, type, dNum, dValNums, dTypes, initBla );

      dimMatches = new short[ getDimNum() ];
      for( int i = 0; i < getDimNum(); i++ )
           featureWeights[ i ] = wArr[ i ];
   }

   public void reinit( int type, Vector con )
   {
      availableFlag = false;
      activatedFlag = false;
      super.reinit( type, con );
   }

/************************ GET AND SET METHODS ********************************/
   public int getConclusionDimIdx( )
   {
       return CONC_IDX;
   }

   public boolean getAvailableFlag()
   {
       return availableFlag;
   }

   public boolean getActivatedFlag()
   {
       return activatedFlag;
   }

   public void getConclusion( short[] result )
   {
       for( int i = 0; i < result.length; i++ )
            result[ i ] = ( (short[])getContent().elementAt( CONC_IDX ) )[ i ];
   }

   public double getActivation( )
  {
       return activation;
   }

   public float[] getFeatureWeights()
   {
       return featureWeights;
  }

   public void getFeatureWeights( float[] arr )
   {
       for( int i = 0; i < featureWeights.length; i++ )
           arr[ i ] = featureWeights[ i ];
   }

   public void setAvailableFlag( boolean flag )
   {
       availableFlag = flag;
   }

   public void setActivatedFlag( boolean flag )
   {
       activatedFlag = flag;
   }

   public void setConclusionDimIdx( short idx )
   {
       CONC_IDX = idx;
   }

   public void setFeatureWeights( double[] arr )
   {
       for( int i = 0; i < featureWeights.length; i++ )
            featureWeights[ i ] = (float)arr[ i ];
   }

   public void setCueDimIndices( short[] indices )
   {
       cueDims = new short[ indices.length ];
       for( int i = 0; i < indices.length; i++ )
            cueDims[ i ] = indices[ i ];
   }

   // check if current content contains "con".
   public boolean contain( int[] con )
   {
        boolean bVal = true;
        for( int i = 0; i < getDimNum(); i++ )
        {
             short idx = (short)con[ i ];
             if( ( (short[])getContent().elementAt(i) )[idx] == 0 )
             {
                    bVal = false;
                    break;
             }
        }
        return bVal;
   }

/******************** CHUNK MATCHING METHODS **********************************/
   public boolean matchCueByRule( Vector cue )
   {
       for( int j = 0; j < getDimNum(); j++ )
            dimMatches[ j ] = 0;

       int matchCount = 0;
       if( !getContent().isEmpty() )
       {
           for( int j = 0; j < cueDims.length; j++ )
           {
                int idx = cueDims[ j ];
                dimMatches[idx] = CommonMethods.contain(
                                  (short[])( cue.elementAt(idx) ),
                                  (short[])( getContent().elementAt(idx) ) );
                if( dimMatches[idx] == 1 )
                    matchCount++;
            }
       }

       if( matchCount == 0 )
           return false;
       else
           return true;
   }

   public boolean matchCueBySimilarity( Vector cue )
   {
       for( int i = 0; i < getDimNum(); i++ )
            dimMatches[ i ] = 0;

       int matchCount = 0;
       if( !getContent().isEmpty() )
       {
           for( int i = 0; i < cueDims.length; i++ )
           {
                int idx = cueDims[ i ];
                short[] arr1 = (short[])( cue.elementAt(idx) );
                short[] arr2 = (short[])( getContent().elementAt(idx) );
                dimMatches[idx] = CommonMethods.contain( arr1, arr2 );
                dimMatches[idx] += CommonMethods.contain( arr2, arr1 );

                if( dimMatches[idx] == 2 )
                {
                    dimMatches[idx] = 1;
                    matchCount++;
                }
                else
                    dimMatches[idx] = 0;
          }
       }

       if( matchCount == 0 )
           return false;
       else
           return true;
   }

   public void calcActivation( int type )
   {
       if( type == 0 )
       {   // rule-based reasoning.
           float val = 0;
           for( int j = 0; j < getDimNum(); j++ )
                val += dimMatches[j] * featureWeights[j];
           activation = val;
       }
       else
       {  // similarity-based reasoning, use the default similarity measure.
           int count = 0;
           for( int j = 0; j < getDimNum(); j++ )
                if( dimMatches[ j ] == 1 )
                    count++;
           activation = (double)count / Math.pow( getDimNum(), 1.0001 );
       }
   }

   public Object clone() throws CloneNotSupportedException
   {
       GKChunk chunk = new GKChunk( getChunkID( ), getChunkType(), getDimNum(), getDimValNums(),
                                    getDimTypes(), getInitBLA(), featureWeights );
       // from CommonChunk class
       chunk.reinit( getChunkType(), getContent() );
       chunk.setCurrentTime( getCurrentTime() );
       chunk.setUsageNum( getUsageNum() );
       chunk.setUsageRecords( getUsageRecords() );
       chunk.setLabel( getLabel() );

       // from this class
       chunk.setConclusionDimIdx( CONC_IDX );
       chunk.setActivatedFlag( activatedFlag );
       chunk.setAvailableFlag( availableFlag );
       return chunk;
   }
}