/*
   this is the new version of NACS top level: GKS.
*/
package clarion.nacs;
import java.util.Arrays;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class GKS{

  final int AMN_IDX = 0, GKS_IDX = 1;
  // global settings.
  final int SCALE_VAL = 100;   // used for selecting chunks to remove.
  final double ALMOST_ZERO = 0.00001;
  Global global;
  TaskClarion taskClarion;
  // variables for reasoning type.
  int    reasoningType;    // reasoning method being used.
  double p_c, p_a, p_ce, p_ae, threshold_ce, threshold_ae;
  double threshold_simi;   // used for similarity reasoning type.
  // variables for reasoning process.
  int    iterationNum;     // # of reasoning iterations to get result.
  double gksActivations[];    // chunk activations by cue.
  double gksConcStrengths[];  // chunk strengths from GKS.
  double amnConcStrengths[];  // chunk strengths from AMN.
  double chunkStrengths[]; //  combined chunk strengths by GKS and AMN.
  int c21, c25;            // used for MCS's minotoring conclusions.
  GKSChunk[] combResults;  // the final combined results.
  Chunk[] allResults;      // all retrievals from both GKS and AMN for monitoring.
  Chunk[] gksResults;      // the top level retrievals for monitoring.
  Chunk[] amnResults;      // the bottom level retrievals for monitoring.
  // variables for chunks.
  int MAX_CHUNK_NUM;    // maximum # of chunks allowed.
  int chunkTypeNum;     // # of chunk type.
  int curChunkNum;      // current # of existing chunks in GKS.
  int chosenChunkIdx;   // the index of chosen chunk.
  // variables for retrieval.
  Feature cue[];             // cue used for retrieval.
  String  cueLabel;          // the label of the cue.
  double  cueActivation;     // the activation of the cue.
  short   cueType;           // type of cue.
  short   concDimIdx;
  // variables for BLA.
  long   currentTime;       // used for calculating BLA.
  // other variables.
  Vector existChunkList;    // indicate current existing chunks.
  Vector emptyChunkList;    // indicate current empty chunks.
  GKSChunk chunks[];        // all the chunk chunks in GKS.
  boolean epiChunkOn;       // episodic chunk in EM or not.
  boolean emOn;             // whether EM is on.
  int  acsNetNum;           // the # of acs nets.
  int  amnNetNum;           // the # of amn nets.
  EM[] acsEMs, nacsEMs;       // EM for storing episodic chunks.
  StochasticDecider nodeDecider;   // used for determining the result.
  // for RT.
  int responseTime;
  boolean chunkBlaOn;
  int chunkBlaOption;

  public GKS( Global g )
  {
      global = g;
      taskClarion = global.getTaskClarion();
      // initialize variables for reasoning.
      reasoningType = global.getGksReasoningMethod();
      p_c = global.getChunkEncodeProb();
      p_a = global.getAssocEncodeProb();
      p_ce = global.getChunkExtProb();
      p_ae = global.getAssocExtProb();
      threshold_ce = global.getChunkExtThreshold();
      threshold_ae = global.getAssocExtThreshold();
      threshold_simi = global.getThreshold_sim();
      iterationNum = global.getGksReasonIterNum();
      // initialize variables for chunks.
      MAX_CHUNK_NUM = global.getGksMaxChunkNum();
      gksActivations = new double[ MAX_CHUNK_NUM ];
      gksConcStrengths = new double[ MAX_CHUNK_NUM ];
      amnConcStrengths = new double[ MAX_CHUNK_NUM ];
      // initialize the lists for managing chunks.
      existChunkList = new Vector( MAX_CHUNK_NUM );
      emptyChunkList = new Vector( MAX_CHUNK_NUM );
      for( int i = 0; i < MAX_CHUNK_NUM; i++ ){
          Integer elem = new Integer( i );
          emptyChunkList.addElement( elem );
      }
      // initialize the chunks.
      chunks = new GKSChunk[ MAX_CHUNK_NUM ];
      if( global.isAssocEncodeExternOn() || global.isChunkEncodeExternOn() )
          encodeExternal();
      epiChunkOn = global.isEpiChunkOn();
      emOn = global.isEMon();
      if( emOn ){
          acsNetNum = global.getAcsNetNum();
          acsEMs = new EM[ acsNetNum ];
          for( int i = 0; i < acsNetNum; i++ )
               acsEMs[ i ] = new EM( i, global, Global.ACS );
          amnNetNum = global.getAmnNum();
          nacsEMs = new EM[ amnNetNum ];
          for( int i = 0; i < amnNetNum; i++ )
               nacsEMs[ i ] = new EM( i, global, Global.NACS );
      }
      // initialize the conclusions for monitoring.
      if( global.isMcsOn() ){
          c21 = global.getMonitorNacsTopConcNum();
          c25 = global.getMonitorNacsBottomConcNum();
      }
      // initialize node decider.
      nodeDecider = new StochasticDecider( global.getGksOutputTemperature() );
      // initialize RT variables.
      chunkBlaOn = global.isNacsChunkBlaOn();
      chunkBlaOption = chunkBlaOn ? global.getChunkBlaOption() : Global.EMPTY;
  }

  public void reinit()
  {
      currentTime = 0;
      existChunkList.removeAllElements();
      emptyChunkList.removeAllElements();
      for( int i = 0; i < MAX_CHUNK_NUM; i++ )
      {
          Integer elem = new Integer( i );
          emptyChunkList.addElement( elem );
      }
      chunks = new GKSChunk[ MAX_CHUNK_NUM ];
  }

/***************************** GET METHODS **********************************/
  public GKSChunk[] getCombinedResults()
  {
      return combResults;
  }

  public Chunk[] getTopLevelMonitorConclusions()
  {
     return gksResults;
  }

  public Chunk[] getBottomLevelMonitorConclusions()
  {
     return amnResults;
  }

  public Chunk[] getAllMonitorConclusions()
  {
     return allResults;
  }

  public EMAcsSample[] getEmAcsSamples( int netIdx, int trainType )
  {
     return acsEMs[netIdx].getAcsSamples( trainType );
  }

  public EMNacsSample[] getEmNacsSamples( int netIdx )
  {
     return nacsEMs[netIdx].getNacsSamples();
  }

  public GKSChunk getChunk( int idx )
  {
     return chunks[idx];
  }

  // get the chunk index with the label.
  public int getChunkIdx( String label )
  {
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 ){
         System.out.println( "no chunk in GKS (label)!" );
         return -1;
     }
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          if( chunks[ idx ].getLabel().compareTo( label ) == 0 ){
              chunks[ idx ].resetAge();
              if( chunkBlaOption == global.VARIED_OPTION )
                  chunks[ idx ].increaseUsage( currentTime );
              return idx;
          }
     }
     System.out.println( "no chunk with the label: " + label );
     return -1;
  }

  public int getChunkIdx( Feature[] features )
  {
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 ){
         System.out.println( "no chunk in GKS (feature)!" );
         return -1;
     }
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          if( chunks[ idx ].isEqual( features ) ){
              chunks[ idx ].resetAge();
              if( chunkBlaOption == global.VARIED_OPTION )
                  chunks[ idx ].increaseUsage( currentTime );
              return idx;
          }
     }
     System.out.println( "no chunk with the features !" );
     return -1;
  }

  public int getRT()
  {
     return responseTime;
  }

  public void resetRT()
  {
     responseTime = 0;
  }

/******************************** SET METHODS  ********************************/
  public void setRetrievalCue( Feature[] content )
  {
     cueType = (short)global.CHK_CUE;
     cueLabel = global.CHUNK_TYPE_LABELS[cueType];
     cue = new Feature[ content.length ];
     for( int i = 0; i < content.length; i++ )
          cue[ i ] = content[ i ];
  }

  public void setRetrievalCue( Feature[] content, int concIdx )
  {
     cueType = (short)global.CHK_CUE;
     cueLabel = global.CHUNK_TYPE_LABELS[cueType];
     cue = new Feature[ content.length ];
     for( int i = 0; i < content.length; i++ )
          cue[ i ] = content[ i ];
     concDimIdx = (short)concIdx;
  }

  public void setCurrentTime( long time )
  {
     currentTime = time;
     for( int i = 0; i < existChunkList.size(); i++ ) {
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          chunks[ idx ].setCurrentTime( currentTime );
     }
  }

/********************** INSERT AND REMOVE CHUNK METHODS ***********************/
  // Vector 'con' should include the feature information.
  public int insertChunk( GKSChunk chunk )
  {
     int availableChunkIdx = 0;
     try{
         if( emptyChunkList.isEmpty() ){
             System.out.println( "max # of chunks is =" + MAX_CHUNK_NUM );

             double[] inputs = new double[ MAX_CHUNK_NUM ];
             for( int i = 0; i < MAX_CHUNK_NUM; i++ )
                  inputs[ i ] = ( (double)chunks[ i ].getAge() ) / SCALE_VAL;
             availableChunkIdx = nodeDecider.getRecommendation( inputs );
             chunks[ availableChunkIdx ] = (GKSChunk)(chunk.clone());
         }
         else{
             availableChunkIdx = (( Integer )( emptyChunkList.remove( 0 ) )).intValue();
             existChunkList.addElement( new Integer( availableChunkIdx ) );
             if( chunks[ availableChunkIdx ] == null )
                 chunks[ availableChunkIdx ] = (GKSChunk)(chunk.clone());
         }
         chunks[ availableChunkIdx ].setAvailableFlag( true );
     }
     catch( Exception e ){
            System.out.println( "max # of chunks is =" + MAX_CHUNK_NUM );

            System.out.println( e + " happens in clone GKS chunk in GKS.java" );
     }

     return availableChunkIdx;
  }

  public int insertChunk( int idnIdx, int amnIdx, int type, String label, Feature[] features )
  {
     int availableChunkIdx = 0;
     if( emptyChunkList.isEmpty() ){
         //System.out.println( MAX_CHUNK_NUM );

         double[] inputs = new double[ MAX_CHUNK_NUM ];
         for( int i = 0; i < MAX_CHUNK_NUM; i++ )
              inputs[ i ] = ( (double)chunks[ i ].getAge() ) / SCALE_VAL;
         availableChunkIdx = nodeDecider.getRecommendation( inputs );
         chunks[ availableChunkIdx ] = new GKSChunk( availableChunkIdx, type,
                                           label, features, idnIdx, amnIdx, global );
     }
     else{
         availableChunkIdx = (( Integer )( emptyChunkList.remove( 0 ) )).intValue();
         existChunkList.addElement( new Integer( availableChunkIdx ) );
         if( chunks[ availableChunkIdx ] == null )
             chunks[ availableChunkIdx ] = new GKSChunk( availableChunkIdx, type,
                                           label, features, idnIdx, amnIdx, global );
     }
     chunks[ availableChunkIdx ].setAvailableFlag( true );
     return availableChunkIdx;
  }

  public void removeChunk( int chunkIdx )
  {
     Integer elem = new Integer( chunkIdx );
     existChunkList.removeElement( elem );
     emptyChunkList.addElement( elem );

     // remove all the rules which has chunkIdx as one of its condition.
     for( int i = 0; i < existChunkList.size(); i++ )
     {
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          chunks[ idx ].removeCondition( chunkIdx );
     }

     chunks[ chunkIdx ].setLabel( "" );
     chunks[ chunkIdx ].setAvailableFlag( false );
  }

  public void insertRule( short[] condition, short conclusion )
  {
     if( chunks[ conclusion ] == null || !chunks[ conclusion ].getAvailableFlag() )
     {
         System.out.println( " cannot insert this rule !" );
         return;
     }

     chunks[ conclusion ].insertRule( condition, conclusion );
  }

  /*
  public void removeRule( short[] condition, short conclusion )
  {
     if( chunks[ conclusion ] == null || !chunks[ conclusion ].getAvailableFlag() )
         System.out.println( " cannot remove such rule !" );

     chunks[ conclusion ].removeRule( condition, conclusion );
  }
  */

/***************************** RETRIEVAL METHOD *******************************/
  public boolean reasoning()
  {
     if( global.isRTOn() )
         responseTime = 0;
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 )
         return false;
     for( int i = 0; i < gksConcStrengths.length; i++ ){
          gksActivations[ i ] = 0.0;
          gksConcStrengths[ i ] = 0.0;
     }
     // activate chunks.
     int activateCount = activateChunks( cue, GKS_IDX );
     /*
     System.out.println( "activated num = " + activateCount );
     System.out.println( "gks activation = " );
     for( int i = 0; i < 12; i++ ){
          System.out.println( "chunk " + i + " = " + gksActivations[i] );
     }
     */

     // no activated chunks.
     if( activateCount == 0 )
         return false;
     //draw conclusion.
     drawConclusion();
     // fill in top level results.
     gksResults = fillInFinalResults( GKS_IDX, gksConcStrengths );

     /*
     System.out.println( "gks strengths = " );
     for( int i = 8; i < 12; i++ ){
          System.out.println( "chunk " + i + " = " + gksConcStrengths[i] );
     }
     */
     /*
     System.out.println( "gks conclusions = " );
     for( int i = 0; i < gksResults.length; i++ ){
          System.out.println( gksResults[i] );
     }
     */

     return true;
  }

  public void bottomUp( int amnIdx, Feature[] amnResult )
  {
     /*
     System.out.println( "the amn retrieval result = " );
     for( int i = 0; i < amnResult.length; i++ )
          System.out.println( amnResult[i] );
     */

     for( int i = 0; i < amnConcStrengths.length; i++ )
          amnConcStrengths[ i ] = 0.0;
     curChunkNum = existChunkList.size();
     // activate all of the compatible chunks.
     activateChunks( amnResult, AMN_IDX );
     // fill in bottom level results.
     amnResults = fillInFinalResults( AMN_IDX, amnConcStrengths );

     /*
     System.out.println( "amn strengths = " );
     for( int i = 8; i < 12; i++ ){
          System.out.println( "chunk " + i + " = " + amnConcStrengths[i] );
     }
     */
     /*
     System.out.println( "amn  conclusions = " );
     for( int i = 0; i < amnResults.length; i++ ){
          System.out.println( amnResults[i] );
     }
     */

     // extract the explicit knowledge.
     extractExplicitKnowledge( amnIdx, amnResult );
  }

  // combine the result from both GKS and AMN.
  public boolean formResults( int retrievalMode )
  {
     curChunkNum = existChunkList.size();
     combResults = null;
     allResults = null;
     // get results.
     int concNum = 0;
     chunkStrengths = new double[ curChunkNum ];
     short[] chunkIndices = new short[ curChunkNum ];
     for( int i = 0; i < curChunkNum; i++ ) {
          chunkStrengths[ i ] = 0.0;
          chunkIndices[ i ] = -1;
     }
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          double maxValue = Math.max( gksConcStrengths[idx], amnConcStrengths[idx] );
          if( maxValue > 0.0 ){
              chunkStrengths[ concNum ] = maxValue;
              chunkIndices[ concNum++ ] = (short)idx;
          }
     }
     // no conclusion has been drawn.
     if( concNum == 0 )
         return false;
     // fill in all the results from GKS and AMN for monitoring.
     allResults = fillInFinalResults( concNum, chunkStrengths, chunkIndices );

     /*
     System.out.println( "final strengths = " );
     for( int i = 0; i < 4; i++ ){
          System.out.println( "chunk " + (i+8) + " = " + chunkStrengths[i] );
     }
     */

     // obtain drawn conclusions.
     try{
         if( retrievalMode == Global.RETRIEVAL_ALL ){
             combResults = new GKSChunk[ concNum ];
             for( int i = 0; i < concNum; i++ ){
                  int idx = chunkIndices[ i ];
                  combResults[ i ] = (GKSChunk)(chunks[ idx ].clone());
                  combResults[ i ].setStrength( chunkStrengths[i] );
             }
         }
         if( retrievalMode == Global.RETRIEVAL_ONE ){
             combResults = new GKSChunk[ 1 ];
             int idx = nodeDecider.getRecommendation( chunkStrengths, 0, concNum );
             combResults[ 0 ] = (GKSChunk)(chunks[ chunkIndices[idx] ].clone());
             combResults[ 0 ].setStrength( chunkStrengths[idx] );
         }
     }
     catch( CloneNotSupportedException e ){
         System.out.println( "clone error " );
     }
     return true;
  }

  // search all of the conclusions with the condition.
  public Vector searchConclusions( short[] conditions )
  {
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 ){
         System.out.println( "No available chunks!" );
         return (Vector)null;
     }
     // retrieve the conditions.
     int retrievalTime = 0;
     for( int i = 0; i < conditions.length; i++ ) {
          //System.out.println("condtion=" + conditions[i]);
          chunks[conditions[i]].resetAge();
          chunks[conditions[i]].calcRetrievalTime();
          retrievalTime += chunks[conditions[i]].getRetrievalTime();
     }
     int concNum = 0;
     Vector vect = new Vector( curChunkNum );
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          if( chunks[ idx ].searchRule( conditions ) ){
              // retrieve conclusions.
              chunks[idx].resetAge();
              chunks[idx].calcRetrievalTime();
              retrievalTime += chunks[idx].getRetrievalTime()+chunks[idx].getAssocRuleAppTime();
              vect.addElement( new Integer(idx) );
              concNum++;
          }
     }
     Vector conclusions = new Vector( concNum );
     Vector sands = new Vector( concNum );
     for( int i = 0; i < concNum; i++ )
          sands.addElement( new Integer( i ) );
     for( int i = 0; i < concNum; i++ ){
          int idx = ((Integer)sands.remove((int)(Math.random()*sands.size()))).intValue();
          conclusions.addElement( vect.elementAt( idx ) );
     }
     if( global.isRTOn() )
         responseTime += retrievalTime;
     return conclusions;
  }

/***************************** GKS UPDATE METHODS *****************************/
  public void update()
  {
     curChunkNum = existChunkList.size();
     if( curChunkNum == 0 ){
         //System.out.println( "No available chunks!" );
         return;
     }
     // search all the chunks should be removed.
     int removeNum = 0;
     int[] removeChunks = new int[ curChunkNum ];
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          boolean needRemoved = chunks[ idx ].update();
          if( needRemoved ){
              //System.out.println("remove chunk = " + chunks[idx] );
              removeChunks[ removeNum++ ] = idx;
          }
     }
     // remove all of the less used chunks.
     for( int i = 0; i < removeNum; i++ )
          removeChunk( removeChunks[ i ] );
     // update EM if exist.
     if( emOn )
         checkEM();
  }

  // add chunks into EM.
  public void updateEM( int type, int idx )
  {
     if( !epiChunkOn )
         return;
     if( type == Global.NACS ){
         int amnIdx = chunks[idx].getAmnIdx();
         if( amnIdx != global.EMPTY )
             nacsEMs[ amnIdx ].insertItem( chunks[idx] );
     }
     if( type == Global.ACS ){
         int idnIdx = chunks[idx].getIdnIdx();
         if( idnIdx != global.EMPTY )
             acsEMs[ idnIdx ].insertItem( chunks[idx] );
     }
  }

/**************************** OVERRIDED METHODS *****************************/
  public String toString()
  {
     curChunkNum = existChunkList.size();

     String str = "GKS( # of chunks = " + String.valueOf( curChunkNum ) + " ) = \n";
     str += "{ ";
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          str += chunks[ idx ];
     }
     str += " } \n";

     return str;
  }

/**************************** OTHER METHODS *****************************/
  public void calcChosenChunkBLA()
  {
     chunks[ chosenChunkIdx ].calcBLA( currentTime );
  }

/******************************* PRIVATE METHODS ******************************/
  private void encodeExternal()
  {
     GKSChunk[] gksChunks = taskClarion.getExternalNacsChunks();
     for( int i = 0; i < gksChunks.length; i++ ){
          if( Math.random() < p_c )
              this.insertChunk( gksChunks[i] );

          //System.out.println( gksChunks[i] );
     }
     /*
     for( int i = 0; i < assocRuleGroups.length; i++ ){
          int chunkIdx = assocRuleGroups[i].getGroupID();
          chunks[ chunkIdx ].setRules( assocRuleGroups[i] );

          System.out.println("the rule group " + assocRuleGroups[i] );
     }
     */
  }

  // get the activated chunks by the information: features.
  private int activateChunks( Feature[] features, int whichLevel )
  {
     int activateCount = 0;
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          // if the chunk is activated.
          boolean bVal = chunks[ idx ].activateChunk( features, whichLevel );
          // the chunk is evocated, set strength of the activated chunk.
          if( bVal ) {
              if( whichLevel == GKS_IDX ) {   // GKS.
                  //gksConcStrengths[ idx ] = chunks[ idx ].getStrength();
                  gksActivations[ idx ] = chunks[ idx ].getStrength();
                  //if( emOn )
                  //    updateEM( Global.NACS, idx );
              }
              else {   // AMN.
                  amnConcStrengths[ idx ] = chunks[ idx ].getStrength();
                  //if( emOn && amnConcStrengths[idx] > threshold_ce )
                  //    updateEM( Global.NACS, idx );
              }
              activateCount++;
              // the chunk is evocated, reset age.
              chunks[ idx ].resetAge();
              // the chunk is evocated, record this time being used.
              // update RT, need further change.
              if( global.isRTOn() ){
                  chunks[idx].calcRetrievalTime();
                  int rt = chunks[idx].getRetrievalTime();
                  if( responseTime < rt )
                      responseTime = rt;
              }
              if( chunkBlaOption == global.VARIED_OPTION )
                  chunks[ idx ].increaseUsage( currentTime );
          }
     }
     return activateCount;
  }

  // draw conclusion.
  private void drawConclusion()
  {
     Vector deducedChunks = new Vector( curChunkNum );    // deduced chunks.
     Vector newDeducedChunks = new Vector( curChunkNum ); // new deduced chunks.
     // load all of the activated chunks.
     for( int i = 0; i < curChunkNum; i++ ){
          int idx = ( (Integer)existChunkList.elementAt( i ) ).intValue();
          if( Math.abs( gksActivations[idx] - 1.0 ) < ALMOST_ZERO ||
              ( global.isAssocPartMatchOn() && gksActivations[idx]>0.0 ) )
             deducedChunks.addElement( new Integer( idx ) );
          chunks[idx].setInEmOn( false );
     }
     // draw conclusions.
     for( int i = 0; i < iterationNum; i++ ){
          newDeducedChunks.removeAllElements();
          int RT = 0;
          for( int j = 0; j < curChunkNum; j++ ){
               int idx = ((Integer)existChunkList.elementAt( j )).intValue();
               // calculate support from rule-based and similarity-based.
               if( chunks[idx].getRuleNum() == 0 )
                   continue;
               chunks[ idx ].calculateStrength( deducedChunks );
               double strength = chunks[ idx ].getStrength();
               // if a new deduced chunk or not.
               if( strength > gksConcStrengths[ idx ] ){
                   // new deduced chunk, the chunk is evocated, reset age.
                   if( gksConcStrengths[ idx ] == 0.0 ){
                       newDeducedChunks.addElement( new Integer( idx ) );
                       chunks[ idx ].resetAge();
                   }
                   chunks[ idx ].setStrength( strength );
                   gksConcStrengths[ idx ] = strength;
                   // update EM if EM is on.
                   if( emOn && strength > threshold_ae && !chunks[idx].isInEmOn() ){
                       chunks[ idx ].setInEmOn( true );
                       short[][] activeConditions = chunks[idx].getActiveRuleConditions();
                       updateEM( activeConditions, idx );
                   }
                   // update RT, need further change.
                   if( global.isRTOn() ){
                       chunks[idx].calcRetrievalTime();
                       int rt = chunks[idx].getRetrievalTime()+chunks[idx].getAssocRuleAppTime();
                       if( RT < rt )
                           RT = rt;
                   }
                   if( chunkBlaOption == global.VARIED_OPTION )
                       chunks[ idx ].increaseUsage( currentTime );
               }
          }
          // no new chunk is concluded.
          if( newDeducedChunks.size() == 0 )
              break;
          deducedChunks.removeAllElements();
          for( int j = 0; j < newDeducedChunks.size(); j++ )
               deducedChunks.addElement( newDeducedChunks.elementAt(j) );
          if( global.isRTOn() )
              responseTime += RT;
     }
     // update conclusion strengths.
     for( int j = 0; j < curChunkNum; j++ ){
          int idx = ((Integer)existChunkList.elementAt( j )).intValue();
          if( chunks[ idx ].getRuleNum() == 0 )
              gksConcStrengths[idx] = 0.0;
     }
  }

  private void  extractExplicitKnowledge( int amnIdx, Feature[] amnResult )
  {
     // chunk from AMN is not allowed.
     if( !global.isChunkFromAmnOn() || !global.isAssocExtractOn() )
         return;
     // calculate the activation of AMN result.
     double amnActivation = calcActivation( amnResult );
     cueActivation = calcActivation( cue );
     int cueChunkIdx = -1;
     int amnResultChunkIdx = -1;
     int amnCombChunkIdx = -1;
     if( amnActivation > threshold_ce )
     {   //  chunk extraction considered.
         if( Math.random() < p_ce ){
             // extract cue.
             cueChunkIdx = getChunkIdx( cue );
             if( cueChunkIdx == -1 )
                 cueChunkIdx = insertChunk( global.EMPTY, amnIdx, cueType, cueLabel, cue );
             // extract AMN result.
             amnResultChunkIdx = getChunkIdx( amnResult );
             if( amnResultChunkIdx == -1 ){
                 int resultType = global.CHK_RETRIEVAL;
                 String resultLabel = global.CHUNK_TYPE_LABELS[resultType];
                 amnResultChunkIdx = insertChunk( global.EMPTY, amnIdx, resultType, resultLabel, amnResult );
             }
             // extract cue and AMN result together.
             Feature[] amnCombResult = new Feature[ cue.length + amnResult.length ];
             for( int i = 0; i < cue.length; i++ )
                  amnCombResult[ i ] = (Feature)(cue[ i ].clone());
             for( int i = 0; i < amnResult.length; i++ )
                  amnCombResult[ cue.length + i ] = (Feature)(amnResult[ i ].clone());
             amnCombChunkIdx = getChunkIdx( amnCombResult );
             if( amnCombChunkIdx == -1 ) {
                 int combType = global.CHK_COMB;
                 String combLabel = global.CHUNK_TYPE_LABELS[combType];
                 amnCombChunkIdx = insertChunk( global.EMPTY, amnIdx, combType, combLabel, amnCombResult );
             }
         }
     }
     if( !global.isAssocExtractOn() )
         return;
     if( amnActivation > threshold_ae )
     {  //  associative rule extraction considered.
         if( Math.random() < p_ae ){
             // extract cue.
             if( cueChunkIdx == -1 )
                 cueChunkIdx = insertChunk( global.EMPTY, amnIdx, cueType, cueLabel, cue );
             // extract AMN result.
             if( amnResultChunkIdx == -1 ){
                 int resultType = global.CHK_RETRIEVAL;
                 String resultLabel = global.CHUNK_TYPE_LABELS[resultType];
                 amnResultChunkIdx = insertChunk( global.EMPTY, amnIdx, resultType, resultLabel, amnResult );
             }
             // associative rule extraction
             curChunkNum = existChunkList.size();
             for( int i = 0; i < curChunkNum; i++ ) {
                  int idx = ((Integer)existChunkList.elementAt( i )).intValue();
                  if( amnConcStrengths[ idx ] > 0.0 ){
                      short[] condition = new short[1];
                      condition[ 0 ] = (short)cueChunkIdx;
                      chunks[ idx ].insertRule( condition, (short)idx );
                  }
             }
         }
     }
  }

  private double calcActivation( Feature[] features )
  {
     double actVal = 0.0;
     int activationMode = global.getActivationMeasure();

     if( activationMode == 0 )  // minimum measurement.
     {
         actVal = 1.0;
         for( int i = 0; i < features.length; i++ )
         {
              double[] activations = features[ i ].getActivations();
              for( int j = 0; j < activations.length; j++ )
                   if( activations[ j ] > 0.0 && activations[ j ] < actVal )
                       actVal = activations[ j ];
         }
     }
     else  // average measurement.
     {
         for( int i = 0; i < features.length; i++ )
         {
              int count = 0;
              double val = 0.0;
              double[] activations = features[ i ].getActivations();
              for( int j = 0; j < activations.length; j++ )
              {
                   if( activations[ j ] > 0.0 )
                   {
                       val += activations[ j ];
                       count++;
                   }
              }
              if( count == 0 )
                  count++;
              val /= count;

              actVal += val;
         }
         actVal /= features.length;
     }
     return actVal;
  }

  // add associative rules into EM.
  private void updateEM( GKSChunk condition, GKSChunk conclusion )
  {
     if( !epiChunkOn )
         return;
     GKSChunk[] ruleChunk = new GKSChunk[2];
     try{
         ruleChunk[0] = (GKSChunk)(condition.clone());
         ruleChunk[1] = (GKSChunk)(conclusion.clone());
         nacsEMs[condition.getAmnIdx()].insertItem(
             new Chunk(Global.CHK_ASSOC,Global.CHUNK_TYPE_LABELS[global.CHK_ASSOC],ruleChunk,global) );
     }
     catch( Exception e ){
     }
  }

  // add associative rules into EM.
  private void updateEM( short[][] condIndices, int concIndex )
  {
     if( !epiChunkOn )
         return;
     try{

       //System.out.println( "the condition indices len = " + condIndices.length );

       for( int i = 0; i < condIndices.length; i++ ){
            int chunkNum = condIndices[i].length + 1;
            GKSChunk[] ruleChunk = new GKSChunk[chunkNum];

            //System.out.println( "the condition chunk idx = " + condIndices[i][0] );

            for( int j = 0; j < chunkNum-1; j++ )
                 ruleChunk[j] = (GKSChunk)(chunks[condIndices[i][j]].clone());
            ruleChunk[chunkNum-1] = (GKSChunk)(chunks[concIndex].clone());
            int amnIdx = chunks[concIndex].getAmnIdx();
            Chunk chunk = new Chunk( Global.CHK_ASSOC,Global.CHUNK_TYPE_LABELS[Global.CHK_ASSOC],
                                     ruleChunk, global );

            //System.out.println( "the saved chunk = \n" + chunk );

            nacsEMs[amnIdx].insertItem( chunk );
       }
     }
     catch( Exception e ){

           System.out.println( e );
     }
  }

  // check EM if activations of some chunks are lower than the threshold.
  public void checkEM()
  {
     for( int i = 0; i < acsEMs.length; i++ )
          acsEMs[i].checkValid();
     for( int i = 0; i < nacsEMs.length; i++ )
          nacsEMs[i].checkValid();
  }

  // fill in the most activated conclusions from top level.
  private Chunk[] fillInFinalResults( int level, double[] strengths )
  {
     int resultNum = 0;
     for( int i = 0; i < strengths.length; i++ )
          if( strengths[i] > 0.0 )
              resultNum++;
     SortableElement[] sortableElements = new SortableElement[ resultNum ];
     int num = 0;
     for( int i = 0; i < strengths.length; i++ ){
          if( strengths[i] > 0.0 )
              sortableElements[num++] = new SortableElement( strengths[i], i );
     }
     Arrays.sort( sortableElements );
     int c = (level == GKS_IDX)?Math.min( c21, resultNum ):Math.min( c25, resultNum );
     int len = (global.isMcsOn())? c : resultNum;
     Chunk[] results = new Chunk[ len ];
     try{
         for( int i = 0; i < len; i++ ){
              int idx =  sortableElements[resultNum-i-1].getChunkIdx();
              results[i] = (GKSChunk)(chunks[idx].clone());
              results[i].setStrength( strengths[idx] );
         }
     }
     catch( Exception e ){
            System.out.println( e + ", clone chunk error in GKS.java." );
     }
     return results;
  }

  // fill in the most activated conclusions from top level.
  private Chunk[] fillInFinalResults( int concNum, double[] strengths, short[] indices )
  {
     int resultNum = concNum;
     SortableElement[] sortableElements = new SortableElement[ resultNum ];
     for( int i = 0; i < resultNum; i++ )
          sortableElements[i] = new SortableElement( strengths[i], indices[i] );
     Arrays.sort( sortableElements );
     int c = Math.min( Math.max(c21, c25), resultNum );
     int len = (global.isMcsOn())? c : resultNum;

     //System.out.println( "result len = " + len );

     Chunk[] results = new Chunk[ len ];
     try{
         for( int i = 0; i < len; i++ ){
              int idx =  sortableElements[resultNum-i-1].getChunkIdx();
              results[i] = (GKSChunk)(chunks[idx].clone());
              double strength = sortableElements[resultNum-i-1].getKeyValue();
              results[i].setStrength( strength );
         }
     }
     catch( Exception e ){
            System.out.println( e + ", clone chunk error in GKS.java." );
     }
     return results;
  }
}