package clarion.nacs;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;

public class NACS {

  final int BL = Global.BL;
  final int TL = Global.TL;
  final int COMB = 2;
  final double ALMOST_ZERO = 0.0000001;
  // global settings.
  protected Global global;
  protected GKS    gks;
  protected AMNet  amns[];
 //retrieval mode: no retrieval, retrieval_one, retrieval_all.
  protected int   retrievalMode = Global.RETRIEVAL_ALL ;
  protected int   activeAmnIdx;
  protected int   acsNetNum;
  protected int[] acsStateDimNums;
  protected int[] acsActionDimNums;
  protected int[] acsRewardDimNums;
  protected int[] acsStepFeatNums;
  protected boolean esChunkOn;
  protected boolean aemOn;
  protected boolean amnMultiFlag;
  protected double  aemTrainedByEmProb;
  protected AbsEM  absEMs[];
  protected Feature[] curState;
  protected Feature[] curAction;
  protected Feature[] nextState;
  protected Feature[] reward;
  protected GKSChunk results[];

  public NACS(){}

  public NACS( Global g )
  {
      global = g;
      global.setNACS( this );
      gks = new GKS( global );
      amns = new AMNet[ global.getAmnNum() ];
      for( int i = 0; i < amns.length; i++ )
           amns[ i ] = new AMNet( i, global, gks );
      amnMultiFlag = global.isAmnMultiFlagOn();
      // initialize variables relative to acs.
      acsNetNum = global.getAcsNetNum();
      acsStateDimNums = new int[ acsNetNum ];
      acsActionDimNums = new int[ acsNetNum ];
      acsRewardDimNums = new int[ acsNetNum ];
      acsStepFeatNums = new int[ acsNetNum ];
      for( int i = 0; i < acsNetNum; i++ ){
           acsStateDimNums[i] = global.getIdnInputDimVals(i).length;
           acsActionDimNums[i] = global.getIdnOutputDimVals(i).length;
           acsRewardDimNums[i] = 1;
           if( global.getIdnLearningOption(i) == global.QL )
               acsStepFeatNums[i]= 2*acsStateDimNums[i]+acsActionDimNums[i]+acsRewardDimNums[i];
           else
               acsStepFeatNums[i]= acsStateDimNums[i]+acsActionDimNums[i]+acsRewardDimNums[i];
      }
      esChunkOn = global.isESChunkOn();
      aemOn = global.isAEMon();
      if( aemOn ){
          aemTrainedByEmProb = global.getAemTrainedByEmProb();
          absEMs = new AbsEM[ acsNetNum ];
          for( int i = 0; i < acsNetNum; i++ )
               absEMs[ i ] = new AbsEM( i, global );
      }
  }

  public GKSChunk[] getResults()
  {
      return results;
  }

  public Chunk[][] getAllConclusions()
  {
      Chunk[][] chunks = new Chunk[3][];
      chunks[TL] = gks.getTopLevelMonitorConclusions();
      chunks[BL] = gks.getBottomLevelMonitorConclusions();
      chunks[COMB] = gks.getAllMonitorConclusions();
      /*
      System.out.println( "all conclusions from GKS = " );
      for( int i = 0; i < chunks[TL].length; i++ )
           System.out.println( chunks[TL][i] );

      System.out.println( "all conclusions from AMN = " );
      for( int i = 0; i < chunks[BL].length; i++ )
           System.out.println( chunks[BL][i] );
       */

      return chunks;
  }

  public AbsEMSample[] getAemSamples( int netIdx )
  {
      return absEMs[ netIdx ].getSamples();
  }

  public EMAcsSample[] getEmAcsSamples( int netIdx, int trainType )
  {
      return gks.getEmAcsSamples( netIdx, trainType );
  }

  public EMNacsSample[] getEmNacsSamples( int netIdx )
  {
      return gks.getEmNacsSamples( netIdx );
  }

  public int getRT()
  {
     //System.out.println( "GKS RT = " + gks.getRT() );
     //System.out.println( "AMN RT = " + amns[activeAmnIdx].getRT() );
     return gks.getRT() + amns[activeAmnIdx].getRT();
  }

  public GKS getGks()
  {
     return gks;
  }

  public void setActiveAmn( int activeAmnIdx )
  {
     this.activeAmnIdx = activeAmnIdx;
  }

  public void setCurrentTime( long curTime )
  {
     gks.setCurrentTime( curTime );
  }

/********************* SET METHODS FOR CUE PARAMETERS ***********************/
  public void setRetrievalMode( int mode )
  {
     retrievalMode = mode;
  }

  public void setGksCue( Feature[] cue, int concIdx )
  {
      gks.setRetrievalCue( cue, concIdx );
  }

  public void setGksCue( Feature[] cue )
  {
      gks.setRetrievalCue( cue );
  }

  public void setAmnCue( int idx, Feature[] input )
  {
      amns[ idx ].setInput( input );
  }

  // need be refined further.
  public void observeState( Feature[] state )
  {
     //if( global.isEMon() )
     //    gks.checkEM();
     if( esChunkOn ) {
         // store current state as a chunk.
         curState = new Feature[ state.length ];
         for( int i = 0; i < curState.length; i++ )
              curState[i] = (Feature)(state[ i ].clone());
         int chunkIdx = gks.insertChunk( 0, 0, global.CHK_STATE,
                        global.CHUNK_TYPE_LABELS[global.CHK_STATE], curState );
         if( global.isEMon() )
             gks.updateEM( Global.ACS, chunkIdx );
         //System.out.println( "current state = " + gks.getChunk(chunkIdx) );
     }
  }

  public void run( int idx )
  {
      // retrieval from both GKS and AMN.
      gks.reasoning();
      amns[ idx ].retrieval( amnMultiFlag );
      // retrieval result of AMN is sent up.
      Feature[] blRetrievals = amns[ idx ].getResults();
      gks.bottomUp( idx, blRetrievals );
      // get the final results.
      gks.formResults( retrievalMode );
      results = gks.getCombinedResults();

      /*
      System.out.println( "results = " );
      for( int i = 0; i < results.length; i++ )
           System.out.println( results[i] );
      */
  }

  // need be refined further.
  public void update( int amnIdx, short[][] chosenActions, Feature[] nState, double[] rewards )
  {
      gks.update();
      amns[ amnIdx ].update();
      if( esChunkOn ) {
          // store the current action, current step as a chunk.
          nextState = nState;
          for( int i = 0; i < acsNetNum; i++ ){
               if( chosenActions[i][0] == Global.EMPTY )
                   continue;
               curAction = new Feature[chosenActions[i].length];
               for( int j = 0; j < curAction.length; j++ ){
                    int type = global.getIdnOutputFeatureInfo(i)[j][0];
                    int order = global.getIdnOutputFeatureInfo(i)[j][1];
                    int len = global.getFeatureLength( type, order );
                    String label = global.getFeatureNames( type, order );
                    short[] vals = new short[1];
                    vals[0] = chosenActions[i][j];
                    curAction[j] = new Feature( label, (short)type, (short)order, (short)len, vals, global );
               }
               reward = new Feature[1];
               double[] actVals = new double[1];
               actVals[0] = rewards[i];
               reward[0] = new Feature( "reward", (short)0, (short)0, actVals, global );
               Feature[] curStep = fetchAcsStep( i );
               int chunkIdx = gks.insertChunk( i, global.EMPTY, global.CHK_ACTION,
                                               global.CHUNK_TYPE_LABELS[ global.CHK_ACTION ], curAction );
               if( global.isEMon() )
                   gks.updateEM( Global.ACS, chunkIdx );
               chunkIdx = gks.insertChunk( i, global.EMPTY, global.CHK_STEP,
                                           global.CHUNK_TYPE_LABELS[global.CHK_STEP], curStep );
               //System.out.println( "current step = " + gks.getChunk(chunkIdx) );
               if( global.isEMon() )
                   gks.updateEM( Global.ACS, chunkIdx );
          }
      }
      // offline training AEM using EM.
      if( aemOn && Math.random() < aemTrainedByEmProb )
          for( int i = 0; i < acsNetNum; i++ )
               offlineTrainAem( i );
  }

  public String testAem( int acsNetIdx )
  {
      return absEMs[ acsNetIdx ].testing();
  }

/************************ PRIVATE METHODS *************************************/
  private Feature[] fetchAcsStep( int netIdx )
  {
     int dimNum = acsStepFeatNums[ netIdx ];
     Feature[] features = new Feature[dimNum];
     int dimOffset = 0;
     for( int i = 0; i < curState.length; i++ )
         features[ dimOffset+i] = (Feature)(curState[ i ].clone());
     dimOffset += curState.length;
     for( int i = 0; i < curAction.length; i++ )
         features[ dimOffset+i] = (Feature)(curAction[ i ].clone());
     dimOffset += curAction.length;
     if( global.getIdnLearningOption(netIdx) == global.QL ){
         for( int i = 0; i < nextState.length; i++ )
              features[ dimOffset+i] = (Feature)(nextState[ i ].clone());
         dimOffset += nextState.length;
     }

     //System.out.println( "dimNum = " + dimNum + ", dimOffset = " + dimOffset );

     for( int i = 0; i < reward.length; i++ )
          features[ dimOffset+i ] = (Feature)(reward[ i ].clone());
     return features;
  }

  // offline training AEM using EM.
  private void offlineTrainAem( int acsNetIdx )
  {
     EMAcsSample[] acsSamples = gks.getEmAcsSamples( acsNetIdx, global.AEM );
     int sampleNum = acsSamples.length;
     int iterNum = 1;
     if( sampleNum < global.getAemTrainedByEmItemNum() ){
         //System.out.println("insuffient EM item # training AEM!" );
         return;
     }
     Vector sands = new Vector( sampleNum * iterNum );
     for( int j = 0; j < iterNum; j++ )
       for( int k = 0; k < sampleNum; k++ )
            sands.addElement( new Integer(k) );
     while( !sands.isEmpty() ){
       int idx = ((Integer)(sands.remove((int)(sands.size()*Math.random())))).intValue();
       //System.out.println("current sample = " + acsSamples[idx] );
       absEMs[ acsNetIdx ].offlineTrain( acsSamples[idx] );
     }

     //debugging and testing.
     //for( int i = 0; i < sampleNum; i++ )
     //     System.out.println( acsSamples[i] );
     //absEMs[ acsNetIdx ].testing();
  }
}