package clarion.options;
import clarion.system.*;

public class Parameter implements Cloneable {
  int curNum;
  int majorType;   // 0: primitive, 1: array.
  int minorType;   // 0: boolean, 1: int, 2: double, 3: string.
  String label;
  String value;
  String values[];

  public Parameter( int majorType, int minorType, String label, String value )
  {
     this.majorType = majorType;
     this.minorType = minorType;
     this.label = label;
     this.value = value;
  }

  public Parameter( int majorType, int minorType, String label, String[] vals )
  {
     this.majorType = majorType;
     this.minorType = minorType;
     this.label = label;
     this.values = new String[ vals.length ];
     curNum = vals.length;
     for( int i = 0; i < curNum; i++ )
          values[i] = new String( vals[i] );
  }

  public Parameter( int majorType, int minorType, String label, int num, String[] vals )
  {
     this.majorType = majorType;
     this.minorType = minorType;
     this.label = label;
     this.values = new String[ num ];
     curNum = num;
     for( int i = 0; i < curNum; i++ )
          values[i] = new String( vals[i] );
  }

  public int getMajorType()
  {
     return majorType;
  }

  public int getMinorType()
  {
     return minorType;
  }

  public String getLabel()
  {
     return label;
  }

  public int getCount()
  {
     return curNum;
  }

  public boolean getBoolValue()
  {
     return (Boolean.valueOf( value )).equals( Boolean.TRUE );
  }

  public int getIntValue()
  {
     return Integer.parseInt( value );
  }

  public double getDoubleValue()
  {
     return Double.parseDouble( value );
  }

  public String getStringValue()
  {
     return value;
  }

  public boolean[] getBoolValues()
  {
     boolean[] array = new boolean[ curNum ];
     for( int i = 0; i < curNum; i++ )
          array[i] = Boolean.valueOf( values[i] ).equals( Boolean.TRUE );
     return array;
  }

  public short[] getIntValues()
  {
     short[] array = new short[ curNum ];
     for( int i = 0; i < curNum; i++ )
          array[i] = Short.parseShort( values[i] );
     return array;
  }

  public double[] getDoubleValues()
  {
     double[] array = new double[ curNum ];
     for( int i = 0; i < curNum; i++ )
          array[i] = Double.parseDouble( values[i] );
     return array;
  }

  public String[] getStringValues()
  {
     String[] array = new String[ curNum ];
     for( int i = 0; i < curNum; i++ )
          array[i] = values[i];
     return array;
  }

  public void setLabel( String str )
  {
     label = str;
  }

  public void setCount( int num )
  {
     if( num <= curNum ){
         curNum = num;
         return;
     }
     final String defaultValues[] = { "false", "1", "1.0", "val" };
     String[] temp = new String[ curNum ];
     for( int i = 0; i < curNum; i++ )
          temp[i] = new String( values[i] );
     values = new String[ num ];
     for( int i = 0; i < num; i++ )
          if( i < curNum )
              values[i] = new String( temp[i] );
          else{
              String postfix = (minorType != Global.STRING) ? "" : String.valueOf(i);
              values[i] = new String ( defaultValues[minorType] + postfix );
          }
     curNum = num;
  }


  public void setBoolValue( boolean val )
  {
     value = String.valueOf( val );
  }

  public void setIntValue( int val )
  {
     value = String.valueOf( val );
  }

  public void setDoubleValue( double val )
  {
     value = String.valueOf( val );
  }

  public void setStringValue( String val )
  {
     value = val;
  }

  public void setBoolValue( int idx, boolean val )
  {
     values[idx] = String.valueOf( val );
  }

  public void setIntValue( int idx, int val )
  {
     values[idx] = String.valueOf( val );
  }

  public void setDoubleValue( int idx, double val )
  {
     values[idx] = String.valueOf( val );
  }

  public void setStringValue( int idx, String val )
  {
     values[idx] = new String( val );
  }

  /*
  public void setBoolValues( boolean[] vals )
  {
     curNum = vals.length;
     for( int i = 0; i < vals.length; i++ )
          values[i] = String.valueOf( vals[i] );
  }

  public void setIntValues( short[] vals )
  {
     curNum = vals.length;
     for( int i = 0; i < vals.length; i++ )
          values[i] = String.valueOf( vals[i] );
  }

  public void setDoubleValues( double[] vals )
  {
     curNum = vals.length;
     for( int i = 0; i < vals.length; i++ )
          values[i] = String.valueOf( vals[i] );
  }

  public void setStringValues( String[] vals )
  {
     curNum = vals.length;
     for( int i = 0; i < vals.length; i++ )
          values[i] = vals[i];
  }
  */

  public String toString()
  {
     String str = "Parameter major type : " + majorType + "\n";
     str += "Parameter minor type : " + minorType + "\n";
     str += "Parameter label : " + label + "\n";
     if( majorType == 0 )
         str += "Parameter value : " + value + "\n";
     else{
         str += "Parameter value : ";
         for( int i = 0; i < curNum; i++ )
              if( i < curNum - 1 )
                  str += values[i] + ",";
              else
                  str += values[i];
         str += "\n";
     }
     return str;
  }

  public Object clone() //throws CloneNotSupportedException
  {
      Parameter param;
      if( majorType == 0 )
          param = new Parameter( majorType, minorType, label, value );
      else{
          //System.out.println( "in clone, curNum = " + curNum );
          param = new Parameter( majorType, minorType, label, curNum, values );
      }
      return param;
  }
}