package clarion.system;
import clarion.common.*;
import clarion.tools.*;
import clarion.nacs.*;

/**
 *  This class is for implementing what a task subject does in a task. In order
 *  for a task subject to communicate with CLARION properly, all of the methods
 *  in this class should be overrided by user codes.
 */

public abstract class TaskAgent{

   // global settings each running agent has.
   protected Global  global;
   // clarion model.
   protected Clarion clarion;
   protected TaskClarion taskClarion;
   protected Feature[] curSensoryInput;
   protected Feature[] curNacsInput;
   protected Feature[] nacsResults;
   protected Feature[] desiredOutput[];
   protected double reinforcements[];
   protected int activeAmnIdx;
   protected boolean learningTestOn;
   protected WriteFile dataFile;

   public TaskAgent(){}

   public TaskAgent( Global g )
   {
      global = g;
      clarion = new Clarion( global, this );
      taskClarion = global.getTaskClarion();
   }

   public void reinit()
   {
      clarion = new Clarion( global, this );
   }

   public boolean isInLearningTest()
   {
      return learningTestOn;
   }

   public int getActiveAmnIdx()
   {
      return activeAmnIdx;
   }

   public String[] getRerRules( int netIdx )
   {
      return clarion.getRerRules( netIdx );
   }

   public String[] getIrlRules( int netIdx )
   {
      return clarion.getIrlRules( netIdx );
   }

   public String[] getFixRules( int netIdx )
   {
      return clarion.getFixRules( netIdx );
   }

   public Feature[] getCurSensoryInput()
   {
      return curSensoryInput;
   }

   public Feature[] getCurNacsInput()
   {
      return curNacsInput;
   }

   public Feature[][] getCurDesiredOutput()
   {
      return desiredOutput;
   }

   public void setLearningTest( boolean flag )
   {
      learningTestOn = flag;
   }

   public void saveOutput( String results )
   {
      dataFile.saveStringData( results );
   }

   // return if the chosen action is an action for controlling NACS.
   abstract public boolean isControlNacsAction( int netIdx, short[] chosenAction );
   // get the otuput of current step.
   abstract public void getOutput( Object arr );
   // determine which AMN will be used given current action.
   abstract public int determineAmnNet( int netIdx, short[] action );
   // send the retrieval results to ACS components.
   abstract public void setRetrievalResults( int netIdx, GKSChunk[] results );
   // perform the chosen action.
   abstract public void performAction( int netIdx, short[] action );
   // simulate a subject performance by step.
   abstract public void run();
}
