package clarion.system;
import java.util.Arrays;
import clarion.common.*;
import clarion.tools.*;
import clarion.nacs.*;
import clarion.acs.net.*;
import clarion.acs.net.auxiliary.*;
import clarion.acs.gs.*;
import clarion.acs.wm.*;

/**
 *  This class is for task specific part of CLARION. In order for a specific task
 *  to communicate with CLARION properly, all of the methods in this class should
 *  be overrided by user codes.
 */
public class TaskClarion{
  final public static int SUB_SYS_NUM = 3, ACS = 0, NACS = 1, RT = 2; // subsystem.
  final public static int ACS_SAMPLE_NUM = 19, NACS_SAMPLE_NUM = 4, RT_SAMPLE_NUM = 9;
  final public static int GOAL = 0, GOAL_DIM = 1, BU = 2, TG = 3, ELIG = 4, REIN = 5;
  final public static int RER_WGT = 6, RER_UTL = 7, RER_POS = 8, RER_IG = 9;
  final public static int IRL_WGT = 10, IRL_UTL = 11, IRL_POS = 12, IRL_IG = 13, IRL_PARAM = 14;
  final public static int FR_WGT = 15, FR_UTL = 16, FR_CON = 17, FR_ACT = 18;
  final public static int ASSOC_WGT = 0, ASSOC_UTL = 1, SIM = 2;
  final public static int OVER_RT = 0, BL_RT = 1, TL_RT = 2, PT_TL = 3, DT_TL = 4, AT_TL = 5;
  final public static int ASSOC_APP = 6, CHUNK_RETR = 7, AMN_PASS = 8;
  final static String className = "TaskClarionPC";
  final static String[][] methodNames = {
        {
           "getGoals",  "getGoalDims", "runBUCorrection",  "runTGCorrection",
           "getEligibility",  "getReinforcement", "getRerWeights", "getRerUtility",
           "getRerPositivity",  "getRerIG", "getIrlWeights",  "getIrlUtility",
           "getIrlPositivity",  "getIrlIG", "getIrlRule", "getFrWeights",
           "getFrUtility",  "getFrCondition", "getFrAction"
        },
        {
           "getExternalNacsChunks", "getExternalNacsRules",
           "getAssocWeights",  "getAssocUtility",
           "getSimilarity",   "getAamDataPattern",
        },
        {
           "getOverallRT", "getBLOverallRT", "getTLOverallRT", "getPT_TL", "getDT_TL",
           "getAT_TL",  "getAssocAppTime", "getChunkRetrtime", "getAmnPassTime"
        }
  };
  final static String[][] methodForms = {
        {
           "public short[][] getGoals()",
           "public short[] getGoalDims()",
           "public void runBUCorrection()",
           "public void runTGCorrection()",
           "public short[][][] getEligibility( int netIdx )",
           "public double getReinforcement( int netIdx, Feature[] input,         \n" +
           "       GKSChunk[] wmChunks, int goal, short[] action, Feature[] newInput,\n" +
           "       GKSChunk[] newWMChunks, int newGoal ) ",
           "public double[] getRerWeights( int netIdx ) ",
           "public double getRerUtility( int netIdx, double PM, double NM, \n" +
           "       double cost ) ",
           "public boolean getRerPositivity( int learnOpt, double reward,  \n" +
           "       double qDiscount, double newMaxQVal, double qVal,       \n" +
           "       double threshold ) ",
           "public double getRerIG( int netIdx, double PMa, double NMa,    \n" +
           "       double PMb, double NMb )",
           "public double[] getIrlWeights( int netIdx )",
           "public double getIrlUtility( int netIdx, double PM, double NM, \n" +
           "       double cost )",
           "public boolean getIrlPositivity( int netIdx, double reward,    \n" +
           "       double qDiscount, double newMaxQVal, double qVal,       \n" +
           "       double threshold ) ",
           "public double getIrlIG( int netIdx, double PMa, double NMa )",
           "public IrlRuleForm getIrlRule(int netIdx, int setIdx, int ruleIdx)",
           "public double[] getFrWeights( int netIdx )",
           "public double getFrUtility( int netIdx, double PM, double NM,  \n" +
           "       double cost )",
           "public short[][] getFrCondition(int netIdx, int ruleIdx)",
           "public short[] getFrAction(int netIdx, int ruleIdx)"
        },
        {
           "public GKSChunk[] getExternalNacsChunks()",
           "public AssocRule[] getExternalNacsRules()",
           "public double[] getAssocWeights()",
           "public double getAssocUtility( double PM, double NM, double cost ) ",
           "public double getSimilarity(GKSChunk chunk1, GKSChunk chunk2)",
           "public short[] getAamDataPattern(int netIdx, int ruleIdx)"
        },
        {
           "public int getOverallRT( int RT_BL, int RT_TL )",
           "public int getBLOverallRT( int PT_BL, int DT_BL, int AT_BL )",
           "public int getTLOverallRT( int PT_TL, int DT_TL, int AT_TL )",
           "public int getPT_TL( int PT_BL )",
           "public int getDT_TL( int DT_BL, int actionTime, double ruleBla, \n" +
           "           double chunkBla )",
           "public int getAT_TL( int action, int assocAppTime, int chunkRetrTime )",
           "public int getAssocAppTime( double assocRuleBla )",
           "public int getChunkRetrtime( double chunkBla )",
           "public int getAmnPassTime()"
        }
  };
  final static String[][] descriptions = {
        {
           "to get all of goals used in goal structure in the form of its values   \n" +
           "in each parameter dimension.                                           \n" +
           "1st dimension - the goal NO.                                           \n" +
           "2nd dimension - eligible value in each parameter dimension.  ",
           "to get goal parameter dimension-value information. Currently, the goal \n" +
           "structure has several goal items representing goal and its parameters. \n" +
           "Those dimensions help describing what a specific goal wants to achieve.\n" +
           "The goal parameter dimension information CLARION wants to know is how  \n" +
           "many values in each dimension. How such dimensions works on goal action\n" +
           "decision is task-specific and so needs user's code to implement it     \n" +
           "whether in another option, user choose goal action by top level( Fixed \n" +
           "Rules) or both levels.             ",
           "the process of Bottom-Up correction",
           "the process of Top-Guidance correction",
           "to get eligible conditions of an external network. Eligible condition  \n" +
           "determines which network is eligible given current input. Each network \n" +
           "has an eligible condition for current input to match.                  \n" +
           "1. Eligible condition of a network has the format:                     \n" +
           "   if ( C1 or C2...or Cn ) then use net K. ( 'or' conditions )         \n" +
           "2. Ci has the format (ci1 and ci2... and cim). ( 'and' subconditions ) \n" +
           "   each condition cij is an array representing an input dimension.     \n" +
           "   In the array, only the elements those values are 1 is the eligible  \n" +
           "   elements in that dimension.                                         \n" +
           "3. The eligible conditions of an external network is represented by 3- \n" +
           "   dimension array.                                                    \n" +
           "   1st dimension - the 'or' eligible conditions of an external network \n" +
           "   that means, if any one of the condition is satisfied, the network   \n" +
           "   is eligible.                                                        \n" +
           "   2nd dimension - the 'and' subconditions composing a 'or' condition. \n" +
           "   In this dimension, the first element, 1-dimension array, stores the \n" +
           "   involved dimension indices on the input.                            \n" +
           "   3rd dimension - the eligible elements in one dimension composing    \n" +
           "   an 'and' subcondition.                                              \n" +
           "4. In user code, the eligible conditions should be given in the above  \n" +
           "   format. ",
           "to get reinforcement given current input, current working memory status\n" +
           "if WM is set, current goal if GS is set; current action; the new input,\n" +
           "the new working memory status if WM is set, new goal if GS is set      \n" +
           "resulting from the firing of current action. ",
           "to get dimension weights of RER condition which is used for          \n" +
           "calculating the rule support for the action of the RER rule",
           "to get RER Utility which measures the effectiveness, or usefulness of\n" +
           "a rule",
           "to get RER positivity which means whether or not the aciton chosen   \n" +
           "according to a rule is reasonably good.",
           "to get Information Gain of RER rule A over RER rule B.",
           "to get dimension weights of IRL condition which is used for          \n" +
           "calculating the rule support for the action of the RER rule",
           "to get IRL utility which measures the effectiveness, or usefulness of\n" +
           "a rule",
           "to get IRL positivity which means whether or not the aciton chosen   \n" +
           "according to a rule is reasonably good.",
           "to get Information Gain of IRL rule A over IRL rule B.",
           "to get an IRL rule. A IRL rule is composed of a set of parameters    \n" +
           "Most parameters are relevant to an input dimension and some are not. \n" +
           "Only knowing the parameters is not enough to make an action decision,\n" +
           "but also we need know how the parameters operates on action decision.\n" +
           "Since the action decision is task-specific, user still has to implement\n" +
           "the codes for how IRL rules work on action decision. That is a method \n" +
           "in class IrlRuleForm.java for user to implement. ",
           "to get dimension weights of Fixed Rule condition which is used for   \n" +
           "calculating the rule support for the action of the RER rule    ",
           "to get Fixed Rule Utility which measures the effectiveness, or       \n" +
           "usefulness of a rule ",
           "to get a Fixed Rule condition. A Fixed Rule condition is represented \n" +
           "by a 2-dimension array.                                              \n" +
           "1st dimension - represents an input dimension.                       \n" +
           "2nd dimension - represents the eligible elements in that dimension.  \n" +
           "The value of each element is 0 or 1, 0: not eligible, 1: eligible.   \n" +
           "Actually, there are two kinds of Fixed Rules, one has fixed aciton   \n" +
           "and the other hasn't. In the first case, differently from IRL rules, \n" +
           "knowing Fixed Rule condition is enough to make an aciton decision by \n" +
           "simply doing a match of current input with the condition. In the     \n" +
           "second case, we still need user's code to implement how Fixed Rule   \n" +
           "works. Also this is a method in the class TaskAgent.java for user to \n" +
           "override.            ",
           "to get the action of the Fixed Rule. Originally, each network has an \n" +
           "action list, to get the action means to get the action index on that \n" +
           "action list.  ",
        },
        {
           "to get the externally encoded NACS chunks. \n",
           "to get the externally encoded NACS rules.  \n",
           "to get dimension weights of associative rule condition which is used\n" +
           "for calculating the rule support for the action of the RER rule",
           "to get associative rule Utility which measures the effectiveness, or\n" +
           "usefulness of a rule",
           "to get the similarity of two NACS chunks",
           "to get the type of data pattern for training AAM."
        },
        {
           "to get overall response time in each step, user should give a new   \n" +
           "formula to calculate it.",
           "to get overall Bottom-Level response time in each step, user should \n" +
           "give a new formula to calculate it.",
           "to get overall Top-Level response time in each step, user should    \n" +
           "give a new formula to calculate it.",
           "to get Top-Level Perceptual Time in each step, user should set a    \n" +
           "number to this variable or a new formula to calculate it.",
           "to get Top-Level Decision Time in each step, user should set a      \n" +
           "number to this variable or a new formula to calculate it.",
           "to get Top-Level Actuation Time in each step, user should set a     \n" +
           "number to this variable or a new formula to calculate it.",
           "to get Associatve rule Application Time of one iteration, user      \n" +
           "should set a number to this variable or a new formula to calculate  \n" +
           "it.",
           "to get Chunk Retrieval Time of one iteration, user should set a     \n" +
           "number to this variable or a new formula to calculate it.",
           "to get AMN one-pass time from mapping input to output, user should  \n" +
           "set a number to this variable or a new formula to calculate it.",
        }
  };
  final static String[][] params = {
        {
           "null",
           "null",
           "null",
           "null",
           "int netIdx - the external network NO.              ",
           "int netIdx - the network NO.                                     \n" +
           "Feature[] input - current input,                                 \n" +
           " ( see the class Feature.java in subdirectory: clarion.common )  \n" +
           "WMChunk[] wmChunks - current working memory status.              \n" +
           " ( see the class WMChunk.java in subdirectory: clarion.acs.wm )  \n" +
           "int goal - current goal.                                         \n" +
           "int action - current action.                                     \n" +
           "Feature[] newInput - new input resulting from firing the action. \n" +
           "WMChunk[] newWMChunks - new working memory status.               \n" +
           "int newGoal - new goal.                             ",
           "int netIdx - the network NO.                        ",
           "int netIdx - the network NO.                                         \n" +
           "double PM - Positive Match of the rule, equals the number of times   \n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is positive.                                          \n" +
           "double NM - Negative Match of the rule, equals the number of times   \n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is negative.                                          \n" +
           "double cost - the execution cost of the rule.         ",
           "int netIdx - the network NO.                                         \n" +
           "double reward - currently received reward.                           \n" +
           "double qDiscount - the discount factor for Q-Learning that favors    \n" +
           "reinfocement received sooner relative to that received later.        \n" +
           "double newMaxQVal - maximum Q-value of the new state resuling from   \n" +
           "current action in current state, Q(next state, next possible action) \n" +
           "double qVal - current q-value, Q(current state, current action).     \n" +
           "double threshold - the threshold for rule positivity. ",
           "int netIdx - the network NO.                                         \n" +
           "double PMa - Positive Match of the rule a, equals the number of times\n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is positive.                                          \n" +
           "double NMa - Negative Match of the rule a, equals the number of times\n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is negative.                                          \n" +
           "double PMb - Positive Match of the rule b.                           \n" +
           "double NMb - Negative Match of the rule b.            ",
           "int netIdx - the network NO.                          ",
           "int netIdx - the network NO.                                         \n" +
           "double PM - Positive Match of the rule, equals the number of times   \n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is positive.                                          \n" +
           "double NM - Negative Match of the rule, equals the number of times   \n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is negative.                                          \n" +
           "double cost - the execution cost of the rule.          ",
           "int netIdx - the network NO.                                         \n" +
           "double reward - currently received reward.                           \n" +
           "double qDiscount - the discount factor for Q-Learning that favors    \n" +
           "reinfocement received sooner relative to that received later.        \n" +
           "double newMaxQVal - maximum Q-value of the new state resuling from   \n" +
           "current action in current state, Q(next state, next possible action) \n" +
           "double qVal - current q-value, Q(current state, current action).     \n" +
           "double threshold - the threshold for rule positivity.  ",
           "int netIdx - the network NO.                                         \n" +
           "double PMa - Positive Match of the rule a, equals the number of times\n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is positive.                                          \n" +
           "double NMa - Negative Match of the rule a, equals the number of times\n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is negative.                                          \n" +
           "double threshold - compared threshold value.      ",
           "int netIdx - the network NO.                                         \n" +
           "int setIdx - the rule subset NO.                                     \n" +
           "int ruleIdx - the rule NO in that rule subset.    ",
           "int netIdx - the network NO.                      ",
           "int netIdx - the network NO.                                        \n" +
           "double PM - Positive Match of the rule, equals the number of times  \n" +
           "that an input state matches rule condition, rule action is performed\n" +
           "and the result is positive.                                         \n" +
           "double NM - Negative Match of the rule, equals the number of times  \n" +
           "that an input state matches rule condition, rule action is performed\n" +
           "and the result is negative.                                         \n" +
           "double cost - the execution cost of the rule.     ",
           "int netIdx - the network NO.                                     \n" +
           "int ruleIdx - the rule NO.                        ",
           "int netIdx - the network NO.                                     \n" +
           "int ruleIdx - the rule NO.                        ",
        },
        {
           "null",
           "null",
           "null",
           "double PM - Positive Match of the rule, equals the number of times  \n" +
           "that an input state matches rule condition, rule action is performed\n" +
           "and the result is positive.                                         \n" +
           "double NM - Negative Match of the rule, equals the number of times  \n" +
           "that an input state matches rule condition, rule action is performed\n" +
           "and the result is negative.                                         \n" +
           "double cost - the execution cost of the rule.   ",
           "GKSChunk chunk1 - chunk for comparison.                         \n" +
           " ( see the class GKSChunk.java in subdirectory: clarion.nacs )  \n" +
           "GKSChunk chunk2 - chunk for comparison.              ",
           "null"
        },
        {
           "int RT_BL - Bottom-Level Response Time.   \n" +
           "int RT_TL - Top-Level Response Time.     ",
           "int PT_BL - Bottom-Level Perceptual Time. \n" +
           "int DT_BL - Bottom-Level Decision Time.   \n" +
           "int AT_BL - Bottom-Level Actuation Time. ",
           "int PT_TL - Top-Level Perceptual Time.    \n" +
           "int DT_TL - Top-Level Decision Time.      \n" +
           "int AT_TL - Top-Level Actuation Time.    ",
           "int PT_BL - Bottom-Level Perceptual Time.",
           "int DT_BL - Bottom-Level Decision Time.   \n" +
           "int actionTime - time determined by the operation dictated by \n" +
           "a particular rule.                                            \n" +
           "double ruleBla - rule base-level activation.                  \n" +
           "double chunkBla - action chunk base-level activation.        ",
           "int action - current action, deciding if NACS is involved or not.   \n" +
           "int assocAppTime - time of one iteration of associative applicaiton.\n" +
           "int chunkRetrTime - time of one iteration of chunk retrieval. ",
           "double assocRuleBla - associative rule base-level activation. ",
           "double chunkBla - nacs chunk base-level activation.           ",
           "null"
        }
  };
  final static String[][] returns = {
        {
           "the goals represented by parameter dimension-value. ",
           "goal parameters' dimension-value information",
           "null",
           "null",
           "the eligible condtions of an external network.",
           "current reinforcement",
           "the dimension weights of RER condtion",
           "the utility of RER",
           "if the action chosen according to a RER rule is reasonably good",
           "the information gain of RER rule A over RER rule B",
           "the dimension weights of IRL condtion",
           "the utility of IRL",
           "if the action chosen according to a IRL rule is reasonably good",
           "the information gain of IRL rule A over a random rule",
           "the IRL rule form",
           "the dimension weights of Fixed rule condtion",
           "the utility of Fixed rule",
           "the condition of Fixed Rule",
           "the action of Fixed Rule"
        },
        {
           "the externally encoded NACS chunks.",
           "the externally enocded NACS rules.",
           "the dimension weights of Associative rule condtion",
           "the utility of Associative rule",
           "the similarity between two NACS chunks using use defined measure",
           "return the dimension-value infomation of data pattern for training AAM"
        },
        {
           "the overall Response Time in each step",
           "the overall Bottom-Level Response Time in each step",
           "the overall Top-Level Response Time in each step",
           "the Top-Level Perceptual Time in each step",
           "the Top-Level Decision Time in each step",
           "the Top-Level Actuation Time in each step",
           "the time in GKS for one iteration of parallel applications of \n" +
           "associative rules",
           "the time of one iteration of chunk retrieval",
           "the time of one iteration of AMN mapping input to output"
        }
  };
  final static String[][] seeSamples = {
        {
           "goals set is empty",
           "Such as in Tower of Hanoi task, each goal item has six parameter \n" +
           "dimensions: (DSIZE, FROM, TO, DSIZE1, FROM1, TO1). Each parameter\n" +
           "dimension has some possible values.  ",
           "null",
           "null",
            "public short[][][] getEligibility( int netIdx )                           \n" +
            "{                                                                         \n" +
            "  //**                                                                    \n" +
            "   * suppose the input has 4 dimensions and each dimension has 5 values.  \n" +
            "   * Eligibility of net 0 is :                                             \n" +
            "   * if ((c1 = v11 or v12) and (c2 > v23)) or ((c3 <= v34) and (c4 >= v42))) \n" +
            "   * then use net 0.                                                    \n" +
            "   * ( vij means the jth value of dimension i. )                        \n" +
            "   *//                                                                  \n" +
            "    short[][][] eligCondition = {                                       \n" +
            "           {                                                              \n" +
            "                {0,1},        /// relevant dims: 0, 1, in this condition. \n" +
            "                {1,1,0,0,0},  /// c1 = v11 or v12.                        \n" +
            "                {0,0,0,1,1}   /// c2 > v23.                               \n" +
            "           },                                                             \n" +
            "           {                                                              \n" +
            "                 {2,3},       /// relevant dims: 2, 3, in this condition. \n" +
            "                 {1,1,1,1,0}, /// c3 <= v34.                              \n" +
            "                 {0,1,1,1,1}  /// c4 >= v42.                              \n" +
            "           }                                                              \n" +
            "    };                                                                    \n" +
            "    return eligCondition;                                                 \n" +
            "}                                                                         ",
           " the method - private double determineFeedback( int val ) \n" +
           " in AgentPC.java \n" +
           " private double determineFeedback( int val )\n" +
           " {                                          \n" +
           "   double fb;                               \n" +
           "   int diff = Math.abs( val - TARGET );     \n" +
           "   switch( diff )                           \n" +
           "   {                                        \n" +
           "    case 0:                                 \n" +
           "    case 1:                                 \n" +
           "         fb = POS_FEEDBACK;                 \n" +
           "         break;                             \n" +
           "    default:                                \n" +
           "         fb = NEG_FEEDBACK * ( diff - 1 );  \n" +
           "         break;                             \n" +
           "   }                                        \n" +
           "   return fb;                               \n" +
           " }                                          \n" +
           "",
           "the default RER condition weights",
           "other RER utility measures in this option",
           "the default RER positivity criterion",
           "the default RER information gain measure",
           "the default IRL condition weights",
           "other IRL utility measures in this option",
           "the default IRL positivity criterion",
           "the default IRL information gain measure",
           "/*\n In Process Control task, there are four sets of IRL rules.  \n" +
           "1. P = a * W + b           \n" +
           "2. P = a * W + c * P1 + b  \n" +
           "3. P = a * W1 + b          \n" +
           "4. P = a * W1 + c * P2 + b \n" +
           "a = {1,2},  b = {-1,-2,0,1,2},  c = {-1,-2,1,2}.            \n" +
           "P = the desired system output level(the goal),              \n" +
           "W = the current input to the system(to be determined),      \n" +
           "W1 = the previous input to the system,                      \n" +
           "P1 = the previous system output level(under W1),            \n" +
           "P2 = the system output level at the time step before P1.    \n" +
           "In this case, elements from the set a, b, c compose the IRL \n" +
           "parameters.   \n */                   \n\n" +
           "public class IrlRuleFormPC implements IrlFuleForm{       \n" +
           " int a, b, c, target, ruleSetIdx;                        \n" +
           " int curAction, nextAction;                              \n" +
           " public IrlRuleFormPC( int a, int b, int c, int target, int ruleSetIdx, int curAction ) \n" +
           " {                                                      \n" +
           "   this.a = a;                                          \n" +
           "   this.b = b;                                          \n" +
           "   this.c = c;                                          \n" +
           "   this.target = target;                                \n" +
           "   this.ruleSetIdx = ruleSetIdx;                        \n" +
           "   this.curAction = curAction;                          \n" +
           "   nextAction = ( int ) ( target - b ) / a;             \n" +
           " }               \n\n" +
           " public int determineAction( Feature[] input )          \n" +
           " {                                                      \n" +
           "    int p1 = input[0].getValues()[0];                   \n" +
           "    int chosenAction = Math.round( ( target - c * p1 - b ) / a ); \n" +
           "    if ( chosenAction < 0 )                             \n" +
           "         chosenAction = 0;                              \n" +
           "    else if ( chosenAction > 11 )                       \n" +
           "              chosenAction = 11;                        \n" +
           "    switch( type )                                      \n" +
           "    {                                                   \n" +
           "        case 0:  // action should be returned for this step. \n" +
           "        case 1:                                         \n" +
           "             curAction = chosenAction;                  \n" +
           "             return curAction;                          \n" +
           "        case 2:  // action should be returned for next step. \n" +
           "        case 3:                                         \n" +
           "             curAction = nextAction;                    \n" +
           "             nextAction = chosenAction;                 \n" +
           "             return curAction;                          \n" +
           "    }                                                   \n" +
           "    return curAction;                                   \n" +
           " }                                                      \n" +
           "}               ",
           "the default Fixed Rule condition weights",
           "other Fixed Rule utility measures in this option",
           "public short[][] getFrCondition(int netIdx, int ruleIdx)      \n" +
           "{                              \n" +
           "    //*                        \n" +
           "     * Suppose the input has 4 dimensions, one possible condtion of a Fixed Rule: \n" +
           "     * { {1,1,1,1}, {1,0,1}, {0,0,1}, {1,0} }                      \n" +
           "     *//                                                           \n" +
           "   short[][] conditions = { {1,1,1,1}, {1,0,1}, {0,0,1}, {1,0} };  \n" +
           "   return conditions;                                              \n" +
           "}                                ",
           " Suppose there are 4 actions, \n" +
           " the conclusion of a Fixed Rule may be : do action 1"
        },
        {
           "null",
           "null",
           "the default associative rule condition weights",
           "other associative rule utility measures in this option",
           "the default similarity measure in this option",
           "the default data pattern - experienced state"
        },
        {
           "the default overall response time formula",
           "the default overall bottome-level response time formula",
           "the default overall top-level response time formula",
           "the default top-level perceptual time formula",
           "the default top-level decision time formula",
           "public int getAT_TL( int action, int assocAppTime, int chunkRetrTime ) \n" +
           "{                                                 \n" +
           "  return 100 + 5 * assocAppTime + chunkRetrTime;  \n" +
           "}                                                   ",
           "the default associative rule application time formula",
           "the default chunk retrieval time formula",
           "the default AMN one-pass time formula"
        }
  };
  final static Sample[][] samples = new Sample[SUB_SYS_NUM][];{
        samples[0] = new Sample[ACS_SAMPLE_NUM];
        samples[1] = new Sample[NACS_SAMPLE_NUM];
        samples[2] = new Sample[RT_SAMPLE_NUM];
        for( int i = 0; i < SUB_SYS_NUM; i++ )
           for( int j = 0; j < samples[i].length; j++ ){
                samples[i][j] = new Sample( className, methodNames[i][j], methodForms[i][j],
                                descriptions[i][j], params[i][j], returns[i][j], seeSamples[i][j]);
                //System.out.println( samples[i][j] );
           }
  }
  // global variables.
  protected Global global;

  public TaskClarion()
  {
  }

  public static String getSample( int subSysIdx, int sampleIdx )
  {
     return samples[subSysIdx][sampleIdx].toString();
  }

  public void setGlobal( Global global )
  {
     this.global = global;
  }

  // the followings need be override by user's own codes.
  // goals.
  public short[][] getGoals()
  {
     return (short[][])null;
  }

  // each goal parameter dimension-value info.
  public short[] getGoalDims()
  {
     return (short[])null;
  }

  // bottom-up correction.
  public void runBUCorrection()
  {
  }

  // top-down guidance correction.
  public void runTGCorrection()
  {
  }

  // eligibility of ACS networks, return the eligible condition of the network.
  public short[][][] getEligibility( int netIdx )
  {
     /**
      * suppose the input has 4 dimensions and each dimension has 5 values.
      * Eligibility of net 0 is :
      * if ((c1 = v11 or v12) and (c2 > v23)) or ((c3 <= v34) and (c4 >= v42)))
      * then use net 0.
      * ( vij means the jth value of dimension i. )
      */
    short[][][] eligCondition = {
                   {
                      {0,1},          // relevant dims: 0, 1, in this condition.
                      {1,1,0,0,0},    // c1 = v11 or v12.
                      {0,0,0,1,1}     // c2 > v23.
                   },
                   {
                      {2,3},          // relevant dims: 2, 3, in this condition.
                      {1,1,1,1,0},    // c3 <= v34.
                      {0,1,1,1,1}     // c4 >= v42.
                   }
                };
     return eligCondition;
  }

  // reinforcement of ACS networks.
  public double getReinforcement( int netIdx, Feature[] input, Chunk[] wmItems, GoalChunk goalItem,
                short[] action, Feature[] newInput, Chunk[] newWmItems, GoalChunk newGoalItem )
  {
     /**
      * int netIdx - the network NO.
        Feature[] input - current input,
        ( see the class Feature.java in subdirectory: clarion.common )
        Chunk[] wmChunks - current working memory status.
        ( see the class WMChunk.java in subdirectory: clarion.acs.wm )
        int goal - current goal.
        short[] action - current action.
        Feature[] newInput - new input resulting from firing the action.
        Chunk[] newWMChunks - new working memory status.
        int newGoal - new goal.
     */
     int TARGET = 5;
     int P = newInput[0].getValues()[0];
     int diff = Math.abs( P - TARGET );
     double fb = 0.0;
     switch( diff ){
         case 0:
         case 1:
              fb = 1.0;
              break;
         default:
              fb = -0.2 * ( diff - 1 );
              break;
     }
     return fb;
  }

  // each dimension weight in RER condition.
  public double[] getRerWeights( int netIdx )
  {
      /**
       * the sum of the weights should be 1.0.
       */
     int dimNum = global.getARInputDimNum( netIdx, Global.RER );
     double[] weights = new double[dimNum];
     for( int i = 0; i < dimNum; i++ )
          weights[i] = 1.0/dimNum;
     return weights;
  }

  // utility measure of RER
  public double getRerUtility( int netIdx, int setIdx, double PM, double NM )
  {
     /**
      * int netIdx - the network NO.
      * int setIdx - the rule subset NO.
      * double PM - Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
        double NM - Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
      */
     double cost = 1.0, v = 1.0, benefit = 2.0, c7, c8, utl = 0.0;
     int option = global.getRuleUtilityOption( netIdx, Global.RER, setIdx );
     if( option == Global.CONSTANT_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_V];
         benefit = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_BENE];
     }
     if( option == Global.VARIED_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_V];
         c7 = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_C7];
         c8 = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_C8];
         benefit = (c7 + PM) / (c8 + PM + NM );
     }
     utl = benefit - v * cost;
     return utl;
  }

  // RER positivity criterion, return if it is positive or not.
  public boolean getRerPositivity( int learnOption, double reward, double qDiscount,
                                   double newMaxQVal, double qVal, double threshold )
  {
     /**
      * int learnOption - learning method.
      * double reward - currently received reward.
        double qDiscount - the discount factor for Q-Learning that favors
        reinfocement received sooner relative to that received later.
        used in q-learning method.
        double newMaxQVal - maximum Q-value of the new state resuling from
        current action in current state, Q(next state, next possible action).
        used in q-learning method.
        double qVal - current q-value, Q(current state, current action).
        double threshold - the threshold for rule positivity.
      */
     boolean positivity = false;
     if( learnOption == Global.QL ) {
         if( reward + qDiscount * newMaxQVal - qVal > threshold )
             positivity = true;
         else
             positivity = false;
     }
     else if( learnOption == Global.SQL ){
              if( reward > threshold )
                  positivity = true;
              else
                  positivity = false;
          }
          else{
              if( reward > 0 )
                  positivity = true;
              else
                  positivity = false;
          }
     return positivity;
  }

  // RER Information Gain criterion
  public double getRerIG( int netIdx, double PMa, double NMa, double PMb, double NMb )
  {
     /**
      * int netIdx - the network NO. decdies the values of c1 and c2.
      * double PMa - Positive Match of the rule a, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
        double NMa - Negative Match of the rule a, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
        double PMb - Positive Match of the rule b.
        double NMb - Negative Match of the rule b.
      */
      double c1 = 1.0, c2 = 2.0;
      c1 = global.getRerIGParaAbove( netIdx );
      c2 = global.getRerIGParaUnder( netIdx );
      double log2 = Math.log( 2 );
      double rateA = (PMa + c1) / ( PMa + NMa + c2 );
      double rateB = (PMb + c1) / ( PMb + NMb + c2 );
      double ig = Math.log( rateA / rateB ) / log2;
      return ig;
  }

  // each dimension weight in IRL condition.
  public double[] getIrlWeights( int netIdx )
  {
      /**
       * the sum of the weights should be 1.0.
       */
     int dimNum = global.getARInputDimNum( netIdx, Global.IRL );
     double[] weights = new double[dimNum];
     for( int i = 0; i < dimNum; i++ )
          weights[i] = 1.0/dimNum;
     return weights;
  }

  // utility measure of IRL
  public double getIrlUtility( int netIdx, int setIdx, double PM, double NM )
  {
     /**
      * int netIdx - the network NO.
      * int setIdx - the rule subset NO.
      * double PM - Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
        double NM - Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
      */
     double cost = 1.0, v = 1.0, benefit = 2.0, c7, c8, utl = 0.0;
     int option = global.getRuleUtilityOption( netIdx, Global.IRL, setIdx );
     if( option == Global.CONSTANT_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_V];
         benefit = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_BENE];
     }
     if( option == Global.VARIED_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_V];
         c7 = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_C7];
         c8 = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_C8];
         benefit = (c7 + PM) / (c8 + PM + NM );
     }
     utl = benefit - v * cost;
     return utl;
  }

  // IRL positivity criterion, return if it is positive
  public boolean getIrlPositivity( int learnOption, double reward, double qDiscount,
                                   double newMaxQVal, double qVal, double threshold )
 {
     /**
      * int learnOption - learning method.
      * double reward - currently received reward.
        double qDiscount - the discount factor for Q-Learning that favors
        reinfocement received sooner relative to that received later.
        used in q-learning method.
        double newMaxQVal - maximum Q-value of the new state resuling from
        current action in current state, Q(next state, next possible action).
        used in q-learning method.
        double qVal - current q-value, Q(current state, current action).
        double threshold - the threshold for rule positivity.
      */
     boolean positivity = false;
     if( learnOption == Global.QL ) {
         if( reward + qDiscount * newMaxQVal - qVal > threshold )
             positivity = true;
         else
             positivity = false;
     }
     else if( learnOption == Global.SQL ){
              if( reward > threshold )
                  positivity = true;
              else
                  positivity = false;
          }
          else {
              if( reward > 0 )
                  positivity = true;
              else
                  positivity = false;
          }
     return positivity;
  }

  // IRL Information Gain criterion
  public double getIrlIG( int netIdx, double PM, double NM, double threshold )
  {
     /**
      * int netIdx - the network NO. decdies the values of c5 and c6.
      * double PM - Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
        double NM - Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
        double threshold - compared threshold value.
      */
      double c5 = 1.0, c6 = 2.0;
      c5 = global.getIrlIGParaAbove( netIdx );
      c6 = global.getIrlIGParaUnder( netIdx );
      double log2 = Math.log( 2 );
      double rate = (PM + c5) / ( PM + NM + c6 );
      double ig = Math.log( rate / threshold ) / log2;
      return ig;
  }

  // IRL parameters to decide an action
  public IrlRuleForm getIrlRule( int groupIdx, int netIdx, int setIdx, int ruleIdx )
  {
   /**
    * In Process Control task, there are four sets of IRL rules.
       1. P = a * W + b
       2. P = a * W + c * P1 + b
       3. P = a * W1 + b
       4. P = a * W1 + c * P2 + b
       a = {1,2},  b = {-1,-2,0,1,2},  c = {-1,-2,1,2}.
       P = the desired system output level(the goal),
       W = the current input to the system(to be determined),
       W1 = the previous input to the system,
       P1 = the previous system output level(under W1),
       P2 = the system output level at the time step before P1.
       In this case, elements from the set a, b, c compose the IRL
       parameters.

     smaple code is the following:

     public class IrlRuleFormPC implements IrlRuleForm{
          int a, b, c, target, ruleSetIdx;
          int curAction, nextAction;

          public IrlRuleFormPC( int a, int b, int c, int target, int ruleSetIdx, int curAction )
          {
                this.a = a;
                this.b = b;
                this.c = c;
                this.target = target;
                this.ruleSetIdx = ruleSetIdx;
                this.curAction = curAction;
                nextAction = ( int ) ( target - b ) / a;
          }

          public double determineAction( short[] input, short[] suggestedAction )
          {
             int p1 = 0;
             for( int i = 0; i < input.length; i++ )
                  if( input[i] == 1 ){
                      p1 = i;
                      break;
                  }
             int chosenAction = Math.round( ( target - c * p1 - b ) / a );
             if ( chosenAction < 0 )
                  chosenAction = 0;
             else if ( chosenAction > 11 )
                       chosenAction = 11;
             switch( ruleSetIdx )
             {
                 case 0:  // action should be returned for this step.
                 case 1:
                      curAction = chosenAction;
                      suggestedAction[0] = (short)curAction;
                      break;
                 case 2:  // action should be returned for next step.
                 case 3:
                      curAction = nextAction;
                      nextAction = chosenAction;
                      suggestedAction[0] = (short)curAction;
                      break;
             }
             return 1.0;
          }

          public void generalize()
          {
          }

          public void specialize()
          {
          }
     }
     */
     return (IrlRuleForm)null;
  }

  // each dimension weight in FR condition.
  public double[] getFrWeights( int netIdx )
  {
      /**
       * the sum of the weights should be 1.0.
       */
     int dimNum = global.getARInputDimNum( netIdx, Global.FR );
     double[] weights = new double[dimNum];
     for( int i = 0; i < dimNum; i++ )
          weights[i] = 1.0/dimNum;
     return weights;
  }

  // utility measure of Fixed Rule
  public double getFrUtility( int netIdx, int setIdx, double PM, double NM )
  {
     /**
      * int netIdx - the network NO.
      * int setIdx - the rule subset NO.
      * double PM - Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
        double NM - Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
      */
     double cost = 1.0, v = 1.0, benefit = 2.0, c7, c8, utl = 0.0;
     int option = global.getRuleUtilityOption( netIdx, Global.FR, setIdx );
     if( option == Global.CONSTANT_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_V];
         benefit = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_BENE];
     }
     if( option == Global.VARIED_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_V];
         c7 = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_C7];
         c8 = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_C8];
         benefit = (c7 + PM) / (c8 + PM + NM );
     }
     utl = benefit - v * cost;
     return utl;
  }

  // FR positivity criterion, return if it is positive or not.
  public boolean getFrPositivity( int learnOption, double reward, double qDiscount,
                                   double newMaxQVal, double qVal, double threshold )
  {
     /**
      * int learnOption - learning method.
      * double reward - currently received reward.
        double qDiscount - the discount factor for Q-Learning that favors
        reinfocement received sooner relative to that received later.
        used in q-learning method.
        double newMaxQVal - maximum Q-value of the new state resuling from
        current action in current state, Q(next state, next possible action).
        used in q-learning method.
        double qVal - current q-value, Q(current state, current action).
        double threshold - the threshold for rule positivity.
      */
     boolean positivity = false;
     if( learnOption == Global.QL ) {
         if( reward + qDiscount * newMaxQVal - qVal > threshold )
             positivity = true;
         else
             positivity = false;
     }
     else if( learnOption == Global.SQL ){
              if( reward > threshold )
                  positivity = true;
              else
                  positivity = false;
          }
          else{
              if( reward > 0 )
                  positivity = true;
              else
                  positivity = false;
          }
     return positivity;
  }

  // Fixed Rule conditions
  public short[][] getFrCondition(int groupIdx, int netIdx, int ruleIdx)
  {
      /**
       * Suppose the input has 4 dimensions, each dimension has 4 values.
       * the value of each element: if the element is in the condition, = 1, otherwise = 0.
       * condtion of a Fixed Rule:
       */
     short[][] conditions = { {1,1,1,1}, {1,0,1,0}, {0,0,0,1}, {1,0,1,0} };
     return conditions;
  }

  // Fixed Rule actions
  public short[] getFrAction(int groupIdx, int netIdx, int ruleIdx)
  {
      /**
       * Suppose the output has 4 dimensions, so the action represented by 1-dimension
       * array with 4-elements, the value of each element is the value of the action
       * in that output dimension.
       */
     short[] action = { 1, 0, 2, 3 };
     return action;
  }

  /**
   * used for the action of a Fixed Rule is variable and depends on current input.
   * fill in the suggestedAction under current input and return the rule support for
   * current input. the following sample is from PC task.
   */
  public double decideFrAction( int groupIdx, int netIdx, int setIdx, Feature[] sensoryInput,
                GoalChunk gsItem, GKSChunk[] wmItems, short[] suggestedAction )
  {
     /**
      *  sample code of PC task.
      *  target: the target production.
      *  begin:  the start pooint of P1 dimension in the input array.
      *  end:    the end pooint of P1 dimension in the input array.
      *  idx:    current P1 value index on the P1 dimension.
      */
     int target = 5, begin = 0, end = 11, idx = 0;
     short[] state = sensoryInput[0].getContent();
     for( int i = begin; i <= end; i++ ){
          if( state[i] == 1 ){
              idx = i;
              break;
          }
     }
     suggestedAction[0] = (short)(( idx + target ) / 2);
     return 1.0;
  }

  // get externally encoded NACS chunks.
  public GKSChunk[] getExternalNacsChunks()
  {
     return (GKSChunk[])null;
  }

  // get externally encoded NACS rules.
  public AssocRule[] getExternalNacsRules()
  {
     return (AssocRule[]) null;
  }

  // condition weights of Associative rule.
  public double[] getAssocWeights( int condNum )
  {
      /**
       * the sum of the weights should be 1.0.
       */
     double[] weights = new double[condNum];
     for( int i = 0; i < condNum; i++ )
          weights[i] = 1.0/condNum;
     return weights;
  }

  // dimension weights of Associative rule.
  public double[] getAssocDimWeights( int dimNum )
  {
      /**
       * the sum of the weights should be 1.0.
       */
     //int dimNum = global.getGksInputDimNum();
     double[] weights = new double[dimNum];
     for( int i = 0; i < dimNum; i++ )
          weights[i] = 1.0/dimNum;
     return weights;
  }

  // utility measure of Associative rule.
  public double getAssocUtility( double PM, double NM )
  {
     /**
      * int netIdx - the network NO.
      * int setIdx - the rule subset NO.
      * double PM - Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
        double NM - Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
      */
     double cost = 1.0, v = 1.0, benefit = 2.0, c7, c8, utl = 0.0;
     int option = global.getAssocUtilityOption();
     if( option == Global.CONSTANT_OPTION ) {
         cost = global.getAssocUtilityParas()[Global.UTL_COST];
         v = global.getAssocUtilityParas()[Global.UTL_V];
         benefit = global.getAssocUtilityParas()[Global.UTL_BENE];
     }
     if( option == Global.VARIED_OPTION ) {
         cost = global.getAssocUtilityParas()[Global.UTL_COST];
         v = global.getAssocUtilityParas()[Global.UTL_V];
         c7 = global.getAssocUtilityParas()[Global.UTL_C7];
         c8 = global.getAssocUtilityParas()[Global.UTL_C8];
         benefit = (c7 + PM) / (c8 + PM + NM );
     }
     utl = benefit - v * cost;
     return utl;
  }

  // chunk similarity measure: chunk1/chunk2
  public double getSimilarity(GKSChunk chunk1, GKSChunk chunk2)
  {
       double similarity = 0.0;
       if( chunk1 == null || chunk2 == null )
           return similarity;
       // calculate similarity.
       Feature[] featList1 = chunk1.getFeatures();
       Feature[] featList2 = chunk2.getFeatures();
       int matchCount = 0;
       for( int i = 0; i < featList1.length; i++ ){
            // search the index of the ith feature in list1 on list2.
            int idx = chunk2.searchFeature( featList1[i].getLabel() );
            if( idx == -1 || featList1[i].getType() != featList2[idx].getType() )
                continue;
            short[] arr1 = featList1[i].getValues();
            short[] arr2 = featList2[idx].getValues();
            Arrays.sort( arr1 );
            Arrays.sort( arr2 );
            if( Arrays.equals( arr1, arr2 ) )
                matchCount++;
       }
       // similarity-based reasoning, use the default similarity measure.
       double val = Math.pow( featList2.length, 1.001 );
       similarity = (double)matchCount / val;
       return similarity;
  }

  // type of data pettern for training AAM.
  public short[] getAamDataPattern(int netIdx, int ruleIdx)
  {
     short[] types = {0};
     return types;
  }

  // overall RT formula
  public int getOverallRT( int RT_BL, int RT_TL )
  {
     /**
      * int RT_BL - Bottom-Level Response Time.
        int RT_TL - Top-Level Response Time.
      */
     return Math.max( RT_BL, RT_TL );
  }

  // overall Bottom-Level RT formula
  public int getBLOverallRT( int PT_BL, int DT_BL, int AT_BL )
  {
     /**
      * int PT_BL - Bottom-Level Perceptual Time.
        int DT_BL - Bottom-Level Decision Time.
        int AT_BL - Bottom-Level Actuation Time.
      */
     return PT_BL + DT_BL + AT_BL;
  }

  // overall Top-Level RT formula
  public int getTLOverallRT( int PT_TL, int DT_TL, int AT_TL )
  {
     /**
      * int PT_TL - Top-Level Perceptual Time.
        int DT_TL - Top-Level Decision Time.
        int AT_TL - Top-Level Actuation Time.
      */
     return PT_TL + DT_TL + AT_TL;
  }

  // PT of TL
  public int getPT_TL( int PT_BL )
  {
     /**
      * int PT_BL - Bottom-Level Perceptual Time.
      */
     return PT_BL + 150;
  }

  // DT of TL
  public int getDT_TL( int DT_BL, int actionTime, double ruleBla, double chunkBla )
  {
     /**
      * int DT_BL - Bottom-Level Decision Time.
      * int actionTime - time determined by the operation dictated by
        a particular rule.
        double ruleBla - rule base-level activation.
        double chunkBla - action chunk base-level activation.
      */
     double t0 = global.getDT_TL_t0();
     double t1 = global.getDT_TL_t1();
     int TLDT = (int)( actionTime + t0 / ruleBla + t1 / chunkBla );
     return TLDT;
  }

  // AT of TL, the action
  public int getAT_TL( boolean flag, int assocAppTime, int chunkRetrTime )
  {
     /**
      * boolean flag - decides if NACS is involved or not.
        int assocAppTime - time of one iteration of associative applicaiton.
        int chunkRetrTime - time of one iteration of chunk retrieval.
      */
     if( flag )
         return 200 + assocAppTime + chunkRetrTime;
     else
         return 500;
  }

  // Associative rule application time.
  public int getAssocAppTime( double assocRuleBla )
  {
     /**
      * double assocRuleBla - associative rule base-level activation.
      */
      double t4 = global.getAssocAppTime_t4();
      double t5 = global.getAssocAppTime_t5();
      int assocRuleAppTime = (int)(t4 + t5 / assocRuleBla );
      return assocRuleAppTime;
  }

  // NACS chunk retrieval time.
  public int getChunkRetrTime( double chunkBla )
  {
     /**
      *  double chunkBla - nacs chunk base-level activation.
      */
      double t2 = global.getChunkRetTime_t2();
      double t3 = global.getChunkRetTime_t3();
      int chunkRetTime = (int)(t2 + t3 / chunkBla);
      return chunkRetTime;
  }

  // AMN one-pass time.
  public int getAmnPassTime()
  {
     return global.getAmnOnePassTime();
  }

  class Sample{
     String cls;
     String method;
     String form;
     String description;
     String params;
     String returns;
     String seeSample;

     Sample( String cls, String method, String form, String description, String params,
             String returns, String seeSample )
     {
        this.cls = cls;
        this.method = method;
        this.form = form;
        this.description = description;
        this.params = params;
        this.returns = returns;
        this.seeSample = seeSample;
     }

     public String toString()
     {
        String str = "";
        str += "USER CLASS : \n" + cls + "\n\n";
        str += "USER METHOD : \n" + method + "\n\n";
        str += "METHOD CALLING FORM : \n" + form + "\n\n";
        str += "METHOD DESCRIPTION : \n" + description + "\n\n";
        str += "PARAMETERS : \n" + params + "\n\n";
        str += "RETURNS : \n" + returns + "\n\n";
        str += "SAMPLES : \n" + seeSample + "\n\n";
        return str;
     }
  }
}