// will use Boltzman probability distribution to choose from among its elements
// p(value)=e^(value)/SUM_every value(e^(evalue))
package clarion.tools;
import java.lang.*;

public class StochasticDecider
{
    protected int size = 0;
    //protected double threshold = 0.0005;
    protected double threshold = 0.0;
    protected double temperature;
    protected double BoltzmannSum = 0;
    protected double[] elementsArr; // used for const num of elems

    public StochasticDecider()
    {
    }

    public StochasticDecider( int nSize, double temper )
    {
        size = nSize;
        temperature = temper;
        elementsArr = new double[ nSize ];
    }

    public StochasticDecider( double temper )
    {
        temperature = temper;
    }

    public void setSize( int length )
    {
        size = length;
    }

    public void setTemperature( double value )
    {
        temperature = value;
    }

    public void setElements( double[] els )
    {
        BoltzmannSum = 0;
        for ( int i = 0; i < size; i++ ) {
              elementsArr[i] = Math.exp( ( els[i] ) / temperature );
              BoltzmannSum += elementsArr[i];
        }
    }

    public int getRecommendation()
    {
        int decision = 0;
        double lottery = Math.random() * BoltzmannSum;
        double currentSum = elementsArr[0];

        while( ( lottery - currentSum ) > threshold )
        {
           decision++;
           currentSum += elementsArr[decision];
        }
        return decision;
    }

    public int getRecommendation( double[] els )
    {
        size = els.length;
        double[] inputArr = new double[ size ];

        // fill in the input array of Boltzmann distribution.
        BoltzmannSum = 0;
        for ( int i = 0; i < size; i++ )
        {
           inputArr[i] = Math.exp( ( els[i] ) / temperature );
           BoltzmannSum += inputArr[i];
        }

        // make a decision according to Boltzmann distribution.
        int decision = 0;
        double lottery = Math.random() * BoltzmannSum;
        double currentSum = inputArr[0];
        while( ( lottery - currentSum ) > threshold ){
               decision++;
               currentSum += inputArr[ decision ];
        }

        return decision;
    }

    public int getRecommendation( double[] els, int len )
    {
        size = len;
        double[] inputArr = new double[ size ];

        // fill in the input array of Boltzmann distribution.
        BoltzmannSum = 0;
       	for ( int i = 0; i < size; i++ ){
         	    inputArr[i] = Math.exp( ( els[i] ) / temperature );
         	    BoltzmannSum += inputArr[i];
        }

        // make a decision according to Boltzmann distribution.
        int decision = 0;
        double lottery = Math.random() * BoltzmannSum;
        double currentSum = inputArr[0];
        while( ( lottery - currentSum ) > threshold ) {
               decision++;
               currentSum += inputArr[ decision ];
        }

        return decision;
    }

    // the chosen action is the distance from start.
    public int getRecommendation( double[] els, int start, int end )
    {
        size = end - start;
        double[] inputArr = new double[ size ];
        // fill in the input array of Boltzmann distribution.
        BoltzmannSum = 0;
        for( int i = start; i < end; i++ ){
            inputArr[i-start] = Math.exp( ( els[i] ) / temperature );
            BoltzmannSum += inputArr[i-start];
        }
        // make a decision according to Boltzmann distribution.
        int decision = 0;
        double lottery = Math.random() * BoltzmannSum;
        double currentSum = inputArr[0];
        while( ( lottery - currentSum ) > threshold ){
            decision++;
            currentSum += inputArr[ decision ];
        }
        return decision;
    }

    public short[] getRecommendation( double[] els, short[] dimVals  )
    {
       short[] decision = new short[ dimVals.length ];
       int offset = 0;
       for( int i = 0; i < decision.length; i++ ){
            decision[i] = (short)getRecommendation( els, offset, offset + dimVals[i] );
            offset += dimVals[i];
       }
       return decision;
    }
}
