package main;
import java.awt.*;
import java.io.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.options.*;
import clarion.tools.*;
import tasksimulators.*;
import tasksimulators.oldsimu.*;

/**
 *  this is the main class of CLARION model. it includes both the simulated tasks
    of old version and the simulated tasks of new version.
*/
public class ClarionMain extends JFrame {
/***************************VARIABLES DECLARATION BEGIN************************/
  // general constant variables for designing window in which results are shown.
   final int WINDOW_WIDTH = 430,  WINDOW_HEIGHT = 520;
   final int OLD = 0, NEW = 1;
   public final static String CLARION_TASK_TEXT[] ={
       "PROCESS CONTROL TASK", "LETTER COUNTING TASK", "TOWER OF HANOI TASK",
       "CATEGORY INFERENCE TASK", "DISCOVERY TASK", "ARTIFICIAL GRAMMAR LEARNING TASK"
   };
   int simuType, taskNo, expNo, paraMainOptInfo[][];
   double defParaArray[], paraValArray[];
   String loadTaskName;
   String[] newTaskNames = new String[Global.MAX_TASK_NUM];
   JFrame thisFrame;
   // caption label.
   JLabel capLabel = new JLabel();
   // old task simulation panel.
   JPanel oldTaskPanel = new JPanel();
   JLabel oldTaskLabel = new JLabel();
   JButton oldTaskButtons[];
   // old task simulation panel.
   JPanel newTaskPanel = new JPanel();
   JLabel newTaskLabel = new JLabel();
   JLabel loadLabel = new JLabel();
   JButton loadButton = new JButton();
   JButton newTaskButton = new JButton();
   //JButton clarionButton = new JButton();
   JPopupMenu loadTaskMenu;
   JMenuItem loadTaskItem[];
   // exit button
   JButton exitButton = new JButton();
   // the tasks.
   PCTask   pcTask;
   LCTask   lcTask;
   THTask   thTask;
   CITask   ciTask;
   DSTask   dsTask;
   Task     newTask;
   TaskClarion taskClarion;
   // the simulators.
   Simulator oldSimulators[];
   Simulator newSimulator;

 public ClarionMain() {

     System.out.println( System.getProperty("user.dir") );

     simuType = OLD;
     taskNo = World.PC_TASK;
     expNo = World.CLARION_MODE;
     defParaArray = new double[ World.TASK_PARA_NUM ];
     paraValArray = new double[ World.TASK_PARA_NUM ];
     if( expNo == World.CLARION_MODE )
         for( int i = 0; i < World.TASK_PARA_NUM; i++ )
              paraValArray[ i ] = World.DEF_TASK_PARA_ARRAY[ taskNo ] [ i ];
     paraMainOptInfo = new int[ World.MAIN_OPTION_NUM ][];
     for( int i = 0; i < World.MAIN_OPT_ITEM_NUM.length; i++ ){
          paraMainOptInfo[ i ] = new int[World.MAIN_OPT_ITEM_NUM[i]];
          World.MODEL_PARA_ARRAY[ i ] = new int[World.MAIN_OPT_ITEM_NUM[i]];
     }
     thisFrame = this;
     // the simulated tasks.
     pcTask = new PCTask( thisFrame );
     lcTask = new LCTask( thisFrame );
     thTask = new THTask( thisFrame );
     ciTask = new CITask( thisFrame );
     dsTask = new DSTask( thisFrame );
     // old task simulators.
     oldSimulators = new Simulator[ World.CLARION_TASK_NUM ];
     oldSimulators[0] = new PCsimulator( this, pcTask );
     oldSimulators[1] = new Simulator( this, lcTask );
     oldSimulators[2] = new Simulator( this, thTask );
     oldSimulators[3] = new Simulator( this, ciTask );
     oldSimulators[4] = new Simulator( this, dsTask );
     initGui();
 }

 private void initGui()
 {
     final int WIDTH = WINDOW_WIDTH - 30;
     setSize( WINDOW_WIDTH, WINDOW_HEIGHT );
     this.getContentPane().setLayout(null);
     // the caption.
     capLabel.setFont(new java.awt.Font("Dialog", 0, 18));
     capLabel.setText("CLARION  MODEL");
     capLabel.setBounds(new Rectangle(110, 2, 160, 26));
     // new simulation panel.
     newTaskPanel.setLayout( null );
     newTaskPanel.setBorder( BorderFactory.createLineBorder(Color.black) );
     newTaskPanel.setBounds( new Rectangle(10, 30, WIDTH, 160) );
     newTaskLabel.setText( "Simulating a new task," );
     newTaskLabel.setBounds( new Rectangle((WIDTH-200)/2, 4, 200, 20) );
     loadLabel.setText( "No task loaded" );
     loadLabel.setBounds( new Rectangle(230, 40, 160, 20) );
     loadButton.setText( "LOAD A TASK" );
     loadButton.setBounds(new Rectangle(100, 40, 120, 20));
     loadButton.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(ActionEvent e) {
           loadButton_actionPerformed( e );
        }
     });
     newTaskButton.setText( "NEW TASK SIMULATOR" );
     newTaskButton.setBounds(new Rectangle((WIDTH-200)/2, 80, 200, 20));
     newTaskButton.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(ActionEvent e) {
          newTaskButton_actionPerformed(e);
        }
     });
     newTaskButton.setEnabled( false );
     /*
     clarionButton.setText("CLARION SETTINGS");
     clarionButton.setBounds(new Rectangle((WIDTH-200)/2, 120, 200, 20));
     clarionButton.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(ActionEvent e) {
          clarionButton_actionPerformed(e);
        }
     });
     clarionButton.setEnabled( false );
     */
     // old simulation panel.
     oldTaskPanel.setLayout( null );
     oldTaskPanel.setBorder( BorderFactory.createLineBorder(Color.black) );
     oldTaskPanel.setBounds( new Rectangle(10, 200, WIDTH, 240) );
     oldTaskLabel.setText( "Old simulation" );
     oldTaskLabel.setBounds( new Rectangle((WIDTH-180)/2, 4, 180, 20) );
     OldTaskButtonHandler handler = new OldTaskButtonHandler();
     oldTaskButtons = new JButton[ World.CLARION_TASK_NUM ];
     for( int i = 0; i < World.CLARION_TASK_NUM; i++ ){
         oldTaskButtons[i] = new JButton( CLARION_TASK_TEXT[i] );
         oldTaskButtons[i].setBounds(new Rectangle((WIDTH-260)/2, 40+40*i, 260, 20));
         oldTaskButtons[i].setEnabled( true );
         oldTaskButtons[i].addActionListener( handler );
         oldTaskPanel.add( oldTaskButtons[i] );
     }
     exitButton.setText("EXIT");
     exitButton.setBounds(new Rectangle((WIDTH-100)/2, 460, 100, 20));
     exitButton.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(ActionEvent e) {
          exitButton_actionPerformed(e);
      }
     });
     // add components.
     newTaskPanel.add( newTaskLabel );
     newTaskPanel.add( loadLabel );
     newTaskPanel.add( loadButton );
     newTaskPanel.add( newTaskButton );
     //newTaskPanel.add( clarionButton );
     oldTaskPanel.add( oldTaskLabel );
     this.getContentPane().add(capLabel, null);
     this.getContentPane().add(newTaskPanel, null);
     this.getContentPane().add(oldTaskPanel, null);
     this.getContentPane().add(exitButton, null);
     validate();
     setEnabled( true );
     setVisible( true );
     // for old task simulatons.
     setModelParaSet();
     setTaskParaSet();
 }

 public int getExpNo( )
 {
     return expNo;
 }

 public int getTaskNo()
 {
    return taskNo;
 }

 public int[][] getModelParaSet()
 {
    return paraMainOptInfo;
 }

 public void setModelParaSet()
 {
    for( int i = 0; i < paraMainOptInfo.length; i++ )
     for( int j = 0; j < paraMainOptInfo[ i ].length; j++ )
          paraMainOptInfo[i][j] =  World.DEF_MODEL_PARA_ARRAY[taskNo][i][j];
 }

 public double[] getTaskParaSet()
 {
    return paraValArray;
 }

 public void setTaskParaSet()
 {
    for( int i = 0; i < paraValArray.length; i++ )
         paraValArray[ i ] =  World.DEF_TASK_PARA_ARRAY[ taskNo ][ i ];
 }


 public static void main( String args[] )
 {
    ClarionMain aDemo = new ClarionMain();
 }

/********************************PRIVATE METHODS *******************************/
  // task selecting buttons handler
  private class OldTaskButtonHandler implements ActionListener
  {
     public void actionPerformed( ActionEvent e ){
       for( int i = 0; i < World.CLARION_TASK_NUM; i++ )
          if( e.getSource() == oldTaskButtons[ i ] ){
              simuType = OLD;
              taskNo = i;
              setModelParaSet();
              setTaskParaSet();
              setEnabled( false );
              setVisible( false );
              oldSimulators[taskNo].setEnabled(true);
              oldSimulators[taskNo].setVisible(true);
          }
     }
  }

  private void loadButton_actionPerformed(ActionEvent e)
  {
     loadTaskName = Global.getTaskClassName();
     int currentTaskNum = searchCurrentTasks( newTaskNames );
     loadTaskMenu = new JPopupMenu( "EXISTING TASKS" );
     loadTaskItem = new JMenuItem[ currentTaskNum ];
     for( int i = 0; i < currentTaskNum; i++ ){
          loadTaskItem[i] = new JMenuItem( newTaskNames[i] );
          loadTaskItem[i].setFont( new Font( "Serif", Font.PLAIN, 14 ) );
          final int focus = i;
          loadTaskItem[ focus ].addActionListener( new ActionListener(){
             public void actionPerformed( ActionEvent e){
                 if( !newTaskNames[focus].equals( loadTaskName ) &&
                     JOptionPane.showConfirmDialog( null, "want to load this task?",
                         "WARNING", JOptionPane.YES_NO_OPTION ) == JOptionPane.YES_OPTION ) {
                     loadTaskName = newTaskNames[focus];
                     loadLabel.setText( loadTaskName  + " is loaded" );
                     Global.setTaskName( loadTaskName );
                     newTaskButton.setEnabled( true );
                     //clarionButton.setEnabled( true );
                     ClarionClassLoader classLoader = new ClarionClassLoader();
                     String loadTaskDir = Global.getTaskDir() + loadTaskName;
                     try{
                          System.out.println( loadTaskDir );
                          String[] classNames = new File( loadTaskDir ).list();
                          for( int i = 0; i < classNames.length; i++ ){
                              String className = classNames[i].substring(0, classNames[i].length()-6 );
                              String fullClassName = Global.getTaskPackageName() + loadTaskName +"."+className;
                              Object obj = classLoader.loadClass( fullClassName ).newInstance();
                              if( obj instanceof Task ){     // load Task
                                  newTask = (Task)obj;
                                  Global.setTaskClassName( className );
                              }
                              if( obj instanceof TaskClarion ){   // load TaskClarion
                                  taskClarion = (TaskClarion)obj;
                                  //Global.setTaskClarion( taskClarion );
                                  Global.setTaskClarionClass( taskClarion.getClass() );
                                  System.out.println( "samples are loaded!" );
                              }
                          }
                          Global.setTask( newTask );
                          newSimulator = new NewSimulator( (ClarionMain)thisFrame, (SimuTask)newTask );

                     }
                     catch( Exception e1 ){
                            System.out.println( e1 );
                     }
                }
            }
          } );
          loadTaskMenu.add( loadTaskItem[ i ] );
     }
     loadTaskMenu.show( loadButton, 10, 10 );
  }

  private void newTaskButton_actionPerformed(ActionEvent e)
  {
       simuType = NEW;
       setEnabled( false );
       setVisible( false );
       newSimulator.setEnabled(true);
       newSimulator.setVisible(true);
  }

  /*
  private void clarionButton_actionPerformed(ActionEvent e)
  {
       simuType = NEW;
       guiClarion.setMainFrame( thisFrame );
       setEnabled( false );
       setVisible( false );
       guiClarion.setEnabled( true );
       guiClarion.setVisible( true );
  }
  */

  private void exitButton_actionPerformed(ActionEvent e) {
       System.exit( 0 );
  }

  public int searchCurrentTasks( String[] names )
  {
     String path = Global.getTaskDir();
     File file = new File( path );
     int len = file.list().length;
     for( int i = 0; i < len; i++ )
          names[i] = file.list()[i];
     return len;
  }
}