package tasksimulators.newsimu.AGL;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;
import java.util.Vector;

public class AgentAGL extends TaskAgent{

 final int SESSION_TIME = 1200000;
 final int SESSION_NUM = 2, PHASE_NUM = 2;
 //final int SESSION_NUM = 2, PHASE_NUM = 2;
 final int CONTROL_TYPE = 0, IMPLICIT_TYPE = 1;
 final int EXPLICIT_TYPE = 2, SYNERGY_TYPE = 3;
 final int TRAIN = 0, TEST = 1;
 //final double LOCK_PROB = 0.8;
 final double LOCK_PROB = 0.9;
 static short[][][] trainSet;
 static short[][][] testSet;
 char[] groupType;  // one of the groups: "CC", "IE", "EI", "II", "EE", "SS"
 int sessionIdx, stageIdx;
 int learningType;
 int currentTime;
 int trainSampleNum;
 int testSampleNum;
 short[] attempts;
 double[] quality, efficiency;
 String letterFeatureName;
 String learnTypeFeatureName;
 String stageFeatureName;
 String positionFeatureName;

 public AgentAGL(){}

 public AgentAGL( Global g, short[][][] trainSet, short[][][] testSet )
 {
    super( g );
    this.trainSet = trainSet;
    this.testSet = testSet;
    trainSampleNum = trainSet.length;
    testSampleNum = testSet.length;
    reinforcements = new double[ 1 ];
    attempts = new short[ SESSION_NUM ];
    quality = new double[ SESSION_NUM ];
    efficiency = new double[ SESSION_NUM ];
    letterFeatureName = global.getFeatureNames( 0, 0 );
    learnTypeFeatureName =  global.getFeatureNames( 0, 1 );
    stageFeatureName =  global.getFeatureNames( 0, 2 );
    positionFeatureName =  global.getFeatureNames( 0, 5 );
 }

 public void reinit()
 {
    super.reinit();
    currentTime = 0;
    sessionIdx = stageIdx = 0;
    for( int i = 0; i < SESSION_NUM; i++ ){
         attempts[i] = 0;
         quality[i] = 0.0;
         efficiency[i] = 0.0;
    }
 }

 public boolean isControlNacsAction( int netIdx, short[] chosenAction )
 {
    return true;
 }

 public void getAttempts( short[] arr )
 {
    for( int i = 0; i < arr.length; i++ )
         arr[ i ] = attempts[ i ];
 }

 public void getQuality( double[] arr )
 {
    for( int i = 0; i < arr.length; i++ )
         arr[ i ] = quality[i] / attempts[i];
 }

 public void getEfficiency( double[] arr )
 {
    for( int i = 0; i < arr.length; i++ )
         arr[ i ] = efficiency[i] / attempts[i];
 }

 public void getOutput( Object arr )
 {
 }

 public int getCurSensoryInputType()
 {
    return 0;
 }

 public String getCurSensoryInputLabel()
 {
    return "label";
 }

 public void setRetrievalResults( int netIdx, GKSChunk[] results )
 {
 }

 public void setGroupType( char[] groupType )
 {
    this.groupType = groupType;
 }

 public short[] decideIrlAction( int netIdx, int setIdx, int ruleNo, short[] input )
 {
    return (short[])null;
 }

 public short[] decideFrAction( int netIdx, int setIdx, short[] input )
 {
    return (short[])null;
 }

 public int determineAmnNet( int netIdx, short[] action )
 {
    activeAmnIdx = 0;
    return activeAmnIdx;
 }

 public void performAction( int netIdx, short[] action )
 {
    //if( !isControlNacsAction( netIdx, action ) )
    //    reinforcements[ 0 ] = 0.0;
    //else
     //   reinforcements[ 0 ] = 1.0;

    determineAmnNet( netIdx, action );
    NacsAGL nacs = (NacsAGL)(clarion.getNACS());
    nacs.setActiveAmn( activeAmnIdx );
    if( stageIdx == TRAIN ){
        nacs.setTrainSample( curNacsInput );
        nacs.training( learningType );
    }
    if( stageIdx == TEST ){
        decideLearningType();
        nacs.setTestSample( curNacsInput );
        nacs.testing( learningType );
        nacsResults = ((NacsAGL)(clarion.getNACS())).getCompletion();
    }
    reinforcements[ 0 ] = taskClarion.getReinforcement( netIdx, curSensoryInput, null,
                           null, action, null, null, null );
 }

 public void run()
 {
    setupBeginFeatures();
    currentTime = 1000;
    for( sessionIdx = 0; sessionIdx < SESSION_NUM; sessionIdx++ ){
         getLearningType( groupType[sessionIdx] );
         stageIdx = 0;
         training();
         stageIdx = 1;
         testing();
    }
 }

/*************************** PRIVATE METHODS *********************************/
 private void setupBeginFeatures()
 {
    NacsAGL nacs = (NacsAGL)(clarion.getNACS());
    Feature[][] features = new Feature[ 2 ][ 1 ];
    short[] values = new short[ 1 ];
    values[ 0 ] = 0;
    int LETTER_DIM_VAL_NUM = 7;
    features[0][0] = new Feature( letterFeatureName, (short)0, (short)0,
                        (short)LETTER_DIM_VAL_NUM, values, global );
    int POSITION_DIM_VAL_NUM = 12;
    features[1][0] = new Feature( positionFeatureName, (short)0, (short)5,
                        (short)POSITION_DIM_VAL_NUM, values, global );
    nacs.setBeginFeatures( features );
 }

 private void getLearningType( char ch )
 {
    learningType = global.EMPTY;
    switch( ch ){
     case 'C':
          learningType = CONTROL_TYPE;
          break;
     case 'I':
          learningType = IMPLICIT_TYPE;
          break;
     case 'E':
          learningType = EXPLICIT_TYPE;
          break;
     case 'S':
          learningType = SYNERGY_TYPE;
          break;
    }
 }

 private void training()
 {
    System.out.println( "in training" );
    Vector sands = new Vector( testSampleNum );
    for( int i = 0; i < trainSampleNum; i++ )
         sands.addElement( new Integer(i) );
    int TRAIN_ITER_NUMS[] = { 1, 6, 6, 20 };
    //int TRAIN_ITER_NUMS[] = { 1, 1, 1, 1 };
    for( int i = 0; i < TRAIN_ITER_NUMS[ learningType ]; i++ ){
         System.out.println( "current iter no = " + i );
         Vector s = (Vector)(sands.clone());
         for( int j = 0; j < trainSampleNum; j++ ){
              clarion.setCurrentTime( currentTime );
              int sampleIdx = ((Integer)(s.remove((int)(Math.random()*s.size())))).intValue();
              initAcsSensory( sampleIdx, trainSet );
              initNacsSensory( sampleIdx, trainSet );
              clarion.training();
              currentTime += clarion.getACS().getRT();
         }
    }
 }

 private void testing()
 {
    System.out.println( "in testing" );

    //((NacsAGL)clarion.getNACS()).showGKS();
    ((NacsAGL)clarion.getNACS()).resetTotSearchNum();

    Vector sands = new Vector( testSampleNum );
    for( int i = 0; i < testSampleNum; i++ )
         sands.addElement( new Integer(i) );
    int deltaTime = 0;
    int impNum = 0, expNum = 0;
    while( deltaTime < SESSION_TIME && !sands.isEmpty() ){
       clarion.setCurrentTime( currentTime );
       int sampleIdx = ((Integer)(sands.remove((int)(Math.random()*sands.size())))).intValue();
       initAcsSensory( sampleIdx, testSet );
       initNacsSensory( sampleIdx, testSet );
       clarion.testing();
       attempts[ sessionIdx ]++;
       checkCompletion();

       if( learningType == IMPLICIT_TYPE )
           impNum++;
       if( learningType == EXPLICIT_TYPE )
           expNum++;
       //System.out.println( "learning type = " + learningType );

       int RT = clarion.getRT();
       deltaTime += RT;
       currentTime += RT;
    }

    System.out.println( "imp num = " + impNum );
    System.out.println( "exp num = " + expNum );
 }

 private void initAcsSensory( int sampleIdx, short[][][] dataSet )
 {
    int DIM_NUM = global.getIdnInputDimNum( 0 );
    int DIM_VAL_NUM = global.getFeatureLength(0,0);
    short[] con = new short[ DIM_VAL_NUM ];
    for( int i = 0; i < con.length; i++ )
         con[ i ] = 0;
    curSensoryInput = new Feature[ DIM_NUM ];
    // letter features.
    for( int i = 0; i < DIM_NUM-2; i++ )
         if( i < dataSet[sampleIdx].length )
             curSensoryInput[ i ] = new Feature( letterFeatureName, (short)0,
                      (short)0, dataSet[sampleIdx][i], global );
         else
             curSensoryInput[ i ] = new Feature( letterFeatureName, (short)0,
                      (short)0, con, global );
    // learning type feature.
    short[] learnTypeValues = new short[global.getFeatureLength(0,1)];
    for( int i = 0; i < learnTypeValues.length; i++ )
         learnTypeValues[ i ] = 0;
    learnTypeValues[ learningType ] = 1;
    curSensoryInput[ DIM_NUM-2 ] = new Feature( learnTypeFeatureName, (short)0,
                    (short)1, learnTypeValues, global );
    // stage feature.
    short[] stageValues = new short[ global.getFeatureLength(0,2) ];
    for( int i = 0; i < stageValues.length; i++ )
         stageValues[ i ] = 0;
    stageValues[ stageIdx ] = 1;
    curSensoryInput[ DIM_NUM-1 ] = new Feature( stageFeatureName, (short)0,
                    (short)2, stageValues, global );
 }

 private void initNacsSensory( int sampleIdx, short[][][] dataSet )
 {
    int DIM_NUM = dataSet[sampleIdx].length;
    curNacsInput = new Feature[ DIM_NUM ];
    for( int i = 0; i < DIM_NUM; i++ )
         curNacsInput[ i ] = new Feature( letterFeatureName, (short)0, (short)0,
                      dataSet[sampleIdx][i], global );
 }

 private void checkCompletion()
 {
    char[] letters = { '$', 'C', 'P', 'S', 'T', 'V', 'X', '-' };
    char[][][] stateTrans = {
       {{'S','1'},{'C','2'}},  {{'C','3'}},            {{'V','1'},{'X','3'}},
       {{'T','4'},{'P','5'}},  {{'S','4'},{'X','6'}},  {{'T','5'},{'V','7'}},
       {{'S','8'},{'X','5'}},  {{'P','6'},{'V','8'}}
    };
    int length = curNacsInput.length;
    int validNum = 0;
    int stateIdx = 0;
    for( int i = 0; i < length; i++ ){
         if( nacsResults[i].getValueNum() == 0 || stateIdx == 8 )
             break;
         if( curNacsInput[i].getValueNum() > 0 && !curNacsInput[i].equals(nacsResults[i]) )
             break;
         char ch = letters[ nacsResults[i].getValues()[0] ];
         for( int j = 0; j < stateTrans[stateIdx].length; j++ ){
              if( ch == stateTrans[stateIdx][j][0] ){
                  validNum++;
                  stateIdx = Integer.parseInt(String.valueOf(stateTrans[stateIdx][j][1]));
                  if( stateIdx == 8 )
                      break;
              }
         }
    }
    //System.out.println( "rate = " + (double)validNum / length );
    if( validNum == length ){
        if( stateIdx == 8 )
            quality[ sessionIdx ]++;
    }
    if( (double)validNum / length >= 0.7 )
        efficiency[ sessionIdx ]++;
 }

 private void decideLearningType()
 {
    if( sessionIdx == 0 || groupType[0] == groupType[1] )
        return;
    if( Math.random() < LOCK_PROB )
        getLearningType( groupType[0] );
    else
        getLearningType( groupType[1] );
 }
}