package tasksimulators.newsimu.AGL;
import java.util.Vector;
import java.util.Stack;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;
import clarion.tools.*;

public class NacsAGL extends NACS{

  final int CONTROL_TYPE = 0, IMPLICIT_TYPE = 1;
  final int EXPLICIT_TYPE = 2, SYNERGY_TYPE = 3;
  final int COND_NUM = 1;
  final int ITER_NUMS[] = { 1, 1, 1, 1 };
  final double RULE_APP_PROB = 1.0;
  //final double RULE_APP_PROB = 0.9;
  int activeAmnIdx;
  int iterIdx;
  int trainIdx;
  int searchNum;
  int totSearchNum;
  Feature[] gksInput;
  Feature[] amnInput;
  Feature[] completion;
  Feature[] gksOutput;
  Feature[] amnOutput;
  Feature[] trainSample;
  Feature[] testSample;
  Feature[][] beginFeatures;

  public NacsAGL(){}

  public NacsAGL( Global g )
  {
     super( g );
  }

  public Feature[] getCompletion()
  {
     return completion;
  }

  public void setIterIdx( int idx )
  {
     iterIdx = idx;
  }

  public void setTrainIdx( int idx )
  {
     trainIdx = idx;
  }

  public void resetTotSearchNum()
  {
     totSearchNum = 0;
  }

  public void setTrainSample( Feature[] curNacsInput )
  {
     trainSample = curNacsInput;
  }

  public void setTestSample( Feature[] curNacsInput )
  {
     testSample = curNacsInput;
  }

  public void setBeginFeatures( Feature[][] features )
  {
     beginFeatures = new Feature[ features.length ][];
     for( int i = 0; i < features.length; i++ ){
       beginFeatures[i] = new Feature[ features[i].length ];
       for( int j = 0; j < features[i].length; j++ )
            beginFeatures[i][j] = (Feature)features[i][j].clone();
     }
  }

  public void training( int learningType )
  {
    //showFeatures( "train sample = ", trainSample );
    trainAmn( learningType );
    if( learningType == EXPLICIT_TYPE || learningType == SYNERGY_TYPE )
        trainGks();
  }

  public void testing( int learningType )
  {
    //showFeatures( "test sample = ", testSample );

    searchNum = 0;
    gks.resetRT();
    amns[ activeAmnIdx ].resetRT();
    if( learningType == CONTROL_TYPE || learningType == IMPLICIT_TYPE ){
        amnInput = testSample;
        testAmn();
        completion = amnOutput;
    }
    if( learningType == EXPLICIT_TYPE ){
        gksInput = testSample;
        if( testGks() )
            completion = gksOutput;
        else{
            amnInput = gksOutput;
            testAmn();
            completion = amnOutput;
        }
    }
    if( learningType == SYNERGY_TYPE ){
        amnInput = testSample;
        while( true ) {
          testAmn();
          gksInput = amnOutput;
          if( checkByGks() ){
              completion = amnOutput;
              break;
          }
          else
              amnInput = gksOutput;
        }
    }

    //System.out.println( "Search Num = " + searchNum );
    //totSearchNum += searchNum;
    //System.out.println( "Total search Num = " + totSearchNum );

  }

  public void showGKS()
  {
     System.out.println( gks );
  }

/****************************** PRIVATE METHODS *******************************/
  private void trainGks()
  {
     //System.out.println( "in training GKS" );
     short[] conditions = new short[ 2 ];
     int beginLetterChunkIdx = gks.getChunkIdx( "begin letter" );
     if( beginLetterChunkIdx == global.EMPTY ){
         System.out.println( "create begin letter chunk in AGLnacs.java" );
         setupBeginLetterChunk();
         beginLetterChunkIdx = gks.getChunkIdx( "begin letter" );
     }
     int beginPositionChunkIdx = gks.getChunkIdx( "begin position" );
     if( beginPositionChunkIdx == global.EMPTY ){
         System.out.println( "create begin position chunk in AGLnacs.java" );
         setupBeginPositionChunk();
         beginPositionChunkIdx = gks.getChunkIdx( "begin position" );
     }
     conditions[ 0 ] = (short)beginLetterChunkIdx;
     conditions[ 1 ] = (short)beginPositionChunkIdx;
     int length = trainSample.length;
     for( int i = 0; i < length; i++ ){
          Feature[] features = new Feature[ 1 ];
          features[ 0 ] = trainSample[ i ];
          short conclusion = (short)gks.getChunkIdx( features );
          if( conclusion == global.EMPTY )
              conclusion = (short)gks.insertChunk( 0, 0, 0, "letterChunk", features );
          gks.insertRule( conditions, conclusion );
          // update the conditions.
          conditions[ 0 ] = conclusion;
          short[] values = new short[ 1 ];
          values[ 0 ] = (short)(i + 1);
          features[ 0 ] = new Feature(global.getFeatureNames(0,5),(short)0,(short)5,
                              (short)global.getFeatureLength(0,5),values,global);
          conditions[ 1 ] = (short)gks.getChunkIdx( features );
          if( conditions[1] == global.EMPTY )
              conditions[1] = (short)gks.insertChunk( 0, 0, 0, "positionChunk", features );
     }
  }

  private void trainAmn( int learningType )
  {
     //System.out.println( "in training AMN" );
     int inputLength = trainSample.length;
     int iterNum = ITER_NUMS[ learningType ];
     // generate samples.
     Vector sands = new Vector( inputLength );
     for( int i = 0; i < inputLength; i++ )
          sands.add( new Integer( i ) );
     int[] sampleNums = new int[ inputLength - 1 ];
     int sampleNum = calcSampleNums( inputLength, sampleNums );
     Vector[] samples = new Vector[ sampleNum ];
     int offset = 0;
     for( int i = 0; i < inputLength-1; i++ ){
        for( int j = 0; j < sampleNums[i]; j++ ){
             int size = i + 2;
             samples[ offset + j ] = new Vector( size );
             Vector s = (Vector)(sands.clone());
             for( int k = 0; k < size; k++ )
                  samples[ offset + j ].addElement(s.remove((int)(Math.random()*s.size())));
        }
        offset += sampleNums[i];
     }
     // train AMN.
     amns[activeAmnIdx].setDesiredOutput( trainSample );
     for( int i = 0; i < iterNum; i++ ){
          sands = new Vector( sampleNum );
          for( int j = 0; j < sampleNum; j++ )
               sands.add( new Integer( j ) );
          for( int j = 0; j < sampleNum; j++ ){
               int sampleIdx = ((Integer)(sands.remove((int)(Math.random()*sands.size())))).intValue();
               int size = samples[sampleIdx].size();
               Feature[] inputs = new Feature[ inputLength ];
               for( int k = 0; k < size; k++ ){
                    int dimIdx = ((Integer)(samples[sampleIdx].elementAt( k ))).intValue();
                    inputs[dimIdx] = (Feature)(trainSample[dimIdx].clone());
               }
               amns[activeAmnIdx].setInput( inputs );
               amns[activeAmnIdx].training();
          }
     }
  }

  // reasoning by using DFS search.
  private boolean testGks()
  {
     boolean completeFlag = false;
     int length = gksInput.length;
     short[] conditions = new short[ 2 ];
     conditions[ 0 ] = (short)gks.getChunkIdx( "begin letter" );
     conditions[ 1 ] = (short)gks.getChunkIdx( "begin position" );
     short[] maxPartial = new short[ length + COND_NUM ];
     short[] dynPartial = new short[ length + COND_NUM ];
     for( int i = 0; i < length + COND_NUM; i++ ){
          maxPartial[ i ] = (short)global.EMPTY;
          dynPartial[ i ] = (short)global.EMPTY;
     }
     maxPartial[ 0 ] = conditions[ 0 ];
     dynPartial[ 0 ] = conditions[ 0 ];
     // get the filled-in positions.
     short[] filledLocs = new short[ 2 ];
     int count = 0;
     for( int i = 0; i < length; i++ )
          if( gksInput[i].getValueNum() == 1 )
              filledLocs[count++] = (short)i;
     // construct stack for Depth First Search.
     Stack stack = new Stack();
     Vector beginElem = new Vector( COND_NUM );
     beginElem.addElement( new Integer( conditions[0] ) );
     stack.push( beginElem );
     Vector conc = gks.searchConclusions( conditions );
     int idx = 0;
     while( idx < conc.size() ){
          if( Math.random() > RULE_APP_PROB )
              conc.removeElementAt( idx );
          else
              idx++;
     }
     stack.push( conc );
     searchNum++;
     int curPos = 1;
     // init current position variable
     short[] posValues = new short[ 1 ];
     posValues[0] = (short)curPos;
     int exitCode = 0;
     while( exitCode == 0 ){
        if( curPos == 0 ){
            exitCode = -1;
            continue;
        }
        Vector top = (Vector)stack.peek();
        int curChunkIdx;
        Feature feature;
        if( top.isEmpty() ){ // this is dead path, needs replace the last letter if exists.
            stack.pop();
            curPos--;
            posValues[0] = (short)curPos;
            ((Vector)stack.peek()).removeElementAt( 0 );
            continue;
        }
        if( curPos == filledLocs[1] + 1 ){ // reach the second filled letter( the end )
            curChunkIdx = ((Integer)top.elementAt(0)).intValue();
            feature = gks.getChunk(curChunkIdx).getFeature(0);
            if( feature.equals( gksInput[ filledLocs[1] ] ) ){
                dynPartial[ curPos ] = (short)curChunkIdx;
                for( int i = 0; i < curPos + 1; i++ )
                     maxPartial[i] = dynPartial[i];
                exitCode = 1;
                continue;
            }
            else
                top.removeElementAt(0);
        }
        else{
            if( curPos == filledLocs[0] + 1 ){ // reach the first filled letter.
                idx = 0;
                while( idx < top.size() ){
                   curChunkIdx = ((Integer)top.elementAt(idx)).intValue();
                   feature = gks.getChunk(curChunkIdx).getFeature(0);
                   // if mismatched, remove it.
                   if( !feature.equals( gksInput[ filledLocs[0] ] ) )
                        top.removeElementAt( idx );
                   else{ // if matched, record maximum partial string currently available.
                        dynPartial[ curPos ] = (short)curChunkIdx;
                        for( int i = 0; i < curPos + 1; i++ )
                             maxPartial[i] = dynPartial[i];
                        idx++;
                   }
                }
                if( top.isEmpty() ) // no match exists.
                    continue;
            }
            // fill in the conditions.
            curChunkIdx = ((Integer)top.elementAt(0)).intValue();
            dynPartial[ curPos ] = (short)curChunkIdx;
            conditions[ 0 ] = dynPartial[ curPos ];
            Feature[] features = new Feature[ 1 ];
            features[0] = new Feature(global.getFeatureNames(0,5),(short)0,(short)5,
                              (short)global.getFeatureLength(0,5),posValues,global);
            conditions[ 1 ] = (short)gks.getChunkIdx( features );
            if( conditions[1] == global.EMPTY ){
                System.out.println( "conditions error in AGLnacs.java" );
                exitCode = -1;
                continue;
            }
            conc = gks.searchConclusions( conditions );
            idx = 0;
            while( idx < conc.size() ){
                   if( Math.random() > RULE_APP_PROB )
                       conc.removeElementAt( idx );
                   else
                       idx++;
            }
            stack.push( conc );
            searchNum++;
            curPos++;
            posValues[0] = (short)curPos;

            /*
            String str = "condition[0] = " + gks.getChunk(conditions[0]).getFeature(0);
            str += ", condition[1] = " + gks.getChunk(conditions[1]).getFeature(0);
            str +=", conclusion = ";
            for( int i = 0; i < conc.size(); i++ )
                 str += gks.getChunk(((Integer)conc.elementAt(i)).intValue()).getFeature(0) + ", ";
            System.out.println( str );
            */

        }
     }
     // fill in the filled-in positions of the testing sample.
     if( exitCode == -1 ){ // fails.
         for( int i = 0; i < 2; i++ )
            if( maxPartial[filledLocs[i] + COND_NUM] == global.EMPTY ){
                Feature[] features = new Feature[ 1 ];
                features[0] = gksInput[filledLocs[i]];
                maxPartial[filledLocs[i] + COND_NUM] = (short)gks.getChunkIdx( features );
            }
     }
     // get output from GKS.
     gksOutput = new Feature[ length ];
     for( int i = 0; i < length; i++ ){
          if( maxPartial[i + COND_NUM] != global.EMPTY ){
              gksOutput[i] = gks.getChunk( maxPartial[i + COND_NUM] ).getFeature( 0 );
          }
          else
              gksOutput[i] = (Feature)null;
     }

     //showFeatures( "gks output = ", gksOutput );

     if( exitCode == 1 )
         completeFlag = true;
     return completeFlag;
  }

  private void testAmn()
  {
     amns[activeAmnIdx].setInput( amnInput );
     amns[activeAmnIdx].retrieval( false );
     amnOutput = amns[activeAmnIdx].getResults();
  }

  private boolean checkByGks()
  {
     boolean completeFlag = false;
     int length = testSample.length;
     short[] conditions = new short[ 2 ];
     conditions[ 0 ] = (short)gks.getChunkIdx( "begin letter" );
     conditions[ 1 ] = (short)gks.getChunkIdx( "begin position" );
     short[] dynPartial = new short[ length + COND_NUM ];
     for( int i = 0; i < length + COND_NUM; i++ )
          dynPartial[ i ] = (short)global.EMPTY;
     dynPartial[ 0 ] = conditions[ 0 ];
     // init current position
     short[] posValues = new short[ 1 ];
     posValues[ 0 ] = 1;
     int idx = 0;    // indicates current position needs validation.
     while( idx < length ){
          Vector conclusions = gks.searchConclusions( conditions );
          searchNum++;
          Feature[] features = new Feature[ 1 ];
          features[ 0 ] = gksInput[ idx ];
          int curChunkIdx = gks.getChunkIdx( features );
          boolean inFlag = false;
          if( !conclusions.isEmpty() ){
              for( int i = 0; i < conclusions.size(); i++ )
                   if( ((Integer)conclusions.elementAt(i)).intValue() == curChunkIdx ){
                       inFlag = true;
                       break;
                   }
          }
          if( inFlag ){
              dynPartial[ idx + COND_NUM ] = (short)curChunkIdx;
              idx++;
              if( idx < length ){
                  posValues[0] = (short)(idx + 1);
                  // fill in conditions.
                  conditions[ 0 ] = (short)curChunkIdx;
                  features = new Feature[ 1 ];
                  features[0] = new Feature(global.getFeatureNames(0,5),(short)0,(short)5,
                                    (short)global.getFeatureLength(0,5),posValues,global);
                  conditions[ 1 ] = (short)gks.getChunkIdx( features );
                  if( conditions[1] == global.EMPTY ){
                      System.out.println("conditions error in AGLnacs.java" );
                      System.exit( 1 );
                  }
               }
               continue;
          }
          // no match.
          // conclusions is empty
          idx += COND_NUM + 1;
          posValues[0] = (short)(idx + 1);
          if( idx < length ){
              if( gksInput[idx-1].getValueNum() > 0 ){
                  for( int i = 0; i < COND_NUM; i++ ){
                       features[ 0 ] = gksInput[ idx - COND_NUM + i ];
                       conditions[ i ] = (short)gks.getChunkIdx( features );
                       if( conditions[ i ] == global.EMPTY ){
                           System.out.println( "chunk null error in condition" );
                           System.out.println( "feature = " + features[0] );
                           System.exit( 1 );
                       }
                  }
              }
              else{
                  System.out.println( "feature null error in AGLnacs" );
                  break;
              }
          }
     } // end of while( idx < length )
     // fill in GKS output.
     int failNum = 0;
     gksOutput = new Feature[ length ];
     for( int i = 0; i < length; i++ ){
          if( dynPartial[i + COND_NUM] != global.EMPTY ){
              gksOutput[i] = gks.getChunk( dynPartial[i + COND_NUM] ).getFeature( 0 );
          }
          else{
              gksOutput[i] = (Feature)null;
              failNum++;
          }
     }

     //showFeatures( "gks output = ", gksOutput );

     //if( Math.random() >= (double)failNum/length )
     completeFlag = true;
     return completeFlag;
  }

  public void setupBeginLetterChunk()
  {
     gks.insertChunk( 0, 0, 0, "begin letter", beginFeatures[0] );
  }

  public void setupBeginPositionChunk()
  {
     gks.insertChunk( 0, 0, 0, "begin position", beginFeatures[1] );
  }

  private int calcSampleNums( int length, int[] sampleNums )
  {
     int sampleNum = 0;
     for( int i = 2; i <= length; i++ ){
          long num = 1;
          for( int j = length; j > length-i; j-- )
               num *= j;
          for( int j = 1; j < i+1; j++ )
               num /= j;
          sampleNums[ i-2 ] = 1 + (int)num/2;
          sampleNum += 1 + (int)num/2;
     }
     return sampleNum;
  }

  private void showFeatures( String head, Feature[] features )
  {
     String str = head;
     for( int i = 0; i < features.length; i++ )
          str += features[i];
     System.out.println( str );
  }
}