package tasksimulators.newsimu.MCM;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;

public class AgentMCM extends TaskAgent{

  static final int BL = Global.BL, TL = Global.TL, COMB = 2;
  static final int BLOCK_NUM = 20, TRIALS_PER_BLOCK = 2;
  static final double FULL_SCORE = 7;
  int INPUT_DIM_NUM, OUTPUT_DIM_NUM, INPUT_NUM, OUTPUT_NUM;
  int blockIdx, trialIdx;
  int acsNetNum;
  boolean[] correctionFlags;
  double warmth, warmths[];
  String featureNames[][];
  GKSChunk retrievalResults[];
  GKSChunk rightChunk;

  public AgentMCM(){}

  public AgentMCM( Global g )
  {
     super( g );
     INPUT_DIM_NUM = global.getIdnInputDimNum( 0 );
     OUTPUT_DIM_NUM = global.getIdnOutputDimNum( 0 );
     INPUT_NUM = global.getIdnInputNum( 0 );
     OUTPUT_NUM = global.getIdnOutputNum( 0 );
     featureNames = global.getFeatureNames();
     reinforcements = new double[ 1 ];
     curSensoryInput = new Feature[ INPUT_DIM_NUM ];
     correctionFlags = new boolean[ BLOCK_NUM ];
     warmths = new double[ BLOCK_NUM ];
     initSensory();
  }

  public void reinit()
  {
     super.reinit();
     for( int i = 0; i < BLOCK_NUM; i++ )
          warmths[ i ] = 0;
  }

  public boolean isControlNacsAction( int netIdx, short[] chosenAction )
  {
     return false;
  }

  public boolean[] isCorrect()
  {
     return correctionFlags;
  }

  public void getOutput( Object arr )
  {
     for( int i = 0; i < BLOCK_NUM; i++ )
          ((double[])arr)[ i ] = warmths[ i ];
  }

  public void setRetrievalResults( int netIdx, GKSChunk[] results  )
  {
  }

  public int determineAmnNet( int netIdx, short[] action )
  {
      activeAmnIdx = 0;
      return activeAmnIdx;
  }

  public void run()
  {
      // get the right solution.
      rightChunk = ((TaskClarionMCM)taskClarion).getCorrectConclusion();
      for( blockIdx = 0; blockIdx < BLOCK_NUM; blockIdx++ ) {
           if( blockIdx % 50 == 0 )
               System.out.println( "block idx = " + blockIdx );
          do{
                initSensory();
                //clarion.testing();
                clarion.training();
                if( clarion.getRetrievalResults() != null )
                    retrievalResults = clarion.getRetrievalResults();
                if( retrievalResults != null ){
                    correctionFlags[blockIdx] = rightChunk.equals( retrievalResults[0] );
                }

                //System.out.println( "GS active goal = " + global.getGS().getActiveGoal() );
          }
          while( !global.getGS().isEmpty() );
          warmths[blockIdx] = warmth;
      }
  }

  public void performAction( int netIdx, short[] action )
  {
     if( action[Global.D_TYPE] == 0 || action[Global.D_WM] == 0 ) // do-nothing action or no WM slot.
         return;
     if( action[Global.D_ACT] == 1 && action[Global.D_INFER] == 1 ){ // inference
         global.getNACS().setRetrievalMode( Global.RETRIEVAL_ONE );
         global.getNACS().setGksCue( curNacsInput );
         global.getNACS().setAmnCue( 0, curNacsInput );
         global.getNACS().run( 0 );
          // store the NACS retrieval results into WM.
         global.getACS().setRetrievalResult( global.getNACS().getResults(), action[Global.D_WM]-1 );
     }
     if( action[Global.D_ACT] == 0 && action[Global.D_REPORT] == 1 ){ // reporting.
         if( retrievalResults == null || retrievalResults[0] == null ){
             System.out.println( "run in disorder !" );
             return;
         }

         double[] relStrengths = global.getMCS().getMonitorBuffer().getNacsRelStrengths( COMB );
         Vector monitorChunks = global.getMCS().getMonitorBuffer().getNacsChunks( COMB );

         /*
         System.out.println( "relative strengths = " );
         for( int i = 0; i < relStrengths.length; i++ )
              System.out.println( relStrengths[i] );
         */

         //System.out.println( "monitored chunks = " );
         //for( int i = 0; i < monitorChunks.size(); i++ )
         //     System.out.println( monitorChunks.elementAt(i) );

         //System.out.println( "final conclusion = " );
         //System.out.println( retrievalResults[0] );

         for( int i = 0; i < monitorChunks.size(); i++ ){
              if( monitorChunks.elementAt(i).equals( retrievalResults[0] ) ){
                  warmth = relStrengths[i] * FULL_SCORE;

                  //System.out.println( "warmth = " + warmth );

                  break;
              }
         }
     }
  }

/**************************** PRIVATE METHODS *********************************/
  private void initSensory()
  {
      curSensoryInput = new Feature[INPUT_DIM_NUM];
      for( int i = 0; i < INPUT_DIM_NUM; i++ ){
           short featType = global.getIdnInputFeatureInfo(0)[i][0];
           short featOrder = global.getIdnInputFeatureInfo(0)[i][1];
           short featLen = (short)global.getFeatureLength( featType, featOrder );
           String featLabel = global.getFeatureNames( featType, featOrder );
           short[] featValues = new short[1];
           featValues[0] = (short)0;
           curSensoryInput[i] = new Feature( featLabel, featType, featOrder, featLen, featValues, global );
      }
      curNacsInput = curSensoryInput;
  }
}
