package tasksimulators.newsimu.MCM;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;
import clarion.acs.gs.*;
import clarion.acs.wm.*;
import clarion.acs.net.auxiliary.*;

public class TaskClarionMCM extends TaskClarion {

  final int INFER = 0, REPORT = 1, CLEAR = 2;
  final int INPUT_DIM_NUM = 4, OUTPUT_DIM_NUM = 1;
  // dims for ACS input.
  final int D_TYPE = 0, D_ACT = 1, D_LEARNING = 2, D_INFER = 3, D_REPORT = 4;
  final int D_CUE = 5, D_OUTPUT = 6, D_TIME = 7, D_WM = 8;
  // dims for NACS input.
  final int D_DATE = 0, D_COIN = 1, D_INFO = 2, D_CALL = 3;
  /*
   - input dimensions
     dimension 0 : the date of coin.      # of values = 1.
                   B.C.
     dimension 1 : the type of coin.      # of values = 2.
                   bronze, gold
     dimension 2 : info about coin.       # of vlaues = 2.
                   much, little
     dimension 3 : the type of action.    # of values = 2.
                   call police or not
   - output dimensions of NACS.
     dimension 0 : the solution.          # of values = 4.
                   the date is wrong,
                   bronze is not ivented in 544 B.C.
                   the curator knew the man is a thief.
                   some coins were stolen from the museum a few days ago.
   - condition chunks. # of chunks = 8;
   - conlusion chunks. # of chunks = 4;
  */
  final int GKS_CHUNK_NUM = 12, COND_CHUNK_NUM = 8, CONC_CHUNK_NUM = 4;
  final int RIGHT_COND_CHUNK_IDX = 0;
  final int RIGHT_CONC_CHUNK_IDX = COND_CHUNK_NUM;
  // fixed rules
  short[] frConditions[][][];
  short[] frActions[][];
  GKSChunk[] gksChunks;

  public TaskClarionMCM()
  {
  }

  // eligibility of ACS networks, return the eligible condition of the network.
  public short[][][] getEligibility( int netIdx )
  {
    short[][][] eligCondition = {
                   {
                      {0},            // relevant dims: 0 in this condition.
                      {1,1,1,1,1},    // all values are eligible.
                   },
                };
     return eligCondition;
  }

  // Fixed Rule conditions
  public short[][] getFrCondition( int groupIdx, int netIdx, int ruleIdx )
  {
     return (short[][])null;
  }

  // Fixed Rule actions
  public short[] getFrAction( int groupIdx, int netIdx, int ruleIdx )
  {
     return (short[])null;
  }

  // reinforcement of ACS networks.
  public double getReinforcement( int netIdx, Feature[] input, Chunk[] wmItems, GoalChunk goalItem,
                short[] action, Feature[] newInput, Chunk[] newWmItems, GoalChunk newGoalItem )
  {
     short[] gsState = ( goalItem == null ) ? null : goalItem.getFeature( 0 ).getContent();
     boolean gsEmpty = ( goalItem == null ) ? true : false;
     boolean wmEmpty = ( wmItems[0] == null ) ? true : false;
     double reward = 0.0;
     if( global.isExNet( netIdx ) ){  // external net.
         if( !gsEmpty && gsState[INFER] == 1 && wmEmpty ){
             reward += (action[D_TYPE]==1) ? 0.1:0.0;     // do sth.
             reward += (action[D_ACT]==1) ? 0.1:0.0;      // activate NACS
             reward += (action[D_LEARNING]==0) ? 0.1:0.0; // no learning
             reward += (action[D_INFER]==1) ? 0.1:0.0;    // inference
             reward += (action[D_REPORT]==0) ? 0.1:0.0;   // no reporting
             reward += (action[D_CUE]==0) ? 0.1:0.0;      // no use
             reward += (action[D_OUTPUT]==0) ? 0.1:0.0;   // no use
             reward += (action[D_TIME]==0) ? 0.1:0.0;     // no time limit
             reward += (action[D_WM]==1) ? 0.1:0.0;       // slot 0 for storing
         }
         else if( !gsEmpty && gsState[REPORT] == 1 && !wmEmpty ){
                  reward += (action[D_TYPE]==1) ? 0.1:0.0;     // do sth.
                  reward += (action[D_ACT]==0) ? 0.1:0.0;      // not activate NACS
                  reward += (action[D_LEARNING]==0) ? 0.1:0.0; // no learning
                  reward += (action[D_INFER]==0) ? 0.1:0.0;    // inference
                  reward += (action[D_REPORT]==1) ? 0.1:0.0;   // no reporting
                  reward += (action[D_CUE]==0) ? 0.1:0.0;      // no use
                  reward += (action[D_OUTPUT]==0) ? 0.1:0.0;   // no use
                  reward += (action[D_TIME]==0) ? 0.1:0.0;     // no time limit
                  reward += (action[D_WM]==1) ? 0.1:0.0;       // slot 0 for storing
               }
               else{
                   reward += (action[D_TYPE]==0) ? 0.1:0.0;     // do sth.
                   reward += (action[D_ACT]==0) ? 0.1:0.0;      // activate NACS
                   reward += (action[D_LEARNING]==0) ? 0.1:0.0; // no learning
                   reward += (action[D_INFER]==0) ? 0.1:0.0;    // inference
                   reward += (action[D_REPORT]==0) ? 0.1:0.0;   // no reporting
                   reward += (action[D_CUE]==0) ? 0.1:0.0;      // no use
                   reward += (action[D_OUTPUT]==0) ? 0.1:0.0;   // no use
                   reward += (action[D_TIME]==0) ? 0.1:0.0;     // no time limit
                   reward += (action[D_WM]==0) ? 0.1:0.0;       // slot 0 for storing
               }
     }
     if( global.isWmNet( netIdx ) ){ // working memory net.
         if( !gsEmpty && gsState[CLEAR] == 1 && !wmEmpty )
             reward += (action[0]==2) ? 1.0:0.0;  // remove the WM item.
         else
             reward += (action[0]==global.getWM().getDoNothingAction()) ? 1.0:0.0;  // do-nothing action.
     }
     if( global.isGsNet( netIdx ) ){  // goal structure net.
         if( gsEmpty ){
             if( wmEmpty )
                 reward += (action[0]==2) ? 1.0:0.0;  // set goal = clear-up.
             else
                 reward += (action[0]==global.getGS().getDoNothingAction())?1.0:0.0; // do nothing action.
         }
         else if( gsState[CLEAR] == 1 && wmEmpty )
                  reward += (action[0]==1) ? 1.0:0.0;  // set goal = reporting.
              else if( gsState[REPORT] == 1 && wmEmpty )
                       reward += (action[0]==0) ? 1.0:0.0;  // set goal = inference.
                   else if( gsState[INFER] == 1 && !wmEmpty )
                            reward += (action[0]==3) ? 1.0:0.0;  // pop or remove goal = inference.
                        else if( gsState[REPORT] == 1 && !wmEmpty )
                                 reward += (action[0]==4) ? 1.0:0.0;  // pop or remove goal = reporting.
                             else if( gsState[CLEAR] == 1 && !wmEmpty )
                                      reward += (action[0]==5) ? 1.0:0.0; // pop or remove goal = clear.
                                  else // do nothing action.
                                      reward += (action[0]==global.getGS().getDoNothingAction())?1.0:0.0;
     }
     return reward;
  }

  public double decideFrAction( int groupIdx, int netIdx, int setIdx, Feature[] sensoryInput,
                                GoalChunk gsItem, GKSChunk[] wmItems, short[] suggestedAction )
  {
     short[] gsState = ( gsItem == null ) ? null : gsItem.getFeature( 0 ).getContent();
     boolean gsEmpty = ( gsItem == null ) ? true : false;
     boolean wmEmpty = ( wmItems[0] == null ) ? true : false;

     //System.out.println( "net idx " + netIdx );
     //System.out.println( "set idx " + setIdx );
     //System.out.println( "GS empty = " + gsEmpty );
     //System.out.println( "WM empty = " + wmEmpty );
     //if( !gsEmpty )
     //     System.out.println( "gs feat = " + gsItem.getFeature( 0 ) );


     double support = 0.0;
     // differen FR rules.
     if( global.isExNet( netIdx ) ){  // external FRs.
         for( int i = 0; i < suggestedAction.length; i++ )
              suggestedAction[i] = 0;
         if( gsEmpty )
             return support;
         switch( setIdx ){
            case 0:
                 if( gsState[INFER] == 1 && wmEmpty ){
                     suggestedAction[D_TYPE] = 1;        // do sth.
                     suggestedAction[D_ACT] = 1;         // activate NACS
                     suggestedAction[D_LEARNING] = 0;    // no learning
                     suggestedAction[D_INFER] = 1;       // inference
                     suggestedAction[D_REPORT] = 0;      // no reporting
                     suggestedAction[D_CUE] = 0;         // no use
                     suggestedAction[D_OUTPUT] = 0;      // no use
                     suggestedAction[D_TIME] = 0;        // no time limit
                     suggestedAction[D_WM] = 1;          // slot 0 for storing
                     support = 1.0;
                 }
                 break;
            case 1:
                 if( gsState[REPORT] == 1 && !wmEmpty ){
                     suggestedAction[D_TYPE] = 1;        // do sth.
                     suggestedAction[D_ACT] = 0;         // not activate NACS
                     suggestedAction[D_LEARNING] = 0;    // no learning
                     suggestedAction[D_INFER] = 0;       // no inference
                     suggestedAction[D_REPORT] = 1;      // reporting
                     suggestedAction[D_CUE] = 0;         // no use
                     suggestedAction[D_OUTPUT] = 0;      // no use
                     suggestedAction[D_TIME] = 0;        // no time limit
                     suggestedAction[D_WM] = 1;          // slot 0 for reporting
                     support = 1.0;
                 }
                 break;
         }
     }
     if( global.isWmNet( netIdx ) ){ // working memory FRs.
         suggestedAction[0] = (short)global.getWM().getDoNothingAction(); // do-nothing action.
         switch( setIdx ){
           case 0:
                if( !gsEmpty && gsState[CLEAR] == 1 && !wmEmpty ){
                    suggestedAction[0] = 2;  // remove the WM item.
                    support = 1.0;
                }
                break;
         }

         //System.out.println( "WM suggest action = " + suggestedAction[0] );
     }
     if( global.isGsNet( netIdx ) ){  // GS FRs.
         suggestedAction[0] = (short)global.getGS().getDoNothingAction();  // do-nothing aciton.
         switch( setIdx ){
           case 0:
                if( gsEmpty && wmEmpty ){
                    suggestedAction[0] = 2; // set goal = clear-up.
                    support = 1.0;
                }
                break;
           case 1:
                if( !gsEmpty && gsState[CLEAR] == 1 && wmEmpty ){
                    suggestedAction[0] = 1; // set goal = reporting.
                    support = 1.0;
                }
                break;
           case 2:
                if( !gsEmpty && gsState[REPORT] == 1 && wmEmpty ){
                    suggestedAction[0] = 0;  // set goal = regular-inference.
                    support = 1.0;
                }
                break;
           case 3:
                if( !gsEmpty && gsState[INFER] == 1 && !wmEmpty ){
                    suggestedAction[0] = 3;  // pop or remove goal = regular-inference.
                    support = 1.0;
                }
                break;
           case 4:
                if( !gsEmpty && gsState[REPORT] == 1 && !wmEmpty ){
                    suggestedAction[0] = 4;  // pop or remove goal = reporting.
                    support = 1.0;
                }
                break;
           case 5:
                if( !gsEmpty && gsState[CLEAR] == 1 && !wmEmpty ){
                    suggestedAction[0] = 5;  // pop or remove goal = clear.
                    support = 1.0;
                }
                break;
         }

         //System.out.println( "GS suggest action = " + suggestedAction[0] );
     }

     //System.out.println( "support = " + support );
     return support;
  }

  // get externally encoded NACS chunks.
  public GKSChunk[] getExternalNacsChunks()
  {
     setupGKS();
     return gksChunks;
  }

  public GKSChunk getCorrectConclusion()
  {
     return gksChunks[RIGHT_CONC_CHUNK_IDX];
  }

  public double[] getAssocDimWeights( int dimNum )
  {
     //double[] weights = { 0.6, 0.15, 0.15, 0.1 };
     double[] weights = { 0.25, 0.25, 0.25, 0.25 };
     return weights;
  }

/*********************************Private Methods *****************************/
  private void setupGKS()
  {
     String[][] chunkNames = {
          { "condition 0", "condition 1", "condition 2", "condition 3",
            "condition 4", "condition 5", "condition 8", "condition 9"
          },
          { "conclusion 0", "conclusion 1", "conclusion 2", "conclusion 3", "conclusion 4" }
     };
     short[][][] chunkDimVals = {
               { {0,0,0,0},{0,0,0,1},{0,0,1,0},{0,0,1,1},{0,1,0,0},{0,1,0,1},{0,1,1,0},{0,1,1,1} },
               { {0}, {1}, {2}, {3} }
           };
     gksChunks = new GKSChunk[GKS_CHUNK_NUM];
     // set up condition chunks.
     for( int i = 0; i < COND_CHUNK_NUM; i++ ){
          Feature[] features = new Feature[INPUT_DIM_NUM];
          for( int j = 0; j < INPUT_DIM_NUM; j++ ){
               short featType = global.getGksInputFeatureInfo()[j][0];
               short featOrder = global.getGksInputFeatureInfo()[j][1];
               short featLen = (short)global.getFeatureLength( featType, featOrder );
               String featLabel = global.getFeatureNames( featType, featOrder );
               short[] featValues = new short[1];
               featValues[0] = chunkDimVals[0][i][j];
               features[j] = new Feature( featLabel, featType, featOrder, featLen, featValues, global );
          }
          gksChunks[i] = new GKSChunk( 0, Global.CHK_ASSOC, chunkNames[0][0], features, 0, 0, global );
     }
     // set up conclusion chunks.
     short featType = global.getGksOutputFeatureInfo()[0][0];
     short featOrder = global.getGksOutputFeatureInfo()[0][1];
     short featLen = (short)global.getFeatureLength( featType, featOrder );
     String featLabel = global.getFeatureNames( featType, featOrder );
     for( int i = 0; i < CONC_CHUNK_NUM; i++ ){
          Feature[] features = new Feature[OUTPUT_DIM_NUM];
          short[] featValues = new short[1];
          featValues[0] = chunkDimVals[1][i][0];
          features[0] = new Feature( featLabel, featType, featOrder, featLen, featValues, global );
          gksChunks[RIGHT_CONC_CHUNK_IDX+i] = new GKSChunk( i, Global.CHK_ASSOC, chunkNames[1][i],
                                                            features, 0, 0, global );
     }
     // set up rules.
     double[] distriProbs = { 0.34, 0.16, 0.16, 0.34 };
     double[] rightRuleProbs = { 0.2, 0.2, 0.8, 0.8 };
     //double[] rightRuleProbs = { 0.5, 0.5, 0.5, 0.5 };
     double[] probSums = { 0.34, 0.5, 0.66, 1.0 };
     int concNum = 1;
     double prob = Math.random();
     for( int i = 0; i < probSums.length; i++ )
          if( prob < probSums[i] ){
              concNum = i + 1;
              break;
          }

     //System.out.println( "conc num = " + concNum );

     AssocRuleGrp[] assocRuleGroups = new AssocRuleGrp[ GKS_CHUNK_NUM ];
     Vector condSands = new Vector( COND_CHUNK_NUM );
     for( int i = 0; i < COND_CHUNK_NUM; i++ )
          condSands.addElement( new Integer(i) );
     Vector sands = new Vector( CONC_CHUNK_NUM );
     for( int i = RIGHT_CONC_CHUNK_IDX; i < GKS_CHUNK_NUM; i++ )
          sands.addElement( new Integer(i) );
     Vector sands1 = new Vector( concNum );
     Vector sands2 = new Vector( concNum );
     for( int i = 0; i < concNum; i++ ){
          Integer elem = (Integer)(sands.remove((int)(Math.random()*sands.size())));
          sands1.addElement( elem );
          sands2.addElement( elem );
     }
     // set up right rule if allowed.
     if( Math.random() < rightRuleProbs[ concNum-1 ] ){
         boolean bVal1 = sands1.remove( new Integer( RIGHT_CONC_CHUNK_IDX ) );
         if( bVal1 ){ // set up right rule.
             int conclusion = RIGHT_CONC_CHUNK_IDX;
             short[] condition = new short[1];
             condition[0] = RIGHT_COND_CHUNK_IDX;
             boolean bVal2 = condSands.remove( new Integer( RIGHT_COND_CHUNK_IDX ) );

             assocRuleGroups[conclusion] = new AssocRuleGrp( conclusion, global );
             assocRuleGroups[conclusion].insertRule( condition, (short)conclusion );
         }

         System.out.println( "right rule setup =" + bVal1 );
     }
     int num = sands1.size();

     //System.out.println( "the size =" + num );

     // set up one rule for each selected conclusion.
     while( sands1.size() > 0 ){
          int conclusion = ((Integer)(sands1.remove((int)(Math.random()*sands1.size())))).intValue();
          short[] condition = new short[1];
          condition[0] = (short)((Integer)(condSands.remove((int)(Math.random()*condSands.size())))).intValue();
          if( conclusion == RIGHT_CONC_CHUNK_IDX && condition[0] == RIGHT_COND_CHUNK_IDX ){
              sands1.add( new Integer( conclusion ) );
              condSands.add( new Integer( condition[0] ) );
              //assocRuleGroups[conclusion] = new AssocRuleGrp( conclusion, global );
              continue;
          }
          assocRuleGroups[conclusion] = new AssocRuleGrp( conclusion, global );
          assocRuleGroups[conclusion].insertRule( condition, (short)conclusion );
     }
     // set up other rules for each selected conclusion.
     int ruleNum = (int)( Math.round( (COND_CHUNK_NUM-concNum)*Math.random() ) );
     //int ruleNum = COND_CHUNK_NUM-concNum;
     //int ruleNum = 0;

     //System.out.println( "rule num = " + (concNum+ruleNum) );

     while( ruleNum > 0 ){
          short[] condition = new short[1];
          condition[0] = (short)((Integer)(condSands.remove((int)(Math.random()*condSands.size())))).intValue();
          int conclusion = ((Integer)(sands2.elementAt((int)(Math.random()*sands2.size())))).intValue();
          if( conclusion == RIGHT_CONC_CHUNK_IDX && condition[0] == RIGHT_COND_CHUNK_IDX ){
              if( concNum > 1 || condSands.size() > 1 ){
                  condSands.add( new Integer( condition[0] ) );
                  continue;
             }
          }
          assocRuleGroups[conclusion].insertRule( condition, (short)conclusion );
          ruleNum--;
     }
     // fill in the rules.
     for( int i = 0; i < GKS_CHUNK_NUM; i++ )
          if( assocRuleGroups[i] != null )
              gksChunks[i].setRules( assocRuleGroups[i] );
  }
}