package tasksimulators.newsimu.MCM;
import javax.swing.*;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class TaskMCM extends Task implements SimuTask {

  final int BLOCK_NUM = 20, RECORD_NUM = 20;
  final int EXP_NUM = 1, GROUP_NUM = 1, AGENT_NUM = 134;
  boolean[][] solutions;
  double[][] warmths;
  int expIdx;
  String[] groupOptionNames = { "frGroup" };
  WriteFile[] agentDataFiles, groupDataFiles, statDataFiles;

  public TaskMCM()
  {
     solutions = new boolean[ AGENT_NUM ][ BLOCK_NUM ];
     warmths = new double[ AGENT_NUM ][ BLOCK_NUM ];
     createDataFiles();
  }

  // from SimuTask.java
  public String getTaskName()
  {
     return "Meta-Cognitive Monitoring";
  }

  // from SimuTask.java
  public void start( int expMode, int blockNum, WriteFile wFile )
  {
  }

  public void init()
  {
     super.settingNames = groupOptionNames;
     super.init();
  }

  // from Task.java
  public void start()
  {
     globals[0].setAgentGroupId( 0 );
     agents = new AgentMCM[ 1 ];
     for( expIdx = 0; expIdx < EXP_NUM; expIdx++ ){
        for( int i = 0; i < AGENT_NUM; i++ ){
             System.out.println( "cur agent no = " + i );

             agents[0] = new AgentMCM( globals[0] );
             ((AgentMCM)agents[0]).run();
             ((AgentMCM)agents[0]).getOutput( warmths[i] );
             solutions[i] = ((AgentMCM)agents[0]).isCorrect();
             saveAgentData( i );
             //agents[0].reinit();
             //agents[0] = new AgentMCM( globals[0] );
        }
        saveGroupData();
     }

     System.out.println( "the task is finished! " );
  }

  /***************************** SAVE FUNCTIONS ****************************/
  public void saveAgentData( int agentIdx )
  {
      String str = "Agent no = " + String.valueOf( agentIdx ) + "\n";
      str += "the solution = " + String.valueOf( solutions[agentIdx] ) + "\n";
      for( int i = 0; i < RECORD_NUM; i++ ){
           str += "last " + (RECORD_NUM - i);
           str += "warmth = " + String.valueOf( warmths[ agentIdx ][ BLOCK_NUM-RECORD_NUM + i ] ) + "\n";
      }
      agentDataFiles[expIdx].saveStringData( str );
  }

  public void saveGroupData()
  {
      for( int j = 0; j < RECORD_NUM; j++ ){
           int rightNum = 0;
           int wrongNum = 0;
           double rightAvgScore = 0.0;
           double wrongAvgScore = 0.0;
           for( int i = 0; i < AGENT_NUM; i++ ){
                int idx = BLOCK_NUM - RECORD_NUM + j;
                if( warmths[i][idx] > 0.0 ){
                    rightNum += (solutions[i][idx]) ? 1 : 0;
                    wrongNum += (!solutions[i][idx]) ? 1 : 0;
                    rightAvgScore += (solutions[i][idx]) ? warmths[i][idx] : 0.0;
                    wrongAvgScore += (!solutions[i][idx]) ? warmths[i][idx] : 0.0;
                }
           }
           rightAvgScore /= (rightNum==0) ? 1 : rightNum;
           wrongAvgScore /= (wrongNum==0) ? 1 : wrongNum;
           String str = "the episode no = last " + (RECORD_NUM - j) + "\n";
           //str += "the experiment no = " + String.valueOf( expIdx ) + "\n";
           str += "the # of agents with correct solution = " + String.valueOf( rightNum ) + "\n";
           str += "the average warmth of agents with correct solutions = " +
                  String.valueOf(rightAvgScore)+ "\n";
           str += "the # of agents with wrong solution = " + String.valueOf( wrongNum ) + "\n";
           str += "the average warmth of agents with wrong solutions = " +
                  String.valueOf(wrongAvgScore)+ "\n";
           groupDataFiles[expIdx].saveStringData( str );
      }
  }

/***************************** PRIVATE METHODS ********************************/
  private void createDataFiles()
  {
      agentDataFiles = new WriteFile[ EXP_NUM ];
      groupDataFiles = new WriteFile[ EXP_NUM ];
      statDataFiles  = new WriteFile[ EXP_NUM ];
      String fileDirName = "dataFiles\\generic\\MCM\\";
      String[] fileNames = { "agtExp.txt", "grpExp.txt", "statExp.txt" };
      for( int i = 0; i < EXP_NUM; i++ ){
           agentDataFiles[i] = new WriteFile();
           groupDataFiles[i] = new WriteFile();
           statDataFiles[i] = new WriteFile();
           agentDataFiles[i].openFile( 1, fileDirName + fileNames[0] );
           groupDataFiles[i].openFile( 1, fileDirName + fileNames[1] );
           statDataFiles[i].openFile( 1, fileDirName + fileNames[2] );
      }
  }
}