/* Organizational Structure task */
package tasksimulators.newsimu.OS;
import clarion.common.*;
import clarion.nacs.*;
import clarion.tools.*;
import clarion.options.*;
import clarion.system.*;
import java.util.Random;
import java.io.*;

/* agents[0-8] are the subordinates
   agents[9] is the supervisor
 */
public class TaskOS extends Task implements SimuTask {

  /* CONSTANTS */
  static final int team = 1;       // macro values: team vs. hierarchy
  static final int hierarchy = 2;

  static final int dist = 1;       // macro values: distributed vs. blocked
  static final int block = 2;

  static final int NUM_PROBLEMS = 100;  // number of problems in training set
  static final int NUMBER_OF_RUNS = 30000;
  static final int NUM_REPETITIONS = 1;  // number of repetitions under each setting

  // data about resource access structure
  short blocked[][] =
    {{0,1,2},{0,1,2},{0,1,2},{3,4,5},{3,4,5},{3,4,5},{6,7,8},{6,7,8},{6,7,8}};
  short distributed[][] =
    {{0,1,2},{3,4,5},{6,7,8},{0,3,6},{1,4,7},{2,5,8},{1,2,3},{4,5,6},{7,8,0}};

  String basedir = "d:\\output\\";
  /* END CONSTANTS */

  /* SETTINGS */
  /*
  double _learn[] = {.5, .25, .1, .05};
  double _temp[] = {.05, .1, .15};
  double _probBL[] = {.95, .75, .5, .25};
  int _org[] = {1, 2};
  int _str[] = {1, 2};
  */
  double _learn[] = {.5};
  double _temp[] = {.05};
  double _probBL[] = {.95};
  int _org[] = {1};
  int _str[] = {1};
  /* END SETTINGS */

  public String getTaskName(){
    return "Organizational Structure";
  }

  public void start( int expMode, int blockNum, WriteFile wFile )
  {
  }

  public TaskOS()
  {
    setRunningAgentNum(10);
  }

  // initialize the global of each agent without GUI.
  public void init()
  {
     String [] settings =
     {"subordinate", "subordinate", "subordinate", "subordinate", "subordinate",
      "subordinate", "subordinate", "subordinate", "subordinate", "supervisor"};
     super.settingNames = settings;
     super.init();
  }

  // the start point of the task.
  public void start()
  {
    agents = new AgentOS[10];
    for(int i = 0; i < 10; i++)
      agents[i] = new AgentOS( globals[i] );

    int taskNum = 0;

    for(int i = 0; i < _temp.length; i++)
      for(int j = 0; j < _learn.length; j++)
        for(int k = 0; k < _probBL.length; k++)
          for(int l = 0; l < _org.length; l++)
            for(int m = 0; m < _str.length; m++)
            { System.out.println(i + " " + j + " " + k + " " + l + " " + m);
              String paramlist = _temp[i] + " " + _learn[j] + " " +
                _probBL[k] + " " + _org[l] + " " + _str[m];

              try
              {
                //BufferedWriter out = new BufferedWriter(new FileWriter (basedir + paramlist + ".txt"));
                System.out.println((++taskNum) + ") Beginning " + paramlist);
                //begin(_org[l], _str[m], _temp[i], _learn[j], _probBL[k], out);
                begin(_org[l], _str[m], _temp[i], _learn[j], _probBL[k], null);
                //out.close();

                System.out.println("PRINTING RULES");
                String [] arr = agents[0].getRerRules(0);
                for( int ij = 0; ij < arr.length; ij++ )
                     System.out.println(arr[ij]);
                System.exit(0);

              }
              catch (Exception e)
              {
                //System.out.println("Couldn't open output file " + paramlist);
                e.printStackTrace();
                System.exit( 1 );
              }
            }
  }

  private void begin(int org, int strc, double temp, double learnRate,
    double probBL, BufferedWriter writer)
  {
    int decisions[] = new int[9];
    int finalDecision;
    int outputs[] = new int[1];

    Random rand = new Random();
//    if(org == hierarchy)
//      setRunningAgentNum(10);
//    else
//      setRunningAgentNum(9);

    for(int i = 0; i < getRunningAgentNum(); i++)
      ((AgentOS) agents[i]).setOptions(temp, learnRate, probBL);

    short ras[][];
    if(strc == dist)
      ras = distributed;
    else
      ras = blocked;

    // place to store the data the agents will "see" based on the
    // distribtion structure (blocked/distributed)
    short agentData[][] = new short[9][3];

    // create the set of problems that will be seen
    // aka - set up training data
    short problems[][] = new short[NUM_PROBLEMS][9];
    for(int i = 0; i < NUM_PROBLEMS; i++)
      newProblem(problems, i);

    short problem[] = new short[9];

    System.out.println();

    for(int i = 0; i < NUMBER_OF_RUNS; i++)
    {
      if(i % 100 == 0)
        System.out.println( "\n current step = " + i + "...");

      // choose a random problem from the training set
      problem = problems[rand.nextInt(NUM_PROBLEMS)];

      // get decision of all nine subordinates
      for(int j = 0; j < 9; j++)
      {
        for(int k = 0; k < 3; k++)
          agentData[j][k] = problem[ras[j][k]]; // ras - resource access structure
                                                // used to map correct problem
                                                // data to individual agent

        // get the decision of the agent
        ((AgentOS) agents[j]).setSensoryInput(agentData[j], correctDecision(problem));
        agents[j].run();

        ((AgentOS) agents[j]).getOutput(outputs);
        decisions[j] = outputs[0];

        /*
        if( j == 0 ){
            System.out.println( "agent = " + j );
            System.out.println( "RULES IN CURRENT STEP = ");
            String [] arr = agents[j].getRerRules(0);
            for( int ij = 0; ij < arr.length; ij++ )
                 System.out.println( arr[ij] );
        }
        */
      }

      // if we have a supervisor pass the sub's decisions to him
      if(org == hierarchy)
      {
        ((AgentOS) agents[9]).setSensoryInput(toShortArray(decisions), correctDecision(problem));
        agents[9].run();
        ((AgentOS) agents[9]).getOutput(outputs);
        finalDecision = outputs[0];
      }
      else  // otherwise use a majority vote
        finalDecision = majority(decisions);

      // see if the final decision was correct for the given problem
      if(finalDecision == correctDecision(problem))
      {
        try // print to file
        {
          //writer.write("1\n");
          //writer.flush();
        }
        catch(Exception e) { System.out.println("error"); }
      }
      else
      {
        try
        {
          //writer.write("0\n");
          //writer.flush();
        }
        catch(Exception e) { System.out.println("error"); }
      }
    }
  }

  // returns the actual correct response to a
  // given problem
  private static int correctDecision(short prob[])
  {

     int sum=0;
     for(int i = 0; i < 9; i++)
       sum += (prob[i] + 1);

     if(sum < 17) // friendly
       return(0);
     else if(sum > 19) // hostile
       return(2);
     else // neutral
       return(1);

  }

  // returns the majority decision
  private static int majority(int decisions[])
  {

     int votes[] = new int[3];  // number of votes for each possibility

     for(int i = 0; i < 3; i++)
       votes[i] = 0; // zero out the votes initialy

     for(int i = 0; i < 9; i++)
       votes[decisions[i]]++;

     if(votes[0] > votes[1] && votes[0] > votes[2])
       return(0);
     else if(votes[1] > votes[0] && votes[1] > votes[2])
       return(1);
     else if(votes[2] > votes[0] && votes[2] > votes[1])
       return(2);
     else
       return(1);

  }

  // this is a lot of work to generate a problem, but it beats
  // creating all possible problems and selecting without replacement
  private static short[] newProblem(short problems[][], int q)
  {
     short prob[] = new short[9];
     Random rand = new Random();

     for(int i = 0; i < 9;i++)
       prob[i] = (short) rand.nextInt(3);

     while(!uniqueProblem(problems,q,prob))
       for(int i = 0; i < 9; i++)
         prob[i] = (short) rand.nextInt(3);

     return(problems[q] = prob);
  }


  private static boolean uniqueProblem(short pastProblems[][], int q, short prob[])
  {
     for(int i = 0; i < q; i++)
       if(problemsEqual(pastProblems[i],prob))
         return false;

     return true;
  }


  private static boolean problemsEqual(short pastProblems[], short prob[])
  {
     for(int i = 0; i < 9; i++)
       if(pastProblems[i] != prob[i])
         return false;

     return true;
  }

  private short[] toShortArray(int[] array)
  {
    short[] newarray = new short[array.length];
    for(int i = 0; i < array.length; i++)
      newarray[i] = (short) array[i];

    return newarray;
  }
}