package tasksimulators.newsimu.PC;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;

public class AgentPC extends TaskAgent{

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  final int TARGET = 5;
  //final int TARGET = 2;
  final double POS_FEEDBACK = 1.0, NEG_FEEDBACK = -0.2;
  int DIM_NUM, DIM_VAL_NUM, ACTION_NUM;
  int TRIALS_PER_BLOCK, BLOCK_NUM;
  int blockIdx, trialIdx, curKnowType;
  short onTargetNums[], testOnTargetNums[][];
  //int curIrlAction, nextIrlAction;
  //double irlRuleParams[][][];
  boolean preTrainFlag, taskLearnTestOn;
  int acsNetNum, KNOW_NUM;
  boolean taskSets[][], learningTestSets[][];
  String featureNames[][];

  public AgentPC(){}

  public AgentPC( Global g )
  {
     super( g );
     DIM_NUM = global.getIdnInputDimNum( 0 );
     DIM_VAL_NUM = global.getIdnInputDimVals( 0 )[0];
     ACTION_NUM = global.getIdnOutputDimVals( 0 )[0];
     //TRIALS_PER_BLOCK = global.getEpisodeLength();
     TRIALS_PER_BLOCK = 11;
     //BLOCK_NUM = global.getEpisodeNum( 0, 0, 0 );
     BLOCK_NUM = 20;
     featureNames = global.getFeatureNames();
     //irlRuleParams = global.getIrlRuleParameters( 0 );
     reinforcements = new double[ 1 ];
     curSensoryInput = new Feature[ DIM_NUM ];
     onTargetNums = new short[ BLOCK_NUM ];
     taskLearnTestOn = global.isLearningTestOn();
     //KNOW_NUM = global.getAcsKnowledgeNum();
     KNOW_NUM = 4;
     if( taskLearnTestOn ){
         acsNetNum = global.getAcsNetNum();
         testOnTargetNums = new short[ KNOW_NUM ][ BLOCK_NUM ];
         taskSets =  new boolean[ acsNetNum ][ KNOW_NUM ];
         learningTestSets = new boolean[ acsNetNum ][ KNOW_NUM ];
         for( int i = 0; i < acsNetNum; i++ ){
              taskSets[i][global.BL] = global.isIdnOn(i);
              taskSets[i][global.RER] = global.isRerOn(i);
              taskSets[i][global.IRL] = global.isIrlOn(i);
              taskSets[i][global.FR] = global.isFrOn(i);
         }
     }
     initSensory();
  }

  public void reinit()
  {
     super.reinit();
     for( int i = 0; i < BLOCK_NUM; i++ )
          onTargetNums[ i ] = 0;
     if( taskLearnTestOn ){
         for( int i = 0; i < KNOW_NUM; i++ )
          for( int j = 0; j < BLOCK_NUM; j++ )
               testOnTargetNums[i][j] = 0;
         for( int i = 0; i < acsNetNum; i++ ){
              taskSets[i][global.BL] = global.isIdnOn(i);
              taskSets[i][global.RER] = global.isRerOn(i);
              taskSets[i][global.IRL] = global.isIrlOn(i);
              taskSets[i][global.FR] = global.isFrOn(i);
         }
     }
  }

  public boolean isControlNacsAction( int netIdx, short[] chosenAction )
  {
     return false;
  }

  public void getOutput( Object arr )
  {
     for( int i = 0; i < BLOCK_NUM; i++ )
          ((int[])arr)[ i ] = onTargetNums[ i ];
  }

  public void getLearningTestData( int[][] arr )
  {
     for( int i = 0; i < KNOW_NUM; i++ )
       for( int j = 0; j < BLOCK_NUM; j++ )
             arr[i][j] = testOnTargetNums[i][j];
  }

  public String getAemTestString()
  {
     String aemTestStr = clarion.testAem( 0 );
     return aemTestStr;
  }

  public int getCurSensoryInputType()
  {
     return 0;
  }

  public String getCurSensoryInputLabel()
  {
     return "sensory input";
  }

  public void setRetrievalResults( int netIdx, GKSChunk[] results  )
  {
  }

  public void setPreTrainFlag( boolean flag )
  {
     preTrainFlag = flag;
  }

  public int determineAmnNet( int netIdx, short[] action )
  {
      activeAmnIdx = 0;
      return activeAmnIdx;
  }

  public void run()
  {
     int runNum = 1;
     if( preTrainFlag )
         runNum++;
     for( int i = 0; i < runNum; i++ ) {
       for( blockIdx = 0; blockIdx < BLOCK_NUM; blockIdx++ ) {
            //System.out.println( "block idx = " + blockIdx );
            initSensory();
            // training.
            for( trialIdx = 1; trialIdx < TRIALS_PER_BLOCK; trialIdx++ ){
                 //System.out.println( "trial idx = " + trialIdx );
                 clarion.training();
            }
            // learning testing including every knowledge type.
            if( !preTrainFlag && taskLearnTestOn ){
                learningTestOn = true;
                for( curKnowType = 0; curKnowType < KNOW_NUM; curKnowType++ ){
                     if( taskSets[0][curKnowType] ){
                         setLearningTestSetting();
                         for( trialIdx = 1; trialIdx < TRIALS_PER_BLOCK; trialIdx++ )
                              clarion.training();
                     }
                }
                resetLearningTestSetting();
                learningTestOn = false;
            }
       }
       preTrainFlag = false;
     }
     // test AEM.
     //String aemTestStr = clarion.testAem( 0 );
     //System.out.println( aemTestStr );
  }

  public void performAction( int netIdx, short[] action )
  {
     int curOutput = calcOutput( action[0] );
     //reinforcements[ 0 ] = determineFeedback( curOutput );
     short[] values = new short[ 1 ];
     values[ 0 ] = (short)curOutput;
     Feature[] newSensoryInput = new Feature[1];
     int CUR_OUTPUT_DIM = 1;
     newSensoryInput[ 0 ] = new Feature( featureNames[0][CUR_OUTPUT_DIM], (short)0,
                                         (short)CUR_OUTPUT_DIM, (short)DIM_VAL_NUM, values, global );
     reinforcements[ 0 ] = taskClarion.getReinforcement( netIdx, curSensoryInput, null,
                           null, action, newSensoryInput, null, null );

     //System.out.println( "reinf = " + reinforcements[0] );

     if( !preTrainFlag && reinforcements[ 0 ] > 0 ){
         if( !learningTestOn )
             onTargetNums[ blockIdx ]++;
         else
             testOnTargetNums[ curKnowType ][ blockIdx ]++;
     }
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
     /*
     for( int i = DIM_NUM-2; i >= 2; i-- ){
          short[] values = new short[ curSensoryInput[i-2].getValueNum() ];
          curSensoryInput[i-2].getValues( values );
          curSensoryInput[i].setValues( values );
     }
     for( int i = 0; i < 2; i++ ){
          short[] values = new short[ 1 ];
          if( i == 0 )
              values[ 0 ] = (short)action;
          else
              values[ 0 ] = (short)curOutput;
          curSensoryInput[i].setValues( values );
     }
     */
     //short[] values = new short[ 1 ];
     //values[ 0 ] = (short)curOutput;
     curSensoryInput[0].setValues( values );
  }

  /*
  public short[] decideIrlAction( int netIdx, int setIdx, int ruleNo, short[] input )
  {
     double sum = 0.0;
     for( int i = 0; i < DIM_NUM; i++ )
      for( int j = 0; j < DIM_VAL_NUM; j++ )
           if( input[ i * DIM_VAL_NUM + j ] == 1 )
               sum += irlRuleParams[setIdx][ruleNo][i] * (j+1);
     // add the constant term.
     //sum += irlRuleParams[setIdx][ruleNo][DIM_NUM] * 1;
     //
     // add the target term.
     sum += irlRuleParams[setIdx][ruleNo][DIM_NUM] * (TARGET+1);
     // add the constant term.
     sum += irlRuleParams[setIdx][ruleNo][DIM_NUM+1] * 1;
     if( setIdx == 0 || setIdx == 1 )
         curIrlAction = (int)sum - 1;
     else
     {
         curIrlAction = nextIrlAction;
         nextIrlAction = (int)sum - 1;
     }
     if( curIrlAction < 0 )
         curIrlAction = 0;
     else if( curIrlAction >= ACTION_NUM -1 )
             curIrlAction = ACTION_NUM -1 ;
     // revised on 5/1/03
     short[] irlAction = new short[1];
     irlAction[0] = (short)curIrlAction;
     return irlAction;
  }
  */

/**************************** PRIVATE METHODS *********************************/
  private void initSensory()
  {
    /*
    for( int i = 0; i < 2; i++ )
    {
         short[] values = new short[1];
         values[0] = (short)( Math.random() * DIM_VAL_NUM );
         curSensoryInput[ i ] = new Feature( featureNames[0][i],(short)0,(short)i,
                                             (short)DIM_VAL_NUM, values, global );
    }
    for( int i = 2; i < DIM_NUM; i++ )
    {
         short[] content = new short[DIM_VAL_NUM];
         for( int j = 0; j < DIM_VAL_NUM; j++ )
              content[ i ] = 0;
         if( i == DIM_NUM - 1 )     // target dim.
              content[ TARGET ] = 1;
         curSensoryInput[ i ] = new Feature( featureNames[0][i],(short)0,(short)i,
                                             content,global );
    }
    */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    int CUR_OUTPUT_DIM = 1;
    short[] values = new short[1];
    values[0] = (short)( Math.random() * DIM_VAL_NUM );
    curSensoryInput[ 0 ] = new Feature( featureNames[0][CUR_OUTPUT_DIM], (short)0,
                                        (short)CUR_OUTPUT_DIM, (short)DIM_VAL_NUM, values, global );
    // revised on 04/15/03
    curNacsInput = curSensoryInput;
  }

  public void setLearningTestSetting()
  {
     for( int i = 0; i < acsNetNum; i++ ){
          learningTestSets[i][global.BL] = false;
          learningTestSets[i][global.RER] = false;
          learningTestSets[i][global.IRL] = false;
          learningTestSets[i][global.FR] = false;
     }
     for( int i = 0; i < acsNetNum; i++ )
          learningTestSets[i][curKnowType] = true;
     for( int i = 0; i < acsNetNum; i++ ){
          global.setRerOn( i, learningTestSets[i][global.RER] );
          global.setIrlOn( i, learningTestSets[i][global.IRL] );
          global.setFrOn( i, learningTestSets[i][global.FR] );
     }
  }

  public void resetLearningTestSetting()
  {
     for( int i = 0; i < acsNetNum; i++ ){
          global.setIdnOn( i, taskSets[i][global.BL] );
          global.setRerOn( i, taskSets[i][global.RER] );
          global.setIrlOn( i, taskSets[i][global.IRL] );
          global.setFrOn( i, taskSets[i][global.FR] );
     }
  }

 /* CALCULATE THE OUTPUT ACCORDING TO DIFFERENT EQUATIONS.*/
  private int calcOutput( int inputVal )
  {
      int a, b, w, p0, n, p1, iRand;
      a = 2;
      b = 0;
      w  = inputVal + 1;
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
      p0 = curSensoryInput[0].getValues()[0] + 1;
      //p0 = curSensoryInput[1].getValues()[0] + 1;
      iRand = ( int ) ( Math.random() * 3 );
      switch ( iRand ) {
          case 0:
                  n = 1;
                  break;
          case 1:
                  n = 0;
                  break;
          case 2:
                  n = - 1;
                  break;
          default:
                  n = 0;
       }
       //n = 0;

       p1 = a * w + b - p0 + n;
       p1--;
       if( p1 < 0 )
           p1 = 0;
       else if( p1 >= ACTION_NUM - 1 )
                p1 = ACTION_NUM - 1;
       return p1;
  }

  private double determineFeedback( int val )
  {
        double fb;
       	int diff = Math.abs( val - TARGET );
       	switch( diff )
        {
           case 0:
           case 1:
              	fb = POS_FEEDBACK;
                break;
           default:
              	fb = NEG_FEEDBACK * ( diff - 1 );
                break;
        }
        return fb;
  }
}
