package tasksimulators.newsimu.PC;
import java.util.Vector;
import clarion.system.*;
import clarion.common.*;
import clarion.acs.net.auxiliary.*;

public class TaskClarionPC extends TaskClarion {

  final int CONTROL_GROUP = 0, ORIGINAL_GROUP = 1;
  final int MEMORY_GROUP = 2, SIMPLE_GROUP = 3;
  final int INPUT_DIM_NUM = 1, ACTION_NUM = 12, DIM_VAL_NUM = 12, TARGET = 5;
  final int IRL_SET_NUM = 4;
  Vector[] irlRules;
  // parameters of rule set
  int[] a = { 1, 2 }; // a's b's and c's
  int[] b = { -2, -1, 0, 1, 2 }; // values to try in the rules
  int[] c = { -2,- 1, 1, 2 };
  // fixed rules
  short[] frConditions[][][];
  short[] frActions[][];
  { // initialize FR rules.
    frConditions = new short[ 1 ][ ACTION_NUM ][ INPUT_DIM_NUM ][ DIM_VAL_NUM ];
    for( int i = 0; i < 1; i++ )
      for( int j = 0; j < ACTION_NUM; j++ ){
           for( int k = 0; k < INPUT_DIM_NUM; k++ ){
                    for( int m = 0; m < DIM_VAL_NUM; m++ )
                         frConditions[i][j][k][m] = 0;
                    frConditions[i][j][k][j] = 1;
                }
      }
    frActions = new short[ 1 ][ ACTION_NUM ][ 1 ];
    for( int i = 0; i < ACTION_NUM; i++ ) {
         int n = 0, iRand = ( int ) ( Math.random() * 3 );
         switch( iRand ) {
         case 0:  n = 1;
                  break;
         case 1:  n = 0;
                  break;
         case 2:  n = -1;
                  break;
         }
        frActions[0][i][0] = (short)(( TARGET + i + n ) / 2 );
    }
  }

  public TaskClarionPC()
  {
     generateIrlRules();
  }

  // eligibility of ACS networks, return the eligible condition of the network.
  public short[][][] getEligibility( int netIdx )
  {
    short[][][] eligCondition = {
                   {
                      {0},            // relevant dims: 0 in this condition.
                      {1,1,1,1,1},    // all values are eligible.
                   },
                };
     return eligCondition;
  }

  // IRL parameters to decide an action
  public IrlRuleForm getIrlRule( int groupIdx, int netIdx, int setIdx, int ruleIdx )
  {
     return (IrlRuleForm)(irlRules[setIdx].elementAt(ruleIdx));
  }

  // Fixed Rule conditions
  public short[][] getFrCondition( int groupIdx, int netIdx, int ruleIdx )
  {
     if( groupIdx == MEMORY_GROUP )
         return frConditions[netIdx][ruleIdx];
     else
         return (short[][]) null;
  }

  // Fixed Rule actions
  public short[] getFrAction( int groupIdx, int netIdx, int ruleIdx )
  {
     if( groupIdx == MEMORY_GROUP )
         return frActions[netIdx][ruleIdx];
     else
         return (short[])null;
  }

  private void generateIrlRules()
  {
     irlRules = new Vector[IRL_SET_NUM];
     for( int generation = 1; generation < 5; generation++ ){
        int ruleType = -1;
        int[] tempB = new int[5];
        int[] tempC;
        if( ( generation % 2 ) == 0 )
            tempC = new int[4];
        else
            tempC = new int[1];
        switch( generation ){
           case 1:
                ruleType = 0;
                tempB = b;
                tempC[0] = 0;
                break;
           case 2:
                ruleType = 1;
                tempB = b;
                tempC = c;
                break;
           case 3:
                ruleType = 2;
                tempB = b;
                tempC[0] = 0;
                 break;
           case 4:
                 ruleType = 3;
                 tempB = b;
                 tempC = c;
                 break;
        }
        // generate a new rule set.
        irlRules[generation-1] = new Vector();
        for( int i = 0; i < a.length; i++ )
          for( int j = 0; j < tempB.length; j++ )
             for( int k = 0; k < tempC.length; k++ )
                  irlRules[generation-1].addElement( new IrlRuleFormPC( a[i], tempB[j], tempC[k], ruleType,
                                                  0, TARGET ) );
    }
  }
}