package tasksimulators.newsimu.PC;
import javax.swing.*;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class TaskPC extends Task implements SimuTask {

  final int INPUT_DIM_NUM = 1, DIM_VAL_NUM = 12, ACTION_NUM = 12;
  final int STATE_LEN = 12, TARGET = 5;
  //final int INPUT_DIM_NUM = 5, DIM_VAL_NUM = 12;
  //final int ACTION_NUM = 12;
  //final int STATE_LEN = 60, TARGET = 5;
  final int TRIALS_PER_BLOCK = 11, BLOCK_NUM = 20;
  final int PRE_TRAIN_BLOCK_NUM = 20;
  final int COMP_NUM = 4, DATA_TYPE_NUM = 2;
  final int TASK_NUM = 2, GROUP_NUM = 4, AGENT_NUM = 30;
  final int CONTROL_GROUP = 0, ORIGINAL_GROUP = 1;
  final int MEMORY_GROUP = 2, SIMPLE_GROUP = 3;
  final int DC = 0, PT = 1;
  final String[][] groupOptionNames = {
        {"controlGroup"}, {"originalGroup"}, {"memoryGroup"}, {"simpleGroup"} };
  final double DC_HUMAN_DATA[] = { 1.97, 2.57, 4.63, 4.00 };
  final double PT_HUMAN_DATA[] = { 2.85, 3.75, 5.53, 5.91 };
  final boolean WITH_RER = true, WITHOUT_RER = false;
  final boolean WITH_IRL = true, WITHOUT_IRL = false;
  final boolean initRuleSet[][][] =  {
          { { WITH_RER, WITHOUT_IRL }, { WITH_RER, WITH_IRL } },
          { { WITHOUT_RER, WITH_IRL }, { WITH_RER, WITH_IRL } },
          { { WITH_RER, WITH_IRL }, { WITH_RER, WITHOUT_IRL },
          { WITHOUT_RER, WITH_IRL } }
  };
  boolean preTrainFlag, finishFlag, guiFlag;
  boolean taskFinishFlag = true;
  boolean learningTestOn = false;
  int runNo = 2, taskIdx, groupIdx, agentIdx;
  int onTargetNums[][][], testOnTargetNums[][][][];
  double onTargetRates[][], totOnTargetRates[];
  double dcMeanSqErr, ptMeanSqErr, totMeanSqErr;
  double MODEL_DATA[][];
  String aemTestString;
  String rerRules[], irlRules[], fixRules[];
  WriteFile modelDataFile, dataFiles[][][][];
  String  ruleSetName;

  public TaskPC()
  {
     setRunningAgentNum( 1 );
     ruleSetName = "FullModel";
     createDataFiles();
     onTargetNums     = new int[ GROUP_NUM ][ AGENT_NUM ][ BLOCK_NUM ];
     testOnTargetNums = new int[ GROUP_NUM ][ AGENT_NUM ][ COMP_NUM ][ BLOCK_NUM ];
     onTargetRates    = new double[ GROUP_NUM ][ AGENT_NUM ];
     totOnTargetRates = new double[ GROUP_NUM ];
     MODEL_DATA    = new double[ TASK_NUM ][ GROUP_NUM ];
     rerRules = new String[3];
     irlRules = new String[3];
     fixRules = new String[3];
  }

  // from SimuTask.java
  public String getTaskName()
  {
     return "DYNAMIC CONTROL TASK";
  }

  // from SimuTask.java
  public void start( int expMode, int blockNum, WriteFile wFile )
  {
  }

  public void init()
  {
     super.settingNames = groupOptionNames[groupIdx];
     super.init();
  }

  // from Task.java
  public void start()
  {
      agents = new AgentPC[1];
      guiFlag = globals[0].isGuiMode();
      if( taskFinishFlag )
          initNewTask();
      else
          initNewGroup();
      finishFlag = false;
      while( !finishFlag ){
          ((AgentPC)agents[0]).setPreTrainFlag( preTrainFlag );
          agents[0].run();
          agents[0].getOutput( onTargetNums[ groupIdx ][ agentIdx ] );
          if( globals[0].isLearningTestOn() )
             ((AgentPC)agents[0]).getLearningTestData( testOnTargetNums[groupIdx][agentIdx] );
          if( globals[0].isAEMon() )
              aemTestString = ((AgentPC)agents[0]).getAemTestString();
          gotoEndThisAgent();
      }
  }

  /*************************** intialization functions ************************/
  public void gotoEndThisAgent()
   {
        saveOnTargets( 0 );
        saveRules();
        if( agentIdx + 1 < AGENT_NUM ){
            agentIdx++;
            initNewAgent();
        }
        else{
            saveOnTargets( 1 );
            if( groupIdx + 1 < GROUP_NUM ){
                groupIdx++;
                if( checkFinish( 0 ) )
                    return;
                initNewGroup();
            }
            else{
                for( int i = 0; i < GROUP_NUM; i++ ) {
                    MODEL_DATA[taskIdx][i] = totOnTargetRates[ i ];
                    MODEL_DATA[taskIdx][i] *= TRIALS_PER_BLOCK - 1;
                }
                if( taskIdx + 1 < TASK_NUM ) {
                    taskIdx++;
                    if( checkFinish( 1 ) )
                        return;
                    initNewTask();
                }
                else{
                     saveModelData();
                     allFinish();
                }
            }
        }
   }

   public void initNewAgent()
   {
      System.out.println( "cur agent no = " + agentIdx );
      if( taskIdx == 1 )
          preTrainFlag = true;
      else
          preTrainFlag = false;
      globals[0].setAgentGroupId( groupIdx );
      if( agents[0] == null ){
          agents[0] = new AgentPC( globals[0] );
      }
      else
          agents[0].reinit();
   }

   public void initNewGroup()
   {
       String[] groupNames = { "CONTROL GROUP", "ORIGINAL GROUP", "MEMORY GROUP", "SIMPLE GROUP" };
       System.out.println( "cur group = " + groupNames[groupIdx] );
       init();
       agentIdx = 0;
       initNewAgent();
   }

   public void initNewTask()
   {
       String[] taskNames = {"Process Control Task", "Human Task" };
       System.out.println( "cur task = " + taskNames[taskIdx] );
       for( int i = 0; i <  GROUP_NUM; i++ )
         for( int j = 0; j <  AGENT_NUM; j++ )
            for( int k = 0; k <  BLOCK_NUM; k++ )
                 onTargetNums[ i ][ j ][ k ] = 0;
       for( int i = 0; i <  GROUP_NUM; i++ )
         for( int j = 0; j <  AGENT_NUM; j++ )
           for( int k = 0; k <  COMP_NUM; k++ )
             for( int m = 0; m <  BLOCK_NUM; m++ )
                 testOnTargetNums[i][j][k][m] = 0;
       groupIdx = 0;
       initNewGroup();
   }

   public void allFinish()
   {
       System.out.println( "Experiment is finished! " );
       preTrainFlag = false;
       taskFinishFlag = finishFlag = true;
       groupIdx = taskIdx = 0;
       for( int i = 0; i <  GROUP_NUM; i++ )
         for( int j = 0; j <  AGENT_NUM; j++ )
           for( int k = 0; k <  BLOCK_NUM; k++ )
                onTargetNums[ i ][ j ][ k ] = 0;
       for( int i = 0; i <  GROUP_NUM; i++ )
         for( int j = 0; j <  AGENT_NUM; j++ )
           for( int k = 0; k <  COMP_NUM; k++ )
             for( int m = 0; m <  BLOCK_NUM; m++ )
                 testOnTargetNums[i][j][k][m] = 0;
   }

   /***************************** SAVE FUNCTIONS ****************************/
   public void saveOnTargets( int mode )
   {
      // record the training block numbers of the current agent.
      String str = "Group NO = " + String.valueOf( groupIdx ) + "  ";
      if( mode == 0 )
          str += "Agent NO = " + String.valueOf( agentIdx ) + "  ";
       str += "On-targets of each block = \n";
       if( mode == 0 )
           dataFiles[taskIdx][groupIdx][0][0].saveBlockData(
                    onTargetNums[groupIdx][agentIdx], BLOCK_NUM, 1 );
       else
       {
          int[] sum = new int[ BLOCK_NUM ];
          for( int i = 0; i < BLOCK_NUM; i++ )
             for( int j = 0; j < AGENT_NUM; j++ )
                  sum[i] +=  onTargetNums[groupIdx][j][i];
          dataFiles[taskIdx][groupIdx][0][0].saveBlockData( sum,
                                        BLOCK_NUM, AGENT_NUM );
       }
       if( mode == 0 )
           saveAgentData();
       else
           saveGroupData();
   }

   public void saveAgentData()
   {
      double sum = 0, rate = 0;
      for( int i = 0; i < BLOCK_NUM; i++ )
           sum += onTargetNums[ groupIdx ][ agentIdx ][ i ];
      rate = ( double ) sum / ( BLOCK_NUM * ( TRIALS_PER_BLOCK  - 1 ) );
      onTargetRates[ groupIdx ][ agentIdx ] = rate;
      String eachLRStr = "Group NO = " + String.valueOf( groupIdx ) + ",  " +
                         "Agent NO = " + String.valueOf( agentIdx ) +  ",  " +
                         "On-target rate = ";
      eachLRStr += String.valueOf( onTargetRates[ groupIdx ][ agentIdx ] );
      dataFiles[taskIdx][groupIdx][0][0].saveStringData( eachLRStr );
      if( globals[0].isLearningTestOn() ){
          String head = "Group NO = " + String.valueOf( groupIdx ) + ",  " +
                        "Agent NO = " + String.valueOf( agentIdx ) + "\n";
          dataFiles[taskIdx][groupIdx][1][4].saveStringData( head + aemTestString );
      }
  }

  public void saveGroupData()
  {
      int totBlockNum = 0, sum = 0;
      for( int i = 0; i < AGENT_NUM; i++ ){
         for( int j = 0; j < BLOCK_NUM; j++ )
              sum += onTargetNums[ groupIdx ][ i ][ j ];
         totBlockNum += BLOCK_NUM;
      }
      totOnTargetRates[ groupIdx ] = (double)sum /
                      ( totBlockNum * ( TRIALS_PER_BLOCK - 1 ) );
      String totalLRStr = "Group NO = " + String.valueOf( groupIdx ) + ",  " +
                          "Total trained blocks = " + String.valueOf( totBlockNum ) +
                           ",  Total on-target rate =  ";
      totalLRStr += String.valueOf( totOnTargetRates[ groupIdx ] );
      totalLRStr += "\n";
      dataFiles[taskIdx][groupIdx][0][0].saveStringData( totalLRStr );
      // save training curve data.
      String str = "";
      for( int j = 0; j < BLOCK_NUM; j++ ){
           str += String.valueOf( j+1 ) + " ";
           sum = 0;
           for( int i = 0; i < AGENT_NUM; i++ )
                sum += onTargetNums[ groupIdx ][ i ][ j ];
           str += String.valueOf( (double)sum/AGENT_NUM ) + "\n";
      }
      dataFiles[taskIdx][groupIdx][0][1].saveStringData( str );
      // save learning testing data.
      if( globals[0].isLearningTestOn() ){
          for( int i = 0; i < COMP_NUM; i++ ){
               str = "";
               for( int j = 0; j < BLOCK_NUM; j++ ){
                    str += String.valueOf( j+1 ) + " ";
                    sum = 0;
                    for( int k = 0; k < AGENT_NUM; k++ )
                         sum += testOnTargetNums[ groupIdx ][ k ][ i ][ j ];
                    str += String.valueOf( (double)sum/AGENT_NUM ) + "\n";
               }
               dataFiles[taskIdx][groupIdx][1][i].saveStringData( str );
          }
      }
  }

  public void saveModelData()
  {
      double sum = 0.0;
      String str = "\n rule setting is: " + ruleSetName;
      str += "\n The Dynamic Control task human data are: ";
      for( int i = 0; i < GROUP_NUM; i++ )
           str += String.valueOf( DC_HUMAN_DATA[ i ] ) + "    ";
      str += "\n The Dynamic Control task model data are: ";
      for( int i = 0; i < GROUP_NUM; i++ )
           str += String.valueOf( MODEL_DATA[ DC ][ i ] ) + "    ";
      str += "\n The mean square error of DC task is: ";
      sum = 0.0;
      for( int i = 0; i < GROUP_NUM; i++ )
           sum += Math.pow(( MODEL_DATA[DC][i] - DC_HUMAN_DATA[i] ),2);
      sum /= GROUP_NUM;
      dcMeanSqErr = sum;
      str += String.valueOf( dcMeanSqErr );
      if( taskIdx == 1 )
      {   // person task.
          str += "\n The Person Task human data are: ";
          for( int i = 0; i < GROUP_NUM; i++ )
               str += String.valueOf( PT_HUMAN_DATA[ i ] ) + "    ";
          str += "\n The Person Task model data are: ";
          for( int i = 0; i < GROUP_NUM; i++ )
               str += String.valueOf( MODEL_DATA[ PT ][ i ] ) + "    ";
          str += "\n The mean square error of person task is: ";
          sum = 0.0;
          for( int i = 0; i < GROUP_NUM; i++ )
               sum += Math.pow(MODEL_DATA[ PT ][ i ] - PT_HUMAN_DATA[ i ], 2);
          sum /= GROUP_NUM;
          ptMeanSqErr = sum;
          str += String.valueOf( ptMeanSqErr );
      }
      /*
      str += "\n The total mean square error of the two tasks is : ";
      sum = ( ptMeanSqErr + dcMeanSqErr ) / 2;
      totMeanSqErr = sum;
      str += String.valueOf( totMeanSqErr ) + "\n";
      */
      modelDataFile.saveStringData( str );
   }

   /************************* SAVE THE LEARNED RULES ************************/
   public void saveRules()
   {
      String rerRuleText[] = {
        "RER'S MATCH-CONDITION RULES: \n", "RER'S MATCH-BOTH RULES: \n",
        "RER'S ALL RULES: \n",
      };
      String irlRuleText[] = {
        "IRL'S MATCH-CONDITION RULES: \n", "IRL'S MATCH-BOTH RULES: \n",
        "IRL'S ALL RULES: \n",
      };
      String fixRuleText[] = {
        "FR'S MATCH-CONDITION RULES: \n", "FR'S MATCH-BOTH RULES: \n",
        "FR'S ALL RULES: \n",
      };
      dataFiles[taskIdx][groupIdx][0][0].saveStringData( " After the training of an agent,  " );
      if( globals[0].isRerOn( 0 )  )
      {
         rerRules = agents[0].getRerRules( 0 );
         for( int i = 0; i < 3; i++ )
              dataFiles[taskIdx][groupIdx][0][0].saveStringData(rerRuleText[i]+rerRules[i]);
      }
      if( globals[0].isIrlOn( 0 ) )
      {
        irlRules = agents[0].getIrlRules( 0 );
        for( int i = 0; i < 3; i++ )
             dataFiles[taskIdx][groupIdx][0][0].saveStringData(irlRuleText[i]+irlRules[i]);
      }
      if( globals[0].isFrOn( 0 )  )
      {
         fixRules = agents[0].getFixRules( 0 );
         for( int i = 0; i < 3; i++ )
              dataFiles[taskIdx][groupIdx][0][0].saveStringData(fixRuleText[i]+fixRules[i]);
      }
   }

/***************************** PRIVATE METHODS ********************************/
  private void createDataFiles()
  {
      int dataTypeNum = 1;
      if( learningTestOn )
          dataTypeNum = DATA_TYPE_NUM;
      modelDataFile = new WriteFile();
      dataFiles = new WriteFile[TASK_NUM][GROUP_NUM][dataTypeNum][];
      String fileDirName = "dataFiles\\generic\\PC\\";
      String modelFileName = "ModelDat";
      modelDataFile.openFile( 1, fileDirName + modelFileName );
      String[] taskNames = { "SugarTask", "PersonTask" };
      String[] groupNames = { "Control", "Original", "Memory", "Simple" };
      String[] dataNames[] = { {"Train", "TrainCurve"},{"TestBL","TestRER","TestIRL","TestFR","TestAem"} };
      for( int j = 0; j < TASK_NUM; j++ )
       for( int k = 0; k < GROUP_NUM; k++ )
        for( int m = 0; m < dataTypeNum; m++ ){
             dataFiles[j][k][m] = new WriteFile[ dataNames[m].length ];
             for( int n = 0; n < dataNames[m].length; n++ ){
                  dataFiles[j][k][m][n] = new WriteFile();
                  String fileName = ruleSetName + taskNames[j] + groupNames[k] + dataNames[m][n];
                  dataFiles[j][k][m][n].openFile( 1, fileDirName + fileName );
            }
       }
  }

  private boolean checkFinish( int code )
  {
     if( !guiFlag )
         return false;
     String str = "";
     switch( code ){
        case 0: str = "Run with a different subject group?";
                taskFinishFlag = false;
                break;
        case 1: str = "Run the human task?";
                taskFinishFlag = true;
                break;
     }
     if( JOptionPane.showConfirmDialog( null, str, "WARNING",
         JOptionPane.YES_NO_OPTION ) == JOptionPane.NO_OPTION ){
         switch( code ){
            case 0: for( int i = 0; i < GROUP_NUM; i++ ) {
                         if( i < groupIdx )
                             MODEL_DATA[taskIdx][i] = totOnTargetRates[ i ];
                         else
                             MODEL_DATA[taskIdx][i] = 0.0;
                         MODEL_DATA[taskIdx][i] *= TRIALS_PER_BLOCK - 1;
                    }
                    saveModelData();
                    allFinish();
                    break;
            case 1: saveModelData();
                    allFinish();
                    break;
         }
     }
     else
         finishFlag = true;
     return true;
  }
}