/**
 * XOR Task
 * This class defines the task-specific part of the Clarion model's behavior.
 * Each instance of TaskClarionXOR is associated with a single agent.
 */
package tasksimulators.newsimu.XOR;
import java.util.Arrays;
import clarion.common.*;
import clarion.tools.*;
import clarion.nacs.*;
import clarion.acs.net.*;
import clarion.acs.net.auxiliary.*;
import clarion.acs.gs.*;
import clarion.acs.wm.*;
import clarion.system.*;

public class TaskClarionXOR extends TaskClarion{

  public TaskClarionXOR()
  {
  }

  // Determines which of the IDN networks is eligible at each step:
  // * The outermost array stores all the eligibility data.
  // * The middle array (there is only one) indicates that there's only one IDN.
  // * The inner arrays indicate, respectively,
  //   - That for each possible input dimension (0 or 1)
  //   - Both values of the first dimension are eligible (0 and 1)
  //   - And both values of the second dimension are eligible (0 and 1)
  public short[][][] getEligibility( int netIdx )
  {
    short[][][] eligCondition = {
                   {
                      {0,1},
                      {1,1},
                      {1,1}
                   },
                };
     return eligCondition;
  }

  // Returns the reinforcement awarded the agent, based on the previous inputs
  // to the agent, and on the agent's subsequent decision.
  public double getReinforcement( int netIdx, Feature[] input, Chunk[] wmChunks, GoalChunk goal,
                        short[] action, Feature[] newInput, Chunk[] newWMChunks, GoalChunk newGoal )
  {
     // Retrieve the most recent inputs to the agent
     short val1 = input[0].getValues()[0];
     short val2 = input[1].getValues()[0];

     // Determine if the agent was right or wrong, and give a corresponding
     // reward of 1 or 0.
     if( (val1 == 1 && val2 == 1) || (val1 == 0 && val2 == 0) ){
       if(action[0] == 0)
         return 1;
       else
         return 0;
     }

     else if( (val1 == 0 && val2 == 1) || (val1 == 1 && val2 == 0) ){
       if(action[0] == 1)
         return 1;
       else
         return 0;
     }

     return 0; // should never happen
  }
}