/**
 * XOR Task
 * Author: Yizchak Naveh-Benjamin
 * Last modified: 7/1/2003
 *
 * This is a simple, bare-bones task designed to illustrate how the Clarion
 * package works.  Only the action-centered subsystem (ACS) is used, and only
 * the minimum set of functions is implemented.
 *
 * Since newcomers may find it hard to write a task from scratch, the code can
 * serve as a useful starting point for writing your own task code.  Just copy
 * this code, and modify it according to your needs.
 */
package tasksimulators.newsimu.XOR;
import clarion.common.*;
import clarion.nacs.*;
import clarion.tools.*;
import clarion.options.*;
import clarion.system.*;

public class TaskXOR extends Task implements SimuTask{
  final int NUMBER_OF_RUNS = 10000;

  // from the SimuTask interface
  public String getTaskName(){
    return "XOR Task";
  }

  // from the SimuTask interface
  public void start( int expMode, int blockNum, WriteFile wFile )
  {
  }

  // Constructor: initializes data structures, etc. for the task.
  // Should not be used to configure the agents' settings (e.g., learning rate)
  // since that is properly the function of init().
  public TaskXOR(){
    setRunningAgentNum(1);
  }

  // Loads the agent settings from the configuration files.
  // This method is automatically called (unless the user decides to configure
  // the agents interactively using the GUI).
  public void init()
  {
     // initialize the first (and only) agent with the "xorGroup1" settings
     String [] settings = {"xorGroup1"};
     super.settingNames = settings;

     // load options from config files; apply them to individual agents
     super.init();
  }

  // the starting point of the task
  public void start()
  {
    // Instantiate an agent, and initialize it using globals[0] (which will
    // contain the settings for xorGroup1).
    agents = new AgentXOR[1];
    agents[0] = new AgentXOR( globals[0] );

    // initialize the input and output to the agent
    short [][] inputs = { {1, 0}, {0, 1}, {0, 0}, {1, 1} };
    int [] outputs = new int[1];

    for(int ct = 0; ct < NUMBER_OF_RUNS; ct++)
      for(int i = 0; i < inputs.length; i++)
      {
        // set the agent's input at each step
        ( (AgentXOR) agents[0]).setSensoryInput(inputs[i]);
        // run a single learning pass on the agent
        ( (AgentXOR) agents[0]).run();
        // collect the agent's output for this step
        agents[0].getOutput(outputs);

        if( ct % 50 == 0 )
          System.out.println("Run " + ct + "; #" + i + ": " + outputs[0]);
      }

    // report the RER rules learned by the agent
    String [] rules = agents[0].getRerRules(0);
    for(int i = 0; i < rules.length; i++)
      System.out.println(rules[i]);
  }

}