// largely based on Chris Terry's CLARION code
package tasksimulators.oldsimu;
import java.util.Random;
import clarion.system.*;
import clarion.common.*;

public class BackPropNet
 {
    protected int nInput, nHidden, nOutput;
    protected double[] Input,Output,Hidden;
    protected double[] OutputMomentum,HiddenMomentum;

    protected double[] OutputThresholds,HiddenThresholds;
    //they get _added_ so most likely should be negative

    protected double[] OutputDeriv,HiddenDeriv;
    protected double[] Errors, HiddenErrors;
    protected double[] DesiredOutput;
    protected double[][] InputToHiddenWeights, HiddenToOutputWeights;
    protected double[][] ItoHWeights, HtoOWeights;
    protected double[][] InputToHiddenMomentum, HiddenToOutputMomentum;

    protected double  HINITWEIGHT = 0.01;    // brackets for the initial weights
    protected double  LINITWEIGHT = -0.01;

    protected double  HINITTHRESHOLD = 2;    // brackets for the initial weights
    protected double  LINITTHRESHOLD = -2;

    protected double Momentum;   // used for update of weights and
                                 // will be modified in controlsim.java
    protected double Eta;        // learning rate, will be modified
                                 // in controlsim.java

    protected double RZero = 0;          // tolerance of error...sorta
    protected double tolerance = 0.3;

    protected boolean noError = false;        // all errors are below tolerance
    protected boolean debug = false;
    protected double sumSqErrors = 0;

    public BackPropNet(int numInput, int numHidden, int numOutput )
    {
        // # of units in input, output, hidden layers
        nInput = numInput;
        nOutput = numOutput;
        nHidden = numHidden;

        // weight vectors
	       InputToHiddenWeights = new double[nInput][nHidden];
        HiddenToOutputWeights = new double[nHidden][nOutput];

        ItoHWeights = new double[nInput][nHidden];
        HtoOWeights = new double[nHidden][nOutput];

       	InputToHiddenMomentum = new double[nInput][nHidden];
        HiddenToOutputMomentum = new double[nHidden][nOutput];

        HiddenMomentum = new double[nHidden];
        OutputMomentum = new double[nOutput];

        // value of input, hidden and output vectors
       	Input = new double[nInput];
        Hidden = new double[nHidden];
        Output = new double[nOutput];

        // error terms in hidden and output layers
        HiddenErrors = new double[nHidden];
      	Errors = new double[nOutput];

        // bias in hidden and output layers
        HiddenThresholds = new double[nHidden];
        OutputThresholds = new double[nOutput];

        // derivative values of activation functions in
        // hidden and output layers
        HiddenDeriv = new double[nHidden];
        OutputDeriv = new double[nOutput];

        // target output
	DesiredOutput = new double[nOutput];

        double v1 = World.TASK_PARA_ARRAY[ World.NET_LRATE ];
        double v2 = World.TASK_PARA_ARRAY[ World.NET_MOMENTUM ];

        setLearningRate( v1 );
        setMomentum( v2 );

        // initialize weights.
	initWeights( 0 );
    }

    public BackPropNet( int netIdx, Global g )
    {
        // # of units in input, output, hidden layers
        nInput = g.getAmnInputNum( netIdx );
        nOutput = g.getAmnOutputNum( netIdx );
        nHidden = g.getAmnHiddenNum( netIdx );

        // weight vectors
       	InputToHiddenWeights = new double[nInput][nHidden];
        HiddenToOutputWeights = new double[nHidden][nOutput];

        ItoHWeights = new double[nInput][nHidden];
        HtoOWeights = new double[nHidden][nOutput];

       	InputToHiddenMomentum = new double[nInput][nHidden];
        HiddenToOutputMomentum = new double[nHidden][nOutput];

        HiddenMomentum = new double[nHidden];
        OutputMomentum = new double[nOutput];

        // value of input, hidden and output vectors
       	Input = new double[nInput];
        Hidden = new double[nHidden];
        Output = new double[nOutput];

        // error terms in hidden and output layers
        HiddenErrors = new double[nHidden];
      	Errors = new double[nOutput];

        // bias in hidden and output layers
        HiddenThresholds = new double[nHidden];
        OutputThresholds = new double[nOutput];

        // derivative values of activation functions in
        // hidden and output layers
        HiddenDeriv = new double[nHidden];
        OutputDeriv = new double[nOutput];

        // target output
        DesiredOutput = new double[nOutput];

        setLearningRate( g.getAmnBPLearningRate( netIdx ) );
        setMomentum( g.getAmnBPMomentum( netIdx ) );

        // initialize weights.
       	initWeights( 0 );
    }

    void initWeights( int mode )
    {
       	for (int i = 0; i < nInput; i++)
            for (int j = 0; j < nHidden; j++)
      		InputToHiddenMomentum[i][j] = Momentum;

       	for (int i = 0; i < nHidden; i++)
        {
	      //  HiddenThresholds[i] = randomThresholds();
                HiddenThresholds[i] = 1;
              		HiddenMomentum[i] = 0;
        }

       for (int i = 0; i < nHidden; i++)
        for (int j = 0; j < nOutput; j++)
                     HiddenToOutputMomentum[i][j] = Momentum;

       for (int i = 0; i < nOutput; i++)
        {
	    //OutputThresholds[i] = randomThresholds();
            OutputThresholds[i] = 1;
            OutputMomentum[i] = 0;
        }

        if( mode == 0 )
        {
            for( int i = 0; i < nInput; i++ )
                for ( int j = 0; j < nHidden; j++ )
                {
                   InputToHiddenWeights[i][j] = randomWeights();
                   ItoHWeights[i][j] = InputToHiddenWeights[i][j];
                }

            for (int i = 0; i < nHidden; i++)
                for (int j = 0; j < nOutput; j++)
                {
                     HiddenToOutputWeights[i][j] = randomWeights();
                     HtoOWeights[i][j] = HiddenToOutputWeights[i][j];
                }

        }
        else
        {
              for( int i = 0; i < nInput; i++ )
                  for ( int j = 0; j < nHidden; j++ )
                    InputToHiddenWeights[i][j] = ItoHWeights[i][j];

              for (int i = 0; i < nHidden; i++)
                  for (int j = 0; j < nOutput; j++)
                              HiddenToOutputWeights[i][j] = HtoOWeights[i][j];
        }
    }

    public void reinitWeights()
    {
         initWeights( 0 );
    }

    public void restoreInitWeights()
    {
         initWeights( 1 );
    }

    double randomWeights()
    {
        double sign,value;
        if ( java.lang.Math.random() < 0.5 )
             sign = LINITWEIGHT;
        else
             sign = HINITWEIGHT;

        value = java.lang.Math.random();
        return( value * sign );
    }

    public void adjustHOWeights( double ratio )
    {
        for (int i = 0; i < nOutput; i++)
        {
            OutputThresholds[i] /= ratio;
            OutputMomentum[i]  /= ratio;
        }

        for (int i = 0; i < nHidden; i++)
         for (int j = 0; j < nOutput; j++)
         {
             HiddenToOutputWeights[i][j] /= ratio;
             HiddenToOutputMomentum[i][j] /= ratio;
         }
    }

    double randomThresholds()
    {
        double sign,value;
        if ( java.lang.Math.random() < 0.5 )
             sign = LINITTHRESHOLD;
        else
             sign = HINITTHRESHOLD;

        value = java.lang.Math.random();
        return( value * sign );
    }

    public double[][] getInitInputToHiddenWeights( )
    {
        return ItoHWeights;
    }

    public double[][] getInitHiddenToOutputWeights()
    {
        return HtoOWeights;
    }

    public void setInitInputToHiddenWeights( double[][] w1 )
    {
        for( int i = 0; i < nInput; i++ )
          for ( int j = 0; j < nHidden; j++ )
          {
             InputToHiddenWeights[i][j] = w1[i][j];
             ItoHWeights[i][j] = InputToHiddenWeights[i][j];
          }
    }

    public void setInitHiddenToOutputWeights( double[][] w1 )
    {
        for( int i = 0; i < nHidden; i++ )
          for ( int j = 0; j < nOutput; j++ )
          {
             HiddenToOutputWeights[i][j] = w1[i][j];
             HtoOWeights[i][j] = HiddenToOutputWeights[i][j];
          }
    }

    public double[][] getInputToHiddenWeights()
    {
       return InputToHiddenWeights;
    }

    public double[][] getHiddenToOutputWeights()
    {
       return HiddenToOutputWeights;
    }

    public void getHiddenThresholds( double[] arr )
    {
       for( int i = 0; i < nOutput; i++ )
            arr[ i ] = HiddenThresholds[ i ];
    }

    public void getOutputThresholds( double[] arr )
    {
       for( int i = 0; i < nOutput; i++ )
            arr[ i ] = OutputThresholds[ i ];
    }

    // activation function
    double activation(double value){
          // can switch other activation functions here
          // simple sigmoid function
          return (1.0 / (1.0 + java.lang.Math.exp( -value)));
    }

    // derivative of the activation function
    double derivative( double value )
    {
          return( value * ( 1 - value ) );
    }

    // utility methods
    public void getInput( double[] fillMe )
    {
        for( int i = 0; i < nInput; i++ )
             fillMe[i] = Input[i];
    }

    public void getOutput( double[] out )
    {
        for( int i = 0; i < nOutput; i++ )
             out[ i ] = Output[i];
    }

    public double getOutput( int idx )
    {
       return Output[ idx ];
    }

    public void setInput( double[] newInputs )
    {
        for( int i = 0; i < nInput; i++ )
             Input[i] = newInputs[i];
    }

    public void setInput( int index, double value )
    {
        Input[index] = value;
    }

    public void setDesiredOutput( double[] dOutputs )
    {
	for( int i = 0; i < nOutput; i++ )
            DesiredOutput[i] = dOutputs[i];
    }

    public void setDesiredOutput( int index, double value )
    {
	DesiredOutput[index] = value;
    }

    public int getnInput()
    {
	return nInput;
    }
    public int getnHidden()
    {
	return nHidden;
    }
    public int getnOutput()
    {
	return nOutput;
    }

    public void getOutputErrors( double arr[] )
    {
        for( int i = 0; i < nOutput; i++ )
              arr[ i ] = Errors[ i ];
    }

    public void setMomentum(double value)
    {
        Momentum = value;
    }
    public void setTolerance(double value)
    {
        tolerance = value;
    }
    public void setRZero(double value)
    {
        RZero = value;
    }
    public void setDisplayDebug(boolean value)
    {
	debug = value;
    }
    public void setLearningRate(double value)
    {
	Eta = value;
    }
    public double getLearningRate()
    {
        return Eta;
    }
    public double getSumSqErrors()
    {
	return sumSqErrors;
    }

    public boolean noError()
    {
	return noError;
    }

    public void displayHidden()
    {
        System.out.print("Hidden unit activation: ");
	for(int i=0;i<nHidden;i++){
	    System.out.print(Hidden[i]+" ");
	}
        System.out.println();
    }

    public void displayOutput()
    {
        System.out.print("Output unit activation: ");
	for(int i=0;i<nOutput;i++){
	    System.out.print(Output[i]+" ");
	}
        System.out.println();
    }

    /********************* forward pass methods **************************/

    public void computeHiddenActivation()
    {
	double sum,v;

	for (int j = 0; j < nHidden; j++)
        {
		sum = HiddenThresholds[j];

		for ( int i = 0; i < nInput; i++ )
	             sum += Input [i] * InputToHiddenWeights[i][j];

		v = activation( sum );

		Hidden[j] = v;
		HiddenDeriv[j] = derivative(v);
	}
    }

    public void computeOutputActivation ()
    {
	double sum,v;

	for (int j = 0; j < nOutput; j++)
        {
		sum = OutputThresholds[j];

		for( int i = 0; i < nHidden; i++ )
	             sum += Hidden[i] * HiddenToOutputWeights[i][j];

		v = activation(sum);

		Output[j] = v;
		OutputDeriv[j] = derivative(v);
	}
    }

    public void forwardPass()
    {
	computeHiddenActivation();
        computeOutputActivation();
    }

    /****************** backward pass methods *****************************/

    // computer the errors of output layer
    public void computeErrors()
    {
        boolean satisfied = true;

	sumSqErrors = 0;

	for ( int i = 0; i < nOutput; i++ )
        {
	    double err = DesiredOutput[i] - Output[i];
	    Errors[i] = err * OutputDeriv[i];
	    sumSqErrors += err * err;
            if( java.lang.Math.abs( err ) > tolerance )
                satisfied = false;
        }

	sumSqErrors = sumSqErrors / 2;

        if ( debug )
             System.out.println(sumSqErrors);
        if ( satisfied )
             noError = true;
        else
             noError = false;
    }

    public void computeHiddenErrors ()
    {
	double h_err;

	for ( int i = 0; i < nHidden; i++ )
        {
		h_err = 0.0;

		for ( int j = 0; j < nOutput; j++ )
	              h_err += HiddenToOutputWeights[i][j] * Errors[j];

		HiddenErrors[i] = h_err * HiddenDeriv[i];
	}
    }

    // modify the weights of hidden to output layer
    public void modifyHiddenToOutput ()
    {
	double x;

	for ( int j = 0; j < nOutput; j++ )
        {
	    if ( java.lang.Math.abs( Errors[j] ) > RZero )
            {
		x = Errors[j];

                // udpate HiddenToOutputWeights
                for ( int i = 0; i < nHidden; i++ )
                {
	             if ( Momentum > RZero )
                     {
			/* momentum */
			HiddenToOutputMomentum[i][j] = Eta * x * Hidden[i] +
                                   Momentum * HiddenToOutputMomentum[i][j];
			HiddenToOutputWeights[i][j] += HiddenToOutputMomentum[i][j];
		     } else
		       	HiddenToOutputWeights[i][j] += Eta * x * Hidden[i];
               }

               // update OutputThresholds
               if ( Momentum > RZero )
               {
		    OutputMomentum[j] =  Eta * x + Momentum * OutputMomentum[j];
		    OutputThresholds[j] += OutputMomentum[j];
	       }
               else
		    OutputThresholds[j] += Eta * x;
	    }
	}
    }

    // modify the weights of input to hidden layer
    public void modifyInputToHidden ()
    {
	double x;

	for ( int j = 0; j < nHidden; j++ )
        {
	    if ( java.lang.Math.abs( HiddenErrors[j] ) > RZero )
            {
		x = HiddenErrors[j];

                // update the InputToHiddenWeights
		for (int i = 0; i < nInput; i++)
                {
		    if ( Momentum > RZero )
                    {
			InputToHiddenMomentum[i][j] = Eta * x* Input[i]
                                       + Momentum * InputToHiddenMomentum[i][j];
			InputToHiddenWeights[i][j] += InputToHiddenMomentum[i][j];
		    } else
			InputToHiddenWeights[i][j] += Eta * x * Input[i];
		}

                // upate the HiddenThresholds
		if (Momentum > RZero)
                {
		    HiddenMomentum[j] = Eta * x + Momentum * HiddenMomentum[j];
		    HiddenThresholds[j] += HiddenMomentum[j];
		} else
		    HiddenThresholds[j] += Eta * x;
	    }
	}
    }

    public void backwardPass()
    {
	//computeErrors();
        computeHiddenErrors();
        modifyHiddenToOutput();
        modifyInputToHidden();
    }

}

