package tasksimulators.oldsimu;
import java.util.Vector;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.coordinate.*;
import clarion.common.*;
import clarion.tools.*;

public class CIAgent{

   // debug flag.
   final boolean debug = true;

   // experiments variables.
   final int EXP_NUM = 4, EXP_ARG_NUM = 225;
   final int EXP_QUESTION_NUMS[] = { 108, 72, 75, 72 };
   final int EXP_GRP_QUEST_NUMS[][] =
             { { 1, 108 }, { 9, 8 }, { 3, 25 }, { 9, 8 } };
   final int ARG_TYPE_PAIRS[] = { 9, 9, 90, 9 };

   // used for combination of BL and TL.
   final int COMB_COMPONENT_NUM = 4;
   final int BL_IDX = 0, RER_IDX = 1, IRL_IDX = 2, FIX_IDX = 3;
   final int FIX_MODE = 0, VARIED_MODE = 1;
   final int AMN_IDX = 0, GKN_IDX = 1;
   // constant variables relevant to combination of the levels of NACS.
   //final double nacsBetaArr[][] = { { 0.11, 0.89 } };  // used for exp1
   //final double nacsBetaArr[][] = { { 0.15, 0.85 } };  // used for exp2
   // final double nacsBetaArr[][] = { { 0.2, 0.8 } };  // used for exp4
   //final double nacsBetaArr[][] = { { 0.5, 0.5 } };  // used for exp5
   final double nacsBetaArr[][] = { { 0.1, 0.9 } };  // used for exp5

   // used for the reasoning methods.
   final int RULE_BASED = 0, SIMILARITY_BASED = 1, RULE_SIM_BASED = 2;

   // constant variables relevant to arguments
   final int DOMAIN_NUM = 3, PHEN_NUM = 2, IDEN_ARGS_PER_DOMAIN = 3;
   final int KEY_PAIRS_PER_DOMAIN = 3, FILT_PAIRS_PER_DOMAIN = 30;

   // different types of arguements
   final int ARGU_TYPE_NUM = 4;
   final int INCL_SIMI_TYPE = 0, PREM_SPEC_TYPE = 1, FILT_TYPE = 2, IDEN_TYPE = 3;

   // different types of chunks.
   final int CHUNK_TYPE_NUM = 3;
   final int FACT_PREM_IDX = 0, FACT_CONC_IDX = 1, ARGU_PREM_IDX = 2;

   // constant variables relevant to chunk.
   final int CHUNK_ROW_NUM = 20, CHUNK_COLUM_NUM = 20;
   final int CHUNK_DIM_NUM = 20, CHUNK_VAL_NUM = 30;
   final int    MAX_NODE_AGE = 200;
   final double CHUNK_ENCODE_PROB = 1.0;
   final double CHUNK_EXTRACT_THRESHOLD = 0.2;
   final double CHUNK_EXT_SUCC_PROB = 0.1;

   // constant variables relevant to associative rules.
   final int    MAX_LINK_AGE = 200;
   final double ASSOC_ENCODE_PROB = 1.0;
   final double ASSOC_EXTRACT_THRESHOLD = 0.2;
   final double ASSOC_EXT_SUCC_PROB = 0.1;

   // cnstant variables of thresholds.
   final double ACTIVATION_THRESHOLD = 0.5;
   final double STRENTH_THRESHOLD = 0.2;
   final double SIMILARITY_THRESHOLD = 0.2;

   // constant variables relevant to QBackPropNet.
   //final int IDNET_INPUT_NUM = 1800, IDNET_HIDDEN_NUM = 200, IDNET_OUTPUT_NUM = 117;
   final int IDNET_INPUT_NUM = 1808, IDNET_HIDDEN_NUM = 100, IDNET_OUTPUT_NUM = 29;
   final int IDN_SI_IDX = 0, IDN_WM_IDX = 1800, IDN_GS_IDX = 1807;
   final int IDN_NACS_IDX = 0, IDN_INFER_IDX = 1, IDN_REPORT_IDX = 12;

   // constant variables relevant to QBackPropNet.
   final int AMNET_INPUT_NUM = 1800, AMNET_HIDDEN_NUM = 200, AMNET_OUTPUT_NUM = 117;
   double LEARNING_RATE, MOMENTUM;

   // constant variables for reasoning.
   final double C14_15_ARR[][] = {
                { 0.5, 1.0 }, { 0.5, 1.0 }, { 1.0, 1.0 }, { 0.88, 1.0 } };
   final double SIM_THRESHOLD = 0.2;

   // used for WM.
   final int MAX_WM_ITEM_NUM = 7;

   // current experiment.
   int expNo;

   // used for setting up GKNET.
   int reasoningType = RULE_SIM_BASED;
   int chunkDVNums[], chunkDimTypes[];
   double INIT_BLA = 0.0;

   // used for combination of levels.
   int actionTypeNum;
   int acsChosenComp, nacsChosenComp;
   double acsBetaArr[][];

   // # of values of each dimension used in QBP net.
   int acsQbpDimValNums[], nacsQbpDimValNums[];

   // arguments variables.
   short arguments[][][][][], arguIndices[][][], argGroupOrder[], conclusion[];

   // actions from IDN or ARN.
   short action[], actionBL[], actionTL[];

   // indices of conclusion chunks in GKN.
   short concChunkIndices[];

   // if a report is done.
   boolean REPORT_FLAG = false;

   // used for outputs of QBP net.
   double strengthsBL[], strengthsTL[], strengths[];

   Vector answerList;
   GKNet gkNet;
   AMNet amNet;
   IDNet idNet;
   ExWM wm;
   StochasticComb acsStochCombiner, nacsStochCombiner;
   StochasticDecider actionDecider;  // action decider

   public CIAgent()
   {
      // initialize conclusion from AMNET.
      conclusion = new short[ AMNET_OUTPUT_NUM ];
      // initialize indices of conclusion chunks in GKN.
      concChunkIndices = new short[ AMNET_OUTPUT_NUM ];
      // initialize actions of IDN and ARN.
      actionTL = new short[ IDNET_OUTPUT_NUM ];
      actionBL = new short[ IDNET_OUTPUT_NUM ];

      // initialize strengths of the chosen chunks.
      strengths = new double[ 2 ];
      strengthsBL = new double[ 2 ];
      strengthsTL = new double[ 2 ];

      // answers to the questions.
      answerList = new Vector( EXP_NUM );
      for( int i = 0; i < EXP_NUM; i++ )
      {
           if( i == 0 )
           {
               short[] elem = new short[ EXP_QUESTION_NUMS[i] ];
               answerList.addElement( elem );
           }
           else
           {
               double[][] elem = new double[ EXP_QUESTION_NUMS[i] ][ 2 ];
               answerList.addElement( elem );
           }
      }

      // combination of two levels of ACS.
      actionTypeNum = 1;
      acsBetaArr = new double[ actionTypeNum ][ COMB_COMPONENT_NUM ];
      for( int i = 0; i < COMB_COMPONENT_NUM; i++ )
           acsBetaArr[ 0 ][ i ] = World.TASK_PARA_ARRAY[ World.WBL + i ];
      acsStochCombiner = new StochasticComb( FIX_MODE, COMB_COMPONENT_NUM, acsBetaArr[0] );

      // combination of two levels of NACS.
      nacsStochCombiner = new StochasticComb( FIX_MODE, 2, nacsBetaArr[0]  );

      // initialize action decider.
      actionDecider = new StochasticDecider( World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );

      LEARNING_RATE = World.TASK_PARA_ARRAY[ World.NET_LRATE ];
      MOMENTUM =  World.TASK_PARA_ARRAY[ World.NET_MOMENTUM ];

      nacsQbpDimValNums = new int[ 3 * CHUNK_DIM_NUM ];
      for( int i = 0; i < nacsQbpDimValNums.length; i++ )
           nacsQbpDimValNums[ i ] = CHUNK_VAL_NUM;
      amNet = new AMNet( AMNET_INPUT_NUM, AMNET_HIDDEN_NUM, AMNET_OUTPUT_NUM,
                           3*CHUNK_DIM_NUM, nacsQbpDimValNums );
      amNet.setLearningRate( LEARNING_RATE );
      amNet.setMomentum( MOMENTUM );

      acsQbpDimValNums = new int[ 3 * CHUNK_DIM_NUM ];
      for( int i = 0; i < acsQbpDimValNums.length; i++ )
           acsQbpDimValNums[ i ] = CHUNK_VAL_NUM;
      idNet = new IDNet( IDNET_INPUT_NUM, IDNET_HIDDEN_NUM, IDNET_OUTPUT_NUM,
                           3*CHUNK_DIM_NUM, acsQbpDimValNums );
      idNet.setLearningRate( LEARNING_RATE );
      idNet.setMomentum( MOMENTUM );

      // initial working memory.
      wm = new ExWM( MAX_WM_ITEM_NUM );
   }

   public void initGKNet()
   {
      chunkDVNums = new int[ CHUNK_DIM_NUM ];
      for( int i = 0; i < CHUNK_DIM_NUM; i++ )
           chunkDVNums[ i ] = CHUNK_VAL_NUM;
      chunkDimTypes = new int[ CHUNK_DIM_NUM ];
      for( int i = 0; i < CHUNK_DIM_NUM; i++ )
           chunkDimTypes[ i ] = 0;
      gkNet = new GKNet( CHUNK_ROW_NUM, CHUNK_COLUM_NUM, CHUNK_DIM_NUM,
                         reasoningType, chunkDVNums, chunkDimTypes, INIT_BLA );
      gkNet.setMaxAges( MAX_NODE_AGE, MAX_LINK_AGE );
      gkNet.setActivationThreshold( ACTIVATION_THRESHOLD );
      gkNet.setStrengthThreshold( STRENTH_THRESHOLD );
      gkNet.setSimilarityThreshold( SIMILARITY_THRESHOLD );
   }

   public void reinit()
   {
      amNet.reinitWeights();
      idNet.reinitWeights();
   }

/***************************** GET METHODS ***************************************/
   public void getAnswers( int expIdx, short[] arr )
   {
       for( int i = 0; i < arr.length; i++ )
            arr[ i ] = ( (short[])answerList.elementAt(0) )[ i ];
   }

   public void getAnswers( int expIdx, double[][] arr )
   {
       for( int i = 0; i < arr.length; i++ )
        for( int j = 0; j < arr[ i ].length; j++ )
             arr[ i ][ j ] = ( (double[][])answerList.elementAt(expIdx) )[ i ][ j ];
   }

/***************************** SET METHODS ***************************************/

   public void setExpIdx( int eNo )
   {
       expNo = eNo;
   }

   public void setQuestGroupOrder( short[] ord )
   {
      argGroupOrder = new short[ ord.length ];
      for( int i = 0; i < ord.length; i++ )
           argGroupOrder[ i ] = ord[ i ];
   }

   public void setArgIndices( short[][][] indices )
   {
     // arguIndices = (short[][][]) indices.clone();
     arguIndices = new short[ indices.length ][][];
     for( int i = 0; i < indices.length; i++ )
     {
          arguIndices[ i ] = new short[ indices[ i ].length ][];
          for( int j = 0; j < indices[ i ].length; j++ )
          {
               arguIndices[ i ][ j ] = new short[ indices[ i ][ j ].length ];
               for( int k = 0; k < indices[ i ][ j ].length; k++ )
                    arguIndices[ i ][ j ][ k ] = indices[ i ][ j ][ k ];
          }
     }
   }

   public void setArguments( short[][][][][] args )
   {
    //  arguments = (short[][][][][])args.clone();
     arguments = new short[ args.length ][][][][];
     for( int i = 0; i < args.length; i++ )
     {
          arguments[ i ] = new short[ args[ i ].length ][][][];
          for( int j = 0; j < args[ i ].length; j++ )
          {
               arguments[ i ][ j ] = new short[ args[ i ][ j ].length ][][];
               for( int k = 0; k < args[ i ][ j ].length; k++ )
               {
                    arguments[ i ][ j ][ k ] = new short[ args[ i ][ j ][ k ].length ][];
                    for( int m = 0; m < args[ i ][ j ][ k ].length; m++ )
                    {
                         arguments[ i ][ j ][ k ][ m ] = new short[ args[ i ][ j ][ k ][ m ].length ];
                         for( int n = 0; n < args[ i ][ j ][ k ][ m ].length; n++ )
                              arguments[ i ][ j ][ k ][ m ][ n ] = args[ i ][ j ][ k ][ m ][ n ];
                    }
               }
          }
     }
   }

   public void setupGKN()
   {
      // SET UP GKNET.
      int indices[] = new int[ CHUNK_DIM_NUM ];
      for( int i = 0; i < CHUNK_DIM_NUM; i++ )
           indices[ i ] = i;
      gkNet.setCueDimIndices( indices );

      int count = 0;
      //final int FACT_PREM_IDX = 0, FACT_CONC_IDX = 1, ARGU_PREM_IDX = 2;
      // create inclusion-similarity chunks.
      for( int i = 0; i < DOMAIN_NUM; i++ )
       for( int j = 0; j < KEY_PAIRS_PER_DOMAIN; j++ )
       {
            int idx = 2 * ( i * KEY_PAIRS_PER_DOMAIN + j );
            int from = gkNet.insertNode( arguments[ INCL_SIMI_TYPE ][ idx ][ FACT_PREM_IDX ], "chunk" );
            int to = gkNet.insertNode( arguments[ INCL_SIMI_TYPE ][ idx ][ FACT_CONC_IDX ], "chunk" );
            // create inclusion-similarity links.
            gkNet.insertLink( (short)from, (short)to );
            concChunkIndices[ count++ ] = (short)to;
            if( debug )
            {
            }
       }

      //final int FACT_PREM_IDX = 0, FACT_CONC_IDX = 1, ARGU_PREM_IDX = 2;
      // create premise-specificity chunks.
      for( int i = 0; i < DOMAIN_NUM; i++ )
       for( int j = 0; j < KEY_PAIRS_PER_DOMAIN; j++ )
        {
            int idx = 2 * ( i * KEY_PAIRS_PER_DOMAIN + j );
            int from1 = gkNet.insertNode( arguments[ PREM_SPEC_TYPE ][ idx ][ FACT_PREM_IDX ], "chunk" );
            int to1 = gkNet.insertNode( arguments[ PREM_SPEC_TYPE ][ idx ][ FACT_CONC_IDX ], "chunk" );
            // create premise-specificity links.
            gkNet.insertLink( (short)from1, (short)to1 );
            int from2 = gkNet.insertNode( arguments[ PREM_SPEC_TYPE ][ idx + 1 ][ FACT_PREM_IDX ], "chunk" );
            int to2 = to1;
            // create premise-specificity links.
            gkNet.insertLink( (short)from2, (short)to2 );
            concChunkIndices[ count++ ] = (short)to1;
            if( debug )
            {
            }
        }

      //final int FACT_PREM_IDX = 0, FACT_CONC_IDX = 1, ARGU_PREM_IDX = 2;
      // create filter chunks.
      for( int i = 0; i < DOMAIN_NUM; i++ )
       for( int j = 0; j < FILT_PAIRS_PER_DOMAIN; j++ )
        {
            int idx = 2 * ( i * FILT_PAIRS_PER_DOMAIN + j );
            int from = gkNet.insertNode( arguments[ FILT_TYPE ][ idx ][ FACT_PREM_IDX ], "chunk" );
            int to = gkNet.insertNode( arguments[ FILT_TYPE ][ idx ][ FACT_CONC_IDX ], "chunk" );
            // create filter links.
            gkNet.insertLink( (short)from, (short)to );
            from = gkNet.insertNode( arguments[ FILT_TYPE ][ idx + 1 ][  FACT_PREM_IDX ], "chunk" );
            // create filter links.
            gkNet.insertLink( (short)from, (short)to );
            concChunkIndices[ count++ ] = (short)to;
        }

      //final int FACT_PREM_IDX = 0, FACT_CONC_IDX = 1, ARGU_PREM_IDX = 2;
      // create identity chunks.
      for( int i = 0; i < DOMAIN_NUM; i++ )
       for( int j = 0; j < IDEN_ARGS_PER_DOMAIN; j++ )
        {
            int idx = i * IDEN_ARGS_PER_DOMAIN + j;
            int from = gkNet.insertNode( arguments[ IDEN_TYPE ][ idx ][ FACT_PREM_IDX ], "chunk" );
            int to = gkNet.insertNode( arguments[ IDEN_TYPE ][ idx ][ FACT_CONC_IDX ], "chunk" );
            // create identity links.
            gkNet.insertLink( (short)from, (short)to );
            concChunkIndices[ count++ ] = (short)to;
        }

      gkNet.setConclusionIndices( concChunkIndices );
      if( debug )
      {
          //System.out.println( gkNet.showLinks() );
      }

      // train AMN
      //trainBL();
   }

/***************************** MAIN METHOD ***************************************/
   public void answerForceQuestions( )
   {
       // train AMN
       //trainBL();

       gkNet.setReasoningParas( C14_15_ARR[ 0 ] );

       int questIdx = 0;
       for( int i = 0; i < EXP_GRP_QUEST_NUMS[ 0 ][ 0 ]; i++ )
         for( int j = 0; j < EXP_GRP_QUEST_NUMS[ 0 ][ 1 ]; j++ )
         {
              double[] answer = new double[ 1 ];
              doQuestion( questIdx, answer );
              ( (short[])answerList.elementAt(0) )[questIdx] = (short)answer[0];
              questIdx++;
         }
   }

   public void answerRatingQuestions()
   {
       // train AMN
       //trainBL();

       gkNet.setReasoningParas( C14_15_ARR[ expNo ] );

       int questIdx = 0;
       for( int i = 0; i < EXP_GRP_QUEST_NUMS[ expNo ][ 0 ]; i++ )
         for( int j = 0; j < EXP_GRP_QUEST_NUMS[ expNo ][ 1 ]; j++ )
         {
              doQuestion( questIdx, ((double[][])answerList.elementAt( expNo ))[questIdx] );
              questIdx++;
         }
   }

   private void doQuestion( int questIdx, double[] result )
   {
       int[] argInfo = new int[ 3 ];
       getArgumentInfo( questIdx, argInfo );
       int argType = argInfo[ 0 ];
       int argIdx = argInfo[ 1 ];
       int absIdx = argInfo[ 2 ];
       int NUM_PER_PAIR = 2;
       if( argType == IDEN_TYPE )
           NUM_PER_PAIR = 1;

/******************************************************************************
       // train an association in AMN.
       for( int i = 0; i < NUM_PER_PAIR; i++ )
       {
            int offset = NUM_PER_PAIR * argIdx + i;
            trainAMN( argType, offset, absIdx );
       }
/******************************************************************************/

       //wm.resetActivations();
       wm.actionResetAll();
       REPORT_FLAG = false;

       acsChosenComp = acsStochCombiner.selectComponent();
       nacsChosenComp = nacsStochCombiner.selectComponent();

       while( !REPORT_FLAG )
       {
          for( int i = 0; i < NUM_PER_PAIR; i++ )
          {
               int offset = NUM_PER_PAIR * argIdx + i;
               // get action from the IDN.
               short condition[] = new short[ IDNET_INPUT_NUM ];
               setIDNInputs( argType, offset, condition );
               idNet.setInput( condition );
               idNet.forwardPass();
               idNet.getOutput( actionBL );
               // get action from the ARN.
               formActionTL( condition );
               boolean right = compareActions( );
               if( acsChosenComp == BL_IDX )
               {   // controlled by IDN.
                   if( right )
                   {
                       action = actionBL;
                       if( actionBL[ IDN_INFER_IDX ] == 1 )
                       {
                           doInference( argType, offset, i, absIdx );
                           trainAMN();
                           int loc = 0;
                           if( actionBL[ IDN_INFER_IDX + 5 ] == 1 )
                               loc = 1;
                           wm.setActivation( loc, strengths[ i ] );
                       }
                       else
                           doReporting( result );
                   }
                   else
                   {
                      strengths[ i ] = 0.5;
                      if( i == NUM_PER_PAIR )
                          doReporting( result );
                   }
               }
               else
               {   // controlled by ARN.
                   action = actionTL;
                   if( actionTL[ IDN_INFER_IDX ] == 1 )
                   {
                       doInference( argType, offset, i, absIdx );
                       trainAMN();
                       if( actionTL[ IDN_INFER_IDX + 4 ] == 1 )
                           wm.setActivation( 0, strengths[ i ] );
                       else
                           wm.setActivation( 1, strengths[ i ] );
                   }
                   else
                       doReporting( result );
               }
/******************************************************************************/
               trainIDN( );
/******************************************************************************/
          }
       }
   }

   private void getArgumentInfo( int questIdx, int[] info )
   {
        int TYPE = 0, IDX = 1, ABS_IDX = 2;
        // get the argument's type and its index.
        int orderIdx = questIdx / EXP_GRP_QUEST_NUMS[ expNo ][ 1 ];
        int offset = questIdx % EXP_GRP_QUEST_NUMS[ expNo ][ 1 ];
        int groupIdx = argGroupOrder[ orderIdx ];

        info[ TYPE ] = arguIndices[ groupIdx ][ offset ][ 0 ];
        info[ IDX ] = arguIndices[ groupIdx ][ offset ][ 1 ];

        // calculate the absolute index of the argument among all the arguments
        info[ ABS_IDX ] = 0;
        for( int i = 0; i < info[ TYPE ]; i++ )
             info[ ABS_IDX ] += ARG_TYPE_PAIRS[ i ];
        info[ ABS_IDX ] += info[ IDX ];
   }

   private void setIDNInputs( int argType, int offset, short[] input )
   {
      int idx = 0;
      short sensory[][][] = arguments[ argType ][ offset ];
      for( int i = 0; i < sensory.length; i++ )
        for( int j = 0; j < sensory[ i ].length; j++ )
          for( int k = 0; k < sensory[ i ][ j ].length; k++ )
               input[ idx++ ] = sensory[ i ][ j ][ k ];

      for( int i = 0; i < MAX_WM_ITEM_NUM; i++ )
           input[ IDN_WM_IDX + i ] = (short)wm.isOccupied( i );

      if( expNo == 0 )  // choice task.
          input[ IDN_GS_IDX ] = 1;
      else              // rating task.
          input[ IDN_GS_IDX ] = 0;
   }

   private void formActionTL( short[] condition )
   {
      // IDN_NACS_IDX = 0, IDN_INFER_IDX = 1, IDN_REPORT_IDX = 12;
      for( int i = 0; i < actionTL.length; i++ )
           actionTL[ i ] = 0;

      // set START NACS to YES.
      actionTL[ IDN_NACS_IDX ] = 1;

      if( condition[ IDN_WM_IDX ] == 0 || condition[ IDN_WM_IDX+1 ] == 0 )
      {   // inference action.
          actionTL[ IDN_INFER_IDX ] = 1;
          // use the conclusion of the argument as conclusion of reasoning.
          actionTL[ IDN_INFER_IDX + 1 ] = 1;
          // use the premise P2 of the argument as cue of reasoning.
          actionTL[ IDN_INFER_IDX + 3 ] = 1;
          // use the empty item in WM for storing activation.
          if( condition[ IDN_WM_IDX ] == 0 )
              actionTL[ IDN_INFER_IDX + 4 ] = 1;
          else
              actionTL[ IDN_INFER_IDX + 5 ] = 1;
      }
      else
      {   // reporting action.
          actionTL[ IDN_REPORT_IDX ] = 1;
          // the item in WM in which the first activation is stored.
          actionTL[ IDN_REPORT_IDX + 1 ] = 1;
          // the item in WM in which the second activation is stored.
          actionTL[ IDN_REPORT_IDX + 9 ] = 1;
          // output bigger or both
          if( condition[ IDN_GS_IDX ] == 1 )
              actionTL[ IDN_REPORT_IDX + 15 ] = 1;
          else
              actionTL[ IDN_REPORT_IDX + 16 ] = 1;
      }
   }

   private boolean compareActions( )
   {
      /*
      // START NACS = NO
      if( actionBL[ IDN_NACS_IDX ] == 0 )
          return false;

      // neithor INFERENCE nor REPORING action is set.
      if( actionBL[ IDN_INFER_IDX ] == 0 && actionBL[ IDN_REPORT_IDX ] == 0 )
          return false;

      // both INFERENCE and REPORING action are set.
      if( actionBL[ IDN_INFER_IDX ] == 1 && actionBL[ IDN_REPORT_IDX ] == 1 )
          return false;

      return true;
      */

     for( int i = 0; i < actionBL.length; i++ )
          if( actionBL[ i ] != actionTL[ i ] )
              return false;
     return true;
   }

   private void doInference( int argType, int offset, int idxInpair, int absIdx )
   {
      // get strengths of the conclusion from the AMN.
      short condition[][][] = arguments[ argType ][ offset ];
      amNet.setInput( condition );
      amNet.forwardPass();
      strengthsBL[ idxInpair ] = amNet.getOutput( absIdx );

      // get strengths of the conclusion from GKN.
      gkNet.setRetrievalCue( arguments[ argType ][ offset ] [ ARGU_PREM_IDX ] );
      //gkNet.reasoningBetweenNode();
      boolean retVal = gkNet.reasoningBetweenNode(
                       arguments[ argType ][ offset ][ FACT_PREM_IDX ],
                       arguments[ argType ][ offset ][ FACT_CONC_IDX ] );
      strengthsTL[ idxInpair ] = gkNet.getChosenChunkStrength();

      if( nacsChosenComp == AMN_IDX )
          strengths[ idxInpair ] = strengthsBL[ idxInpair ];
      else
          strengths[ idxInpair ] = strengthsTL[ idxInpair ];

      if( debug )
      {
          if( argType == 0 || argType == 1 )
          {
              //if( nacsChosenComp == AMN_IDX )
              //{
              String strArr1[] = { "IDN", "RER", "IRL", "ARN" };
              String strArr2[] = { "AMN", "GKN" };
              System.out.print( "ACS = " + strArr1[acsChosenComp] );
              System.out.print( ", NACS = " + strArr2[nacsChosenComp] );
              System.out.print( ", absIdx = " + absIdx );
              System.out.print( ", type = " + argType );
              System.out.print( ", idxInpair = " + idxInpair );
              System.out.print( ", strengthBL = " + strengthsBL[ idxInpair ] );
              System.out.println( ", strengthTL = " + strengthsTL[ idxInpair ] );
              //}
          }
      }
   }

   private void doReporting( double[] result )
   {
      REPORT_FLAG = true;

      // output both.
      if( actionTL[ IDN_REPORT_IDX + 16 ] == 1 )
      {
          result[ 0 ] = wm.getActivation( 0 );
          result[ 1 ] = wm.getActivation( 1 );
          return;
      }

      strengths[ 0 ] = wm.getActivation( 0 );
      strengths[ 1 ] = wm.getActivation( 1 );

      /*
      if( strengths[ 0 ] >= strengths[ 1 ] )
          result[ 0 ] = 0;
      else
          result[ 0 ] = 1;
      */
      if( strengths[ 0 ] > strengths[ 1 ] )
          result[ 0 ] = 0;
      else if( strengths[ 0 ] < strengths[ 1 ] )
               result[ 0 ]  = 1;
           else
           {
               if( Math.random() >= 0.5 )
                   result[ 0 ]  = 0;
               else
                   result[ 0 ]  = 1;
           }
   }

   private void trainAMN( int argType, int offset, int absIdx )
   {
      // get strengths of the conclusion from the NACS's bottom level
      short condition[][][] = arguments[ argType ][ offset ];
      amNet.setInput( condition );
      amNet.forwardPass();

      double activations[] = new double[ AMNET_OUTPUT_NUM ];
      //for( int i = 0; i< activations.length; i++ )
       //    activations[ i ] = 0.0;
      amNet.getOutput( activations );
      activations[ absIdx ] = 1.0;
      amNet.setDesiredOutput( activations );
      amNet.backwardPass();
   }

   private void trainIDN( )
   {
      idNet.setDesiredOutput( actionTL );
      idNet.computeErrors();
      idNet.backwardPass();
   }

   private void trainAMN( )
   {
      double activations[] = new double[ AMNET_OUTPUT_NUM ];
      gkNet.getConclusionStrengths( activations );
      amNet.setDesiredOutput( activations );
      amNet.backwardPass();
   }

   private void trainBL()
   {
      for( int times = 0; times < 1; times++ )
      {
           System.out.println( times );
           Vector seeds1 = new Vector( EXP_ARG_NUM );
           for( int i = 0; i < arguments.length; i++ )
            for( int j = 0; j < arguments[ i ].length; j++ )
            {
                 int[] elem = new int[ 2 ];
                 elem[ 0 ] = i;
                 elem[ 1 ] = j;
                 seeds1.addElement( elem );
            }

           // start offset of each argument type in the ACS's BL output.
           int[] offsets = new int[ ARGU_TYPE_NUM ];
           offsets[ 0 ] = 0;
           for( int i = 1; i < ARGU_TYPE_NUM; i++ )
                offsets[ i ] = offsets[ i - 1 ] + ARG_TYPE_PAIRS[ i - 1 ];

           // training the ACS's BL.
           for( int i = 0; i < EXP_ARG_NUM; i++ )
           {
                int index = (int)( seeds1.size() * Math.random() );
                int argType = ((int[])seeds1.elementAt( index ))[0];
                int argIdx =  ((int[])seeds1.elementAt( index ))[1];

                short condition[][][] = arguments[ argType ][ argIdx ];
                amNet.setInput( condition );
                amNet.forwardPass();

                int idx = 0;
                if( argType < IDEN_TYPE )
                    idx = offsets[ argType ] + argIdx / 2;
                else
                    idx = offsets[ argType ] + argIdx;
                double activations[] = new double[ AMNET_OUTPUT_NUM ];
                amNet.getOutput( activations );
                activations[ idx ] = 1.0;

                amNet.setDesiredOutput( activations );
                amNet.backwardPass();

               seeds1.removeElementAt( index );
           }
       }
  }
}