package tasksimulators.oldsimu;
import java.awt.*;
import java.awt.Graphics;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.Vector;
import java.util.Arrays;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class CITask extends JFrame implements SimuTask {

   // debug flag.
   final boolean debug = true;

   // the user interface
   final int WINDOW_WIDTH = 400, WINDOW_HEIGHT = 300;

   // experiments variables.
   final int EXP_NUM = 4;
   //final int EXP_AGENT_NUMS[] = { 73, 18, 27, 18 };
   final int EXP_AGENT_NUMS[] = { 73, 18, 27, 18 };
   //final int EXP_ARG_NUMS[] = { 216, 144, 141, 144 };
   //final int EXP_CHUNK_NUMS[] = { 522, 342, 282, 342 };
   final int EXP_QUEST_NUMS[] = { 108, 72, 75, 72 };

   final int GRP_QUEST_NUMS[][] =
             { { 1, 108 }, { 9, 8 }, { 3, 25 }, { 9, 8 } };
   final int GRP_QUEST_TYPE_PAIRS[][] =
             { { 9, 9, 90, 0 }, { 1, 1, 6, 0 },
               { 3, 3, 16, 3 }, { 1, 1, 6, 0 } };

   // # of pairs of arguments with different types.
   final int ARG_TYPE_PAIRS[] = { 9, 9, 90, 9 };

   // # of arguments with different types.
   final int ARG_TYPE_NUMS[] = { 18, 18, 180, 9 };

   // constant variables relevant to arguments
   final int DOMAIN_NUM = 3, PHEN_NUM = 2, IDEN_ARGS_PER_DOMAIN = 3;
   final int KEY_PAIRS_PER_DOMAIN = 3, FILT_PAIRS_PER_DOMAIN = 30;
   final int KEY_ARG_PAIRS = KEY_PAIRS_PER_DOMAIN * DOMAIN_NUM;

   // different types of arguements
   final int ARGU_TYPE_NUM = 4;
   final int INCL_SIMI_TYPE = 0, PREM_SPEC_TYPE = 1, FILT_TYPE = 2, IDEN_TYPE = 3;

   // different types of chunks.
   final int CHUNK_TYPE_NUM = 3;
   final int FACT_PREM_IDX = 0, FACT_CONC_IDX = 1, ARGU_PREM_IDX = 2;

   // constant variables relevant to chunk.
   //final int CHUNK_ROW_NUM = 20, CHUNK_COLUM_NUM = 20;
   final int CHUNK_DIM_NUM = 20, CHUNK_VAL_NUM = 30;
   final int INTERVAL = CHUNK_VAL_NUM / DOMAIN_NUM;

   final int LESS_SIM_TYPE = 0, MID_SIM_TYPE = 1, MORE_SIM_TYPE = 2;
   final double[][][] SIMI_VALS =
         {
             //{ { 0.84, 0.86 }, { 0.86, 0.87 }, { 0.92, 0.98 } },
             //{ { 0.84, 0.86 }, { 0.86, 0.87 }, { 0.92, 0.98 } },
             //{ { 0.84, 0.86 }, { 0.86, 0.87 }, { 0.92, 0.98 } }
             { { 0.75, 0.85 }, { 0.82, 0.92 }, { 0.92, 0.98 } },
             { { 0.75, 0.85 }, { 0.82, 0.92 }, { 0.92, 0.98 } },
             { { 0.75, 0.85 }, { 0.82, 0.92 }, { 0.92, 0.98 } }
         };

   short groupOrder[], arguments[][][][][], arguIndices[][][];
   boolean exitFlag;

   // variables for saving data.
   Vector answerList, agentResultList, groupResultList, finalResultList;

   // used for user interface.
   Container c;
   JFrame mainFrame;
   TextArea arguArea1, resultArea1, showArea1;
   JButton beginButton, exitButton;

   WriteFile datFile, expDatFiles[];
   CIAgent ciAgent;

   public CITask( JFrame mFrame )
   {
      // initialize user interface.
      initInterface( mFrame );

      // initialize arguments.
      arguments = new short[ ARGU_TYPE_NUM ][][][][];

      // initialize the variable relevant to the answers to the questions.
      answerList = new Vector( EXP_NUM );
      for( int i = 0; i < EXP_NUM; i++ )
      {
           if( i == 0 )
           {
               short[] elem = new short[ EXP_QUEST_NUMS[i] ];
               answerList.addElement( elem );
           }
           else
           {
               double[][] elem = new double[ EXP_QUEST_NUMS[i] ][ 2 ];
               answerList.addElement( elem );
           }
      }

      // initialize the variable of agent's results of the experiments.
      agentResultList = new Vector( EXP_NUM );
      for( int i = 0; i < EXP_NUM; i++ )
      {
           if( i == 0 )
           {
               short[][][] elem = new short[ 2 ][ KEY_ARG_PAIRS ][ 2 ];
               agentResultList.addElement( elem );
           }
           else
           {
               double[][][] elem = new double[ 2 ][ KEY_ARG_PAIRS ][ 2 ];
               agentResultList.addElement( elem );
           }
      }

      // initialize the variable of final results of the experiments.
      finalResultList = new Vector( EXP_NUM );
      for( int i = 0; i < EXP_NUM; i++ )
      {
           if( i == 0 )
           {
               short[][][] elem = new short[ 2 ][ KEY_ARG_PAIRS ][ 2 ];
               finalResultList.addElement( elem );
           }
           else
           {
               double[][][] elem = new double[ 2 ][ KEY_ARG_PAIRS ][ 2 ];
               finalResultList.addElement( elem );
           }
      }

      // initialize the variable of group's results of the experiments.
      groupResultList = new Vector( EXP_NUM );
      for( int i = 0; i < EXP_NUM; i++ )
      {
           if( i == 0 )
           {
               short[][][][] elem = new short[ EXP_AGENT_NUMS[i] ][ 2 ][ KEY_ARG_PAIRS ][ 2 ];
               groupResultList.addElement( elem );
           }
           else
           {
               double[][][][] elem = new double[ EXP_AGENT_NUMS[i] ][ 2 ][ KEY_ARG_PAIRS ][ 2 ];
               groupResultList.addElement( elem );
           }
      }
   }

   public String getTaskName()
   {
      return "CATEGORY INFERENCE TASK";
   }

   public void start( int expMode, int blockNum, WriteFile dFile )
   {
       datFile = dFile;
       setEnabled( true );
       setVisible( true );

       // create the files for saving data.
       createExpDataFiles();

       ciAgent = new CIAgent();
       ciAgent.initGKNet();
       createArguments();
       //ciAgent.setArguments( incSimArgs, preSpecArgs, filterArgs, idenArgs );
       ciAgent.setArguments( arguments );
       ciAgent.setupGKN();
   }

/***************************** PRIVATE METHOD ***************************************/
  private void initInterface( JFrame mFrame )
   {
      // setup user interface.
      c = getContentPane();
      c.removeAll();
      c.setLayout( new BorderLayout( 5, 0 ) );
      buildInterface();

      setSize( WINDOW_WIDTH, WINDOW_HEIGHT );
      setEnabled( false );
      setVisible( false );

      mainFrame = mFrame;
      addWindowListener(
          new WindowAdapter(){
             public void windowClosing( WindowEvent e )
             {
               if ( exitFlag == true )
               {
                    datFile.closeFile();
                    setVisible( false );
                    setEnabled( false );
                    mainFrame.setVisible( true );
                    mainFrame.setEnabled( true );
               }
             }
      }  );
   }

  private void buildInterface()
  {
      JPanel showPanel = new JPanel();
      showPanel.setLayout( new BorderLayout() );
      c.add( showPanel, BorderLayout.CENTER );

      JPanel capPanel = new JPanel();
      capPanel.setLayout( new FlowLayout( FlowLayout.CENTER, 100, 5 ) );
      showPanel.add( capPanel, BorderLayout.NORTH );
      JLabel taskCaption = new JLabel( "CATEGORICAL INFERENCE TASK" );
      taskCaption.setFont( new Font( "Serif", Font.BOLD, 20 ) );
      capPanel.add( taskCaption );

      JPanel mainPanel = new JPanel();
      mainPanel.setLayout( new GridLayout( 1, 1, 5, 5 ) );
      showPanel.add( mainPanel, BorderLayout.CENTER );

      JPanel resultPanel = new JPanel();
      resultPanel.setLayout( new GridLayout( 3, 1, 5, 5 ) );
      mainPanel.add( resultPanel );
      arguArea1 = new TextArea( "ARGUMENTS IN EXPETIMENTS \n" );
      resultPanel.add( arguArea1 );
      resultArea1 = new TextArea( "THE RESULTS ARE: " );
      resultPanel.add( resultArea1 );
      showArea1 = new TextArea( "EXPERIMENT PROCESS SHOW \n" );
      resultPanel.add( showArea1 );

      JPanel buttonPanel = new JPanel();
      buttonPanel.setLayout( new FlowLayout() );
      c.add( buttonPanel, BorderLayout.SOUTH );

      beginButton = new JButton( "BEGIN" );
      beginButton.setEnabled( true );
      beginButton.addActionListener(
         new  ActionListener() {
                 public void actionPerformed( ActionEvent e )
                 {
                     beginButton.setEnabled( false );
                     exitButton.setEnabled( false );
                     exitFlag = false;
                     startExp();
                     exitButton.setEnabled( true );
                 }
        }
      );

      exitButton = new JButton( "Exit" );
      exitButton.addActionListener(
                 new  ActionListener() {
                     public void actionPerformed( ActionEvent e )
                     {
                        closeAllFiles();
                        setVisible( false );
                        setEnabled( false );
                        mainFrame.setVisible( true );
                        mainFrame.setEnabled( true );
                     }
                 }
      );

      buttonPanel.add( beginButton );
      buttonPanel.add( exitButton );
   }

   private void closeAllFiles()
   {
      datFile.closeFile();
      for( int i = 0; i < EXP_NUM; i++ )
           expDatFiles[i].closeFile();
   }

   private void createExpDataFiles()
   {
      expDatFiles = new WriteFile[ EXP_NUM ];
      for( int i = 0; i < EXP_NUM; i++ )
      {
           expDatFiles[i] = new WriteFile();
           String str = "exp";
           if( i < 2 )
               str += String.valueOf( i + 1 );
           else
               str += String.valueOf( i + 2 );

           expDatFiles[i].openFile( 1, str );
      }
   }

/***************************** MAIN METHOD ***************************************/
   private void startExp()
   {
      //doExp1();
      doExp2();
      //doExp4();
      //doExp5();
   }

   private void doExp1()
   {
      createQuestionnaires( 0 );
      ciAgent.setArgIndices( arguIndices );
      ciAgent.setExpIdx( 0 );

      // do experiment.
      for( int i = 0; i < EXP_AGENT_NUMS[ 0 ]; i++ )
      {
           System.out.println( "current agent = " + i );
           // generate question group order.
           groupOrder = new short[ GRP_QUEST_NUMS[ 0 ][ 0 ] ];
           createQuestGroupOrder( 0, i, groupOrder );
           ciAgent.setQuestGroupOrder( groupOrder );
           ciAgent.answerForceQuestions();
           ciAgent.getAnswers( 0, (short[])answerList.elementAt( 0 ) );
           doStatistics( 0, i );
           saveAgentData( 0, i );
           ciAgent.reinit();
      }

      saveExpData( 0 );
   }

   private void doExp2()
   {
      createQuestionnaires( 1 );
      ciAgent.setArgIndices( arguIndices );
      ciAgent.setExpIdx( 1 );

      //final int GRP_QUEST_NUMS[][] =
      //       { { 1, 108 }, { 9, 8 }, { 3, 25 }, { 9, 8 } };
      // do experiment.
      for( int i = 0; i < EXP_AGENT_NUMS[ 1 ]; i++ )
      {
           System.out.println( "current agent = " + i );
           groupOrder = new short[ GRP_QUEST_NUMS[ 1 ][ 0 ] ];
           createQuestGroupOrder( 1, i, groupOrder );
           ciAgent.setQuestGroupOrder( groupOrder );
           ciAgent.answerRatingQuestions();
           ciAgent.getAnswers( 1, (double[][])answerList.elementAt( 1 ) );
           doStatistics( 1, i );
           saveAgentData( 1, i );
           ciAgent.reinit();
      }

      saveExpData( 1 );
   }

   private void doExp4()
   {
      createQuestionnaires( 2 );
      ciAgent.setArgIndices( arguIndices );
      ciAgent.setExpIdx( 2 );

      for( int i = 0; i < EXP_AGENT_NUMS[ 2 ]; i++ )
      {
           System.out.println( "current agent = " + i );
           groupOrder = new short[ GRP_QUEST_NUMS[ 2 ][ 0 ] ];
           createQuestGroupOrder( 2, i, groupOrder );
           ciAgent.setQuestGroupOrder( groupOrder );
           ciAgent.answerRatingQuestions();
           ciAgent.getAnswers( 2, (double[][])answerList.elementAt( 2 ) );
           doStatistics( 2, i );
           saveAgentData( 2, i );
           ciAgent.reinit();
      }

      saveExpData( 2 );
   }

   private void doExp5()
   {
      createQuestionnaires( 3 );
      ciAgent.setArgIndices( arguIndices );
      ciAgent.setExpIdx( 3 );

      for( int i = 0; i < EXP_AGENT_NUMS[ 3 ]; i++ )
      {
           System.out.println( "current agent = " + i );
           groupOrder = new short[ GRP_QUEST_NUMS[ 3 ][ 0 ] ];
           createQuestGroupOrder( 3, i, groupOrder );
           ciAgent.setQuestGroupOrder( groupOrder );
           ciAgent.answerRatingQuestions();
           ciAgent.getAnswers( 3, (double[][])answerList.elementAt( 3 ) );
           doStatistics( 3, i );
           saveAgentData( 3, i );
           ciAgent.reinit();
      }

      saveExpData( 3 );
   }

   private void createQuestGroupOrder( int expIdx, int agentIdx, short[] order )
   {
       int groupNum = GRP_QUEST_NUMS[ expIdx ][ 0 ];
       int agentPerGroupNum = EXP_AGENT_NUMS[ expIdx ] / groupNum;

       if( agentIdx % agentPerGroupNum != 0 )
           return;

       int startIdx = agentIdx / agentPerGroupNum;
       for( int i = 0; i < order.length; i++ )
            order[ i ] = (short)( ( startIdx + i ) % order.length );
   }

   private void createArguments()
   {
      createInclusionSimArguments();
      createPremiseSpecArguments();
      createFilterArguments();
      createIdentityArguments();
   }

   private void createInclusionSimArguments( )
   {
      //final int ARG_TYPE_NUMS[] = { 18, 18, 180, 9 };
      arguments[ INCL_SIMI_TYPE ] = new short[ARG_TYPE_NUMS[INCL_SIMI_TYPE]][CHUNK_TYPE_NUM][CHUNK_DIM_NUM][CHUNK_VAL_NUM];

      for( int i = 0; i < DOMAIN_NUM; i++ )
       for( int j = 0; j < KEY_PAIRS_PER_DOMAIN; j++ )
       {
          for( int k = 0; k < 2; k++ )
          {
             int idx = 2 * ( i * KEY_PAIRS_PER_DOMAIN + j ) + k;
             for( int p = 0; p < CHUNK_TYPE_NUM; p++ )
             {
                  if( p < 2 )
                  {
                     if( k == 0 )
                         // create super and attribute chunk
                         createNormChunk( i, j, arguments[ INCL_SIMI_TYPE ][idx][p] );
                     else
                         //copy the super and attribute chunks from previous argument.
                         for( int q = 0; q < CHUNK_DIM_NUM; q++ )
                           for( int m = 0; m < CHUNK_VAL_NUM; m++ )
                                arguments[ INCL_SIMI_TYPE ][idx][p][q][m] = arguments[ INCL_SIMI_TYPE ][idx-1][p][q][m];
                  }
                  else
                  {
                      if( k == 0 )
                          // create sub chunk
                          createSubChunk( i, j, MORE_SIM_TYPE, arguments[ INCL_SIMI_TYPE ][idx][p], arguments[ INCL_SIMI_TYPE ][idx][0] );
                      else
                          // create sub chunk
                          createSubChunk( i, j, LESS_SIM_TYPE, arguments[ INCL_SIMI_TYPE ][idx][p], arguments[ INCL_SIMI_TYPE ][idx][0] );
                  }
             }
          }
          /*
          if( debug )
          {
              int idx = 2 * ( i * KEY_PAIRS_PER_DOMAIN + j );
              System.out.println( "IS fact premise : " +
              EqualTest.isEqual(arguments[ INCL_SIMI_TYPE ][idx][0], arguments[ INCL_SIMI_TYPE ][idx+1][0] ));
              System.out.println( "IS fact conclusion : " +
              EqualTest.isEqual(arguments[ INCL_SIMI_TYPE ][idx][1], arguments[ INCL_SIMI_TYPE ][idx+1][1] ));
              System.out.println( "IS argument 1 premise : " +
              EqualTest.isEqual(arguments[ INCL_SIMI_TYPE ][idx][2], arguments[ INCL_SIMI_TYPE ][idx][0] ));
              System.out.println( "IS argument 2 premise : " +
              EqualTest.isEqual(arguments[ INCL_SIMI_TYPE ][idx+1][2], arguments[ INCL_SIMI_TYPE ][idx+1][0] ));
          }
          */
       }
   }

   private void createPremiseSpecArguments( )
   {
      //final int ARG_TYPE_NUMS[] = { 18, 18, 180, 9 };
      arguments[ PREM_SPEC_TYPE ] = new short[ARG_TYPE_NUMS[PREM_SPEC_TYPE]][CHUNK_TYPE_NUM][CHUNK_DIM_NUM][CHUNK_VAL_NUM];

      for( int i = 0; i < DOMAIN_NUM; i++ )
       for( int j = 0; j < KEY_PAIRS_PER_DOMAIN; j++ )
       {
          for( int k = 1; k >= 0; k-- )
          {
             int idx = 2 * ( i * KEY_PAIRS_PER_DOMAIN + j ) + k;
             for( int p = 0; p < CHUNK_TYPE_NUM; p++ )
             {
                  switch( p )
                  {
                     case 0: if( k == 1 ) //create super chunk
                                 createNormChunk( i, j, arguments[ PREM_SPEC_TYPE ][idx][p] );
                             else // create sub chunk
                                 createSubChunk( i, j, MID_SIM_TYPE, arguments[ PREM_SPEC_TYPE ][idx][p],
                                                 arguments[ PREM_SPEC_TYPE ][idx+1][p] );
                             break;
                     case 1: if( k == 1 ) //create norm chunk
                                 createNormChunk( i, j, arguments[ PREM_SPEC_TYPE ][idx][p] );
                             else
                                 //copy the norm chunk from previous argument.
                                 for( int q = 0; q < CHUNK_DIM_NUM; q++ )
                                   for( int m = 0; m < CHUNK_VAL_NUM; m++ )
                                        arguments[ PREM_SPEC_TYPE ][idx][p][q][m] = arguments[ PREM_SPEC_TYPE ][idx+1][p][q][m];
                             break;
                     case 2: if( k == 0 )
                             {   //create subsub chunk
                                 createSubChunk( i, j, MORE_SIM_TYPE, arguments[ PREM_SPEC_TYPE ][idx][p],
                                                 arguments[ PREM_SPEC_TYPE ][idx][0] );
                                 //copy the subsub chunk from aleady-set argument.
                                 for( int q = 0; q < CHUNK_DIM_NUM; q++ )
                                   for( int m = 0; m < CHUNK_VAL_NUM; m++ )
                                        arguments[ PREM_SPEC_TYPE ][idx+1][p][q][m] = arguments[ PREM_SPEC_TYPE ][idx][p][q][m];
                             }
                             break;
                  }
             }
          }
          /*
          if( debug )
          {
              int idx = 2 * ( i * KEY_PAIRS_PER_DOMAIN + j );
              System.out.println( "PS fact premise : " +
              EqualTest.isEqual(arguments[ PREM_SPEC_TYPE ][idx][0], arguments[ PREM_SPEC_TYPE ][idx+1][0] ));
              System.out.println( "PS fact conclusion : " +
              EqualTest.isEqual(arguments[ PREM_SPEC_TYPE ][idx][1], arguments[ PREM_SPEC_TYPE ][idx+1][1] ));
              System.out.println( "PS argument premise : " +
              EqualTest.isEqual(arguments[ PREM_SPEC_TYPE ][idx][2], arguments[ PREM_SPEC_TYPE ][idx+1][2] ));
              System.out.println( "PS argument 1 premise : " +
              EqualTest.isEqual(arguments[ PREM_SPEC_TYPE ][idx][2], arguments[ PREM_SPEC_TYPE ][idx][0] ));
              System.out.println( "PS argument 2 premise : " +
              EqualTest.isEqual(arguments[ PREM_SPEC_TYPE ][idx+1][2], arguments[ PREM_SPEC_TYPE ][idx+1][0] ));
          }
          */
      }
   }

   private void createFilterArguments( )
   {
      //final int ARG_TYPE_NUMS[] = { 18, 18, 180, 9 };
      arguments[ FILT_TYPE ] = new short[ARG_TYPE_NUMS[FILT_TYPE]][CHUNK_TYPE_NUM][CHUNK_DIM_NUM][CHUNK_VAL_NUM];
      for( int i = 0; i < DOMAIN_NUM; i++ )
       for( int j = 0; j < FILT_PAIRS_PER_DOMAIN; j++ )
       {
          for( int k = 0; k < 2; k++ )
          {
             int idx = 2 * ( i * FILT_PAIRS_PER_DOMAIN + j ) + k;
             for( int p = 0; p < CHUNK_TYPE_NUM; p++ )
             {
                  if( p != FACT_CONC_IDX  || k == 0 )
                      // create norm chunk
                      createNormChunk( i, j, arguments[ FILT_TYPE ][idx][p] );
                  else
                  {
                     //copy the norm chunks from previous argument.
                     for( int q = 0; q < CHUNK_DIM_NUM; q++ )
                       for( int m = 0; m < CHUNK_VAL_NUM; m++ )
                            arguments[ FILT_TYPE ][idx][p][q][m] = arguments[ FILT_TYPE ][idx-1][p][q][m];
                  }
             }
          }
          /*
          if( debug )
          {
              int idx = 2 * ( i * KEY_PAIRS_PER_DOMAIN + j );
              System.out.println( "FILT fact premise : " +
              EqualTest.isEqual(arguments[ FILT_TYPE ][idx][0], arguments[ FILT_TYPE ][idx+1][0] ));
              System.out.println( "FILT fact conclusion : " +
              EqualTest.isEqual(arguments[ FILT_TYPE ][idx][1], arguments[ FILT_TYPE ][idx+1][1] ));
              System.out.println( "FILT argument 1 premise : " +
              EqualTest.isEqual(arguments[ FILT_TYPE ][idx][2], arguments[ FILT_TYPE ][idx][0] ));
              System.out.println( "FILT argument 2 premise : " +
              EqualTest.isEqual(arguments[ FILT_TYPE ][idx+1][2], arguments[ FILT_TYPE ][idx+1][0] ));
          } */
      }
   }

   private void createIdentityArguments( )
   {
      //final int ARG_TYPE_NUMS[] = { 18, 18, 180, 9 };
      arguments[ IDEN_TYPE ] = new short[ARG_TYPE_NUMS[IDEN_TYPE]][CHUNK_TYPE_NUM][CHUNK_DIM_NUM][CHUNK_VAL_NUM];
      for( int i = 0; i < DOMAIN_NUM; i++ )
       for( int j = 0; j < IDEN_ARGS_PER_DOMAIN; j++ )
        {
             int idx = i * IDEN_ARGS_PER_DOMAIN + j;
             for( int p = 0; p < CHUNK_TYPE_NUM; p++ )
             {
                  if( p != ARGU_PREM_IDX )
                      // create norm chunk
                      createNormChunk( i, j, arguments[ IDEN_TYPE ][idx][p] );
                  else
                  {
                     //copy the norm chunks from previous argument.
                     for( int q = 0; q < CHUNK_DIM_NUM; q++ )
                       for( int m = 0; m < CHUNK_VAL_NUM; m++ )
                            arguments[ IDEN_TYPE ][idx][p][q][m] = arguments[ IDEN_TYPE ][idx][FACT_PREM_IDX][q][m];
                  }
             }
             /*
             if( debug )
             {
                 System.out.println( "IDEN argument premise : " +
                 EqualTest.isEqual(arguments[ IDEN_TYPE ][idx][2], arguments[ IDEN_TYPE ][idx][0] ));
             }
             */
       }

   }

   // create super, attribute, or identity chunk.
   private void createNormChunk( int domainIdx, int argIdx, short[][] arr )
   {
      int downBound =  domainIdx * INTERVAL;
      int upBound = ( domainIdx + 1 ) * INTERVAL;
      for( int q = 0; q < CHUNK_DIM_NUM; q++ )
      {
           int valNum = (int)( 1 + ( INTERVAL * 0.5 ) * Math.random() );
           Vector seeds = genSeeds( 0, INTERVAL );
           for( int m = 0; m < valNum; m++ )
           {
                int idx = (int)( seeds.size() * Math.random() );
                int val = ((int[])seeds.elementAt( idx ))[0];
                arr[ q ][ downBound + val ] = 1;
                seeds.removeElementAt( idx );
           }
      }
   }

   private void createSubChunk( int domainIdx, int argIdx, int simType,
                                short[][] arr, short[][] superArr )
   {
      //generate the # of dims of sub chunks identical to super chunk.
      int valNum = 0;

     // final int LESS_SIM_TYPE = 0, MID_SIM_TYPE = 1, MORE_SIM_TYPE = 2;
     /* final double[][][] SIMI_VALS =
         {
             { { 0.4, 0.6 }, { 0.5, 0.8 }, { 0.7, 0.9 } },
             { { 0.4, 0.6 }, { 0.5, 0.8 }, { 0.7, 0.9 } },
             { { 0.4, 0.6 }, { 0.5, 0.8 }, { 0.7, 0.9 } }
         };
      */

      // create chunk with some similiarity to super chunk.
      double downBound = SIMI_VALS[ domainIdx ][ simType ][ 0 ] * CHUNK_DIM_NUM;
      double upBound = SIMI_VALS[ domainIdx ][ simType ][ 1 ] * CHUNK_DIM_NUM;
      double span = upBound - downBound;
      valNum = (int)( downBound + span * Math.random() );

      //find those dims of sub chunks identical to super chunk.
      int[] dimIdenInfo = new int[ CHUNK_DIM_NUM ];
      for( int p = 0; p < CHUNK_DIM_NUM; p++ )
           dimIdenInfo[ p ] = 0;

      Vector seeds = genSeeds( 0, CHUNK_DIM_NUM );
      for( int p = 0; p < valNum; p++ )
      {
           int idx = (int)( seeds.size() * Math.random() );
           int val = ((int[])seeds.elementAt( idx ))[0];
           dimIdenInfo[ val ] = 1;
           seeds.removeElementAt( idx );
      }

      // create the sub chunks.
      downBound =  domainIdx * INTERVAL;
      upBound = ( domainIdx + 1 ) * INTERVAL;
      for( int p = 0; p < CHUNK_DIM_NUM; p++ )
      {
           for( int q = 0; q < CHUNK_VAL_NUM; q++ )
                arr[p][q] = superArr[p][q];
           if( dimIdenInfo[ p ] == 0 )  // the dim is not the same.
           {
               int num = (int)( 1 + ( INTERVAL * 0.1 ) * Math.random() );
               int count = 0;
               while( count < num )
               {
                      int offset = (int)(downBound + INTERVAL * Math.random());
                      if( arr[p][offset] == 0 )
                      {
                          arr[p][offset] = 1;
                          count++;
                      }
               }
           }
      }
   }

   private void createQuestionnaires( int expIdx )
   {
      // ARG_TYPE_PAIRS[] = { 9, 9, 90, 9 };
      Vector seeds[] = new Vector[ ARGU_TYPE_NUM ];
      for( int i = 0; i < ARGU_TYPE_NUM; i++ )
           seeds[ i ] = genSeeds( 0, ARG_TYPE_PAIRS[ i ] );

      // GRP_QUEST_TYPE_PAIRS[][] =
      //                          { { 9, 9, 90, 0 }, { 1, 1, 6, 0 },
      //                            { 3, 3, 16, 3 }, { 1, 1, 6, 0 } };
      short[][][] tempIndices =
                  new short[GRP_QUEST_NUMS[expIdx][0]][GRP_QUEST_NUMS[expIdx][1]][2];
      for( int i = 0; i < GRP_QUEST_NUMS[ expIdx ][ 0 ]; i++ )
      {
           int offset = 0;
           for( int  j = 0; j < ARGU_TYPE_NUM; j++ )
           {
                int num = GRP_QUEST_TYPE_PAIRS[ expIdx ][ j ];
                for( int m = 0; m < num; m++ )
                {
                     int idx = ( int )( seeds[ j ].size() * Math.random() );
                     int val = (( int[] )seeds[ j ].elementAt( idx ) )[ 0 ];
                     tempIndices[ i ][ offset + m ][ 0 ] = (short)j;
                     tempIndices[ i ][ offset + m ][ 1 ] = (short)val;
                     seeds[j].removeElementAt( idx );
                }
                offset += num;
          }
      }

      // final int GRP_QUEST_NUMS[][] =
      //       { { 1, 108 }, { 9, 8 }, { 3, 25 }, { 9, 8 } };
      arguIndices = new short[GRP_QUEST_NUMS[expIdx][0]][GRP_QUEST_NUMS[expIdx][1]][2];
      for( int i = 0; i < GRP_QUEST_NUMS[ expIdx ][ 0 ]; i++ )
      {
           int num = GRP_QUEST_NUMS[ expIdx ][ 1 ];
           Vector seed = genSeeds( 0, num );
           for( int j = 0; j < num; j++ )
           {
                int idx = (int)( seed.size() * Math.random() );
                int val = ((int[])seed.elementAt( idx ))[0];
                arguIndices[ i ][ j ][ 0 ] = tempIndices[ i ][ val ][ 0 ];
                arguIndices[ i ][ j ][ 1 ] = tempIndices[ i ][ val ][ 1 ];
                seed.removeElementAt( idx );
           }
      }

      /*
      if( debug )
      {
          for( int i = 0; i < GRP_QUEST_NUMS[ expIdx ][ 0 ]; i++ )
           for( int j = 0; j < GRP_QUEST_NUMS[ expIdx ][ 1 ]; j++ )
           {
               System.out.print( "group idx = " + i + ", " );
               System.out.print( "array idx = " + j + ", " );
               System.out.print( "arg type = " +  arguIndices[i][j][0] + ", " );
               System.out.println( "arg idx = " + arguIndices[i][j][1] );
           }
      }
      */
   }

   private Vector genSeeds( int from, int to )
   {
      int len  = to - from;
      Vector seeds = new Vector( len );
      for( int i = from; i < to; i++ )
      {
           int[] elem = new int[ 1 ];
           elem[ 0 ] = i;
           seeds.addElement( elem );
      }

      return seeds;
   }


   private void doStatistics( int expIdx, int agentIdx )
   {
      if( expIdx == 0 )
          doFCStatics( agentIdx );
      else
          doRateStatics( expIdx, agentIdx );
   }

   private void doFCStatics( int agentIdx )
   {
      short[] fcAnswers = (short[])answerList.elementAt( 0 );
      short[][][] agentResults = (short[][][])agentResultList.elementAt( 0 );
      short[][][] groupResults = ((short[][][][])groupResultList.elementAt( 0 ))[agentIdx];
      short[][][] finalResults = (short[][][])finalResultList.elementAt( 0 );

      int idx1 = GRP_QUEST_NUMS[ 0 ][ 0 ];
      int idx2 = GRP_QUEST_NUMS[ 0 ][ 1 ];
      for( int i = 0; i < idx1; i++ )
        for( int j = 0; j < idx2; j++ )
        {
             int questIdx = i * GRP_QUEST_NUMS[ 0 ][ 1 ] + j;
             int argType = arguIndices[ i ][ j ][ 0 ];
             int keyIdx = arguIndices[ i ][ j ][ 1 ];
             switch( argType ) {
               case INCL_SIMI_TYPE:
               case PREM_SPEC_TYPE:
                    agentResults[ argType ][ keyIdx ][ fcAnswers[questIdx] ] = 1;
                    agentResults[ argType ][ keyIdx ][ 1-fcAnswers[questIdx] ] = 0;
                    groupResults[ argType ][ keyIdx ][ fcAnswers[questIdx] ] = 1;
                    groupResults[ argType ][ keyIdx ][ 1-fcAnswers[questIdx] ] = 0;
                    finalResults[ argType ][ keyIdx ][ fcAnswers[questIdx] ]++;
                    break;
               case FILT_TYPE:
                    break;
               case IDEN_TYPE:
                    break;
             }
        }
   }

   public void doRateStatics( int expIdx, int agentIdx )
   {
      double[][] fcAnswers = (double[][])answerList.elementAt( expIdx );
      double[][][] agentResults = (double[][][])agentResultList.elementAt( expIdx );
      double[][][] groupResults = ((double[][][][])groupResultList.elementAt( expIdx ))[agentIdx];
      double[][][] finalResults = (double[][][])finalResultList.elementAt( expIdx );

      int idx1 = GRP_QUEST_NUMS[ expIdx ][ 0 ];
      int idx2 = GRP_QUEST_NUMS[ expIdx ][ 1 ];

      for( int i = 0; i < idx1; i++ )
        for( int j = 0; j < idx2; j++ )
        {
             int questIdx = i * idx2 + j;
             int argType = arguIndices[ groupOrder[i] ][ j ][ 0 ];
             int keyIdx = arguIndices[ groupOrder[i] ][ j ][ 1 ];

             switch( argType ) {
               case INCL_SIMI_TYPE:
               case PREM_SPEC_TYPE:
                    for( int k = 0; k < 2; k++ )
                    {
                         agentResults[ argType ][ keyIdx ][ k ] = fcAnswers[questIdx][k];
                         groupResults[ argType ][ keyIdx ][ k ] = fcAnswers[questIdx][k];
                         finalResults[ argType ][ keyIdx ][ k ] += fcAnswers[questIdx][k];
                    }
                    break;
               case FILT_TYPE:
                    break;
               case IDEN_TYPE:
                    break;
             }
        }
   }

   private void saveAgentData( int expIdx, int agentIdx )
   {
      if( expIdx == 0 )
          saveFCAgentData( agentIdx );
      else
          saveRateAgentData( expIdx, agentIdx );
   }

   private void saveFCAgentData( int agentIdx )
   {
      short[][][] agentResults = (short[][][])agentResultList.elementAt( 0 );

      String str = "Current agent is: " + String.valueOf( agentIdx ) + "\n";
      str += "The result of choice is: " + "\n";

      for( int i = 0; i < 2; i++ )
      {
           if( i == 0 )
               str += "The first argument is chosen in a pair of inclusion similarity: \n";
           else
               str += "The first argument is chosen in a pair of premise specificit: \n";

           for( int j = 0; j < KEY_ARG_PAIRS; j++ )
                str += String.valueOf( agentResults[ i ][ j ][ 0 ] ) + "\n";
      }

      str += "\n";
      expDatFiles[ 0 ].saveStringData( str );
   }

   private void saveRateAgentData( int expIdx, int agentIdx )
   {
      double[][][] agentResults = (double[][][])agentResultList.elementAt( expIdx );

      String str = "Current agent is: " + String.valueOf( agentIdx ) + "\n";
      str += "The result of rating is: " + "\n";

      for( int i = 0; i < 2; i++ )
      {
           if( i == 0 )
               str += "conditional probabilities of each argument in a pair of inclusion similarity: \n";
           else
               str += "conditional probabilities of each argument in a pair of premise specificit: \n";

           for( int j = 0; j < KEY_ARG_PAIRS; j++ )
           {
                for( int k = 0; k < 2; k++ )
                     str += String.valueOf( agentResults[ i ][ j ][ k ] ) + " ";
                str += "\n";
           }
      }

      str += "\n";
      expDatFiles[ expIdx ].saveStringData( str );
   }

   private void saveExpData( int expIdx )
   {
      if( expIdx == 0 )
          saveFCExpData();
      else
          saveRateExpData( expIdx );
   }

   private void saveFCExpData()
   {
      short[][][] finalResults = (short[][][])finalResultList.elementAt( 0 );

      String str = "Current experiment is: " + String.valueOf( 1 ) + "\n";
      str += "The result is: \n";

      for( int i = 0; i < 2; i++ )
      {
           if( i == 0 )
               str += "Percentage of times the more similar argument from each pair of \n" +
                      "inclusion similarity arguments: \n";
           else
               str += "Percentage of times the more similar argument from each pair of \n" +
                      "premise specificity arguments: \n";

           double sum = 0.0;
           for( int j = 0; j < KEY_ARG_PAIRS; j++ )
                sum += finalResults[ i ][ j ][ 0 ];

           sum /= KEY_ARG_PAIRS * EXP_AGENT_NUMS[ 0 ];
           str += String.valueOf( sum ) + "\n";
      }

      str += "\n";
      datFile.saveStringData( str );
   }

   private void saveRateExpData( int expIdx )
   {
      double[][][] finalResults = (double[][][])finalResultList.elementAt( expIdx );
      int[] expNums = { 1, 2, 4, 5 };
      String[] simiStrs = { "More similar : ", "Less similar : " };

      String str = "Current experiment is: " + String.valueOf( expNums[ expIdx ] ) + "\n";
      str += "The result is: \n";

      str += "Mean conditional probability: \n";
      for( int i = 0; i < 2; i++ )
      {
           if( i == 0 )
               str += "Inclusion similarity arguments: \n";
           else
               str += "Premise specificity arguments: \n";

           double[] means = new double[ 2 ];
           for( int j = 0; j < 2; j++ )
           {
                str += simiStrs[ j ];
                means[ j ] = 0.0;
                for( int k = 0; k < KEY_ARG_PAIRS; k++ )
                     means[ j ] += finalResults[ i ][ k ][ j ];

                means[ j ] /= KEY_ARG_PAIRS * EXP_AGENT_NUMS[ expIdx ];
                str += String.valueOf( means[ j ] ) + "\n";
           }
      }

      str += "\n";
      datFile.saveStringData( str );
   }
}