package tasksimulators.oldsimu;
import java.util.Vector;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.coordinate.*;
import clarion.common.*;
import clarion.tools.*;

public class DSAgent{

   // used for combination of BL and TL.
   final int COMB_COMPONENT_NUM = 4;
   final int BL_IDX = 0, RER_IDX = 1, IRL_IDX = 2, FIX_IDX = 3;
   final int FIX_MODE = 0, VARIED_MODE = 1;
   final int AMN_IDX = 0, GKN_IDX = 1;

   // constant variables for the experiment.
   final int ITEM_NUM = 10, WORDS_PER_ITEM = 16;
   // constant variable for training.
   //final int TRAIN_DATA_LENGTH = 20, TRAIN_INTERVAL = 50, LEFT_INTERVAL = 40;
   final int TRAIN_DATA_LENGTH = 20, TRAIN_INTERVAL = 250, LEFT_INTERVAL = 200;
   //final int BASE_NUM = 1, FIRST_TIMES = 4, SECOND_TIMES = 17;
   final int BASE_NUM = 5, FIRST_TIMES = 4, SECOND_TIMES = 17;
   final int FIRST_WORD_NUM = 12, SECOND_WORD_NUM = 3;

   // constant variables relevant to IDN
   final int IDNET_DIM_NUM = 161, IDNET_INPUT_NUM = 1602, IDNET_HIDDEN_NUM = 200, IDNET_OUTPUT_NUM = 24;
   // constant variables relevant to AMN
   final int AMNET_DIM_NUM = 160, AMNET_INPUT_NUM = 1600, AMNET_HIDDEN_NUM = 200, AMNET_OUTPUT_NUM = 100;
     // constant variables for WM.
   final int WM_ITEM_NUM = 15, GS_ITEM_NUM = 2;
   // used for combination of two levels of NACS.
   final double nacsBetaArr[][] = { { 0.5, 0.5 } };

    // constant variables relevant to chunk.
   final int CHUNK_ROW_NUM = 20, CHUNK_COLUM_NUM = 20;
   final int CHUNK_DIM_NUM = 10, CHUNK_DIM_VAL_NUM = 10;
   final double NOISE_LEVEL = 0.2;

   // thresholds for chunks.
   //final double d_c = 0.01;
   final double d_c = 0.004;
   //final double p_ce = 0.7;
   final double p_ce = 1.0;
   //final double threshold_ce = 0.2;
   final double threshold_ce = 0.5;
   final double p_c = 0.5;

   // thresholds for associative rules.
   //final double d_a = 0.01;
   final double d_a = 0.004;
   //final double p_ae = 0.7;
   final double p_ae = 1.0;
   //final double threshold_ae = 0.2;
   final double threshold_ae = 0.5;
   final double p_a = 0.5;

   // constant for hunch or fianl solution.
   final double H_THRESHOLD = 0.25, W_THRESHOLD = 0.3;
   //final double H_THRESHOLD = 0.24, W_THRESHOLD = 0.28;

   // constant for reasoning types.
   final int RULE_BASED = 0, SIMI_BASED = 1;

   int acsQbpDimValNums[], nacsQbpDimValNums[];
   int ruleStats[], hunchIndices[], winIndices[];
   int chunkDVNums[], chunkDimTypes[], dimValNums[];
   double hunchIdx, winIdx;
   double activations[][], amnActivation[][], gknActivation[][];

   // used for combination of levels.
   int actionTypeNum;
   int acsChosenComp, nacsChosenComp;
   double acsBetaArr[][];

   // all the words used.
   short wordItems[][][][];
   // actions from IDN or ARN.
   short action[], actionBL[], actionTL[];

   // used for indicating a chunk already exists.
   short chunkExists[][];
   // used for retrieval cue.
   int cueDimIndices[];
   // test flag for getting the testing data
   boolean testFlag = true;
   double trainData[], testData[][];
   String rules, chunks;

   // agent no
   int agentIdx;

   GKNet gkNet;
   AMNet amNet;
   IDNet idNet;
   ExWM wm;
   StochasticComb acsStochCombiner, nacsStochCombiner;

   public DSAgent()
   {
      hunchIndices = new int[ ITEM_NUM ];
      winIndices = new int[ ITEM_NUM ];
      ruleStats = new int[ WORDS_PER_ITEM - 1 ];

      activations = new double[ ITEM_NUM ][ WORDS_PER_ITEM - 1 ];
      amnActivation = new double[ ITEM_NUM ][ WORDS_PER_ITEM - 1 ];
      gknActivation = new double[ ITEM_NUM ][ WORDS_PER_ITEM - 1 ];

      // initialize actions of IDN and ARN.
      actionTL = new short[ IDNET_OUTPUT_NUM ];
      actionBL = new short[ IDNET_OUTPUT_NUM ];

      // used for indicating a chunk already exists.
      chunkExists = new short[ ITEM_NUM ][ WORDS_PER_ITEM ];
      for( int i = 0; i < ITEM_NUM; i++ )
       for( int j = 0; j < WORDS_PER_ITEM; j++ )
            chunkExists[ i ][ j ] = -1;

      // combination of two levels of ACS.
      actionTypeNum = 1;
      acsBetaArr = new double[ actionTypeNum ][ COMB_COMPONENT_NUM ];
      for( int i = 0; i < COMB_COMPONENT_NUM; i++ )
           acsBetaArr[ 0 ][ i ] = World.TASK_PARA_ARRAY[ World.WBL + i ];
      acsStochCombiner = new StochasticComb( FIX_MODE, COMB_COMPONENT_NUM, acsBetaArr[0] );

      // combination of two levels of NACS.
      nacsStochCombiner = new StochasticComb( FIX_MODE, 2, nacsBetaArr[0]  );

      // initialize the GKN.
      chunkDVNums = new int[ CHUNK_DIM_NUM ];
      for( int i = 0; i < CHUNK_DIM_NUM; i++ )
           chunkDVNums[ i ] = CHUNK_DIM_VAL_NUM;
      chunkDimTypes = new int[ CHUNK_DIM_NUM ];
      for( int i = 0; i < CHUNK_DIM_NUM; i++ )
           chunkDimTypes[ i ] = 0;
      gkNet = new GKNet( ITEM_NUM, WORDS_PER_ITEM, CHUNK_DIM_NUM, RULE_BASED,
                         chunkDVNums, chunkDimTypes, 0.0 );
      gkNet.setMaxAges( (int)( 1.0 / d_c ), (int)( 1.0 / d_a ) );

      cueDimIndices = new int[ CHUNK_DIM_NUM ];
      for( int i = 0; i < CHUNK_DIM_NUM; i++ )
           cueDimIndices[ i ] = i;
      gkNet.setCueDimIndices( cueDimIndices );

      // initialize AMN.
      nacsQbpDimValNums = new int[ AMNET_DIM_NUM ];
      for( int i = 0; i < nacsQbpDimValNums.length; i++ )
           nacsQbpDimValNums[ i ] = CHUNK_DIM_VAL_NUM;
      amNet = new AMNet( AMNET_INPUT_NUM, AMNET_HIDDEN_NUM, AMNET_OUTPUT_NUM,
                           AMNET_DIM_NUM, nacsQbpDimValNums );

      // initialize IDN.
      acsQbpDimValNums = new int[ IDNET_DIM_NUM ];
      for( int i = 0; i < acsQbpDimValNums.length - 1; i++ )
           acsQbpDimValNums[ i ] = CHUNK_DIM_VAL_NUM;
      acsQbpDimValNums[ acsQbpDimValNums.length - 1 ] = GS_ITEM_NUM;
      idNet = new IDNet( IDNET_INPUT_NUM, IDNET_HIDDEN_NUM, IDNET_OUTPUT_NUM,
                           IDNET_DIM_NUM, acsQbpDimValNums );

      // initial working memory.
      wm = new ExWM( WM_ITEM_NUM );

      // initialization of variables for testing performance of AMN and IDN
      trainData = new double[ TRAIN_DATA_LENGTH ];
      testData = new double[ ITEM_NUM ][ WORDS_PER_ITEM - 1 ];
   }

   public void reinit()
   {
      gkNet.reinit();
      amNet.reinitWeights();
      idNet.reinitWeights();
   }

   public void getResults( double[] arr )
   {
       arr[ 0 ] = hunchIdx;
       if( hunchIdx < 0.0 || winIdx < 0.0 )
           arr[ 1 ] = -1.0;
       else
           arr[ 1 ] = winIdx - hunchIdx;
   }

   public void getTestDataInTraining( double[] arr )
   {
       for( int i = 0; i < arr.length; i++ )
            arr[ i ] = trainData[ i ];
   }

   public void getTestDataInTesting( double[] arr )
   {
      for( int i = 0; i < testData.length; i++ )
       for( int j = 0; j < testData[ 0 ].length; j++ )
            arr[ j ] += (testData[ i ][ j ]/testData.length);
   }

   public void getActivations( double[][] arr )
   {
      for( int i = 0; i < arr.length; i++ )
       for( int j = 0; j < arr[ 0 ].length; j++ )
            arr[ i ][ j ] = activations[ i ][ j ];
   }

   public void getRuleStats( int[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           arr[ i ] = ruleStats[ i ];
   }

   public String getRules()
   {
      return rules;
   }

   public String getChunks()
   {
      return chunks;
   }

   public void setAgentIdx( int idx )
   {
       agentIdx = idx;
   }

   public void setWordItems( short[][][][] items )
   {
      wordItems = new short[ ITEM_NUM ][ WORDS_PER_ITEM ][ CHUNK_DIM_NUM ][ CHUNK_DIM_VAL_NUM ];
      for( int i = 0; i < items.length; i++ )
       for( int j = 0; j < items[ 0 ].length; j++ )
        for( int k = 0; k < items[ 0 ][ 0 ].length; k++ )
         for( int m = 0; m < items[ 0 ][ 0 ][ 0 ].length; m++ )
              wordItems[ i ][ j ][ k ][ m ] = items[ i ][ j ][ k ][ m ];
   }

   public void training()
   {
      int wordNum = ( WORDS_PER_ITEM - 1 ) * ITEM_NUM;
      Vector seeds = new Vector( wordNum );
      for( int i = 0; i < ITEM_NUM; i++ )
       for( int j = 0; j < WORDS_PER_ITEM - 1; j++ )
       {
            int[] item = new int[ 3 ];
            item[ 0 ] = i;
            item[ 1 ] = j;
            if( j < FIRST_WORD_NUM )
                item[ 2 ] = BASE_NUM * FIRST_TIMES;
            else
                item[ 2 ] = BASE_NUM * SECOND_TIMES;
            seeds.addElement( item );
       }

      System.out.println( "IN TRAINING " );
      for( int i = 0; i < trainData.length; i++ )
           trainData[ i ] = 0.0;

      for( int i = 0; i < ITEM_NUM; i++ )
       for( int j = 0; j < WORDS_PER_ITEM; j++ )
            chunkExists[ i ][ j ] = -1;

      int trainCount = 0;
      double activation = 0.0;
      while( seeds.size() > 0 )
      {
           int idx = (int)( Math.random() * seeds.size() );
           int itemIdx = ((int[])seeds.elementAt( idx ))[ 0 ];
           int wordIdx = ((int[])seeds.elementAt( idx ))[ 1 ];
           trainPair( itemIdx, wordIdx );
           activation = extraction( itemIdx, wordIdx );

           if( testFlag )
           {
               if( trainCount / TRAIN_INTERVAL < TRAIN_DATA_LENGTH - 1 )
                   trainData[ (trainCount++) / TRAIN_INTERVAL ] += ( activation / TRAIN_INTERVAL );
               else
                   trainData[ (trainCount++) / TRAIN_INTERVAL ] += ( activation / LEFT_INTERVAL );

               if( trainCount % TRAIN_INTERVAL == 0 )
               {
                   System.out.println("trainCount = " + trainCount );
                   //getCompatibleChunks();
               }
           }

           int leftNum = ((int[])seeds.elementAt( idx ))[ 2 ];
           if( leftNum > 1 )
           {
               int[] item = new int[ 3 ];
               item[ 0 ] = itemIdx;
               item[ 1 ] = wordIdx;
               item[ 2 ] = leftNum - 1;
               seeds.setElementAt( item, idx );
           }
           else
               seeds.removeElementAt( idx );
      }
   }

   public void test()
   {
      for( int i = 0; i < ITEM_NUM; i++ )
           hunchIndices[ i ] = -1;
      for( int i = 0; i < ITEM_NUM; i++ )
           winIndices[ i ] = -1;
      for( int i = 0; i < WORDS_PER_ITEM - 1; i++ )
           ruleStats[ i ] = 0;

      System.out.println( "IN TESTING " );
      for( int i = 0; i < ITEM_NUM; i++ )
      {
           System.out.println( "current item no = " + String.valueOf( i ));

           double[] inputs = new double[ AMNET_INPUT_NUM ];
           for( int j = 0; j < inputs.length; j++ )
                inputs[ j ] = 0.0;

           for( int j = 0; j < WORDS_PER_ITEM - 1; j++ )
           {
                // get activation from AMN.
                getInputFromWM( i, j, inputs );
                amNet.setInput( inputs );
                amNet.forwardPass();
                amnActivation[ i ][ j ] = calcActivations( i );

                if( testFlag )
                    testData[ i ][ j ] = amnActivation[ i ][ j ];

                // get activation from GKN.
                gkNet.setRetrievalCue( wordItems[ i ][ j ] );
                boolean bVal = gkNet.reasoningBetweenNode();
                if( bVal )
                {
                    if( chunkExists[ i ][ WORDS_PER_ITEM - 1 ] == gkNet.getChosenChunkIdx() )
                        gknActivation[ i ][ j ] = gkNet.getChosenChunkStrength();
                    else
                    {
                        System.out.println( "Target is not reached !" );
                        gknActivation[ i ][ j ] = 0.0;
                    }
                }
                else
                    gknActivation[ i ][ j ] = 0.0;

                // decide activation from AMN or GKN.
                nacsChosenComp = nacsStochCombiner.selectComponent();
                if( nacsChosenComp == AMN_IDX )
                    activations[ i ][ j ] = amnActivation[ i ][ j ];
                else
                {
                    if( bVal )
                    {
                        gkNet.resetNodeAge( (short)gkNet.getChosenChunkIdx() );
                        double actVal = gknActivation[ i ][ j ];
                        for( int k = 0; k < j + 1; k++ )
                        {
                             gkNet.resetNodeAge( chunkExists[i][k] );
                             if( gknActivation[ i ][ k ] > 0.0 )
                                 gkNet.resetLinkAge( chunkExists[i][k],
                                                     (short)gkNet.getChosenChunkIdx() );

                             if( gknActivation[ i ][ k ] > actVal )
                                 actVal = gknActivation[ i ][ k ];
                        }
                        activations[ i ][ j ] = actVal;
                    }
                    else
                        activations[ i ][ j ] = amnActivation[ i ][ j ];
                }

                extraction( i, j );
                gkNet.update();
                storeIntoWM( j );
          }
          resetWM();
      }

      // get all of the associative rules.
      if( testFlag )
      {
          rules = "";
          chunks = "";

          System.out.println( " GET ALL THE RULES AND CHUNKS ! " );
          for( int i = 0; i < ITEM_NUM; i++ )
          {
              System.out.println( "current item = " + i );
               for( int j = 0; j < WORDS_PER_ITEM; j++ )
               {
                    String label = "";
                    if( j < WORDS_PER_ITEM - 1 )
                        label = "Item"+String.valueOf( i+1 )+"-Word"+String.valueOf( j+1 );
                    else
                        label = "Item"+String.valueOf( i+1 )+"-Word"+"target";

                    int nodeIdx = gkNet.getChunkIdx( label );
                    if( nodeIdx == -1 )
                        continue;

                    if( gkNet.isAvailableChunk( nodeIdx ) )
                    {
                        rules += gkNet.getRules( nodeIdx, true );
                        chunks += gkNet.getChunk( nodeIdx );
                        if( j < WORDS_PER_ITEM - 1 )
                            ruleStats[ j ] += gkNet.getRuleNum( nodeIdx );
                    }
               }
          }
      }
   }

   // modified on 08/20/02.
   private void trainPair( int itemIdx, int wordIdx )
   {
      amNet.setInput( wordItems[ itemIdx ][ wordIdx ] );
      amNet.forwardPass();
      amNet.setDesiredOutput( wordItems[ itemIdx ][ WORDS_PER_ITEM-1 ] );
      amNet.backwardPass();
   }

   private double extraction( int itemIdx, int wordIdx )
   {
      double activation = calcActivations( itemIdx );

      String wordLabel = "Item" + String.valueOf( itemIdx + 1 ) +
                         "-Word" + String.valueOf( wordIdx + 1 );
      String targetLabel = "Item" + String.valueOf( itemIdx + 1 ) +
                           "-Word" + "target";
      // check if the word chunk exists.
      int wordNodeIdx = gkNet.getChunkIdx( wordLabel );
      chunkExists[ itemIdx ][ wordIdx ] = (short)wordNodeIdx;

      // check if the word target exists.
      int targetNodeIdx = gkNet.getChunkIdx( targetLabel );
      chunkExists[ itemIdx ][ WORDS_PER_ITEM - 1 ] = (short)targetNodeIdx;

      if( chunkExists[ itemIdx ][ wordIdx ] == -1 ) // the word chunk not exists.
      {
          if( Math.random() < p_ce ) // this word chunk is encoded.
          {
              chunkExists[ itemIdx ][ wordIdx ] =
                    (short)gkNet.insertNode( wordItems[ itemIdx ][ wordIdx ], wordLabel );
          }
      }
      else // the word chunk exists.
          gkNet.resetNodeAge( chunkExists[ itemIdx ][ wordIdx ] );


      // chunk extraction
      if( activation > threshold_ce )
      {
          if( chunkExists[ itemIdx ][ WORDS_PER_ITEM - 1 ] == -1 )
          {   // target chunk doesn't exist.
              if( Math.random() < p_ce )
              {
                  chunkExists[ itemIdx ][ WORDS_PER_ITEM - 1 ] =
                   (short)gkNet.insertNode( wordItems[ itemIdx ][ WORDS_PER_ITEM - 1 ], targetLabel );
              }
          }
          else  // target chunk exists.
             gkNet.resetNodeAge( chunkExists[ itemIdx ][ WORDS_PER_ITEM - 1 ] );
      }

      // associative rule extraction
      if( chunkExists[ itemIdx ][ wordIdx ] != -1 &&
          chunkExists[ itemIdx ][ WORDS_PER_ITEM - 1 ] != -1 )
      {   // associative rule extraction
          if( activation > threshold_ae )
          {
              boolean existed = gkNet.checkLinkExist( chunkExists[ itemIdx ][ wordIdx ],
                                                      chunkExists[ itemIdx ][ WORDS_PER_ITEM - 1 ] );
              if( !existed )
              {
                   if( Math.random() < p_ae )
                       gkNet.insertLink( chunkExists[ itemIdx ][ wordIdx ],
                                         chunkExists[ itemIdx ][ WORDS_PER_ITEM - 1 ] );
              }
              else
                  gkNet.resetLinkAge( chunkExists[ itemIdx ][ wordIdx ],
                                      chunkExists[ itemIdx ][ WORDS_PER_ITEM - 1 ] );
          }
      }

      gkNet.update();
      return activation;
   }

   private String arrayToString( short[][] arr )
   {
       String str = "[ ";
       for( int i = 0; i < CHUNK_DIM_NUM; i++ )
       {
            str += "( ";
            int count = 0;
            for( int j = 0; j < arr[ i ].length; j++ )
            {
                 if( arr[ i ][ j ] == 1 )
                 {
                     if( count == 0 )
                         str += String.valueOf( j + 1 );
                     else
                         str += ", " + String.valueOf( j + 1 );
                     count++;
                 }
             }
            str += " )";
       }

       str += " ] \n";
       return str;
   }

   public void getCompatibleChunks()
   {
      double[] output = new double[ AMNET_OUTPUT_NUM ];
      amNet.getOutput( output );

      int count = 0;
      short[][] arr = new short[ CHUNK_DIM_NUM ][ CHUNK_DIM_VAL_NUM ];
      for( int i = 0; i < arr.length; i++ )
       for( int j = 0; j < arr[ 0 ].length; j++ )
       {
            int idx = i * CHUNK_DIM_VAL_NUM + j;
            if( output[ idx ] > NOISE_LEVEL )
            {
                arr[ i ][ j ] = 1;
                count++;
            }
            else
                arr[ i ][ j ] = 0;
       }

       System.out.println( "# of feature values is:" + count );
   }

   // calculate activations in the measurement of minumum activation.
   // modified on 08/20/02.
   public double calcActivations( int itemIdx )
   {
      double[] output = new double[ AMNET_OUTPUT_NUM ];
      amNet.getOutput( output );

      double activation = 1.0;
      for( int i = 0; i < CHUNK_DIM_NUM; i++ )
        for( int j = 0; j < CHUNK_DIM_VAL_NUM; j++ )
             if( wordItems[ itemIdx ][ WORDS_PER_ITEM-1 ][ i ][ j ] == 1 )
             {
                 if( output[ i * CHUNK_DIM_VAL_NUM + j ] < activation )
                     activation = output[ i * CHUNK_DIM_VAL_NUM + j ];
             }

      return activation;
   }

   public void getInputFromWM( int itemNo, int wordNo, double[] inputs )
   {
       int offset = wordNo * CHUNK_DIM_NUM * CHUNK_DIM_VAL_NUM;
       for( int i = 0; i < CHUNK_DIM_NUM; i++ )
         for( int j = 0; j < CHUNK_DIM_VAL_NUM; j++ )
         {
              inputs[ offset + i * CHUNK_DIM_VAL_NUM + j ] = inputs[ i * CHUNK_DIM_VAL_NUM + j ];
              inputs[ i * CHUNK_DIM_VAL_NUM + j ] = wordItems[ itemNo ][ wordNo ][ i ][ j ];
         }
   }

/*****************************************************************************************/
   public void getInputFromWM( int itemNo, int wordNo, double[] inputs, double[] inputsWM )
   {
       int offset = wordNo * CHUNK_DIM_NUM * CHUNK_DIM_VAL_NUM;
       for( int i = 0; i < CHUNK_DIM_NUM; i++ )
         for( int j = 0; j < CHUNK_DIM_VAL_NUM; j++ )
         {
              //inputs[ offset + i * CHUNK_DIM_VAL_NUM + j ] = inputs[ i * CHUNK_DIM_VAL_NUM + j ];
              inputs[ i * CHUNK_DIM_VAL_NUM + j ] = wordItems[ itemNo ][ wordNo ][ i ][ j ];
         }

       for( int i = 0; i < CHUNK_DIM_NUM; i++ )
         for( int j = 0; j < CHUNK_DIM_VAL_NUM; j++ )
         {
              inputsWM[ offset + i * CHUNK_DIM_VAL_NUM + j ] = inputsWM[ i * CHUNK_DIM_VAL_NUM + j ];
              inputsWM[ i * CHUNK_DIM_VAL_NUM + j ] = wordItems[ itemNo ][ wordNo ][ i ][ j ];
         }
   }
/*******************************************************************************************/

   public void storeIntoWM( int idx )
   {
      wm.setActivation( idx, 1.0 );
   }

   public void resetWM()
   {
      wm.resetAllActivations();
   }
}