package tasksimulators.oldsimu;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.Vector;
import java.util.Arrays;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;
import clarion.nacs.*;
import clarion.acs.*;

public class DSTask  extends JFrame implements SimuTask {
   // debug flag.
   final boolean debug = true;

   // the user interface
   final int WINDOW_WIDTH = 400, WINDOW_HEIGHT = 300;

   // constant variables for the experiment.
   final int AGENT_NUM = 20, ITEM_NUM = 10, WORDS_PER_ITEM = 16;
   final int DIM_NUM = 10, DIM_VAL_NUM = 10;
   final int MAX_DIM_VAL_NUM = 2, MAX_OVERLAP_NUM = 8;
   final int TRAIN_DATA_LENGTH = 20, TRAIN_INTERVAL = 50;
   final int CREATE_MODE = 0, IMPORT_MODE = 1;

   // used for word item generation.
   //int generateMode = CREATE_MODE;
   int generateMode = IMPORT_MODE;
   int ruleStats[][];
   short wordItems[][][][];
   double expData[][], expActivations[][][];
   double trainData[][], testData[][];
   String rules, chunks;
   boolean exitFlag;

   // used for user interface.
   Container c;
   JFrame mainFrame;
   TextArea arguArea1, resultArea1, showArea1;
   JButton beginButton, exitButton;

   String fullWordsFromFileName =
    "d:\\clarionexp\\CLARIONTASK\\datFiles\\DS\\data\\0802\\DSwords0.txt";
   //"d:\\clarionexp\\CLARIONTASK\\datFiles\\DSwords0.txt";
   RandomAccessFile wordsFromFile;
   WriteFile datFile, expDatFiles[];
   DSAgent dsAgent;

   public DSTask( JFrame mFrame )
   {
      wordItems = new short[ ITEM_NUM ][ WORDS_PER_ITEM ][ DIM_NUM ][ DIM_VAL_NUM ];
      expData = new double[ AGENT_NUM ][ 2 ];
      expActivations = new double[ AGENT_NUM ][ ITEM_NUM ][ WORDS_PER_ITEM - 1 ];

      // initialization of rule statistics.
      ruleStats = new int[ AGENT_NUM ][ WORDS_PER_ITEM - 1 ];

      // initialization of variables for testing performance of AMN and IDN
      trainData = new double[ AGENT_NUM ][ TRAIN_DATA_LENGTH ];
      testData = new double[ AGENT_NUM ][ WORDS_PER_ITEM - 1 ];

      // initialize user interface.
      initInterface( mFrame );
   }

   public String getTaskName()
   {
       return "DISCOVERY TASK";
   }

/***************************** PUBLIC METHOD ***************************************/
   public void start( int expMode, int blockNum, WriteFile dFile )
   {
       datFile = dFile;
       setEnabled( true );
       setVisible( true );

       // create the files for saving data.
       createExpDataFiles();
       dsAgent = new DSAgent();
   }

/***************************** PRIVATE METHOD ***************************************/
  private void initInterface( JFrame mFrame )
   {
      // setup user interface.
      c = getContentPane();
      c.removeAll();
      c.setLayout( new BorderLayout( 5, 0 ) );
      buildInterface();

      setSize( WINDOW_WIDTH, WINDOW_HEIGHT );
      setEnabled( false );
      setVisible( false );

      mainFrame = mFrame;
      addWindowListener(
          new WindowAdapter(){
             public void windowClosing( WindowEvent e )
             {
               if ( exitFlag == true )
               {
                    datFile.closeFile();
                    setVisible( false );
                    setEnabled( false );
                    mainFrame.setVisible( true );
                    mainFrame.setEnabled( true );
               }
             }
      }  );
   }

  private void buildInterface()
  {
      JPanel showPanel = new JPanel();
      showPanel.setLayout( new BorderLayout() );
      c.add( showPanel, BorderLayout.CENTER );

      JPanel capPanel = new JPanel();
      capPanel.setLayout( new FlowLayout( FlowLayout.CENTER, 100, 5 ) );
      showPanel.add( capPanel, BorderLayout.NORTH );
      JLabel taskCaption = new JLabel( "DISCOVERY TASK" );
      taskCaption.setFont( new Font( "Serif", Font.BOLD, 20 ) );
      capPanel.add( taskCaption );

      JPanel mainPanel = new JPanel();
      mainPanel.setLayout( new GridLayout( 1, 1, 5, 5 ) );
      showPanel.add( mainPanel, BorderLayout.CENTER );

      JPanel resultPanel = new JPanel();
      resultPanel.setLayout( new GridLayout( 3, 1, 5, 5 ) );
      mainPanel.add( resultPanel );
      arguArea1 = new TextArea( "WORDS IN EXPETIMENTS \n" );
      resultPanel.add( arguArea1 );
      resultArea1 = new TextArea( "THE RESULTS ARE: " );
      resultPanel.add( resultArea1 );
      showArea1 = new TextArea( "EXPERIMENT PROCESS SHOW \n" );
      resultPanel.add( showArea1 );

      JPanel buttonPanel = new JPanel();
      buttonPanel.setLayout( new FlowLayout() );
      c.add( buttonPanel, BorderLayout.SOUTH );

      beginButton = new JButton( "BEGIN" );
      beginButton.setEnabled( true );
      beginButton.addActionListener(
         new  ActionListener() {
                 public void actionPerformed( ActionEvent e )
                 {
                     beginButton.setEnabled( false );
                     exitButton.setEnabled( false );
                     exitFlag = false;
                     startExp();
                     exitButton.setEnabled( true );
                 }
        }
      );

      exitButton = new JButton( "Exit" );
      exitButton.addActionListener(
                 new  ActionListener() {
                     public void actionPerformed( ActionEvent e )
                     {
                        closeAllFiles();
                        setVisible( false );
                        setEnabled( false );
                        mainFrame.setVisible( true );
                        mainFrame.setEnabled( true );
                     }
                 }
      );

      buttonPanel.add( beginButton );
      buttonPanel.add( exitButton );
   }

   private void startExp()
   {
     if( generateMode == CREATE_MODE )
     {
         boolean flag = false;
         while( !flag )
         {
            createWordItems();
            flag = checkWordItems();
         }
         saveWordItems();
     }
     else
        importWordItems();

     dsAgent.setWordItems( wordItems );
     for( int i = 0; i < AGENT_NUM; i++ )
     {
         System.out.println( "current agent no = " + String.valueOf( i ));
         dsAgent.setAgentIdx( i );
         dsAgent.training();
         dsAgent.test();
         dsAgent.getResults( expData[ i ] );
         dsAgent.getActivations( expActivations[ i ] );
         dsAgent.getTestDataInTraining( trainData[ i ] );
         dsAgent.getTestDataInTesting( testData[ i ] );
         dsAgent.getRuleStats( ruleStats[ i ] );
         rules = dsAgent.getRules();
         chunks = dsAgent.getChunks();

         dsAgent.reinit();
         saveAgentData( i );
     }
     saveExpData();
   }

   private void createWordItems()
   {
      int[] countArr = new int[ DIM_NUM * DIM_VAL_NUM ];
      double[] inputs = new double[ DIM_VAL_NUM ];
      for( int i = 0; i < countArr.length; i++ )
           countArr[ i ] = 0;
      for( int i = 0; i < inputs.length; i++ )
           inputs[ i ] = 0.0;

      for( int i = 0; i < wordItems.length; i++ )
        for( int j = 0; j < wordItems[0].length; j++ )
          for( int k = 0; k < wordItems[0][0].length; k++ )
            for( int m = 0; m < wordItems[0][0].length; m++ )
                 wordItems[ i ][ j ][ k ][ m ] = 0;

      StochasticDecider decider = new StochasticDecider( DIM_VAL_NUM, 10 );
      for( int i = 0; i < ITEM_NUM; i++ )
       for( int j = 0; j < WORDS_PER_ITEM; j++ )
        for( int k = 0; k < DIM_NUM; k++ )
        {
             int num = 1 + (int)( Math.random() * MAX_DIM_VAL_NUM );
             int count = 0;
             for( int m = 0; m < DIM_VAL_NUM; m++ )
                  inputs[ m ] = countArr[ k * DIM_VAL_NUM + m ];
             while( count < num )
             {
                    decider.setElements( inputs );
                    int idx = decider.getRecommendation();
                    wordItems[ i ][ j ][ k ][ idx ] = 1;
                    inputs[ idx ] -= 1.0;
                    countArr[ k * DIM_VAL_NUM + idx ]--;
                    count++;
             }
        }

      System.out.println(" word sets has been created !" );
   }

   public void saveWordItems()
   {
      for( int itemIdx = 0; itemIdx < ITEM_NUM; itemIdx++ )
       for( int wordIdx = 0; wordIdx < WORDS_PER_ITEM; wordIdx++ )
       {
            String label = "";
            if( wordIdx < WORDS_PER_ITEM - 1 )
                label = "Item"+String.valueOf( itemIdx+1 )+"-Word"+String.valueOf( wordIdx+1 );
            else
                label = "Item"+String.valueOf( itemIdx+1 )+"-Word"+"target";

            String str = "word \"" + label + "\" = " ;
            str += "[ ";
            for( int i = 0; i < DIM_NUM; i++ )
            {
                 str += "( ";
                 int count = 0;
                 for( int j = 0; j < DIM_VAL_NUM; j++ )
                 {
                      if( wordItems[itemIdx][wordIdx][i][j] == 1 )
                      {
                          if( count == 0 )
                              str += String.valueOf( j + 1 );
                          else
                              str += ", " + String.valueOf( j + 1 );
                         count++;
                      }
                 }
                 str += " )";
             }
             str += " ] ";
             expDatFiles[ 1 ].saveStringData( str );
       }
   }

   public void importWordItems()
   {
       System.out.println( " import word items from a file! " );
       try {
          wordsFromFile  = new RandomAccessFile( fullWordsFromFileName, "r" );
          wordsFromFile.seek( 0 );
          while( wordsFromFile.getFilePointer() < wordsFromFile.length() )
          {
             String str = wordsFromFile.readLine();
             int len = str.length();
             if( len < 10 )
                 continue;
             int itemIdx = Integer.parseInt( str.substring( str.indexOf("Item") + "Item".length(),
                                                                         str.indexOf("-") ) ) - 1;
             int wordIdx = 0;
             String subStr = str.substring( str.indexOf("Word") + "Word".length(),
                                            str.lastIndexOf("\"" ) );
             if( subStr.compareTo( "target") != 0 )
                 wordIdx = Integer.parseInt( subStr ) - 1;
              else
                 wordIdx = WORDS_PER_ITEM - 1;

             int dimIdx = 0;
             int startIdx = str.indexOf( "(" );
             while( startIdx < str.length() )
             {
                if( startIdx == -1 )
                     break;
                int endIdx = str.indexOf( ")", startIdx );
                if( endIdx != -1 )
                {
                    String dimStr = str.substring( startIdx + 1, endIdx + 1 );
                    int[] locs = new int[ dimStr.length() ];
                    locs[ 0 ] = -1;
                    int idx = 0, count = 0;
                    while( idx < dimStr.length() )
                    {
                       if( dimStr.substring( idx, idx + 1 ).compareTo( "," ) == 0 )
                           locs[ ++count ] = idx;
                       idx++;
                    }
                    locs[ count+1 ] = dimStr.length() - 1;
                    for( int i = 0; i < count + 1; i++ )
                    {
                         int val = Integer.parseInt( dimStr.substring( locs[i]+1, locs[i+1] ).trim() );
                         wordItems[ itemIdx ][ wordIdx ][ dimIdx ][ val-1 ] = 1;
                    }

                    dimIdx++;
                    startIdx = str.indexOf( "(", endIdx );
                }
                else
                   break;
             }
          }
          wordsFromFile.close();
       }
       catch ( IOException e ) {
               JOptionPane.showMessageDialog( null, "File Operation Error",
                          "Invalid File Name", JOptionPane.ERROR_MESSAGE );
       }
       System.out.println( " finish importing word items from a file! " );
   }

   public boolean checkWordItems()
   {
      int wordNum = WORDS_PER_ITEM * ITEM_NUM;
      for( int i = 0; i < wordNum; i++ )
      {
           int itemIdx1 = i / WORDS_PER_ITEM;
           int wordIdx1 = i % WORDS_PER_ITEM;
           for( int j = i + 1; j < wordNum; j++ )
           {
                int itemIdx2 = j / WORDS_PER_ITEM;
                int wordIdx2 = j % WORDS_PER_ITEM;
                int count = 0;
                for( int k = 0; k < DIM_NUM; k++ )
                  for( int m = 0; m < DIM_VAL_NUM; m++ )
                  {
                       if( ( wordItems[ itemIdx1 ][ wordIdx1 ][ k ][ m ] == 1 ) &&
                           ( wordItems[ itemIdx2 ][ wordIdx2 ][ k ][ m ] == 1 ) )
                           count++;
                       if( count >= MAX_OVERLAP_NUM )
                           return false;
                  }
           }
      }

      System.out.println(" word sets has been created successfully!" );
      return true;
   }

   private void closeAllFiles()
   {
      datFile.closeFile();
      for( int i = 0; i < expDatFiles.length; i++ )
           expDatFiles[i].closeFile();
   }

   private void  createExpDataFiles()
   {
      String[] fileNames = { "DSexp", "DSwords", "DSamnTrain", "DSamnTest",
                             "DSchunks", "DSrules", "DSruleStat" };
      expDatFiles = new WriteFile[ fileNames.length ];
      for( int i  = 0; i < fileNames.length; i++ )
      {
           expDatFiles[ i ] = new WriteFile();
           expDatFiles[ i ].openFile( 1, fileNames[ i ] );
      }
   }

   private void saveAgentData( int agentIdx )
   {
      String str = "Current agent is: ";
      str += String.valueOf( agentIdx ) + "\n";

      /*
      str += "The index of the word which a hunch is arrived at: ";
      str += String.valueOf( expData[ agentIdx ][ 0 ] ) + "\n";
      str += "The number of clue words it took for subjects to go from a hunch to a conviction is: ";
      str += String.valueOf( expData[ agentIdx ][ 1 ] ) + "\n\n";
      */

      str += "Current activations are: \n";
      for( int j = 0; j < WORDS_PER_ITEM - 1; j++ )
      {
           double activation = 0.0;
           for( int i = 0; i < ITEM_NUM; i++ )
                activation += expActivations[ agentIdx ][ i ][ j ];
           activation /= ITEM_NUM;
           str += String.valueOf( activation ) + "\n";
      }
      expDatFiles[ 0 ].saveStringData( str );

      str = "Current agent is: ";
      str += String.valueOf( agentIdx ) + "\n";
      str += "All the rules are: \n";
      str += rules;
      expDatFiles[ 5 ].saveStringData( str );
   }

   private void saveExpData()
   {
      /*
      int count1 = 0, count2 = 0;
      double sum1 = 0.0, sum2 = 0.0;
      for( int i = 0; i < AGENT_NUM; i++ )
      {
           if( expData[ i ][ 0 ] >= 0.0 )
           {
               sum1 += expData[ i ][ 0 ];
               count1++;
           }
           if( expData[ i ][ 1 ] >= 0.0 )
           {
               sum2 += expData[ i ][ 1 ];
               count2++;
           }
      }

      if( count1 > 0 )
          sum1 /= count1;
      if( count2 > 0 )
          sum2 /= count2;

      double sd1 = 0.0, sd2 = 0.0;
      for( int i = 0; i < AGENT_NUM; i++ )
      {
           if( expData[ i ][ 0 ] >= 0.0 )
               sd1 += Math.pow( sum1 - expData[ i ][ 0 ], 2 );
           if( expData[ i ][ 1 ] >= 0.0 )
               sd2 += Math.pow( sum2 - expData[ i ][ 1 ], 2 );
      }

      sd1 = Math.sqrt( sd1 / ( count1 - 1 ) );
      sd2 = Math.sqrt( sd2 / ( count2 - 1 ) );
      */

      String str = "";
      /*
      str += "The average index of the word which a hunch is arrived at: ";
      str += "M = "+ String.valueOf( sum1 ) + ", SD = " + String.valueOf( sd1 ) + "\n";
      str += "The average number of clue words it took for subjects to go from a hunch to a conviction is: ";
      str += "M = "+ String.valueOf( sum2 ) + ", SD = " + String.valueOf( sd2 ) + "\n";
      */

      str += "\n The activation curve is: \n";
      double[] activations = new double[ WORDS_PER_ITEM - 1 ];
      for( int i = 0; i < WORDS_PER_ITEM - 1; i++ )
      {
           activations[ i ] = 0.0;
           for( int j = 0; j < AGENT_NUM; j++ )
             for( int k = 0; k < ITEM_NUM; k++ )
                  activations[ i ] += expActivations[ j ][ k ][ i ];
           activations[ i ] /= AGENT_NUM * ITEM_NUM;
           str += String.valueOf( i + 1 ) + " " + String.valueOf( activations[ i ] ) + "\n";
      }
      datFile.saveStringData( str );

      for( int j = 0; j < TRAIN_DATA_LENGTH; j++ )
      {
           double sum = 0.0;
           String str1 = String.valueOf( j + 1 ) + " ";
           for( int i = 0; i < AGENT_NUM; i++ )
                sum += trainData[ i ][ j ];
           sum /= AGENT_NUM;
           str1 += String.valueOf( sum );
           expDatFiles[ 2 ].saveStringData( str1 );
      }

      for( int j = 0; j < WORDS_PER_ITEM - 1; j++ )
      {
           double sum = 0.0;
           String str1 = String.valueOf( j + 1 ) + " ";
           for( int i = 0; i < AGENT_NUM; i++ )
                sum += testData[ i ][ j ];
           sum /= AGENT_NUM;
           str1 += String.valueOf( sum );
           expDatFiles[ 3 ].saveStringData( str1 );
      }

      str = "All the chunks are: \n";
      str += chunks;
      expDatFiles[ 4 ].saveStringData( str );

      str = "Rule statistics is: \n";
      for( int i = 0; i < WORDS_PER_ITEM - 1; i++ )
      {
           str += String.valueOf( i + 1 ) + " ";
           double num = 0.0;
           for( int j = 0; j < AGENT_NUM; j++ )
                num += ruleStats[ j ][ i ];
           num /= AGENT_NUM;
           str += String.valueOf( num ) + "\n";
      }
      expDatFiles[ 6 ].saveStringData( str );
   }
}