// Draw input & output class
package tasksimulators.oldsimu;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.Graphics;


class DrawPanel extends JPanel {

    private int  X_BASE_POINT = 40,     Y_BASE_POINT = 30,
                 XPIXELS_OF_STEP = 26,  YPIXELS_OF_STEP = 16,
                 PIXELS_OF_SCALE = 3,   LEVELS  = 12,
                 TRIALS_PER_BLOCK = 11;

     private int preXPoint, preYPoint;
     private int curXPoint, curYPoint;
     private int width, height, UNIT = 50;
     private int curNum, curChoice, curInputVal;
     private int receivedValues[];
     private int targetVal, targetX1, targetY1,targetX2, targetY2;

     public DrawPanel( int w, int h, int tVal, int choice )
     {
         width =  w;
         height = h;
         targetVal = tVal;
         curChoice = choice;

         curNum = 0;
         receivedValues = new int[ TRIALS_PER_BLOCK + 1 ];
         if( choice == 1 )
             calcTargetLine();
        // repaint();
     }

     public void setUnitVal( int val)
     {
         UNIT = val;
         repaint();
     }

     public void resetDraw()
     {
         for( int i = 0; i < TRIALS_PER_BLOCK + 1; i++)
              receivedValues[ i ] = 0;

         curNum = 0;

         preXPoint = 0;
         preYPoint = 0;
         curXPoint = 0;
         curYPoint = 0;

         repaint();
     }

     public void setHAxis( int w, int h )
     {
         preXPoint = X_BASE_POINT;
         preYPoint = h - Y_BASE_POINT;
         curXPoint = w - X_BASE_POINT;
         curYPoint = h - Y_BASE_POINT;
     }

     public void setVAxis( int w, int h )
     {
         preXPoint = X_BASE_POINT;
         preYPoint = h - Y_BASE_POINT;
         curXPoint = X_BASE_POINT;
         curYPoint = 24;
     }

     public void calcTargetLine()
     {
         targetX1 = X_BASE_POINT;
         targetY1 = height - Y_BASE_POINT - (targetVal + 1)* YPIXELS_OF_STEP;
         targetX2 = width - X_BASE_POINT;
         targetY2 = height - Y_BASE_POINT - (targetVal + 1) * YPIXELS_OF_STEP;
     }

    public void drawTargetLine( int val )
    {
       targetVal = val;
       calcTargetLine();
       repaint();
    }

    public void setScalePoints( int mode, int iStep )
    {
        if ( mode == 0) {
             preXPoint = X_BASE_POINT + iStep * XPIXELS_OF_STEP;
             preYPoint = height - Y_BASE_POINT;
             curXPoint = X_BASE_POINT + iStep * XPIXELS_OF_STEP;
             curYPoint = height - Y_BASE_POINT - PIXELS_OF_SCALE;
        }
        else {
              preXPoint = X_BASE_POINT;
              preYPoint = height - Y_BASE_POINT - iStep * YPIXELS_OF_STEP;
              curXPoint = X_BASE_POINT + PIXELS_OF_SCALE;
              curYPoint = height - Y_BASE_POINT - iStep * YPIXELS_OF_STEP;
        }
     }

     public void setResultPoint( int iStep )
     {
        curXPoint = X_BASE_POINT + ( iStep + 1 )* XPIXELS_OF_STEP;
        curYPoint = height -Y_BASE_POINT - receivedValues[ iStep ]
                                         * YPIXELS_OF_STEP / UNIT;

        if( iStep > 0 ) {
               preXPoint = X_BASE_POINT + iStep * XPIXELS_OF_STEP;
               preYPoint = height - Y_BASE_POINT - receivedValues[ iStep  - 1 ]
                                        * YPIXELS_OF_STEP / UNIT;
        }
        else {
               preXPoint = curXPoint;
               preYPoint = curYPoint;
        }
     }

     public void showValue( int val2 )
     {
         //curInputVal = val1;
         receivedValues[ curNum ] = val2;
         repaint();
         curNum++;
     }

     public void paintComponent( Graphics g )
     {
         super.paintComponent( g );

         setHAxis( width, height );
         g.drawLine( preXPoint, preYPoint, curXPoint, curYPoint );
         setVAxis( width, height );
         g.drawLine( preXPoint, preYPoint, curXPoint, curYPoint );

         if ( curChoice == 0 )
              g.drawString( "Input UNIT = " + String.valueOf( UNIT ), 15, 10 );
         else
         {
              g.drawString( "PSI   UNIT = " + String.valueOf( UNIT ), 15, 10 );
              g.drawLine( targetX1, targetY1,targetX2, targetY2 );
         }

        for ( int i = 0; i < TRIALS_PER_BLOCK; i++ ) {
               setScalePoints( 0, i + 1 );
               g.drawLine( preXPoint, preYPoint, curXPoint, curYPoint );
               g.drawString( String.valueOf( i ), curXPoint - 5, curYPoint  + 20 );
         }

        for ( int i = 0; i < LEVELS; i++ ) {
               setScalePoints( 1, i + 1 );
               g.drawLine( preXPoint, preYPoint, curXPoint, curYPoint );
               if( i < 9 )
                  g.drawString( String.valueOf( i + 1 ),
                                   curXPoint - 20, curYPoint + 5);
               else
                  g.drawString( String.valueOf( i + 1 ),
                                   curXPoint - 24, curYPoint + 5);
         }

         g.setColor ( new Color(100, 100, 200) );
         g.setFont ( new Font("Serif", Font.BOLD, 12) );

         g.drawString( "Trials", 15, height - Y_BASE_POINT  + 18 );

         if ( curChoice == 0 )
              g.drawString( "Input", 15, height + 5 );
         else
              g.drawString( "Output", 15, height + 5 );

         if ( curNum == 0 && receivedValues[ 0 ] == 0 )
              return;

         for ( int i = 0; i < curNum; i++ ) {
               setResultPoint( i );
               g.drawOval( curXPoint - 2, curYPoint - 2, 4, 4 );
               g.fillOval( curXPoint - 2, curYPoint - 2, 4, 4 );
               if ( i < curNum && i > 0 )
                    g.drawLine( preXPoint, preYPoint, curXPoint, curYPoint );
         }

        int value = 0;
        if ( curNum > 0 )
            value = receivedValues[ curNum - 1 ];

        if ( curNum == 1 )
             g.drawString( String.valueOf( value ), curXPoint - 10, height + 5);
        else if ( value == UNIT )
                  g.drawString( String.valueOf( value ), curXPoint - 5, height + 5 );
             else
                  g.drawString( String.valueOf( value ), curXPoint - 10,  height + 5 );

      }
}
