 /******************** TO REALIZE THE FUNCTION OF GOAL STACK*************/
 package tasksimulators.oldsimu;
 import java.util.Stack;
 import clarion.common.*;
 import clarion.tools.*;

 public class GS {

    final int WM_ACTION = 0, EX_ACTION = 1, GS_ACTION = 2;
    final int PUSH_TYPE = 0, POP_TYPE = 1;
    final int TOP_LEVEL = 0, DOWN_LEVEL = 1;

    int GS_NET_INFO[];
    //int GS_NET_INFO[] = { 8, 30, 60, 13 };
    int DIM_VAL_INFO[];
    //int DIM_VAL_INFO[] = { 6, 3, 3, 3, 3, 3, 3, 6 };

    int MAX_TOW_SIZE, DIM_NUM, DIM_VAL_NUM;
    int preState[], curState[], gsState[];
    int action, levelFlag;
    int activeItem[], goal[];

    double gsQValue1, gsQValue2;
    double feedback, threshold, qbpArray[];

    Stack gs;
    QBackPropNet  gsNet;
    StochasticDecider actionDecider;

    public GS( int mSize, int dimNum, int gsStateLen,
                      int[] netInfoArr, int[] dimInfoArr )
    {
       MAX_TOW_SIZE = mSize;
       DIM_NUM = dimNum;

       activeItem = new int[ DIM_NUM ];

       gsState = new int[ gsStateLen ];
       preState = new int[ netInfoArr[1] ];
       curState = new int[ netInfoArr[1] ];

       GS_NET_INFO = new int[ netInfoArr.length ];
       for( int i = 0; i < netInfoArr.length; i++ )
            GS_NET_INFO[ i ] = netInfoArr[ i ];

       DIM_VAL_INFO = new int[ dimInfoArr.length ];
       for( int i = 0; i < netInfoArr.length; i++ )
            DIM_VAL_INFO[ i ] = dimInfoArr[ i ];

       DIM_VAL_NUM = DIM_VAL_INFO[ 0 ];

       initGS();
       initQbpNet();
    }

    public void initGS()
    {
       goal = new int[ DIM_NUM ];
       goal[ 0 ] = MAX_TOW_SIZE;
       goal[ 1 ] = 0;
       goal[ 2 ] = 2;
       goal[ 3 ] = MAX_TOW_SIZE + 1;
       goal[ 4 ] = 0;
       goal[ 5 ] = 1;

       gs = new Stack();
       gs.push( goal );
    }

    public void initQbpNet()
    {
       qbpArray = new double[ GS_NET_INFO[3] ];
       gsNet = new QBackPropNet( GS_NET_INFO[1], GS_NET_INFO[2],
                                 GS_NET_INFO[3], GS_NET_INFO[0], DIM_VAL_INFO );

       actionDecider = new StochasticDecider( GS_NET_INFO[3],
                           World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );
    }

    public void reinit( int tSize )
    {
       reinitGS( tSize );
       reinitQbpNet();
    }

    public void reinitGS( int tSize )
    {
        MAX_TOW_SIZE = tSize;
        goal[ 0 ] = MAX_TOW_SIZE;
        goal[ 1 ] = 0;
        goal[ 2 ] = 2;
        goal[ 3 ] = MAX_TOW_SIZE + 1;
        goal[ 4 ] = 0;
        goal[ 5 ] = 1;
        gs.push( goal );
    }

    public void reinitQbpNet()
    {
        gsNet.reinit();
    }

    public double getSuccRate()
    {
        double succRate = gsNet.getSuccRate();
        return succRate;
    }

    public int getAction( int[] inputs )
    {
        gsNet.setCurState( inputs );
        gsNet.setInput();
        gsNet.forwardPass();
	gsNet.getQvals( qbpArray );
        actionDecider.setElements( qbpArray );

        // recommended action from the bottom level
        action = actionDecider.getRecommendation();
        return action;
    }

    public boolean getState( int[] arr )
    {
        boolean bVal = getActiveItem();
        if( !bVal )
             return bVal;

        for( int i = 0; i < gsState.length; i++ )
             gsState[ i ] = 0;

        int offset = 0;
        for( int i = 0; i < activeItem.length; i++ )
        {
             gsState[ offset + activeItem[ i ] ] = 1;
             offset += DIM_VAL_NUM;
        }

        for( int i = 0; i < arr.length; i++ )
             arr[ i ] = gsState[ i ];

        return bVal;
    }

    public boolean getActiveItem( )
    {
       if( gs.isEmpty() )
           return false;
       else
       {
          activeItem = ( int[] )gs.peek();
          return true;
       }
    }

    public boolean getActiveItem( int[] arr )
    {
       if( gs.isEmpty() )
           return false;
       else
       {
          activeItem = ( int[] )gs.peek();
          for( int i = 0; i < arr.length; i++ )
               arr[ i ] = activeItem[ i ];
          return true;
       }
    }

    public void setLevelFlag( int level )
    {
       levelFlag = level;
    }

    public void setFeedback( double val )
    {
        feedback = val;
    }

    public void setPreState( int[] state )
    {
       for( int i = 0; i < state.length; i++ )
            preState[ i ] = state[ i ];
    }

    public void setCurState( int[] state )
    {
       for( int i = 0; i < state.length; i++ )
            curState[ i ] = state[ i ];
    }

    public boolean checkAction( int act )
    {
       return true;
    }

    public boolean checkAction( int[] arr )
    {
       return true;
    }

    public void performAction( int[] arr )
    {
      if( arr[ 0 ] != GS_ACTION )
          return;

      if( arr[ 1 ] == PUSH_TYPE )
      {
          int[] goal = new int[ arr.length - 2 ];
          for( int i = 0; i < goal.length; i++ )
               goal[ i ] = arr[ i + 2 ];
          gs.push( goal );
      }

      if( arr[ 1 ] == POP_TYPE )
      {
          boolean bVal = gs.empty();
          if( !bVal )
              gs.pop();
      }
    }

    public void updateAction( int chosenAction )
    {
        action = chosenAction;
    }

    public void illegalUpdate()
    {
       gsQValue1 = gsNet.getQVal( action );

       if( levelFlag == DOWN_LEVEL )
       {
           gsNet.updateMatchCount( false );
           gsNet.update( action, feedback, gsQValue1 );
       }
    }

    public void legalUpdate()
    {
        gsQValue1 = gsNet.getQVal( action );

       //obtain ngst state QValue2.
	gsNet.setCurState( curState );
	gsNet.setInput();
        gsNet.forwardPass();
        gsQValue2 = gsNet.getMaxQVal();

/******************************************************************************/
        // update the QBP network.
        gsNet.update( action, feedback, gsQValue1, gsQValue2 );
/******************************************************************************/

        // update RER rule set if the RER flag is set.
        double r = gsNet.getReinforcement();
        double d = gsNet.getDiscount();

        if( r + d * gsQValue2 - gsQValue1 > threshold )
           gsNet.updateMatchCount(  true );
        else
           gsNet.updateMatchCount( false );
    }

    public void update( boolean flag )
    {
      if( flag )
          legalUpdate();
      else
          illegalUpdate();
    }
 }
