/* the LETTER COUNTING AGENT FOR LETTER COUNTING TASK  */
package tasksimulators.oldsimu;
import java.lang.*;
import java.util.Vector;
import clarion.common.*;
import clarion.tools.*;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.coordinate.*;

public class LCAgent {

    final int SHOW_INTERVAL = 50;
    final int WITH_BBT = 0, WITHOUT_BBT = 1;
    final boolean debug = true;
    final double ACTIVATION_THRESHOLD = 0.5;

    // used for combination of BL and TL.
    final int COMB_COMPONENT_NUM = 4;
    final int BL_IDX = 0, RER_IDX = 1, IRL_IDX = 2, FIX_IDX = 3;
    final int FIX_MODE = 0, VARIED_MODE = 1, WS_COMB = 0, STOCH_COMB = 1;

    final int ACS_SIM = 0, NACS_SIM = 1;   // for simulation.
    final double POS_REWARD = 1.0, NEG_REWARD = 0.0; // for QBP.

    // for each component.
    int COUNT_TYPE = 0, RETRIEVE_TYPE = 1;               // FOR FIXED RULES.
    int NO_MERGE = 0, TEN_MERGE = 1, EACH_MERGE = 2;     // FOR RER.
    int TOP_LEVEL = 0, DOWN_LEVEL = 1;

    // used for chunks.
    final int CHUNK_ROW_NUM = 6, CHUMK_COLUM_NUM = 6;
    final int MAX_CHUNK_AGE = 250, MAX_LINK_AGE = 250;
    final double CHUNK_ENCODING_PROB = 0.4;

    // used for different simulations.
    int BBT_IDX = WITHOUT_BBT;

    // used for calculate RT.
    double FAIL_RETRIEVAL_TIME = 500;
    double INIT_BLA = 0.06;
    double PRE_CHUNK_BLA = 10.0;
    double RER_RULE_ACTION_TIME = 200;
    double RER_RULE_COST = 0.05;
    double[][] RER_RULE_PRE_BLAS =
    {
        { 15.0, 150.0 },  // with BBT.
        { 10.0, 100.0 }   // without BBT.
    };
    double[] FR_RULE_ACTION_TIMES = { 0, 0, 200, 200, 200, 200 };
    double[][] FR_RULE_COSTS =
    {
        { 0.6, 0.6, 0.54, 0.54, 0.54, 0.54 }, //with BBT
        { 0.6, 0.6, 0.4, 0.4, 0.4, 0.4 }      //without BBT
    };
    double[][][] FR_RULE_PRE_BLAS =
    {
       {
          { 800, 800, 10, 10, 10, 10 },  // used for simu1  with BBT.
          { 800, 800, 10, 10, 10, 10 },  // used for simu2  with BBT.
          { 700, 700, 10, 10, 10, 10 },  // used for simu1' with BBT.
          { 700, 700, 10, 10, 10, 10 }   // used for simu2' with BBT.
       },
       {
          { 660, 660, 10, 10, 10, 10 },  // used for simu1  without BBT.
          { 660, 660, 10, 10, 10, 10 },  // used for simu2  without BBT.
          { 700, 700, 10, 10, 10, 10 },  // used for simu1' without BBT.
          { 660, 660, 10, 10, 10, 10 }   // used for simu2' without BBT.
       }
    };

    //for training.
    int[][][] BTWN_TRIAL_NUMS =
    {
      { { 2000, 14000 }, { 1000, 7000 } }, // used for with BBT.
      { { 0, 0 }, { 0, 0 }  }              // used for without BBT.
    };

    int PRE_TRAIN_BLOCK = 1, PRE_TRAIN_PROB_NUM = 36;

    int simNo, expIdx, groupIdx, agentIdx, blockNum, trialNum;
    int stageIdx, blockIdx, trialIdx, concDimIdx, levelFlag;
    int fixRuleNum, fixRuleType, reasoningType;
    int chunkDimNum, chunkDimTypes[], chunkDVNums[];
    int BLOCK_NUMS[], TRIAL_NUMS[], probSequ[];

    // used for BL.
    int NET_DIM_NUM, NET_INPUT_NUM;
    int NET_HIDDEN_NUM, NET_OUTPUT_NUM;
    int DIM_VAL_NUM;

    // used for combination of BL and TL.
    int combMode1, combMode2, actionTypeNum;
    short actionNumArr[];
    //int action, actionBL, actionRER, actionIRL, actionFR, chosenComp;
    int chosenComp, chosenAction, chosenActions[], actionResult[];
    int dimValNums[], curState[], nodeInfoArr[];

    // problems arrays.
    int answer, responseTime, CURRENT_TIME, TOT_TRIAL_NUM;
    int preTrainProbArr[][], actTrainProbArr[][], wholeProbSet[][];
    int trainProbArr[][], transProbArr[][], trainRTArr[], transRTArr[];

    // used for NACS.
    short dims[][];

    // used for testing.
    int upCounts[][][], downCounts[][][], rerRuleUsedNums[][][], fixRuleUsedNums[][][][];

/***********************************************************************/
    int onTargetArr[][];
    double onTargetRates[][];
/***********************************************************************/
    boolean PRE_TRAIN_FLAG, BTWN_TRAIN_FLAG, FR_FLAG, RER_FLAG;
    boolean BL_RETRIEVAL_FLAG, RER_RETRIEVAL_FLAG, CHUNK_RETRIEVAL_FLAG;

    double PRIOR_TARGET_RATE[] = { 0.1, 0.1 };
    double trainPossArr[], transPossArr[], trainRetrArr[], transRetrArr[];
    double trainContArr[], transContArr[], trainBttmArr[], transBttmArr[];
    double trainRateArr[], transRateArr[];
    int trainBLretr1[], trainBLretr2[];

    double qValue, feedback, feedbacks[], BLA[], betaArr[][];
    double compSuccRates[][], activationArr[][][];

    String add = "+";
    String sub = "-";

    String count = "COUNT";
    String retrieve = "RETRIEVE";

    String nameArr[][] =
    {
       { "ADDITION", "START_LETTER", "STEP" },
       { "SUBTRACTION", "START_LETTER", "STEP"  }
    };

    // used to store recommended actions. use Boltzmann distribution to
    // decide probabilitically on an action to be performed.
    StochasticDecider actionDecider;  // action decider
    QBackPropNet qbpNet;      // bottom level network

    Vector      retrievalCue;
    ConItem[]   countConclusion;
    ConItem[]   retrieveConclusion;

    LCFixRuleSet  fixRuleSet;   // top level rule set.
    T_RerRuleSet  rerRuleSet;  // top level rule set.

    GS          goalStack;
    GKNet       gkNet;

    StochasticComb stochCombiner;
    WeightSumComb wsCombinder;

    StringBuffer fixRuleStrings[];
    StringBuffer rerRuleStrings[];

    public LCAgent( int sNo, int rNum, int rType, int totNum, boolean frFlag, boolean rerFlag,
                    int[] gkDVNums, int[] qDVNums, int[] dTypes, LCFixRule[] rSet )
    {
           simNo = sNo;
           fixRuleNum = rNum;
           reasoningType = rType;
           TOT_TRIAL_NUM = totNum;
           FR_FLAG = frFlag;
           RER_FLAG = rerFlag;

           // used for testing.
           upCounts = new int[ 2 ][ 2 ][ 2 ];
           downCounts = new int[ 2 ][ 2 ][ 2 ];
           fixRuleUsedNums = new int[ 2 ][ 2 ][ 2 ][ fixRuleNum ];
           rerRuleUsedNums = new int[ 2 ][ 2 ][ 2 ];

           actionTypeNum = 1;
           actionResult = new int[ actionTypeNum ];
           chosenActions = new int[ COMB_COMPONENT_NUM ];
           feedbacks = new double[ COMB_COMPONENT_NUM ];

           // NACS initialization.
           if( simNo == NACS_SIM )
           {
               chunkDimNum = gkDVNums.length;
               chunkDVNums = new int[ chunkDimNum ];
               for( int i = 0; i < chunkDimNum; i++ )
                    chunkDVNums[ i ] = gkDVNums[ i ];

               chunkDimTypes = new int[ chunkDimNum ];
               for( int i = 0; i < chunkDimNum; i++ )
                    chunkDimTypes[ i ] = dTypes[ i ];

               dims = new short[ chunkDimNum ][];
               for( int i = 0; i < chunkDimNum; i++ )
                    dims[i] = new short[ chunkDVNums[i] ];

               retrievalCue = new Vector( chunkDimNum );
               gkNet = new GKNet( CHUNK_ROW_NUM, CHUMK_COLUM_NUM, chunkDimNum,
                                  reasoningType, gkDVNums, chunkDimTypes, INIT_BLA );
               gkNet.setMaxAges( MAX_CHUNK_AGE, MAX_LINK_AGE );
               gkNet.setActivationThreshold( ACTIVATION_THRESHOLD );
           }

           // bottom level initialization.
           NET_DIM_NUM = ( int )World.TASK_PARA_ARRAY[World.NET_DIM_NUM ];
           NET_INPUT_NUM = (int)World.TASK_PARA_ARRAY[World.NET_INPUT_NUM ];
           NET_HIDDEN_NUM =(int)World.TASK_PARA_ARRAY[World.NET_HIDDEN_NUM ];
           NET_OUTPUT_NUM = (int)World.TASK_PARA_ARRAY[World.NET_OUTPUT_NUM];
           if( simNo == NACS_SIM && !FR_FLAG && RER_FLAG )
               NET_OUTPUT_NUM++;

           DIM_VAL_NUM = NET_INPUT_NUM / NET_DIM_NUM;

	          actionDecider = new StochasticDecider( NET_OUTPUT_NUM,
                             World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );

           curState = new int[ NET_INPUT_NUM ];
           dimValNums = new int[ qDVNums.length ];
           for( int i = 0; i < qDVNums.length; i++ )
                dimValNums[ i ] = qDVNums[ i ];

       	   qbpNet = new QBackPropNet( NET_INPUT_NUM, NET_HIDDEN_NUM,
                                      NET_OUTPUT_NUM, NET_DIM_NUM, qDVNums );

           countConclusion = new ConItem[ chunkDimNum ];
           retrieveConclusion = new ConItem[ chunkDimNum + 1 ];
           for( int i = 0; i < chunkDimNum + 1; i++ )
           {
                Vector item = new Vector( 2 );
                item.add( 0, "dim_name" + i );
                int[] elem = new int[1];
                elem[ 0 ] = 0;
                item.add( 1, elem );
                retrieveConclusion[ i ] = new ConItem( item );
           }

           // fixed rules initialization.
           if( FR_FLAG )
           {
               fixRuleSet = new LCFixRuleSet( World.LC_TASK, fixRuleNum, rSet );

               fixRuleStrings = new StringBuffer[3];
               for( int i = 0; i < 3; i++ )
                    fixRuleStrings[ i ] = new StringBuffer();

               fixRuleSet.setPreChunkBLA( PRE_CHUNK_BLA );
               fixRuleSet.setRuleCosts( FR_RULE_COSTS[ BBT_IDX ] );
               fixRuleSet.setRuleActionTimes( FR_RULE_ACTION_TIMES );
               if( !RER_FLAG )
                   if( simNo == ACS_SIM )
                       fixRuleSet.setRulePreBLAs( FR_RULE_PRE_BLAS[ BBT_IDX ][ 0 ] );
                   else
                       fixRuleSet.setRulePreBLAs( FR_RULE_PRE_BLAS[ BBT_IDX ][ 1 ] );
               else
                   if( simNo == ACS_SIM )
                       fixRuleSet.setRulePreBLAs( FR_RULE_PRE_BLAS[ BBT_IDX ][ 2 ] );
                   else
                       fixRuleSet.setRulePreBLAs( FR_RULE_PRE_BLAS[ BBT_IDX ][ 3 ] );
           }

           // RER initialization.
           if( RER_FLAG )
           {
               nodeInfoArr = new int[ 5 ];
               nodeInfoArr[ 0 ] = NET_DIM_NUM;
               nodeInfoArr[ 1 ] = NET_DIM_NUM;
               nodeInfoArr[ 2 ] = NET_INPUT_NUM;
               nodeInfoArr[ 3 ] = NET_HIDDEN_NUM;
               nodeInfoArr[ 4 ] = NET_OUTPUT_NUM;

               rerRuleSet = new T_RerRuleSet( 0, NO_MERGE, nodeInfoArr, true );

               rerRuleStrings = new StringBuffer[3];
               for ( int i = 0; i < 3; i++ )
                     rerRuleStrings[i] = new StringBuffer();

               rerRuleSet.setInitBLA( INIT_BLA );
               rerRuleSet.setRuleCosts( RER_RULE_COST );
               rerRuleSet.setRuleActionTimes( RER_RULE_ACTION_TIME );
               if( FR_FLAG )
                   rerRuleSet.setRulePreBLAs( RER_RULE_PRE_BLAS[ BBT_IDX ][ 0 ] );
               else
                   rerRuleSet.setRulePreBLAs( RER_RULE_PRE_BLAS[ BBT_IDX ][ 1 ] );

               rerRuleSet.setTotTrialNum( TOT_TRIAL_NUM );
               rerRuleSet.setRTFlag( true );
           }

           // initialization of variables for combination of TL and BL.
           actionNumArr  = new short[ actionTypeNum ];
           actionNumArr[ 0 ] = (short)NET_OUTPUT_NUM;

           betaArr       = new double[ actionTypeNum ][ COMB_COMPONENT_NUM ];
           compSuccRates = new double[ actionTypeNum ][ COMB_COMPONENT_NUM ];
           activationArr = new double[actionTypeNum][COMB_COMPONENT_NUM][NET_OUTPUT_NUM];

           if( World.MODEL_PARA_ARRAY[ World.CROSS_COMB_OPT ][ 0 ] == 1 )
           {
               combMode2 = FIX_MODE;
               if( World.MODEL_PARA_ARRAY[ World.FIX_COMB_OPT ][ 0 ] == 1 )
                   combMode1 = WS_COMB;
               else
                   combMode1 = STOCH_COMB;
           }
           else
           {
               combMode2 = VARIED_MODE;
               if( World.MODEL_PARA_ARRAY[ World.VAR_COMB_OPT ][ 0 ] == 1 )
                   combMode1 = WS_COMB;
               else
                   combMode1 = STOCH_COMB;
           }

           if( combMode1 == WS_COMB )
           {
               for( int i = 0; i < COMB_COMPONENT_NUM; i++ )
                    betaArr[ 0 ][ i ] = World.TASK_PARA_ARRAY[ World.WBL + i ];
               wsCombinder = new WeightSumComb( combMode2, actionNumArr[0], betaArr[0],
                                                World.TASK_PARA_ARRAY[World.TEMPERATURE] );
           }
           else
           {
               for( int i = 0; i < COMB_COMPONENT_NUM; i++ )
                    betaArr[ 0 ][ i ] = World.TASK_PARA_ARRAY[ World.PBL + i ];
               stochCombiner = new StochasticComb( combMode2, COMB_COMPONENT_NUM, betaArr[0] );
           }
   }

   public double[][] getInitIHWeights()
   {
      return qbpNet.getInitInputToHiddenWeights();
   }

   public double[][] getInitHOWeights()
   {
      return qbpNet.getInitHiddenToOutputWeights();
   }

   public void setInitWeights( double[][] ItoHWeights, double[][] HtoOWeights  )
   {
      qbpNet.setInitInputToHiddenWeights( ItoHWeights );
      qbpNet.setInitHiddenToOutputWeights( HtoOWeights );
   }

   public double[][] getIHWeights()
   {
      return qbpNet.getInputToHiddenWeights();
   }

   public double[][] getHOWeights()
   {
      return qbpNet.getHiddenToOutputWeights();
   }

   public void getThresholds( double hThresholds[], double oThresholds[] )
   {
      qbpNet.getHiddenThresholds( hThresholds );
      qbpNet.getOutputThresholds( oThresholds );
   }

  /*************************************************************************/
  /**** the following functions are served as simulating the process *******/
  /**********************  of thinking ofa subject *************************/
   public void initAgent()
   {
      CURRENT_TIME = 0;
   }

   public void reinitAgent()
   {
       initAgent();
       qbpNet.reinit();

       if( FR_FLAG )
           fixRuleSet.reinit();

       if( RER_FLAG )
           rerRuleSet.reinit();

       if( simNo == NACS_SIM )
           gkNet.reinit();
   }

   public void setWholeProbSet( int[][] probSet )
   {
     int len1 = probSet.length;
     int len2 = probSet[0].length;

     wholeProbSet = new int[ len1 ][ len2 ];
     for( int i = 0; i < len1; i++ )
       for( int j = 0; j < len2; j++ )
            wholeProbSet[i][j] = probSet[i][j];
   }

   public void setVariables( int eNo, int gNo, int num1, int num2,
                             int[] arr1, int[] arr2 )
   {
       expIdx = eNo;
       groupIdx = gNo;

       trainRTArr = new int[ num1 ];
       transRTArr = new int[ num2 ];

       trainPossArr = new double[ num1 ];
       transPossArr = new double[ num2 ];
       trainRetrArr = new double[ num1 ];
       transRetrArr = new double[ num2 ];
       trainContArr = new double[ num1 ];
       transContArr = new double[ num2 ];
       trainBttmArr = new double[ num1 ];
       transBttmArr = new double[ num2 ];
       trainRateArr = new double[ num1 ];
       transRateArr = new double[ num2 ];

       trainBLretr1 = new int[ 2 ];
       trainBLretr2 = new int[ 2 ];

       BLOCK_NUMS = new int[2];
       for( int i = 0; i < arr1.length; i++ )
            BLOCK_NUMS[ i ] = arr1[ i ];
       TRIAL_NUMS = new int[2];
       for( int i = 0; i < arr2.length; i++ )
            TRIAL_NUMS[ i ] = arr2[ i ];
   }

   public void setAgentNo( int no )
   {
       agentIdx = no;
   }

   public void obtainProblems( int[][] trainProbs, int[][] transProbs )
   {
      int len1 = trainProbs.length;
      int len2 = trainProbs[0].length;
      trainProbArr = new int[ len1 ][ len2 ];

      for( int i = 0; i < len1; i++ )
       for( int j = 0; j < len2; j++ )
            trainProbArr[ i ][ j ] = trainProbs[ i ][ j ];

      len1 = transProbs.length;
      len2 = transProbs[0].length;
      transProbArr = new int[ len1 ][ len2 ];

      for( int i = 0; i < len1; i++ )
       for( int j = 0; j < len2; j++ )
            transProbArr[ i ][ j ] = transProbs[ i ][ j ];
   }

   public void initNewStage()
   {
    /***********************************************************************/
     if( simNo == ACS_SIM && expIdx == 1 )
     {
         qbpNet.resetPM();
         qbpNet.resetNM();
         if( FR_FLAG )
             fixRuleSet.reinit();
         if( RER_FLAG )
             rerRuleSet.reinit();
     }
    /***********************************************************************/
   }

   public void initPreTrainProbArr()
   {
      int len1 = PRE_TRAIN_PROB_NUM;
      int len2 = 2;
      preTrainProbArr = new int[ len1 ][ len2 ];

      int MAX_PROB_NUM = wholeProbSet.length;
      Vector probInfoVect = new Vector( MAX_PROB_NUM );
      for( int i = 0; i < MAX_PROB_NUM; i++ )
      {
           int[] item = new int[1];
           item[0] = i;
           probInfoVect.addElement( item );
      }

      for( int i = 0; i < len1; i++ )
      {
           int range = probInfoVect.size();
           int val = (int)( Math.random() * range );
           int idx = ((int[])( probInfoVect.elementAt( val ) ))[0];
           preTrainProbArr[ i ][ 0 ] = wholeProbSet[idx][0];
           preTrainProbArr[ i ][ 1 ] = wholeProbSet[idx][1];
           probInfoVect.removeElementAt( val );
      }
   }

   public void setActiveTrainProbSet( int mode )
   {
       if( mode == 0 )
       {
         int len1 = preTrainProbArr.length;
         int len2 = preTrainProbArr[0].length;
         actTrainProbArr = new int[ len1 ][ len2 ];
         for( int i = 0; i < len1; i++ )
          for( int j = 0; j < len2; j++ )
               actTrainProbArr[i][j] = preTrainProbArr[i][j];
       }
       else
       {
         int len1 = trainProbArr.length;
         int len2 = trainProbArr[0].length;
         actTrainProbArr = new int[ len1 ][ len2 ];
         for( int i = 0; i < len1; i++ )
          for( int j = 0; j < len2; j++ )
               actTrainProbArr[i][j] = trainProbArr[i][j];
       }
   }

   public void setProbSequ( int num )
   {
      Vector vect = new Vector( num );
      for( int i = 0; i < num; i++ )
      {
         int[] arr = new int[1];
         arr[0] = i;
         vect.addElement( arr );
      }

      probSequ = new int[ num ];
      for( int i = 0; i < num; i++ )
      {
           int len = vect.size();
           int idx = (int)( Math.random() * len );
           probSequ[ i ] = ( (int[]) vect.elementAt( idx ) )[0];
           vect.removeElementAt( idx );
      }

      /*
      probSequ = new int[ num ];
      for( int i = 0; i < num; i++ )
           probSequ[ i ] = i;
      */
   }

   public void solveProblems()
   {
       getPriorKnowledge();
       training();
   }

   public void getPriorKnowledge()
   {
      CURRENT_TIME = 0;
      stageIdx = 0;
      blockIdx = 0;
      double rate = 0.0;

      initPreTrainProbArr();
      setActiveTrainProbSet( 0 );
      PRE_TRAIN_FLAG = true;
      for( int i = 0; i < PRE_TRAIN_BLOCK; i++ )
      {
          int hitNum = 0;
          trialNum = PRE_TRAIN_PROB_NUM;

          setProbSequ( trialNum );
          for( trialIdx = 0; trialIdx < trialNum; trialIdx++ )
          {
               int probIdx = probSequ[ trialIdx ];
               ConItem[] perceptData = new ConItem[3];
               int[] problem = new int[3];
               perceptProblem( problem, perceptData, probIdx );

               setTime();
               decideAction( problem, perceptData );
               calcFeedbacks();
               if( feedbacks[ BL_IDX ] > 0 )
                   hitNum++;

               calcRT( problem );
               updateTime();
               updateModel( problem );
          }

/***********************************************************************
          rate = (double)hitNum / trialNum;
          if( (++blockIdx) % 100 == 0 )
          {
             System.out.print( " blockIdx = " );
             System.out.print( blockIdx );
             System.out.print( ",   rate = " );
             System.out.println( rate );
          }
/***********************************************************************/
      }
      PRE_TRAIN_FLAG = false;
   }

   public void betweenBlockTraining()
   {
      int BTWN_TRIAL_NUM, hitNum = 0;
      if( !RER_FLAG )
          BTWN_TRIAL_NUM = BTWN_TRIAL_NUMS[ BBT_IDX ][ 0 ][ groupIdx ];
      else
          BTWN_TRIAL_NUM = BTWN_TRIAL_NUMS[ BBT_IDX ][ 1 ][ groupIdx ];

      BTWN_TRAIN_FLAG = true;
      for( trialIdx = 0; trialIdx < BTWN_TRIAL_NUM; trialIdx++ )
      {
           int probIdx = (int)( Math.random() * trialNum );
           ConItem[] perceptData = new ConItem[3];
           int[] problem = new int[3];

           perceptProblem( problem, perceptData, probIdx );
           decideAction( problem, perceptData );

           // get feedback from performing the chosen action.
           calcFeedbacks();
           if( feedbacks[ BL_IDX ] > 0 )
               hitNum++;

           updateModel( problem );
       }
       BTWN_TRAIN_FLAG = false;

       if( debug )
           debugBBT();
   }

   private void debugBBT()
   {
     /***********************************************************************
      double rate = (double)hitNum / BTWN_TRIAL_NUM;
      System.out.print( " blockIdx = " );
      System.out.print( blockIdx );
      System.out.print( ",   rate = " );
      System.out.println( rate );
     /***********************************************************************/
   }

   public void training()
   {
       onTargetArr = new int[2][];
       onTargetRates = new double[2][];
       BLA = new double[ 2 ];

       BTWN_TRAIN_FLAG = false;
       setActiveTrainProbSet( 1 );
       for( stageIdx = 0; stageIdx < 2; stageIdx++ )
       {
          blockNum = BLOCK_NUMS[stageIdx];
          trialNum = TRIAL_NUMS[stageIdx];
          onTargetArr[ stageIdx ] = new int [blockNum];
          onTargetRates[ stageIdx ] = new double [blockNum];

          for( blockIdx = 0; blockIdx < blockNum; blockIdx++ )
          {
               setProbSequ( TRIAL_NUMS[ stageIdx ] );
               for( trialIdx = 0; trialIdx < trialNum; trialIdx++ )
               {
                    int probIdx = probSequ[ trialIdx ];
                    ConItem[] perceptData = new ConItem[3];
                    int[] problem = new int[3];
                    perceptProblem( problem, perceptData, probIdx );

                    setTime();
                    decideAction( problem, perceptData );
                    calcFeedbacks();
                    if( feedbacks[ BL_IDX ] > 0 )
                        onTargetArr[ stageIdx ][ blockIdx ]++;

                    calcRT( problem );
                    recordRT();
                    updateTime();
                    updateModel( problem );
                    calcCorrctRate();

                    if( debug )
                        debugTrialResult();
              }
              calcTargetRate();
              if( stageIdx == 0  && blockIdx < blockNum-1 )
                  betweenBlockTraining();
          }

          if( debug )
              debugStageResult();

          if( stageIdx == 0 )
              initNewStage();
       }

       // get the BLA OF the first rule in the rule set.
       if( FR_FLAG )
           BLA[ expIdx ] = fixRuleSet.getBLA( expIdx );
   }

   // record the correct rate of application of the two levels
   private void calcCorrctRate()
   {
      int idx = blockIdx * trialNum + trialIdx;
      if( feedbacks[ chosenComp ] > 0 )
      {
          if( stageIdx == 0 )
              trainRateArr[ idx ] = 1;
          else
              transRateArr[ idx ] = 1;
      }
      else
      {
           if( stageIdx == 0 )
               trainRateArr[ idx ] = 0;
           else
               transRateArr[ idx ] = 0;
      }
   }

   private void debugTrialResult()
   {
       if( levelFlag == TOP_LEVEL )
       {
           upCounts[ expIdx ][ groupIdx ][ stageIdx ]++;
           if( chosenComp == FIX_IDX )
           {
               if( feedback > 0 )
               {
                   int ruleIdx = fixRuleSet.getChosenRule();
                   fixRuleUsedNums[ expIdx ][ groupIdx ][ stageIdx ][ ruleIdx ]++;
               }
           }
           else
               if( feedback > 0 )
                   rerRuleUsedNums[ expIdx ][ groupIdx ][ stageIdx ]++;
       }
       else
           if( feedback > 0 )
               downCounts[ expIdx ][ groupIdx ][ stageIdx ]++;
/***********************************************************************
       if( expIdx == 1 && stageIdx == 1 )
       {
           System.out.print(" the BLA = " );
           double val = fixRuleSet.getBLA( expIdx );
           System.out.println( val );
       }
/***********************************************************************/
   }

   private void debugStageResult()
   {
      if( ( agentIdx + 1 ) % SHOW_INTERVAL == 0 )
      {
           System.out.print( "current exp = " );
           System.out.println( expIdx );
           System.out.print( "current stage = " );
           System.out.println( stageIdx );

           /*
           System.out.print( "# of RETRIEVAL CHOSEN BY BL is: " );
           System.out.println( trainBLretr1[ groupIdx ] );
           System.out.print( "# of EXECUTED BL RETRIEVAL is: " );
           System.out.println( trainBLretr2[ groupIdx ] );
           */

           System.out.print( "# of BL used is: " );
           System.out.println( downCounts[ expIdx ][ groupIdx ][ stageIdx ] );

           for( int i = 0; i < fixRuleNum; i++ )
           {
                System.out.print( "# of fixed rule " + i + " used is: " );
                System.out.println( fixRuleUsedNums[ expIdx ][ groupIdx ][ stageIdx ][ i ] );
           }

           if( RER_FLAG )
           {
               System.out.print( "# of RER RULE used is: " );
               System.out.println( rerRuleUsedNums[ expIdx ][ groupIdx ][ stageIdx ] );
           }

           String str = "";
           for( int i = 0; i < blockNum; i++ )
           {
               if( i < 10 )
                   str += "blockIdx=0" + String.valueOf( i ) + " : ";
               else
                   str += "blockIdx=" + String.valueOf( i ) + " : ";

               for( int j = 0; j < trialNum; j++ )
               {
                  if( stageIdx == 0 )
                  {
                      int val = trainRTArr[ i * trialNum + j ];
                      switch ( String.valueOf( val ).length()) {
                        case 2: str += String.valueOf( -100 );
                                break;
                        case 3: str += "0" + String.valueOf( val );
                                break;
                        case 4: str += String.valueOf( val );
                                break;
                        case 5: str += String.valueOf( 9999 );
                                break;
                      }
                      str += " ";
                  }
                  else
                  {
                      int val = transRTArr[ i * trialNum + j ];
                      switch ( String.valueOf( val ).length()) {
                        case 2: str += String.valueOf( -100 );
                                break;
                        case 3: str += "0" + String.valueOf( val );
                                break;
                        case 4: str += String.valueOf( val );
                                break;
                        case 5: str += String.valueOf( 9999 );
                                break;
                      }
                      str += " ";
                  }
               }
               str += "\n";
           }
           System.out.println( str );
       }
   }

   // percept a problem.
   public void perceptProblem( int[] problem, ConItem[] con, int probIdx )
   {
       for( int i = 0; i < 3; i++ )
       {
          Vector item = new Vector( 2 );
          if( ( stageIdx == 0 ) || ( expIdx == 0 ) )
          {
              problem[ 0 ] = 1;
              item.add( nameArr[ 0 ][ i ] );
          }
          else
          {
              problem[ 0 ] = 2;
              item.add( nameArr[ 1 ][ i ] );
          }

          if( i > 0 )
          {
              int valArr[] = new int[1];
              if( stageIdx == 0 )
                  valArr[0] = actTrainProbArr[ probIdx ][ i-1 ];
              else
                  valArr[0] = transProbArr[ probIdx ][ i-1 ];

              item.add( valArr );
              problem[ i ] = valArr[0];
           }
           con[i] = new ConItem( item );
       }

       // calculate the right answer.
       if( problem[0] == 1 )
           answer = problem[1] + problem[2] - 1;
       else
           answer = problem[1] - problem[2] - 1;
   }

   public void setCurState( int[] problem )
   {
       for( int i = 0; i < curState.length; i++ )
            curState[ i ] = 0;

       int offset = 0;
       for( int i = 0; i < NET_DIM_NUM; i++ )
       {
            int idx = problem[ i ] - 1;
            curState[ offset + idx ] = 1;
            offset += DIM_VAL_NUM;
       }
   }

   public void decideAction( int[] problem, ConItem[] con  )
   {
        setCurState( problem );

        // get the Q values from the bottom level
        qbpNet.setState( problem );
        qbpNet.setInput();
        qbpNet.forwardPass();
	qbpNet.getQvals( activationArr[ 0 ][ BL_IDX ] );

        // recommended action from the bottom level
        // decide action by boltmann distribution.
        actionDecider.setElements( activationArr[ 0 ][ BL_IDX ] );
        chosenActions[ BL_IDX ] = actionDecider.getRecommendation();

        if( BTWN_TRAIN_FLAG )
       // if( BTWN_TRAIN_FLAG || PRE_TRAIN_FLAG )
        {
           chosenComp = BL_IDX;
           chosenAction = chosenActions[ BL_IDX ];
           levelFlag = DOWN_LEVEL;
        }
        else
        {
          // components combination.
          // recommended action from the bottom level
           if( simNo == NACS_SIM && !FR_FLAG && chosenActions[ BL_IDX ] == NET_OUTPUT_NUM-1 )
           {
               BL_RETRIEVAL_FLAG = true;
               fillConclusion( problem, con );
               chosenActions[ BL_IDX ] = ruleFire( problem );
           }
           else
               BL_RETRIEVAL_FLAG = false;

          // recommended action from the top level
          // recommended action from fixed rules.
          if( FR_FLAG )
          {
              fixRuleType = fixRuleSet.getRecommendedRule( con );
              fixRuleSet.fillConclusion( con );
              if( fixRuleType == COUNT_TYPE )
                  countConclusion = fixRuleSet.getAction();
              else
                  retrieveConclusion = fixRuleSet.getAction();
              chosenActions[ FIX_IDX ] = ruleFire( problem );
              fixRuleStrings[0] = fixRuleSet.getRules();
          }

          // recommended action from RER rules.
          if( RER_FLAG )
          {
              chosenActions[ RER_IDX ] = rerRuleSet.getAction( curState );
              if( simNo == NACS_SIM && !FR_FLAG && chosenActions[ RER_IDX ] == NET_OUTPUT_NUM-1 )
              {
                  RER_RETRIEVAL_FLAG = true;
                  fillConclusion( problem, con );
                  chosenActions[ RER_IDX ] = ruleFire( problem );
              }
              else
                  RER_RETRIEVAL_FLAG = false;

              rerRuleStrings[0] = rerRuleSet.getAllRules();
              rerRuleStrings[1] = rerRuleSet.getMatchRules();
          }

          if( combMode1 == WS_COMB )
          {
              if( FR_FLAG )
              {
                  for( int i = 0; i < NET_OUTPUT_NUM; i++ )
                       activationArr[ 0 ][ FIX_IDX ][ i ] = 0;
                  if( chosenActions[ FIX_IDX ] != -1 )
                      activationArr[ 0 ][ FIX_IDX ][ chosenActions[ FIX_IDX ] ] = 1;
              }

              if( RER_FLAG )
                  rerRuleSet.getActions( curState, activationArr[ 0 ][ RER_IDX ] );
          }

          if( combMode2 == VARIED_MODE )
          {
              compSuccRates[ 0 ][ BL_IDX ] = qbpNet.getSuccRate();
              if( RER_FLAG )
                  compSuccRates[ 0 ][ RER_IDX ] = rerRuleSet.getSuccRate();
              else
                  compSuccRates[ 0 ][ RER_IDX ] = 0;

              compSuccRates[ 0 ][ IRL_IDX ] = 0;
              if( FR_FLAG )
                  compSuccRates[ 0 ][ FIX_IDX ] = fixRuleSet.getSuccRate();
              else
                  compSuccRates[ 0 ][ FIX_IDX ] = 0;
          }

          if( combMode1 == WS_COMB )
          {
              if( combMode2 == VARIED_MODE )
              {
                  wsCombinder.setSuccRate( compSuccRates[0] );
                  wsCombinder.setActivations( activationArr[0] );
                  wsCombinder.calcWeights();
              }
              actionResult[0] = wsCombinder.combination();
              chosenAction = actionResult[ 0 ];
          }
          else
          {
              if( combMode2 == VARIED_MODE )
              {
                  stochCombiner.setSuccRate( compSuccRates[ 0 ] );
                  stochCombiner.calcProbabilities();
              }
              chosenComp = stochCombiner.selectComponent();
              switch( chosenComp )
              {
                 case BL_IDX :
                               chosenAction = chosenActions[ BL_IDX ];
                               levelFlag = DOWN_LEVEL;
                               break;
                 case RER_IDX :
                               chosenAction = chosenActions[ RER_IDX ];
                               levelFlag = TOP_LEVEL;
                               break;
                 case FIX_IDX :
                               chosenAction = chosenActions[ FIX_IDX ];
                               levelFlag = TOP_LEVEL;
                               break;
              }
          }

          if( chosenAction == -1 )
          {
              chosenComp = BL_IDX;
              chosenAction = chosenActions[ BL_IDX ];
              levelFlag = DOWN_LEVEL;
          }

          // record the retrieval-rules-executed rate, counting-rules-executed rate
          // and BL-exectued rate.
          calcComponentExecutedRates();
        }
   }

   public void calcComponentExecutedRates()
   {
        if( expIdx == 0 && stageIdx == 0 )
        {
            if( BL_RETRIEVAL_FLAG )
            {
                trainBLretr1[ groupIdx ]++;
                if( chosenComp == BL_IDX && chosenAction != -1 )
                    trainBLretr2[ groupIdx ]++;
            }
        }

        if( !PRE_TRAIN_FLAG )
        {
            int idx = blockIdx * trialNum + trialIdx;
            if( stageIdx == 0 )
            {
                switch( chosenComp ){
                      case BL_IDX:
                           trainRetrArr[ idx ] = 0.0;
                           trainContArr[ idx ] = 0.0;
                           trainBttmArr[ idx ] = 1.0;
                           break;
                      case RER_IDX:
                           trainRetrArr[ idx ] = 0.0;
                           trainContArr[ idx ] = 0.0;
                           trainBttmArr[ idx ] = 0.0;
                           break;
                      case FIX_IDX:
                           if( fixRuleType == RETRIEVE_TYPE )
                           {
                               trainRetrArr[ idx ] = 1.0;
                               trainContArr[ idx ] = 0.0;
                               trainBttmArr[ idx ] = 0.0;
                           }
                           else
                           {
                               trainRetrArr[ idx ] = 0.0;
                               trainContArr[ idx ] = 1.0;
                               trainBttmArr[ idx ] = 0.0;
                           }
                           break;
                }
            }
            else
            {
                switch( chosenComp ){
                      case BL_IDX:
                           transRetrArr[ idx ] = 0.0;
                           transContArr[ idx ] = 0.0;
                           transBttmArr[ idx ] = 1.0;
                           break;
                      case RER_IDX:
                           transRetrArr[ idx ] = 0.0;
                           transContArr[ idx ] = 0.0;
                           transBttmArr[ idx ] = 0.0;
                           break;
                      case FIX_IDX:
                           if( fixRuleType == RETRIEVE_TYPE )
                           {
                               transRetrArr[ idx ] = 1.0;
                               transContArr[ idx ] = 0.0;
                               transBttmArr[ idx ] = 0.0;
                           }
                           else
                           {
                               transRetrArr[ idx ] = 0.0;
                               transContArr[ idx ] = 1.0;
                               transBttmArr[ idx ] = 0.0;
                           }
                           break;
                }
            }
        }

        if( !PRE_TRAIN_FLAG )
        {
             /*
             // record the probabilities of application of the two levels
             int idx = blockIdx * trialNum + trialIdx;
             if( stageIdx == 0 )
                 trainPossArr[ idx ] = probFR;
             else
                 transPossArr[ idx ] = probFR;
             */
        }
   }

   public void fillConclusion( int[] problem, ConItem[] con )
   {
        if( problem[ 0 ] == 1 )
        {
            double val = Math.random();
            if( val < 0.8 )
            {
                retrieveConclusion[2].setSlots( 0, "+" );
                retrieveConclusion[1].setSlots( 1, con[1].getSlots().elementAt(1) );
                concDimIdx = 3;
            }
            else
            {
                retrieveConclusion[2].setSlots( 0, "-" );
                retrieveConclusion[4].setSlots( 1, con[1].getSlots().elementAt(1) );
                concDimIdx = 0;
            }
        }
        else
        {
            double val = Math.random();
            if( val < 0.2 )
            {
                retrieveConclusion[2].setSlots( 0, "-" );
                retrieveConclusion[1].setSlots( 1, con[1].getSlots().elementAt(1) );
                concDimIdx = 3;
            }
            else
            {
                retrieveConclusion[2].setSlots( 0, "+" );
                retrieveConclusion[4].setSlots( 1, con[1].getSlots().elementAt(1) );
                concDimIdx = 0;
            }
        }

        retrieveConclusion[3].setSlots( 1, con[2].getSlots().elementAt(1) );
   }

   public int ruleFire( int[] problem )
   {
        int result = 0, opType = 0;
        if( FR_FLAG )
        {
            if( fixRuleType == COUNT_TYPE )
            {
                opType = problem[ 0 ];
                if( opType == 1 )
                    result = problem[ 1 ] + problem[ 2 ] - 1;
                else
                    result = problem[ 1 ] - problem[ 2 ] - 1;
            }
            else
            {
                formRetrievalCue();
                result = performRetrieval();
            }
        }
        else
        {
            formRetrievalCue();
            result = performRetrieval();
        }

        return result;
   }

   public void formRetrievalCue( )
   {
       if( FR_FLAG )
           concDimIdx = fixRuleSet.getConcDimIdx();

       int count = 0;
       int[] indices = new int[ chunkDimNum - 1 ];
       for( int i = 0; i < chunkDimNum; i++ )
            if( i != concDimIdx )
                 indices[ count++ ] = i;
       gkNet.setCueDimIndices( indices );
       retrievalCue.removeAllElements();

       for( int i = 0; i < dims.length; i++ )
         for( int j = 0; j < dims[i].length; j++ )
              dims[i][j] = 0;

       for( int i = 0; i < chunkDimNum; i++ )
       {
           int idx = -1;
           if( i != concDimIdx )
           {
             if( i == 1 )
             {
                 String str = (String)retrieveConclusion[i+1].getSlots().elementAt(0);
                 if( str.compareTo( add ) == 0 )
                     idx = 1;
                 else
                     idx = 2;
             }
             else
                 idx = ((int[])retrieveConclusion[i+1].getSlots().elementAt(1))[0];

             dims[ i ][ idx - 1 ] = 1;
          }

          retrievalCue.addElement( dims[ i ] );
       }
   }

   public int performRetrieval()
   {
       int[] result = new int[ chunkDVNums[concDimIdx] ];
       for( int i = 0; i < result.length; i++ )
            result[ i ] = 0;

       gkNet.setConclusionDimIdx( concDimIdx );
       gkNet.setRetrievalCue( retrievalCue );
       CHUNK_RETRIEVAL_FLAG = gkNet.reasoningWithinNode();

       /*
       if( CHUNK_RETRIEVAL_FLAG )
       {
           System.out.println("stop here");
       }*/

       gkNet.getResults( result );

       int i = 0;
       while( i < result.length && result[ i ] == 0 )
              i++;

       if( i == result.length )
           return -1;
       else
           return i;
   }

   public void updateModel( int[] problem )
   {
       //update positive or negative matches.
       if( feedbacks[ BL_IDX ] > 0 )
           qbpNet.updateMatchCount( true );
       else
           qbpNet.updateMatchCount( false );

       if( !BL_RETRIEVAL_FLAG || chosenComp != BL_IDX )
       {
           // obtain the current state QValue
           qValue = qbpNet.getQVal( chosenAction );
           // update the bottom level.
           qbpNet.update( chosenAction, feedback, qValue );
       }
       else
       {
           // obtain the current state QValue
           qValue = qbpNet.getQVal( NET_OUTPUT_NUM - 1 );
           // update the bottom level.
           qbpNet.update( NET_OUTPUT_NUM - 1, feedback, qValue );
       }

       if( BTWN_TRAIN_FLAG )
       //if( BTWN_TRAIN_FLAG || PRE_TRAIN_FLAG )
           return;

        // update each component.
        if( combMode1 == STOCH_COMB )
        {
            switch( chosenComp ) {
            case BL_IDX:
            case RER_IDX:
                        if( RER_FLAG )
                        {
                            if( feedback > 0 )
                                rerRuleSet.updateRuleSet( curState, chosenAction,
                                                          true, false, false );
                            else if( chosenAction != -1 )
                                     rerRuleSet.updateRuleSet( curState, chosenAction,
                                                               false, false, false );
                            if( chosenComp == RER_IDX )
                                rerRuleStrings[ 2 ] = rerRuleSet.getChosenRules();
                        }
                        break;
            case FIX_IDX:
                        if( feedback > 0 )
                            fixRuleSet.update( true );
                        else
                            fixRuleSet.update( false );
                        break;
            }
        }
        else
        {
            if( RER_FLAG )
            {
                if( feedback > 0 )
                    rerRuleSet.updateRuleSet( curState, chosenAction,
                                              true, false, false );
                else
                    rerRuleSet.updateRuleSet( curState, chosenAction,
                                              false, false, false );
            }

            if( feedbacks[ FIX_IDX ] > 0 )
                fixRuleSet.update( true );
            else
                fixRuleSet.update( false );

        }

        if( simNo == ACS_SIM )
            return;

        // update the NACS upper level.
        if( combMode1 == STOCH_COMB )
        {
            if( chosenComp != FIX_IDX || fixRuleType == COUNT_TYPE )
            {
                double prob = Math.random();
                if( prob < CHUNK_ENCODING_PROB )
                {
                    if( chosenAction != -1 )
                    {
                        int[] content = new int[ chunkDimNum ];
                        content[ 0 ] = problem[ 1 ] - 1;
                        content[ 1 ] = problem[ 0 ] - 1;
                        content[ 2 ] = problem[ 2 ] - 1;
                        content[ 3 ] = chosenAction;
                        gkNet.update( content, "chunk" );
                    }
                }
            }
            else
            {
          // modified on 02/14/02, all retrievals can affect the chunk's BLA.
                if( feedback > 0 )
                    gkNet.update( true );
                else if( chosenActions[ FIX_IDX ] != -1 )
                         gkNet.update( false );
                     else
                         gkNet.update();
            }
        }
        else
        {
            double prob = Math.random();
            if( prob < CHUNK_ENCODING_PROB )
            {
                int[] content = new int[ chunkDimNum ];
                content[ 0 ] = problem[ 1 ] - 1;
                content[ 1 ] = problem[ 0 ] - 1;
                content[ 2 ] = problem[ 2 ] - 1;
                content[ 3 ] = chosenAction;
                gkNet.update( content, "chunk" );
            }
        }
   }

   // calculate the feedbacks for each component: BL, RER and FR.
   public void calcFeedbacks()
   {
      for( int i = 0; i < COMB_COMPONENT_NUM; i++ )
           if( chosenActions[ i ] == answer )
               feedbacks[ i ] = POS_REWARD;
           else
               feedbacks[ i ] = NEG_REWARD;

      if( chosenAction == answer )
          feedback = POS_REWARD;
      else
          feedback = NEG_REWARD;
   }

/*****************************************************************************/
   public void calcTargetRate()
   {
       onTargetRates[ stageIdx ][ blockIdx ] =
        (double)onTargetArr[ stageIdx ][ blockIdx ] / trialNum;
   }

   public double[] getTrainHitRate()
   {
      return onTargetRates[0];
   }

   public double[] getTransHitRate()
   {
      return onTargetRates[1];
   }
/*****************************************************************************/

   public void setTime()
   {
       if( FR_FLAG )
           fixRuleSet.setCurrentTime( CURRENT_TIME );

       if( simNo == NACS_SIM )
           gkNet.setCurrentTime( CURRENT_TIME );

       if( RER_FLAG )
           rerRuleSet.setCurrentTime( CURRENT_TIME );
   }

  public void updateTime()
  {
      CURRENT_TIME += responseTime;
  }

  public void calcRT( int[] problem )
  {
     if( combMode1 == STOCH_COMB )
     {
        switch( chosenComp ) {
          case BL_IDX: responseTime = qbpNet.getResponseTime();
                       if( BL_RETRIEVAL_FLAG )
                       {
                           if( chosenActions[ BL_IDX ] != -1 )
                           {
                               gkNet.calcChosenChunkBLA();
                               double chunkBLA = gkNet.getChosenChunkBLA();
                               responseTime += PRE_CHUNK_BLA * Math.pow( chunkBLA, -1 );
                           }
                           else
                               responseTime += FAIL_RETRIEVAL_TIME;
                       }
                       break;
          case RER_IDX:
                       rerRuleSet.calcRT();
                       responseTime = rerRuleSet.getResponseTime();
                       if( RER_RETRIEVAL_FLAG )
                       {
                           if( chosenActions[ RER_IDX ] != -1 )
                           {
                               gkNet.calcChosenChunkBLA();
                               double chunkBLA = gkNet.getChosenChunkBLA();
                               responseTime += PRE_CHUNK_BLA * Math.pow( chunkBLA, -1 );
                           }
                           else
                               responseTime += FAIL_RETRIEVAL_TIME;
                       }
                       break;
          case FIX_IDX:
                       if( fixRuleType == COUNT_TYPE )
                       {
                           fixRuleSet.calcBLA();
                           fixRuleSet.calcTLDT( problem[2] );
                           responseTime = fixRuleSet.getResponseTime();
                       }
                       else
                       {
                           fixRuleSet.calcBLA();
                           fixRuleSet.calcTLDT();
                           if( chosenActions[ FIX_IDX ] != -1 )
                           {
                               gkNet.calcChosenChunkBLA();
                               double chunkBLA = gkNet.getChosenChunkBLA();
                               fixRuleSet.calcTLAT( chunkBLA );
                           }
                           responseTime = fixRuleSet.getResponseTime();
                           if( chosenActions[ FIX_IDX ] == -1 )
                               responseTime += FAIL_RETRIEVAL_TIME;
                       }
                       break;
        }
     }
     else
     {
        double[] rtArr = new double[ COMB_COMPONENT_NUM ];
        for( int i = 0; i < COMB_COMPONENT_NUM; i++ )
             rtArr[ i ] = 0.0;

        double[][] weights = new double[ actionTypeNum ][ COMB_COMPONENT_NUM ];
        wsCombinder.getWeights( weights[0] );

        double rt = 0.0, totWeight = 0.0;
        for( int i = 0; i < COMB_COMPONENT_NUM; i++ )
        {
             if( feedbacks[ i ] > 0 )
             {
                 rt += rtArr[ i ] * weights[ 0 ][ i ];
                 totWeight += weights[ 0 ][ i ];
             }
        }

        responseTime = (int)( rt / totWeight );
     }
   }

   public void recordRT()
   {
      int time;

      if( feedback > 0 )
          time = responseTime;
      else
          time = -1;

      int idx = blockIdx * trialNum + trialIdx;
      if( stageIdx == 0 )
          trainRTArr[idx] = time;
      else
          transRTArr[idx] = time;
   }

   public double[] getBLA()
   {
      return BLA;
   }

   public void getRTArrays( int[] arr1, int[] arr2 )
   {
      for( int i = 0; i < arr1.length; i++ )
           arr1[ i ] = trainRTArr[ i ];
      for( int i = 0; i < arr2.length; i++ )
           arr2[ i ] = transRTArr[ i ];
   }

   public void getPerformanceArrays( double[] arr1, double[] arr2, double[] arr3, double[] arr4,
                                     double[] arr5, double[] arr6, double[] arr7, double[] arr8,
                                     double[] arr9, double[] arr10  )
   {
      for( int i = 0; i < arr1.length; i++ )
           arr1[ i ] = trainPossArr[ i ];
      for( int i = 0; i < arr2.length; i++ )
           arr2[ i ] = transPossArr[ i ];
      for( int i = 0; i < arr3.length; i++ )
           arr3[ i ] = trainRetrArr[ i ];
      for( int i = 0; i < arr4.length; i++ )
           arr4[ i ] = transRetrArr[ i ];
      for( int i = 0; i < arr5.length; i++ )
           arr5[ i ] = trainContArr[ i ];
      for( int i = 0; i < arr6.length; i++ )
           arr6[ i ] = transContArr[ i ];
      for( int i = 0; i < arr7.length; i++ )
           arr7[ i ] = trainBttmArr[ i ];
      for( int i = 0; i < arr8.length; i++ )
           arr8[ i ] = transBttmArr[ i ];
      for( int i = 0; i < arr9.length; i++ )
           arr9[ i ] = trainRateArr[ i ];
      for( int i = 0; i < arr10.length; i++ )
           arr10[ i ] = transRateArr[ i ];
   }

   public void fillFixRule( StringBuffer str )
   {
      str.delete( 0, str.length() );
      str.insert( 0, fixRuleStrings[ 0 ].substring( 0 ) );
   }

   public void fillRerRule( StringBuffer[] str )
   {
      String ruleHeads[] = {
       "\nAll rules are: \n", "\nMatch rules are: \n", "\nChosen rules are: \n"
      };

      for( int i = 0; i < 3; i++ )
           str[i].append( ruleHeads[ i ] + rerRuleStrings[i].substring( 0 ) );
   }
}
