package tasksimulators.oldsimu;
import clarion.common.*;
import clarion.tools.*;

public class LCFixRuleSet {

   int ruleNum, chosenRule, chosenAction;
   int matchCount, matchedRule[];

   double SUCC_C3 = 1.0, SUCC_C4 = 2.0;
   double succRate, sumPM, sumNM;

   StochasticDecider ruleDecider;  // action decider
   LCFixRule lcRuleSet[];

   public LCFixRuleSet( int tNo, int num, LCFixRule[] rSet )
   {
      sumPM = 0.0;
      sumNM = 0.0;
      succRate = ( SUCC_C3 + sumPM ) / ( SUCC_C4 + sumPM + sumNM );

      ruleNum = num;
      lcRuleSet = new LCFixRule[ ruleNum ];
      for( int i = 0; i < num; i++ )
           lcRuleSet[ i ] = rSet[ i ];

      matchedRule = new int[ ruleNum ];
      // ruleDecider = new StochasticDecider( ruleNum,
      //                  World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );
     ruleDecider = new StochasticDecider( ruleNum, 1.0 );
   }

   public LCFixRuleSet( int tNo, int num )
   {
      sumPM = 0.0;
      sumNM = 0.0;
      succRate = ( SUCC_C3 + sumPM ) / ( SUCC_C4 + sumPM + sumNM );

      ruleNum = num;
   }

   public void reinit()
   {
      sumPM = 0.0;
      sumNM = 0.0;
      succRate = ( SUCC_C3 + sumPM ) / ( SUCC_C4 + sumPM + sumNM );

      for( int i = 0; i < ruleNum; i++ )
           lcRuleSet[ i ].reinit();
   }

   public double getBLA( int idx )
   {
       return lcRuleSet[ idx ].getBLA();
   }

   public int getResponseTime()
   {
       return lcRuleSet[ chosenRule ].getRT();
   }

   public void setCurrentTime( int curTime )
   {
       for( int i = 0; i < ruleNum; i++ )
            lcRuleSet[i].setCurTime( curTime );
   }

   public void setRuleCosts( double[] costs )
   {
      for( int i = 0; i < lcRuleSet.length; i++ )
           lcRuleSet[ i ].setRuleCost( costs[ i ] );
   }

   public void setRuleActionTimes( double[] actionTimes )
   {
      for( int i = 0; i < lcRuleSet.length; i++ )
           lcRuleSet[ i ].setRuleActionTime( actionTimes[ i ] );
   }

   public void setRulePreBLAs( double[] preBLAs )
   {
      for( int i = 0; i < lcRuleSet.length; i++ )
           lcRuleSet[ i ].setRulePreBLA( preBLAs[ i ] );
   }

   public void setPreChunkBLA( double val )
   {
      for( int i = 0; i < lcRuleSet.length; i++ )
           lcRuleSet[ i ].setPreChunkBLA( val );
   }

   public double getSuccRate()
   {
      double val = 0.0;
      sumPM = 0.0;
      sumNM = 0.0;

      for( int i = 0; i < ruleNum; i++ )
      {
           val = lcRuleSet[i].getPM();
           sumPM += val;
           val = lcRuleSet[i].getNM();
           sumNM += val;
      }

      succRate = ( SUCC_C3 + sumPM ) / ( SUCC_C4 + sumPM + sumNM );
      return succRate;
   }

   public int getChosenRule()
   {
       return chosenRule;
   }


   // used for Letter Counting.
   public int getRecommendedRule( ConItem[] problem )
   {
      matchCount = 0;

      // get the indices of all the matched rules.
      for( int i = 0; i < ruleNum; i++ )
      {
         boolean bVal = lcRuleSet[i].patternMatch( problem );
         if( bVal )
             matchedRule[ matchCount++ ] = i;
      }

      if( matchCount == 0 )
      {
          System.out.println( "no matching rule!" );
          return -1;
      }

      // get the utilities of all the matched rules.
      double UTL_ARRAY[] = new double[ matchCount ];
      for( int i = 0; i < matchCount; i++ )
      {
           lcRuleSet[ matchedRule[i] ].calcUtility();
           UTL_ARRAY[ i ] = lcRuleSet[ matchedRule[i] ].getUtility( );
      }

      // use Boltzmann distribution to decide probabilitically on
      // which rule will be fired by the utilities.
      ruleDecider.setSize( matchCount );
      ruleDecider.setElements( UTL_ARRAY );
      int idx = ruleDecider.getRecommendation();
      chosenRule = matchedRule[ idx ];

      if( chosenRule < 2 )
          return 0;
      else
          return 1;
   }

   public void fillConclusion( ConItem[] problem )
   {
      lcRuleSet[ chosenRule ].fillConclusion( chosenRule, problem );
   }

   public ConItem[] getAction()
   {
      return lcRuleSet[ chosenRule ].getConclusion();
   }

   public int getConcDimIdx()
   {
      return lcRuleSet[ chosenRule ].getConcDimIdx( chosenRule );
   }

   public void calcBLA()
   {
      lcRuleSet[ chosenRule ].calcBLA();
   }

   public void calcTLDT ( int num )
   {
      lcRuleSet[ chosenRule ].calcTLDT( num );
   }

   public void calcTLDT()
   {
      lcRuleSet[ chosenRule ].calcTLDT();
   }

   public void calcTLAT( double chunkBLA )
   {
      lcRuleSet[ chosenRule ].calcTLAT( chunkBLA );
   }

   public void update( boolean succFlag )
   {
      lcRuleSet[ chosenRule ].update( succFlag );
   }

   public StringBuffer getRules()
   {
      StringBuffer strBuf = new StringBuffer("aaaa");
      return strBuf;
   }
 }