/******************************************************************************/
/************************* LETTER COUNTING TASK  ***************************/
/******************************************************************************/
package tasksimulators.oldsimu;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.Vector;
import java.util.Arrays;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;

public class LCTask extends JFrame implements SimuTask{

   int WINDOW_WIDTH = 800, WINDOW_HEIGHT = 600;
   int MAX_TIME = 8000;
   double INIT_BLA = 0.06;

   int ACS_SIM = 0, NACS_SIM = 1;
   //int simNo = ACS_SIM;
   int simNo = NACS_SIM;
   //boolean FR_FLAG = false;
   boolean FR_FLAG = true;
   boolean RER_FLAG = false;
   //boolean RER_FLAG = true;

   int MAX_PROB_NUM = 135;
   int ADD_TYPE = 0, SUB_TYPE = 1;
   int EXP_NUM = 2, GROUP_NUM = 2, STAGE_NUM = 2, FILE_TYPE_NUM = 7, AGENT_NUM = 50;
   int LETTER_NUM = 26, ADDEND_NUM = 6, OP_TYPE_NUM = 2;
   int CON_PROB_NUM = 12, VAR_PROB_NUM = 72, TRANS_PROB_NUM = 12;
   int EPISODE_NUM = 36, TRAIN_TRIAL_NUM = 432, TRANS_TRIAL_NUM = 36;

   int BLOCK_NUMS[][] = { { 36, 3 },  { 6, 3 } };
   int TRIAL_NUMS[][] = { { 12, 12 }, { 72, 12 } };

   int QBP_DIM_VAL_NUMS[] = { OP_TYPE_NUM, LETTER_NUM, ADDEND_NUM };
   int GKN_DIM_VAL_NUMS[] = { LETTER_NUM, OP_TYPE_NUM, ADDEND_NUM, LETTER_NUM };

   // used for different simulation types
   int RULE_BASED = 0, SIMILARITY_BASED = 1, RULE_SIM_BASED = 2;

   int taskNo, expIdx, groupIdx, agentIdx, blockNum;
   int startLetter, opType, addend;
   int conProbArr[][], varProbArr[][];
   int transAddProbArr[][], transSubProbArr[][];
   int trainRTArray[][][], transRTArray[][][];
   int wholeProbSet[][];

   double trainPossArr[][][], transPossArr[][][];
   double trainRetrArr[][][], transRetrArr[][][];
   double trainContArr[][][], transContArr[][][];
   double trainBttmArr[][][], transBttmArr[][][];
   double trainRateArr[][][], transRateArr[][][];
/***************************************************************************/
   double blaInfo[][], trainHitRates[], transHitRates[];
/***************************************************************************/

   int ruleNumArr[] = { 2, 6 };
   int ruleNum;

   int[] CHUNK_DIM_TYPES = { 0, 0, 0, 0 };
   String ruleConditionArr[][] =
   {
       { "ADDITION", "START_LETTER", "STEP" },
       { "SUBTRACTION", "START_LETTER", "STEP"  },
       { "ADDITION", "START_LETTER", "STEP" },
       { "ADDITION", "START_LETTER", "STEP" },
       { "SUBTRACTION", "START_LETTER", "STEP"  },
       { "SUBTRACTION", "START_LETTER", "STEP"  }
   };

   String ruleActionArr[][] =
   {
        { "COUNT", "START_LETTER", "+", "STEP" },
        { "COUNT", "START_LETTER", "-", "STEP" },
        { "RETRIEVE", "START_LETTER", "+", "STEP", "RESULT" },
        { "RETRIEVE", "RESULT", "-", "STEP", "START_LETTER" },
        { "RETRIEVE", "START_LETTER", "-", "STEP", "RESULT" },
        { "RETRIEVE", "RESULT", "+", "STEP", "START_LETTER" }
   };

   boolean exitFlag = false;

   Vector probInfoVect;
   ConItem condition[][], action[][];
   LCFixRule lcRules[];
   WriteFile dataFile, agentDatFile, ruleDatFile, expDatFiles[][][][];
   LCAgent expAgent;

   Container c;
   JFrame mainFrame;
   TextArea probArea1, probArea2;
   TextArea ruleArea1, ruleArea2;
   TextArea showArea1, showArea2;
   JButton beginButton, exitButton;

   // the contructor of class LCTASK
   public LCTask( JFrame mFrame )
   {
      opType = 0;
      trainRTArray = new int[ GROUP_NUM ][ AGENT_NUM ][ TRAIN_TRIAL_NUM ];
      transRTArray = new int[ GROUP_NUM ][ AGENT_NUM ][ TRANS_TRIAL_NUM ];
      blaInfo = new double[ AGENT_NUM ][ 2 ];

      trainPossArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRAIN_TRIAL_NUM ];
      transPossArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRANS_TRIAL_NUM ];
      trainRetrArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRAIN_TRIAL_NUM ];
      transRetrArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRANS_TRIAL_NUM ];
      trainContArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRAIN_TRIAL_NUM ];
      transContArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRANS_TRIAL_NUM ];
      trainBttmArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRAIN_TRIAL_NUM ];
      transBttmArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRANS_TRIAL_NUM ];
      trainRateArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRAIN_TRIAL_NUM ];
      transRateArr = new double[ GROUP_NUM ][ AGENT_NUM ][ TRANS_TRIAL_NUM ];

      initProblems();
      initInterface( mFrame );
   }

   public void initProblems()
   {
      conProbArr = new int[ CON_PROB_NUM ][ 2 ];
      varProbArr = new int[ VAR_PROB_NUM ][ 2 ];
      transAddProbArr = new int[ TRANS_PROB_NUM ][ 2 ];
      transSubProbArr = new int[ TRANS_PROB_NUM ][ 2 ];

      wholeProbSet = new int[ MAX_PROB_NUM ][ 2 ];
      genWholeProbSet( 0 );

      initProbInfoVect();
      genProblems( 0, VAR_PROB_NUM, varProbArr );
      genProblems( 1, TRANS_PROB_NUM, transAddProbArr );
      genConProblems();
      genInvertProblems();
   }

   public void initProbInfoVect()
   {
      probInfoVect = new Vector( MAX_PROB_NUM );
      for( int j = 0; j < MAX_PROB_NUM; j++ )
      {
           int[] item = new int[1];
           item[0] = j;
           probInfoVect.addElement( item );
      }
   }

   public void genWholeProbSet( int mode )
   {
       int offset = 0;
       int BORDER_VAL[] = { LETTER_NUM - ADDEND_NUM - 1, ADDEND_NUM };

       for( int i = 0; i < LETTER_NUM; i++ )
       {
          int END_NUM = 0;
          if( ( ( mode == 0 ) && ( i <= BORDER_VAL[0] ) ) ||
              ( ( mode == 1 ) && ( i >= BORDER_VAL[1] ) ) )
              END_NUM = ADDEND_NUM;
          else
          {
             if( mode == 0 )
                 END_NUM = ADDEND_NUM + BORDER_VAL[ mode ] - i;
             else
                 END_NUM = i;
          }

          for( int j = 0; j < END_NUM; j++ )
          {
               wholeProbSet[ offset + j ][ 0 ] = i + 1;
               wholeProbSet[ offset + j ][ 1 ] = j + 1;
          }

          offset += END_NUM;
       }
   }

  public void genProblems( int mode, int num, int[][] arr )
  {
      int[] counters = new int[ ADDEND_NUM  ];
      for( int i = 0; i < ADDEND_NUM; i++ )
           counters[ i ] = 0;

      int i = 0;
      while( i < num )
      {
          int range = probInfoVect.size();
          int val = (int)( Math.random() * range );
          int idx = ((int[])( probInfoVect.elementAt( val ) ))[0];
          int addend = wholeProbSet[idx][1];
         // if( mode==1 || counters[ addend-1 ]<( VAR_PROB_NUM/ADDEND_NUM ) )
          if( ( mode==0 && counters[ addend-1 ]<( VAR_PROB_NUM/ADDEND_NUM ) ) ||
              ( mode==1 && counters[ addend-1 ]<( TRANS_PROB_NUM/ADDEND_NUM ) ) )
          {
              arr[ i ][ 0 ] = wholeProbSet[ idx ] [0 ];
              arr[ i ][ 1 ] = wholeProbSet[ idx ][ 1 ];
              probInfoVect.removeElementAt( val );
              counters[ addend-1 ]++;
              i++;
          }
      }
  }

  public void genConProblems()
  {
      int[] counters = new int[ ADDEND_NUM  ];
      for( int i = 0; i < ADDEND_NUM; i++ )
           counters[ i ] = 0;

     Vector vect = new Vector( VAR_PROB_NUM );
     for( int j = 0; j < VAR_PROB_NUM; j++ )
     {
          int[] item = new int[1];
          item[0] = j;
          vect.addElement( item );
     }

     int i = 0;
     while( i < CON_PROB_NUM )
     {
          int range = vect.size();
          int val = (int)( Math.random() * range );
          int idx = ((int[])( vect.elementAt( val ) ))[0];
          int addend = varProbArr[ idx ][ 1 ];
          if( counters[ addend - 1 ] < ( CON_PROB_NUM/ADDEND_NUM ) )
          {
              conProbArr[ i ][ 0 ] = varProbArr[ idx ][ 0 ];
              conProbArr[ i ][ 1 ] = varProbArr[ idx ][ 1 ];
              vect.removeElementAt( val );
              counters[ addend-1 ]++;
              i++;
          }
     }
  }

  public void genInvertProblems()
  {
     Vector vect = new Vector( CON_PROB_NUM );
     for( int j = 0; j < CON_PROB_NUM; j++ )
     {
          int[] item = new int[1];
          item[0] = j;
          vect.addElement( item );
     }

     for( int i = 0; i < CON_PROB_NUM; i++ )
     {
          int range = vect.size();
          int val = (int)( Math.random() * range );
          int idx = ((int[])( vect.elementAt( val ) ))[0];
          transSubProbArr[i][0] = conProbArr[idx][0] + conProbArr[idx][1];
          transSubProbArr[i][1] = conProbArr[idx][1];
          vect.removeElementAt( val );
     }
  }

  public void initInterface( JFrame mFrame )
   {
      // setup user interface.
      c = getContentPane();
      c.removeAll();
      c.setLayout( new BorderLayout( 5, 0 ) );
      buildInterface();

      setSize( WINDOW_WIDTH, WINDOW_HEIGHT );
      setEnabled( false );
      setVisible( false );

      mainFrame = mFrame;
      addWindowListener(
          new WindowAdapter(){
             public void windowClosing( WindowEvent e )
             {
               if ( exitFlag == true )
               {
                    dataFile.closeFile();
                    setVisible( false );
                    setEnabled( false );
                    mainFrame.setVisible( true );
                    mainFrame.setEnabled( true );
               }
             }
      }  );
   }

  public void buildInterface()
  {
      JPanel showPanel = new JPanel();
      showPanel.setLayout( new BorderLayout() );
      c.add( showPanel, BorderLayout.CENTER );

      JPanel capPanel = new JPanel();
      capPanel.setLayout( new FlowLayout( FlowLayout.CENTER, 200, 5 ) );
      showPanel.add( capPanel, BorderLayout.NORTH );
      JLabel taskCaption = new JLabel( "LETTER  COUNTING  TASK" );
      taskCaption.setFont( new Font( "Serif", Font.BOLD, 20 ) );
      capPanel.add( taskCaption );

      JPanel mainPanel = new JPanel();
      mainPanel.setLayout( new GridLayout( 1, 2, 5, 5 ) );
      showPanel.add( mainPanel, BorderLayout.CENTER );

      JPanel trainPanel = new JPanel();
      trainPanel.setLayout( new GridLayout( 3, 1, 5, 5 ) );
      mainPanel.add( trainPanel );
      probArea1 = new TextArea( "PROBLEMS IN TRAINING \n" );
      trainPanel.add( probArea1 );
      ruleArea1 = new TextArea( "All the fixed rules are: " );
      trainPanel.add( ruleArea1 );
      showArea1 = new TextArea( "TRAINING PROCESS SHOW \n" );
      trainPanel.add( showArea1 );

      JPanel transPanel = new JPanel();
      transPanel.setLayout( new GridLayout( 3, 1, 5, 5 ) );
      mainPanel.add( transPanel );
      probArea2 = new TextArea( "PROBLEMS IN TRANSFER \n" );
      transPanel.add( probArea2 );
      ruleArea2 = new TextArea( "All the fixed rules are: " );
      transPanel.add( ruleArea2 );
      showArea2 = new TextArea( "TRANSFER PROCESS SHOW \n" );
      transPanel.add( showArea2 );

      JPanel buttonPanel = new JPanel();
      buttonPanel.setLayout( new FlowLayout() );
      c.add( buttonPanel, BorderLayout.SOUTH );

      beginButton = new JButton( "BEGIN" );
      beginButton.setEnabled( true );
      beginButton.addActionListener(
         new  ActionListener() {
                 public void actionPerformed( ActionEvent e )
                 {
                     beginButton.setEnabled( false );
                     exitButton.setEnabled( false );
                     exitFlag = false;
                     startExp();
                     exitButton.setEnabled( true );
                 }
        }
      );

      exitButton = new JButton( "Exit" );
      exitButton.addActionListener(
                 new  ActionListener() {
                     public void actionPerformed( ActionEvent e )
                     {
                        closeAllFiles();
                        setVisible( false );
                        setEnabled( false );
                        mainFrame.setVisible( true );
                        mainFrame.setEnabled( true );
                     }
                 }
      );

      buttonPanel.add( beginButton );
      buttonPanel.add( exitButton );
   }

   public void closeAllFiles()
   {
      dataFile.closeFile();
      for( int i = 0; i < EXP_NUM; i++ )
        for( int j = 0; j < GROUP_NUM; j++ )
           for( int k = 0; k < STAGE_NUM; k++ )
            for( int p = 0; p < FILE_TYPE_NUM; p++ )
                 expDatFiles[i][j][k][p].closeFile();
   }

   public void showProblems()
   {
     int len1, len2, len3, len4, probArr1[][], probArr2[][];

     if( groupIdx == 0 )
     {
         len1 =  conProbArr.length;
         len2 =  conProbArr[0].length;
         probArr1 = new int[ len1 ][ len2 ];
         for( int i = 0; i < len1; i++ )
           for( int j = 0; j < len2; j++ )
                probArr1[ i ][ j ] = conProbArr[ i ][ j ];
     }
     else
     {
         len1 =  varProbArr.length;
         len2 =  varProbArr[0].length;
         probArr1 = new int[ len1 ][ len2 ];
         for( int i = 0; i < len1; i++ )
           for( int j = 0; j < len2; j++ )
         probArr1[ i ][ j ] = varProbArr[ i ][ j ];
     }

     if( expIdx == 0 )
     {
         len3 =  transAddProbArr.length;
         len4 =  transAddProbArr[0].length;
         probArr2 = new int[ len3 ][ len4 ];
         for( int i = 0; i < len3; i++ )
            for( int j = 0; j < len4; j++ )
                 probArr2[ i ][ j ] = transAddProbArr[ i ][ j ];
     }
     else
     {
         len3 =  transSubProbArr.length;
         len4 =  transSubProbArr[0].length;
         probArr2 = new int[ len3 ][ len4 ];
         for( int i = 0; i < len3; i++ )
            for( int j = 0; j < len4; j++ )
                 probArr2[ i ][ j ] = transSubProbArr[ i ][ j ];
     }

     String str = "PROBLEMS IN TRAINING: \n";
     for( int i = 0; i < len1; i++ )
       for( int j = 0; j < len2; j++ )
       {
          if( j == 0 )
          {
              char ch = (char)(probArr1[ i ][ j ] + 64);
              int chVal = (probArr1[ i ][ j ] + 64);
              str += "The start letter is:" + String.valueOf( ch ) + " ";
              str += "an the val is:" + String.valueOf( chVal );
          }
          else
          {
              int num = probArr1[ i ][ j ];
              str += " and the number is:" + String.valueOf( num ) + "\n";
          }
       }

      probArea1.setText( str );
/****************************************************************************
      dataFile.saveStringData( str );
/****************************************************************************/

      str = "PROBLEMS IN TRANSFER: \n";
      for( int i = 0; i < len3; i++ )
       for( int j = 0; j < len4; j++ )
       {
          if( j == 0 )
          {
              char ch = (char)(probArr2[ i ][ j ] + 64);
              int chVal = (probArr2[ i ][ j ] + 64);
              str += "The start letter is:" + String.valueOf( ch ) + " ";
              str += "an the val is:" + String.valueOf( chVal );
          }
          else
          {
              int num = probArr2[ i ][ j ];
              str += " and the number is:" + String.valueOf( num ) + "\n";
          }
       }

      probArea2.setText( str );
/****************************************************************************
      dataFile.saveStringData( str );
/****************************************************************************/
   }

   public void showRules()
   {
   }

   public void startExp()
   {
      expAgent.setWholeProbSet( wholeProbSet );
      //for( expIdx = 1; expIdx < EXP_NUM; expIdx++ )
      for( expIdx = 0; expIdx < EXP_NUM; expIdx++ )
      {
         System.out.print(" the current experiment is: " );
         System.out.println( expIdx );

         showRules();
         for( groupIdx = 0; groupIdx < GROUP_NUM; groupIdx++ )
         {
            System.out.print(" the current group is: " );
            System.out.println( groupIdx );

            showProblems();
            setProblemSet();
            setEnvironment();

/************************************************************************/
            blockNum = BLOCK_NUMS[ groupIdx ][ 0 ];
            trainHitRates = new double[blockNum];
            for( int i = 0; i < blockNum; i++ )
                 trainHitRates[ i ] = 0;
            transHitRates = new double[3];
            for( int i = 0; i < 3; i++ )
                 transHitRates[ i ] = 0;
/************************************************************************/

            for( agentIdx = 0; agentIdx < AGENT_NUM; agentIdx++ )
            {
               System.out.print(" the current agent is: " );
               System.out.println( agentIdx );
               setAgentNo( agentIdx );
               solveProblems();

/***************************************************************************/
               double[] rates1 = new double[blockNum];
               rates1 = expAgent.getTrainHitRate();
               for( int i = 0; i < blockNum; i++ )
                    trainHitRates[ i ] += rates1[ i ];

               double[] rates2 = new double[3];
               rates2 = expAgent.getTransHitRate();
               for( int i = 0; i < 3; i++ )
                    transHitRates[ i ] += rates2[ i ];
/***************************************************************************/

               getAgentRTInfo();
               getAgentBLAInfo();
               getAgentParaInfo();

               saveAgentData();
               initNewAgent();
            }

/***********************************************************************/
            //saveOnTargets();
/***********************************************************************/

            saveGroupData();
            initNewGroup();
         }
      }
   }

   public void setProblemSet()
   {
       if( expIdx == 0 )
       {
           if( groupIdx == 0 )
               expAgent.obtainProblems( conProbArr,  transAddProbArr );
           else
               expAgent.obtainProblems( varProbArr,  transAddProbArr );
       }
       else
       {
           if( groupIdx == 0 )
               expAgent.obtainProblems( conProbArr,  transSubProbArr );
           else
               expAgent.obtainProblems( varProbArr,  transSubProbArr );
       }
   }

   public void setEnvironment()
   {
      expAgent.setVariables( expIdx, groupIdx, TRAIN_TRIAL_NUM, TRANS_TRIAL_NUM,
                             BLOCK_NUMS[ groupIdx ], TRIAL_NUMS[ groupIdx ] );
   }

   public void setAgentNo( int no )
   {
      expAgent.setAgentNo( no );
   }

   public void solveProblems()
   {
      expAgent.solveProblems();
   }

   public void getAgentRTInfo()
   {
      expAgent.getRTArrays( trainRTArray[ groupIdx ][ agentIdx ],
                            transRTArray[ groupIdx ][ agentIdx ] );
   }

   public void getAgentBLAInfo()
   {
      blaInfo[ agentIdx ] = expAgent.getBLA();
   }

   public void getAgentParaInfo()
   {
      expAgent.getPerformanceArrays( trainPossArr[ groupIdx ][ agentIdx ],
                                     transPossArr[ groupIdx ][ agentIdx ],
                                     trainRetrArr[ groupIdx ][ agentIdx ],
                                     transRetrArr[ groupIdx ][ agentIdx ],
                                     trainContArr[ groupIdx ][ agentIdx ],
                                     transContArr[ groupIdx ][ agentIdx ],
                                     trainBttmArr[ groupIdx ][ agentIdx ],
                                     transBttmArr[ groupIdx ][ agentIdx ],
                                     trainRateArr[ groupIdx ][ agentIdx ],
                                     transRateArr[ groupIdx ][ agentIdx ]
                                   );
   }

   public void initNewAgent()
   {
      expAgent.reinitAgent();
   }

   public void initNewGroup()
   {
      for( int i = 0; i < AGENT_NUM; i++ )
         for( int j = 0; j < TRAIN_TRIAL_NUM; j++ )
              trainRTArray[groupIdx][i][j] = 0;
      for( int i = 0; i < AGENT_NUM; i++ )
         for( int j = 0; j < TRANS_TRIAL_NUM; j++ )
              transRTArray[groupIdx][i][j] = 0;
   }

   public void saveOnTargets()
   {
     String str = "CURRENT EXPERIMENT IS: ";
     str += String.valueOf( expIdx + 1 ) + "\n";
     str += "CURRENT GROUP IS: ";
     if( groupIdx == 0 )
         str += " consistent " + "\n";
     else
         str += " varied " + "\n";

     str += "The ON-TARGET RATE OF EACH BLOCK IN TRAINING IS: \n";
     double sum  = 0.0;
     for( int i = 0; i < blockNum; i++ )
     {
          double rate = trainHitRates[i] / AGENT_NUM;
          sum += rate;
          str += String.valueOf( rate ) + "\n";
     }
     double avg = sum / blockNum;
     str += "The AVERAGE ON-TARGET RATE IN TRAINING IS: \n";
     str += String.valueOf( avg ) + "\n";

     str += "The ON-TARGET RATE OF EACH BLOCK IN TRANSFER IS: \n";
     sum = 0.0;
     for( int i = 0; i < 3; i++ )
     {
          double rate = transHitRates[i] / AGENT_NUM;
          sum += rate;
          str += String.valueOf( rate ) + "\n";
     }
     avg = sum / 3;
     str += "The AVERAGE ON-TARGET RATE IN TRANSFER IS: \n";
     str += String.valueOf( avg ) + "\n";

     str += "\n";

     dataFile.saveStringData( str );
   }

   public void saveCombProbilities()
   {
     String str1, str2, str3, str4;

     str1 = str2 = "";
     for( int i = 0; i < TRAIN_TRIAL_NUM; i++ )
     {
        double sum = 0.0;
        for( int j = 0; j < AGENT_NUM; j++ )
             sum += trainPossArr[ groupIdx ][ j ][ i ];
        sum /= AGENT_NUM;
        str1 += String.valueOf( i + 1 ) + " ";
        str1 += String.valueOf( sum ) + "\n";
        str2 += String.valueOf( i + 1 ) + " ";
        str2 += String.valueOf( 1-sum ) + "\n";

     }
     expDatFiles[ expIdx ][ groupIdx ][ 0 ][ 1 ].saveStringData( str1 );
     expDatFiles[ expIdx ][ groupIdx ][ 0 ][ 2 ].saveStringData( str2 );

     str1 = str2 = "";
     for( int i = 0; i < TRANS_TRIAL_NUM; i++ )
     {
        double sum = 0.0;
        for( int j = 0; j < AGENT_NUM; j++ )
             sum += transPossArr[ groupIdx ][ j ][ i ];
        sum /= AGENT_NUM;
        str1 += String.valueOf( i + 1 ) + " ";
        str1 += String.valueOf( sum ) + "\n";
        str2 += String.valueOf( i + 1 ) + " ";
        str2 += String.valueOf( 1-sum ) + "\n";
     }

     expDatFiles[ expIdx ][ groupIdx ][ 1 ][ 1 ].saveStringData( str1 );
     expDatFiles[ expIdx ][ groupIdx ][ 1 ][ 2 ].saveStringData( str2 );

     str1 = str2 = str3 = str4 = "";
     for( int i = 0; i < EPISODE_NUM; i++ )
     {
        double sum1 = 0.0, sum2 = 0.0, sum3 = 0.0, sum4 = 0.0;
        for( int j= 0; j < CON_PROB_NUM; j++ )
          for( int k = 0; k < AGENT_NUM; k++ )
          {
               sum1 += trainRetrArr[ groupIdx ][ k ][ i * CON_PROB_NUM + j ];
               sum2 += trainContArr[ groupIdx ][ k ][ i * CON_PROB_NUM + j ];
               sum3 += trainBttmArr[ groupIdx ][ k ][ i * CON_PROB_NUM + j ];
               sum4 += trainRateArr[ groupIdx ][ k ][ i * CON_PROB_NUM + j ];
          }

        sum1 /= AGENT_NUM * CON_PROB_NUM;
        sum2 /= AGENT_NUM * CON_PROB_NUM;
        sum3 /= AGENT_NUM * CON_PROB_NUM;
        sum4 /= AGENT_NUM * CON_PROB_NUM;

        str1 += String.valueOf( i + 1 ) + " ";
        str1 += String.valueOf( sum1 ) + "\n";
        str2 += String.valueOf( i + 1 ) + " ";
        str2 += String.valueOf( sum2 ) + "\n";
        str3 += String.valueOf( i + 1 ) + " ";
        str3 += String.valueOf( sum3 ) + "\n";
        str4 += String.valueOf( i + 1 ) + " ";
        str4 += String.valueOf( sum4 ) + "\n";
     }

     expDatFiles[ expIdx ][ groupIdx ][ 0 ][ 3 ].saveStringData( str1 );
     expDatFiles[ expIdx ][ groupIdx ][ 0 ][ 4 ].saveStringData( str2 );
     expDatFiles[ expIdx ][ groupIdx ][ 0 ][ 5 ].saveStringData( str3 );
     expDatFiles[ expIdx ][ groupIdx ][ 0 ][ 6 ].saveStringData( str4 );

     str1 = str2 = str3 = str4 = "";
     for( int i = 0; i < EPISODE_NUM; i++ )
     {
        double sum1 = 0.0, sum2 = 0.0, sum3 = 0.0, sum4 = 0.0;
        for( int j= 0; j < 1; j++ )
          for( int k = 0; k < AGENT_NUM; k++ )
          {
               sum1 += transRetrArr[ groupIdx ][ k ][ i * 1 + j ];
               sum2 += transContArr[ groupIdx ][ k ][ i * 1 + j ];
               sum3 += transBttmArr[ groupIdx ][ k ][ i * 1 + j ];
               sum4 += transRateArr[ groupIdx ][ k ][ i * 1 + j ];
          }

        sum1 /= AGENT_NUM * 1;
        sum2 /= AGENT_NUM * 1;
        sum3 /= AGENT_NUM * 1;
        sum4 /= AGENT_NUM * 1;

        str1 += String.valueOf( i + 1 ) + " ";
        str1 += String.valueOf( sum1 ) + "\n";
        str2 += String.valueOf( i + 1 ) + " ";
        str2 += String.valueOf( sum2 ) + "\n";
        str3 += String.valueOf( i + 1 ) + " ";
        str3 += String.valueOf( sum3 ) + "\n";
        str4 += String.valueOf( i + 1 ) + " ";
        str4 += String.valueOf( sum4 ) + "\n";
     }

     expDatFiles[ expIdx ][ groupIdx ][ 1 ][ 3 ].saveStringData( str1 );
     expDatFiles[ expIdx ][ groupIdx ][ 1 ][ 4 ].saveStringData( str2 );
     expDatFiles[ expIdx ][ groupIdx ][ 1 ][ 5 ].saveStringData( str3 );
     expDatFiles[ expIdx ][ groupIdx ][ 1 ][ 6 ].saveStringData( str4 );
   }

   public void saveAgentData()
   {
     String str = "CURRENT EXPERIMENT IS: ";
     str += String.valueOf( expIdx + 1 ) + "\n";

     str += "CURRENT GROUP IS: ";
     if( groupIdx == 0 )
         str += " consistent " + "\n";
     else
         str += " varied " + "\n";

     str += "CURRENT AGENT IS: ";
     str += String.valueOf( agentIdx ) + "\n";

     String headStr = str;

     str += "The MEDIAN RESPONSE TIME OF EACH BLOCK IN TRAINING IS: \n";
     for( int i = 0; i < EPISODE_NUM; i++ )
     {
          int medianVal = getMedianVal( trainRTArray[groupIdx][agentIdx],
                                        i, CON_PROB_NUM );
          str += String.valueOf( medianVal ) + "\n";
     }

     str += "The MEDIAN RESPONSE TIME IN TRANSFER IS: \n";
     int medianVal = getMedianVal( transRTArray[groupIdx][agentIdx],
                                   0, EPISODE_NUM );
     str += String.valueOf( medianVal ) + "\n";
     agentDatFile.saveStringData( str );

     // save the RER RULES.
     ruleDatFile.saveStringData( headStr );
     if( RER_FLAG )
     {
         StringBuffer ruleSet[] = new StringBuffer[3];
         for( int i = 0; i < 3; i++ )
              ruleSet[ i ] = new StringBuffer( "" );

         expAgent.fillRerRule( ruleSet );
         for( int i = 0; i < 3; i++ )
              ruleDatFile.saveStringData( ruleSet[i].substring( 0 ) );
     }
   }

   public void saveGroupData()
   {
     int[] trainRTDat = new int[ EPISODE_NUM ];
     int[] transRTDat = new int[ EPISODE_NUM ];

     saveCombProbilities();

     String str = "CURRENT EXPERIMENT IS: ";
     str += String.valueOf( expIdx + 1 ) + "\n";
     str += "CURRENT GROUP IS: ";
     if( groupIdx == 0 )
         str += " consistent " + "\n";
     else
         str += " varied " + "\n";

     str += "The MEAN RESPONSE TIME OF EACH BLOCK IN TRAINING IS: \n";
     String str1 = "";
     for( int i = 0; i < EPISODE_NUM; i++ )
     {
        int count = 0, sum = 0;
        for( int j = 0; j < AGENT_NUM; j++ )
        {
             int medianVal = getMedianVal( trainRTArray[groupIdx][j],
                                            i, CON_PROB_NUM );
             if( medianVal > 0 )
             {
                 sum += medianVal;
                 count++;
             }
        }

        if( count == 0 )
        {
            if( i > 0 )
                sum = trainRTDat[ i - 1 ];
             else
                sum = MAX_TIME;
            count++;
        }
        sum /= count;
        trainRTDat[ i ] = sum;

        str1 += String.valueOf( i + 1 ) + " ";
        str1 += String.valueOf( trainRTDat[ i ] ) + "\n";
     }
/****************************************************************************/
     expDatFiles[expIdx][groupIdx][0][0].saveStringData( str1 );
/****************************************************************************/
     str += str1 + "\n";

     /*
     str += "The MEAN RESPONSE TIME OF EACH BLOCK IN TRANSFER IS: \n";
     String str2 = "";
     for( int i = 0; i < EPISODE_NUM; i++ )
     {

        int count = 0, sum = 0;
        for( int j = 0; j < AGENT_NUM; j++ )
        {
             if( transRTArray[groupIdx][j][i] > -1 )
             {
                 sum += transRTArray[groupIdx][j][i];
                 count++;
             }
        }

        if( count == 0 )
            count++;

        sum /= count;
        transRTDat[ i ] = sum;
        str2 += String.valueOf( transRTDat[ i ] ) + "\n";
     }
/***************************************************************************
     expDatFiles[expIdx][groupIdx][1][0].saveStringData( str2 );
/***************************************************************************
     str += str2 + "\n";
     */

     str += "The MEAN RESPONSE TIME IN TRANSFER IS: \n";
     int count = 0, sum = 0;
     for( int j = 0; j < AGENT_NUM; j++ )
     {
          int medianVal = getMedianVal( transRTArray[groupIdx][j],
                                        0, EPISODE_NUM );
          if( medianVal > 0 )
          {
              sum += medianVal;
              count++;
          }
     }

     if( count == 0 )
     {
         sum = 0;
         count++;
     }
     sum /= count;
     str += String.valueOf( sum ) + "\n\n";

     dataFile.saveStringData( str );
   }

   public int getMedianVal( int[] dat, int episode, int length )
   {
      int count = 0;
      int[] arr = new int[ length ];
      for( int i = 0; i < length; i++ )
      {
           if( dat[ episode * length + i ] != -1 )
               arr[ count++ ] = dat[ episode * length + i ];
      }

      if( count == 0 )
          return -1;

      Arrays.sort( arr, 0, count );
      int val = 0;
      if( ( count % 2 ) == 1 )
            val = arr[ count / 2 ];
      else
            val = ( arr[count/2 - 1] + arr[count/2] ) / 2;
      return val;
   }

   public String getTaskName()
   {
      return "LETTER COUNTING TASK";
   }

   public void start( int expMode, int blockNum, WriteFile dFile )
   {
       dataFile = dFile;
       setEnabled( true );
       setVisible( true );
       createExpDataFiles();
       initFixedRules();
       initAgent();
   }

   public void createExpDataFiles()
   {
      agentDatFile = new WriteFile();
      agentDatFile.openFile( 1, "lcAgentDat" );

      ruleDatFile = new WriteFile();
      ruleDatFile.openFile( 1, "lcRerRuleDat" );

      String nameArr[] = { "mean", "probUp", "probDown",
                           "retrPerc", "contPerc", "bttmPerc",
                           "correctRate" };
      expDatFiles = new WriteFile[ EXP_NUM ][ GROUP_NUM ][ STAGE_NUM ][ FILE_TYPE_NUM ];
      for( int i = 0; i < EXP_NUM; i++ )
        for( int j = 0; j < GROUP_NUM; j++ )
           for( int k = 0; k < STAGE_NUM; k++ )
            for( int p = 0; p < FILE_TYPE_NUM; p++ )
            {
               expDatFiles[i][j][k][p] = new WriteFile();
               String str = "exp" + String.valueOf( i + 1 ) +
                            "grp" + String.valueOf( j + 1 ) +
                            "stg" + String.valueOf( k + 1 ) + nameArr[ p ];

               expDatFiles[i][j][k][p].openFile( 1, str );
            }
   }

   public void initFixedRules()
   {
       ruleNum = ruleNumArr[ simNo ];
       lcRules = new LCFixRule[ ruleNum ];
       condition = new ConItem[ ruleNum ][];
       action = new ConItem[ ruleNum ][];

       for( int  i = 0; i < ruleNum; i++ )
       {
          int len = ruleConditionArr[i].length;
          condition[ i ] = new ConItem[ len ];
          for( int j = 0; j < len; j++ )
          {
             Vector item1 = new Vector( 1 );
             item1.addElement( ruleConditionArr[ i ][ j ] );
             condition[i][j] = new ConItem( item1 );
          }

          len = ruleActionArr[i].length;
          action[ i ] = new ConItem[ len ];
          for( int j = 0; j < len; j++ )
          {
             Vector item2 = new Vector( 2 );
             item2.addElement( ruleActionArr[ i ][ j ] );
             int[] elem = new int[1];
             elem[ 0 ] = 0;
             item2.addElement( elem );
             action[i][j] = new ConItem( item2 );
          }

          lcRules[ i ] = new LCFixRule( condition[i], action[i], World.LC_TASK,
                             TRAIN_TRIAL_NUM + TRANS_TRIAL_NUM, INIT_BLA );
       }
   }

   public void initAgent()
   {
       expAgent = new LCAgent( simNo, ruleNumArr[simNo], RULE_BASED,
                  TRAIN_TRIAL_NUM + TRANS_TRIAL_NUM, FR_FLAG, RER_FLAG,
                  GKN_DIM_VAL_NUMS, QBP_DIM_VAL_NUMS, CHUNK_DIM_TYPES, lcRules );
   }
}