// process control task.
package tasksimulators.oldsimu;
import java.lang.*;
import clarion.acs.*;
import clarion.common.*;
import clarion.tools.*;

public class PCAgent {

            private final int NO_MERGE = 0;
            private final int TEN_TRIALS_MERGE = 1;
            private final int EACH_TRIAL_MERGE = 2;
            private final static double POS_FEEDBACK = 1;
	    private final static double NEG_FEEDBACK = -0.2;


	    private int start, target, trialIdx, curState[];
            private int NET_DIM_NUM, NET_INPUT_NUM;
            private int NET_HIDDEN_NUM, NET_OUTPUT_NUM;

	    int action, P1, startAction, startP1, nodeInfoArr[];
	    double qValue1, qValue2;  // used in computing the error in the
                                      // bottom level.
            double feedback;          // get feeback after the chosen action
                                      // is performed.

            double threshold;      // for testing RER success;

            // modified by xi zhang 02/05/01
            double [] weightQR;

            double [] qbpArray;
            double [] irlArray;
            double [] rerArray;
            double [] fixArray;    // all are used to store recommended
                                    // actions.

            boolean irlRuleFlag;
            boolean rerRuleFlag;     // Flags to test if IRL OR RER is used.
            boolean fixRuleFlag;
            boolean smpRuleFlag;

	    PCIrlRule     irlRule;
            T_RerRuleSet  rerRule;
            PCFixRule   fixRule;
            SimpleRule  smpRule;

            double [] totDeciderInputs;
                                    // input of Boltzmann distribution to
                                    // decide probabilitically on an action
                                    // to be performed.
	    StochasticDecider totDecider;

            QBackPropNet mycon; // implicit rules need not be aware of target
	    			   // so long as target is constant

            StringBuffer rerRuleStrings[];
            StringBuffer irlRuleStrings[];
            StringBuffer fixRuleStrings[];

    public PCAgent( int v1, int v2, int targetLevel )
    {
           NET_DIM_NUM = ( int )World.TASK_PARA_ARRAY[World.NET_DIM_NUM ];
           NET_INPUT_NUM = (int)World.TASK_PARA_ARRAY[World.NET_INPUT_NUM ];
           NET_HIDDEN_NUM =(int)World.TASK_PARA_ARRAY[World.NET_HIDDEN_NUM ];
           NET_OUTPUT_NUM = (int)World.TASK_PARA_ARRAY[World.NET_OUTPUT_NUM];

           nodeInfoArr = new int[ 5 ];
           nodeInfoArr[ 0 ] = NET_DIM_NUM;
           nodeInfoArr[ 1 ] = NET_DIM_NUM;
           nodeInfoArr[ 2 ] = NET_INPUT_NUM;
           nodeInfoArr[ 3 ] = NET_HIDDEN_NUM;
           nodeInfoArr[ 4 ] = NET_OUTPUT_NUM;

           startAction = v1;
           startP1 = v2;
           initVariables( v1, v2 );

           target = targetLevel;
           threshold = World.TASK_PARA_ARRAY[ World.HIGH_RER_THRESHOLD ];

            qbpArray = new double[NET_OUTPUT_NUM];
	    irlArray = new double[NET_OUTPUT_NUM];
	    rerArray = new double[NET_OUTPUT_NUM];
	    fixArray = new double[NET_OUTPUT_NUM];

	    totDeciderInputs = new double[ NET_OUTPUT_NUM ];
	    totDecider = new StochasticDecider( NET_OUTPUT_NUM,
                                   World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );

	    mycon = new QBackPropNet( NET_INPUT_NUM, NET_HIDDEN_NUM,
                        NET_OUTPUT_NUM, NET_DIM_NUM, target );

            curState = new int[ NET_INPUT_NUM ];

            irlRule = new PCIrlRule( target, action, P1 );
            irlRule.setParams( action, P1 );
            irlRuleStrings = new StringBuffer[3];
            for ( int i = 0; i < 3; i++ )
                  irlRuleStrings[i] = new StringBuffer();

            rerRule = new T_RerRuleSet( 0, EACH_TRIAL_MERGE, nodeInfoArr, false );

            rerRuleStrings = new StringBuffer[3];
            for ( int i = 0; i < 3; i++ )
                  rerRuleStrings[i] = new StringBuffer();

            smpRule = new SimpleRule( target, action, P1 );
            fixRule = new PCFixRule( target, action, P1 );
            fixRuleStrings = new StringBuffer[3];
            for ( int i = 0; i < 3; i++ )
                  fixRuleStrings[ i ] = new StringBuffer();
   }

   public void initVariables( int startInput, int startOutput )
   {
      action = startInput;
      P1 = startOutput;
   }

   public void reinitBlock( int startInput, int startOutput )
   {
       initVariables( startInput, startOutput );
       mycon.reinitBlock();
   }

   public void reinitAgent()
   {
       reinitBlock( startAction, startP1 );
       mycon.restoreInitWeights();

       if( irlRuleFlag )
           irlRule.reinit( startAction, startP1 );

       if( rerRuleFlag )
           rerRule.reinit();
   }

   public void getInitVals( int[] arr )
   {
       arr[ 0 ] = startAction;
       arr[ 1 ] = startP1;
   }

   public void setRules( boolean[] flagArray )
   {
      irlRuleFlag = flagArray[ 0 ];
      rerRuleFlag = flagArray[ 1 ];
      fixRuleFlag = flagArray[ 2 ];
      smpRuleFlag = flagArray[ 3 ];
   }

   public double[][] getInitIHWeights()
   {
      return mycon.getInitInputToHiddenWeights();
   }

   public double[][] getInitHOWeights()
   {
      return mycon.getInitHiddenToOutputWeights();
   }

   public void setInitWeights( double[][] ItoHWeights, double[][] HtoOWeights  )
   {
      mycon.setInitInputToHiddenWeights( ItoHWeights );
      mycon.setInitHiddenToOutputWeights( HtoOWeights );
   }

   public double[][] getIHWeights()
   {
      return mycon.getInputToHiddenWeights();
   }

   public double[][] getHOWeights()
   {
      return mycon.getHiddenToOutputWeights();
   }

   public void getThresholds( double hThresholds[], double oThresholds[] )
   {
      mycon.getHiddenThresholds( hThresholds );
      mycon.getOutputThresholds( oThresholds );
   }

   public void setRuleThreshold()
   {
      if( rerRuleFlag )
      {
          threshold = World.TASK_PARA_ARRAY[ World.LOW_RER_THRESHOLD ];
          rerRule.setThreshold( World.TASK_PARA_ARRAY[ World.LOW_RER_THRESHOLD1 ],
                                World.TASK_PARA_ARRAY[ World.LOW_RER_THRESHOLD2 ] );
      }
      if( irlRuleFlag )
          irlRule.setThreshold( World.TASK_PARA_ARRAY[ World.HIGH_IRL_THRESHOLD ]);
   }

   public int decideAction( int idx )
   {
        // get the Q values from the bottom level
        trialIdx = idx;
        mycon.setCurState( action, P1, trialIdx );
        mycon.setInput();
        mycon.forwardPass();
	mycon.getQvals( qbpArray );

        // decide current action, it has four types
        // no rules applied
        if( !irlRuleFlag && !rerRuleFlag && !fixRuleFlag )
        {
            weightQR = new double[1];
            weightQR[ 0 ] = 1;

            for (int i = 0; i < NET_OUTPUT_NUM; i++ )
                 totDeciderInputs[i] = weightQR[ 0 ] * qbpArray[i];

            totDecider.setElements( totDeciderInputs );
            action = totDecider.getRecommendation();
        }

       // only IRL rule applied
        if( irlRuleFlag && !rerRuleFlag && !fixRuleFlag )
        {
            weightQR = new double[2];
            weightQR[ 0 ] = 0.8;
            weightQR[ 1 ] = 0.2;

            irlRule.getRecommendedAction( irlArray );
            irlRuleStrings[0] = irlRule.getAllRules();
            irlRuleStrings[1] = irlRule.getMatchRules();

            for (int i = 0; i < NET_OUTPUT_NUM; i++ )
                 totDeciderInputs[i] = qbpArray[i] * weightQR[ 0 ] +
                                       irlArray[i] * weightQR[ 1 ];

            //decide action by boltmann distribution.
            totDecider.setElements( totDeciderInputs );
            action = totDecider.getRecommendation();
        }

        // only RER rule applied
        if( !irlRuleFlag && rerRuleFlag && !fixRuleFlag )
        {
           weightQR = new double[2];
           weightQR[ 0 ] = 0.8;
           weightQR[ 1 ] = 0.2;

           mycon.getState( curState );
           rerRule.getActions( curState, rerArray );
           rerRuleStrings[0] = rerRule.getAllRules();
           rerRuleStrings[1] = rerRule.getMatchRules();

	   for( int i = 0; i < NET_OUTPUT_NUM; i++ )
                totDeciderInputs[i] = qbpArray[i] * weightQR[ 0 ] +
                                      rerArray[i] * weightQR[ 1 ];

           //decide action by boltmann distribution.
           totDecider.setElements( totDeciderInputs );
           action = totDecider.getRecommendation();
        }

        // both IRL and RER rule applied
        if( irlRuleFlag && rerRuleFlag && !fixRuleFlag )
        {
           weightQR = new double[3];
           weightQR[ 0 ] = World.TASK_PARA_ARRAY[ World.WBL ] +
                           World.TASK_PARA_ARRAY[ World.WFIX ];
           weightQR[ 1 ] = World.TASK_PARA_ARRAY[ World.WIRL ];
           weightQR[ 2 ] = World.TASK_PARA_ARRAY[ World.WRER ];

           mycon.getState( curState );
           rerRule.getActions( curState, rerArray );
           rerRuleStrings[0] = rerRule.getAllRules();
           rerRuleStrings[1] = rerRule.getMatchRules();

           irlRule.getRecommendedAction( irlArray );
           irlRuleStrings[0] = irlRule.getAllRules();
           irlRuleStrings[1] = irlRule.getMatchRules();

	   for( int i = 0; i < NET_OUTPUT_NUM; i++ )
                totDeciderInputs[i] = qbpArray[i] * weightQR[ 0 ] +
                                      irlArray[i] * weightQR[ 1 ] +
                                      rerArray[i] * weightQR[ 2 ] ;

           //decide action by boltmann distribution.
           totDecider.setElements( totDeciderInputs );
           action = totDecider.getRecommendation();
        }

        // only simple rule applied.
        if( !irlRuleFlag && !rerRuleFlag && fixRuleFlag )
        {
           weightQR = new double[ 2 ];
           weightQR[ 0 ] = 0.8;
           weightQR[ 1 ] = 0.2;

           if( !smpRuleFlag )
           {
               fixArray = fixRule.getRecommendedAction();
               fixRuleStrings[0] = fixRule.getRules();
           }
           else
           {
               fixArray = smpRule.getRecommendedAction();
               fixRuleStrings[0] = smpRule.getRules();
           }

	   for( int i = 0; i < NET_OUTPUT_NUM; i++ )
                totDeciderInputs[i] = qbpArray[i] * weightQR[ 0 ] +
                                      fixArray[i] * weightQR[ 1 ];

           //decide action by boltmann distribution.
           totDecider.setElements( totDeciderInputs );
           action = totDecider.getRecommendation();
        }

        // simple rule & IRL rule applied.
        if( irlRuleFlag && !rerRuleFlag && fixRuleFlag )
        {
           weightQR = new double[ 3 ];
           weightQR[ 0 ] = 0.6;
           weightQR[ 1 ] = 0.2;
           weightQR[ 2 ] = 0.2;

           irlRule.getRecommendedAction( irlArray );
           irlRuleStrings[0] = irlRule.getAllRules();
           irlRuleStrings[1] = irlRule.getMatchRules();

           if( !smpRuleFlag )
           {
               fixArray = fixRule.getRecommendedAction();
               fixRuleStrings[0] = fixRule.getRules();
           }
           else
           {
               fixArray = smpRule.getRecommendedAction();
               fixRuleStrings[0] = smpRule.getRules();
           }
	   for( int i = 0; i < NET_OUTPUT_NUM; i++ )
                totDeciderInputs[i] = qbpArray[i] * weightQR[ 0 ] +
                                      irlArray[i] * weightQR[ 1 ] +
                                      fixArray[i] * weightQR[ 2 ];

           //decide action by boltmann distribution.
           totDecider.setElements( totDeciderInputs );
           action = totDecider.getRecommendation();
        }

        // simple rule & RER rule applied.
        if( !irlRuleFlag && rerRuleFlag && fixRuleFlag )
        {
           weightQR = new double[ 3 ];
           weightQR[ 0 ] = 0.6;
           weightQR[ 1 ] = 0.2;
           weightQR[ 2 ] = 0.2;

           mycon.getState( curState );
           rerRule.getActions( curState, rerArray );
           rerRuleStrings[0] = rerRule.getAllRules();
           rerRuleStrings[1] = rerRule.getMatchRules();

           if( !smpRuleFlag )
           {
               fixArray = fixRule.getRecommendedAction();
               fixRuleStrings[0] = fixRule.getRules();
           }
           else
           {
               fixArray = smpRule.getRecommendedAction();
               fixRuleStrings[0] = smpRule.getRules();
           }

	   for( int i = 0; i < NET_OUTPUT_NUM; i++ )
                totDeciderInputs[i] = qbpArray[i] * weightQR[ 0 ] +
                                      rerArray[i] * weightQR[ 1 ] +
                                      fixArray[i] * weightQR[ 2 ];

           //decide action by boltmann distribution.
           totDecider.setElements( totDeciderInputs );
           action = totDecider.getRecommendation();
        }

        // simple rule & IRL rule & RER rule applied.
        if( irlRuleFlag && rerRuleFlag && fixRuleFlag )
        {
           weightQR = new double[ 4 ];
           weightQR[ 0 ] = World.TASK_PARA_ARRAY[ World.WBL ];
           weightQR[ 1 ] = World.TASK_PARA_ARRAY[ World.WIRL ];
           weightQR[ 2 ] = World.TASK_PARA_ARRAY[ World.WRER ];
           weightQR[ 3 ] = World.TASK_PARA_ARRAY[ World.WFIX ];

           mycon.getState( curState );
           rerRule.getActions( curState, rerArray );
           rerRuleStrings[0] = rerRule.getAllRules();
           rerRuleStrings[1] = rerRule.getMatchRules();

           irlRule.getRecommendedAction( irlArray );
           irlRuleStrings[0] = irlRule.getAllRules();
           irlRuleStrings[1] = irlRule.getMatchRules();

           if( !smpRuleFlag )
           {
               fixArray = fixRule.getRecommendedAction();
               fixRuleStrings[0] = fixRule.getRules();
           }
           else
           {
               fixArray = smpRule.getRecommendedAction();
               fixRuleStrings[0] = smpRule.getRules();
           }

	   for( int i = 0; i < NET_OUTPUT_NUM; i++ )
                totDeciderInputs[i] = qbpArray[i] * weightQR[ 0 ] +
                                      irlArray[i] * weightQR[ 1 ] +
                                      rerArray[i] * weightQR[ 2 ] +
                                      fixArray[i] * weightQR[ 3 ];

           //decide action by boltmann distribution.
           totDecider.setElements( totDeciderInputs );
           action = totDecider.getRecommendation();
        }

        // obtain the current state QVaule1
        qValue1 = mycon.getQVal( action );
        return action;
   }

   public boolean update( int newOutput, int onTargetNum[],
                          int blockIdx,  int trialIdx )
   {
       //obtain next state QValue2.
        P1 = newOutput;
	mycon.setCurState( action, P1, trialIdx + 1 );
	mycon.setInput();
        mycon.forwardPass();
        qValue2 = mycon.getMaxQVal();

	// get feedback from the new output resulting from performing
        // the chosen action.
        feedback = determineFeedback( P1 );

        // count the numbers of ON-TARGET.
        if( feedback > 0 )
            onTargetNum[ blockIdx ]++;

	//let QBP know what action was performed
	mycon.setPerformedAction( action );

        //update the reinforcement
        mycon.setReinforcement( feedback );

        // update the Q-values of the previous state.
        mycon.setPreState( trialIdx );
        mycon.setInput();
        mycon.forwardPass();
	mycon.computeErrors( qValue1, qValue2 );
	mycon.backwardPass();

        if( fixRuleFlag )
        {
           if( !smpRuleFlag )
                fixRule.setParams( action, P1 );
           else
	        smpRule.setParams( action, P1 );
        }

        //update upper RER rule set if the RER flag is set.
        if( rerRuleFlag )
        {
            double r = mycon.getReinforcement();
            double d = mycon.getDiscount();

            boolean discountFlag = false;
            if( trialIdx == NET_OUTPUT_NUM - 1 )
                discountFlag = true;
            else
                discountFlag = false;

            if( r + d * qValue2 - qValue1 > threshold )
                rerRule.updateRuleSet( curState, action, true, discountFlag, true );
            else
                rerRule.updateRuleSet( curState, action, false, discountFlag, true );

            rerRuleStrings[ 2 ] = rerRule.getChosenRules();
        }

        //update upper level IRL rule set if the IRL flag is set.
        if( irlRuleFlag )
        {
	    irlRule.setParams( action, P1 );
            boolean retVal = irlRule.updateRule( action, feedback, trialIdx  );
            irlRuleStrings[ 2 ] = irlRule.getChosenRules();
            return retVal;
        }
        else
            return true;
   }

   public void fillRerRule( StringBuffer str[] )
   {
        for( int i = 0; i < 3; i++ )
        {
            str[i].delete( 0, str[i].length() );
            str[i].insert( 0, rerRuleStrings[i].substring( 0 ) );
        }
   }

   public void fillIrlRule( StringBuffer str[] )
   {
        for( int i = 0; i < 3; i++ )
        {
            str[i].delete( 0, str[i].length() );
            str[i].insert( 0, irlRuleStrings[i].substring( 0 ) );
        }
   }

   public void fillFixRule( StringBuffer str )
   {
      str.delete( 0, str.length() );
      str.insert( 0, fixRuleStrings[ 0 ].substring( 0 ) );
   }

   public int getMatchCounter()
   {
      int val;
      if( rerRuleFlag )
      {
          val = rerRule.getMatchCounter();
          return val;
      }
      else
         return -1;
   }

   public int getChosenCounter()
   {
      int val;
      if( rerRuleFlag )
      {
          val = rerRule.getChosenCounter();
          return val;
      }
      else
         return -1;
   }

   // get feedback after a chosen action is performed.
   //public static double determineFeedback( int val )
   public double determineFeedback( int val )
   {
        double fb;
	int diff = Math.abs( val - target );
	switch( diff )
        {
           case 0:
           case 1:
		fb = POS_FEEDBACK;
                break;
           default:
		fb = NEG_FEEDBACK * ( diff - 1 );
                break;
        }
        return fb;
   }
}
