// Rule reasoning part for the dynamic control task
package tasksimulators.oldsimu;
import java.lang.*;
import java.util.Vector;
import java.awt.*;
import clarion.common.*;

public class PCIrlRule
{
    private final int MAX_AGE  = 50;

    int generation, action, P1, target;
                  // action: the previous input. P1: the previous output.
    Vector rules;
    boolean debug = true;

    // parameters of rule set
    int[] a = { 1, 2 }; // a's b's and c's
    int[] b = { -2, -1, 0, 1, 2 }; // values to try in the rules
    int[] c = { -2,- 1, 1, 2 };

    // minimum number of times the rule was applicable
    // before its usefulness is evaluated
    int MinApp;

    // if evaluateUsefulness() returns a number
    // below this threshold, rule will be deleted
    double threshold; //gets reset in the sim

    StringBuffer allRules;
    StringBuffer matchRules;
    StringBuffer chosenRules;

    public PCIrlRule( int goal, int initAction, int initOutput )
    {
	generation = 0;
	MinApp = World.RULE_MIN_APP;

        target = goal;
        P1 = initOutput;
        action = initAction;

	rules = new Vector();
	generateRules( action );

        allRules = new StringBuffer();
        matchRules = new StringBuffer();
        chosenRules = new StringBuffer();

        double v1 =  World.TASK_PARA_ARRAY[ World.LOW_IRL_THRESHOLD ];
        setThreshold( v1 );
    }

    public void setThreshold( double th )
    {
	threshold = th;
    }

    public void reinit( int initAction, int initOutput )
    {
	generation = 0;
        P1 = initOutput;
        action = initAction;

        if( !rules.isEmpty() )
            rules.clear();
	generateRules( action );
        setThreshold( World.TASK_PARA_ARRAY[ World.LOW_IRL_THRESHOLD ] );
    }

   /* public void setParams(int p1, int p2, int w1, int t)
      {
	P1 = p1; // previous system output
        P2 = p2; // system output for t-2
        W1 = w1; // previous system input
	target = t; // target output value in this trial
      }  */

      // set previous action and output after updating rules.
      public void setParams( int oldAction, int oldOutput )
      {
          action = oldAction;
          P1 = oldOutput;
      }

    /* make sure actions have been set basically, by calling this
       AFTER getQvals() "action" is the action CLARION has chosen
       based on various recommendations "result" indicates is the
       result was positive (on target)
    */
    public boolean updateRule( int chosenAction, double result, int trialIdx )
    {
	Rule currentR;

        chosenRules.delete( 0, allRules.length() );

        int idx = 0;
	while ( idx < rules.size() )
        {
	    currentR = ( Rule ) rules.elementAt( idx );
            currentR.incAge();

            // set match counts for rules which recommended this action
	    if ( currentR.getCurAction() == chosenAction )
            {
                currentR.setAge( 0 );
                String str = convertToString( currentR );
                chosenRules.append( str );

		if( result > 0 )
                   currentR.incPM();
		else
                   currentR.incNM();

                double eval = currentR.evaluateUsefulness();
                if( ( currentR.timesApplied() >= MinApp ) && ( eval < threshold ) )
                      rules.removeElementAt( idx );
                else
                      idx++;
                continue;
            }

            // delete the useless rule
            int age = currentR.getAge();
            if( age == MAX_AGE )
            {
                rules.removeElementAt( idx );
                continue;
            }
            else
               idx++;

            // delete unsuccessful rules
        }

      	if ( rules.isEmpty() )
        {
           boolean retVal = generateRules( chosenAction );
           return retVal;
        }

        if( trialIdx == ((int)World.TASK_PARA_ARRAY[World.NET_OUTPUT_NUM]-1) )
            for( int i = 0; i < rules.size(); i++ )
            {
                 currentR = ( Rule ) rules.elementAt(i);
                 currentR.discountStatistics();
            }

        return true;
    }

    public void setDebug( boolean d )
    {
	debug = d;
    }

   //determine the desirabilities of all the actions by the current rule set.
    public void getRecommendedAction( double[] FillMe )
    {
	//zero out the FillMe array
	for( int j = 0; j < (int)World.TASK_PARA_ARRAY[World.NET_OUTPUT_NUM]; j++ )
             FillMe[j] = 0;

	int RelevantVariable = -1;

	switch( generation )
        {
	   case 1:
	         RelevantVariable = 0;
	         break;
	   case 2:
	         RelevantVariable = P1;
	         break;
	   case 3:
	         RelevantVariable = 0;
	         break;
	   case 4:
	         RelevantVariable = P1;
	         break;
        }

	Rule curR;
	for( int i = 0; i < rules.size(); i++ )
        {
	    curR = ( Rule ) rules.elementAt(i);
	    int action = curR.determineAction( RelevantVariable );
	    FillMe[ action ] = World.MAX_ACTIVATION_VAL;
        }
    }

    // generates new set of rules with the initAction as current
    // chosen action, invoked at the beginning of a generation
    private boolean generateRules( int initAction )
    {
	generation++;

	if ( generation > 4 )
        {
	    System.out.println( "No more rules to generate!" );
	    return false;
        }

        /*
	if ( ! rules.isEmpty() )
        {
	    System.out.println( "Rules vector not empty!" );
	    return false;
        }

	if ( debug )
             System.out.println("  Generation "+generation+". Generating rules.\n");
        */

        int ruleType = -1;
	int[] tempB = new int[5];
        int[] tempC;

        if( ( generation % 2 ) == 0 )
            tempC = new int[4];
        else
            tempC = new int[1];

	switch( generation )
        {
           case 1:
                ruleType = 0;
                tempB = b;
                tempC[0] = 0;
	        break;
           case 2:
                ruleType = 1;
	        tempB = b;
                tempC = c;
                break;
           case 3:
                ruleType = 2;
	        tempB = b;
                tempC[0] = 0;
        	 break;
           case 4:
                 ruleType = 3;
	         tempB = b;
                 tempC = c;
                 break;
        }

        // generate a new rule set.
	for( int i = 0; i < a.length; i++ )
          for( int j = 0; j < tempB.length; j++ )
             for( int k = 0; k < tempC.length; k++ )
                  rules.addElement( new Rule( a[i], tempB[j], tempC[k],
                                    ruleType, initAction, target ) );

       return true;
    }

    public String convertToString( Rule element )
    {
        int a = element.getA();
        int b = element.getB();
        int c = element.getC();

        String str, strA, strB, strC;
        strA = String.valueOf( a );

        if ( b > 0 )
             strB = " + " + String.valueOf( Math.abs( b ) );
        else if( b < 0 )
                 strB = " - " + String.valueOf( Math.abs( b ) );
             else
                 strB = "";

        if ( c >= 0 )
             strC = " + " + String.valueOf( Math.abs( c ) );
        else
             strC = " - " + String.valueOf( Math.abs( c ) );

        switch( element.getType() )
        {
           case 0:
                  str = strA + " * w" + strB + "\n";
                  break;
           case 1:
                  str = strA + " * w" + strC + " * p1" + strB + "\n";
                  break;
           case 2:
                  str = strA + " * w1" + strB + "\n";
                  break;
           case 3:
                  str = strA + " * w1" + strC + " * p2" + strB + "\n";
                  break;
           default:
                  str = "";
                  break;
       }

       return str;
    }

    public StringBuffer getAllRules()
    {
        allRules.delete( 0, allRules.length() );

	for( int i = 0; i < rules.size(); i++ )
        {
	    Rule currentR = ( Rule ) rules.elementAt(i);
            String str = convertToString( currentR );
            allRules.append( str );
       }

        return allRules;
    }

    public StringBuffer getMatchRules()
    {
        matchRules.delete( 0, matchRules.length() );
        matchRules.append( allRules.substring( 0 ) );
        return matchRules;
    }

    public StringBuffer getChosenRules()
    {
        return chosenRules;
    }

}

/* a , b, c are multipliers in the three-variable rules in this task
   rules are of the form P = a*W + b + c*V1
   Type 0:   P = a * W + b;   Type 1:   P = a * W + b + c * P1;
   Type 2:   P = a * W1 + b;  Type 3:   P = a * W1 + b + c * P2;
*/

class Rule {

    static double DISCOUNT = 0.9;
    int C5 = 0, C6 = 0;   // used for IG measure in IRL.
    int a, b, c, pm, nm, target;
    int curAction, nextAction;
    int age, type;                // 4 types of rules as paper Page36 says

    Rule( int first, int second, int third, int rType, int cAction, int  t )
    {
       	a = first;
        b = second;
        c = third;

        type = rType;
        target = t;

        curAction = cAction;
        nextAction = ( int ) ( target - b ) / a;

        pm = 0;              // positive and negative match counts
       	nm = 0;
        age = 0;
    }

    public int getA()
    {
      	return a;
    }

    public int getB()
    {
      	return b;
    }

    public int getC()
    {
	      return c;
    }

    public int getType()
    {
       return type;
    }


    public int getAge()
    {
       return age;
    }

    public int getNM()
    {
      	return nm;
    }

    public int getPM()
    {
      	return pm;
    }

    public void incAge()
    {
       age++;
    }

    public void setAge( int val )
    {
        age = val;
    }

    public void incPM()
    {
       	pm++;
    }

    public void incNM()
    {
       	nm++;
    }

    public void resetMatchCounts()
    {
       	pm = 0;
        nm = 0;
    }

    public int getCurAction()
    {
       	return curAction;
    }

    public boolean mayBeTHERule()
    {
        if( type != 1 )
            return false;

       	if( ( a == 2 ) && ( b == 0 ) && ( c == -1 ) )
              return true;
   	    else
              return false;
    }


    /* action ("W" from the paper) which this particular rule recommends
       at this step it's somewhat unclear what action selection strategy
       should be adopted for rules which do not depend on current user
       response */
   /* according to the paper, the rule set 1 & 2 refers to 'W'(current
      action which should be decided now and will be applied in this trial),
      the rule set 3 & 4 refers to 'W1'( the action which can be decided
      now and should be used in next trial ), in this case, the action
      applied in this trial has been decided in the last trial.  */
    public int determineAction( int v1 )
    {
       int chosenAction = Math.round( ( target - c * v1 - b ) / a );
       if ( chosenAction < 0 )
            chosenAction = 0;
       else if ( chosenAction > 11 )
                 chosenAction = 11;

       switch( type )
       {
           case 0:             // action should be returned for this step.
           case 1:
                curAction = chosenAction;
        	return curAction;
           case 2:             // action should be returned for next step.
           case 3:
                curAction = nextAction;
                nextAction = chosenAction;
        	return curAction;
       }

       return curAction;
    }

    // how many times the rule was applicable
    public int timesApplied()
    {
	return ( pm + nm );
    }

    public void discountStatistics()
    {
        pm *= DISCOUNT;
        nm *= DISCOUNT;
    }

    // is this rule any good?
    // cannot use IG(information gain) in this task
    public double evaluateUsefulness()
    {
	return( ( double )( pm + C5 ) / ( (double)( pm + nm + C6 ) ) );
    }
}

