/******************************************************************************/
/************************* PROCESSING CONTROL TASK  ***************************/
/******************************************************************************/
package tasksimulators.oldsimu;
import java.awt.*;
import java.awt.Graphics;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;
import tasksimulators.*;

public class PCTask extends JFrame implements SimuTask {

/***************************VARIABLES DECLARATION BEGIN************************/
   final int   INPUT_NUM = 12,        OUTPUT_NUM = 12,
               MIN_LEVEL = 1,         MAX_LEVEL = 12,
               MIN_BLOCK_NUM = 1,     MAX_BLOCK_NUM = 200,
               MAX_AGENT_NUM = 10,    TRIALS_PER_BLOCK = 11,
               INPUT = 0, OUTPUT = 1, ACTION = 2;

  // general constant variables for designing window in which results are shown.
   final int WINDOW_WIDTH = 850,       WINDOW_HEIGHT = 680,
             BOTTOM_PANEL_WIDTH = 850, BOTTOM_PANEL_HEIGHT = 300,
             DRAW_WIDTH = 380,         DRAW_HEIGHT = 250,
             Y_WIDTH = 30,             Y_HEIGHT = 380;

   final int STATUS_NUMBER = 5;
   final int STATUS_EQU_TYPE = 0, STATUS_EQU_PARA = 1;
   final int STATUS_TIMER = 2, STATUS_TIME_LIMIT = 3, STATUS_NOISE = 4;
   final int RULE_AREA_NUMBER = 3;
   final int ALL_RULE = 0, MATCH_RULE = 1, CHOSEN_RULE = 2;
   final int LINEAR = 0, U_SHAPE = 1, POWER = 2;

   private int trialIdx, blockIdx, agentIdx, blockNum, expMode;
   private int noiseArray[], dataArray[][][];
   private int MIN_INPUT, MAX_INPUT, MIN_OUTPUT, MAX_OUTPUT;
   private long curTime, oldTime;

   TaskManager taskManager;
   WriteFile   dataFile;
   Timer       inputTimer;
   ActionEventHandler timerHandler;
   List  yInputList;
   ItemEventHandler inputListHandler;

   // graphical components for user's graphical interface
   private Container   c;
   public  JFrame      mainFrame;
   private TextArea    statusArea[];
   private JTextField  fromField, toField;
   private JLabel      statusLabel, npLabel, promptLabel1;
   private JPanel      userInputPanel, statusPanel;
   private JPanel      inputPanel, buttonPanel1,buttonPanel2, listPanel;
   private JPanel      cPanel, dPanel, labelPanel, drawPanel, rPanel;
   private DrawPanel   dInputPanel, dOutputPanel;

  //graphical components for setting the program environment.
   private int ITEMS_PER_MENU[] = { 3, 3, 1 };
   private int EQU_NUM = 3, TIME_INTERVAL_NUM = 8;
   private int NOISE_RATIO_NUM = 5, TARGET_CHOICE_NUM = 5;
   private int RANGE_SELECT_CHOICE_NUM = 12;

   private int equIdx, parIdx;

   private JPopupMenu statusMenu, equMenu,parMenu[];
   private JPopupMenu timerMenu, timeMenu, noiseMenu;
   private JPopupMenu targetMenu, rangeMenu, rangeSelectMenu;
   private JMenuItem  equItem[], parItem[][], statusItem[];
   private JMenuItem  timerItem[], timeItem[], noiseItem[];
   private JMenuItem  targetItem[], rangeItem[], rangeSelectItem[];
   private JButton    equButton, parButton, statusButton, rangeButton, confirmButton;
   private JButton    timerButton, timeSetButton, noiseSetButton, targetButton;
   private JButton    beginButton, exitButton, fromButton, toButton;

   private JFrame       noiseFrame;
   private Container    noiseContainer;
   private JButton      noiseExitButton, noiseCancelButton;
   private JLabel       noiseLabelArray[];
   private JTextField   noiseFieldArray[];

   private boolean taskManagerFlag = false;

   String equationStr;
   String PROMPT =
      "PROMPT: select values as inputs in sequence for 10 times. an input will"+
      " be generated automatically after 2 seconds nor response from user";

   private boolean  STATUS_FLAG = false;
   private String statusChoices[] =
   {
       "ON",  "OFF"
   };

   private boolean TIMER_FLAG = false;
   private String timerChoices[] =
   {
       "ON",  "OFF"
   };

   private int TIME_INTERVAL = 6000;
   private int TIME_INTERVAL_ARRAY[] =
   {
       3000, 4000, 5000, 6000, 7000, 8000, 9000, 10000
   };
   private String timeIntervals[] =
   {
      " 3000 ", " 4000 ", " 5000 ", " 6000 ",
      " 7000 ", " 8000 ", " 9000 ", " 10000"
   };

   private double NOISE_TIMES;
   private double NOISE_RATIO;
 //  private double NOISE_RATIO_ARRAY[];

   private String NOISE_LABEL_ARRAY[] =
   {
      "RATIO OF UP TWO LEVEL",    "RATIO OF UP ONE LEVEL",
      "RATIO OF NO NOISE",    "RATIO OF DOWN ONE LEVEL",
      "RATIO OF DOWN TWO LEVEL"
   };

   private double NOISE_RATIO_ARRAY[] =
   {
      0.1, 0.2, 0.4, 0.2, 0.1
   };

   /*
   private double NOISE_RATIO_ARRAY[] =
   {
      0.0, 0.1, 0.2, 0.3, 0.33, 0.4, 0.5
   }; */

   private String noiseRatios[] =
   {
      " 0.0 ", " 0.1 ", " 0.2 ", " 0.3 ", "0.33", " 0.4 ", " 0.5 "
   };

   private int TARGET_LEVEL;
   private int TARGET_CHOICE_ARRAY[] =
   {
       3, 4, 5, 6, 7
   };
   private String targetChoices[] =
   {
      "LEVEL 4", "LEVEL 5", "LEVEL 6", "LEVEL 7", "LEVEL 8"
   };

   private boolean RANGE_FLAG;
   private String rangeChoices[] =
   {
      "ON", "OFF"
   };

   private boolean FROM_FLAG;
   private String rangeSelectChoices[] =
   {
      "100", "200", "300", "400", "500", "600",
      "700", "800", "900", "1000", "1100", "1200"
   };

   // input data in the list for user to select.
   private String INPUT_NUMBERS[] =
   {
       "1200", "1100", "1000", "900", "800",  "700",
        "600", "500",  "400",  "300",  "200",  "100"
   };

  // level of output to be shown.
   private String OUTPUT_NUMBERS[] =
    {
        "12", "11", "10", " 9", " 8", " 7",
        " 6", " 5", " 4", " 3", " 2", " 1"
    };

  // equation types in the equation selecting menu.
   private String equTypes[] =
   {
       "Linear",  "U-shaped",  "Power"
   };

   // parameter names in the parameter selecting menus.
   private String parNames[][] =
    {
       {
           "a = 20, b = 0", "a = 10, b = 6000",  "a = 30, b = -6000"
       },
       {
           "a = 300, b = 0.01", "a = 600, b = 0.01",  "a = 900, b = 0.01"
       },
       {
           "a=2, b=1, c=9.0, d=500"
       },
   };

  // true value of the parameters of the selected equation.
  private double paramVal[][][] =
   {
      { {20, 0},    {10, 6000}, {30, -6000}  },
      { {300, 0.01}, {600, 0.01}, {900, 0.01} },
      { {2, 1, 9.0, 500} },
   };

   // the unit value corresponding to each type of equations
   private int  iUnitVal[] =
   {
       100, 100, 100
   };

   private int  oUnitVal[] =
   {
        1000, 200, 200
   };

/**********************METHODS DEFINATION BEGIN HERE **************************/
   // Constructor method
   public PCTask( JFrame mFrame )
   {
      mainFrame = mFrame;
      // setup user interface.
      c = getContentPane();
      c.removeAll();
      c.setLayout( new BorderLayout( 5, 0 ) );

      buildInputSelection();
      buildDrawArea();
      buildBottomArea();

      setSize( WINDOW_WIDTH, WINDOW_HEIGHT );
      setEnabled( false );
      setVisible( false );

      // timer initialization
      timerHandler = new ActionEventHandler();
      inputTimer = new Timer( TIME_INTERVAL, timerHandler );

      // noise ratio intialization.
      NOISE_RATIO = 0.33;

      // target level.
      TARGET_LEVEL = World.OUTPUT_TARGET_LEVEL;

      dataArray = new int[ MAX_BLOCK_NUM ][ TRIALS_PER_BLOCK ][ 3 ];
      noiseArray = new int[ TRIALS_PER_BLOCK ];

      // add noise window
       noiseFrame = new JFrame( "NOISE SELECTION" );
       noiseFrame.setSize( 400, 200 );
       noiseContainer = noiseFrame.getContentPane();
       noiseContainer.setLayout( new GridLayout( NOISE_RATIO_NUM + 1, 2, 0, 0 ) );

       noiseLabelArray = new JLabel[ NOISE_RATIO_NUM ];
       noiseFieldArray = new JTextField[ NOISE_RATIO_NUM ];
       for( int i = 0; i < NOISE_RATIO_NUM; i++ )
       {
          noiseLabelArray[ i ] = new JLabel( NOISE_LABEL_ARRAY[i] );
          noiseFieldArray[ i ] = new JTextField( String.valueOf(NOISE_RATIO_ARRAY[i]) );
          noiseContainer.add( noiseLabelArray[ i ] );
          noiseContainer.add( noiseFieldArray[ i ] );
       }

       noiseExitButton = new JButton("RETURN");
       noiseExitButton.setEnabled( true );
       noiseContainer.add( noiseExitButton );
       noiseExitButton.addActionListener( new ActionListener()
      {
         public void actionPerformed( ActionEvent e)
         {
            double sum = 0.0;
            boolean numFlag = true;
            for( int i = 0; i < NOISE_RATIO_NUM; i++ )
            {
               try {
                   double v1 = Double.parseDouble( noiseFieldArray[i].getText());
                   sum += v1;
               }
               catch( NumberFormatException nfe ){
                       showMessage(0);
                   numFlag = false;
                   break;
               }
            }

            if( numFlag )
            {
               if ( Math.abs( sum - 1.0 ) > 0.001 )
                    showMessage(1);
               else
               {
                   for( int i = 0; i < NOISE_RATIO_NUM; i++ )
                        NOISE_RATIO_ARRAY[ i ] = Double.parseDouble(
                                           noiseFieldArray[i].getText() );
                   noiseFrame.setEnabled(false);
                   noiseFrame.setVisible(false);
               }
            }
         }
      }  );

       noiseCancelButton = new JButton("CANCEL");
       noiseCancelButton.setEnabled( true );
       noiseContainer.add( noiseCancelButton );
       noiseCancelButton.addActionListener( new ActionListener()
       {
         public void actionPerformed( ActionEvent e)
         {
              noiseFrame.setEnabled(false);
              noiseFrame.setVisible(false);
         }
       } );

       addWindowListener(
          new WindowAdapter(){
             public void windowClosing( WindowEvent e )
             {
                saveData();
                finishTheTask( false );

                setVisible( false );
                setEnabled( false );
                taskManagerFlag = false;
                mainFrame.setVisible( true );
                mainFrame.setEnabled( true );
             }
        }  );
   }

   public void showMessage( int mode )
   {
    if(mode == 0)
       JOptionPane.showMessageDialog( noiseFrame, "Not an valid number!",
                       "WARNING",   JOptionPane.WARNING_MESSAGE );
    else
       JOptionPane.showMessageDialog( noiseFrame, "Sum should be 1.0!",
                       "WARNING", JOptionPane.WARNING_MESSAGE );
   }

   public void setExpMode( int mode )
   {
      expMode = mode;
   }

   public void setBlockNum( int num )
   {
      blockNum = num;
   }

   public String getTaskName()
   {
      return "PROCESS CONTROL TASK";
   }

   public void start( int mode, int bNum, WriteFile dFile )
   {
      expMode = mode;
      blockNum = bNum;
      dataFile = dFile;
      if( expMode == World.CLARION_MODE )
      {
         if( !taskManagerFlag )
         {
              taskManager = new TaskManager( blockNum, dataFile );
              taskManagerFlag = true;
         }
      }
      blockIdx = 0;
      trialIdx = 0;
      setEnabled( true );
      setVisible( true );
   }

/***************************************************************************/
/****************BUILD WINDOW ENVIRONMENT IN WHICH PROGRAM RUNS************/
/**************************************************************************/
   public void buildInputSelection()
   {
      inputPanel = new JPanel();
      inputPanel.setSize( Y_WIDTH, Y_HEIGHT );
      inputPanel.setLayout( new BorderLayout( 10, 5 ) );

      // for status, equation, parameter, timer, time, noise, target, range buttons.
      buttonPanel1 = new JPanel();
      buttonPanel1.setLayout( new GridLayout( 8, 1, 20, 5 ) );
      addStatusButton();
      addEquButton();
      addTimerButton();
      addTimeButton();
      addNoiseButton();
      addTargetButton();
      addRangeButton();

      // for begin and exit buttons.
      buttonPanel2 = new JPanel();
      buttonPanel2.setLayout( new GridLayout( 3, 1, 20, 5 ) );
      addBeginButton();
      addExitButton();

      // for input list.
      addInputList();

      inputPanel.add( buttonPanel1, BorderLayout.NORTH );
      inputPanel.add( buttonPanel2, BorderLayout.SOUTH );
      inputPanel.add( listPanel, BorderLayout.CENTER );
      c.add( inputPanel, BorderLayout.WEST );
   }

   public void buildDrawArea()
   {
      cPanel = new JPanel();
      //cPanel.setLayout( new BorderLayout( 10, 10 ) );
      cPanel.setLayout( new GridLayout( 2, 1, 5, 5 ) );

      addDrawArea();
      addRangeChoice();

      //cPanel.add( dPanel, BorderLayout.CENTER );
      //cPanel.add( rPanel, BorderLayout.SOUTH );
      cPanel.add( dPanel );
      cPanel.add( rPanel );
      c.add( cPanel, BorderLayout.CENTER );
   }

   public void buildBottomArea()
   {
      statusPanel = new JPanel( new GridLayout( 1, STATUS_NUMBER, 0, 0 ) );
      statusArea = new TextArea[ STATUS_NUMBER ];

      for( int i = 0; i < STATUS_NUMBER; i++ )
      {
           statusArea[ i ] = new TextArea( "", 2, 25 );
           statusPanel.add( statusArea[ i ] );
           statusArea[ i ].setEnabled( false );
      }

      c.add( statusPanel, BorderLayout.SOUTH );
   }

  /***********************************************************************/
  /*************************ADD COMPONENT METHODS ************************/
  /***********************************************************************/
   private void addStatusButton()
   {
      statusButton = new JButton("SHOW SETTING");
      statusButton.setEnabled( true );
      statusButton.addActionListener( new ActionListener()
      {
         public void actionPerformed( ActionEvent e)
         {
            statusMenu.show( inputPanel, 40, 10 );
         }
      }  );
      buttonPanel1.add( statusButton );

      statusMenu = new JPopupMenu("STATUS");
      statusItem = new JMenuItem[ 2 ];
      for( int i = 0; i < 2; i++ )
      {
         statusItem[ i ] = new JMenuItem( statusChoices[ i ] );
         statusItem[i].setFont( new Font( "Serif", Font.PLAIN, 14 ) );
         statusItem[ i ].addActionListener( new ActionListener()
         {
            public void actionPerformed( ActionEvent e)
            {
               for( int i = 0; i < 2; i++ )
                  if ( e.getSource() == statusItem[ i ] )
                  {
                     if( i == 0 )
                        STATUS_FLAG = true;
                     else
                        STATUS_FLAG = false;
                     break;
                  }
               showSettings( STATUS_FLAG );
            }
          }    );
         statusMenu.add( statusItem[ i ] );
      }
      inputPanel.add( statusMenu );
   }

   private void addEquButton()
   {
      equButton = new JButton("EQUATIONS");
      equButton.setEnabled( true );

      equButton.addActionListener( new ActionListener()
      {
        public void actionPerformed( ActionEvent e )
        {
          equMenu.show(inputPanel, 20, 20 );
        }
      } );

      parButton = new JButton("PARAMETERS");
      parButton.setEnabled( false );
      parButton.addActionListener( new ActionListener()
      {
        public void actionPerformed( ActionEvent e )
        {
           parMenu[equIdx].show( inputPanel, 10, 40 );
        }
      } );

      buttonPanel1.add( equButton );
      buttonPanel1.add( parButton );

      equMenu = new JPopupMenu("Equation");
      parMenu = new JPopupMenu[ EQU_NUM ];
      parMenu[0] = new JPopupMenu("Param0");
      parMenu[1] = new JPopupMenu("Param1");
      parMenu[2] = new JPopupMenu("Param2");
      equItem = new JMenuItem[ EQU_NUM ];
      parItem = new JMenuItem[ EQU_NUM ][ EQU_NUM ];

      for( int i = 0; i < EQU_NUM; i++ )
          parItem[i] = new JMenuItem[ ITEMS_PER_MENU[i] ];

      for( int i = 0; i < equTypes.length; i++ )
      {
         equItem[ i ] = new JMenuItem( equTypes[ i ] );
         equItem[i].setFont( new Font( "Serif", Font.PLAIN, 14 ) );

         equItem[ i ].addActionListener( new ActionListener()
         {
            public void actionPerformed( ActionEvent e)
            {
               parButton.setEnabled(true);
               for( int i = 0; i < ITEMS_PER_MENU[0]; i++ )
                  if ( e.getSource() ==  equItem[ i ] )
                  {
                       equIdx = i;
                       break;
                  }

               beginButton.setEnabled( false );
               showEquType( STATUS_FLAG );
            }
          }    );
         equMenu.add( equItem[ i ] );
      }
      inputPanel.add( equMenu );

      for ( int i = 0; i < parNames.length; i++ )
      {
          for ( int j = 0; j < ITEMS_PER_MENU[i]; j++)
          {
              parItem[i][j] = new JMenuItem( parNames[i][j] );
              parMenu[i].add( parItem[i][j] );
              parItem[i][j].setFont( new Font( "Serif", Font.PLAIN, 14 ) );
              parItem[i][j].addActionListener( new ActionListener()
              {
                 public void actionPerformed( ActionEvent e )
                 {
                    for( int i = 0; i < ITEMS_PER_MENU[equIdx]; i++ )
                       if ( e.getSource() ==  parItem[equIdx][i] )
                       {
                          parIdx = i;
                          break;
                       }

                    showEquPara( STATUS_FLAG );
                    setEnviron( equIdx, parIdx );
                 }
               }   );
          }
          inputPanel.add( parMenu[i] );
       }
   }

   private void addTimerButton()
   {
      timerButton = new JButton( "USE TIMER" );
      timerButton.setEnabled( true );
      timerButton.addActionListener( new ActionListener()
      {
          public void actionPerformed( ActionEvent e)
         {
            timerMenu.show( inputPanel, 40, 60 );
         }
      }  );
      buttonPanel1.add( timerButton );

      timerMenu = new JPopupMenu("TIMER");
      timerItem = new JMenuItem[ 2 ];
      for( int i = 0; i < 2; i++ )
      {
         timerItem[ i ] = new JMenuItem( timerChoices[ i ] );
         timerItem[i].setFont( new Font( "Serif", Font.PLAIN, 14 ) );
         timerItem[ i ].addActionListener( new ActionListener()
         {
            public void actionPerformed( ActionEvent e)
            {
               for( int i = 0; i < 2; i++ )
                  if ( e.getSource() == timerItem[ i ] )
                  {
                     if( i == 0 )
                     {
                        TIMER_FLAG = true;
                        timeSetButton.setEnabled(true);
                     }
                     else
                     {
                        TIMER_FLAG = false;
                        timeSetButton.setEnabled(false);
                     }
                     break;
                  }
               showTimer( STATUS_FLAG );
            }
          }    );
         timerMenu.add( timerItem[ i ] );
      }

      inputPanel.add( timerMenu );
   }

   private void addTimeButton()
   {
      timeSetButton = new JButton( " TIME SETTING " );
      timeSetButton.setEnabled( false );
      timeSetButton.addActionListener( new ActionListener()
      {
        public void actionPerformed( ActionEvent e )
        {
            timeMenu.show( inputPanel, 20, 60 );
        }
      } );
      buttonPanel1.add( timeSetButton );

      timeMenu = new JPopupMenu("TIME");
      timeItem = new JMenuItem[ TIME_INTERVAL_NUM ];
      for( int i = 0; i < TIME_INTERVAL_NUM; i++ )
      {
         timeItem[ i ] = new JMenuItem( timeIntervals[ i ] );
         timeItem[i].setFont( new Font( "Serif", Font.PLAIN, 14 ) );

         timeItem[ i ].addActionListener( new ActionListener()
         {
            public void actionPerformed( ActionEvent e)
            {
               for( int i = 0; i < TIME_INTERVAL_NUM; i++ )
                  if ( e.getSource() ==  timeItem[ i ] )
                  {
                       TIME_INTERVAL = TIME_INTERVAL_ARRAY[  i ];
                       inputTimer.setDelay( TIME_INTERVAL );
                       inputTimer.setInitialDelay( TIME_INTERVAL );
                       break;
                  }
                showTimeLimit( STATUS_FLAG );
            }
          }    );
         timeMenu.add( timeItem[ i ] );
      }

      inputPanel.add( timeMenu );
   }

   private void addNoiseButton()
   {
      noiseSetButton = new JButton( " NOISE SETTING " );
      noiseSetButton.setEnabled( true );
      noiseSetButton.addActionListener( new ActionListener()
      {
        public void actionPerformed( ActionEvent e )
        {
           // noiseMenu.show( inputPanel, 0, 20 );
           noiseFrame.setLocation(160, 180);
           noiseFrame.setEnabled(true);
           noiseFrame.setVisible( true );
        }
      } );
      buttonPanel1.add( noiseSetButton );

      noiseMenu = new JPopupMenu("NOISE");
      noiseItem = new JMenuItem[ NOISE_RATIO_NUM ];
      for( int i = 0; i < NOISE_RATIO_NUM; i++ )
      {
         noiseItem[ i ] = new JMenuItem( noiseRatios[ i ] );
         noiseItem[i].setFont( new Font( "Serif", Font.PLAIN, 14 ) );
         noiseItem[ i ].addActionListener( new ActionListener()
         {
            public void actionPerformed( ActionEvent e)
            {
               for( int i = 0; i < NOISE_RATIO_NUM; i++ )
                  if ( e.getSource() ==  noiseItem[ i ] )
                  {
                      // NOISE_TIMES = i;
                       NOISE_RATIO = NOISE_RATIO_ARRAY[ i ];
                       break;
                  }

               //showNoiseRatio( STATUS_FLAG );
            }
          }    );

         noiseMenu.add( noiseItem[ i ] );
      }
      inputPanel.add( noiseMenu );
   }

   private void addTargetButton()
   {
      targetButton = new JButton( " TARGET SETTING " );
      targetButton.setEnabled( true );
      targetButton.addActionListener( new ActionListener()
      {
        public void actionPerformed( ActionEvent e )
        {
            targetMenu.show( inputPanel, 10, 100 );
        }
      } );
      buttonPanel1.add( targetButton );

      targetMenu = new JPopupMenu("TARGET");
      targetItem = new JMenuItem[ TARGET_CHOICE_NUM ];
      for( int i = 0; i < TARGET_CHOICE_NUM; i++ )
      {
         targetItem[ i ] = new JMenuItem( targetChoices[ i ] );
         targetItem[i].setFont( new Font( "Serif", Font.PLAIN, 14 ) );
         targetItem[ i ].addActionListener( new ActionListener()
         {
            public void actionPerformed( ActionEvent e)
            {
               for( int i = 0; i < TARGET_CHOICE_NUM; i++ )
                  if ( e.getSource() ==  targetItem[ i ] )
                  {
                       TARGET_LEVEL = TARGET_CHOICE_ARRAY[ i ];
                       break;
                  }

               showTarget();
            }
          }    );

         targetMenu.add( targetItem[ i ] );
      }
      inputPanel.add( targetMenu );
   }

   private void addRangeButton()
   {
      rangeButton = new JButton( " RANGE CHOICE " );
      rangeButton.setEnabled( true );
      rangeButton.addActionListener( new ActionListener()
      {
          public void actionPerformed( ActionEvent e )
          {
             rangeMenu.show( inputPanel, 10, 200 );
          }
      } );
      buttonPanel1.add( rangeButton );

      rangeMenu = new JPopupMenu("RANGE");
      rangeItem = new JMenuItem[ 2 ];
      for( int i = 0; i < 2; i++ )
      {
         rangeItem[ i ] = new JMenuItem( rangeChoices[ i ] );
         rangeItem[i].setFont( new Font( "Serif", Font.PLAIN, 14 ) );
         rangeItem[ i ].addActionListener( new ActionListener()
         {
            public void actionPerformed( ActionEvent e)
            {
               for( int i = 0; i < 2; i++ )
               {
                    if ( e.getSource() == rangeItem[ i ] )
                    {
                         if( i == 0 )
                         {
                             if( !RANGE_FLAG )
                             {
                                 RANGE_FLAG = true;
                                 rPanel.setVisible( true );
                                 rPanel.setEnabled( true );
                             }
                         }
                         else
                         {
                             if( RANGE_FLAG )
                             {
                                 RANGE_FLAG = false;
                                 rPanel.setVisible( false );
                                 rPanel.setEnabled( false );
                             }
                         }
                         break;
                     }
                }
            }
          }    );
         rangeMenu.add( rangeItem[ i ] );
      }
      inputPanel.add( rangeMenu );
   }

   private void addBeginButton()
   {
      beginButton = new JButton( "BEGIN" );
      beginButton.setEnabled( false );
      beginButton.addActionListener(
         new  ActionListener() {
                 public void actionPerformed( ActionEvent e )
                 {
                     yInputList.setEnabled( true );
                     statusButton.setEnabled( false );
                     equButton.setEnabled( false );
                     parButton.setEnabled( false );
                     timerButton.setEnabled( false );
                     timeSetButton.setEnabled( false );
                     noiseSetButton.setEnabled( false );
                     targetButton.setEnabled( false );
                     rangeButton.setEnabled( false );
                     fromButton.setEnabled( true );
                     toButton.setEnabled( true );
                     confirmButton.setEnabled( true );

                     beginButton.setEnabled( false );
                     exitButton.setEnabled( false );

                     if( expMode == World.CLARION_MODE ){
                         taskManager.trainMain();
                         finishTheTask( true );
                     }
                     else
                         startTraining();
                 }
        }
      );

      buttonPanel2.add( beginButton );
   }

   private void addExitButton()
   {
      exitButton = new JButton( "Exit" );
      exitButton.addActionListener(
                 new  ActionListener() {
                     public void actionPerformed( ActionEvent e )
                     {
                        saveData();
                        //dataFile.closeFile();   // modifed on 08/18/02
                        setVisible( false );
                        setEnabled( false );
                        taskManagerFlag = false;
                        mainFrame.setVisible( true );
                        mainFrame.setEnabled( true );
                     }
                 }
      );
      buttonPanel2.add( exitButton );
   }

   private void addInputList()
   {
      listPanel = new JPanel();
      listPanel.setLayout( new BorderLayout( 8, 8 ) );

      npLabel = new JLabel( "     Next Input" );
      npLabel.setFont( new Font( "Serif", Font.BOLD, 18 ) );
      listPanel.add( npLabel, BorderLayout.NORTH );

      yInputList = new List(INPUT_NUM, false);
      for( int i = 0; i < INPUT_NUMBERS.length; i++ )
           yInputList.add( INPUT_NUMBERS[ i ] );

      yInputList.setLocation( 0, 0 );
      yInputList.makeVisible( INPUT_NUM  );
      yInputList.setFont( new Font( "Serif", Font.BOLD, 16 ) );

      inputListHandler = new ItemEventHandler();
      yInputList.addItemListener( inputListHandler );
      yInputList.setEnabled(false);

      listPanel.add( yInputList, BorderLayout.CENTER );
    }

    public void addDrawArea()
    {
      dPanel = new JPanel();
      dPanel.setLayout( new BorderLayout() );

      labelPanel = new JPanel();
      labelPanel.setLayout( new FlowLayout(FlowLayout.CENTER, 100, 10) );
      promptLabel1 = new JLabel( PROMPT );
      promptLabel1.setFont( new Font( "Serif", Font.PLAIN, 12 ) );
      labelPanel.add( promptLabel1 );

      dPanel.add( labelPanel, BorderLayout.NORTH );

      drawPanel = new JPanel();
      drawPanel.setSize( DRAW_WIDTH * 2, DRAW_HEIGHT );
      drawPanel.setLayout( new GridLayout() );
      dInputPanel = new DrawPanel( DRAW_WIDTH, DRAW_HEIGHT, 0, 0 );
      drawPanel.add( dInputPanel );
      dOutputPanel = new DrawPanel( DRAW_WIDTH, DRAW_HEIGHT,
                                    World.OUTPUT_TARGET_LEVEL, 1 );
      drawPanel.add( dOutputPanel );

      dPanel.add( drawPanel, BorderLayout.CENTER );
    }

   private void addRangeChoice( )
   {
      rPanel = new JPanel ( new FlowLayout( FlowLayout.CENTER, 5, 5 ) );

      JLabel rangeLabel = new JLabel( "Select range( 100 to 1200 ) : " );
      rangeLabel.setFont( new Font( "Serif", Font.BOLD, 18 ) );

      ActionListener listen = new ActionListener()
      {
          public void actionPerformed(ActionEvent e)
          {
             checkRangeInput();
          }
      };

      fromButton = new JButton( " FROM " );
      fromButton.setEnabled( false );
      fromField = new JTextField( "500", 3 );
      fromField.addActionListener( listen );
      fromField.setEnabled( false );
      fromButton.addActionListener( new ActionListener()
      {
         public void actionPerformed( ActionEvent e )
         {
             FROM_FLAG = true;
             rangeSelectMenu.show( rPanel, 310, 30 );
         }
      } );

      toButton = new JButton( " TO " );
      toButton.setEnabled( false );
      toField = new JTextField( "700", 3 );
      toField.addActionListener( listen );
      toField.setEnabled( false );
      toButton.addActionListener( new ActionListener()
      {
         public void actionPerformed( ActionEvent e )
         {
             FROM_FLAG = false;
             rangeSelectMenu.show( rPanel, 420, 30 );
         }
      } );

      rangeSelectMenu = new JPopupMenu("RANGE_SELCT");
      //rangeSelectMenu.setPopupSize( 50, 100 );
      rangeSelectItem = new JMenuItem[ RANGE_SELECT_CHOICE_NUM ];
      for( int i = 0; i < RANGE_SELECT_CHOICE_NUM; i++ )
      {
         rangeSelectItem[ i ] = new JMenuItem( rangeSelectChoices[ i ] );
         rangeSelectItem[i].setFont( new Font( "Serif", Font.PLAIN, 12 ) );
         rangeSelectItem[ i ].addActionListener( new ActionListener()
         {
            public void actionPerformed( ActionEvent e)
            {
               for( int i = 0; i < RANGE_SELECT_CHOICE_NUM; i++ )
                  if ( e.getSource() == rangeSelectItem[ i ] )
                  {
                       if( FROM_FLAG )
                           fromField.setText( rangeSelectChoices[ i ] );
                       else
                           toField.setText( rangeSelectChoices[ i ] );
                       break;
                  }
            }
          }    );

         rangeSelectMenu.add( rangeSelectItem[ i ] );
      }
      rPanel.add( rangeSelectMenu );

      confirmButton = new JButton( "CONFIRM" );
      confirmButton.setEnabled( false );
      confirmButton.addActionListener( new ActionListener()
        {
           public void actionPerformed( ActionEvent e )
           {
              if( TIMER_FLAG )
                  stopInputTimer();

              checkRangeInput();

              // record the user's response time.
              curTime = System.currentTimeMillis();
              dataArray[ blockIdx][ trialIdx ][ ACTION ] = ( int )( curTime - oldTime );

              int from = Integer.parseInt( fromField.getText() );
              int to = Integer.parseInt( toField.getText() );
              int increment = (int)( (int)Math.round( Math.random()*(to-from)/iUnitVal[equIdx] )
                                     * iUnitVal[equIdx] );
              int value = from + increment;

              dataArray[ blockIdx ][ trialIdx ][ INPUT ] = value;
              dataArray[ blockIdx][ trialIdx ][ OUTPUT ] = calcOutput( value );
              dInputPanel.showValue( value );
              dOutputPanel.showValue( dataArray[ blockIdx][ trialIdx ][ OUTPUT ]);
              gotoNextTrial();

              curTime = System.currentTimeMillis();
              oldTime = curTime;

              if( TIMER_FLAG )
                  startInputTimer();
           }
        } );

      rPanel.add( rangeLabel );
      rPanel.add( fromButton );
      rPanel.add( fromField );
      rPanel.add( toButton );
      rPanel.add( toField );
      rPanel.add( confirmButton );

      rPanel.setEnabled( false );
      rPanel.setVisible( false );
   }

/******************************************************************************/
/************************** Action Event Listener *****************************/
/******************************************************************************/
/************************ Action Event Handler for timer.**********************/
   private class ActionEventHandler implements ActionListener
   {
       public void actionPerformed( ActionEvent e )
      {
         if( trialIdx == 0 )
             return;

         if ( trialIdx < TRIALS_PER_BLOCK )
         {
             yInputList.setEnabled(false);

             // record the user's response time.
             curTime = System.currentTimeMillis();
             dataArray[ blockIdx][ trialIdx ][ ACTION ] = ( int )( curTime - oldTime );

             // modified on Jan 28, 2002.
             //int value = World.INPUT_MEAN_LEVEL * iUnitVal[ equIdx ];
             int value = iUnitVal[ equIdx ] / 2; //( version of 01/28/2002 )

             dataArray[ blockIdx ][ trialIdx ][ INPUT ] = value;
             dataArray[ blockIdx][ trialIdx ][ OUTPUT ] = calcOutput( value );
             dInputPanel.showValue( value );
             dOutputPanel.showValue( dataArray[ blockIdx][ trialIdx ][ OUTPUT ]);
             gotoNextTrial();
             yInputList.setEnabled( true );

             curTime = System.currentTimeMillis();
             oldTime = curTime;
         }
       }
   }

 /*************** INPUT LIST LISTENER HANDLER IMPLEMENTATION ***************/
   private class ItemEventHandler implements  ItemListener {
      public void itemStateChanged( ItemEvent e )
      {
         if( trialIdx == 0 )
             return;

         if ( trialIdx < TRIALS_PER_BLOCK )
         {
            if( TIMER_FLAG )
                stopInputTimer();

            //record the user's response time.
            curTime = System.currentTimeMillis();
            dataArray[ blockIdx][ trialIdx ][ ACTION ] = ( int )( curTime - oldTime );

            int value = Integer.parseInt( ( String ) INPUT_NUMBERS[ yInputList.
                                            getSelectedIndex() ] );
            dataArray[ blockIdx ][ trialIdx ][ INPUT ] = value;
            dataArray[ blockIdx][ trialIdx ][ OUTPUT ] = calcOutput( value );
            dInputPanel.showValue( value );
            dOutputPanel.showValue( dataArray[blockIdx][trialIdx][ OUTPUT ] );
            gotoNextTrial();

            if( TIMER_FLAG )
                startInputTimer();

            curTime = System.currentTimeMillis();
            oldTime = curTime;
         }
      }
   }


/******************************************************************************/
/***************************** OTHER SUB-FUNCTIONS ****************************/
/******************************************************************************/

/**************** ACTIVATED BY BEGIN BUTTON ***********************************/
   public void startTraining()
   {
       checkRangeInput();

       blockIdx = 0;
       trialIdx = 0;
       genFirstOutput();
   }

   public void genFirstOutput( )
   {
      dInputPanel.resetDraw();
      dOutputPanel.resetDraw();

      int value = ( int ) ( Math.random() * ( MAX_LEVEL - MIN_LEVEL ) );
      value += MIN_LEVEL;
      value *= iUnitVal[ equIdx ];
      dataArray[ blockIdx ][ trialIdx ][ INPUT ] = value;
      dInputPanel.showValue( value );

      value = ( int ) ( Math.random() * ( MAX_LEVEL - MIN_LEVEL ) );
      value += MIN_LEVEL;
      value *= oUnitVal[ equIdx ];
      dataArray[ blockIdx ][ trialIdx ][ OUTPUT ] = value;
      dOutputPanel.showValue( value );
      trialIdx++;

      if( TIMER_FLAG )
          startInputTimer();

      curTime = System.currentTimeMillis();
      oldTime = curTime;
  }

  public void genNoiseSequence()
  {
     int range = TRIALS_PER_BLOCK - 1;
     boolean[] existArray = new boolean[ range ];
     for( int i = 0; i <= range; i++ )
     {
          existArray[ i ] = false;
          noiseArray[ i ] = 0;
     }

     int counter = 0;
     while( counter < 2 * NOISE_TIMES )
     {
          int value = ( int ) Math.round( ( range * Math.random() ) );
          if( value < 1 )
              continue;
          if( existArray[ value ] )
              continue;
          else
          {
              existArray[ value ] = true;
              if( ( counter % 2 ) == 0 )
                  noiseArray[ value ] = oUnitVal[ equIdx ];
              else
                  noiseArray[ value ] = -oUnitVal[ equIdx ];
              counter++;
          }
     }
  }

/************************** METHODS ACTIVATED BY LISNTERS **********************/

/****************************** TIMER METHODS  *****************************/
   private void startInputTimer()
   {
       if( inputTimer.isRunning() )
           inputTimer.restart();
       else
           inputTimer.start();
   }

   private void stopInputTimer()
   {
       inputTimer.stop();
   }

  /* CALCULATE THE OUTPUT ACCORDING TO DIFFERENT EQUATIONS.*/
   private int calcOutput( int inputVal )
   {
      int w, n, iRand;
      double a, b, c,d, p0, p1;

      if ( trialIdx == 0 )
      {
          iRand  = ( int ) ( Math.random() * ( MAX_LEVEL - MIN_LEVEL ) );
          iRand += MIN_LEVEL;

          iRand *= oUnitVal[ equIdx ];
          return iRand;
      }

      w  = inputVal;
      p0 = dataArray[ blockIdx][ trialIdx  - 1 ][ OUTPUT ];
      //n =  noiseArray[ trialIdx ];
      n = genNoise();

      switch( equIdx )
      {
        case LINEAR:
                a = paramVal[equIdx][parIdx][0];
                b = paramVal[equIdx][parIdx][1];
                p1 = ( int )( a * w + b - p0 + n);
                break;
        case U_SHAPE:
                a = paramVal[equIdx][parIdx][0];
                b = paramVal[equIdx][parIdx][1];
                p1 = ( int )( b * Math.pow( (w - a), 2 ) - p0 + n );
                break;
        case POWER:
                a = paramVal[equIdx][parIdx][0];
                b = paramVal[equIdx][parIdx][1];
                //p1 = (int)(b * Math.pow(a, (double)(w/100)) - p0 + n);
                /*
                p1 = (int)(b*Math.pow(a,9.0+((double)w/500.0))-p0+n);
                double v = 500.0*(Math.log((double)1200+p1)/Math.log(2.0)-9.0);
                System.out.println( v );
                */
                c = paramVal[equIdx][parIdx][2];
                d = paramVal[equIdx][parIdx][3];
                p1 = (int)(b * Math.pow(a, (double)(c + w/d)) - p0 + n);
                break;
        default:
                a = paramVal[1][1][0];
                b = paramVal[1][1][1];
                p1 = ( int ) ( a * w + b - p0 + n );
                break;
       }

       if ( p1 > MAX_OUTPUT )
            p1 = MAX_OUTPUT;
       else if ( p1 < MIN_OUTPUT )
                 p1 = MIN_OUTPUT;

       return (int) p1;
   }

  private int genNoise()
  {
     int retVal = 0, MEAN_NUM = NOISE_RATIO_NUM / 2;
     double value = Math.random();
     double sum = 0.0;

     for( int i = 0; i < NOISE_RATIO_NUM; i++ )
     {
        sum += NOISE_RATIO_ARRAY[i];
        if( value <= sum )
        {
           retVal = (i - MEAN_NUM) * oUnitVal[ equIdx ];
           break;
        }
     }

     return retVal;
  }

  private void gotoNextTrial()
  {
        trialIdx++;
        if ( trialIdx == TRIALS_PER_BLOCK )
        {
             if ( ( blockIdx + 1  ) < blockNum )
             {
                    blockIdx++;
                    initNewBlock();
             }
             else
             {
                  saveData();
                  if( ( agentIdx + 1 ) < MAX_AGENT_NUM )
                  {
                      agentIdx++;
                      initNewAgent();
                  }
                  else
                      finishTheTask( true );
             }
        }
  }

  private void initNewBlock()
  {
       if( TIMER_FLAG )
           stopInputTimer();

       JOptionPane.showMessageDialog( this, "Go to next block",
                                  "Next BLOCK",  JOptionPane.PLAIN_MESSAGE );
       trialIdx = 0;
       genFirstOutput();
  }

  private void initNewAgent()
  {
       if( TIMER_FLAG )
           stopInputTimer();

       JOptionPane.showMessageDialog( this, "Go to next session",
                                  "Next session",  JOptionPane.PLAIN_MESSAGE );
       blockIdx = 0;
       trialIdx = 0;

       yInputList.setEnabled( false );
       statusButton.setEnabled( true );
       equButton.setEnabled( true );
       parButton.setEnabled( true );
       timerButton.setEnabled( true );
       timeSetButton.setEnabled( TIMER_FLAG );
       noiseSetButton.setEnabled( true );
       targetButton.setEnabled( true );
       rangeButton.setEnabled( true );
       fromButton.setEnabled( false );
       toButton.setEnabled( false );
       confirmButton.setEnabled( false );

       beginButton.setEnabled( true );
       exitButton.setEnabled( true );
  }

  private void finishTheTask( boolean flag )
  {
       if( TIMER_FLAG )
           stopInputTimer();

       if( flag )
           JOptionPane.showMessageDialog( this, "All the sessions done!",
                                        "All done",  JOptionPane.PLAIN_MESSAGE );

       blockIdx = 0;
       trialIdx = 0;
       agentIdx = 0;
       TIMER_FLAG = false;

       //dataFile.closeFile();      // modifed on 08/18/02
       //stopInputTimer();

       dInputPanel.resetDraw();
       dOutputPanel.resetDraw();

       yInputList.setEnabled( false );
       statusButton.setEnabled( true );
       equButton.setEnabled( true );
       parButton.setEnabled( false );
       timerButton.setEnabled( true );
       timeSetButton.setEnabled( false );
       noiseSetButton.setEnabled( true );
       targetButton.setEnabled( true );
       rangeButton.setEnabled( true );
       fromButton.setEnabled( false );
       toButton.setEnabled( false );
       confirmButton.setEnabled( false );

       beginButton.setEnabled( false );
       exitButton.setEnabled( true );
  }

/*****************************************************************************/
 private void checkRangeInput()
 {
       try {
          int fromLevel = Integer.parseInt( fromField.getText() );
          int toLevel = Integer.parseInt( toField.getText() );
          if( fromLevel <= 0 || fromLevel > 1200 ||
              toLevel <= 0 || toLevel > 1200 || fromLevel > toLevel )
          {
              JOptionPane.showMessageDialog( null, "Number invalid, reset !",
                                             "Error", JOptionPane.ERROR_MESSAGE );
              fromField.setText( "500" );
              toField.setText( "700" );
          }
       }
       catch( NumberFormatException e1 ){
              JOptionPane.showMessageDialog( null, "Not a number, reset !",
              "Error", JOptionPane.ERROR_MESSAGE );
              fromField.setText( "500" );
              toField.setText( "700" );
       }
 }

/********************* ACTIVATED BY EQU AND PARA BUTTON **********************/
  private void setEnviron( int idx1, int idx2 )
  {
     MIN_INPUT  = MIN_LEVEL * iUnitVal[idx1];
     MAX_INPUT  = MAX_LEVEL * iUnitVal[idx1];
     MIN_OUTPUT = MIN_LEVEL * oUnitVal[idx1];
     MAX_OUTPUT = MAX_LEVEL * oUnitVal[idx1];
     dInputPanel.setUnitVal( iUnitVal[idx1] );
     dOutputPanel.setUnitVal( oUnitVal[idx1] );
     beginButton.setEnabled( true );
  }

/*****************************************************************************/
/******************************* SHOW FUNCTIONS ******************************/
/*****************************************************************************/
   public void showSettings( boolean flag )
   {
      for( int i  = 0; i < STATUS_NUMBER; i++ )
           statusArea[ i ].setEnabled( STATUS_FLAG );

      showEquPara( flag );
      showTimer( flag );
      showTimeLimit( flag );
     // showNoiseRatio( flag );
   }

   public void showEquType( boolean flag )
   {
       String str = "equation type: ";
       switch ( equIdx )
       {
        case LINEAR:
             str += "LINER";
             break;
        case U_SHAPE:
             str += "U_SHAPE";
             break;
        case POWER:
             str += "POWER";
             break;
       }

       String str1 = "equation parameters needed!";
       if( flag )
       {
           statusArea[ STATUS_EQU_TYPE ].setText( str );
           statusArea[ STATUS_EQU_PARA ].setText( str1 );
       }
       else
       {
           statusArea[ STATUS_EQU_TYPE ].setText( "" );
           statusArea[ STATUS_EQU_PARA ].setText( "" );
       }
   }

   public void showEquPara( boolean flag )
   {
       String str = "p1 = ";
       String a = String.valueOf( paramVal[equIdx][parIdx][0] );
       String b = String.valueOf( paramVal[equIdx][parIdx][1] );

       switch ( equIdx )
       {
        case LINEAR:
             if( parIdx != 2 )
                str += a + " * w  + " + b + " - p0 + N ";
             else
                str += a + " * w " + b + " - p0 + N ";
             break;
        case U_SHAPE:
             str += b + " * ( w  - " + a + " )^2 " + " - p0 + N";
             break;
        case POWER:
             String c = String.valueOf( paramVal[equIdx][parIdx][2] );
             String d = String.valueOf( paramVal[equIdx][parIdx][3] );
             str += b + "*" + a + "^" + "( " + c + "+ w/"+ d + " )" + " - p0 + N";
             break;
       }
       equationStr = str;
       if( flag )
           statusArea[ STATUS_EQU_PARA ].setText( str );
       else
           statusArea[ STATUS_EQU_PARA ].setText( "" );
   }

   public void showTimer( boolean flag )
   {
      String str = "Timer ";
      String str1 = "";
      if( TIMER_FLAG )
      {
          str += "will be used";
          str1 = "Time limit needed!";
      }
      else
          str += "will not be used";

      if( flag )
      {
          statusArea[ STATUS_TIMER ].setText( str );
          statusArea[ STATUS_TIME_LIMIT ].setText( str1 );
      }
      else
      {
          statusArea[ STATUS_TIMER ].setText( "" );
          statusArea[ STATUS_TIME_LIMIT ].setText( "" );
      }
   }

   public void showTimeLimit( boolean flag )
   {
       String str = "";
       if( TIMER_FLAG )
       {
           str = "Time limit: ";
           str += String.valueOf( TIME_INTERVAL ) + " ms ";
       }
       else
           str = "No time limit";

       if( flag )
           statusArea[ STATUS_TIME_LIMIT ].setText( str );
       else
           statusArea[ STATUS_TIME_LIMIT ].setText( "" );
   }

   public void showNoiseRatio( boolean flag )
   {
       String str = "Noise probability: ";
       str += String.valueOf( NOISE_RATIO * 100 ) + "%" + " up(or down)";

       if( flag )
           statusArea[ STATUS_NOISE ].setText( str );
       else
           statusArea[ STATUS_NOISE ].setText( "" );
   }

   public void showTarget()
   {
      dOutputPanel.drawTargetLine( TARGET_LEVEL );
   }

   public void showDecidedAction( )
   {
   }

  /************************* SAVE THE TRAINING DATA ************************/
   public void saveData()
   {
       String str = "The current equation type is:  ";
       str += equTypes[ equIdx ];
       dataFile.saveStringData( str );

       str = "The current equation is:  ";
       str += equationStr;
       dataFile.saveStringData( str );

       str = "The current time limit is:  ";
       str += String.valueOf( TIME_INTERVAL );
       dataFile.saveStringData( str );

       str = "The current noise setting is:  \n";
       //str += String.valueOf( NOISE_RATIO );
       for( int i = 0; i < NOISE_RATIO_NUM; i++ )
       {
            str += NOISE_LABEL_ARRAY[ i ] + " = ";
            str += String.valueOf( NOISE_RATIO_ARRAY[i]);
            str += "\n";
       }
       dataFile.saveStringData( str );

       checkRangeInput();
       str = "The lastest input-level range is : from level " + fromField.getText() +
             " to level " + toField.getText() + "\n";
       dataFile.saveStringData( str );

       dataFile.saveNumericData( dataArray, blockNum, TRIALS_PER_BLOCK,
                                  blockIdx, trialIdx );
   }
 }