// implements Q learning over a backprop neural net, as described
// in detail in the CLARION model.
// input is a state, and output is Q values of each action
package tasksimulators.oldsimu;
import clarion.common.*;

public class QBackPropNet extends BackPropNet implements Cloneable{

    protected int BLRT, BLPT, BLDT, BLAT;
    protected int RecommendedAction, actionPerformed;
    protected int BLOCK, blockNum, target;
    protected int targetDimIdx, timeSeqDimIdx;
    protected int dimValNums[];

    protected double reinforcement, discount;
    protected double PM = 0.0, NM = 0.0, succRate;
    protected double SUCC_C3 = 1.0, SUCC_C4 = 2.0;
    protected double[] state, preState, oldState;

    public QBackPropNet(int inputs,int hidden,int outputs,int dimNum,int tLevel )
    {
	super( inputs, hidden, outputs );

        target = tLevel;
	actionPerformed = -1;
        discount = World.TASK_PARA_ARRAY[World.Q_DISCOUNT_RATE];

	BLOCK = 2 * nOutput;
        blockNum = dimNum / 2;
        targetDimIdx = dimNum - 2;
        timeSeqDimIdx = dimNum - 1;

    	state = new double[ inputs ];
    	preState = new double[ inputs ];
    	oldState = new double[ BLOCK ];
//	initializeState( );
    }

    public QBackPropNet(int inputs,int hidden,int outputs,int dimNum,int[] dValNums )
    {
	super( inputs, hidden, outputs );

	actionPerformed = -1;
	BLOCK = 2 * nOutput;
        discount = World.TASK_PARA_ARRAY[World.Q_DISCOUNT_RATE];

        dimValNums = new int[ dimNum ];
        for( int i = 0; i < dimNum; i++ )
             dimValNums[ i ] = dValNums[ i ];

    	state = new double[ inputs ];
    	preState = new double[ inputs ];
    	oldState = new double[ BLOCK ];
	//initializeState( );
    }

    public QBackPropNet( int inputs, int hidden, int outputs )
    {
	super( inputs, hidden, outputs );
	actionPerformed = -1;
	BLOCK = 2 * nOutput;
        discount = World.TASK_PARA_ARRAY[World.Q_DISCOUNT_RATE];
    	state = new double[ inputs ];
    	preState = new double[ inputs ];
    	oldState = new double[ BLOCK ];
    }

    public void reinit()
    {
        //PM = NM = 0.0;
        resetPM();
        resetNM();
	actionPerformed = -1;
	initializeState( );
        reinitWeights();
    }

    public void reinitBlock( )
    {
	actionPerformed = -1;
	initializeState( );
    }

    public void resetPM()
    {
        PM = 0.0;
    }

    public void resetNM()
    {
        NM = 0.0;
    }

    public void initializeState( )
    {
	for( int i = 0; i < nInput; i++ )
             state[i] = 0;

	for( int i = 0; i < nInput; i++ )
             preState[i] = 0;

	for( int i = 0; i < BLOCK; i++ )
             oldState[i] = 0;
    }

    public double getSuccRate()
    {
       succRate = ( SUCC_C3 + PM ) / ( SUCC_C4 + PM + NM );
       return succRate;
    }

    public void getState( int[] fillMe )
    {
    	for(int i = 0; i < nInput; i++ )
             fillMe[i] = (int) state[i];
    }

    public void setState( int[] arr )
    {
	for( int i = 0; i < nInput; i++ )
             state[i] = 0;

        int offset = 0;
        for( int i = 0; i < arr.length; i++ )
        {
             state[ arr[i] + offset - 1 ] = 1;
             offset += dimValNums[i];
        }
    }

    public void setState( short[][] arr )
    {
        int offset = 0;
        for( int i = 0; i < arr.length; i++ )
        {
             for( int j = 0; j < dimValNums[i]; j++ )
                  state[ offset + j ] = arr[ i ][ j ];
             offset += dimValNums[i];
        }
    }

    public void setState( short[][][] arr )
    {
        int offset = 0, dimOffset = 0;
        for( int i = 0; i < arr.length; i++ )
        {
             for( int j = 0; j < arr[i].length; j++ )
             {
                  for( int k = 0; k < dimValNums[ dimOffset ]; k++ )
                       state[ offset + k ] = arr[ i ][ j ][ k ];
                  offset += dimValNums[ dimOffset ];
                  dimOffset++;
             }
        }
    }

    public void setPreState( )
    {
        for( int i = 0; i < nInput; i++ )
             state[ i ] = preState[ i ];
    }

    public void setCurState( int[] curState )
    {
        for( int i = 0; i < nInput; i++ )
             preState[ i ] = state[ i ];

        for( int i = 0; i < nInput; i++ )
             state[ i ] = curState[ i ];
    }

    public void setPreState( int trialNO )
    {
        if( blockNum > 2 )
        {
           for( int i = 0; i < blockNum - 2; i++ )
              for( int j = 0; j < BLOCK; j++ )
                   state[ i * BLOCK + j ] = state[ ( i + 1 ) * BLOCK  + j ];

        }

        for( int i = 0; i < BLOCK; i++ )
             state[ i + ( blockNum - 2 ) * BLOCK ] = oldState[ i ];

        for( int i = ( blockNum - 1 ) * BLOCK; i < blockNum * BLOCK; i++ )
             state[ i ] = 0;

        state[ target +  targetDimIdx * nOutput ] = 1;
        state[ trialNO + timeSeqDimIdx * nOutput ] = 1;
    }

    public void setCurState(int action, int result, int trialNO )
    {
        // store the Input & Output of time step t - 2
	for( int i = 0; i < BLOCK; i++ )
	     oldState[ i ] = state[ i + ( blockNum - 2 ) * BLOCK ];

        if( blockNum > 2 )
        {
           for( int i = blockNum - 2; i > 0; i-- )
              for( int j = 0; j < BLOCK; j++ )
                   state[ i * BLOCK + j ] = state[ ( i - 1 ) * BLOCK  + j ];
        }

	//map attitude(t)=nextAttitude, action(t-1)=RecommendedAction
	for( int i = 0; i < BLOCK; i++ )
             state[i] = 0;

	state[ result ] = 1;
	state[ action + nOutput ] = 1;

        for( int i = ( blockNum - 1 ) * BLOCK; i < blockNum * BLOCK; i++ )
             state[ i ] = 0;

        state[ target +  targetDimIdx * nOutput ] = 1;
        state[ trialNO + timeSeqDimIdx * nOutput ] = 1;
    }

    public void setInput()
    {
	setInput( state );
    }

    public double getDiscount()
    {
	return discount;
    }

    public void setDiscount( double value )
    {
	discount = value;
    }

    public double getReinforcement()
    {
       return reinforcement;
    }

    public void setReinforcement(double value)
    {
	reinforcement = value;
    }

    public int getResponseTime()
    {
         BLPT = 200;
         BLAT = 500;
         BLDT = 200;
         BLRT = BLPT + BLDT + BLAT;
         return BLRT;
    }

    public double getQVal( int index )
    {
       return Output[ index ];
    }

    // get the system's recommendations
    public void getQvals (double[] FillMe)
    {
    	for (int i = 0; i < nOutput; i++ )
              FillMe[i] = Output[i];
    }

   // to find the max Q value of all the Q values.
    public double getMaxQVal()
    {
       double highest = Output[ 0 ];
       for( int i = 1; i < nOutput; i++ )
            if( Output[ i ] > highest )
                highest = Output[ i ];
        return highest;
    }

    // let the system know what action was performed
    // this speeds up the backprop of error, since error on all other
    // output units is 0, call this function before computeErrors!!!
    public void setPerformedAction(int action)
    {
    	if ( action > ( nOutput - 1 ) )
        {
	    System.out.println("Set Performed Action > nOutput");
	    System.exit(1);
	}
        else
           actionPerformed = action;
    }

/************************** QBP version****************************************/

    // optimized, since nInput-window inputs are always zero (half on average)
    //further optimized by noting that even in active blocks only 1 out of
    // nOutput unit is active
    public void computeHiddenActivation()
    {
	double sum, v;

	for ( int j = 0; j < nHidden; j++ )
        {
	    sum = HiddenThresholds[j];
            for (int i = 0; i < nInput; i++)
            {
		if( Input[i] > 0.1 )
                    sum += Input [i] * InputToHiddenWeights[i][j];
            }
            v = activation(sum);
	    Hidden[j] = v;
	    HiddenDeriv[j] = derivative(v);
	}
    }

    // compute the errors of QBP network
    public void computeErrors( double qValue1, double qValue2 )
    {
       for ( int i = 0; i < nOutput; i++ )
       {
           if( i != actionPerformed )
                 Errors[i] = 0;
           else {
		    double err = reinforcement + discount * qValue2 - qValue1;
/******************************************************************************/
                    if( err > 1.0 )
                        err = 1.0;
                    else if( err < -1.0 )
                             err = -1.0;
/*****************************************************************************/
		    Errors[i] = OutputDeriv[i] * err;
		    sumSqErrors = err * err / 2;

		    if (debug)
                      System.out.println( "Activn[" + i + "]=" + Output[i]
                            + ", Err= " + err + ", Reinf= " + reinforcement
                            + ", ErrTerm= " + Errors[i]
                            + " OutDer= " + OutputDeriv[i] );

		    //if (debug) System.out.println(sumSqErrors);
		    if( Math.abs( err ) < tolerance )
                        noError = true;
		    else
                        noError = false;
		}
       }
    }

    // compute the errors of simplified QBP network
    public void computeErrors( double qValue )
    {
       for ( int i = 0; i < nOutput; i++ )
       {
           if( i != actionPerformed )
                 Errors[i] = 0;
           else {
		    double err = reinforcement - qValue;
/******************************************************************************/
                    if( err > 1.0 )
                        err = 1.0;
                    else if( err < -1.0 )
                             err = -1.0;
/*****************************************************************************/
		    Errors[i] = OutputDeriv[i] * err;
		    sumSqErrors = err * err / 2;

		    //if (debug) System.out.println(sumSqErrors);
		    if( Math.abs( err ) < tolerance )
                        noError = true;
		    else
                        noError = false;
		}
       }
    }

    // optimized, since all errors except for one are 0
    public void modifyHiddenToOutput ()
    {
	double x = Errors[ actionPerformed ];
	for ( int i = 0; i < nHidden; i++ )
        {
            // update the HiddenToOutputWeights
	    if ( Momentum > RZero )
            {
		/* momentum */
		HiddenToOutputMomentum[i][actionPerformed] = Eta * x * Hidden[i]
                      + Momentum * HiddenToOutputMomentum[i][actionPerformed];
		HiddenToOutputWeights[i][actionPerformed] +=
                       HiddenToOutputMomentum[i][actionPerformed];
	    }
             else
	       	HiddenToOutputWeights[i][actionPerformed] += Eta * x * Hidden[i];
	}

        // update the OutputThresholds
	if (Momentum > RZero)
        {
	    OutputMomentum[ actionPerformed ] =  Eta * x +
                      Momentum * OutputMomentum[actionPerformed];
	    OutputThresholds[ actionPerformed ] +=
                         OutputMomentum[ actionPerformed ];
         }
        else
	    OutputThresholds[ actionPerformed ] += Eta * x;
    }

    public void update( int action, double feedback, double qValue )
    {
	//let QBP know what action was performed
	setPerformedAction( action );
        //update the reinforcement
        setReinforcement( feedback );
        // update QBPNET
	computeErrors( qValue );
	backwardPass();
    }

    public void update(int action, double feedback, double qVal1, double qVal2 )
    {
	//let QBP know what action was performed
	setPerformedAction( action );
        //update the reinforcement
        setReinforcement( feedback );

        setPreState();
        setInput();
        forwardPass();
	computeErrors( qVal1, qVal2 );
	backwardPass();
    }

    public void updateMatchCount( boolean flag )
    {
        if( flag )
            PM += 1.0;
        else
            NM += 1.0;
    }
}
