package tasksimulators.oldsimu;
import clarion.acs.*;

 public class RerGS extends GS {

   final int DIRECT_TYPE = 2, NOTHING_TYPE = 3, ILLEGAL_TYPE = -1;
   final int POP_ACTION = 12, DO_NOTHING_ACTION = 13;

   int gsAction, gsTLAction, gsBLAction, actionType;
   int discLocs[], nodeInfoArr[];
   double tlSuccRate, blSuccRate;

   T_RerRuleSet rerRuleSet;

   public RerGS( int curSize, int dimNum, int stateLen, int[] netInfoArr,
                 int[] dimValInfo )
   {
       super( curSize, dimNum, stateLen, netInfoArr, dimValInfo );

       nodeInfoArr = new int[ netInfoArr.length + 1 ];
       nodeInfoArr[ 0 ] = netInfoArr[ 0 ];
       for( int i = 0; i < netInfoArr.length; i++ )
            nodeInfoArr[ i +1  ] = netInfoArr[ i ];

       rerRuleSet = new T_RerRuleSet( 1, 0, nodeInfoArr, true );
   }

   public void reinit( int tSize )
   {
       super.reinit( tSize );
       rerRuleSet.reinit();
   }

   public void initDiscLocs( int discNum )
   {
       discLocs = new int[ discNum ];
   }

   public double getBLSuccRate()
   {
       blSuccRate = super.getSuccRate();
       return blSuccRate;
   }

   public double getTLSuccRate()
   {
       tlSuccRate = 1.0;
       return tlSuccRate;
   }

   public int getActionBL( int[] condition )
   {
      gsBLAction = getAction( condition );
      return gsBLAction;
   }

   public int getFRAction()
   {
      int retAction = -1;
      getActiveItem();   // get current goal.

      boolean achieved = checkAchieved();
      if( achieved )
      {
          retAction = POP_ACTION;
          return retAction;
      }

      boolean towerAchieved = checkTowerAchieved();
      if( towerAchieved )
      {
          int size = activeItem[ 3 ];  // largest disc of the current goal.
          int to = activeItem[ 5 ];    // destination peg of current largest disc.
          int loc = 0;

          int BLOCK_SIZE = DIM_VAL_NUM * DIM_VAL_NUM;
          for( int i = 0; i < discLocs.length; i++ )
          {
               int dSize = discLocs[ i ] % DIM_VAL_NUM;
               if( dSize == size )
               {
                   loc = discLocs[ i ] / BLOCK_SIZE;
                   break;
               }
          }

          retAction = 2 * size + ( to - loc + 2 ) % 3;
          return retAction;
      }

      if( activeItem[ 0 ] == 0 )
      {
          int size = activeItem[ 0 ];  // largest disc of the current goal.
          int to = activeItem[ 2 ];    // destination peg of current largest disc.
          int loc = 0;

          int BLOCK_SIZE = DIM_VAL_NUM * DIM_VAL_NUM;
          for( int i = 0; i < discLocs.length; i++ )
          {
               int dSize = discLocs[ i ] % DIM_VAL_NUM;
               if( dSize == size )
               {
                   loc = discLocs[ i ] / BLOCK_SIZE;
                   break;
               }
          }

          retAction = 2 * size + ( to - loc + 2 ) % 3;
          return retAction;
      }

      return retAction;
   }

   public int getActionTL( int[] condition )
   {
      gsTLAction = getFRAction();

      if( gsTLAction == -1 )
          gsTLAction = rerRuleSet.getAction( condition );
      return gsTLAction;
   }

   public double getQVal( int aciton )
   {
      double qval = gsNet.getQVal( action );
      return qval;
   }

   public void setDiscLocs( int[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           discLocs[ i ] = arr[ i ];
   }

   public void setThreshold( double th )
   {
      threshold = th;
   }

   public void setThreshold( double th1, double th2 )
   {
      rerRuleSet.setThreshold( th1, th2 );
   }

   public void setValidNum( int num )
   {
       rerRuleSet.setValidDimNum( num );
   }

   public void setFullFlag( boolean flag )
   {
       rerRuleSet.setFullFlag( flag );
   }

   // check if the chosen action is legal.
   public boolean checkAction( int chosenAction )
   {
      gsAction = chosenAction;
      actionType = ILLEGAL_TYPE;

      if( chosenAction == - 1 )
      {
          actionType = ILLEGAL_TYPE;
          return false;
      }

      if( chosenAction == DO_NOTHING_ACTION )
      {
          actionType = NOTHING_TYPE;
          return false;
      }

      if( chosenAction == POP_ACTION ) // case of pop actions.
      {
         /*
          boolean achieved = checkAchieved();
          if( achieved )
          {
              actionType = POP_TYPE;
              return true;
          }
          else
          {
              actionType = ILLEGAL_TYPE;
              return false;
          }
          */
          boolean bVal = getActiveItem();
          if( bVal )   // goal stack is not empty.
          {
              actionType = POP_TYPE;
              return true;
          }
          else
          {
              actionType = ILLEGAL_TYPE;
              return false;
          }
      }
      else   // cases of push action or direct-move action.
      {
          int size = chosenAction / 2;
          if( size > MAX_TOW_SIZE )
          {
              actionType = ILLEGAL_TYPE;
              return false;
          }

          boolean achieved = checkAchieved();
          if( achieved )
          {
              /*
              actionType = ILLEGAL_TYPE;
              return false;
              */
              actionType = PUSH_TYPE;
              return true;
          }

          int tSize = activeItem[ 0 ];  // sub-tower size of the current goal.
          boolean towerAchieved = checkTowerAchieved();
          if( towerAchieved )
          {
              int dSize = activeItem[ 3 ]; // the largest disc of the current goal.
              if( size == dSize )
              {
                  actionType = DIRECT_TYPE;
                  return true;           // direct-move action.
              }
              else
              {
                /*
                  actionType = ILLEGAL_TYPE;
                  return false;
                */
                actionType = PUSH_TYPE;
                return true;
              }
          }
          else
          {
               if( size > tSize )
               {
                   /*
                   actionType = ILLEGAL_TYPE;
                   return false;
                   */
                   actionType = PUSH_TYPE;
                   return true;
               }
               else if( size == tSize )
                    {
                       if( tSize == 0 )
                       {
                           actionType = DIRECT_TYPE;
                           return true;     // direct-move action.
                       }
                       else
                       {
                           /*
                           actionType = ILLEGAL_TYPE;
                           return false;
                           */
                           actionType = PUSH_TYPE;
                           return true;
                       }
                    }
                    else
                    {
                       actionType = PUSH_TYPE;
                       return true;         // push action.
                    }
           }
      }
   }

   public boolean checkAchieved( )
   {
      boolean towerAchieved = checkTowerAchieved();
      if( !towerAchieved )
          return false;

      int size = activeItem[ 3 ];  // largest disc of the current goal.
      int to = activeItem[ 5 ];     // destination peg of current largest disc.

      int BLOCK_SIZE = DIM_VAL_NUM * DIM_VAL_NUM;
      for( int i = 0; i < discLocs.length; i++ )
      {
         int dSize = discLocs[ i ] % DIM_VAL_NUM;
         if( dSize == size )
         {
             int loc = discLocs[ i ] / BLOCK_SIZE;
             if( loc != to )
                 return false;
         }
      }
      return true;
   }

   public boolean checkTowerAchieved( )
   {
      int tSize = activeItem[ 0 ];  // sub-tower size of the current goal.
      int to = activeItem[ 2 ];     // destination peg of sub-tower.

      int BLOCK_SIZE = DIM_VAL_NUM * DIM_VAL_NUM;
      for( int i = 0; i < discLocs.length; i++ )
      {
         int dSize = discLocs[ i ] % DIM_VAL_NUM;
         if( dSize <= tSize )
         {
             int loc = discLocs[ i ] / BLOCK_SIZE;
             if( loc != to )
                 return false;
         }
      }
      return true;
   }

   // check if current chosen action is a direct move or not.
   // if it was not, return -2( nothing action ) or -1( push or pop );
   // otherwise convert it into disc-move action.
   public int checkDirectMove()
   {
      int action = 0;
      if( actionType == NOTHING_TYPE )
      {
          action = -2;
          return action;
      }

      if( actionType != DIRECT_TYPE )
      {
          action = -1;
          return action;
      }

      int size = gsAction / 2;
      int pegIdx = getPegIdx( size );
      int offset = gsAction % 2;
      action = 2 * pegIdx + offset;

      return action;
   }

   // get the peg from which the GS action is.
   public int getPegIdx( int size )
   {
      int pegIdx = 0;
      //int size = gsAction / 2;
      int BLOCK_SIZE = DIM_VAL_NUM * DIM_VAL_NUM;
      for( int i = 0; i < discLocs.length; i++ )
      {
           int dSize = discLocs[ i ] % DIM_VAL_NUM;
           if( dSize == size )
           {
               pegIdx = discLocs[ i ] / BLOCK_SIZE;
               break;
           }
      }
      return pegIdx;
   }

   // perform the chosen action.
   public void performAction()
   {
      if( actionType == DIRECT_TYPE || actionType == NOTHING_TYPE )
          return;

      int[] item = new int[ 2 + DIM_NUM ];
      item[ 0 ] = GS_ACTION;
      item[ 1 ] = actionType;

      if( actionType == PUSH_TYPE )
      {
          /*
          int size = gsAction / 2;
          int from = getPegIdx( size );
          item[ 2 ] = size;
          item[ 3 ] = from;       // current loc of sub-tower
          if( activeItem[1] != activeItem[2] )
              item[ 4 ] = 3 - activeItem[ 1 ] - activeItem[ 2 ];
          else
              item[ 4 ] = 2 - activeItem[ 1 ];

          item[ 5 ] = activeItem[ 0 ];
          item[ 6 ] = activeItem[ 1 ];
          item[ 7 ] = activeItem[ 2 ];
          */
          int dSize = gsAction / 2;
          int from1 = getPegIdx( dSize );
          int to1 = ( from1 + 1 + ( gsAction % 2 ) ) % 3;
          int tSize = 0;
          if( dSize > 0 )
              tSize = dSize - 1;
          else
              tSize = dSize;
          int from = getPegIdx( tSize );
          int to = 3 - from1 - to1;

          item[ 2 ] = tSize;
          item[ 3 ] = from;  // current loc of sub-tower
          item[ 4 ] = to;
          item[ 5 ] = dSize;
          item[ 6 ] = from1;  // current loc of focal disc.
          item[ 7 ] = to1;
      }
      performAction( item );
   }

   // update the RER_GS system.
   public void update(  boolean flag )
   {
       updateAction( gsAction );
       if( action < 0 )
           return;

       if( flag )
       {
           legalUpdate();
           // update RER rule set if the RER flag is set.
           double r = gsNet.getReinforcement();
           double d = gsNet.getDiscount();

           if( r + d * gsQValue2 - gsQValue1 > threshold )
              rerRuleSet.updateRuleSet( preState, action, true, false, true );
           else
              rerRuleSet.updateRuleSet( preState, action, false, false, true );
       }
       else
       {
           illegalUpdate();
           if( levelFlag == TOP_LEVEL )
               rerRuleSet.deleteRule( preState, action );
       }
   }

   public StringBuffer getAllRules()
   {
       return rerRuleSet.getAllRules();
   }

   public StringBuffer getMatchRules()
   {
       return rerRuleSet.getMatchRules();
   }
   public StringBuffer getChosenRules()
   {
       return rerRuleSet.getChosenRules();
   }
 }
