package tasksimulators.oldsimu;

public class RuleFeature {

   // TASK PARAMETER SETTING.
   int UTL_SEL_IDX, BLA_SEL_IDX;
   boolean DEF_UTL_FLAG, DEF_BLA_FLAG;

  // for LC response time.
   int TIME_PER_STEP = 300;

   int TLRT, TLPT, TLDT, TLAT;
   int BLRT, BLPT, BLDT, BLAT;

   // for base-level activation.
   int TOT_TRIAL_NUM, usedNum, usageInfoArr[];
   double INIT_BLA, BLA, PRE_BLA, EXP_BLA = -1;
   double BLA_C = 2.0, BLA_P = -0.5;
   double RULE_ACTION_TIME, PRE_CHUNK_BLA;

   // used for utility
   double BENEFIT_C7 = 1.0, BENEFIT_C8 = 2.0, COST_V = 1.0;
   double utility, benefit, PM, NM, cost;

   // counstructor of the class RuleReature.
   public RuleFeature( int totTrialNum, double iBLA, int[] selArr, boolean[] defFlags )
   {
       BLPT = 200;
       TLPT = BLPT + 135;
       BLAT = 500;
       TLAT = BLAT;
       BLDT = 200;

       TOT_TRIAL_NUM = totTrialNum;
       INIT_BLA = iBLA;
       usageInfoArr = new int[ TOT_TRIAL_NUM * 5 ];

       BLA_SEL_IDX = selArr[ 0 ];
       UTL_SEL_IDX = selArr[ 1 ];

       DEF_UTL_FLAG = defFlags[ 0 ];
       DEF_BLA_FLAG = defFlags[ 1 ];

       // set the values of parameters about rule utility.
       if( DEF_UTL_FLAG )
       {
         if( UTL_SEL_IDX == 0 )
         {
            benefit = 2;
            cost = 1;
            COST_V = 1.0;
         }
      }
    }

   // counstructor of the class RuleReature.
   public RuleFeature( int[] selArr, boolean[] defFlags )
   {
       BLA_SEL_IDX = selArr[ 0 ];
       UTL_SEL_IDX = selArr[ 1 ];

       DEF_UTL_FLAG = defFlags[ 0 ];
       DEF_BLA_FLAG = defFlags[ 1 ];

       // set the values of parameters about rule utility.
       if( DEF_UTL_FLAG )
       {
         if( UTL_SEL_IDX == 0 )
         {
            benefit = 2;
            cost = 1;
            COST_V = 1.0;
         }
      }
   }

   public void reinit()
   {
       PM = 0.0;
       NM = 0.0;

       // set the values of parameters about rule utility.
       if( DEF_UTL_FLAG )
       {
           if( UTL_SEL_IDX == 0 )
           {
               benefit = 2;
               cost = 1;
               COST_V = 1.0;
           }
       }

       usedNum = 0;
       for( int i = 0; i < usageInfoArr.length; i++ )
            usageInfoArr[i] = 0;
   }

   public double getPM()
   {
        return PM;
   }

   public double getNM()
   {
        return NM;
   }

   public double getUtility()
   {
       return utility;
   }

   public double getBLA()
   {
       return BLA;
   }

   public int getResponseTime()
   {
       TLRT = TLPT + TLDT + TLAT;
       return TLRT;
   }

   public void setPM( double PM )
   {
       this.PM = PM;
   }

   public void setNM( double NM )
   {
       this.NM = NM;
   }

   public void setUsedNum( int usedNum )
   {
      this.usedNum = usedNum;
   }

   public void setUsageInfoArr( int[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           usageInfoArr[ i ] = arr[ i ];
   }

   public void setCost( double c )
   {
       cost = c;
   }

   public void setActionTime( double time )
   {
       RULE_ACTION_TIME = time;
   }

   public void setPreBLA( double val )
   {
       PRE_BLA = val;
   }

   public void setPreChunkBLA( double val )
   {
       PRE_CHUNK_BLA = val;
   }

   public Object clone()
   {
      int[] selArr = new int[ 2 ];
      boolean[] defFlags = new boolean[ 2 ];

      selArr[ 0 ] = BLA_SEL_IDX;
      selArr[ 1 ] = UTL_SEL_IDX;
      defFlags[ 0 ] = DEF_UTL_FLAG;
      defFlags[ 1 ] = DEF_BLA_FLAG;
      RuleFeature feature = new RuleFeature( TOT_TRIAL_NUM, INIT_BLA, selArr, defFlags );

      feature.setUsedNum( usedNum );
      feature.setUsageInfoArr( usageInfoArr );

      feature.setPM( PM );
      feature.setNM( NM );
      feature.setCost( cost );
      feature.setActionTime( RULE_ACTION_TIME );
      feature.setPreBLA( PRE_BLA );
      feature.setPreChunkBLA( PRE_CHUNK_BLA );

      return feature;
   }

   public void updateMatchCounts( boolean succFlag )
   {
       if( succFlag )
           PM = PM + 1;
       else
           NM = NM + 1;
   }

   public void updateUsageInfo( int curTime )
   {
       usageInfoArr[ usedNum ] = curTime;
       usedNum++;
   }

   // for LC counting rules.
   public void calcTLDT( int steps )
   {
        TLDT = (int)( steps * TIME_PER_STEP + PRE_BLA * Math.pow(BLA, EXP_BLA));
   }

   // for LC retrieval rules && RER rules.
   public void calcTLDT( )
   {
        TLDT = (int)( RULE_ACTION_TIME + PRE_BLA * Math.pow(BLA, EXP_BLA));
   }

   // for LC retrieval rules.
   public void calcTLAT( double chunkBLA )
   {
        double val = PRE_CHUNK_BLA * Math.pow( chunkBLA, EXP_BLA );
        TLAT = (int)( BLAT + val );
   }

   public void calcBLA( int curTime )
   {
      if( DEF_BLA_FLAG )
      {
         if( BLA_SEL_IDX == 0 )
             BLA = 0;
         else
         {
             double sum = INIT_BLA;
             for( int i = 0; i < usedNum; i++ )
             {
                  int timeLaps = curTime - usageInfoArr[ i ];
                  double val = BLA_C * Math.pow( (double)timeLaps, BLA_P );
                  sum += val;
             }
             BLA = sum;
         }
      }
   }

   public void calcUtility( )
   {
       if( DEF_UTL_FLAG )
       {
         if( UTL_SEL_IDX == 0 )
         {
             benefit = 2;
             utility = benefit - COST_V * cost;
             if( utility < 0 )
                 utility = 0;
         }
         else
         {
             benefit = ( BENEFIT_C7 + PM ) / ( BENEFIT_C8 + PM + NM );
             utility = benefit - COST_V * cost;
          }
       }
   }

   public void incCost( int val )
   {
       cost += val;
   }

   public void decCost( int val )
   {
       cost -= val;
   }
 }