/* fix rules class for using rule utility, rule BLA, rule support etc. */
package tasksimulators.oldsimu;
import java.lang.*;
import java.util.Vector;
import java.awt.*;

public class THFixRule {

  final int GS_TOW_DSIZ_IDX = 0, GS_TOW_FROM_IDX = 1, GS_TOW_TO_IDX = 2;
  final int GS_BTM_DSIZ_IDX = 3, GS_BTM_FROM_IDX = 4, GS_BTM_TO_IDX = 5;
  final int WM_ACTION = 0, EX_ACTION = 1, GS_ACTION = 2;
  final int PUSH_ACTION = 0, POP_ACTION = 1;

  int EX_DIM_NUM, EX_VAL_NUM, GS_DIM_NUM, GS_VAL_NUM;
  int curMovedDisc, wmCon[], exCon[], gsCon[], resultArr[];
  boolean PUSH_FLAG, POP_FLAG, GOAL_FLAG;

  public THFixRule( int num1, int num2, int num3,  int dNum1, int vNum1 )
  {
      wmCon = new int[ num1 ];
      exCon = new int[ num2 ];
      gsCon = new int[ num3 ];

      EX_DIM_NUM = dNum1;
      EX_VAL_NUM = vNum1;

      curMovedDisc = -1;
      PUSH_FLAG = false;
      POP_FLAG = false;
      GOAL_FLAG = false;
  }

  public void reinit()
  {
      curMovedDisc = -1;
      PUSH_FLAG = false;
      POP_FLAG = false;
      GOAL_FLAG = false;
  }

  public boolean getGoalFlag( )
  {
      return GOAL_FLAG;
  }

  public void getFlags( boolean[] flagArr )
  {
      flagArr[ 0 ] = POP_FLAG;
      flagArr[ 1 ] = PUSH_FLAG;
      flagArr[ 2 ] = GOAL_FLAG;
  }

  public void setCondition( int[] arr1, int[] arr2, int[] arr3 )
  {
      for( int i = 0; i < arr1.length; i++ )
           wmCon[ i ] = arr1[ i ];
      for( int i = 0; i < arr2.length; i++ )
           exCon[ i ] = arr2[ i ];
      for( int i = 0; i < arr3.length; i++ )
           gsCon[ i ] = arr3[ i ];
  }

  public void setFlags()
  {
      POP_FLAG = true;
      PUSH_FLAG = false;
      GOAL_FLAG = true;
  }

  // set current moved disc.
  public void setCurMovedDisc( int dSize )
  {
      curMovedDisc = dSize;
  }

  public int searchLocation( int[] arr, int begin, int end, int val )
  {
     if( begin >= arr.length )
         return 0;

     int idx = begin;
     while( arr[ idx ] != val )
     {
            idx++;
            if( idx == end )
                break;
     }

     if( idx == end )
         return -1;
     else
         return idx - begin;
  }

  // get the peg which the disc with size of dSize is on.
  public int getPegIdx( int dSize )
  {
      int val = 0;
      for( int i = 0; i < 3; i++ )
        if( exCon[ dSize * EX_VAL_NUM + i ] == 1 )
        {
            val = i;
            break;
        }
      return val;
  }

  public void getResultArr( int[] arr )
  {
     for( int i = 0; i < resultArr.length; i++ )
          arr[ i ] = resultArr[ i ];
  }

  // get the "other" peg except "from" peg and "to" peg
  public int getOtherPeg( int from, int to )
  {
      int other = 0;
      if( from != to )
          other = 3 - from - to;
      else if( from > 0 )
               other = 3 - from;
           else
               other = 2 - from;
      return other;
  }

  // check if the sub-tower is achieved.
  // if achieved, return  -1,
  // otherwise return the largest out-of-place disc.
  public int checkSubTower()
  {
      int dSize = gsCon[ GS_TOW_DSIZ_IDX ];
      int to = gsCon[ GS_TOW_TO_IDX ];

      int size = dSize;
      do
      {
          int loc = getPegIdx( size );
          if( loc == to )
              size--;
          else
              break;
      }
      while( size >= 0 );

      return size;
  }

  public void formExAction( int from, int to )
  {
      resultArr = new int[ 3 ];
      resultArr[ 0 ] = EX_ACTION;
      resultArr[ 1 ] = from;
      resultArr[ 2 ] = to;
  }

  public void pushGoal( int dSize, int from1, int to1, int from2, int to2 )
  {
      resultArr = new int[ 8 ];

      resultArr[ 0 ] = GS_ACTION;
      resultArr[ 1 ] = PUSH_ACTION;

      resultArr[ GS_TOW_DSIZ_IDX + 2 ] = dSize-1;
      resultArr[ GS_TOW_FROM_IDX + 2 ] = from1;
      resultArr[ GS_TOW_TO_IDX + 2 ] = to1;
      resultArr[ GS_BTM_DSIZ_IDX + 2 ] = dSize;
      resultArr[ GS_BTM_FROM_IDX + 2 ] = from2;
      resultArr[ GS_BTM_TO_IDX + 2 ] = to2;
  }

  public void popGoal( )
  {
      resultArr = new int[ 2 ];
      resultArr[ 0 ] = GS_ACTION;
      resultArr[ 1 ] = POP_ACTION;
  }

  public void initGoals( int dSize, int from, int to, int other )
  {
      if( dSize > 0 )
          pushGoal( dSize, from, other, from, to );
      else
          formExAction( from, to );
  }

  public void moveDisc( int loc1, int to1 )
  {
      if( loc1 != to1 )
          formExAction( loc1, to1 );
      else
          popGoal();
  }

  // the function performs the actual content of the fixed rule.
  public boolean matchRule()
  {
      int dSize = gsCon[ GS_TOW_DSIZ_IDX ];
      int from = gsCon[ GS_TOW_FROM_IDX ];
      int to = gsCon[ GS_TOW_TO_IDX ];
      int other = getOtherPeg( from, to );

      int dSize1 = gsCon[ GS_BTM_DSIZ_IDX ];
      int from1 = gsCon[ GS_BTM_FROM_IDX ];
      int to1 = gsCon[ GS_BTM_TO_IDX ];
      int other1 = getOtherPeg( from1, to1 );

      if( curMovedDisc == -1 )
      {
          initGoals( dSize, from, to, other );
          return true;
      }

      if( curMovedDisc < dSize1 )
      {
          POP_FLAG = false;
          PUSH_FLAG = true;
      }

      if( POP_FLAG )
      {
          if( curMovedDisc > dSize1 )
          {
              popGoal();
              GOAL_FLAG = false;
          }

          if( curMovedDisc == dSize1 )
          {
              int loc1 = getPegIdx( dSize1 );
              if( loc1 == to1 )
              {
                  POP_FLAG = false;
                  PUSH_FLAG = true;
              }
              else
              {
                 int size = checkSubTower();
                 if( size == -1 )
                     moveDisc( loc1, to1 );
                 else
                 {
                     popGoal();
                     GOAL_FLAG = false;
                 }
              }
         }
      }

      if( PUSH_FLAG )
      {
          int size = checkSubTower();
          if( size == -1 )
          {
              int loc1 = getPegIdx( dSize1 );
              if( loc1 != to1 )
                  moveDisc( loc1, to1 );
              else
                  popGoal();
              return true;
          }

          int loc = getPegIdx( size );
          if( size > 0 )
          {
              int loc1 = getPegIdx( size-1 );
              pushGoal( size, loc1, 3-loc-to, loc, to );
          }
          else
          {
              formExAction( loc, to );
              PUSH_FLAG = false;
          }
      }

      return true;
  }
}
