/* the LETTER COUNTING AGENT FOR LETTER COUNTING TASK  */
package tasksimulators.oldsimu;
import java.lang.*;
import java.util.Vector;
import java.util.Stack;
import clarion.common.*;
import clarion.tools.*;
import clarion.acs.*;

public class THFrAgent {

    final int PEG_NUM = 3, MAX_DISC_NUM = 6;
    final int TOP_LEVEL = 0, DOWN_LEVEL = 1;
    final int MAX_ITEM_VAL = 100, ENLARGE = 10000;
    final int STABLE_EPISODE_NUM = 3;

    // used for QBP networks.
    final int TOT_INPUT_NUM = 30;
    final int EX_INPUT_NUM = 18, WM_INPUT_NUM = 6, GS_INPUT_NUM = 6;
    final int EX_DIM_NUM = 18, GS_DIM_NUM = 6, WM_DIM_NUM = 6;
    final int WM_STATE_LEN = 36, EX_STATE_LEN = 108, GS_STATE_LEN = 36;
    final int EX_NET_INFO[] = { 8, 30, 60, 6 };
    final int WM_NET_INFO[] = { 8, 30, 60, 18 };
    final int GS_NET_INFO[] = { 8, 30, 60, 13 };
    final int DIM_VAL_INFO[] = { 6, 3, 3, 3, 3, 3, 3, 6 };
    final int PUSH_ACTION = 0, POP_ACTION = 1;
    final int WM_ACTION = 0, EX_ACTION = 1, GS_ACTION = 2;

    /* original*/
    final int PUSH_TIME = 2500, POP_TIME = 500;
    final int BLPT = 500, BLDT = 350, BLAT = 1500;
    final int TLPT = BLPT + 135, TLDT = 0, TLAT = 1500;

    /* simluate disc-4
    final int PUSH_TIME = 2200, POP_TIME = 500;
    final int BLPT = 500, BLDT = 350, BLAT = 1500;
    final int TLPT = BLPT + 135, TLDT = 0, TLAT = 1500;
    */

    /* simluate disc-5
    final int PUSH_TIME = 650, POP_TIME = 550;
    final int BLPT = 500, BLDT = 350, BLAT = 500;
    final int TLPT = BLPT + 135, TLDT = 0, TLAT = 500;
    /**/

    final double FULL_REWARD = 1.0, NEG_REWARD = -0.05, ZERO_REWARD = 0.0;

    int simIdx, expIdx, groupIdx, agentIdx, recordIdx, testIdx;
    int expNum, MIN_DISC_NUM, discNum, BLOCK_SIZE, discSelArr[];
    int moveNum, avgMoveNum, minMoveNums[], moveNumArr[][];
    int topCount, botCount, restartCount, decisionCount, levelFlag;
    int MAX_EPISODE_NUM, MAX_MOVE_NUM, EPISODE_BOUND_ARR[];

    int beginPeg, endPeg, beginState[],  endState[];
    int beginIndices[], endIndices[], preIndices[], curIndices[];
    int PEG_VAL_NUM, nodeInfoArr[], pegInfo[];

    int wmActionBL, exActionBL, gsActionBL;
    int curState[], curWMState[], curExState[], curGSState[];
    int action, actionFR, resultFR[];

    // statistics including response time, stop episodes, # of stable episode;
    // # of goal pop, rate of bottom's good move, combination probability
    // both for training and testing; # of bottom move for testing.
    int responseTimeArr[][], stopEpisodeArr[], stableEpisodeNum;
    int TEST_LEN, testBLMoveArr[][];
    int testGoalPopArr[][], testBLGoodMoveArr[][], testCombineProbArr[][];
    int trainGoalPopArr[][], trainBLGoodMoveArr[][], trainCombineProbArr[][];

    double qValue1, qValue2, threshold, feedback, botFeedback, qbpArray[];
    double probFR, probBL, betaBL, betaFR, diffRate, optimalStepNum;

    boolean reinitFlag, stableFlag, solutionFlag;
    boolean checkPermitFlag, blTestFlag, TEST_FLAG;
    boolean wmFlag, exFlag, gsFlag, goalFlag, topProcFlag, frFlagArr[];

    GS     goalStack;
    WorkingMemory workMemory;

    // used to store recommended actions. use Boltzmann distribution to
    // decide probabilitically on an action to be performed.
    QBackPropNet  exNet;      // bottom level network
    StochasticDecider actionDecider;  // action decider
    THFixRuleSet    fixRuleSet;
    Stack         curPegState[];

   public THFrAgent( int num )
   {
       discNum = num;

       qbpArray  =  new double[ EX_NET_INFO[3] ];
       exNet = new QBackPropNet( EX_NET_INFO[1], EX_NET_INFO[2],
                                 EX_NET_INFO[3], EX_NET_INFO[0], DIM_VAL_INFO );

       actionDecider = new StochasticDecider( EX_NET_INFO[3],
                           World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );

       fixRuleSet = new THFixRuleSet( World.TH_TASK, 1 );

       goalStack = new GS( discNum - 1, GS_DIM_NUM, GS_STATE_LEN,
                                  GS_NET_INFO, DIM_VAL_INFO );
       workMemory = new WorkingMemory( discNum, WM_DIM_NUM, WM_STATE_LEN,
                                       WM_NET_INFO, DIM_VAL_INFO );
   }

  /*************************************************************************/
  /**** the following functions are served as simulating the process *******/
  /**********************  of thinking of a subject *************************/
   public int getArrayLen( int idx )
   {
      int len = moveNumArr[ idx ].length;
      return len;
   }

   public void getMoveNums( int[] arr )
   {
      for( int i = 0; i < minMoveNums.length; i++ )
           arr[ i ] = minMoveNums[ i ];
   }

   public int getRestartCount()
   {
      return restartCount;
   }

   public void getStopEpisodeArr( int arr[] )
   {
     for( int i = 0; i < stopEpisodeArr.length; i++ )
          arr[ i ] = stopEpisodeArr[ i ];
   }

   public void getMoveNumArr( int[] arr, int idx )
   {
      for( int i = 0; i < moveNumArr[ idx ].length; i++ )
           arr[ i ] = moveNumArr[ idx ][ i ];
   }

   public void getResponseTimeArr( int[] arr, int idx )
   {
      for( int i = 0; i < responseTimeArr[ idx ].length; i++ )
           arr[ i ] = responseTimeArr[ idx ][ i ] / 1000;
   }

   public void getTrainBLGoodMoveArr( float[] arr, int idx )
   {
      for( int i = 0; i < trainBLGoodMoveArr[ idx ].length; i++ )
           arr[ i ] = (float)trainBLGoodMoveArr[idx][i] / moveNumArr[idx][i];
   }

   public void getTrainGoalPopArr( float[] arr, int idx )
   {
      for( int i = 0; i < trainGoalPopArr[ idx ].length; i++ )
           arr[ i ] = (float)trainGoalPopArr[idx][i];
   }

   public void getTrainCombineProbArr( float[] arr, int idx )
   {
      for( int i = 0; i < trainCombineProbArr[ idx ].length; i++ )
           arr[i] = ( (float)trainCombineProbArr[idx][i]/ENLARGE )/moveNumArr[idx][i];
   }

   public void getTestBLGoodMoveArr( float[] arr, int idx )
   {
      for( int i = 0; i < testBLGoodMoveArr[ idx ].length; i++ )
           arr[ i ] = (float)testBLGoodMoveArr[idx][i] / testBLMoveArr[idx][i];
   }

   public void getTestGoalPopArr( float[] arr, int idx )
   {
      for( int i = 0; i < testGoalPopArr[ idx ].length; i++ )
           arr[ i ] = (float)testGoalPopArr[idx][i];
   }

   public void getTestCombineProbArr( float[] arr, int idx )
   {
      for( int i = 0; i < testCombineProbArr[ idx ].length; i++ )
           arr[i] = ( (float)testCombineProbArr[idx][i]/ENLARGE )/testBLMoveArr[idx][i];
   }

   public void getTestBLMoveArr( int arr[], int idx )
   {
     for( int i = 0; i < testBLMoveArr[ idx ].length; i++ )
          arr[ i ] = testBLMoveArr[ idx ][ i ];
   }

   public void setSimuType( boolean[] flags )
   {
   }

   public void allocateMemory( int eNum, int mNum, int blockSize,
               int[] episode_arr,  double[] betaVals, boolean tFlag )
   {
       expNum = eNum;
       MIN_DISC_NUM = mNum;
       BLOCK_SIZE = blockSize;

       int len = episode_arr.length;
       EPISODE_BOUND_ARR = new int[ len ];
       for( int i = 0; i < len; i++ )
            EPISODE_BOUND_ARR[ i ] = episode_arr[ i ];

       betaBL = betaVals[0];
       betaFR = betaVals[1];
       TEST_FLAG = tFlag;

       resultFR = new int[ 8 ];

       minMoveNums = new int[ expNum ];
       stopEpisodeArr = new int[ expNum ];

       moveNumArr = new int[ expNum ][];
       testBLMoveArr = new int[ expNum ][];
       testGoalPopArr = new int[ expNum ][];
       testBLGoodMoveArr = new int[ expNum ][];
       testCombineProbArr = new int[ expNum ][];
       trainGoalPopArr = new int[ expNum ][];
       trainBLGoodMoveArr = new int[ expNum ][];
       trainCombineProbArr = new int[ expNum ][];
       responseTimeArr = new int[ expNum ][];

       for( int i = 0; i < expNum; i++ )
       {
            int num = EPISODE_BOUND_ARR[ i ];
            moveNumArr[ i ] = new int[ num ];
            trainGoalPopArr[ i ] = new int[ num ];
            trainBLGoodMoveArr[ i ]  = new int[ num ];
            trainCombineProbArr[ i ] = new int[ num ];
            responseTimeArr[ i ] = new int[ num ];
       }

       for( int i = 0; i < expNum; i++ )
       {
            int num = EPISODE_BOUND_ARR[ i ] / BLOCK_SIZE;
            testBLMoveArr[ i ] = new int[ num ];
            testGoalPopArr[ i ] = new int[ num ];
            testBLGoodMoveArr[ i ]  = new int[ num ];
            testCombineProbArr[ i ] = new int[ num ];
       }


       beginState = new int[ EX_INPUT_NUM ];
       endState = new int[ EX_INPUT_NUM ];

       curState = new int[ TOT_INPUT_NUM ];
       curWMState = new int[ WM_INPUT_NUM ];
       curExState = new int[ EX_INPUT_NUM ];
       curGSState = new int[ GS_INPUT_NUM ];
   }

   public void setEnvVarables( int sNo, int gNo, int aNo )
   {
       simIdx = sNo;
       groupIdx = gNo;
       agentIdx = aNo;
   }

   public void reinitAgent()
   {
       exNet.reinit();
       workMemory.reinit( );
       goalStack.reinit( discNum - 1 );

       expIdx = 0;
       initNewExp();
   }

   // set begin state and end state.
   public void setDiscStates()
   {
      discSelArr = new int[ discNum ];
      for( int i = 0; i < discNum; i++ )
           discSelArr[ i ] = i;

      beginIndices = new int[ discNum ];
      endIndices = new int[ discNum ];
      curIndices = new int[ discNum ];
      preIndices = new int[ discNum ];

      for( int i = 0; i < discNum; i++ )
      {
           int size = discSelArr[ i ];
           int idx = discNum - i - 1;
           beginIndices[ idx ] = size * PEG_NUM;
           beginState[ beginIndices[ idx ] ] = 1;
           endIndices[ idx ] = size * PEG_NUM + 2;
           endState[ endIndices[ idx ] ] = 1;
      }
   }

   public void initNewExp()
   {
      avgMoveNum = 0;
      moveNum = recordIdx = 0;
      topCount = botCount = 0;
      decisionCount = 0;
      stableEpisodeNum = 0;
      diffRate = 100;

      blTestFlag = false;
      stableFlag = false;
      reinitFlag = false;

      discNum = expIdx + MIN_DISC_NUM;
      optimalStepNum = Math.pow( 2, discNum ) - 1;
      MAX_MOVE_NUM = (int)optimalStepNum * 100;

      minMoveNums[ expIdx ] = 10000;
      stopEpisodeArr[ expIdx ] = 0;

      MAX_EPISODE_NUM = EPISODE_BOUND_ARR[ expIdx ];
      for( int i = 0; i < MAX_EPISODE_NUM; i++ )
      {
           moveNumArr[ expIdx ][ i ] = 0;
           trainGoalPopArr[ expIdx ][ i ] = 0;
           trainBLGoodMoveArr[ expIdx ][ i ]  = 0;
           trainCombineProbArr[ expIdx ][ i ] = 0;
           responseTimeArr[ expIdx ][ i ] = 0;
      }

      TEST_LEN = MAX_EPISODE_NUM / BLOCK_SIZE;
      for( int i = 0; i < TEST_LEN; i++ )
      {
           testBLMoveArr[ expIdx ][ i ] = 0;
           testGoalPopArr[ expIdx ][ i ] = 0;
           testBLGoodMoveArr[ expIdx ][ i ]  = 0;
           testCombineProbArr[ expIdx ][ i ] = 0;
      }

      setDiscStates();

       //setBLParameters();
       //set temperature.
       //actionDecider.setTemperature( 0.08 );
       actionDecider.setTemperature( 0.1 );
      // set learning rate and momentum.
      // exNet.setLearningRate( 0.3  );
       exNet.setLearningRate( 0.1 );
       //exNet.setMomentum( 0.1 );
       exNet.setMomentum( 0.05 );
   }

   public void setPegState()
   {
     curPegState = new Stack[ 3 ];
     for( int i = 0; i < 3; i++ )
          curPegState[i] = new Stack();

     for( int i = discNum - 1; i >= 0; i-- )
     {
          int[] obj = new int[1];
          obj[0] = discSelArr[i];
          curPegState[0].push( obj );
     }
   }

   public void initNewEpisode()
   {
     if( TEST_FLAG )
     {
         int remainder = (recordIdx + 1) % BLOCK_SIZE;
         if( remainder == 0 )
         {
             if( !blTestFlag )
             {
                 System.out.println("BLB !" );
                 blTestFlag = true;
                 testIdx = ( recordIdx + 1 ) / BLOCK_SIZE - 1;
                 actionDecider.setTemperature( 0.05 );
                 exNet.setLearningRate( 0 );
                 exNet.setMomentum( 0 );
             }
             else
             {
                 System.out.println("BLF !" );
                 blTestFlag = false;
                 testBLMoveArr[ expIdx ][ testIdx ] = moveNum;
                 //actionDecider.setTemperature( 0.08 );
                 actionDecider.setTemperature( 0.1 );
                 // set learning rate and momentum.
                 // exNet.setLearningRate( 0.3  );
                 exNet.setLearningRate( 0.1 );
                 //exNet.setMomentum( 0.1 );
                 exNet.setMomentum( 0.05 );
             }
         }
         else
            blTestFlag = false;
      }

      moveNum = 0;
      topCount = botCount = 0;
      decisionCount = 0;
      solutionFlag = false;

      setPegState();

      // reinit GS.
      goalStack.reinitGS( discNum - 1 );
      // reinit FR.
      fixRuleSet.reinit();

      for( int i = 0; i < beginState.length; i++ )
           curExState[ i ] = beginState[ i ];

      for( int i = 0; i < beginIndices.length; i++ )
      {
           curIndices[ i ] = beginIndices[ i ];
           preIndices[ i ] = beginIndices[ i ];
      }
   }

   public void setBLParameters()
   {
      // set temperature.
      double temp;
      if( recordIdx == 0 )
          temp = 0.1;
      else
      {
          temp = 0.1 / ( 1 + 0.1 * recordIdx );
          if( temp < 0.01 )
              temp = 0.01;
      }
      actionDecider.setTemperature( temp );

      // set learning rate and momentum.
       if( diffRate > 10.0 )
       {
           double rate = 1.0 / ( 2.0 + 0.05 * recordIdx );
           if( rate < 0.05 )
               rate = 0.05;
           exNet.setLearningRate( rate  );
           double mom = 0.5 / ( 2.0 + 0.05 * recordIdx );
           exNet.setMomentum( mom );
       }
       else
       {
           double rate = 0.5 / ( 10.0 + 0.5 * recordIdx );
           if( rate < 0.005 )
               rate = 0.005;
           exNet.setLearningRate( rate );
           double mom = 0.25 / ( 10.0 + 0.5 * recordIdx );
           exNet.setMomentum( mom );
       }
   }

   public void solveProblems()
   {
     expIdx = 0;
     while( expIdx < expNum )
     {
        restartCount = 0;
        initNewExp();
        while( !stableFlag )
        {
            initNewEpisode();
            while( !solutionFlag )
            {
                if( blTestFlag )
                {
                    if( moveNum == MAX_MOVE_NUM )
                    {
                        solutionFlag = true;
                        continue;
                    }
                    if( decisionCount == ( 2 * MAX_MOVE_NUM ) )
                    {
                        moveNum = MAX_MOVE_NUM;
                        solutionFlag = true;
                        continue;
                    }
                }

                if( moveNum > MAX_MOVE_NUM )
                {
                    System.out.print( "move number too big, " );
                    reinitFlag = true;
                    break;
                }

                boolean flag = decideAction();
                if( !flag )
                {
                    reinitFlag = true;
                    break;
                }

                convToAction();
                checkAction();
                update();

            }// end of while( !solutionFlag )

            if( reinitFlag )
            {
                reinitAgent();
                restartCount++;
                System.out.println( "reinit!! " );
                continue;
            }

            if( !blTestFlag )
                checkStable();

            if( TEST_FLAG )
            {
                if( !blTestFlag )
                    recordIdx++;
                if( recordIdx == MAX_EPISODE_NUM )
                    stableFlag = true;
            }
            else
            {
                recordIdx++;
                if( stableFlag )
                    continue;
                if( recordIdx == MAX_EPISODE_NUM )
                {
                    reinitAgent();
                    restartCount++;
                    System.out.println( " not stable, reinit!! " );
                }
            }
       }// end of while( !stableFlag )

       expIdx++;
     }
   }

   // get current state including WM, EX, GS inputs.
   public boolean getCurState()
   {
     boolean flag = goalStack.getActiveItem( curGSState );
     if( !flag )
     {
        System.out.println( "The stack is empty! " );
        return false;
     }

     workMemory.getWMState( curWMState );

     int offset = 0;
     for( int i = 0; i < curWMState.length; i++ )
          curState[ offset + i ] = curWMState[ i ];

     offset += curWMState.length;
     for( int i = 0; i < curExState.length; i++ )
          curState[ offset + i ] = curExState[ i ];

     offset += curExState.length;
     for( int i = 0; i < curGSState.length; i++ )
          curState[ offset + i ] = curGSState[ i ];

      return true;
   }

   // decide an action from the top level or bottom level using variable
   // percentage combination.
   public boolean decideAction()
   {
        boolean flag = getCurState();
        if( !flag )
            return false;

        if( topProcFlag )
        {
           // recommended action from the top level
           for( int i = 0; i < 8; i++ )
                resultFR[ i ] = 0;
           fixRuleSet.getResult( curWMState, curExState, curGSState, resultFR );
           return true;
        }

        double succRateFR = fixRuleSet.getSuccRate();
        double succRateBL = exNet.getSuccRate();
        double pFR = betaFR * succRateFR;
        double pBL = betaBL * succRateBL;
        probFR = pFR / ( pFR + pBL );
        probBL = pBL / ( pFR + pBL );
        double val = Math.random();

        //get combined percentage of both levels
        if( blTestFlag )
            testCombineProbArr[ expIdx ][ testIdx ] += (int)( probFR * ENLARGE );
        else
            trainCombineProbArr[ expIdx ][ recordIdx ] += (int)( probFR * ENLARGE );

        if( blTestFlag )
        {
            levelFlag = DOWN_LEVEL;
            botCount++;
        }
        else
        {
            if( val <= probFR )
            {
                levelFlag = TOP_LEVEL;
                topCount++;
                responseTimeArr[ expIdx ][ moveNum ] = TLPT + TLDT;
            }
            else
            {
                levelFlag = DOWN_LEVEL;
                botCount++;
            }
        }

        decisionCount++;

        // get the Q values from the bottom level
        exNet.setCurState( curState );
        exNet.setInput();
        exNet.forwardPass();
        exNet.getQvals( qbpArray );
        actionDecider.setElements( qbpArray );
        // recommended action from the bottom level
        exActionBL = actionDecider.getRecommendation();
        gsActionBL = goalStack.getAction( curState );
        wmActionBL = workMemory.getAction( curState );

        // recommended action from the top level
        fixRuleSet.setTopFlags();
        for( int i = 0; i < 8; i++ )
             resultFR[ i ] = 0;
        fixRuleSet.getResult( curWMState, curExState, curGSState, resultFR );

        topProcFlag = true;
        checkPermitFlag = false;
        return true;
   }

   //
   public void convToAction()
   {
       frFlagArr = new boolean[ 3 ];

       switch( resultFR[0] )
       {
          case WM_ACTION:
                        actionFR = -1;
                        fixRuleSet.getTopFlags( frFlagArr );
                        workMemory.setFlags( frFlagArr );
                        break;
          case EX_ACTION:
                        actionFR = resultFR[1] * 2;
                        actionFR += ( resultFR[2] - resultFR[1] + 2 ) % 3;
                        if( levelFlag == TOP_LEVEL )
                            responseTimeArr[ expIdx ][ moveNum ] += TLAT;
                        break;
          case GS_ACTION:
                        actionFR = -1;
                        goalFlag = fixRuleSet.getGoalFlag();
                        if( resultFR[ 1 ] == 1 && !goalFlag )
                        {
                            if( blTestFlag )
                                testGoalPopArr[ expIdx ][ testIdx ] += 1;
                            else
                                trainGoalPopArr[ expIdx ][ recordIdx ] += 1;
                        }
                        if( levelFlag == TOP_LEVEL )
                        {
                            if( resultFR[ 1 ] == 0 )
                                responseTimeArr[ expIdx ][ moveNum ] += PUSH_TIME;
                            if( resultFR[ 1 ] == 1 )
                                responseTimeArr[ expIdx ][ moveNum ] += POP_TIME;
                        }
                        break;
       }

       if( levelFlag == DOWN_LEVEL )
           action = exActionBL;
       else
           action = actionFR;

       if( actionFR != -1 )
           checkPermitFlag = true;
   }

   // check if the chosen action is legal.
   public void checkAction( )
   {
      if( levelFlag == DOWN_LEVEL )
      {
         // check WM action.
         wmFlag = workMemory.checkAction( wmActionBL );
         // perform GS action.
         gsFlag = goalStack.checkAction( gsActionBL );
      }
      else
      {
         // check WM action.
         wmFlag = workMemory.checkAction( resultFR );
         // perform GS action.
         gsFlag = goalStack.checkAction( resultFR );
      }

      if( action == -1 )
          return;

      if( !checkPermitFlag )
          return;

      beginPeg = action / 2;
      endPeg = ( beginPeg + 1 + action % 2 ) % 3;

  /******* the following is to check if the action is an illegal move.********/
      // check if the beginning peg is empty.
      if( curPegState[ beginPeg ].isEmpty() )
      {
          solutionFlag = false;
          feedback = NEG_REWARD;
          exFlag = false;
          return;
      }

      if( moveNum < responseTimeArr[ expIdx ].length )
      {
          if( levelFlag == DOWN_LEVEL && !blTestFlag )
              responseTimeArr[ expIdx ][ moveNum ] = BLPT + BLDT + BLAT;
      }

      moveNum++;

      // check if item1 > item2( an illegal move ).
      int[] item1 = new int[1];
      item1 = (int[])curPegState[ beginPeg ].peek();
      int[] item2 = new int[1];
      if( !curPegState[ endPeg ].isEmpty() )
          item2 = (int[])curPegState[ endPeg ].peek();
      else
          item2[0] = MAX_ITEM_VAL;

      if( item1[0] > item2[0] )
      {
          solutionFlag = false;
          feedback = NEG_REWARD;
          exFlag = false;
          return;
      }

      exFlag = true;
   }

   public void update()
   {
      if( gsFlag )
          goalStack.performAction( resultFR );
      if( wmFlag )
          workMemory.performAction( resultFR );

      if( !blTestFlag )
      {
          goalStack.setLevelFlag( levelFlag );
          goalStack.update( gsFlag );
          workMemory.setLevelFlag( levelFlag );
          workMemory.update( wmFlag );
      }

      if( action == -1 )
          return;

      if( actionFR == -1 )
          return;

      if( exFlag )
      {
          performAction();
          int[] item1 = new int[1];
          item1 = (int[])curPegState[ endPeg ].peek();
          fixRuleSet.setCurMovedDisc( item1[0] );
          if( !blTestFlag )
              legalUpdate();
      }
      else
      {
          if( blTestFlag )
              testCombineProbArr[ expIdx ][ testIdx ] -= (int)( probFR * ENLARGE );
          else
          {
              trainCombineProbArr[ expIdx ][ recordIdx ] -= (int)( probFR * ENLARGE );
              illegalUpdate();
          }
      }

      topProcFlag = false;
   }

   // perform the EXternal action.
   public void performAction()
   {
      // move the specific disc.
      int[] item1 = new int[1];
      item1 = (int[])curPegState[ beginPeg ].pop();
      curPegState[ endPeg ].push( item1 );

      for( int i = 0; i < discNum; i++ )
           preIndices[ i ] = curIndices[ i ];

      int idx = 0;
      for( int i = 0; i < PEG_NUM; i++ )
        for( int j = 0; j < curPegState[i].size(); j++ )
        {
             int[] item = (int[])curPegState[i].elementAt( j );
             curIndices[ idx++ ] = item[0] * PEG_NUM + i;
        }

      // update the current state.
      for( int i = 0; i < discNum; i++ )
           curExState[ preIndices[ i ] ] = 0;

      for( int i = 0; i < discNum; i++ )
           curExState[ curIndices[ i ] ] = 1;

      // check if a solution has been found so far.
      for( int i = 0; i < discNum; i++ )
        if( curIndices[i] != endIndices[i] )
        {
            solutionFlag = false;
            getNonNegReward();
            return;
        }

      // a solution has been found.
      solutionFlag = true;
      feedback = FULL_REWARD;
   }

  // check if the current move is a key move.
   public void getNonNegReward()
   {
       if( levelFlag == TOP_LEVEL )
       {
           feedback = 0.1;
           if( action == exActionBL )
           {
               botFeedback = 0.1;
               if( blTestFlag )
                   testBLGoodMoveArr[ expIdx ][ testIdx ] += 1;
               else
                   trainBLGoodMoveArr[ expIdx ][ recordIdx ] += 1;
           }
           else
               botFeedback = 0.0;
       }
       else if( action == actionFR )
            {
                feedback = 0.1;
                botFeedback = 0.1;
                if( blTestFlag )
                    testBLGoodMoveArr[ expIdx ][ testIdx ] += 1;
                else
                    trainBLGoodMoveArr[ expIdx ][ recordIdx ] += 1;
            }
            else
            {
                feedback = 0.0;
                botFeedback = 0.0;
            }
   }

   // update both top and bottom level when an illegal action is chosen
   public void illegalUpdate()
   {
      //obtain the current state QVaule
      qValue1 = exNet.getQVal( action );

      if( levelFlag == DOWN_LEVEL )
          exNet.update( action, feedback, qValue1 );
      else
          System.out.println( "delete a rule !!!!" );
   }

   // update both top and bottom level when an legal action is chosen
   public void legalUpdate()
   {
        //obtain the current state QVaule
        qValue1 = exNet.getQVal( action );

       //obtain next state QValue2.
        boolean flag = getCurState();
        if( !flag )
            return;

	exNet.setCurState( curState );
	exNet.setInput();
        exNet.forwardPass();
        qValue2 = exNet.getMaxQVal();

/******************************************************************************/
        // update the QBP network.
        if( botFeedback > 0 )
            exNet.updateMatchCount( true );
        else if( botFeedback < 0 )
                 exNet.updateMatchCount( false );

        exNet.update( action, feedback, qValue1, qValue2 );
/******************************************************************************/
   }

   // check if the current solution is stable.
   public void checkStable( )
   {
      moveNumArr[ expIdx ][ recordIdx ] = moveNum;
      diffRate = ((double)moveNum - optimalStepNum ) / optimalStepNum;

      if( recordIdx % ( BLOCK_SIZE / 2 ) == 0 )
      {
          System.out.print( "G = " );
          System.out.print( groupIdx );
          System.out.print( ", A = " );
          System.out.print( agentIdx );
          System.out.print( ", D = " );
          System.out.print( discNum );
          System.out.print( ", R = " );
          System.out.println( recordIdx );
          //System.out.print( " compared with optimal: ");
          //System.out.println( diffRate );
      }

/******************************************************************************/
      // original
      double[] threshArr = { 0.6, 0.6, 0.6, 0.6 };

      // for disc-4
      //double[] threshArr = { 0.1, 0.1, 0.1, 0.1 };

      // for disc-5
     // double[] threshArr = { 0.01, 0.01, 0.01, 0.01 };

      double baseRate = threshArr[ expIdx ];
/******************************************************************************/

      if( TEST_FLAG )
      {
          minMoveNums[ expIdx ] = moveNum;
          stopEpisodeArr[ expIdx ] = MAX_EPISODE_NUM - 1;
      }
      else
      {
          if( diffRate <= baseRate )
          {
             int v2, v1 = moveNumArr[ expIdx ][ recordIdx ];
             if( recordIdx > 0 )
                 v2 = moveNumArr[ expIdx ][ recordIdx - 1 ];
             else
                 v2 = 100000;

             stableEpisodeNum++;
             avgMoveNum += moveNum;

             if( stableEpisodeNum == STABLE_EPISODE_NUM )
             {
                 System.out.println( " Success! ");
                 stableFlag = true;
                 int val = avgMoveNum / STABLE_EPISODE_NUM;
                 minMoveNums[ expIdx ] = val;

                 // don't append data.
                 //for( int i = recordIdx; i < MAX_EPISODE_NUM; i++ )
                 //     moveNumArr[ expIdx ][ i ] = val;
                 // store the stop point.
                 stopEpisodeArr[ expIdx ] = recordIdx;
             }
          }
          else
          {
              if( stableEpisodeNum > 0 )
              {
                  stableEpisodeNum = 0;
                  avgMoveNum = 0;
              }
              stableFlag = false;
          }
      }
  }
}
