/* the LETTER COUNTING AGENT FOR LETTER COUNTING TASK  */
package tasksimulators.oldsimu;
import java.lang.*;
import java.util.Vector;
import java.util.Stack;
import clarion.common.*;
import clarion.tools.*;
import clarion.acs.*;

public class THRerAgent {

    final int NO_MERGE = 0, TEN_MERGE = 1, EACH_MERGE = 2;
    final int ORIGIN_PEG = 0, MIDDLE_PEG = 1, GOAL_PEG = 2;
    final int TOP_LEVEL = 0, DOWN_LEVEL = 1;
    final int MAX_ITEM_VAL = 100, STABLE_EPISODE_NUM = 3;

    // used for reward in QBP networks.
    final double FULL_REWARD = 1.0;
    final double NEG_REWARD = -0.1;
    final double ZERO_REWARD = 0.0;

    // used for combination of two levels.
    final double[][] RER_SUCC_ARR =
                { { 1.0, 1.0, 1.0 }, { 5.0, 5.0, 5.0 } };

   // the location map for the rule state.
    int pegMapArr[][] = { { 0, 2, 1 }, { 1, 0, 2 } };
    int PEG_LOC_OFFSETS[][];

    // parameters for different QBP nwtworks.
    final int TOTAL_NET_INPUT_NUM = 180,  EX_TL_INPUT_NUM = 108;
    final int EX_DIM_NUM = 18, GS_DIM_NUM = 6, WM_DIM_NUM = 6;
    final int WM_STATE_LEN = 36, EX_STATE_LEN = 108, GS_STATE_LEN = 36;
    final int EX_NET_INFO_ARR[] = { 30, 180, 80, 7 };
    final int GS_NET_INFO_ARR[] = { 30, 180, 80, 14};
    final int WM_NET_INFO_ARR[] = { 30, 180, 80, 18};

    final int EX_DO_NOTHING_ACTION = 6;
    final int GS_DO_NOTHING_ACTION = 13;

    // variables for controlling, recording, environment-setting.
    int MAX_EPISODE_NUM, MAX_MOVE_NUM, EPISODE_BOUND_ARR[];
    int MAX_DISC_NUM = 6, MIN_DISC_NUM, PEG_VAL_NUM;
    int simIdx, expIdx, groupIdx, agentIdx, recordIdx, stageIdx, postStageIdx;
    int expNum, discNum, maxDiscLoc, discSelArr[], dimValNumArr[];
    int moveNum, avgMoveNum, minMoveNums[], moveNumArr[][];
    int beginPeg, endPeg, levelFlag, stableEpisodeNum;
    int ruleRecorder[][], nodeInfoArr[], pegInfo[];

    // used for some statistics of the agent.
    int topCount, botCount, restartCount;
    int stopEpisodeArr[];

    // the following is used for record the states of EX, GS and WM.
    int action, curState[], directMove;
    int curExState[],curWMState[],curGSState[];
    int wmAction, exAction, gsAction;
    int wmActionBL, exActionBL, gsActionBL;
    int wmActionTL, exActionTL, gsActionTL;

    // used for EXTERNAL.
    int beginExState[],  endExState[];
    int beginExIndices[], endExIndices[];
    int preExIndices[], curExIndices[];

    double diffRate, optimalStepNum;
    double exQValue1, exQValue2;
    double gsQvalue1, gsQvalue2;
    double threshold, feedback;
    double betaBL, betaTL, qbpArray[];

    boolean RER_FLAG, UTL_FLAG, WM_FLAG, GS_FLAG, FR_FLAG, gsFlag;
    boolean actionLegalFlag, exActionLegalFlag;
    boolean wmActionLegalFlag, gsActionLegalFlag;
    boolean reinitFlag, stableFlag, solutionFlag, gsChosenFlag;

    // used to store recommended actions. use Boltzmann distribution to
    // decide probabilitically on an action to be performed.
    QBackPropNet  exNet;      // bottom level network of external
    StochasticDecider exActionDecider;  // action decider
    T_RerRuleSet    exT_RerRuleSet;  // top level rule set.
    RerGS         rerGS;
    RerWM         rerWM;

    StringBuffer  rerRuleStrings[][];
    Stack         curPegState[];

    // used for RER only.
    public THRerAgent( int pvNum )
    {
           PEG_VAL_NUM = pvNum;
           threshold = World.TASK_PARA_ARRAY[ World.HIGH_RER_THRESHOLD ];

           PEG_LOC_OFFSETS = new int[2][3];
           PEG_LOC_OFFSETS[0][0] = 0;
           PEG_LOC_OFFSETS[0][1] = PEG_VAL_NUM;
           PEG_LOC_OFFSETS[0][2] = -PEG_VAL_NUM;
           PEG_LOC_OFFSETS[1][0] = PEG_VAL_NUM;
           PEG_LOC_OFFSETS[1][1] = -PEG_VAL_NUM;
           PEG_LOC_OFFSETS[1][2] = 0;

           dimValNumArr = new int[ GS_NET_INFO_ARR[0] ];
           for( int i = 0; i < dimValNumArr.length; i++ )
                dimValNumArr[ i ] = MAX_DISC_NUM;

           qbpArray  =  new double[ EX_NET_INFO_ARR[3] ];
	   exNet = new QBackPropNet( EX_NET_INFO_ARR[1], EX_NET_INFO_ARR[2],
                                      EX_NET_INFO_ARR[3] );

	   exActionDecider = new StochasticDecider( EX_NET_INFO_ARR[3],
                                World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );
   }

   public double[][] getInitIHWeights()
   {
      return exNet.getInitInputToHiddenWeights();
   }

   public double[][] getInitHOWeights()
   {
      return exNet.getInitHiddenToOutputWeights();
   }

   public void setInitWeights( double[][] ItoHWeights, double[][] HtoOWeights  )
   {
      exNet.setInitInputToHiddenWeights( ItoHWeights );
      exNet.setInitHiddenToOutputWeights( HtoOWeights );
   }

   public double[][] getIHWeights()
   {
      return exNet.getInputToHiddenWeights();
   }

   public double[][] getHOWeights()
   {
      return exNet.getHiddenToOutputWeights();
   }

   public void getThresholds( double hThresholds[], double oThresholds[] )
   {
      exNet.getHiddenThresholds( hThresholds );
      exNet.getOutputThresholds( oThresholds );
   }

  /*************************************************************************/
  /**** the following functions are served as simulating the process *******/
  /**********************  of thinking ofa subject *************************/
   public void getMoveNums( int[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           arr[ i ] = minMoveNums[ i ];
   }

   public int getArrayLen( int idx )
   {
      int len = moveNumArr[ idx ].length;
      return len;
   }

   public void getMoveNumArr( int[] arr, int idx )
   {
      for( int i = 0; i < moveNumArr.length; i++ )
           arr[ i ] = moveNumArr[ idx ][ i ];
   }

   public int getRestartCount()
   {
      return restartCount;
   }

   public void getStopEpisodeArr( int arr[] )
   {
     for( int i = 0; i < arr.length; i++ )
          stopEpisodeArr[ i ] = arr[ i ];
   }

   public void setThreshold( double th )
   {
       threshold = th;
       if( GS_FLAG );
           rerGS.setThreshold( th );
       if( WM_FLAG )
           rerWM.setThreshold( th );
   }

   public void setRERThresholds( double th1, double th2 )
   {
       exT_RerRuleSet.setThreshold( th1, th2 );
       if( GS_FLAG );
           rerGS.setThreshold( th1, th2 );
       if( WM_FLAG )
           rerWM.setThreshold( th1, th2 );
   }

   public void setSimuType( boolean[] flags )
   {
       RER_FLAG = flags[0];
       UTL_FLAG = flags[1];
       WM_FLAG = flags[2];
       GS_FLAG = flags[3];
       FR_FLAG = flags[4];

       nodeInfoArr = new int[ 5 ];
       nodeInfoArr[ 0 ] = EX_DIM_NUM;
       nodeInfoArr[ 1 ] = EX_DIM_NUM;
       nodeInfoArr[ 2 ] = EX_TL_INPUT_NUM;
       nodeInfoArr[ 3 ] = EX_NET_INFO_ARR[2];
       nodeInfoArr[ 4 ] = EX_NET_INFO_ARR[3];

       if( RER_FLAG )
       {
           exT_RerRuleSet = new T_RerRuleSet( 1, NO_MERGE, nodeInfoArr, true );
           exT_RerRuleSet.setFullFlag( false );
       }

       if( GS_FLAG )
       {
           rerGS = new RerGS( discNum - 1, GS_DIM_NUM, GS_STATE_LEN,
                              GS_NET_INFO_ARR, dimValNumArr );
           rerGS.setFullFlag( true );
       }

       if( WM_FLAG )
       {
           rerWM = new RerWM( discNum, WM_DIM_NUM, WM_STATE_LEN,
                              WM_NET_INFO_ARR, dimValNumArr );
           rerWM.setFullFlag( true );
       }

      rerRuleStrings = new StringBuffer[3][3];
      for ( int i = 0; i < 3; i++ )
         for ( int j = 0; j < 3; j++ )
               rerRuleStrings[i][j] = new StringBuffer( "" );
   }

   public void allocateMemory( int eNum, int mNum, int[] episode_arr,
                               double[] betaVals, boolean tFlag )
   {
       expNum = eNum;
       MIN_DISC_NUM = mNum;

       int len = episode_arr.length;
       EPISODE_BOUND_ARR = new int[ len ];
       for( int i = 0; i < len; i++ )
            EPISODE_BOUND_ARR[ i ] = episode_arr[ i ];

       curState = new int[ TOTAL_NET_INPUT_NUM ];
       curWMState = new int[ WM_STATE_LEN ];
       curExState = new int[ EX_STATE_LEN ];
       curGSState = new int[ GS_STATE_LEN ];

       beginExState = new int[ EX_STATE_LEN ];
       endExState = new int[ EX_STATE_LEN ];

       moveNumArr = new int[ expNum ][];
       minMoveNums = new int[ expNum ];
       stopEpisodeArr = new int[ expNum ];

       betaBL = betaVals[0];
       betaTL = betaVals[1];
   }

   public void setEnvVarables( int sNo, int gNo, int aNo )
   {
       simIdx = sNo;
       groupIdx = gNo;
       agentIdx = aNo;
   }

   public void reinitAgent()
   {
       exNet.reinit();
       exT_RerRuleSet.reinit();
       expIdx = 0;
       initNewExp();
   }

   public void setValidDimNums( int num )
   {
       exT_RerRuleSet.setValidDimNum( num );
       rerGS.setValidNum( num + GS_DIM_NUM + WM_DIM_NUM );
       rerWM.setValidNum( num + GS_DIM_NUM + WM_DIM_NUM );
   }

   // set current experiemnt's begin state and end state.
   public void initExStates()
   {
      discSelArr = new int[ discNum ];
      for( int i = 0; i < discNum; i++ )
           discSelArr[ i ] = i;

      for( int i = 0; i < beginExState.length; i++ )
      {
           beginExState[ i ] = 0;
           endExState[ i ] = 0;
      }

      beginExIndices = new int[ discNum ];
      endExIndices = new int[ discNum ];
      curExIndices = new int[ discNum ];
      preExIndices = new int[ discNum ];

      for( int i = 0; i < discNum; i++ )
      {
           int size = discSelArr[discNum - i - 1];
           beginExIndices[i] = PEG_VAL_NUM - (i + 1) * MAX_DISC_NUM + size;
           beginExState[ beginExIndices[i] ] = 1;
           endExIndices[i] = EX_STATE_LEN - (i + 1) * MAX_DISC_NUM + size;
           endExState[ endExIndices[i] ] = 1;
      }

      for( int i = 0; i < beginExState.length; i++ )
           curExState[ i ] = beginExState[ i ];
   }

   public void initNewExp()
   {
      moveNum = 0;
      recordIdx = 0;
      avgMoveNum = 0;
      topCount = 0;
      botCount = 0;
      stableEpisodeNum = 0;
      diffRate = 100;
      minMoveNums[ expIdx ] = 10000;

      stableFlag = false;
      reinitFlag = false;

      discNum = expIdx + MIN_DISC_NUM;
      optimalStepNum = Math.pow( 2, discNum ) - 1;
      MAX_MOVE_NUM = (int)optimalStepNum * 100;
      MAX_EPISODE_NUM = EPISODE_BOUND_ARR[ expIdx ];
      moveNumArr[ expIdx ] = new int[ MAX_EPISODE_NUM ];
      ruleRecorder = new int[ (int)optimalStepNum * 5 ][ discNum + 2 ];

      if( GS_FLAG )
          rerGS.initDiscLocs( discNum );

      if( WM_FLAG )
          rerWM.initDiscLocs( discNum );

      setValidDimNums( discNum );
      initExStates();

      if( expIdx == 0 )
      {
          curExState[ PEG_VAL_NUM - MAX_DISC_NUM ] = 1;
          int GOAL_ACTION = 1, DIM_IDX = 5;
          exT_RerRuleSet.createKnownRule( curExState, GOAL_ACTION, DIM_IDX );
      }

       //setBLParameters();
       // set temperature.
       exActionDecider.setTemperature( 0.08 );
      // set learning rate and momentum.
       exNet.setLearningRate( 0.3  );
       exNet.setMomentum( 0.1 );
   }

   public void initPegState()
   {
     curPegState = new Stack[ 3 ];
     for( int i = 0; i < 3; i++ )
          curPegState[i] = new Stack();

     for( int i = discNum - 1; i >= 0; i-- )
     {
          int[] obj = new int[1];
          obj[0] = discSelArr[i];
          curPegState[0].push( obj );
     }
   }

   public void initNewEpisode()
   {
       moveNum = 0;
       topCount = 0;
       botCount = 0;

       // modified on 02/01/02
       stageIdx = 0;
       postStageIdx = 0;

       solutionFlag = false;
       gsChosenFlag = false;

       for( int i = 0; i < ruleRecorder.length; i++ )
         for( int j = 0; j < ruleRecorder[0].length; j++ )
              ruleRecorder[i][j] = 0;

       for( int i = 0; i < beginExState.length; i++ )
            curExState[ i ] = beginExState[ i ];
       for( int i = 0; i < beginExIndices.length; i++ )
       {
            curExIndices[ i ] = beginExIndices[ i ];
            preExIndices[ i ] = beginExIndices[ i ];
       }

       initPegState();

       //initialize goal stack.
       rerGS.reinitGS( discNum - 1 );
       //initizlize working memory.
       rerWM.reinit();
   }

   public void setBLParameters()
   {
      // set temperature.
      double temp = 0.1;
      /*
      if( recordIdx == 0 )
          temp = 0.1;
      else
      {
          temp = 0.1 / ( 1 + 0.1 * recordIdx );
          if( temp < 0.01 )
              temp = 0.01;
      }
      */

      exActionDecider.setTemperature( temp );

      // set learning rate and momentum.
       if( diffRate > 10.0 )
       {
           double rate = 1.0 / ( 2.0 + 0.05 * recordIdx );
           if( rate < 0.05 )
               rate = 0.05;
           exNet.setLearningRate( rate  );
           double mom = 0.5 / ( 2.0 + 0.05 * recordIdx );
           exNet.setMomentum( mom );
       }
       else
       {
           double rate = 0.5 / ( 10.0 + 0.5 * recordIdx );
           if( rate < 0.005 )
               rate = 0.005;
           exNet.setLearningRate( rate );
           double mom = 0.25 / ( 10.0 + 0.5 * recordIdx );
           exNet.setMomentum( mom );
       }
   }

   public void solveProblems()
   {
     expIdx = 0;
     while( expIdx < expNum )
     {
        restartCount = 0;
        initNewExp();
        while( !stableFlag )
        {
            initNewEpisode();
            while( !solutionFlag )
            {
                if( moveNum > MAX_MOVE_NUM )
                {
                    System.out.print( "move number too big, " );
                    reinitFlag = true;
                    break;
                }

                // modified on 02/01/02.
                // check the current stage.
                //stageIdx = getStageInfo( curExIndices );
                stageIdx = postStageIdx;
                if( stageIdx == - 1 )
                {
                    System.out.print( "stage error,  " );
                    reinitFlag = true;
                    break;
                }

                decideActions();
                checkActions();
                performActions();
                updates();

                if( (moveNum % 100) == 0 )
                {
                     System.out.print( "moveNum = " );
                     System.out.println( moveNum );
                }
            }// end of while( !solutionFlag )

            if( reinitFlag )
            {
               reinitAgent();
               restartCount++;
               System.out.println( "reinit!! " );
               continue;
            }

            checkStable();
            recordIdx++;
            if( stableFlag )
                continue;

            if( recordIdx == MAX_EPISODE_NUM )
            {
                reinitAgent();
                restartCount++;
                System.out.println( " not stable, reinit!! " );
            }
       }// end of while( !stableFlag )

       expIdx++;
     }
     appendRules();
   }

   public int getStageInfo( int[] indices )
   {
      int stage = 0;
      if( discNum == 1 )
      {
          stage = 0;
          return stage;
      }

      // search the location of the max disc.
      for( int i = 0; i < discNum; i++ )
      {
          int discNo = indices[ i ] % MAX_DISC_NUM;
          if( discNo == discSelArr[ discNum - 1 ] )
          {
              maxDiscLoc = i;
              break;
          }
      }

      // check current stage.
      int pegIdx = indices[ maxDiscLoc ] / PEG_VAL_NUM;
      switch( pegIdx )
      {
        case 0: if( indices[ 1 ] < PEG_VAL_NUM ||
                    indices[ discNum - 1 ] >= 2 * PEG_VAL_NUM )
                {
                    stage = 0;
                }
                else
                    stage = 2;
                break;
        case 1: stage = -1;
                break;
        case 2: stage = 1;
                break;
      }

      return stage;
   }

   public boolean getCurState()
   {
     boolean flag = rerGS.getState( curGSState );
     if( !flag )
     {
         System.out.println( "The stack is empty! " );
         rerGS.reinit( discNum );
         rerGS.getState( curGSState );
     }

     rerWM.getWMState( curWMState );

     int offset = 0;
     for( int i = 0; i < curExState.length; i++ )
          curState[ offset + i ] = curExState[ i ];

     offset += curExState.length;
     for( int i = 0; i < curGSState.length; i++ )
          curState[ offset + i ] = curGSState[ i ];

     offset += curGSState.length;
     for( int i = 0; i < curWMState.length; i++ )
          curState[ offset + i ] = curWMState[ i ];

      return true;
   }

   // GET EXTERNAL BL ACTION.
   public int getExActionBL()
   {
        // get the Q values from the bottom level
        exNet.setCurState( curState );
        exNet.setInput();
        exNet.forwardPass();
	exNet.getQvals( qbpArray );

        // recommended action from the bottom level
        exActionDecider.setElements( qbpArray );
        return exActionDecider.getRecommendation();
   }

   // decide an action from the top level or bottom level using variable
   // percentage combination.
   public boolean decideActions()
   {
        // get current state including all of the three parts.
        boolean flag = getCurState();
        if( !flag )
            return false;

        // set previous states.               // 12/06/01
        rerGS.setPreState( curState );
        rerWM.setPreState( curState );

        // set current disc locations.
        rerGS.setDiscLocs( curExIndices );
        rerWM.setDiscLocs( curExIndices );

        // if GS is in charge of move-making.
        if( gsFlag )
        {
            gsActionBL = rerGS.getActionBL( curState );
            gsActionTL = rerGS.getActionTL( curState );

            double succRateTL = RER_SUCC_ARR[ groupIdx ][ 1 ];
            double succRateBL = rerGS.getBLSuccRate();
            double pTL = betaTL * succRateTL;
            double pBL = betaBL * succRateBL;
            double probTL = pTL / ( pTL + pBL );
            double probBL = pBL / ( pTL + pBL );
            double levelVal = Math.random();
            if( levelVal <= probTL && gsActionTL != -1 )
            {
                levelFlag = TOP_LEVEL;
                gsAction = gsActionTL;
                //topCount++;
            }
            else
            {
                levelFlag = DOWN_LEVEL;
                gsAction = gsActionBL;
                //botCount++;
            }
            return true;
        }

        // recommended action from the BOTTOM level
        exActionBL = getExActionBL();
        gsActionBL = rerGS.getActionBL( curState );
        wmActionBL = rerWM.getActionBL( curState );

        // recommended action from the TOP level
        int[] exRuleState = new int[ curExState.length ];
        convToExRuleState( exRuleState, curExIndices, stageIdx, true );
        exActionTL = exT_RerRuleSet.getAction( exRuleState );
        gsActionTL = rerGS.getActionTL( curState );
        wmActionTL = rerWM.getActionTL( curState );

        // decide an action is taken from TL or BL.
        double rerSucc, rerSuccEx, rerSuccGS, rerSuccWM;
        double blSucc, blSuccEx, blSuccGS, blSuccWM;

        rerSuccEx = RER_SUCC_ARR[ groupIdx ][ 0 ];
        rerSuccGS = RER_SUCC_ARR[ groupIdx ][ 1 ];
        rerSuccWM = RER_SUCC_ARR[ groupIdx ][ 2 ];
        rerSucc = rerSuccEx + rerSuccGS + rerSuccWM;
        blSuccEx = exNet.getSuccRate();
        blSuccGS = rerGS.getBLSuccRate();
        blSuccWM = rerWM.getBLSuccRate();
        blSucc = blSuccEx + blSuccGS + blSuccWM;

        double pTL = betaTL * rerSucc;
        double pBL = betaBL * blSucc;
        double probTL = pTL / ( pTL + pBL );
        double probBL = pBL / ( pTL + pBL );
        double levelVal = Math.random();

        if( levelVal <= probTL )
        {
            levelFlag = TOP_LEVEL;
            wmAction = wmActionTL;
            exAction = exActionTL;
            gsAction = gsActionTL;
            topCount++;
        }
        else
        {
            levelFlag = DOWN_LEVEL;
            wmAction = wmActionBL;
            exAction = exActionBL;
            gsAction = gsActionBL;
            botCount++;
        }

        // decide an move is decided by EX or GS.
        if( ( exActionBL == EX_DO_NOTHING_ACTION && levelFlag == DOWN_LEVEL )
            || ( ( exActionTL == -1 ) && levelFlag == TOP_LEVEL ) )
            gsFlag = true;
        else
            gsFlag = false;

        // set the actual action.
        if( !gsFlag )
        {
            if( levelFlag == TOP_LEVEL )
                action = exActionTL;
            else
                action = exActionBL;
        }
        else
            rerGS.setLevelFlag( levelFlag );

        return true;
   }

   // The following three are used to convert the state into rule conditions.
   public void convToExRuleState( int[] ruleState, int[] indices, int stage,
                                  boolean delFlag )
   {
      for( int i = 0; i < ruleState.length; i++ )
           ruleState[ i ] = 0;

      if( !delFlag || discNum == MIN_DISC_NUM )
      {
         for( int i = 0; i < discNum; i++ )
         {
             int offset = ( indices[ i ] / MAX_DISC_NUM ) * MAX_DISC_NUM;
             int discNo = indices[ i ] % MAX_DISC_NUM;
             int idx = 0;
             while( discNo != discSelArr[ idx ] )
                    idx++;
             ruleState[ offset + idx ] = 1;
         }
         return;
      }

      // search the location of the max disc.
      for( int i = 0; i < discNum; i++ )
      {
          int discNo = indices[ i ] % MAX_DISC_NUM;
          if( discNo == discSelArr[ discNum - 1 ] )
          {
              maxDiscLoc = i;
              break;
          }
      }

      // status is stage 2 and the state is used to decide an action.
      if( stage == 2 )
      {
          int offset = ( indices[ maxDiscLoc ] / MAX_DISC_NUM ) * MAX_DISC_NUM;
          int idx = discNum - 1;
          ruleState[ offset + idx ] = 1;
          return;
      }

      // status is stage 0 or stage 1.
      for( int i = 0; i < discNum; i++ )
      {
          int offset = ( indices[ i ] / MAX_DISC_NUM ) * MAX_DISC_NUM;
          int pegIdx = indices[ i ] / PEG_VAL_NUM;
          int discNo = indices[ i ] % MAX_DISC_NUM;

          offset += PEG_LOC_OFFSETS[ stage ][ pegIdx ];
          int maxDiscPegIdx = indices[ maxDiscLoc ] / PEG_VAL_NUM;

         // delete the max disc in order to use the already generated rules.
          if( discNo != discSelArr[ discNum - 1 ] )
          {
              if( maxDiscPegIdx == pegIdx )
                  offset += MAX_DISC_NUM;
              int idx = 0;
              while( discNo != discSelArr[ idx ] )
                     idx++;
              ruleState[ offset + idx ] = 1;
          }
      }
   }

   // check if the chosen action is legal.
   public void checkActions()
   {
      actionLegalFlag = false;
      exActionLegalFlag = actionLegalFlag;

      if( !gsFlag )
      {
          wmActionLegalFlag = rerWM.checkAction( wmAction );
          gsActionLegalFlag = rerGS.checkAction( gsAction );
          actionLegalFlag = checkExAction();
          exActionLegalFlag = actionLegalFlag;
      }
      else
      {
          gsActionLegalFlag = rerGS.checkAction( gsAction );
          if( gsActionLegalFlag )                // GS action is legal
          {
           // action = -2: nothing action; -1: push or pop;  >=0: direct move
              action = rerGS.checkDirectMove();
              if( action == -2 )                 // nothing action
                  action = EX_DO_NOTHING_ACTION;

              if( action != -1 )             // direct move or nothing action
              {
                  actionLegalFlag = checkExAction();
                  exActionLegalFlag = actionLegalFlag;
              }
          }
      }
   }

   // check if the chosen EXTERNAL action is legal.
   public boolean checkExAction( )
   {
      if( action == EX_DO_NOTHING_ACTION )
      {
          solutionFlag = false;
          feedback = NEG_REWARD;
          return false;
      }

      convAction( action, stageIdx );

  /******* the following is to check if the action is an illegal move.********/
      // check if the beginning peg is empty.
      if( curPegState[ beginPeg ].isEmpty() )
      {
          solutionFlag = false;
          feedback = NEG_REWARD;
          return false;
      }

      if( moveNum < ruleRecorder.length )
      {
          for( int i = 0; i < discNum; i++ )
               ruleRecorder[ moveNum ][ i ] = curExIndices[ i ];
          ruleRecorder[ moveNum ][ discNum ] = action;
          ruleRecorder[ moveNum ][ discNum + 1 ] = stageIdx;
      }
      moveNum++;

      // check if item1 > item2( an illegal move ).
      int[] item1 = new int[1];
      item1 = (int[])curPegState[ beginPeg ].peek();
      int[] item2 = new int[1];
      if( !curPegState[ endPeg ].isEmpty() )
          item2 = (int[])curPegState[ endPeg ].peek();
      else
          item2[0] = MAX_ITEM_VAL;

      if( item1[0] > item2[0] )
      {
          solutionFlag = false;
          feedback = NEG_REWARD;
          return false;
      }

      return true;
   }

   public void convAction( int curAction, int stage )
   {
      int begin = curAction / 2;
      int end = ( begin + 1 + curAction % 2 ) % 3;

      // modified on 02/01/02.
      if( gsFlag || discNum == MIN_DISC_NUM || stage == 2 )
      {
          beginPeg = begin;
          endPeg = end;
          return;
      }

      beginPeg = pegMapArr[ stage ][ begin ];
      endPeg = pegMapArr[ stage ][ end ];
   }

   // perform actions of each components.
   public void performActions()
   {
     if( !gsFlag )
     {
        if( wmActionLegalFlag )
            rerWM.performAction();
        if( gsActionLegalFlag )
            rerGS.performAction();
        if( actionLegalFlag )
            performExAction();
     }
     else
     {
        if( gsActionLegalFlag )
            rerGS.performAction();
        if( actionLegalFlag )       // gs makes a legal move.
            performExAction();
     }
   }

   // perform the chosen action to get the current state and the feedback
   public void performExAction()
   {
      // move the specific disc.
      int[] item1 = new int[1];
      item1 = (int[])curPegState[ beginPeg ].pop();
      curPegState[ endPeg ].push( item1 );

      for( int i = 0; i < discNum; i++ )
           preExIndices[ i ] = curExIndices[ i ];

      int offset = 0, idx = 0;
      for( int i = 0; i < 3; i++ )
      {
        offset = PEG_VAL_NUM * ( i + 1 );
        for( int j = 0; j < curPegState[i].size(); j++ )
        {
           int[] item = (int[])curPegState[i].elementAt( j );
           curExIndices[ idx++ ] = offset - (j+1) * MAX_DISC_NUM + item[0];
        }
      }

      // update the current state.
      for( int i = 0; i < discNum; i++ )
           curExState[ preExIndices[ i ] ] = 0;

      for( int i = 0; i < discNum; i++ )
           curExState[ curExIndices[ i ] ] = 1;

      // modified on 02/01/02.
      // get current stage info after moving a disc.
      postStageIdx = getStageInfo( curExIndices );

      // check if a solution has been found so far.
      for( int i = 0; i < discNum; i++ )
        if( curExIndices[i] != endExIndices[i] )
        {
            solutionFlag = false;
            getNonNegReward();
            return;
        }

      // a solution has been found.
      solutionFlag = true;
      feedback = FULL_REWARD;
   }

  // check if the current move is a key move. a key move means that the current
  // moved disc is the largest disc in the two non-goal pegs and the end peg is
  // the GOAL_PEG.
   public void getNonNegReward()
   {
      // modified on 02/01/02.
      // the biggest disc is moved
      /*
      if( stageIdx == 2 )
      {
          feedback = FULL_REWARD;
          return;
      }*/
      if( ( stageIdx == 0 && postStageIdx == 2 ) ||
          ( stageIdx == 2 && postStageIdx == 1 ) )
      {
          feedback = FULL_REWARD;
          return;
      }

      if( ( stageIdx == 2 && postStageIdx == 0 ) ||
          ( stageIdx == 1 && postStageIdx == 2 ) )
      {
          feedback = NEG_REWARD;
          return;
      }

      if( postStageIdx == -1 )
      {
          feedback = NEG_REWARD;
          return;
      }

      // if only one disc.
      if( discNum == 1  )
      {
          if( beginPeg == ORIGIN_PEG && endPeg == GOAL_PEG )
              feedback = FULL_REWARD;
          else
              feedback = ZERO_REWARD;
          return;
      }

      // check if the end peg is the GOAL_PEG.
      if( endPeg != pegMapArr[ stageIdx ] [ GOAL_PEG ] )
      {
          feedback = ZERO_REWARD;
          return;
      }

      // the currently moved disc
      int[] item = new int[1];
      item = (int[])curPegState[ endPeg ].peek();

/***************************************************************************/
      if( item[0] < discSelArr[ discNum - 2 ] )
          feedback = ZERO_REWARD;
      else
          feedback = FULL_REWARD * 0.8;
/***************************************************************************/
   }

   // update different components of the agent.
   public void updates()
   {
     if( !gsFlag )
     {
        //obtain current state
         boolean flag = getCurState();
         if( !flag )
             return;

         if( !actionLegalFlag )
             illegalUpdate();
         else
             legalUpdate();

         rerWM.setCurState( curState );
         if( wmActionLegalFlag )
             rerWM.setFeedback( ZERO_REWARD );
         else
             rerWM.setFeedback( NEG_REWARD );
         rerWM.update( wmActionLegalFlag );

         rerGS.setCurState( curState );
         if( gsActionLegalFlag )
             rerGS.setFeedback( ZERO_REWARD );
         else
             rerGS.setFeedback( NEG_REWARD );
         rerGS.update( gsActionLegalFlag );
     }
     else
     {
        // modified on 02/01/02
        //obtain current state
         boolean flag = getCurState();
         if( !flag )
             return;

         if( action != -1 )  // a physical action is made
         {
             rerGS.setFeedback( feedback );
             if( !actionLegalFlag )
                 illegalUpdate();
             else
                 legalUpdate();
         }
         else
         {
             if( gsActionLegalFlag )
                 rerGS.setFeedback( ZERO_REWARD );
             else
                 rerGS.setFeedback( NEG_REWARD );
         }

         rerGS.setCurState( curState );
         rerGS.update( gsActionLegalFlag );

         if( !gsActionLegalFlag || action != -1 ) // the process of making a
             gsFlag = false;                      // move by GS is finished.
      }
   }

   // update both top and bottom level when an illegal action is chosen
   public void illegalUpdate()
   {
      if( action == -1 )
          return;

      exQValue1 = exNet.getQVal( action );
      exQValue2 = exNet.getMaxQVal();

      int[] exRuleState = new int[ curExState.length ];
      convToExRuleState( exRuleState, curExIndices, stageIdx, true );

      // modified on 02/01/02
      if( levelFlag == DOWN_LEVEL )
      {
          //gs doesn't determine the action.
           if( !gsFlag )
           {
               if( feedback > 0 )
                   exNet.updateMatchCount( true );
               else if( feedback < 0 )
                        exNet.updateMatchCount( false );
           }

          exNet.update( action, feedback, exQValue1 );
      }
      else
          exT_RerRuleSet.deleteRule( exRuleState, action );
   }

   // update both top and bottom level when an legal action is chosen
   public void legalUpdate()
   {
        exQValue1 = exNet.getQVal( action );

       //obtain next state QValue2.
	exNet.setCurState( curState );
	exNet.setInput();
        exNet.forwardPass();
        exQValue2 = exNet.getMaxQVal();

/******************************************************************************/
        // update the QBP network.
        exNet.update( action, feedback, exQValue1, exQValue2 );
/******************************************************************************/

        // update RER rule set if the RER flag is set.
        int[] exRuleState = new int[ curExState.length ];
        convToExRuleState( exRuleState, preExIndices, stageIdx, false );
        int ruleAction = convToRuleAction( action, stageIdx );

        double r = exNet.getReinforcement();
        double d = exNet.getDiscount();

        // modified on 02/01/02
        if( r + d * exQValue2 - exQValue1 > threshold )
        {
           if( !gsFlag )
               exNet.updateMatchCount( true );
           exT_RerRuleSet.updateRuleSet( exRuleState, ruleAction, true, false, true );
        }
        else
        {
           if( !gsFlag )
               exNet.updateMatchCount( false );
           exT_RerRuleSet.updateRuleSet( exRuleState, ruleAction, false, false, true );
        }
   }

   public int convToRuleAction( int curAction, int stage )
   {
       int ruleAction;
       if( discNum == MIN_DISC_NUM )
           ruleAction = curAction;
       else
       {
           convAction( curAction, stage );
           ruleAction = beginPeg * 2;
           int offset = ( endPeg - beginPeg + 2 ) % 3;
           ruleAction += offset;
       }
       return ruleAction;
   }

   // check if the current solution is stable.
   public void checkStable( )
   {
      System.out.print( "GROUP NNUMBER  ===  " );
      System.out.print( groupIdx );
      System.out.print( "   ,    AGENT NUMBER  ===  " );
      System.out.println( agentIdx );

      moveNumArr[ expIdx ][ recordIdx ] = moveNum;
      diffRate = ((double)moveNum - optimalStepNum ) / optimalStepNum;
      System.out.print( "recordIdx = " );
      System.out.println( recordIdx );
      System.out.print( " compared with optimal: ");
      System.out.println( diffRate );

/******************************************************************************/
      double[] threshArr = { 0.6, 0.6, 0.6, 0.6, 0.6 };
      //double baseRate = ( discNum - MIN_DISC_NUM ) * 0.2;
      double baseRate = threshArr[ expIdx ];
/******************************************************************************/

      if( diffRate <= baseRate )
      {
          int v2, v1 = moveNumArr[ expIdx ][ recordIdx ];
          if( recordIdx > 0 )
              v2 = moveNumArr[ expIdx ][ recordIdx - 1 ];
          else
              v2 = 100000;

          if( v1 < v2 )
              recordOptimalRules();

          stableEpisodeNum++;
          avgMoveNum += moveNum;

          if( stableEpisodeNum == STABLE_EPISODE_NUM )
          {
              System.out.println( " Success! ");
              stableFlag = true;
              int val = avgMoveNum / STABLE_EPISODE_NUM;
              minMoveNums[ expIdx ] = val;

              for( int i = recordIdx; i < MAX_EPISODE_NUM; i++ )
                   moveNumArr[ expIdx ][ i ] = val;

              stopEpisodeArr[ expIdx ] = recordIdx;
          }
          /*else if( expIdx == expNum - 1 )
               {
                   System.out.println( " finish! ");
                   stableFlag = true;
                   minMoveNums[ expIdx ] = avgMoveNum;
                   for( int i = recordIdx; i < MAX_EPISODE_NUM; i++ )
                        moveNumArr[ expIdx ][ i ] = avgMoveNum;
               }
          */
      }
      else
      {
          if( stableEpisodeNum > 0 )
          {
              stableEpisodeNum = 0;
              avgMoveNum = 0;
          }
          stableFlag = false;
      }
  }

  public void recordOptimalRules()
  {
     for( int i = 0; i < moveNum; i++ )
     {
        int len = ruleRecorder[i].length;
        for( int j = 0; j < len - 2; j++ )
             preExIndices[ j ] = ruleRecorder[i][j];

        int[] exRuleState = new int[ curExState.length ];
        convToExRuleState( exRuleState, preExIndices, ruleRecorder[i][len-1], false );
        int ruleAction = convToRuleAction( ruleRecorder[ i ][ len - 2 ],
                                           ruleRecorder[ i ][ len - 1 ] );

        exT_RerRuleSet.updateRuleSet( exRuleState, ruleAction, true, false, true );
        exT_RerRuleSet.enforceRule( exRuleState, ruleAction );
     }
  }

    public void appendRules()
   {
       for( int i = 0; i < 3; i++ )
         for( int j = 0; j < 3; j++ )
              rerRuleStrings[i][j].delete( 0, rerRuleStrings[i][j].length() );

       String str = "THE RULES FOR EXTERNAL:\n\n";
       str += "All rules are: \n";
       rerRuleStrings[0][0].append( str + exT_RerRuleSet.getAllRules() );
       str = "\nMatch rules are: \n";
       rerRuleStrings[0][1].append( str + exT_RerRuleSet.getMatchRules() );
       str = "\nChosen rules are: \n";
       rerRuleStrings[0][2].append( str + exT_RerRuleSet.getChosenRules() );

       str = " THE RULES FOR GOAL STACK:\n\n ";
       str += "All rules are: \n";
       rerRuleStrings[1][0].append( str + rerGS.getAllRules() );
       str = "\nMatch rules are: \n";
       rerRuleStrings[1][1].append( str + rerGS.getMatchRules() );
       str = "\nChosen rules are: \n";
       rerRuleStrings[1][2].append( str + rerGS.getChosenRules() );

       str = " THE RULES FOR WORKING MEMORY:\n\n ";
       str += "All rules are: \n";
       rerRuleStrings[2][0].append( str + rerWM.getAllRules() );
       str = "\nMatch rules are: \n";
       rerRuleStrings[2][1].append( str + rerWM.getMatchRules() );
       str = "\nChosen rules are: \n";
       rerRuleStrings[2][2].append( str + rerWM.getChosenRules() );
   }

   public void fillRerRule( StringBuffer[] str )
   {
       for( int i = 0; i < 3; i++ )
        {
            //str[i].delete( 0, str[i].length() );
            for( int j = 0; j < 3; j++ )
                 str[i].append( rerRuleStrings[i][j].substring( 0 ) );
        }
   }

 public void showInfo()
  {
        System.out.print( "recordIdx = " );
        System.out.println( recordIdx - 1 );
        /*
        System.out.print( ",  moveNum = " );
        System.out.println( moveNumArr[ recordIdx - 1 ] );
        System.out.print( "topCount = " );
        System.out.print( topCount );
        System.out.print( ".  botCount = " );
        System.out.println( botCount );
        */
  }

  // show the peg state.
   public void showPegState()
   {
       int offset = 0, idx = 0;
       for( int i = 0; i < 3; i++ )
       {
          offset = PEG_VAL_NUM * ( i + 1 );
          System.out.print("peg " + String.valueOf(i+1)+ "= " );
          System.out.print( "( " );
          for( int j = 0; j < curPegState[i].size(); j++ )
          {
             int[] item = (int[])curPegState[i].elementAt( j );
             System.out.print( String.valueOf( item[0] ) + ", "  );
          }
          System.out.print( " )" );
          System.out.print( ",  " );
       }

       System.out.print( "  the action is " );
       System.out.print( action );
       System.out.print( ",  the move num is " );
       System.out.println( moveNum );

       System.out.println( "qbpArray=" );
       for( int i = 0; i < 6; i++ )
            System.out.println( qbpArray[i] );
   }
}
