/******************************************************************************/
/**************************  TOWER OF HANOI TASK  *****************************/
/******************************************************************************/
 package tasksimulators.oldsimu;
 import java.awt.*;
 import java.awt.event.*;
 import java.io.*;
 import javax.swing.*;
 import javax.swing.event.*;
 import java.util.Vector;
 import clarion.common.*;
 import clarion.system.*;
 import clarion.tools.*;
 import clarion.acs.*;

 public class THTask extends JFrame implements SimuTask {

    int WINDOW_WIDTH = 600, WINDOW_HEIGHT = 400;

    /* used for RER version
    int SIMU_NUM = 1, SIMU_TYPE = 0;
    int GROUP_NUM = 2, AGENT_NUM = 20, EXP_NUM = 5;
    int MAX_DISC_NUM = 6, MIN_DISC_NUM = 1;
    int BLOCK_SIZE = 20, EPISODE_MULTIPLIER = 200;
    boolean exitFlag = false, TEST_FLAG = false;
    */

    /* used for FR version BL TEST
    int SIMU_NUM = 1, SIMU_TYPE = 1;
    int GROUP_NUM = 1, AGENT_NUM = 20, EXP_NUM = 4;
    int MAX_DISC_NUM = 6, MIN_DISC_NUM = 2;
    int BLOCK_SIZE = 2000, EPISODE_MULTIPLIER = 20000;
    boolean exitFlag = false, TEST_FLAG = true;
    */

    /* used for RER + GS + WM version */
    int SIMU_NUM = 1, SIMU_TYPE = 2;
    int GROUP_NUM = 2, AGENT_NUM = 20, EXP_NUM = 5;
    int MAX_DISC_NUM = 6, MIN_DISC_NUM = 1;
    int BLOCK_SIZE = 20, EPISODE_MULTIPLIER = 200;
    boolean exitFlag = false, TEST_FLAG = false;

    int GROUP_FILE_NUM = 12, AGENT_FILE_NUM = 11;
    int PEG_VAL_NUM = MAX_DISC_NUM * MAX_DISC_NUM;

    int expMode, simIdx, groupIdx, agentIdx, discNum;
    int discInfo[], restartInfo[], DIM_VAL_NUMS[];
    int MAX_EPISODE_ARR[], moveNumArr[][][], stopEpisodeArr[][][];

    int nzMoveArr[][], nzTrainRTArr[][], nzTestBLMoveArr[][];
    int nzTrainPopArr[][], nzTestPopArr[][];
    int nzTrainBLHitArr[][], nzTestBLHitArr[][];
    int nzTrainComPercentArr[][], nzTestComPercentArr[][];

    int curMoveArr[], curTrainRTArr[], curTestBLMoveArr[];
    int totMoveArr[][], totTrainRTArr[][], totTestBLMoveArr[][];

    float curTrainPopArr[], curTrainBLHitArr[], curTrainComProbArr[];
    float totTrainPopArr[][], totTrainBLHitArr[][], totTrainComProbArr[][];
    float curTestPopArr[], curTestBLHitArr[], curTestComProbArr[];
    float totTestPopArr[][], totTestBLHitArr[][], totTestComProbArr[][];

    double  BETA_VALS[][][] = {
                               { { 5.0, 1.0 }, { 1.0, 2.0 } }, // FOR RER VERSION.
                               { { 1.0, 1.0 }, { 1.0, 5.0} }, // FOR FR VERSION.
                               { { 2.0, 1.0 }, { 1.0, 40.0 } } // FOR RER+GS+WM VERSION.
                            };

    final boolean YES_RER_FLAG = true, NO_RER_FLAG = false;
    final boolean YES_UTL_FLAG = true, NO_UTL_FLAG = false;
    final boolean YES_WM_FLAG = true, NO_WM_FLAG = false;
    final boolean YES_GS_FLAG = true, NO_GS_FLAG = false;
    final boolean YES_FR_FLAG = true, NO_FR_FLAG = false;
    boolean SIMU_FLAG_ARR[][] =
      {
        { YES_RER_FLAG, YES_UTL_FLAG, NO_WM_FLAG, NO_GS_FLAG, NO_FR_FLAG },
        { NO_RER_FLAG,  NO_UTL_FLAG,  YES_WM_FLAG, YES_GS_FLAG, YES_FR_FLAG },
        { YES_RER_FLAG, YES_UTL_FLAG,  YES_WM_FLAG, YES_GS_FLAG, NO_FR_FLAG },
      };

    WriteFile dataFile, groupDatFiles[][][][], agentDatFiles[][][][][];
    THRerAgent rerAgent;
    THFrAgent frAgent;

    Container c;
    JFrame mainFrame;
    JButton beginButton, exitButton;

    public THTask( JFrame mFrame )
    {
       initInterface( mFrame );
    }

   public void initInterface( JFrame mFrame )
   {
      // setup user interface.
      c = getContentPane();
      c.removeAll();
      c.setLayout( new BorderLayout( 5, 0 ) );
      buildInterface();

      setSize( WINDOW_WIDTH, WINDOW_HEIGHT );
      setEnabled( false );
      setVisible( false );

      mainFrame = mFrame;
      addWindowListener(
          new WindowAdapter(){
             public void windowClosing( WindowEvent e )
             {
               if ( exitFlag == true )
               {
                    dataFile.closeFile();
                    setVisible( false );
                    setEnabled( false );
                    mainFrame.setVisible( true );
                    mainFrame.setEnabled( true );
               }
             }
      }  );
  }

  public void buildInterface()
  {
      JPanel showPanel = new JPanel();
      showPanel.setLayout( new BorderLayout() );
      c.add( showPanel, BorderLayout.CENTER );

      JPanel capPanel = new JPanel();
      capPanel.setLayout( new FlowLayout( FlowLayout.CENTER, 200, 5 ) );
      showPanel.add( capPanel, BorderLayout.NORTH );
      JLabel taskCaption = new JLabel( "TOWER OF HANOI TASK" );
      taskCaption.setFont( new Font( "Serif", Font.BOLD, 20 ) );
      capPanel.add( taskCaption );

      JPanel mainPanel = new JPanel();
      mainPanel.setLayout( new GridLayout( 1, 2, 5, 5 ) );
      showPanel.add( mainPanel, BorderLayout.CENTER );

      JPanel trainPanel = new JPanel();
      trainPanel.setLayout( new GridLayout( 3, 1, 5, 5 ) );
      mainPanel.add( trainPanel );
      JTextArea probArea1 = new JTextArea( "ALL THE RER RULES: \n" );
      trainPanel.add( probArea1 );
      JTextArea ruleArea1 = new JTextArea( "MATCHING RER RULES: " );
      trainPanel.add( ruleArea1 );
      JTextArea showArea1 = new JTextArea( "CHOSEN RER RULES: \n" );
      trainPanel.add( showArea1 );

      JPanel transPanel = new JPanel();
      transPanel.setLayout( new GridLayout( 3, 1, 5, 5 ) );
      mainPanel.add( transPanel );
      JTextArea probArea2 = new JTextArea( "STATUS OF PEG 1: \n" );
      transPanel.add( probArea2 );
      JTextArea ruleArea2 = new JTextArea( "STATUS OF PEG 2: \n" );
      transPanel.add( ruleArea2 );
      JTextArea showArea2 = new JTextArea( "STATUS OF PEG 3: \n" );
      transPanel.add( showArea2 );

      JPanel buttonPanel = new JPanel();
      buttonPanel.setLayout( new FlowLayout() );
      c.add( buttonPanel, BorderLayout.SOUTH );

      beginButton = new JButton( "BEGIN" );
      beginButton.setEnabled( true );
      beginButton.addActionListener(
         new  ActionListener() {
                 public void actionPerformed( ActionEvent e )
                 {
                     beginButton.setEnabled( false );
                     exitButton.setEnabled( false );
                     exitFlag = false;
                     startExp();
                     exitButton.setEnabled( true );
                     exitFlag = true;
                 }
        }
      );

      exitButton = new JButton( "Exit" );
      exitButton.addActionListener(
                 new  ActionListener() {
                     public void actionPerformed( ActionEvent e )
                     {
                        closeAllFiles();
                        setVisible( false );
                        setEnabled( false );
                        mainFrame.setVisible( true );
                        mainFrame.setEnabled( true );
                     }
                 }
      );

      buttonPanel.add( beginButton );
      buttonPanel.add( exitButton );
   }

   public void allocateMemory()
   {
      moveNumArr = new int[ GROUP_NUM ][ AGENT_NUM ][ EXP_NUM ];
      stopEpisodeArr = new int[ GROUP_NUM ][ AGENT_NUM ][ EXP_NUM ];

      restartInfo = new int[ AGENT_NUM ];

      nzMoveArr = new int[ EXP_NUM ][];
      nzTrainRTArr = new int[ EXP_NUM ][];
      nzTrainPopArr = new int[ EXP_NUM ][];
      nzTrainBLHitArr = new int[ EXP_NUM ][];
      nzTrainComPercentArr = new int[ EXP_NUM ][];

      nzTestBLMoveArr = new int[ EXP_NUM ][];
      nzTestPopArr = new int[ EXP_NUM ][];
      nzTestBLHitArr = new int[ EXP_NUM ][];
      nzTestComPercentArr = new int[ EXP_NUM ][];

      totMoveArr = new int[ EXP_NUM ][];
      totTrainRTArr = new int[ EXP_NUM ][];
      totTrainPopArr = new float[ EXP_NUM ][];
      totTrainBLHitArr = new float[ EXP_NUM ][];
      totTrainComProbArr = new float[ EXP_NUM ][];

      totTestBLMoveArr = new int[ EXP_NUM ][];
      totTestPopArr = new float[ EXP_NUM ][];
      totTestBLHitArr = new float[ EXP_NUM ][];
      totTestComProbArr = new float[ EXP_NUM ][];

      MAX_EPISODE_ARR = new int[ EXP_NUM ];
      for( int expIdx = 0; expIdx < EXP_NUM; expIdx++ )
           MAX_EPISODE_ARR[ expIdx ] = EPISODE_MULTIPLIER * ( MIN_DISC_NUM + expIdx );

      for( int expIdx = 0; expIdx < EXP_NUM; expIdx++ )
      {
           int len = MAX_EPISODE_ARR[ expIdx ];
           nzMoveArr[ expIdx ] = new int[ len ];
           nzTrainRTArr[ expIdx ] = new int[ len ];
           nzTrainPopArr[ expIdx ] = new int[ len ];
           nzTrainBLHitArr[ expIdx ] = new int[ len ];
           nzTrainComPercentArr[ expIdx ] = new int[ len ];

           totMoveArr[ expIdx ] = new int[ len ];
           totTrainRTArr[ expIdx ] = new int[ len ];
           totTrainPopArr[ expIdx ] = new float[ len ];
           totTrainBLHitArr[ expIdx ] = new float[ len ];
           totTrainComProbArr[ expIdx ] = new float[ len ];

           int len1 = MAX_EPISODE_ARR[ EXP_NUM - 1 ] / BLOCK_SIZE;
           nzTestBLMoveArr[ expIdx ] = new int[ len1 ];
           nzTestPopArr[ expIdx ] = new int[ len1 ];
           nzTestBLHitArr[ expIdx ] = new int[ len1 ];
           nzTestComPercentArr[ expIdx ] = new int[ len1 ];

           totTestBLMoveArr[ expIdx ] = new int[ len1 ];
           totTestPopArr[ expIdx ] = new float[ len1 ];
           totTestBLHitArr[ expIdx ] = new float[ len1 ];
           totTestComProbArr[ expIdx ] = new float[ len1 ];
      }

      int len = MAX_EPISODE_ARR[ EXP_NUM - 1 ];
      curMoveArr = new int[ len ];
      curTrainRTArr = new int[ len ];
      curTrainPopArr = new float[ len ];
      curTrainBLHitArr = new float[ len ];
      curTrainComProbArr = new float[ len ];

      int len1 = MAX_EPISODE_ARR[ EXP_NUM - 1 ] / BLOCK_SIZE;
      curTestBLMoveArr = new int[ len1 ];
      curTestPopArr = new float[ len1 ];
      curTestBLHitArr = new float[ len1 ];
      curTestComProbArr = new float[ len1 ];
   }

   public void clearMemory()
   {
      for( int i = 0; i < AGENT_NUM; i++ )
           restartInfo[ i ] = 0;

      for( int expIdx = 0; expIdx < EXP_NUM; expIdx++ )
      {
           int len = MAX_EPISODE_ARR[ expIdx ];
           for( int i = 0; i < len; i++ )
           {
                nzMoveArr[ expIdx ][ i ] = 0;
                nzTrainRTArr[ expIdx ][ i ] = 0;
                nzTrainPopArr[ expIdx ][ i ] = 0;
                nzTrainBLHitArr[ expIdx ][ i ] = 0;
                nzTrainComPercentArr[ expIdx ][ i ] = 0;

                totMoveArr[ expIdx ][ i ] = 0;
                totTrainRTArr[ expIdx ][ i ] = 0;
                totTrainPopArr[ expIdx ][ i ] = 0;
                totTrainBLHitArr[ expIdx ][ i ] = 0;
                totTrainComProbArr[ expIdx ][ i ] = 0;
           }

           int len1 = MAX_EPISODE_ARR[ EXP_NUM - 1 ] / BLOCK_SIZE;
           for( int i = 0; i < len1; i++ )
           {
                nzTestBLMoveArr[ expIdx ][ i ] = 0;
                nzTestPopArr[ expIdx ][ i ] = 0;
                nzTestBLHitArr[ expIdx ][ i ] = 0;
                nzTestComPercentArr[ expIdx ][ i ] = 0;

                totTestBLMoveArr[ expIdx ][ i ] = 0;
                totTestPopArr[ expIdx ][ i ] = 0;
                totTestBLHitArr[ expIdx ][ i ] = 0;
                totTestComProbArr[ expIdx ][ i ] = 0;
           }
      }
   }

   public void initAgent()
   {
      if( simIdx == 0 || simIdx == 2 )
      {
          int NET_DIM_NUM = ( int )World.TASK_PARA_ARRAY[World.NET_DIM_NUM ];
          DIM_VAL_NUMS = new int[ NET_DIM_NUM ];
          for( int i = 0; i < NET_DIM_NUM; i++ )
               DIM_VAL_NUMS[ i ] = MAX_DISC_NUM;
          rerAgent = new THRerAgent( PEG_VAL_NUM );
          rerAgent.allocateMemory( EXP_NUM, MIN_DISC_NUM,  MAX_EPISODE_ARR,
                                  BETA_VALS[ simIdx ][ groupIdx ], TEST_FLAG );
      }

      if( simIdx == 1 )
      {
          frAgent = new THFrAgent( MIN_DISC_NUM );
          frAgent.allocateMemory( EXP_NUM, MIN_DISC_NUM,  BLOCK_SIZE,
               MAX_EPISODE_ARR, BETA_VALS[ simIdx ][ groupIdx ], TEST_FLAG );
      }
   }

   public void setSimuType()
   {
     if( simIdx == 0 || simIdx == 2 )
         rerAgent.setSimuType( SIMU_FLAG_ARR[simIdx] );
     else
         frAgent.setSimuType( SIMU_FLAG_ARR[simIdx] );
   }

   public void setThresholds()
   {
     if( groupIdx == 0 )
     {
         rerAgent.setThreshold(World.TASK_PARA_ARRAY[World.HIGH_RER_THRESHOLD]);
         rerAgent.setRERThresholds(World.TASK_PARA_ARRAY[World.HIGH_RER_THRESHOLD1],
                           World.TASK_PARA_ARRAY[World.HIGH_RER_THRESHOLD2] );
     }
     else
     {
         rerAgent.setThreshold(World.TASK_PARA_ARRAY[World.LOW_RER_THRESHOLD]);
         rerAgent.setRERThresholds(World.TASK_PARA_ARRAY[World.LOW_RER_THRESHOLD1],
                           World.TASK_PARA_ARRAY[World.LOW_RER_THRESHOLD2] );
     }
   }

   public void startExp()
   {
     allocateMemory();
     for( simIdx = SIMU_TYPE; simIdx < SIMU_TYPE + 1; simIdx++ )
     {
        initAgent();
        setSimuType();
        {
           for( groupIdx = 0; groupIdx < GROUP_NUM; groupIdx++ )
            {
                if( simIdx == 0 || simIdx == 2 )
                    setThresholds();
                for( agentIdx = 0; agentIdx < AGENT_NUM; agentIdx++ )
                {
                     System.out.print( "=============================== ");
                     System.out.print( " GROUP NO = ");
                     System.out.print(  groupIdx );
                     System.out.print( ",,,,, AGENT NO = ");
                     System.out.println(  agentIdx );
                     runAgent();
                 }
                 saveGroupData();
                 clearMemory();
            }
            saveSimuData();
         }
     }
   }

   public void runAgent()
   {
      if( simIdx == 0 || simIdx == 2 )
      {
          rerAgent.setEnvVarables( simIdx, groupIdx, agentIdx );
          rerAgent.solveProblems();
          rerAgent.getMoveNums(moveNumArr[groupIdx][agentIdx]);
          restartInfo[ agentIdx ] = rerAgent.getRestartCount();
          saveAgentData();
          rerAgent.reinitAgent();
      }

      if( simIdx == 1 )
      {
          frAgent.setEnvVarables( simIdx, groupIdx, agentIdx );
          frAgent.solveProblems();
          frAgent.getMoveNums(moveNumArr[groupIdx][agentIdx]);
          restartInfo[ agentIdx ] = frAgent.getRestartCount();
          saveAgentData();
          frAgent.reinitAgent();
      }
   }

   public void saveAgentData()
   {
     for( int expIdx = 0; expIdx < EXP_NUM; expIdx++ )
     {
       // record the stop episode number of the agent
       if( simIdx == 0 || simIdx == 2 )
           rerAgent.getStopEpisodeArr( stopEpisodeArr[ groupIdx ][ agentIdx ] );
       else
           frAgent.getStopEpisodeArr( stopEpisodeArr[ groupIdx ][ agentIdx ] );

       // record  the performance of the agent.
       int len = 0;
       if( simIdx == 0 || simIdx == 2 )
           len = rerAgent.getArrayLen( expIdx );
       else
           len = frAgent.getArrayLen( expIdx );

       if( simIdx == 0 || simIdx == 2 )
           rerAgent.getMoveNumArr( curMoveArr, expIdx );
       else
       {
           frAgent.getMoveNumArr( curMoveArr, expIdx );
           frAgent.getTrainBLGoodMoveArr( curTrainBLHitArr, expIdx );
           frAgent.getTrainCombineProbArr( curTrainComProbArr, expIdx );
           frAgent.getTrainGoalPopArr( curTrainPopArr, expIdx );
           frAgent.getResponseTimeArr( curTrainRTArr, expIdx );

           frAgent.getTestBLGoodMoveArr( curTestBLHitArr, expIdx );
           frAgent.getTestCombineProbArr( curTestComProbArr, expIdx );
           frAgent.getTestGoalPopArr( curTestPopArr, expIdx );
           frAgent.getTestBLMoveArr( curTestBLMoveArr, expIdx );
       }

       String str1 = "", str2 = "";
       int sum = 0, count = 0, blockNum = 0;
       len = nzMoveArr[ expIdx ].length;
       for( int i = 0; i < len; i++ )
       {
           int val = curMoveArr[ i ];
           if( val > 0 )
           {
               nzMoveArr[ expIdx ][ i ]++;
               totMoveArr[ expIdx ][ i ] += val;

               /*
               str1 += String.valueOf( i + 1 ) + " ";
               str1 += String.valueOf( val ) + "\n";
               count++;
               sum += val;
               if( count == BLOCK_SIZE )
               {
                   blockNum++;
                   sum /= count;
                   str2 += String.valueOf( blockNum ) + " ";
                   str2 += String.valueOf( sum ) + "\n";
                   sum = 0;
                   count = 0;
               }
               */
           }
           else
           {
               /*
               if( count > 0 )
               {
                   blockNum++;
                   sum /= count;
                   str2 += String.valueOf( blockNum ) + " ";
                   str2 += String.valueOf( sum ) + "\n";
                   sum = 0;
                   count = 0;
               }
               break;
               */
           }
       }

       //agentDatFiles[0][groupIdx][agentIdx][expIdx][0].saveStringData(str1);
       //agentDatFiles[0][groupIdx][agentIdx][expIdx][1].saveStringData(str2);

       if( simIdx == 1 )
       {
           int len1 = len / BLOCK_SIZE;
           saveTrainData( expIdx, len );
           saveTestData( expIdx, len1 );
       }

       String SIMU_TYPE_STR[] =
       {
         //"ONLY RER \n",
         "RER + RULE UTILITY \n",
         "WM + GS + FR \n", "WM + GS + RER \n"
       };

       String str = "CURRENT EXPERIMENT IS: TOWER OF HANOI \n";
       str += "The simulation is: " + SIMU_TYPE_STR[ simIdx ];
       str += "CURRENT GROUP IS: ";
       if( groupIdx == 0 )
           str += " no verbalization group" + "\n";
       else
           str += " verbalization group" + "\n";

       str += "CURRENT AGENT NO IS: ";
       str += String.valueOf( agentIdx ) + "\n";

       str += "CURRENT NUMBER OF DISCS IS: ";
       str += String.valueOf( expIdx + MIN_DISC_NUM ) + "\n\n";

       str += "THE NUMBER OF MOVE IS: ";
       str += String.valueOf( moveNumArr[groupIdx][agentIdx][expIdx] );
       str += "\n";
       dataFile.saveStringData( str );
     }

     if( simIdx == 0 || simIdx == 2 )
     {
         StringBuffer ruleSet[] = new StringBuffer[3];
         for( int i = 0; i < 3; i++ )
              ruleSet[ i ] = new StringBuffer( "" );

         rerAgent.fillRerRule( ruleSet );
         for( int i = 0; i < 3; i++ )
              dataFile.saveStringData( ruleSet[i].substring( 0 ) );
     }
   }

   public void saveTrainData( int expIdx, int len )
   {
      String str1 = "";
      for( int i = 0; i < 4; i++ )
      {
           str1 = "";
           for( int j = 0; j < len; j++ )
           {
                float val = 0;
                switch( i )
                {
                   case 0:
                         val = curTrainBLHitArr[ j ];
                         if( val > 0 )
                         {
                             nzTrainBLHitArr[ expIdx ][ j ]++;
                             totTrainBLHitArr[ expIdx ][ j ] += val;
                         }
                         break;
                    case 1:
                         val = curTrainComProbArr[ j ];
                         if( val > 0 )
                         {
                             nzTrainComPercentArr[ expIdx ][ j ]++;
                             totTrainComProbArr[ expIdx ][ j ] += val;
                         }
                         break;
                    case 2:
                         val = curTrainPopArr[ j ];
                         if( val > 0 )
                         {
                             nzTrainPopArr[ expIdx ][ j ]++;
                             totTrainPopArr[ expIdx ][ j ] += val;
                         }
                         break;
                    case 3:
                         val = curTrainRTArr[ j ];
                         if( val > 0 )
                         {
                             nzTrainRTArr[ expIdx ][ j ]++;
                             totTrainRTArr[ expIdx ][ j ] += val;
                         }
                         break;
                   }
             }
             //agentDatFiles[0][groupIdx][agentIdx][expIdx][3+i].saveStringData(str1);
       }
   }

   public void saveTestData( int expIdx, int len )
   {
      String str1 = "";
      for( int i = 0; i < 4; i++ )
      {
           str1 = "";
           for( int j = 0; j < len; j++ )
           {
                float val = 0;
                switch( i )
                {
                   case 0:
                         val = curTestBLHitArr[ j ];
                         if( val > 0 )
                         {
                             nzTestBLHitArr[ expIdx ][ j ]++;
                             totTestBLHitArr[ expIdx ][ j ] += val;
                         }
                         break;
                    case 1:
                         val = curTestComProbArr[ j ];
                         if( val > 0 )
                         {
                             nzTestComPercentArr[ expIdx ][ j ]++;
                             totTestComProbArr[ expIdx ][ j ] += val;
                         }
                         break;
                    case 2:
                         val = curTestPopArr[ j ];
                         if( val > 0 )
                         {
                             nzTestPopArr[ expIdx ][ j ]++;
                             totTestPopArr[ expIdx ][ j ] += val;
                         }
                         break;
                    case 3:
                         val = curTestBLMoveArr[ j ];
                         if( val > 0 )
                         {
                             nzTestBLMoveArr[ expIdx ][ j ]++;
                             totTestBLMoveArr[ expIdx ][ j ] += val;
                         }
                         break;
                   }
             }
             //agentDatFiles[0][groupIdx][agentIdx][expIdx][7+i].saveStringData(str1);
       }
   }

   public void saveGroupData()
   {
    for( int expIdx = 0; expIdx < EXP_NUM; expIdx++ )
    {
       String str = "";
       for( int i = 0; i < AGENT_NUM; i++ )
       {
            int val = stopEpisodeArr[groupIdx][i][expIdx];
            str += String.valueOf( i + 1 ) + " ";
            str += String.valueOf( val + 1 ) + "\n";
       }
       groupDatFiles[0][groupIdx][expIdx][3].saveStringData( str );

       int len = nzMoveArr[expIdx].length;
       /*
       str = "";
       for( int i = 0; i < len; i++ )
       {
          int count = nzMoveArr[ expIdx ][ i ];
          if( count == 0 )
              count++;
          int sum = totMoveArr[ expIdx ][ i ];

          sum /= count;
          str += String.valueOf( i + 1 ) + " ";
          str += String.valueOf( sum ) + "\n";
       }
       groupDatFiles[0][groupIdx][expIdx][0].saveStringData( str );
       */

       str = "";
       int blockNum = len / BLOCK_SIZE;
       if( len % BLOCK_SIZE != 0 )
           blockNum++;

       for( int i = 0; i < blockNum; i++ )
       {
          int END_NUM = 0;
          if( ( len % BLOCK_SIZE == 0 ) || ( i < blockNum - 1 ) )
                END_NUM = BLOCK_SIZE;
          else
                END_NUM = len % BLOCK_SIZE;

          int sum = 0, count = 0;
          for( int j = 0; j < END_NUM; j++ )
          {
              sum += totMoveArr[ expIdx ][ BLOCK_SIZE * i + j ];
              count += nzMoveArr[ expIdx ][ BLOCK_SIZE * i + j ];
          }

          if( count == 0 )
              count++;
          sum /= count;
          str += String.valueOf( i + 1 ) + " ";
          str += String.valueOf( sum ) + "\n";
       }
       groupDatFiles[0][groupIdx][expIdx][1].saveStringData( str );

       str = "CURRENT EXPERIMENT IS: TOWER OF HANOI \n";
       String SIMU_TYPE_STR[] =
       {
           //"ONLY RER \n",
           "RER + RULE UTILITY \n",
           "WM + GS + FR \n", "WM + GS + RER \n"
       };
       str += "The simulation is: " + SIMU_TYPE_STR[ simIdx ];

       str += "CURRENT NUMBER OF DISCS IS: ";
       str += String.valueOf( expIdx + MIN_DISC_NUM ) + "\n\n";

       str += "CURRENT GROUP IS: ";
       if( groupIdx == 0 )
           str += " no verbalization group" + "\n";
       else
           str += " verbalization group" + "\n";

       str += "The number of restart is: \n" ;
       for( int i = 0; i < AGENT_NUM; i++ )
            str += String.valueOf( restartInfo[ i ] ) + "\n";

       groupDatFiles[0][groupIdx][expIdx][2].saveStringData( str );

       len = nzMoveArr[expIdx].length;
       for( int k = 0; k < 4; k++ )
            saveAsTrainBlockData( len, expIdx, k );

       int len1 = len / BLOCK_SIZE;
       for( int k = 0; k < 4; k++ )
            saveAsTestBlockData( len1, expIdx, k );
     }
   }

   public void saveAsTrainBlockData( int len, int eIdx, int arrIdx )
   {
       String str = "";
       int blockSize = 0, blockNum = 0;
       if( arrIdx < 3 )
           blockSize = BLOCK_SIZE;
       else
           blockSize = 1;
       blockNum = len / blockSize;

       if( len % blockSize != 0 )
           blockNum++;

       for( int i = 0; i < blockNum; i++ )
       {
          int END_NUM = 0;
          if( ( len % blockSize == 0 ) || ( i < blockNum - 1 ) )
                END_NUM = blockSize;
          else
                END_NUM = len % blockSize;

          int count = 0;
          float sum = 0;

          for( int j = 0; j < END_NUM; j++ )
          {
               switch( arrIdx )
               {
                  case 0:
                       sum += totTrainBLHitArr[ eIdx ][ blockSize * i + j ];
                       count += nzTrainBLHitArr[ eIdx ][ blockSize * i + j ];
                       break;
                  case 1:
                       sum += totTrainComProbArr[ eIdx ][ blockSize * i + j ];
                       count += nzTrainComPercentArr[ eIdx ][ blockSize * i + j ];
                       break;
                  case 2:
                       sum += totTrainPopArr[ eIdx ][ blockSize * i + j ];
                       count += nzTrainPopArr[ eIdx ][ blockSize * i + j ];
                       break;
                  case 3:
                       sum += totTrainRTArr[ eIdx ][ blockSize * i + j ];
                       count += nzTrainRTArr[ eIdx ][ blockSize * i + j ];
                       break;
               }
          }

          if( count == 0 )
              count++;
          sum /= count;
          if( sum > 0 )
          {
              str += String.valueOf( i + 1 ) + " ";
              str += String.valueOf( sum ) + "\n";
          }
       }

       groupDatFiles[0][groupIdx][eIdx][4+arrIdx].saveStringData( str );
   }

   public void saveAsTestBlockData( int len, int eIdx, int arrIdx )
   {
       String str = "";
       int blockSize = 1;
       int blockNum = len;

       for( int i = 0; i < blockNum; i++ )
       {
          int count = 0;
          float sum = 0;

          switch( arrIdx )
          {
             case 0:
                   sum += totTestBLHitArr[ eIdx ][ blockSize * i ];
                   count += nzTestBLHitArr[ eIdx ][ blockSize * i ];
                   break;
             case 1:
                   sum += totTestComProbArr[ eIdx ][ blockSize * i ];
                   count += nzTestComPercentArr[ eIdx ][ blockSize * i ];
                   break;
             case 2:
                   sum += totTestPopArr[ eIdx ][ blockSize * i ];
                   count += nzTestPopArr[ eIdx ][ blockSize * i ];
                   break;
              case 3:
                   sum += totTestBLMoveArr[ eIdx ][ blockSize * i ];
                   count += nzTestBLMoveArr[ eIdx ][ blockSize * i ];
                   break;
          }

          if( count == 0 )
              count++;
          sum /= count;
          if( sum > 0 )
          {
              str += String.valueOf( i + 1 ) + " ";
              str += String.valueOf( sum ) + "\n";
          }
       }

       groupDatFiles[0][groupIdx][eIdx][8+arrIdx].saveStringData( str );
   }

   public void saveSimuData()
   {
     String str = "CURRENT EXPERIMENT IS: TOWER OF HANOI \n";
     String SIMU_TYPE_STR[] =
     {
        //"ONLY RER \n",
        "RER + RULE UTILITY \n",
        "WM + GS + FR \n", "WM + GS + RER \n"
     };

     str += "The simulation is: " + SIMU_TYPE_STR[ simIdx ];

     for( int expIdx = 0; expIdx < EXP_NUM; expIdx++ )
     {
       str += "CURRENT NUMBER OF DISCS IS: ";
       str += String.valueOf( expIdx + MIN_DISC_NUM ) + "\n";

       for( groupIdx = 0; groupIdx < GROUP_NUM; groupIdx++ )
       {
           str += "CURRENT GROUP IS: ";
           if( groupIdx == 0 )
               str += " no verbalization group" + "\n";
           else
               str += " verbalization group" + "\n";

           int sum = 0;
           for( agentIdx = 0; agentIdx < AGENT_NUM; agentIdx++ )
           {
              str += "CURRENT AGENT NO IS: ";
              str += String.valueOf( agentIdx ) + "\n";
              str += "NUMBER OF MOVE OF THIS AGENT IS: ";
              int num = moveNumArr[groupIdx][agentIdx][expIdx];
              str += String.valueOf( num ) + "\n";
              sum += num;
           }

           str += "THE MEAN OF MOVES OF THIS GROUP IS: ";
           str += String.valueOf( (float)sum/AGENT_NUM );
           str += "\n\n";
        }
      }
      dataFile.saveStringData( str );
   }

   public void createDataFiles()
   {
      String nameStr[] = { "sgl", "blk", "rst", "sec",
                           "Train_bhc", "Train_cpc", "Train_pop", "Train_rtc",
                           "Test_bhc", "Test_cpc", "Test_pop", "Test_blp"
                         };

      groupDatFiles = new WriteFile[SIMU_NUM][GROUP_NUM][EXP_NUM][GROUP_FILE_NUM];
      for( int i = 0; i < SIMU_NUM; i++ )
        for( int j = 0; j < GROUP_NUM; j++ )
          for( int k = 0; k < EXP_NUM; k++ )
            for( int p = 0; p < GROUP_FILE_NUM; p++ )
            {
                 groupDatFiles[i][j][k][p] = new WriteFile();
                 String str = "sim" + String.valueOf( i + 1 ) +
                        "grp" + String.valueOf( j + 1 ) +
                        "dsc" + String.valueOf( MIN_DISC_NUM+k ) + nameStr[p];
                 groupDatFiles[i][j][k][p].openFile( 1,  str );
            }

       /*
      String nameStr1[] = { "sgl", "blk", "rst", "blc",
                            "cpc", "pop", "rtc", "blp" };
      agentDatFiles = new WriteFile[SIMU_NUM][GROUP_NUM][AGENT_NUM][EXP_NUM][AGENT_FILE_NUM];
      for( int i = 0; i < SIMU_NUM; i++ )
        for( int j = 0; j < GROUP_NUM; j++ )
          for( int k = 0; k < AGENT_NUM; k++ )
            for( int p = 0; p < EXP_NUM; p++ )
              for( int q = 0; q < AGENT_FILE_NUM; q++ )
              {
                   agentDatFiles[i][j][k][p][q] = new WriteFile();
                   String str = "sim" + String.valueOf( i + 1 ) +
                                "grp" + String.valueOf( j + 1 ) +
                                "agt" + String.valueOf( k + 1 ) +
                                "dsc" + String.valueOf( MIN_DISC_NUM + p ) +
                                nameStr1[ q ];
                   agentDatFiles[i][j][k][p][q].openFile( 1,  str );
              }
       */
   }

   public void closeAllFiles()
   {
     dataFile.closeFile();
     for( int i = 0; i < SIMU_NUM; i++ )
          for( int j = 0; j < GROUP_NUM; j++ )
            for( int k = 0; k < EXP_NUM; k++ )
                for( int p = 0; p < GROUP_FILE_NUM; p++ )

                   groupDatFiles[i][j][k][p].closeFile();
     /*
     for( int i = 0; i < SIMU_NUM; i++ )
         for( int j = 0; j < GROUP_NUM; j++ )
            for( int k = 0; k < AGENT_NUM; k++ )
               for( int p = 0; p < EXP_NUM; p++ )
                 for( int q = 0; q < AGENT_FILE_NUM; q++ )
                      agentDatFiles[i][j][k][p][q].closeFile();
     */
   }

   public String getTaskName()
   {
     return "TOWER OF HANOI TASK";
   }

   public void start( int expMode, int blockNum, WriteFile dFile )
   {
       this.expMode = expMode;
       dataFile = dFile;
       setEnabled( true );
       setVisible( true );
       createDataFiles();
       //initAgent();
   }
 }
