 /************************** class T_RerRule begin ********************************/
 package tasksimulators.oldsimu;
 import clarion.common.*;
 import clarion.tools.*;

 public class T_RerRule extends Object
   {
     protected final double DISCOUNT = 0.9;

     protected int age, DIMENSION_NUM, DIM_VAL_NUM, TOT_TRIAL_NUM;
     protected boolean UTL_FLAG, MATCH_ALL_FLAG, FULL_FLAG, RT_FLAG;
     /* FULL_FLAG is used in TOH task to represent whether the condition of */
     /* a rule include all the three parts: WM,EX,GS or only one part. */

     protected double INIT_BLA;
     protected float utility;
     protected float posMatch, negMatch, matchRate;
     protected float[][][][] statistics;
     /* the 3rd dimension records statistics of stateX[i][j]
        index = 0: records the statistics if the condition
        "stateX[i][j] = 1" doesn't exists.
        index = 1: records the statistics if the condition
        "stateX[i][j] = 1"  exists.

        the 4th dimension
        index = 0: Positive Matches
        index = 1: Negative Matches
        index = 2: the rate PM + 1/ PN + NM + 2
      */

     protected int DIFF_DIM_IDX;
     protected int suggestedAction;         // conclusion of the rule.
     protected int[][] stateX;              // conditions of the rule.
     protected int[]   stateY;              // used for converting state.
     protected int[] maxValLocPlus, maxValLocMinus;
                                        /* used for recording the location of
                                           the max matching rate in the array
                                           statistics.*/
     protected int[]   indices;             /* numbers of elements in each dimension
                                            of  condition array stateX. */
     protected StringBuffer ruleString;   // represents the content of the rule
                                        // including condition and conclusion.
     RuleFeature feature;

     public T_RerRule( int dimNum, int dimValNum, boolean uFlag, boolean rFlag )
     {
         suggestedAction = -1;
         DIMENSION_NUM = dimNum;
         DIM_VAL_NUM = dimValNum;

         UTL_FLAG = uFlag;
         RT_FLAG = rFlag;

         ruleString = new StringBuffer( new String("") );
     }

     public T_RerRule ( int dimNum,  int dimValNum,
                      boolean uFlag, boolean mFlag, boolean fFlag, boolean rFlag,
                      int[] state, int action, int totNum, double initBLA )
     {
         DIMENSION_NUM = dimNum;
         DIM_VAL_NUM = dimValNum;

         UTL_FLAG = uFlag;
         MATCH_ALL_FLAG = mFlag;
         FULL_FLAG = fFlag;
         RT_FLAG = rFlag;

         suggestedAction = action;
         TOT_TRIAL_NUM = totNum;
         INIT_BLA = initBLA;

         ruleString = new StringBuffer( new String("") );

         indices  = new int [ DIMENSION_NUM ];
         initIndices();

         maxValLocPlus = new int[2];
         maxValLocMinus = new int[2];

         statistics = new float[DIMENSION_NUM][DIM_VAL_NUM][2][3];
         initStatistics();

         stateY = new int[ DIMENSION_NUM ];
         stateX = new int[ DIMENSION_NUM ][ DIM_VAL_NUM ];
         convertToStateX( state );

         if( UTL_FLAG )
             initRuleFeature( initBLA );
     }

/***************************** GET METHODS ************************************/
     public void getOneDimension( int which, int[] arr )
     {
        for( int i = 0; i < DIM_VAL_NUM; i++ )
             arr[ i ] = stateX[ which ][ i ];
     }

     public int getAction()
     {
        return suggestedAction;
     }

     public int getAge()
     {
        return age;
     }

     public int getResponseTime()
     {
        return feature.getResponseTime();
     }

     public int getADimIdx( int idx )
     {
         return indices[ idx ];
     }

     public int getOneDimValue( int which )
     {
        return stateX[ which ][ indices[ which ] ];
     }

     public int[] getIndices()
     {
        return indices;
     }

     /* get the location of the max matching rate in the statistics array. */
     /* used for rule generalization and specialization. */
     /* mode 0: +1 condition, mode 1: -1 condition. */
     public int[] getMaxIGLoc( int mode )
     {
       if( mode == 0 )
           return maxValLocPlus;
       else
           return maxValLocMinus;
     }

     /* search the location of the max matching rate in the array statistics.
        mode 0: generalization, mode 1: specialization. */
     public int[] getOptCondition( int mode )
     {
        double val, maxVal = 0.0;
        int[]  retVal = new int[2];

        for( int i = 0; i < DIMENSION_NUM; i++ )
            for( int j = 0; j < DIM_VAL_NUM; j++ )
                if( stateX[i][j] == mode )
                {
                    if( mode == 0 )
                        val = statistics[i][j][1][2];
                    else
                        val = statistics[i][j][0][2];

                    if( maxVal < val )
                    {
                        maxVal = val;
                        retVal[0] = i;
                        retVal[1] = j;
                    }
                }

         if( mode == 0 )
             maxValLocPlus = retVal;
         else
             maxValLocMinus = retVal;

         return retVal;
     }

     public int[] getStateY()
     {
        return stateY;
     }

     public int[][] getStateX()
     {
        return stateX;
     }

     public float getUtility()
     {
        feature.calcUtility();
        utility = (float)feature.getUtility();
        return utility;
     }

     public float getPosMatch()
     {
        return posMatch;
     }

     public float getNegMatch()
     {
        return negMatch;
     }

     public float getMatchRate()
     {
        matchRate = ( posMatch + 1 ) / ( posMatch + negMatch + 2 );
        return matchRate;
     }

     // get a match rate of a minor variation of the rule condition.
     public float getAStatRate( int idx1, int idx2, int mode )
     {
         return statistics[ idx1 ][ idx2 ][ 1 - mode ][ 2 ];
     }

     public float[][][][] getStatistics()
     {
        return statistics;
     }

     public RuleFeature getFeature()
     {
        return feature;
     }

    // get the text string representing the rule.
     public StringBuffer getRuleString()
     {
         return ruleString;
     }

/***************************** SET METHODS ************************************/
     public void setAction( int action )
     {
        suggestedAction = action;
     }

     public void setAge( int age )
     {
        this.age = age;
     }

     public void setIndices( int[] arr )
     {
        for( int i = 0; i < arr.length; i++ )
              indices[ i ] = arr[ i ];
     }

     public void setStateY( int[] arr )
     {
        for( int i = 0; i < arr.length; i++ )
             stateY[ i ] = arr[ i ];
     }

     public void setStateX( int[][] state )
     {
        for( int i = 0; i < state.length; i++ )
         for( int j = 0; j < state[ i ].length; j++ )
              stateX[ i ][ j ] = state[ i ][ j ];
     }

     public void setMaxIGLocPlus( int[] arr )
     {
       for( int i = 0; i < arr.length; i++ )
            maxValLocPlus[ i ] = arr[ i ];
     }

     public void setMaxIGLocMinus( int[] arr )
     {
       for( int i = 0; i < arr.length; i++ )
            maxValLocMinus[ i ] = arr[ i ];
     }

     public void setPosMatch( float posMatch )
     {
        this.posMatch = posMatch;
     }

     public void setNegMatch( float negMatch )
     {
        this.negMatch = negMatch;
     }

     public void setMatchRate( float matchRate )
     {
        this.matchRate = matchRate;
     }

     public void setStatistics( float[][][][] stats )
     {
        for( int i = 0; i < stats.length; i++ )
         for( int j = 0; j  < stats[ i ].length; j++ )
          for( int k = 0; k < stats[ i ][ j ].length; k++ )
           for( int m = 0; m < stats[ i ][ j ][ k ].length; m++ )
                statistics[ i ][ j ][ k ][ m ] = stats[ i ][ j][ k ][ m ];
     }

     public void setRuleCost( double cost )
     {
        feature.setCost( cost );
     }

     public void setRuleActionTime( double time )
     {
        feature.setActionTime( time );
     }

     public void setRulePreBLA( double val )
     {
        feature.setPreBLA( val );
     }

    public void setFeature( RuleFeature feat )
    {
        feature = (RuleFeature)feat.clone();
    }

   // convert the rule into a text string.
    public String toString()
    {
         int MAX_DIM_NUM = DIMENSION_NUM;
         ruleString.delete( 0, ruleString.length() );
         ruleString.append( "{ " );
         for( int i = 0; i < MAX_DIM_NUM; i++ )
         {
              ruleString.append( "(" );
              for( int j = 0; j < DIM_VAL_NUM; j++ )
              {
                   if( stateX[ i ][ j ] == 1 )
                   {
                      ruleString.append( String.valueOf( j + 1 ) );
                      ruleString.append( " " );
                   }
              }
              ruleString.append( ")" );
              if( i < MAX_DIM_NUM - 1 )
                  ruleString.append( "," );
         }
         ruleString.append( " }" );

         ruleString.append( " --> action " + String.valueOf( suggestedAction + 1));
         ruleString.append( "\n" );

         return ruleString.substring( 0, ruleString.length() );
   }

   public void aging()
   {
        age++;
   }

   public void incUtility( int val )
   {
        feature.decCost( val );
        feature.calcUtility( );
   }

   public void decUtility( int val )
   {
       feature.incCost( val );
       feature.calcUtility( );
   }

   public void resetStatistics()
   {
       initStatistics();
   }

    /* update the rule statistics whenever a match exists.
       S_FLAG:  true if the current step is successful,   Positive Matches
                false if the current step is unsuccessful, Negative Matches.
       ALL_FLAG: true if the rule is a match-all rule.
                 false if the rule is not a match-all rule.
       C_FLAG: true if match current condition and action.
               false if doesn't match current generalized condation  and action
     */
    public void updateStatistics( boolean S_FLAG, boolean ALL_FLAG, boolean C_FLAG )
    {
        // match current condition and action.
        if( C_FLAG )
        {
            if( S_FLAG )
                posMatch += 1;
            else
                negMatch += 1;
            matchRate = ( posMatch + 1) / ( posMatch + negMatch + 2);

            if( UTL_FLAG )
            {
                if( S_FLAG )
                    incUtility( 1 );
                else
                    decUtility( -1 );
            }
        }

        if( ALL_FLAG )
            return;

        int mode;
        if( S_FLAG )
            mode = 0;
        else
            mode = 1;

        // update all statistics regardless of condition
        // "stateX[i][j] = 1" exist.
        for( int i = 0; i < DIMENSION_NUM; i++ )
         for( int j = 0; j < DIM_VAL_NUM; j++ )
         {
              statistics[i][j][0][mode] += 1;
              statistics[i][j][1][mode] += 1;
         }

        // since stateX[i][stateY[i]] = 1 which means condition
        // "stateX[i][stateY[i]] = 1" should exist, so the statistics
        // statistics[i][stateY[i]][0][mode]( the case of nonexisting
        // of "stateX[i][stateY[i]] = 1" )  needn't increase.
        for( int i = 0; i < DIMENSION_NUM; i++ )
        {
             if ( stateY[i] >= 0 )
                  statistics[i][stateY[i]][0][mode] -= 1;
        }

        for( int i = 0; i < DIMENSION_NUM; i++ )
          for( int j = 0; j < DIM_VAL_NUM; j++ )
               for( int k = 0; k < 2; k++ )
               {
                    float val1 = statistics[i][j][k][0];
                    float val2 = statistics[i][j][k][1];
                    statistics[i][j][k][2] = (val1 + 1) / (val1 + val2 + 2);
               }
     }

   // discount the rule statistics after each trial.
   public void discountStatistics()
   {
       posMatch *= DISCOUNT;
       negMatch *= DISCOUNT;
       matchRate = ( posMatch + 1) / ( posMatch + negMatch + 2);

       for( int i = 0; i < DIMENSION_NUM; i++ )
         for( int j = 0; j < DIM_VAL_NUM; j++ )
            for( int k = 0; k < 2; k++ )
            {
                 statistics[i][j][k][0] *= DISCOUNT;
                 statistics[i][j][k][1] *= DISCOUNT;
                 float val1 = statistics[i][j][k][0];
                 float val2 = statistics[i][j][k][1];
                 statistics[i][j][k][2] = (val1 + 1) / (val1 + val2 + 2);
            }
   }

   // add a value of one dimension to the rule.
   // used for generaliztion.
   public void addAValue( int idx1, int idx2 )
   {
        stateX[ idx1 ][ idx2 ] = 1;
        indices[ idx1 ]++;
   }

   // del a value of one dimension to the rule.
   // used for specialization.
   public void delAValue( int idx1, int idx2 )
   {
        stateX[ idx1 ][ idx2 ] = 0;
        indices[ idx1 ]--;
   }

   // test if the current rule can be generalized.
   public boolean canGeneralized()
   {
        for( int i = 0; i < DIMENSION_NUM; i++ )
             if ( indices[ i ] < DIM_VAL_NUM )
                  return true;
        return false;
   }

   // test if the current rule can ba specialized.
   public boolean canSpecialized()
   {
          /*
            for( int i = 0; i < DIMENSION_NUM; i++ )
                 if ( indices[ i ] > 1 )
                      return true;
            return false;
            */                 // modified 12/05/01

        for( int i = 0; i < DIMENSION_NUM; i++ )
              if ( indices[ i ] <= 1 )
                   return false;   // modified 08/15/02
        return true;
   }

/************************** RULE MATCH METHODS *********************************/

   // if current state matches the rule condition.
   public boolean matchCondition( int[] state )
   {
        convertToStateY( state );

        int dNum = DIMENSION_NUM;
        for( int i = 0; i < dNum; i++ )
        {
             if( stateY[ i ] < 0 )
                 continue;
             if( stateX[ i ][ stateY[i] ] == 0 )
                 return false;
        }
        return true;
   }

   // if the chosen action matches the conclusion of the rule.
   public boolean matchAction( int action )
   {
        if( action == suggestedAction )
            return true;
        else
            return false;
   }

   /* successFlag means the current step is successful.
      check if the current state and chosen action match the condition of
      the rule and the conclusion.*/
   public boolean match( int[] state, int action, boolean successFlag )
   {
        boolean retVal = matchCondition( state );
        if( !retVal )
            return false;

        retVal = matchAction( action );
        if( retVal )
        {
            age = 0;            // it means the rule is applied this step.
            updateStatistics( successFlag, false, true );
            return true;
        }
        else
            return false;       // doesn't match the current rule
   }

   // if current state matches the generalized rule condition.
   public boolean matchGeneralCondition( int[] state )
   {
        convertToStateY( state );

        int dNum = DIMENSION_NUM;
        int misCount = 0;
        for( int i = 0; i < dNum; i++ )
        {
             if( stateY[ i ] < 0 )
                 continue;
             if( stateX[ i ][ stateY[i] ] == 0 )
             {
                 DIFF_DIM_IDX = i;
                 misCount++;
             }
             if( misCount > 1 )
                 return false;
        }

        if( misCount == 1 )
            return true;
        else
            return false;
   }

   /* successFlag means the current step is successful.
      check if the current state and chosen action match the generalized
      condition of the rule and the conlusion. */
   public boolean generalMatch( int[] state, int action, boolean successFlag )
   {
        boolean retVal = matchGeneralCondition( state );
        if( !retVal )
             return false;

        retVal = matchAction( action );
        if( retVal )
        {
            updateStatistics( successFlag, false, false );
            return true;
        }
        else
            return false;           // doesn't match
   }

   // the method is only used when the rule is a match-all rule.
   public void updateMatchAll( int[] state, boolean successFlag )
   {
        convertToStateY( state );
        age = 0;          // it means the rule is applied this step.
        updateStatistics( successFlag, true, true );
   }

   public void updateFeature( boolean succFlag, int curTime )
   {
        feature.updateMatchCounts( succFlag );
        if( RT_FLAG && succFlag )
             feature.updateUsageInfo( curTime );
   }

   public void calcBLA( int curTime )
   {
        feature.calcBLA( curTime );
   }

   public void calcTLDT()
   {
        feature.calcTLDT();
   }

/************************** PRIVATE METHODS ************************************/
     private void convertToStateX( int[] state )
     {
        for( int i = 0; i < DIMENSION_NUM; i++ )
          for( int j = 0; j < DIM_VAL_NUM; j++)
                 stateX[i][j] = 0;

        for( int i = 0; i < DIMENSION_NUM; i++ )
          for( int j = 0; j < DIM_VAL_NUM; j++ )
          {
               stateX[ i ][ j ] = state[ i * DIM_VAL_NUM + j ];
               if( state[ i * DIM_VAL_NUM + j] == 1 )
                   indices[ i ]++;
          }
     }

     // stateY used for converting the original input state into
     // a 6-element array.
     private void convertToStateY( int[] state )
     {
        for( int i = 0; i < DIMENSION_NUM; i++ )
             stateY[i] = exist( state, i*DIM_VAL_NUM, (i+1)*DIM_VAL_NUM, 1 );
     }

     // search 'val' in array.
     // return value >= 0: the location;  = -1: nonexist.
     private int exist( int[] array, int minIdx, int maxIdx, int val )
     {
        for( int i = minIdx; i < maxIdx; i++ )
              if ( array[i] == val )
                   return ( i - minIdx );
        return -1;
     }

     private void initIndices()
     {
         for( int i = 0; i < DIMENSION_NUM;i++ )
              indices[ i ] = 0;
     }

     // initialize the rule feature.
     private void initRuleFeature( double initBLA )
     {
        int[] ruleInfoArr = new int[2];

        int len = World.MODEL_PARA_ARRAY[ World.RULE_BLA_OPT ].length;
        for( int i = 0; i < len; i++ )
          if( World.MODEL_PARA_ARRAY[ World.RULE_BLA_OPT ][i] == 1 )
          {
              ruleInfoArr[ 0 ] = i;
              break;
          }

        len = World.MODEL_PARA_ARRAY[ World.RULE_UTL_OPT ].length;
        for( int i = 0; i < len; i++ )
          if( World.MODEL_PARA_ARRAY[ World.RULE_UTL_OPT ][i] == 1 )
          {
              ruleInfoArr[ 1 ] = i;
              break;
          }

        boolean[] defFlagArr = new boolean[2];
        defFlagArr[0] = true;
        defFlagArr[1] = true;

        if( RT_FLAG )
            feature = new RuleFeature( TOT_TRIAL_NUM, initBLA, ruleInfoArr, defFlagArr );
        else
            feature = new RuleFeature( ruleInfoArr, defFlagArr );
     }

/******************************* PROTECTED METHODS *****************************/
   protected Object clone( )
   {
         int[] state = new int[ DIM_VAL_NUM * DIMENSION_NUM ];
         T_RerRule rule = new T_RerRule ( DIMENSION_NUM, DIM_VAL_NUM,
                                      UTL_FLAG, MATCH_ALL_FLAG, FULL_FLAG, RT_FLAG,
                                      state, suggestedAction, TOT_TRIAL_NUM, INIT_BLA );
         rule.setAge( age );
         rule.setIndices( indices );
         rule.setStateY( stateY );
         rule.setStateX( stateX );
         rule.setMaxIGLocPlus( maxValLocPlus );
         rule.setMaxIGLocMinus( maxValLocMinus );

         rule.setPosMatch( posMatch );
         rule.setNegMatch( negMatch );
         rule.setMatchRate( matchRate );
         rule.setStatistics( statistics );

         if( UTL_FLAG )
             rule.setFeature( feature );

         return rule;
   }

     protected void initStatistics()
     {
         posMatch = 0;
         negMatch = 0;
         matchRate = 0;

         for( int i = 0; i < 2; i++ )
         {
              maxValLocPlus[ i ] = 0;
              maxValLocMinus[ i ] = 0;
         }

         for( int i = 0; i < DIMENSION_NUM; i++)
           for( int j = 0; j < DIM_VAL_NUM; j++)
             for( int k = 0; k < 2; k++ )
               for( int m = 0; m < 3; m++ )
                    statistics[i][j][k][m] = 0;
     }
 }