package tasksimulators.oldsimu;

public class T_RerRuleEx extends T_RerRule {

     protected float[][][] specInfo;
     /*  record statistics of all the values in stateX in order to determine
         which dimension will be specialized.
         the 3rd dimension
         index = 0: Positive Matches
         index = 1: Negative Matches
         index = 2: the rate PM + 1/ PN + NM + 2
     */

     protected float[][] genInfo;
     /* record statistics of all the dimensions of stateX in order to determine
        which dimension will be generalized.
        the 2nd dimension
        index = 0: Positive Matches
        index = 1: Negative Matches
        index = 2: the rate PM + 1/ PN + NM + 2
     */

     protected int[][] dimGenInfo;
     /* used for which dimension of stateX can be generalized or specialized */
     /* only record the valid dimensions which has non-zero values in its elements*/
     /* 1st dimension contains the indices of all the valid dimensions */
     /* 2nd dimension indicates the corresonping dimension can be generalized or
     /* specialized. = 0 : means can be generalized; = 1 : can be specialized */

     protected int VALID_DIM_NUM, genIdx, specIdx[];
     /* VALID_DIM_NUM means the # of dimensions which has non-zero elements */
     /* genIdx: indicating the dimension which can generalized at the most priority. */
     /* specIdx: indicating the dimension and one of its elements which can
     /* specialized at at the most priority.*/

 public T_RerRuleEx ( int dimNum,  int vDimNum, int dimValNum,
                    boolean uFlag, boolean mFlag, boolean fFlag, boolean rFlag,
                    int[] state, int action, int totNum, double initBLA )
 {
        super( dimNum, dimValNum, uFlag, mFlag, fFlag, rFlag,
               state,  action, totNum, initBLA );

        VALID_DIM_NUM = vDimNum;

        genInfo = new float[DIMENSION_NUM][3];
        specInfo = new float[DIMENSION_NUM][DIM_VAL_NUM][3];
        dimGenInfo = new int[VALID_DIM_NUM][2];

        if( !MATCH_ALL_FLAG )
            setValidDims( state );
 }

 public void setValidDims( int[] state )
 {
     int count = 0;
     for( int i = 0; i < DIMENSION_NUM; i++ )
      for( int j = 0; j < DIM_VAL_NUM; j++ )
      {
           if( state[ i * DIM_VAL_NUM + j ] == 1 )
           {
               dimGenInfo[count++][0] = i;
               break;
           }
      }
 }

 public int getDimGenInfoIdx( int idx )
 {
        int retVal = 0;
        for( int i = 0; i < dimGenInfo.length; i++ )
             if( dimGenInfo[ i ][ 0 ] == idx )
             {
                 retVal = i;
                 break;
             }
        return retVal;
 }

  /* the following three methods are used in TOH task functioned as
     getOptCondition   */
 public float getAMaxRate( int mode )
 {
        double val;
        if( mode == 0 )
            val = getMaxGenRate();
        else
            val = getMaxSpecRate();
        return (float)val;
 }

 public float getMaxGenRate()
 {
        double val1, val2, val = 0.0, maxVal = 0.0;
        for( int i = 0; i < VALID_DIM_NUM; i++ )
        {
            if( dimGenInfo[i][1] == 0 )
            {
               val1 = genInfo[dimGenInfo[i][0]][0];
               val2 = genInfo[dimGenInfo[i][0]][1];
               val = ( val1 + 1 ) / ( val1 + val2 + 2 );
               if( val > maxVal )
               {
                   maxVal = val;           // modified by xi zhang 11/28/01
                   genIdx = dimGenInfo[i][0];
               }
            }
        }
        return (float)val;
 }

 public float getMaxSpecRate()
 {
        double val1, val2, val = 0.0, maxVal = 0.0;

        specIdx = new int[2];
        for( int i = 0; i < VALID_DIM_NUM; i++ )
        {
            if( dimGenInfo[i][1] == 1 )
            {
               for( int j = 0; j < DIM_VAL_NUM; j++ )
               {
                  val1 = specInfo[dimGenInfo[i][0]][j][0];
                  val2 = specInfo[dimGenInfo[i][0]][j][1];
                  val = ( val1 + 1 ) / ( val1 + val2 + 2 );
                  if( val > maxVal )
                  {
                      maxVal = val;           // modified by xi zhang 11/28/01
                      specIdx[0] = dimGenInfo[i][0];
                      specIdx[1] = j;
                  }
               }
            }
        }
        return (float)val;
 }

 public void setDimGenInfo( int[][] info )
 {
        for( int i = 0; i < info.length; i++ )
         for( int j = 0; j < info[ i ].length; j++ )
             dimGenInfo[ i ][ j ] = info[ i ][ j ];
 }

 public void  setGenInfo( float[][] info )
 {
        for( int i = 0; i < info.length; i++ )
         for( int j = 0; j < info[ i ].length; j++ )
              genInfo[ i ][ j] = info[ i ][ j ];
 }

 public void setSpecInfo( float[][][] info )
 {
        for( int i = 0; i < info.length; i++ )
         for( int j = 0; j < info[ i ].length; j++ )
           for( int k = 0; k < info[ i ][ j ].length; k++ )
                specInfo[ i ][ j ][ k ] = info[ i ][ j ][ k ];
 }

 // used for TOH Task.
 public void incMatchRate( double posVal )
 {
     posMatch += posVal;
 }

 public void decMatchRate( double negVal )
 {
     negMatch += negVal;
 }

 public void generalADim( int dIdx )
 {
        for( int i = 0; i < DIM_VAL_NUM; i++ )
             stateX[ dIdx ][ i ] = 1;

        int idx = getDimGenInfoIdx( dIdx );
        dimGenInfo[ idx ][ 1 ] = 1;    // modified on 11/28/01, it means this
                                       // dimension can only be specialized and
                                       // not generalized in the future.
 }

 public void generalizeADim()
 {
        for( int i = 0; i < DIM_VAL_NUM; i++ )
             stateX[ genIdx ][ i ] = 1;

        int idx = getDimGenInfoIdx( genIdx );
        dimGenInfo[ idx ][ 1 ] = 1;     // modified on 11/28/01, it means this
                                        // dimension can only be specialized
                                        // and not generalized in the future.
 }

 public void specializeADim()
 {
        for( int i = 0; i < DIM_VAL_NUM; i++ )
             stateX[ specIdx[0] ][ i ] = 0;
        stateX[ specIdx[0] ][ specIdx[1] ] = 1;

        int idx = getDimGenInfoIdx( specIdx[0] );
        dimGenInfo[ idx ][ 1 ] = 0;    // modified on 11/28/01, it means this
                                              // dimension can only be generalized
                                              // and not specialized in the future.
 }

 /*
 // convert the rule into a text string.
 public void setRuleString()
 {
     int dimIdx = 0, END_P = 0;
     String[] pString = { "external part: ", "gs part: ", "wm part: " };
     String[] pDimString = { "PEG: ", "DIM: ", "DIM: " };

     int[] pDimArr = { 3, 6, 6 };
     int[] pDimBlkArr = { 6, 1, 1 };
     int[] offset = { 0, 18, 24 };

     if( FULL_FLAG )
         END_P = 3;
     else
         END_P = 1;

     ruleString.delete( 0, ruleString.length() );
     ruleString.append( "{ " );
     for( int p = 0; p < END_P; p++ )
     {
          ruleString.append( pString[ p ] );
          ruleString.append( "[" );

          for( int i = 0; i < pDimArr[ p ]; i++ )
          {
               int begin = offset[ p ] + i * pDimBlkArr[ p ];
               int end = offset[ p ] + ( i + 1 ) * pDimBlkArr[ p ];

               ruleString.append( pDimString[ p ] );
               ruleString.append( String.valueOf( i ) + ": " );

               ruleString.append( "(" );
               while( dimIdx < VALID_DIM_NUM && dimGenInfo[dimIdx][0] >= begin
                      && dimGenInfo[dimIdx][0] < end )
               {
                     ruleString.append( "<" );
                     for( int j = 0; j < DIM_VAL_NUM; j++ )
                     {
                          if( stateX[ dimGenInfo[dimIdx][0] ][ j ] == 1 )
                          {
                              ruleString.append( String.valueOf( j ) );
                              ruleString.append( " " );
                          }
                     }
                     ruleString.append( "> " );
                     dimIdx++;
               }
               ruleString.append( ")" );
               if( i < pDimArr[ p ] - 1 )
                   ruleString.append( ", " );
         }
         ruleString.append( "]" );
         if( p < END_P - 1 )
             ruleString.append( "," );
         ruleString.append( "\n" );
     }

     ruleString.append( " }" );
     ruleString.append( " --> action " + String.valueOf( suggestedAction ));
     ruleString.append( "\n" );
 } */

 /* update the rule statistics whenever a match exists.
    S_FLAG:  true if the current step is successful,   Positive Matches
             false if the current step is unsuccessful, Negative Matches.
    ALL_FLAG: true if the rule is a match-all rule.
              false if the rule is not a match-all rule.
    C_FLAG: true if match current condition and action.
           false if doesn't match current generalized condation  and action
 */
 public void updateStatistics( boolean S_FLAG, boolean ALL_FLAG, boolean C_FLAG )
 {
     // match current condition and action.
     if( C_FLAG )
     {
         if( S_FLAG )
             posMatch += 1;
         else
             negMatch += 1;
         matchRate = ( posMatch + 1) / ( posMatch + negMatch + 2);

         if( UTL_FLAG )
         {
             if( S_FLAG )
                 incUtility( 1 );
             else
                 decUtility( -1 );
         }
     }

     if( ALL_FLAG )
         return;

     if( !C_FLAG )
     {
         if( S_FLAG )
             genInfo[DIFF_DIM_IDX][0] += 1;
         else
             genInfo[DIFF_DIM_IDX][1] += 1;
     }

     for( int i = 0; i < DIMENSION_NUM; i++ )
     {
          if ( stateY[i] >= 0 )
          {
               if( S_FLAG )
                   specInfo[i][stateY[i]][0] += 1;
               else
                   specInfo[i][stateY[i]][1] += 1;
          }
     }
 }

 // discount the rule statistics after each trial.
 public void discountStatistics()
 {
     posMatch *= DISCOUNT;
     negMatch *= DISCOUNT;
     matchRate = ( posMatch + 1) / ( posMatch + negMatch + 2);

     for( int i = 0; i < DIMENSION_NUM; i++ )
     {
          genInfo[i][0] *= DISCOUNT;
          genInfo[i][1] *= DISCOUNT;
          for( int j = 0; j < DIM_VAL_NUM; j++ )
          {
               specInfo[i][j][0] *= DISCOUNT;
               specInfo[i][j][1] *= DISCOUNT;
          }
     }
 }

 // test if the current rule can be generalized.
 public boolean canGeneralized()
 {
     for( int i = 0; i < VALID_DIM_NUM; i++ )
          if( dimGenInfo[i][1] == 0 )
              return true;
     return false;
 }

 // test if the current rule can ba specialized.
 public boolean canSpecialized()
 {
     for( int i = 0; i < VALID_DIM_NUM; i++ )
          if( dimGenInfo[i][1] == 1 )
              return true;
     return false;
 }

/******************************* PROTECTED METHODS *****************************/
 protected Object clone( )
 {
     int[] state = new int[ DIM_VAL_NUM * DIMENSION_NUM ];
     T_RerRuleEx rule = new T_RerRuleEx ( DIMENSION_NUM,  VALID_DIM_NUM, DIM_VAL_NUM,
                                      UTL_FLAG, MATCH_ALL_FLAG, FULL_FLAG, RT_FLAG,
                                      state, suggestedAction, TOT_TRIAL_NUM, INIT_BLA );

     rule.setDimGenInfo( dimGenInfo );
     rule.setGenInfo( genInfo );
     rule.setSpecInfo( specInfo );

     return rule;
 }

 protected void initStatistics()
 {
     posMatch = 0;
     negMatch = 0;
     matchRate = 0;

     for( int i = 0; i < VALID_DIM_NUM; i++ )
       for( int j = 0; j < 2; j++ )
            dimGenInfo[i][j] = 0;

     for( int i = 0; i < DIMENSION_NUM; i++)
     {
          for( int j = 0; j < 3; j++ )
               genInfo[i][j] = 0;
          for( int j = 0; j < DIM_VAL_NUM; j++)
            for( int k = 0; k < 3; k++ )
                 specInfo[i][j][k] = 0;
     }
  }
}