/********************* IMPLEMENTATION OF RER RULE ***************************/
package tasksimulators.oldsimu;
import java.lang.*;
import java.util.*;
import clarion.common.*;
import clarion.tools.*;

public class T_RerRuleSet
{
/************************** Variables of T_RerRuleSet ******************************/
   private final int RULE_NUM = 150;
   private final int OLDEST_AGE = 500;  // used to test if a rule needs deleting.
                                       // according to clarion V5.0
   private final int STEP_INTERVAL = 10;

   private final int BASIC_MODE = 0, EX_MODE = 1;

   private double SUCC_C3 = 1.0, SUCC_C4 = 2.0;
   private double succRate, sumPM, sumNM;
   private double RULE_COST, RULE_ACTION_TIME, RULE_PRE_BLA, INIT_BLA;

   private int genMode, DIMENSION_NUM, VALID_DIM_NUM;
   private int INPUT_NUM, HIDDEN_NUM, OUTPUT_NUM, DIM_VAL_NUM;
   private int currentTime, responseTime, TOT_TRIAL_NUM;
   private int curRuleNum, matchCounter, chosenCounter;
   private short chosenAction, chosenActionNum, chosenActionRules[];

   private int whenMerge;              // when merger flag.
   private int timeSteps;              // steps since the last app of the rule.
   private boolean UTL_FLAG, RT_FLAG, FULL_FLAG = true;
   private double threshold1, threshold2;
                                      /* for testing the success of the
                                         current step or rule. According to
                                         clarion 5.0 */

   private T_LinkedNode ruleLink;             // link used for managing the rule set.

   private T_LinkedNode[] matchAllNode;       // 12 match-all rules, each has a
                                       // different conclusion.
   private StringBuffer allRules;
   private StringBuffer matchRules;
   private StringBuffer chosenRules;

/************************* begin of T_RerRuleSet methods  **************************/

  // constructor method of class T_RerRuleSet.
  //public T_RerRuleSet ( int mode, int mergeFlag, int[] nodeInfo, boolean uFlag )
  public T_RerRuleSet ( int mode, int mergeFlag, int[] nodeInfo, boolean uFlag )
  {
       matchCounter = 0;
       chosenCounter = 0;
       curRuleNum = 0;

       genMode = mode;
       whenMerge = mergeFlag;
       UTL_FLAG = uFlag;

       DIMENSION_NUM = nodeInfo[0];
       VALID_DIM_NUM = nodeInfo[1];
       INPUT_NUM = nodeInfo[2];
       HIDDEN_NUM = nodeInfo[3];
       OUTPUT_NUM = nodeInfo[4];
       DIM_VAL_NUM = INPUT_NUM / DIMENSION_NUM;

       chosenActionRules = new short[ RULE_NUM ];
       matchRules = new StringBuffer();
       allRules = new StringBuffer();
       chosenRules = new StringBuffer();

       ruleLink = new T_LinkedNode( DIMENSION_NUM, DIM_VAL_NUM, UTL_FLAG, RT_FLAG );
       createMatchAllRuleNode();

       double v1 =  World.TASK_PARA_ARRAY[ World.HIGH_RER_THRESHOLD1 ];
       double v2 =  World.TASK_PARA_ARRAY[ World.HIGH_RER_THRESHOLD2 ];
       setThreshold( v1, v2 );
  }

  public void reinit()
  {
      matchCounter = 0;
      chosenCounter = 0;
      curRuleNum = 0;

      ruleLink = null;
      ruleLink = new T_LinkedNode( DIMENSION_NUM, DIM_VAL_NUM, UTL_FLAG, RT_FLAG );

      for( int i = 0; i < OUTPUT_NUM; i++ )   // modified on 12/13/01
           matchAllNode[ i ] = null;
      createMatchAllRuleNode( );

      double v1 =  World.TASK_PARA_ARRAY[ World.HIGH_RER_THRESHOLD1 ];
      double v2 =  World.TASK_PARA_ARRAY[ World.HIGH_RER_THRESHOLD2 ];
      setThreshold( v1, v2 );
  }

   /*  Match-all rule means that the condition of the rule matches all the
       possible states.*/
  public void createMatchAllRuleNode( )
  {
      int[] state = new int[ INPUT_NUM ];

      for( int i = 0; i <  INPUT_NUM; i++ )
           state[ i ] = 1;

      matchAllNode = new T_LinkedNode[ OUTPUT_NUM ];
      for( int i = 0; i < OUTPUT_NUM; i++ )
           matchAllNode[ i ] = new  T_LinkedNode( DIMENSION_NUM, VALID_DIM_NUM, DIM_VAL_NUM,
                                    UTL_FLAG, true, FULL_FLAG, RT_FLAG,
                                    state, i, TOT_TRIAL_NUM, INIT_BLA );
  }

  public int getMatchCounter()
  {
    return matchCounter;
  }

  public int getChosenCounter()
  {
    return chosenCounter;
  }

  public int getResponseTime()
  {
     return responseTime;
  }

  // get current active rule numbers
  public int getActiveRuleNum()
  {
      T_LinkedNode cur = ruleLink.getSible();
      int counter = 0;

      while( cur != null )
      {
          counter++;
          cur = cur.getSible();
      }

      return counter;
  }

  public double getSuccRate()
  {
     T_LinkedNode chd = ruleLink;
     T_LinkedNode cur = ruleLink;
     sumPM = 0.0;
       sumNM = 0.0;
       for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
       {
            sumPM += cur.getElement().getPosMatch();
            sumNM += cur.getElement().getNegMatch();
            chd = cur.getChild();
            while( chd != null )
            {
               sumPM += cur.getElement().getPosMatch();
               sumNM += cur.getElement().getNegMatch();
               chd = chd.getSible();
            }
       }

       succRate = ( SUCC_C3 + sumPM ) / ( SUCC_C4 + sumPM + sumNM );
       return succRate;
  }

  // get all the rules including active or inactive used to be shown.
  public StringBuffer getAllRules()
  {
       String str;
       T_LinkedNode chd = ruleLink;
       T_LinkedNode cur = ruleLink;

       allRules.delete( 0, allRules.length() );

       for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
       {
            str = "ACTIVE RULE: " +
                   cur.getElement().getRuleString().substring( 0 );
            allRules.append( str );

            str = "";
            if( UTL_FLAG )
                str = "The utility is: " +
                    String.valueOf(cur.getElement().getUtility() ) + "\n";
            str += "The match rate is: " +
                    String.valueOf(cur.getElement().getMatchRate() ) + "\n";

            allRules.append( str );

            chd = cur.getChild();
            while( chd != null )
            {
               str = "INACTIVE RULE: " +
                      chd.getElement().getRuleString().substring( 0 );
               allRules.append( str );

               str = "";
               if( UTL_FLAG )
                   str = "The utility is: " +
                         String.valueOf(cur.getElement().getUtility() ) + "\n";
               str += "The match rate is: " +
                      String.valueOf(cur.getElement().getMatchRate() ) + "\n";
               allRules.append( str );

               chd = chd.getSible();
            }
       }

       return allRules;
  }

  // matching rules ( only matching condition ) only from the active rules.
  public StringBuffer getMatchRules()
  {
      return matchRules;
  }

  // chosen fules only from the active rules.
  public StringBuffer getChosenRules()
  {
    return chosenRules;
  }

  // modified by 02/13/02.
  // get all the actions' strengths for being selected from the current RER rule set,
  public void getActions( int[] state, double[] arr )
  {
      double actionActivations[] = new double[ OUTPUT_NUM ];
      for( int i = 0; i < OUTPUT_NUM; i++ )
           actionActivations[ i ] = 0;

      T_LinkedNode cur = ruleLink.getSible();
      if( cur == null )
      {
          for( int i = 0; i < OUTPUT_NUM; i++ )
               arr[ i ] = actionActivations[ i ];
          return;
      }

      matchRules.delete( 0, matchRules.length() );
      for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
      {
           boolean retVal = cur.getElement().matchCondition( state );
           if( retVal )
           {
               matchRules.append( cur.getElement().getRuleString() );
               int action = cur.getElement().getAction();
               if( UTL_FLAG )
               {
                   double util = cur.getElement().getUtility();
                   actionActivations[ action ] += util;
               }
               else
               {
                   double rate = cur.getElement().getMatchRate();
                   actionActivations[ action ] = 1;
               }
               matchCounter++;
          }
      }

      for( int i = 0; i < OUTPUT_NUM; i++ )
           arr[ i ] = actionActivations[ i ];
  }

  // get recommended action from the current RER rule set.
  public int getAction( int[] state )
  {
      short[] actionMatchNums = new short[ OUTPUT_NUM ];
      short[][] ruleMatchInfo = new short[ OUTPUT_NUM ][ RULE_NUM ];
      for( int i = 0; i < OUTPUT_NUM; i++ )
      {
         actionMatchNums[ i ] = 0;
         for( int j = 0; j < RULE_NUM; j++ )
              ruleMatchInfo[ i ][ j ] = 0;
      }

      chosenActionNum = 0;
      for( int i = 0; i < RULE_NUM; i++ )
           chosenActionRules[ i ] = 0;

      short  actionSelections[] = new short[ OUTPUT_NUM ];
      double actionActivations[] = new double[ OUTPUT_NUM ];
      for( int i = 0; i < OUTPUT_NUM; i++ )
      {
           actionSelections[ i ] = 0;
           actionActivations[ i ] = 0;
      }

      T_LinkedNode cur = ruleLink.getSible();
      if( cur == null )
          return -1;

      short matchIdx = 0, ruleIdx = 0, matchNum = 0;
      matchRules.delete( 0, matchRules.length() );
      for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
      {
           boolean retVal = cur.getElement().matchCondition( state );
           if( retVal )
           {
               matchRules.append( cur.getElement().getRuleString() );
               int action = cur.getElement().getAction();
               ruleMatchInfo[ action ][ actionMatchNums[ action ]++ ] = ruleIdx;
               actionSelections[ action ] = 1;
               if( UTL_FLAG )
               {
                   double util = cur.getElement().getUtility();
                   actionActivations[ action ] += util;
               }
               else
               {
                   double rate = cur.getElement().getMatchRate();
                   actionActivations[ action ] = 1;
               }
               matchNum++;
               matchCounter++;
          }
          ruleIdx++;
      }

      if( matchNum == 0 )
          return -1;

      short[]  candidateActions = new short[ matchNum ];
      double[] matchedRuleActivations = new double[ matchNum ];
      int idx = 0;
      for( int i = 0; i < OUTPUT_NUM; i++ )
      {
           if( actionSelections[ i ] == 1 )
           {
               candidateActions[ idx ] = (short)i;
               matchedRuleActivations[ idx++ ] = actionActivations[ i ];
           }
      }

      StochasticDecider actionDecider = new StochasticDecider( matchNum,
                        World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );
      int chosenActionIdx = (short)actionDecider.getRecommendation( matchedRuleActivations );
      chosenAction = candidateActions[ chosenActionIdx ];
      chosenActionNum = actionMatchNums[ chosenAction ];
      for( int i = 0; i < chosenActionNum; i++ )
           chosenActionRules[ i ] = ruleMatchInfo[ chosenAction ][ i ];
      return chosenAction;
  }

  public void setValidDimNum( int num )
  {
     VALID_DIM_NUM = num;
  }

  public void setTotTrialNum( int num )
  {
     TOT_TRIAL_NUM = num;
  }

  public void setCurrentTime( int curTime )
  {
    currentTime = curTime;
  }

  public void setRuleCosts( double val )
  {
     RULE_COST = val;
  }

  public void setRuleActionTimes( double val )
  {
     RULE_ACTION_TIME = val;
  }

  public void setRulePreBLAs( double val )
  {
     RULE_PRE_BLA = val;
  }

  public void setInitBLA( double iBLA )
  {
      INIT_BLA = iBLA;
  }

  // lower RER  thresholds.
  public void setThreshold( double th1, double th2 )
  {
      threshold1 = th1;
      threshold2 = th2;
  }

  public void setRTFlag( boolean flag )
  {
      RT_FLAG = flag;
  }

  public void setUtilityFlag( boolean flag )
  {
      UTL_FLAG = flag;
  }

  public void setFullFlag( boolean flag )
  {
      FULL_FLAG = flag;
  }

  public void calcRT()
  {
       responseTime = 0;
       if( chosenActionNum == 0 )
       {
           System.out.println( "no chosen rule!" );
           return;
       }

       int idx = 0, ruleIdx = 0;
       T_LinkedNode cur = ruleLink.getSible();
       for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
       {
            if( chosenActionRules[ idx ] == ruleIdx )
            {
                idx++;
                cur.getElement().calcBLA( currentTime );
                cur.getElement().setRuleCost( RULE_COST );
                cur.getElement().setRuleActionTime( RULE_ACTION_TIME );
                cur.getElement().setRulePreBLA( RULE_PRE_BLA );
                cur.getElement().calcTLDT();
                responseTime += cur.getElement().getResponseTime();
            }
            ruleIdx++;
       }

       responseTime /= chosenActionNum;
  }

  public void deleteRule( int[] state, int action )
  {
      T_LinkedNode pre, cur;
      boolean retVal;

      pre = ruleLink;
      cur = pre.getSible();
      if ( cur == null )
           return;

      while( cur != null )
      {
          retVal = cur.getElement().match( state, action, false );
          if( retVal )
          {
              deleteNode( 0, pre, cur );
              cur = null;                      // modified on 12/13/01
              cur = pre.getSible();
          }
          else
              cur = cur.getSible();
      }
  }

  // this function is used for TOH task
  public void enforceRule( int[] state, int action )
  {
      T_LinkedNode cur;
      boolean retVal1, retVal2;

      for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
      {
          if ( cur == null )
               break;

          retVal1 = cur.getElement().matchCondition( state );
          if( retVal1 )
          {
              retVal2 = cur.getElement().matchAction( action );
              if( retVal2 )
              {
                  if( UTL_FLAG )
                      cur.getElement().incUtility( 2 );
                  break;
              }
              else
              {
                  if( UTL_FLAG )
                      cur.getElement().decUtility( 2 );
                  break;
              }
          }
      }
  }

  // this function is used for TOH task when there is only one disc used
  public void createKnownRule( int[] initState, int goalAction, int dimIdx )
  {
      /*
      T_LinkedNode node = new  T_LinkedNode( DIMENSION_NUM, VALID_DIM_NUM, DIM_VAL_NUM,
                                         UTL_FLAG, false, FULL_FLAG, RT_FLAG,
                                         initState, goalAction, TOT_TRIAL_NUM, INIT_BLA );
      node.getElement().generalADim( dimIdx );
      extraction( node );
      */
  }

  // main method of the class T_RerRuleSet. update current RER rule set
  // at each trial. Now chosen rules only from active rules.
  public void updateRuleSet( int[] state, int action, boolean successFlag,
                              boolean discountFlag, boolean refineFlag )
  {
      T_LinkedNode cur, chd;
      int counter = 0;
      boolean retVal, rVal;

      /* search all the matching rules both from the active and inactive nodes.
         and update all the statistics of each matching rule. */

      matchAllNode[ action ].getElement().updateMatchAll( state, successFlag );
      chosenRules.delete( 0, chosenRules.length() );

      for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
      {
          if ( cur == null )
               break;

          retVal = cur.getElement().generalMatch( state, action, successFlag );
          retVal = cur.getElement().match( state, action, successFlag );
          if( retVal == true )
          {
              counter++;
               if( UTL_FLAG )
                   cur.getElement().updateFeature( successFlag, currentTime );

              chosenRules.append( cur.getElement().getRuleString() );
              chd = cur.getChild();
              while( chd != null )
              {
                  rVal = cur.getElement().generalMatch( state, action, successFlag );
                  rVal = chd.getElement().match( state, action, successFlag );
                  if( rVal == true )
                  {
                      counter++;
                      if( UTL_FLAG )
                          cur.getElement().updateFeature( successFlag, currentTime );
                  }
                  chd = chd.getSible();
              }
          }
      }

      chosenCounter += counter;
      if( counter == 0 )   // no rule covers this step in the top level.
      {
         if( successFlag )     // extract a new rule.
         {
             T_LinkedNode node = null;
             if( genMode == BASIC_MODE )
                 node = new T_LinkedNode( DIMENSION_NUM, DIM_VAL_NUM,
                                        UTL_FLAG, false, FULL_FLAG, RT_FLAG,
                                        state, action, TOT_TRIAL_NUM, INIT_BLA );
             else
                 node = new T_LinkedNode( DIMENSION_NUM, VALID_DIM_NUM, DIM_VAL_NUM,
                                        UTL_FLAG, false, FULL_FLAG, RT_FLAG,
                                        state, action, TOT_TRIAL_NUM, INIT_BLA );
             extraction( node );
         }
      }
      else
      {
         if( refineFlag )
         {
             //genealization and specialization only for the active rules
             for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
             {
                  retVal = cur.getElement().matchCondition( state );
                  if( !retVal )
                      continue;

                  retVal = cur.getElement().matchAction( action );
                  if( !retVal )
                      continue;

                  boolean genFlag = cur.getElement().canGeneralized();
                  boolean specFlag = cur.getElement().canSpecialized();
                  double ig1 = infoGain( cur.getElement(),
                                         matchAllNode[ action ].getElement() );
                  if( successFlag && genFlag )
                  {
                      double ig2 = maxInfoGain( cur.getElement(), 0 );
                      if( ( ig1 > threshold1 ) && ( ig2 >= 0 ) )
                            generalization( cur );
                  }
                  if( !successFlag && specFlag )
                  {
                      double ig3 = maxInfoGain( cur.getElement(), 1 );
                      if( ( ig1 < threshold2 ) && ( ig3 > 0 ) )
                            specialization( cur );
                  }
           }
        }
      }

      if( genMode == BASIC_MODE )
      {
          handleUselessRule();
          mergeActiveRule();
      }

      updateRule( discountFlag );
  }

/*************************** PRIVATE METHODS **********************************/

   // measure for goodness of a rule
   private double infoGain( T_RerRule rule1, T_RerRule rule2 )
   {
      double log2 = Math.log( 2 );
      double rate1 = rule1.getMatchRate();
      double rate2 = rule2.getMatchRate();

      double ig = Math.log( rate1 / rate2 ) / log2;
      return ig;
   }

   // return maxIG(C',C)
   private double maxInfoGain( T_RerRule rule, int mode )
   {
      int[] retVal = new int[2];
      double rate1 = 0.0, rate2 = 1.0;

      if( genMode == BASIC_MODE )
      {
          retVal = rule.getOptCondition( mode );
          rate1 = rule.getAStatRate( retVal[0], retVal[1], mode );
          rate2 = rule.getMatchRate();
      }

      if( genMode == EX_MODE )
      {
          rate1 = ((T_RerRuleEx)rule).getAMaxRate( mode );
          rate2 = ((T_RerRuleEx)rule).getMatchRate();
      }

      double log2 = Math.log( 2 );
      double ig = Math.log( rate1 / rate2 ) / log2;
      return ig;
   }

   // extract a new rule from the current state and the chosen action.
   private void extraction( T_LinkedNode node )
   {
      curRuleNum = getActiveRuleNum();
      node.getElement().toString();

      if( curRuleNum < RULE_NUM )
          insertNode( 0, ruleLink, node );
      else
      {
          deleteOldestRule();
          insertNode( 0, ruleLink, node );
      }
   }

  /* search the oldest rule according tO the # of steps occurred since the */
  /* last  application of the rule and delete it */
   private void deleteOldestRule( )
   {
      T_LinkedNode oldP, newP, pre, cur;
      int age, maxAge;

      oldP = ruleLink;
      newP = ruleLink.getSible();
      pre = ruleLink;
      cur = ruleLink.getSible();

      maxAge = cur.getElement().getAge();

      for( cur = ruleLink.getSible(); cur !=null; cur = cur.getSible() )
      {
         age = cur.getElement().getAge();
         if( maxAge < age )
         {
            maxAge = age;
            oldP = pre;
            newP = cur;
         }
         pre = cur;
      }

      deleteNode( 0, oldP, newP );

      // free the oldest rule node newP itself.    // modified on 12/13/01
      newP = null;
   }

   // generalization amounts to adding an additional value to one input
   // dimension in the condition of a rule, so that the rule will have
   // more opportunities of matching inputs.
   private void generalization( T_LinkedNode node )
   {
      int[] loc  = new int[2];
      T_RerRule rule = node.getElement();

      if( genMode == BASIC_MODE )
      {
        //System.out.print( "before gen:" + rule.getRuleString().substring( 0 ) );
        loc = rule.getMaxIGLoc( 0 );
        rule.addAValue( loc[0], loc[1] );
        rule.toString();
        //System.out.print( "after gen:" + rule.getRuleString().substring( 0 ) );
      }

      if( genMode == EX_MODE )
      {
          ( (T_RerRuleEx)rule ).generalizeADim();
          ( (T_RerRuleEx)rule ).toString();
      }

      // added on 12/04/01.
      rule.resetStatistics();
   }

   // specialization amounts to removing one value from one input
   // dimension in the condition of a rule, so that the rule will have
   // less opportunities of matching inputs.
   private void specialization( T_LinkedNode node )
   {
      int[] loc = new int[2];
      T_RerRule rule = node.getElement();

      if( genMode == BASIC_MODE )
      {
        //System.out.print( "before spec: " + rule.getRuleString().substring( 0 ) );
        loc = rule.getMaxIGLoc( 1 );
        rule.delAValue( loc[0], loc[1] );
        rule.toString();
        restoreChildrenNode( node );
        //System.out.print( "after spec: " + rule.getRuleString().substring( 0 ) );
      }

      if( genMode == EX_MODE )
      {
          ( (T_RerRuleEx)rule ).specializeADim();
          ( (T_RerRuleEx)rule ).toString();
      }

      // added on 12/04/01.
      rule.resetStatistics();
   }

 /* restore those rules in the children list of the original rule that are
    not covered by the specialized rule and the other existing rules.*/
   private void  restoreChildrenNode( T_LinkedNode node )
   {
      T_LinkedNode pre, cur, activeP, restoreLink;
      int retVal, counter = 0;

      pre = node;
      cur = node.getChild();

      if( cur == null )
          return;

      restoreLink = new T_LinkedNode( DIMENSION_NUM, DIM_VAL_NUM, UTL_FLAG, RT_FLAG );

      while( cur != null )
      {
          counter = 0;
          for( activeP = ruleLink.getSible(); activeP != null;
               activeP = activeP.getSible() )
          {
             retVal = cover( activeP.getElement(), cur.getElement() );
             if ( retVal == -2 )             // is included entirely.
                  counter++;
          }

          if( counter == 0 )          /* isn't included by any existing rule,
                                        restore child. */
          {
             if( pre.equals( node ) )
                 deleteNode( 1, pre, cur );
             else
                 deleteNode( 0, pre, cur );

             insertNode( 0, restoreLink, cur );
             cur = pre.getSible();
          }
          else
          {
             pre = cur;
             cur = cur.getSible();
          }
      }

      if( restoreLink.getSible() != null )
          insertNode( 0, ruleLink, restoreLink.getSible() );
   }

  /* search if any useless rule exists according to the age of the
     rule and delete it. */
  private void handleUselessRule()
   {
      T_LinkedNode pre, cur;
      int age;

      pre = ruleLink;
      cur = ruleLink;

      age = cur.getElement().getAge();

      while ( cur != null )
      {
         age = cur.getElement().getAge();
         if( age >= OLDEST_AGE )
         {
             deleteNode( 0, pre, cur );
             cur = null;              // modified on 12/13/01
             cur = pre.getSible();
         }
         else
         {
            pre = cur;
            cur = cur.getSible();
         }
      }
  }

  // merge rules according to what the page 11 says.
  private void mergeActiveRule()
  {
      T_LinkedNode node;
      T_LinkedNode preCur = ruleLink;
      T_LinkedNode cur = ruleLink.getSible();
      T_LinkedNode pre = ruleLink;
      T_LinkedNode next = ruleLink.getSible();

      int retVal = 0;
      int[] arr1 = new int[ DIM_VAL_NUM ];
      int[] arr2 = new int[ DIM_VAL_NUM ];

      String str = new String("");
      T_RerRule rule = null;

      // no merge at all or rule set is null
      if( ( whenMerge == 0 ) || ( preCur == null ) || ( cur == null ) )
            return;

      // merge every 10 trials.
      if( whenMerge == 1 )
      {
          if ( timeSteps < STEP_INTERVAL )
          {
               timeSteps++;
               return;
          }
          else
             timeSteps = 0;
      }

      while( cur != null )
      {
          pre  = ruleLink;
          next = ruleLink.getSible();

          if( next == null )
              return;

          while ( next != null )
          {
             if( next.equals( cur ) )
             {
                 pre = next;
                 next = next.getSible();
                 continue;
             }

             retVal = cover( cur.getElement(), next.getElement() );

             // not include at all.
             if( retVal == -1 )
             {
                 pre = next;
                 next = next.getSible();
             }

             // include entirely
             if( retVal == -2 )
             {
                 if( next.equals( preCur ) )
                 {
                    deleteNode( 0, pre, next );
                    insertNode( 1, cur, next );
                    preCur = pre;
                    next = pre.getSible();
                 }
                 else
                 {
                    deleteNode( 0, pre, next );
                    insertNode( 1, cur, next );
                    next = pre.getSible();
                 }
             }

             // need merge
             if( retVal >= 0 )
             {
                  /*
                  System.out.println( retVal );
                  str = cur.getElement().getRuleString().substring( 0 );
                  System.out.print( " cur : " + str );
                  str = next.getElement().getRuleString().substring( 0 );
                  System.out.print( " next : " + str );
                   */
               // create a new rule with the same statistics as the bigger rule.
                  //rule.copy( cur.getElement() );
                  if( genMode == BASIC_MODE )
                      rule = (T_RerRule)cur.getElement().clone();
                  else
                      rule = (T_RerRuleEx)cur.getElement().clone();
                  node = new T_LinkedNode( rule );

                  cur.getElement().getOneDimension( retVal, arr1 );
                  next.getElement().getOneDimension( retVal, arr2 );

                  for( int i = 0; i < DIM_VAL_NUM; i++ )
                  {
                     if( arr1[ i ] != arr2[ i ] )
                         node.getElement().addAValue( retVal, i );
                  }

                  node.getElement().toString();

                  /*
                  str = cur.getElement().getRuleString().substring( 0 );
                  System.out.print( " cur : " + str );
                  str = next.getElement().getRuleString().substring( 0 );
                  System.out.print( " next : " + str );

                  str = node.getElement().getRuleString().substring( 0 );
                  System.out.print( " parent : " + str );
                  */
                 if( cur.equals( pre ) )
                 {
                     deleteNode( 0, preCur, cur );
                     deleteNode( 0, preCur, next);
                     insertNode( 0, preCur, node );
                     insertNode( 1, node, cur );
                     insertNode( 1, node, next );
                     cur = node;
                     pre = node;
                     next = pre.getSible();
                 }
                 else if( next.equals( preCur ) )
                      {
                         deleteNode( 0, pre, next );
                         deleteNode( 0, pre, cur );
                         insertNode( 0, pre, node );
                         insertNode( 1, node, cur );
                         insertNode( 1, node, next );
                         preCur = pre;
                         cur =  preCur.getSible();
                         next = pre.getSible();
                      }
                      else
                      {
                         deleteNode( 0, preCur, cur );
                         deleteNode( 0, pre, next );
                         insertNode( 0, preCur, node );
                         insertNode( 1, node, cur );
                         insertNode( 1, node, next );
                         cur = node;
                         next = pre.getSible();
                      }
             }
          }   // end of while next

          preCur = cur;
          if( cur != null )
              cur = cur.getSible();
      } // end of while cur
   }

  //if rule1 covers  rule2, including both condition and action.
  private int cover( T_RerRule rule1, T_RerRule rule2 )
  {
      int rVal = -1, idx = 0, count = 0;
      boolean retVal;

      if( ( rule1 == null ) || ( rule2 == null ) )
          return -1;

      if( rule1.getAction() !=  rule2.getAction() )
          return -1;

      if( rule1.getAction() < 0 )
          return -1;

      for( int i = 0; i < DIMENSION_NUM; i++ )
      {
         retVal = contain( rule1.getStateX()[i], rule2.getStateX()[i], DIM_VAL_NUM );
         if ( !retVal )
         {
             count++;
             idx = i;
         }
      }

      switch ( count )
      {
         case 0:                 // include entirely
             rVal = -2;
             break;
         case 1:                 // include except one dimension.
             rVal = idx;
             break;
         default:                // not include at all
             rVal = -1;
             break;
      }

      return rVal;
  }

  // if array a1 contains array a2.
  private boolean contain( int[] a1, int[] a2, int maxIdx )
  {
       for( int i = 0; i < maxIdx; i++ )
            if ( ( a1[i] == 0 ) && ( a2[i] == 1 ) )
                 return false;

       return true;
  }

   /* update age which is used for deleting an oldest rule when rule number
      reaches  the max number, discount the statistics of all the rules.*/
   private void updateRule(  boolean discountFlag )
   {
       T_LinkedNode chd = ruleLink;
       T_LinkedNode cur = ruleLink;

       for( int i = 0; i < OUTPUT_NUM; i++)
       {
          matchAllNode[i].getElement().aging();

          // modified on 04/16/02
          /*
          if( UTL_FLAG )
              matchAllNode[i].getElement().updateFeature( succFlag );
          */

          if( discountFlag )
               matchAllNode[i].getElement().discountStatistics();
       }

       for( cur = ruleLink.getSible(); cur != null; cur = cur.getSible() )
       {
            cur.getElement().aging();

            // modified on 04/16/02
            /*
            if( UTL_FLAG )
                cur.getElement().updateFeature( succFlag );
            */

            if( discountFlag )
                cur.getElement().discountStatistics();

            chd = cur.getChild();
            while( chd != null )
            {
               chd.getElement().aging();
               // modified on 04/16/02
               /*
               if( UTL_FLAG )
                   cur.getElement().updateFeature( succFlag );
               */

               if( discountFlag )
                   chd.getElement().discountStatistics();

               chd = chd.getSible();
            }
       }
   }

   // insert a node p2 after node p1. mode 0: sible, mode 1: child
   private void insertNode( int mode, T_LinkedNode p1, T_LinkedNode p2 )
   {
      T_LinkedNode chd = p2;

      if ( p1 == null || p2 == null )
           return;

      chd.setChild( null );

      while( chd.getSible() != null )
             chd = chd.getSible();

      if( mode == 0 )
      {
           chd.setSible( p1.getSible() );
           p1.setSible( p2 );
      }
      else
      {
           chd.setSible( p1.getChild() );
           p1.setChild( p2 );
      }
   }

   /* delete a node p2 after the node p1. mode 0: sible, mode 1:child
      whenever a node is deleted, its children will be in the same
      hierarchy as itself, that is, node p2 will be flat without hierarchy */
   private void deleteNode( int mode, T_LinkedNode p1, T_LinkedNode p2 )
   {
       if ( p1 == null || p2 == null )
            return;

       if( mode == 0 )
       {
           p1.setSible( p2.getSible() );
           //make p2 and its children in the same hierarchy.
           p2.setSible( p2.getChild() );
           p2.setChild( null );
       }
       else
       {
           p1.setChild( p2.getSible() );
           free( p2 );
       }
   }

   // free the memory taken by node p.
   private void free( T_LinkedNode p )
   {
        p.setSible( null );
        p.setChild( null );
   }
}