package tasksimulators.oldsimu;
import clarion.common.*;
import clarion.tools.*;

public class TaskManager {

  // variables for agent of clarion learning.
   private PCAgent clarionAgent[];

   private final int CONTROL_GROUP = 0;
   private final int ORIGINAL_GROUP = 1;
   private final int MEMORY_GROUP = 2;
   private final int SIMPLE_GROUP = 3;

   private final int MAX_BLOCK_NUM = 40;
   private final int MAX_AGENT_NUM = 20;
   private final int MAX_GROUP_NUM = 4;
   private final int MAX_TASK_NUM = 2;
   private final int MAX_SET_NUM = 3;

   private final boolean YES_IRL_RULE = true;
   private final boolean NO_IRL_RULE = false;
   private final boolean YES_RER_RULE = true;
   private final boolean NO_RER_RULE = false;
   private final boolean YES_FIX_RULE = true;
   private final boolean NO_FIX_RULE = false;
   private final boolean YES_SMP_RULE = true;
   private final boolean NO_SMP_RULE = false;

   private final int IRL_IDX = 0;
   private final int RER_IDX = 1;
   private final int FIX_IDX = 2;
   private final int SMP_IDX = 3;

   private final int RULE_TYPE_NUM = 4;

   private final boolean initRuleSet[][][] =
           {
             {
                { NO_IRL_RULE, YES_RER_RULE }, { YES_IRL_RULE, YES_RER_RULE }
             },
             {
                { YES_IRL_RULE, NO_RER_RULE }, { YES_IRL_RULE, YES_RER_RULE }
             },
             {
                { YES_IRL_RULE, YES_RER_RULE }, { NO_IRL_RULE, YES_RER_RULE },
                { YES_IRL_RULE, NO_RER_RULE }
             }
           };

   private boolean preTrainFlag, finishFlag, ruleFlagArray[];

   private int runNo = 2;
   private int agentIdx, groupIdx, taskIdx, ruleSetIdx, decidedAction;
   private int startArray[][], trainBlockNums[][], onTargetNums[][][];

   private double onTargetRates[][], totOnTargetRates[];
   private double IHW[][], HOW[][];

   final double[] DC_HUMAN_DATA = { 1.97, 2.57, 4.63, 4.00 };
   final double[] PT_HUMAN_DATA = { 2.85, 3.75, 5.53, 5.91 };

   private double[][] DC_MODEL_DATA, PT_MODEL_DATA;

   private double dcMeanSqErr, ptMeanSqErr, totMeanSqErr;

   static int STATUS_NUMBER = 4;
   static int STATUS_ACTION = 0, STATUS_EQUATION = 1,
              STATUS_EACH = 2,   STATUS_TOTAL = 3;

   private String statusText[] =
   {
       "Action = ",    "Equation = \n",
       "LR = ",        "Group's LR = "
   };

   private String rerRuleText[] =
   {
      "RER'S ALL RULES LIST: \n", " RER'S MATCH RULES LIST: \n",
      "RER'S CHOSEN RULES LIST: \n"
   };

   private String irlRuleText[] =
   {
      "IRL'S ALL RULES LIST: \n", " IRL'S MATCH RULES LIST: \n",
      "IRL'S CHOSEN RULES LIST: \n"
   };

   private StringBuffer rerRules[], irlRules[], fixRules[];
   private String  eachLRStr, totalLRStr;

   private int trialIdx, blockIdx, blockNum;
   private int NET_INPUT_NUM, NET_HIDDEN_NUM, NET_OUTPUT_NUM;
   private int trainArray[][][];

   private static int TRIALS_PER_BLOCK = 11,
                      INPUT = 0,  OUTPUT = 1, ACTION = 2,
                      MIN_LEVEL = 1,    MAX_LEVEL = 12,
                      MIN_OUTPUT = 1,   MAX_OUTPUT = 12,
                      PRE_TRAIN_BLOCK_NUM = 20;

    String ModelFileName = "MODELDAT";
    WriteFile modelDataFile = new WriteFile();
    String FileNameArr[] = { "PCTASK", "PersonTASK" };
    WriteFile dataFile[][] = new WriteFile[MAX_TASK_NUM][MAX_GROUP_NUM];
   //WriteFile  dataFile = new WriteFile();

   // Constructor method
   public TaskManager( int bNum, WriteFile file )
   {
     NET_INPUT_NUM = ( int )World.TASK_PARA_ARRAY[World.NET_INPUT_NUM ];
     NET_HIDDEN_NUM = ( int )World.TASK_PARA_ARRAY[World.NET_HIDDEN_NUM ];
     NET_OUTPUT_NUM = ( int )World.TASK_PARA_ARRAY[World.NET_OUTPUT_NUM ];

     //dataFile = file;
     createDataFiles();
     clarionAgent = new PCAgent[ MAX_AGENT_NUM ];

     // initialize variables on recording the training data.
     trialIdx = 0;
     blockIdx = 0;
     blockNum = bNum;

     trainArray = new int[ MAX_BLOCK_NUM ][ TRIALS_PER_BLOCK ][ 3 ];
     //trainArray = new int[ blockNum ][ TRIALS_PER_BLOCK ][ 3 ] ;

     // learning agent initializaion.
     ruleSetIdx = 0;
     taskIdx = 0;
     groupIdx = 0;
     agentIdx = 0;

     preTrainFlag = false;
     finishFlag = false;

     trainBlockNums   = new int[ MAX_GROUP_NUM ][ MAX_AGENT_NUM ];
     onTargetNums     = new int[ MAX_GROUP_NUM ][ MAX_AGENT_NUM ][ MAX_BLOCK_NUM ];
     //onTargetNums     = new int[ MAX_GROUP_NUM ][ MAX_AGENT_NUM ][ blockNum ];
     onTargetRates    = new double[ MAX_GROUP_NUM ][ MAX_AGENT_NUM ];
     totOnTargetRates = new double[ MAX_GROUP_NUM ];

     DC_MODEL_DATA    = new double[ MAX_SET_NUM ][ MAX_GROUP_NUM ];
     PT_MODEL_DATA    = new double[ MAX_SET_NUM ][ MAX_GROUP_NUM ];

     IHW = new double[ NET_INPUT_NUM ][ NET_HIDDEN_NUM ];
     HOW = new double[ NET_HIDDEN_NUM][ NET_OUTPUT_NUM ];

     eachLRStr  = new String( statusText[ STATUS_EACH ] );
     totalLRStr = new String( statusText[ STATUS_TOTAL ] );

     rerRules = new StringBuffer[3];
     irlRules = new StringBuffer[3];
     fixRules = new StringBuffer[3];

     for( int i = 0; i < 3; i++ )
     {
         rerRules[ i ] = new StringBuffer();
         irlRules[ i ] = new StringBuffer();
         fixRules[ i ] = new StringBuffer();
     }

     startArray = new int[ MAX_AGENT_NUM ][ 2 ];
     for( int i = 0; i < MAX_AGENT_NUM; i++ )
     {
        int v1 = ( int ) ( Math.random() * ( MAX_LEVEL - MIN_LEVEL ) );
        v1 += MIN_LEVEL;
        if( v1 > MAX_LEVEL )
            v1 = MAX_LEVEL;
        startArray[ i ][ INPUT ] = v1;

        int v2 = ( int ) ( Math.random() * ( MAX_LEVEL - MIN_LEVEL ) );
        v2 += MIN_LEVEL;
        if( v2 > MAX_LEVEL )
            v2 = MAX_LEVEL;
        startArray[ i ][ OUTPUT ] = v2;

        clarionAgent[ i ] = new PCAgent( v1, v2, World.OUTPUT_TARGET_LEVEL );
     }

     ruleFlagArray = new boolean[ RULE_TYPE_NUM ];
     initNewRuleSet();
   }

   public void createDataFiles()
   {
       for( int i = 0; i < MAX_TASK_NUM; i++ )
         for( int j = 0; j < MAX_GROUP_NUM; j++ )
         {
              dataFile[i][j] = new WriteFile();
              dataFile[i][j].openFile( 1, FileNameArr[i]);
         }

      modelDataFile.openFile( 1, ModelFileName );
   }


   public void trainMain()
   {
      trialIdx = 1;
      decidedAction = clarionAgent[agentIdx].decideAction( trialIdx );
      trainArray[ blockIdx ][ trialIdx - 1 ][ ACTION ] = decidedAction + 1;

      while( !finishFlag )
      {
           int value = decidedAction + 1;

           trainArray[ blockIdx ][ trialIdx ][ INPUT ] = value;
           trainArray[ blockIdx][ trialIdx ][ OUTPUT ] = calcOutput( value );

           // update clarion agent information.
           int v1 = trainArray[blockIdx][trialIdx][ OUTPUT ] - 1;

           boolean retVal = clarionAgent[ agentIdx ].update( v1,
                   onTargetNums[groupIdx][agentIdx],blockIdx,trialIdx );

           if( retVal )
               gotoNextTrial();
           else
               gotoEndThisAgent();
      }
   }

  /* CALCULATE THE OUTPUT ACCORDING TO DIFFERENT EQUATIONS.*/
   public int calcOutput( int inputVal )
   {
      int w, n, iRand;
      double a, b, p0, p1;

      if ( trialIdx == 0 )
      {
        iRand  = ( int ) ( Math.random() * ( MAX_LEVEL - MIN_LEVEL ) );
        iRand += MIN_LEVEL;

        return iRand;
      }

       w  = inputVal;
       p0 = trainArray[ blockIdx][ trialIdx  - 1 ][ OUTPUT ];
       p0 = p0 * 1;
       iRand = ( int ) ( Math.random() * 3 );

       switch ( iRand ) {
          case 0:
                  n = 1;
                  break;
          case 1:
                  n = 0;
                  break;
          case 2:
                  n = - 1;
                  break;
          default:
                  n = 0;
       }

       a = 2;
       b = 0;
       p1 = ( int )( a * w + b - p0 + n);

       if ( p1 > MAX_OUTPUT )
            p1 = MAX_OUTPUT;
       else if ( p1 < MIN_OUTPUT )
                 p1 = MIN_OUTPUT;

       return (int) p1;
   }

   // generate first output of every block of trials.
   public void genFirstOutput( int mode )
   {
     if( mode == 0 )
     {
         int value = ( int ) ( Math.random() * ( MAX_LEVEL - MIN_LEVEL ) );
         value += MIN_LEVEL;
         trainArray[ blockIdx ][ trialIdx ][ INPUT ] = value;

         value = ( int ) ( Math.random() * ( MAX_LEVEL - MIN_LEVEL ) );
         value += MIN_LEVEL;
         trainArray[ blockIdx ][ trialIdx ][ OUTPUT ] = value;
         //trainArray[ blockIdx][ trialIdx ][ OUTPUT ] = calcOutput( value );
     }
     else
         clarionAgent[ agentIdx ].getInitVals( trainArray[blockIdx][trialIdx] );
  }

  /*************************** intialization functions ************************/
   public void beginTraining()
   {
         blockIdx = 0;
         trialIdx = 0;

         preTrainFlag = false;
         for( int i = 0; i < PRE_TRAIN_BLOCK_NUM; i++ )
              onTargetNums[ groupIdx ][ agentIdx ][ i ] = 0;

         genFirstOutput( 1 );
   }

   public void initNewBlock()
   {
       trialIdx = 0;
       genFirstOutput( 0 );

       int v1 = trainArray[ blockIdx ][ trialIdx ][ INPUT ];
       int v2 = trainArray[ blockIdx ][ trialIdx ][ OUTPUT ];
       clarionAgent[ agentIdx ].reinitBlock( v1, v2 );
   }

   public void initNewAgent()
   {
      if( (agentIdx % 10 ) == 0 )
      {
          System.out.print( " cur agent no =  " );
          System.out.println( agentIdx );
      }

       if( taskIdx == 1 )
           preTrainFlag = true;
       else
           preTrainFlag = false;

       clarionAgent[ agentIdx ].setRules( ruleFlagArray );
       clarionAgent[ agentIdx ].reinitAgent();
       if( groupIdx > CONTROL_GROUP )
           clarionAgent[ agentIdx ].setRuleThreshold();

       blockIdx = 0;
       trialIdx = 0;
       genFirstOutput( 1 );
   }

   public void initNewGroup()
   {
       System.out.print( " cur agent group no =  " );
       System.out.println( groupIdx );

       eachLRStr = statusText[ STATUS_EACH ];

       if ( groupIdx >= MEMORY_GROUP )
            ruleFlagArray[ FIX_IDX ] = YES_FIX_RULE;
       else
            ruleFlagArray[ FIX_IDX ] = NO_FIX_RULE;

       if ( groupIdx == SIMPLE_GROUP )
            ruleFlagArray[ SMP_IDX ] = YES_SMP_RULE;
       else
            ruleFlagArray[ SMP_IDX ] = NO_SMP_RULE;

       agentIdx = 0;
       initNewAgent();
   }

   public void initNewTask()
   {
       System.out.print( " cur task no =  " );
       System.out.println( taskIdx );

       for( int i = 0; i <  MAX_GROUP_NUM; i++ )
         for( int j = 0; j <  MAX_AGENT_NUM; j++ )
            for( int k = 0; k <  MAX_BLOCK_NUM; k++ )
        //   for( int k = 0; k <  blockNum; k++ )
                onTargetNums[ i ][ j ][ k ] = 0;

       groupIdx = 0;
       initNewGroup();
   }

   public void initNewRuleSet()
   {
       System.out.print( " cur rule set no =  " );
       System.out.println( ruleSetIdx );

       ruleFlagArray[ IRL_IDX ] = initRuleSet[ runNo ][ ruleSetIdx ][ 0 ];
       ruleFlagArray[ RER_IDX ] = initRuleSet[ runNo ][ ruleSetIdx ][ 1 ];

       taskIdx = 0;
       initNewTask();
   }

   public void allFinish()
   {
       blockIdx = 0;
       trialIdx = 0;
       agentIdx = 0;
       groupIdx = 0;
       taskIdx = 0;
       ruleSetIdx = 0;

       preTrainFlag = false;
       finishFlag = true;

       for( int i = 0; i <  MAX_GROUP_NUM; i++ )
         for( int j = 0; j <  MAX_AGENT_NUM; j++ )
           for( int k = 0; k <  MAX_BLOCK_NUM; k++ )
           //for( int k = 0; k <  blockNum; k++ )
                onTargetNums[ i ][ j ][ k ] = 0;

       eachLRStr = statusText[ STATUS_EACH ];
       saveModelData();
   }

   public void gotoNextTrial()
   {
        boolean closeFlag = false;

        trialIdx++;

        if ( trialIdx == TRIALS_PER_BLOCK )
        {
             if( preTrainFlag )
             {
                 if ( ( blockIdx + 1 ) == PRE_TRAIN_BLOCK_NUM )
                        beginTraining();
                 else
                 {
                       blockIdx++;
                       initNewBlock();
                 }
             }
             else
             {
                if ( ( blockIdx + 1  ) < blockNum )
                {
                    blockIdx++;
                    initNewBlock();
                }
                else
                {
                  //saveData( closeFlag );
                  saveOnTargets( 0 );
                  saveRules();
                  //saveParameters( agentIdx );

                  if( ( agentIdx + 1 ) < MAX_AGENT_NUM )
                  {
                      agentIdx++;
                      initNewAgent();
                  }
                  else
                  {
                      saveOnTargets( 1 );

                     if( ( groupIdx + 1 ) < MAX_GROUP_NUM )
                     {
                         groupIdx++;
                         initNewGroup();
                     }
                     else
                     {
                        if( ( taskIdx + 1 ) < MAX_TASK_NUM )
                        {
                           for( int i = 0; i < MAX_GROUP_NUM; i++ )
                           {
                                DC_MODEL_DATA[ruleSetIdx][ i ] = totOnTargetRates[ i ];
                                DC_MODEL_DATA[ruleSetIdx][ i ] *= TRIALS_PER_BLOCK - 1;
                           }

                           taskIdx++;
                           initNewTask();
                        }
                        else
                        {
                           for( int i = 0; i < MAX_GROUP_NUM; i++ )
                           {
                              PT_MODEL_DATA[ruleSetIdx][i]=totOnTargetRates[i];
                              PT_MODEL_DATA[ruleSetIdx][i] *= TRIALS_PER_BLOCK-1;
                           }

                           if( ruleSetIdx + 1 < MAX_SET_NUM )
                           {
                               ruleSetIdx++;
                               initNewRuleSet();
                           }
                           else
                           {
                               allFinish();
                               closeFlag = true;
                           }
                        }
                     }
                   }
                }
              }

              trialIdx++;
         }

         decidedAction = clarionAgent[agentIdx].decideAction( trialIdx );
         trainArray[ blockIdx ][ trialIdx - 1 ][ ACTION ] = decidedAction + 1;
   }

   public void gotoEndThisAgent()
   {
        boolean closeFlag = false;

        //saveData( closeFlag );
        saveOnTargets( 0 );
        saveRules();
        //saveParameters( agentIdx );

        if( agentIdx + 1 < MAX_AGENT_NUM )
        {
            agentIdx++;
            initNewAgent();
        }
        else
        {
            saveOnTargets( 1 );
            if( groupIdx + 1 < MAX_GROUP_NUM )
            {
                groupIdx++;
                initNewGroup();
            }
            else
            {
                if( ( taskIdx + 1 ) < MAX_TASK_NUM )
                {
                    for( int i = 0; i < MAX_GROUP_NUM; i++ )
                    {
                        DC_MODEL_DATA[ruleSetIdx][ i ] = totOnTargetRates[ i ];
                        DC_MODEL_DATA[ruleSetIdx][ i ] *= TRIALS_PER_BLOCK - 1;
                     }

                     taskIdx++;
                     initNewTask();
                }
                else
                {
                     for( int i = 0; i < MAX_GROUP_NUM; i++ )
                     {
                        PT_MODEL_DATA[ruleSetIdx][i]=totOnTargetRates[i];
                        PT_MODEL_DATA[ruleSetIdx][i] *= TRIALS_PER_BLOCK-1;
                     }

                      if( ruleSetIdx + 1 < MAX_SET_NUM )
                      {
                          ruleSetIdx++;
                          initNewRuleSet();
                      }
                      else
                      {
                          allFinish();
                          closeFlag = true;
                      }
                }
            }
        }
   }

   /*************************************************************************/
   /***************************** SAVE FUNCTIONS ****************************/
   /*************************************************************************/

   public void saveGroupOnTargetRate()
   {
      int totBlockNum = 0, sum = 0;

      for( int i = 0; i < MAX_AGENT_NUM; i++ )
      {
         for( int j = 0; j < trainBlockNums[ groupIdx ][ i ]; j++ )
              sum += onTargetNums[ groupIdx ][ i ][ j ];

         totBlockNum += trainBlockNums[ groupIdx ][ i ];
      }

      totOnTargetRates[ groupIdx ] = (double)sum /
                      ( totBlockNum * ( TRIALS_PER_BLOCK - 1 ) );

      totalLRStr = "Group NO = " + String.valueOf( groupIdx ) + ",  " +
                   "Total trained blocks = " + String.valueOf( totBlockNum ) +
                   ",  Total on-target rate =  ";

      totalLRStr += String.valueOf( totOnTargetRates[ groupIdx ] );
      totalLRStr += "\n";
      dataFile[taskIdx][groupIdx].saveStringData( totalLRStr );
  }

  public void saveAgentOnTargetRate()
  {
      double sum = 0, rate = 0;

      for( int i = 0; i < blockIdx + 1; i++ )
           sum += onTargetNums[ groupIdx ][ agentIdx ][ i ];

      rate = ( double ) sum / ( ( blockIdx + 1 ) * ( TRIALS_PER_BLOCK  - 1 ) );

      onTargetRates[ groupIdx ][ agentIdx ] = rate;

      eachLRStr = "Group NO = " + String.valueOf( groupIdx ) + ",  " +
                  "Agent NO = " + String.valueOf( agentIdx ) +  ",  " +
                  "Trained Blocks = " + String.valueOf( blockIdx + 1 ) + ",  " +
                  "'On-target rate = ";

      //eachLRStr = "";
      eachLRStr += String.valueOf( onTargetRates[ groupIdx ][ agentIdx ] );

      //statusArea[ STATUS_EACH ].setText( eachLRStr );
      dataFile[taskIdx][groupIdx].saveStringData( eachLRStr );
  }

   /************************* SAVE THE TRAINING DATA ************************/
   public void saveData( boolean closeFlag )
   {
       dataFile[taskIdx][groupIdx].saveNumericData( trainArray, blockNum, TRIALS_PER_BLOCK,
                                  blockIdx, trialIdx );
       /*
       if( closeFlag )
       {
          wdataFile[taskIdx][groupIdx].closeFile( c );
          saveFlag = false;
          c.removeAll();
          startExp();
       } */
   }

   public void saveModelData()
   {
     double sum = 0.0;

     for( int setNo = 0; setNo < MAX_SET_NUM; setNo++ )
     {
        String str = "\n rule set no is:  " + String.valueOf( setNo );
        str += "\n The Dynamic Control task human data are: ";
        for( int i = 0; i < MAX_GROUP_NUM; i++ )
             str += String.valueOf( DC_HUMAN_DATA[ i ] ) + "    ";
        str += "\n The Dynamic Control task model data are: ";
        for( int i = 0; i < MAX_GROUP_NUM; i++ )
             str += String.valueOf( DC_MODEL_DATA[ setNo ][ i ] ) + "    ";
        str += "\n The mean square error of DC task is: ";
        sum = 0.0;
        for( int i = 0; i < MAX_GROUP_NUM; i++ )
             sum += Math.pow(( DC_MODEL_DATA[setNo][i] - DC_HUMAN_DATA[i] ),2);
        sum /= MAX_GROUP_NUM;
        dcMeanSqErr = sum;
        str += String.valueOf( dcMeanSqErr );

        str += "\n The Person Task human data are: ";
        for( int i = 0; i < MAX_GROUP_NUM; i++ )
             str += String.valueOf( PT_HUMAN_DATA[ i ] ) + "    ";
        str += "\n The Person Task model data are: ";
        for( int i = 0; i < MAX_GROUP_NUM; i++ )
             str += String.valueOf( PT_MODEL_DATA[ setNo ][ i ] ) + "    ";
        str += "\n The mean square error of person task is: ";
        sum = 0.0;
        for( int i = 0; i < MAX_GROUP_NUM; i++ )
             sum += Math.pow(PT_MODEL_DATA[setNo][ i ] - PT_HUMAN_DATA[ i ], 2);
        sum /= MAX_GROUP_NUM;
        ptMeanSqErr = sum;
        str += String.valueOf( ptMeanSqErr );

        str += "\n The total mean square error of the two tasks is : ";
        sum = ( ptMeanSqErr + dcMeanSqErr ) / 2;
        totMeanSqErr = sum;
        str += String.valueOf( totMeanSqErr ) + "\n";

        modelDataFile.saveStringData( str );
      }
   }

   /********************** SAVE THE TRAINING PARAMETERS ***********************/
   public void saveParameters( int agentNo )
   {
      double[]  hThresholds = new double[ NET_HIDDEN_NUM ];
      double[]  oThresholds = new double[ NET_OUTPUT_NUM ];

      double[][] ItoHWeights = new double[ NET_INPUT_NUM ][ NET_HIDDEN_NUM ];
      double[][] HtoOWeights = new double[ NET_HIDDEN_NUM][ NET_OUTPUT_NUM ];

      clarionAgent[agentNo].getThresholds( hThresholds, oThresholds );
      ItoHWeights = clarionAgent[agentNo].getIHWeights();
      HtoOWeights = clarionAgent[agentNo].getHOWeights();

      String str = "\n The thresholds of hidden layer are: \n\n";
      for( int i = 0; i < NET_HIDDEN_NUM; i++ )
           str += String.valueOf( hThresholds[ i ] ) + "\n";

      dataFile[taskIdx][groupIdx].saveStringData( str );

      str = "\n The thresholds of output layer are: \n\n";
      for( int i = 0; i < NET_OUTPUT_NUM; i++ )
           str += String.valueOf( oThresholds[ i ] ) + "\n";

      dataFile[taskIdx][groupIdx].saveStringData( str );

      str = "\n The weights of input to hidden layer are: \n\n";
      for( int i = 0; i < NET_INPUT_NUM; i++ )
        for( int j = 0; j < NET_HIDDEN_NUM; j++ )
        {
           str += "["+String.valueOf(i)+"]"+"["+String.valueOf(j)+"]=  ";
           str += String.valueOf( ItoHWeights[i][j] ) + "\n";
        }

      dataFile[taskIdx][groupIdx].saveStringData( str );

      str = "\n The weights of hidden to output layer are: \n\n";
      for( int i = 0; i < NET_HIDDEN_NUM; i++ )
        for( int j = 0; j < NET_OUTPUT_NUM; j++ )
        {
           str += "["+String.valueOf(i)+"]"+"["+String.valueOf(j)+"]=  ";
           str += String.valueOf( HtoOWeights[i][j] ) + "\n";
        }

      dataFile[taskIdx][groupIdx].saveStringData( str );
   }

  /************************* SAVE THE ON-TARGET RATES  ************************/
   public void saveOnTargets( int mode )
   {
      // record the training block numbers of the current agent.
      trainBlockNums[ groupIdx ][ agentIdx ] = blockIdx + 1;
      String str = "Group NO = " + String.valueOf( groupIdx ) + "  ";
      if( mode == 0 )
          str += "Agent NO = " + String.valueOf( agentIdx ) + "  ";
       str += "On-targets of each block = \n";
       //dataFile[taskIdx][groupIdx].saveStringData( str );
       int curBlockNum = trainBlockNums[ groupIdx ][ agentIdx ];
       if( mode == 0 )
           dataFile[taskIdx][groupIdx].saveBlockData( onTargetNums[groupIdx][agentIdx],
                                 curBlockNum, 1 );
       else
       {
          int[] sum = new int[ blockNum ];

          for( int i = 0; i < blockNum; i++ )
             for( int j = 0; j < MAX_AGENT_NUM; j++ )
                  sum[i] +=  onTargetNums[groupIdx][j][i];

          dataFile[taskIdx][groupIdx].saveBlockData( sum,  blockNum, MAX_AGENT_NUM );
       }

       if( mode == 0 )
           saveAgentOnTargetRate();
       else
           saveGroupOnTargetRate();
   }

   /************************* SAVE THE LEARNED RULES ************************/
   public void saveRules()
   {
      dataFile[taskIdx][groupIdx].saveStringData( " After the training of an agent,  " );

      if( ruleFlagArray[ IRL_IDX ] )
      {
        clarionAgent[ agentIdx ].fillIrlRule( irlRules );
        for( int i = 0; i < 3; i++ )
             dataFile[taskIdx][groupIdx].saveStringData(  irlRuleText[ i ] +
                                       irlRules[ i ].substring( 0 ) );
      }

     if( ruleFlagArray[ RER_IDX ]  )
     {
        clarionAgent[ agentIdx ].fillRerRule( rerRules );
        for( int i = 0; i < 3; i++ )
             dataFile[taskIdx][groupIdx].saveStringData(  rerRuleText[ i ] +
                                        rerRules[ i ].substring( 0 ) );

      /*
        String str = " \n The total matching times = ";
        int val = clarionAgent[ groupIdx ][ agentIdx ].getMatchCounter();
        str += String.valueOf( val ) + "\n";
        wDataFile.saveStringData( str );

        str = " \n The total chosen times = ";
        val = clarionAgent[ groupIdx ][ agentIdx ].getChosenCounter();
        str += String.valueOf( val ) + "\n";
        wDataFile.saveStringData( str );
        */
     }

     if( ruleFlagArray[ FIX_IDX ]  )
     {
        clarionAgent[ agentIdx ].fillFixRule( fixRules[ 0 ] );
        dataFile[taskIdx][groupIdx].saveStringData( fixRules[ 0 ].substring( 0 ) );
     }

   }

   /************************ END OF SAVE FUNCTIONS ****************************/
     public void showRules()
     {
       String str = new String("");

       if( ruleFlagArray[ IRL_IDX ]  )
       {
          clarionAgent[ agentIdx ].fillIrlRule( irlRules );
          for( int i = 0; i < 3; i++ )
               str += irlRuleText[ i ] + irlRules[ i ].substring( 0 );
       }

       if( ruleFlagArray[ RER_IDX ]  )
       {
          clarionAgent[ agentIdx ].fillRerRule( rerRules );
          for( int i = 0; i < 3; i++ )
                str += rerRuleText[ i ] +  rerRules[ i ].substring( 0 );
       }
     }
  }