package tasksimulators.oldsimu;
import java.util.Vector;
import clarion.nacs.*;
import clarion.common.*;
import clarion.tools.*;

public class WorkingMemory {

  int WM_NET_INFO[], DIM_VAL_INFO[];
  int TOP_LEVEL = 0, DOWN_LEVEL = 1, levelFlag;

  int size_WM, DIM_NUM, DIM_VAL_NUM;
  int action, preState[], curState[], wmState[];
  long currentTime;

  double wmQValue1, wmQValue2, feedback;
  double threshold, threshold_WM, qbpArray[];

  boolean flags[];

  GKChunk items[];
  QBackPropNet wmNet;
  StochasticDecider actionDecider;

  public WorkingMemory( int size )
  {
      size_WM = size;
  }

  public WorkingMemory( int size, int dimNum, int wmStateLen,
                        int[] netInfoArr, int[] dimInfoArr )
  {
     size_WM = size;
     DIM_NUM = dimNum;

     wmState = new int[ wmStateLen ];
     preState = new int[ netInfoArr[1] ];
     curState = new int[ netInfoArr[1] ];

     WM_NET_INFO = new int[ netInfoArr.length ];
     for( int i = 0; i < netInfoArr.length; i++ )
          WM_NET_INFO[ i ] = netInfoArr[ i ];

     DIM_VAL_INFO = new int[ dimInfoArr.length ];
     for( int i = 0; i < netInfoArr.length; i++ )
          DIM_VAL_INFO[ i ] = dimInfoArr[ i ];
     DIM_VAL_NUM = DIM_VAL_INFO[ 0 ];

     items = new GKChunk[ size_WM ];
     initQbpNet();
  }

  public void initQbpNet()
  {
     qbpArray  =  new double[ WM_NET_INFO[3] ];
     wmNet = new QBackPropNet( WM_NET_INFO[1], WM_NET_INFO[2],
                               WM_NET_INFO[3], WM_NET_INFO[0], DIM_VAL_INFO );
     actionDecider = new StochasticDecider( WM_NET_INFO[3],
                           World.TASK_PARA_ARRAY[ World.TEMPERATURE ] );
  }

  public void reinit()
  {
     reinitWM();
     reinitQbpNet();
  }

  public void reinitWM()
  {
     actionResetAll();
  }

  public void reinitQbpNet()
  {
     wmNet.reinit();
  }

  public void setCurrentTime( long time )
  {
     currentTime = time;
  }

/************************ Working Memory actions ******************************/

  public void actionSet( int idx, GKChunk chunk )
  {
      if( items[ idx ] != null )
      {
         if( items[ idx ].equals( chunk ) )
         {
             items[ idx ].increaseUsage( currentTime );
             return;
         }
      }

      try{
             items[ idx ] = (GKChunk)chunk.clone();
             items[ idx ].reinit( 0 );
             items[ idx ].increaseUsage( currentTime );
         }
      catch( CloneNotSupportedException e )
      {
         System.out.println( "clone error!" );
      }
  }

  public void actionReset( int idx )
  {
      items[ idx ].reinit( 0, (Vector)null );
  }

  public void actionNo()
  {
  }

  public void actionResetAll()
  {
      for( int i = 0; i < size_WM; i++ )
           if( items[ i ] != null )
               actionReset( i );
  }

  public double getSuccRate()
  {
     double succRate = wmNet.getSuccRate();
     return succRate;
  }

  public int getAction( int[] inputs )
  {
     wmNet.setCurState( inputs );
     wmNet.setInput();
     wmNet.forwardPass();
     wmNet.getQvals( qbpArray );
     actionDecider.setElements( qbpArray );

     // recommended action from the bottom level
     action = actionDecider.getRecommendation();
     return action;
  }

/************************* NEED TO BE IMPLEMENTED ****************************/
  /* get the WM part of input to the network. */
  public void getWMState( int[] arr )
  {

  }

  public void setLevelFlag( int level )
  {
     levelFlag = level;
  }

  public void setFeedback( double val )
  {
     feedback = val;
  }

  public void setThresholdWM( double th )
  {
     threshold_WM = th;
  }

  public void setPreState( int[] state )
  {
     for( int i = 0; i < state.length; i++ )
          preState[ i ] = state[ i ];
  }

  public void setCurState( int[] state )
  {
     for( int i = 0; i < state.length; i++ )
          curState[ i ] = state[ i ];
  }

  public void setFlags( boolean[] flagArr )
  {
     flags = new boolean[ flagArr.length ];
     for( int i = 0; i < flagArr.length; i++ )
          flags[ i ] = flagArr[ i ];
  }

  public boolean checkAction( int act )
  {
     return false;
  }

  public boolean checkAction( int[] arr )
  {
     return true;
  }

  public void performAction( int[] arr )
  {
  }

  public void updateAction( int chosenAction )
  {
     action = chosenAction;
  }

  public void update( boolean flag )
  {
     if( flag )
         legalUpdate();
     else
         illegalUpdate();
  }

/************************* PROTECTED METHODS **********************************/
  protected void illegalUpdate()
  {
     wmQValue1 = wmNet.getQVal( action );
     if( levelFlag == DOWN_LEVEL )
         wmNet.update( action, feedback, wmQValue1 );
  }

  protected void legalUpdate()
  {
     wmQValue1 = wmNet.getQVal( action );

     //obtain nwmt state QValue2.
     wmNet.setCurState( curState );
     wmNet.setInput();
     wmNet.forwardPass();
     wmQValue2 = wmNet.getMaxQVal();

/******************************************************************************/
     // update the QBP network.
     wmNet.update( action, feedback, wmQValue1, wmQValue2 );
/******************************************************************************/

     // update RER rule set if the RER flag is set.
     double r = wmNet.getReinforcement();
     double d = wmNet.getDiscount();

     if( r + d * wmQValue2 - wmQValue1 > threshold )
         wmNet.updateMatchCount(  true );
     else
         wmNet.updateMatchCount( false );
  }
}